// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include "CSelection.h"
#include "TImageWindow.h"
#include <math.h>
#include "globals.h"

void	SetModifierKeyDown(XEvent *event) ;
void	SetModifierKeyUp(XEvent *event) ;
void	WriteObjectInfo(char *text) ;

#define log2(x)	log(x)/log(2.0)
extern long near(float) ;

extern	GrafPtr	gCurrentPort ;
#define	HSIZE	3
CSelection::CSelection(Point &pt,TImageWindow *window)
{
	// Point coming in is in image coordinates
	mSignature = 'isel' ;
	mIsSelected = true ;

	SetRect(&mSelectionRect,pt.x,pt.y,pt.x,pt.y) ;
	SetActiveHandles() ;
	
	mWindow = window ;
	Modify(pt) ;
}

CSelection::~CSelection() {
} 

void CSelection::Modify(Point &loc)
{
	Point where = loc ;
	mWindow->ImageToWindowPoint(where) ;
	mWindow->ImageToWindowRect(mTopLeftCorner) ;
	mWindow->ImageToWindowRect(mTopRightCorner) ;
	mWindow->ImageToWindowRect(mBottomLeftCorner) ;
	mWindow->ImageToWindowRect(mBottomRightCorner) ;
	mWindow->ImageToWindowRect(mTopMiddle) ;
	mWindow->ImageToWindowRect(mLeftMiddle) ;
	mWindow->ImageToWindowRect(mBottomMiddle) ;
	mWindow->ImageToWindowRect(mRightMiddle) ;
	mWindow->ImageToWindowRect(mSelectionRect) ;

	mIsSelected	= false ;
	
	if (PtInRect(&where,&mTopLeftCorner)) {
		DragTopLeft(where) ;
	}
	else if (PtInRect(&where,&mTopRightCorner)) {
		DragTopRight(where) ;
	}
	else if (PtInRect(&where,&mBottomLeftCorner)) {
		DragBottomLeft(where) ;
	}
	else if (PtInRect(&where,&mBottomRightCorner)) {
		DragBottomRight(where) ;
	}
	else if (PtInRect(&where,&mTopMiddle)) {
		DragTop(where) ;
	}
	else if (PtInRect(&where,&mLeftMiddle)) {
		DragLeft(where) ;
	}
	else if (PtInRect(&where,&mBottomMiddle)) {
		DragBottom(where) ;
	}
	else if (PtInRect(&where,&mRightMiddle)) {
		DragRight(where) ;
	}
	else {
		Rect theRect = mSelectionRect ;
		InsetRect(&theRect,-2,-2) ;
		if (PtInRect(&where,&theRect)) {
			Shift(where) ;
		}
	}	
	mWindow->WindowToImageRect(mTopLeftCorner) ;
	mWindow->WindowToImageRect(mTopRightCorner) ;
	mWindow->WindowToImageRect(mBottomLeftCorner) ;
	mWindow->WindowToImageRect(mBottomRightCorner) ;
	mWindow->WindowToImageRect(mTopMiddle) ;
	mWindow->WindowToImageRect(mLeftMiddle) ;
	mWindow->WindowToImageRect(mBottomMiddle) ;
	mWindow->WindowToImageRect(mRightMiddle) ;
	mWindow->WindowToImageRect(mSelectionRect) ;
	
	mIsSelected	= true ;
}

void	CSelection::GetObjectInfo(char *text) {
	Rect objectRect ;
	GetSelectionRect(objectRect) ;
	sprintf(text,"Selection: W = %d, H = %d\nSelection Rect: (%d,%d,%d,%d)",
		objectRect.right-objectRect.left,objectRect.bottom-objectRect.top,
		objectRect.left,objectRect.top,objectRect.right,objectRect.bottom) ;
}

void	CSelection::ShowInfo() {
	Rect objectRect ;
	char objectInfo[256] ;
	GetSelectionRect(objectRect) ;
	sprintf(objectInfo,"Selection: W = %d, H = %d\nSelection Rect: (%d,%d,%d,%d)",
		objectRect.right-objectRect.left,objectRect.bottom-objectRect.top,
		objectRect.left,objectRect.top,objectRect.right,objectRect.bottom) ;
	::WriteObjectInfo(objectInfo) ;
}

void CSelection::DoMouseDown(Point &loc)
{
	Modify(loc) ;
}

void CSelection::Zoom(float factor,short x,short y) {
	short width,height,x0,y0 ;
	
	width = mSelectionRect.right - mSelectionRect.left ;
	height = mSelectionRect.bottom - mSelectionRect.top ;
	width = (int) (width*factor + 0.5) ;
	height = (int) (height*factor + 0.5) ;
	x0 = (mSelectionRect.right + mSelectionRect.left)/2 + x ;
	y0 = (mSelectionRect.bottom + mSelectionRect.top)/2 + y ;
	mSelectionRect.left = x0 - width/2 ;
	mSelectionRect.top = y0 - height/2 ;
	mSelectionRect.right = mSelectionRect.left + width ;
	mSelectionRect.bottom = mSelectionRect.top + height ;
	SetActiveHandles() ;
}

void CSelection::Update(Rect *clipRect) {
	RgnHandle	clipRgn,savedClip ;
	GrafPtr	currentPort ;
	
	GetPort(&currentPort) ;
	savedClip = NewRgn() ;
	clipRgn = NewRgn() ;
	RectRgn(savedClip,&currentPort->GetPortRect()) ;
	InsetRect(clipRect,1,1) ;
	RectRgn(clipRgn,clipRect) ;
	SetClip(clipRgn) ;
	Draw() ;
	SetClip(savedClip) ;
	DisposeRgn(clipRgn) ;
	DisposeRgn(savedClip) ;
	InsetRect(clipRect,-1,-1) ;
}

void CSelection::Pan(short dx,short dy) {
	OffsetRect(&mSelectionRect,dx,dy) ;
	SetActiveHandles() ;
}

bool CSelection::InSelection(Point &pt) {
	if (PtInRect(&pt,&mTopLeftCorner)) {
		return true ;
	}
	else if (PtInRect(&pt,&mTopRightCorner)) {
		return true ;
	}
	else if (PtInRect(&pt,&mBottomLeftCorner)) {
		return true ;
	}
	else if (PtInRect(&pt,&mBottomRightCorner)) {
		return true ;
	}
	else if (PtInRect(&pt,&mTopMiddle)) {
		return true ;
	}
	else if (PtInRect(&pt,&mLeftMiddle)) {
		return true ;
	}
	else if (PtInRect(&pt,&mBottomMiddle)) {
		return true ;
	}
	else if (PtInRect(&pt,&mRightMiddle)) {
		return true ;
	}
	else {
		Rect	outerRect = mSelectionRect ;
		Rect	innerRect = mSelectionRect ;
		InsetRect(&outerRect,-3,-3) ;
		InsetRect(&innerRect,3,3) ;
		if (PtInRect(&pt,&outerRect) && !PtInRect(&pt,&innerRect)) 
			return true ;
	}	
	return false ;
}

void CSelection::DrawHandles() {
	SetFgColor(gGreenPixel) ;
	FrameRect(&mTopLeftCorner) ;
	FrameRect(&mTopRightCorner) ;
	FrameRect(&mBottomLeftCorner) ;
	FrameRect(&mBottomRightCorner) ;
	FrameRect(&mTopMiddle) ;
	FrameRect(&mLeftMiddle) ;
	FrameRect(&mBottomMiddle) ;
	FrameRect(&mRightMiddle) ;	
	SetFgColor(gBlackPixel) ;		
}
void CSelection::Draw() {
	SetFgColor(gRedPixel) ;
	Rect	drawRect = mSelectionRect ;
	drawRect.right -= 1 ;
	drawRect.bottom -= 1 ;
	FrameRect(&drawRect) ;	
	SetFgColor(gBlackPixel) ;		
	if (mIsSelected) DrawHandles() ;
}

void 	CSelection::Drag(Point &loc) {
	Point lastPoint,currentPoint,startPoint ;
	int top,left,right,bottom,width,height,ww,n ;
	Rect	imageRect ;
	Rect	boundsRect ;
	
	short	imageWidth = mWindow->GetImage()->Width() ;
	short	imageHeight = mWindow->GetImage()->Height() ;
	
	SetRect(&boundsRect,0,0,imageWidth,imageHeight) ;
	mWindow->ImageToWindowRect(boundsRect) ;
	
	currentPoint = loc ; ;
	lastPoint = currentPoint ;
	startPoint = currentPoint ;
	top = lastPoint.y ;
	bottom = lastPoint.y ;
	left = lastPoint.x ;
	right = lastPoint.x ;
	PenMode(kXOrMode);
	SetRect(&mSelectionRect,left,top,right,bottom) ;
	Draw() ;		// draw (actually empty drawing)
	
	
	extern	bool	gMouseIsDown ;
	XEvent	event ;
	XMotionEvent	*motionEvent =  (XMotionEvent *)&event ;
	
	// Grab the pointer
	XGrabPointer(gDisplay,mWindow->GetXWindow(),False,ButtonPressMask | ButtonReleaseMask | PointerMotionMask,
		GrabModeAsync,GrabModeAsync,None,None,CurrentTime) ;
		
	while (gMouseIsDown) {
		XNextEvent(gDisplay,&event) ;		
		switch(event.type) {
			case ButtonRelease:
				gMouseIsDown = false ;
				break ;
			case MotionNotify:
				currentPoint.x = motionEvent->x ;
				currentPoint.y = motionEvent->y ;
				Draw() ;
				left = (currentPoint.x < startPoint.x ? currentPoint.x : startPoint.x) ;
				right = (currentPoint.x > startPoint.x ? currentPoint.x : startPoint.x) ;
				top = (currentPoint.y < startPoint.y ? currentPoint.y : startPoint.y) ;
				bottom = (currentPoint.y > startPoint.y ? currentPoint.y : startPoint.y) ;
				SetRect(&imageRect,left,top,right,bottom) ;
				if (ShiftKeyDown()) {
					width = right - left ;
					height = bottom - top ;
					ww = width > height ? width : height ;
					if (currentPoint.x < startPoint.x) 
						left = right - ww ;
					else
						right = left + ww ;
					if (currentPoint.y < startPoint.y) 
						top = bottom - ww ;
					else
						bottom = top + ww ;
				}
				if (ControlKeyDown()) {
					mWindow->WindowToImageRect(imageRect) ;
					width = imageRect.right - imageRect.left ;
					height = imageRect.bottom - imageRect.top ;
					ww = width > height ? width : height ;
					n = near(log2((double)ww)) ;
					ww = pow(2.,(double)n) ;
					if (currentPoint.x < startPoint.x) {
						if (imageRect.right - ww >= 0)
							imageRect.left = imageRect.right - ww ;
					}
					else {
						if (imageRect.left + ww <= imageWidth)
							imageRect.right = imageRect.left + ww ;
					}
					if (currentPoint.y < startPoint.y) {
						if (imageRect.bottom - ww >= 0)
							imageRect.top = imageRect.bottom - ww ;
					}
					else {
						if (imageRect.top + ww <= imageHeight)
							imageRect.bottom = imageRect.top + ww ;		
					}
					mWindow->ImageToWindowRect(imageRect) ;	
					left = imageRect.left ;
					right = imageRect.right ;
					top = imageRect.top ;
					bottom = imageRect.bottom ;
				}
				if (left < boundsRect.left) left = boundsRect.left ;
				if (right > boundsRect.right) right = boundsRect.right ;
				if (top < boundsRect.top) top = boundsRect.top ;
				if (bottom > boundsRect.bottom) bottom = boundsRect.bottom ;
				SetRect(&mSelectionRect,left,top,right,bottom) ;
				Draw() ;
				lastPoint = currentPoint ;
				break ;
			default:
				break ;
		}
	}
	XUngrabPointer(gDisplay,CurrentTime) ;
	SetActiveHandles() ;
	PenNormal() ;
}

void 	CSelection::SetActiveHandles() {
	SetRect(&mTopLeftCorner,mSelectionRect.left-HSIZE,mSelectionRect.top-HSIZE,mSelectionRect.left+HSIZE,mSelectionRect.top+HSIZE) ;
	SetRect(&mTopRightCorner,mSelectionRect.right-HSIZE,mSelectionRect.top-HSIZE,mSelectionRect.right+HSIZE,mSelectionRect.top+HSIZE) ;
	SetRect(&mBottomLeftCorner,mSelectionRect.left-HSIZE,mSelectionRect.bottom-HSIZE,mSelectionRect.left+HSIZE,mSelectionRect.bottom+HSIZE) ;
	SetRect(&mBottomRightCorner,mSelectionRect.right-HSIZE,mSelectionRect.bottom-HSIZE,mSelectionRect.right+HSIZE,mSelectionRect.bottom+HSIZE) ;
	SetRect(&mTopMiddle,(mSelectionRect.left+mSelectionRect.right)/2-HSIZE,mSelectionRect.top-HSIZE,
			(mSelectionRect.left+mSelectionRect.right)/2+HSIZE,mSelectionRect.top+HSIZE) ;
	SetRect(&mLeftMiddle,mSelectionRect.left-HSIZE,(mSelectionRect.top+mSelectionRect.bottom)/2-HSIZE,
			mSelectionRect.left+HSIZE,(mSelectionRect.top+mSelectionRect.bottom)/2+HSIZE) ;
	SetRect(&mBottomMiddle,(mSelectionRect.left+mSelectionRect.right)/2-HSIZE,mSelectionRect.bottom-HSIZE,
			(mSelectionRect.left+mSelectionRect.right)/2+HSIZE,mSelectionRect.bottom+HSIZE) ;
	SetRect(&mRightMiddle,mSelectionRect.right-HSIZE,(mSelectionRect.top+mSelectionRect.bottom)/2-HSIZE,
			mSelectionRect.right+HSIZE,(mSelectionRect.top+mSelectionRect.bottom)/2+HSIZE) ;
}

void 	CSelection::DragTopLeft(Point &loc) {
	Point lastPoint,currentPoint,startPoint,topLeft,bottomRight ;
	int top,left,right,bottom,width,height,ww,n ;
	Rect	imageRect ;
	Rect	boundsRect ;
	
	short	imageWidth = mWindow->GetImage()->Width() ;
	short	imageHeight = mWindow->GetImage()->Height() ;
	
	SetRect(&boundsRect,0,0,imageWidth,imageHeight) ;
	mWindow->ImageToWindowRect(boundsRect) ;
	
	startPoint = loc ;
	lastPoint = startPoint ;
	top = mSelectionRect.top ;
	bottom = mSelectionRect.bottom ;
	left = mSelectionRect.left ;
	right = mSelectionRect.right ;
	topLeft.x = left ;
	topLeft.y = top ;
	bottomRight.x = right ;
	bottomRight.y = bottom ;
	PenMode(kXOrMode);
	Draw() ;		// to draw without handles
	
	extern	bool	gMouseIsDown ;
	XEvent	event ;
	XMotionEvent	*motionEvent =  (XMotionEvent *)&event ;
		
	// Grab the pointer
	XGrabPointer(gDisplay,mWindow->GetXWindow(),False,ButtonPressMask | ButtonReleaseMask | PointerMotionMask,
		GrabModeAsync,GrabModeAsync,None,None,CurrentTime) ;
		
	while (gMouseIsDown) {
		XNextEvent(gDisplay,&event) ;		
		switch(event.type) {
			case ButtonRelease:
				gMouseIsDown = false ;
				break ;
			case KeyPress:
				SetModifierKeyDown(&event) ;	
				break ;
			case KeyRelease:
				SetModifierKeyUp(&event) ;
				break ;
			case MotionNotify:
				currentPoint.x = motionEvent->x ;
				currentPoint.y = motionEvent->y ;
				Draw() ;
				topLeft.x += currentPoint.x - lastPoint.x ;
				topLeft.y += currentPoint.y - lastPoint.y ;
				left = (topLeft.x < bottomRight.x ? topLeft.x : bottomRight.x) ;
				right = (topLeft.x > bottomRight.x ? topLeft.x : bottomRight.x) ;
				top = (topLeft.y < bottomRight.y ? topLeft.y : bottomRight.y) ;
				bottom = (topLeft.y > bottomRight.y ? topLeft.y : bottomRight.y) ;
				SetRect(&imageRect,left,top,right,bottom) ;
				if (ShiftKeyDown()) {
					width = right - left ;
					height = bottom - top ;
					ww = width > height ? width : height ;
					if (currentPoint.x < startPoint.x) 
						left = right - ww ;
					else
						right = left + ww ;
					if (currentPoint.y < startPoint.y) 
						top = bottom - ww ;
					else
						bottom = top + ww ;
				}
				if (ControlKeyDown()) {
					mWindow->WindowToImageRect(imageRect) ;
					width = imageRect.right - imageRect.left ;
					height = imageRect.bottom - imageRect.top ;
					ww = width > height ? width : height ;
					n = near(log2((float)ww)) ;
					ww = pow(2.,(double)n) ;
					if (currentPoint.x < startPoint.x) {
						if (imageRect.right - ww >= 0)
							imageRect.left = imageRect.right - ww ;
					}
					else {
						if (imageRect.left + ww <= imageWidth)
							imageRect.right = imageRect.left + ww ;
					}
					if (currentPoint.y < startPoint.y) {
						if (imageRect.bottom - ww >= 0)
							imageRect.top = imageRect.bottom - ww ;
					}
					else {
						if (imageRect.top + ww <= imageHeight)
							imageRect.bottom = imageRect.top + ww ;		
					}
					mWindow->ImageToWindowRect(imageRect) ;	
					left = imageRect.left ;
					right = imageRect.right ;
					top = imageRect.top ;
					bottom = imageRect.bottom ;
				}
				if (left < boundsRect.left) left = boundsRect.left ;
				if (right > boundsRect.right) right = boundsRect.right ;
				if (top < boundsRect.top) top = boundsRect.top ;
				if (bottom > boundsRect.bottom) bottom = boundsRect.bottom ;
				SetRect(&mSelectionRect,left,top,right,bottom) ;
				ShowInfo() ;
				Draw() ;
				lastPoint = currentPoint ;
				break ;
			default:
				break ;
		}
	}
	XUngrabPointer(gDisplay,CurrentTime) ;
	SetActiveHandles() ;
	PenNormal() ;
	
}
void 	CSelection::DragTopRight(Point &loc) {
	Point lastPoint,currentPoint,startPoint,topLeft,bottomRight ;
	int top,left,right,bottom,width,height,ww,n ;
	Rect	imageRect ;
	Rect	boundsRect ;
	
	short	imageWidth = mWindow->GetImage()->Width() ;
	short	imageHeight = mWindow->GetImage()->Height() ;
	
	SetRect(&boundsRect,0,0,imageWidth,imageHeight) ;
	mWindow->ImageToWindowRect(boundsRect) ;
	
	startPoint = loc ;
	lastPoint = startPoint ;
	top = mSelectionRect.top ;
	bottom = mSelectionRect.bottom ;
	left = mSelectionRect.left ;
	right = mSelectionRect.right ;
	topLeft.x = left ;
	topLeft.y = top ;
	bottomRight.x = right ;
	bottomRight.y = bottom ;
	PenMode(kXOrMode);
	Draw() ;		// to draw without handles


	extern	bool	gMouseIsDown ;
	XEvent	event ;
	XMotionEvent	*motionEvent =  (XMotionEvent *)&event ;
	
	// Grab the pointer
	XGrabPointer(gDisplay,mWindow->GetXWindow(),False,ButtonPressMask | ButtonReleaseMask | PointerMotionMask,
		GrabModeAsync,GrabModeAsync,None,None,CurrentTime) ;
		
	while (gMouseIsDown) {
		XNextEvent(gDisplay,&event) ;		
		switch(event.type) {
			case ButtonRelease:
				gMouseIsDown = false ;
				break ;
			case KeyPress:
				SetModifierKeyDown(&event) ;	
				break ;
			case KeyRelease:
				SetModifierKeyUp(&event) ;
				break ;
			case MotionNotify:
				currentPoint.x = motionEvent->x ;
				currentPoint.y = motionEvent->y ;
				Draw() ;
				bottomRight.x += currentPoint.x - lastPoint.x ;
				topLeft.y += currentPoint.y - lastPoint.y ;
				left = (topLeft.x < bottomRight.x ? topLeft.x : bottomRight.x) ;
				right = (topLeft.x > bottomRight.x ? topLeft.x : bottomRight.x) ;
				top = (topLeft.y < bottomRight.y ? topLeft.y : bottomRight.y) ;
				bottom = (topLeft.y > bottomRight.y ? topLeft.y : bottomRight.y) ;
				SetRect(&imageRect,left,top,right,bottom) ;
				if (ShiftKeyDown()) {
					width = right - left ;
					height = bottom - top ;
					ww = width > height ? width : height ;
					if (currentPoint.x < startPoint.x) 
						left = right - ww ;
					else
						right = left + ww ;
					if (currentPoint.y < startPoint.y) 
						top = bottom - ww ;
					else
						bottom = top + ww ;
				}
				if (ControlKeyDown()) {
					mWindow->WindowToImageRect(imageRect) ;
					width = imageRect.right - imageRect.left ;
					height = imageRect.bottom - imageRect.top ;
					ww = width > height ? width : height ;
					n = near(log2((float)ww)) ;
					ww = pow(2.,(double)n) ;
					if (currentPoint.x < startPoint.x) {
						if (imageRect.right - ww >= 0)
							imageRect.left = imageRect.right - ww ;
					}
					else {
						if (imageRect.left + ww <= imageWidth)
							imageRect.right = imageRect.left + ww ;
					}
					if (currentPoint.y < startPoint.y) {
						if (imageRect.bottom - ww >= 0)
							imageRect.top = imageRect.bottom - ww ;
					}
					else {
						if (imageRect.top + ww <= imageHeight)
							imageRect.bottom = imageRect.top + ww ;		
					}
					mWindow->ImageToWindowRect(imageRect) ;	
					left = imageRect.left ;
					right = imageRect.right ;
					top = imageRect.top ;
					bottom = imageRect.bottom ;
				}
				if (left < boundsRect.left) left = boundsRect.left ;
				if (right > boundsRect.right) right = boundsRect.right ;
				if (top < boundsRect.top) top = boundsRect.top ;
				if (bottom > boundsRect.bottom) bottom = boundsRect.bottom ;
				SetRect(&mSelectionRect,left,top,right,bottom) ;
				ShowInfo() ;
				Draw() ;
				lastPoint = currentPoint ;
				break ;
			default:
				break ;
		}
	}
	XUngrabPointer(gDisplay,CurrentTime) ;
	SetActiveHandles() ;
	PenNormal() ;
}
void 	CSelection::DragBottomLeft(Point &loc) {
	Point lastPoint,currentPoint,startPoint,topLeft,bottomRight ;
	int top,left,right,bottom,width,height,ww,n ;
	Rect	imageRect ;
	Rect	boundsRect ;
	
	short	imageWidth = mWindow->GetImage()->Width() ;
	short	imageHeight = mWindow->GetImage()->Height() ;
	
	SetRect(&boundsRect,0,0,imageWidth,imageHeight) ;
	mWindow->ImageToWindowRect(boundsRect) ;
	
	startPoint = loc ;
	lastPoint = startPoint ;
	top = mSelectionRect.top ;
	bottom = mSelectionRect.bottom ;
	left = mSelectionRect.left ;
	right = mSelectionRect.right ;
	topLeft.x = left ;
	topLeft.y = top ;
	bottomRight.x = right ;
	bottomRight.y = bottom ;
	PenMode(kXOrMode);
	Draw() ;		// to draw without handles

	extern	bool	gMouseIsDown ;
	XEvent	event ;
	XMotionEvent	*motionEvent =  (XMotionEvent *)&event ;
	
	// Grab the pointer
	XGrabPointer(gDisplay,mWindow->GetXWindow(),False,ButtonPressMask | ButtonReleaseMask | PointerMotionMask,
		GrabModeAsync,GrabModeAsync,None,None,CurrentTime) ;
		
	while (gMouseIsDown) {
		XNextEvent(gDisplay,&event) ;		
		switch(event.type) {
			case ButtonRelease:
				gMouseIsDown = false ;
				break ;
			case KeyPress:
				SetModifierKeyDown(&event) ;	
				break ;
			case KeyRelease:
				SetModifierKeyUp(&event) ;
				break ;
			case MotionNotify:
				currentPoint.x = motionEvent->x ;
				currentPoint.y = motionEvent->y ;
				Draw() ;
				topLeft.x += currentPoint.x - lastPoint.x ;
				bottomRight.y += currentPoint.y - lastPoint.y ;
				left = (topLeft.x < bottomRight.x ? topLeft.x : bottomRight.x) ;
				right = (topLeft.x > bottomRight.x ? topLeft.x : bottomRight.x) ;
				top = (topLeft.y < bottomRight.y ? topLeft.y : bottomRight.y) ;
				bottom = (topLeft.y > bottomRight.y ? topLeft.y : bottomRight.y) ;
				SetRect(&imageRect,left,top,right,bottom) ;
				if (ShiftKeyDown()) {
					width = right - left ;
					height = bottom - top ;
					ww = width > height ? width : height ;
					if (currentPoint.x < startPoint.x) 
						left = right - ww ;
					else
						right = left + ww ;
					if (currentPoint.y < startPoint.y) 
						top = bottom - ww ;
					else
						bottom = top + ww ;
				}
				if (ControlKeyDown()) {
					mWindow->WindowToImageRect(imageRect) ;
					width = imageRect.right - imageRect.left ;
					height = imageRect.bottom - imageRect.top ;
					ww = width > height ? width : height ;
					n = near(log2((float)ww)) ;
					ww = pow(2.,(double)n) ;
					if (currentPoint.x < startPoint.x) {
						if (imageRect.right - ww >= 0)
							imageRect.left = imageRect.right - ww ;
					}
					else {
						if (imageRect.left + ww <= imageWidth)
							imageRect.right = imageRect.left + ww ;
					}
					if (currentPoint.y < startPoint.y) {
						if (imageRect.bottom - ww >= 0)
							imageRect.top = imageRect.bottom - ww ;
					}
					else {
						if (imageRect.top + ww <= imageHeight)
							imageRect.bottom = imageRect.top + ww ;		
					}
					mWindow->ImageToWindowRect(imageRect) ;	
					left = imageRect.left ;
					right = imageRect.right ;
					top = imageRect.top ;
					bottom = imageRect.bottom ;
				}
				if (left < boundsRect.left) left = boundsRect.left ;
				if (right > boundsRect.right) right = boundsRect.right ;
				if (top < boundsRect.top) top = boundsRect.top ;
				if (bottom > boundsRect.bottom) bottom = boundsRect.bottom ;
				SetRect(&mSelectionRect,left,top,right,bottom) ;
				ShowInfo() ;
				Draw() ;
				lastPoint = currentPoint ;
				break ;
			default:
				break ;
		}
	}
	XUngrabPointer(gDisplay,CurrentTime) ;

	SetActiveHandles() ;
	PenNormal() ;
}
void 	CSelection::DragBottomRight(Point &loc) {
	Point lastPoint,currentPoint,startPoint,topLeft,bottomRight ;
	int top,left,right,bottom,width,height,ww,n ;
	Rect	imageRect ;
	Rect	boundsRect ;
	
	short	imageWidth = mWindow->GetImage()->Width() ;
	short	imageHeight = mWindow->GetImage()->Height() ;
	
	SetRect(&boundsRect,0,0,imageWidth,imageHeight) ;
	mWindow->ImageToWindowRect(boundsRect) ;
	
	startPoint = loc ;
	lastPoint = startPoint ;
	top = mSelectionRect.top ;
	bottom = mSelectionRect.bottom ;
	left = mSelectionRect.left ;
	right = mSelectionRect.right ;
	topLeft.x = left ;
	topLeft.y = top ;
	bottomRight.x = right ;
	bottomRight.y = bottom ;
	PenMode(kXOrMode);
	Draw() ;		// to draw without handles

	extern	bool	gMouseIsDown ;
	XEvent	event ;
	XMotionEvent	*motionEvent =  (XMotionEvent *)&event ;
	
	// Grab the pointer
	XGrabPointer(gDisplay,mWindow->GetXWindow(),False,ButtonPressMask | ButtonReleaseMask | PointerMotionMask,
		GrabModeAsync,GrabModeAsync,None,None,CurrentTime) ;
		
	while (gMouseIsDown) {
		XNextEvent(gDisplay,&event) ;		
		switch(event.type) {
			case ButtonRelease:
				gMouseIsDown = false ;
				break ;
			case KeyPress:
				SetModifierKeyDown(&event) ;	
				break ;
			case KeyRelease:
				SetModifierKeyUp(&event) ;
				break ;
			case MotionNotify:
				currentPoint.x = motionEvent->x ;
				currentPoint.y = motionEvent->y ;
				Draw() ;
				bottomRight.x += currentPoint.x - lastPoint.x ;
				bottomRight.y += currentPoint.y - lastPoint.y ;
				left = (topLeft.x < bottomRight.x ? topLeft.x : bottomRight.x) ;
				right = (topLeft.x > bottomRight.x ? topLeft.x : bottomRight.x) ;
				top = (topLeft.y < bottomRight.y ? topLeft.y : bottomRight.y) ;
				bottom = (topLeft.y > bottomRight.y ? topLeft.y : bottomRight.y) ;
				SetRect(&imageRect,left,top,right,bottom) ;
				if (ShiftKeyDown()) {
					width = right - left ;
					height = bottom - top ;
					ww = width > height ? width : height ;
					if (currentPoint.x < startPoint.x) 
						left = right - ww ;
					else
						right = left + ww ;
					if (currentPoint.y < startPoint.y) 
						top = bottom - ww ;
					else
						bottom = top + ww ;
				}
				if (ControlKeyDown()) {
					mWindow->WindowToImageRect(imageRect) ;
					width = imageRect.right - imageRect.left ;
					height = imageRect.bottom - imageRect.top ;
					ww = width > height ? width : height ;
					n = near(log2((float)ww)) ;
					ww = pow(2.,(double)n) ;
					if (currentPoint.x < startPoint.x) {
						if (imageRect.right - ww >= 0)
							imageRect.left = imageRect.right - ww ;
					}
					else {
						if (imageRect.left + ww <= imageWidth)
							imageRect.right = imageRect.left + ww ;
					}
					if (currentPoint.y < startPoint.y) {
						if (imageRect.bottom - ww >= 0)
							imageRect.top = imageRect.bottom - ww ;
					}
					else {
						if (imageRect.top + ww <= imageHeight)
							imageRect.bottom = imageRect.top + ww ;		
					}
					mWindow->ImageToWindowRect(imageRect) ;	
					left = imageRect.left ;
					right = imageRect.right ;
					top = imageRect.top ;
					bottom = imageRect.bottom ;
				}
				if (left < boundsRect.left) left = boundsRect.left ;
				if (right > boundsRect.right) right = boundsRect.right ;
				if (top < boundsRect.top) top = boundsRect.top ;
				if (bottom > boundsRect.bottom) bottom = boundsRect.bottom ;
				SetRect(&mSelectionRect,left,top,right,bottom) ;
				ShowInfo() ;
				Draw() ;
				lastPoint = currentPoint ;
				break ;
			default:
				break ;
		}
	}
	XUngrabPointer(gDisplay,CurrentTime) ;

	SetActiveHandles() ;
	PenNormal() ;
}
void 	CSelection::DragTop(Point &loc) {
	Point lastPoint,currentPoint,startPoint,topLeft,bottomRight ;
	int top,left,right,bottom ;
	Rect	boundsRect ;
	
	SetRect(&boundsRect,0,0,mWindow->GetImage()->Width(),mWindow->GetImage()->Height()) ;
	mWindow->ImageToWindowRect(boundsRect) ;
	
	startPoint = loc ;
	lastPoint = startPoint ;
	top = mSelectionRect.top ;
	bottom = mSelectionRect.bottom ;
	left = mSelectionRect.left ;
	right = mSelectionRect.right ;
	topLeft.x = left ;
	topLeft.y = top ;
	bottomRight.x = right ;
	bottomRight.y = bottom ;
	PenMode(kXOrMode);
	Draw() ;		// to draw without handles

	extern	bool	gMouseIsDown ;
	XEvent	event ;
	XMotionEvent	*motionEvent =  (XMotionEvent *)&event ;
	
	// Grab the pointer
	XGrabPointer(gDisplay,mWindow->GetXWindow(),False,ButtonPressMask | ButtonReleaseMask | PointerMotionMask,
		GrabModeAsync,GrabModeAsync,None,None,CurrentTime) ;
		
	while (gMouseIsDown) {
		XNextEvent(gDisplay,&event) ;		
		switch(event.type) {
			case ButtonRelease:
				gMouseIsDown = false ;
				break ;
			case KeyPress:
				SetModifierKeyDown(&event) ;	
				break ;
			case KeyRelease:
				SetModifierKeyUp(&event) ;
				break ;
			case MotionNotify:
				currentPoint.x = motionEvent->x ;
				currentPoint.y = motionEvent->y ;
				Draw() ;
				topLeft.y += currentPoint.y - lastPoint.y ;
				top = (topLeft.y < bottomRight.y ? topLeft.y : bottomRight.y) ;
				bottom = (topLeft.y > bottomRight.y ? topLeft.y : bottomRight.y) ;
				if (top < boundsRect.top) top = boundsRect.top ;
				if (bottom > boundsRect.bottom) bottom = boundsRect.bottom ;
				SetRect(&mSelectionRect,left,top,right,bottom) ;
				ShowInfo() ;
				Draw() ;
				lastPoint = currentPoint ;
				break ;
			default:
				break ;
		}
	}
	XUngrabPointer(gDisplay,CurrentTime) ;
	SetActiveHandles() ;
	PenNormal() ;
}
void 	CSelection::DragLeft(Point &loc) {
	Point lastPoint,currentPoint,startPoint,topLeft,bottomRight ;
	int top,left,right,bottom ;
	Rect	boundsRect ;
	
	SetRect(&boundsRect,0,0,mWindow->GetImage()->Width(),mWindow->GetImage()->Height()) ;
	mWindow->ImageToWindowRect(boundsRect) ;
	
	startPoint = loc ;
	lastPoint = startPoint ;
	top = mSelectionRect.top ;
	bottom = mSelectionRect.bottom ;
	left = mSelectionRect.left ;
	right = mSelectionRect.right ;
	topLeft.x = left ;
	topLeft.y = top ;
	bottomRight.x = right ;
	bottomRight.y = bottom ;
	PenMode(kXOrMode);
	Draw() ;		// to draw again without handles

	extern	bool	gMouseIsDown ;
	XEvent	event ;
	XMotionEvent	*motionEvent =  (XMotionEvent *)&event ;
	
	// Grab the pointer
	XGrabPointer(gDisplay,mWindow->GetXWindow(),False,ButtonPressMask | ButtonReleaseMask | PointerMotionMask,
		GrabModeAsync,GrabModeAsync,None,None,CurrentTime) ;
		
	while (gMouseIsDown) {
		XNextEvent(gDisplay,&event) ;		
		switch(event.type) {
			case ButtonRelease:
				gMouseIsDown = false ;
				break ;
			case KeyPress:
				SetModifierKeyDown(&event) ;	
				break ;
			case KeyRelease:
				SetModifierKeyUp(&event) ;
				break ;
			case MotionNotify:
				currentPoint.x = motionEvent->x ;
				currentPoint.y = motionEvent->y ;
				Draw() ;
				topLeft.x += currentPoint.x - lastPoint.x ;
				left = (topLeft.x < bottomRight.x ? topLeft.x : bottomRight.x) ;
				right = (topLeft.x > bottomRight.x ? topLeft.x : bottomRight.x) ;
				if (left < boundsRect.left) left = boundsRect.left ;
				if (right > boundsRect.right) right = boundsRect.right ;
				SetRect(&mSelectionRect,left,top,right,bottom) ;
				ShowInfo() ;
				Draw() ;
				lastPoint = currentPoint ;
				break ;
			default:
				break ;
		}
	}
	XUngrabPointer(gDisplay,CurrentTime) ;
	SetActiveHandles() ;
	PenNormal() ;
}
void 	CSelection::DragBottom(Point &loc) {
	Point lastPoint,currentPoint,startPoint,topLeft,bottomRight ;
	int top,left,right,bottom ;
	Rect	boundsRect ;
	
	SetRect(&boundsRect,0,0,mWindow->GetImage()->Width(),mWindow->GetImage()->Height()) ;
	mWindow->ImageToWindowRect(boundsRect) ;
	
	startPoint = loc ;
	lastPoint = startPoint ;
	top = mSelectionRect.top ;
	bottom = mSelectionRect.bottom ;
	left = mSelectionRect.left ;
	right = mSelectionRect.right ;
	topLeft.x = left ;
	topLeft.y = top ;
	bottomRight.x = right ;
	bottomRight.y = bottom ;
	PenMode(kXOrMode);
	Draw() ;		// to draw again without handles


	extern	bool	gMouseIsDown ;
	XEvent	event ;
	XMotionEvent	*motionEvent =  (XMotionEvent *)&event ;
	
	// Grab the pointer
	XGrabPointer(gDisplay,mWindow->GetXWindow(),False,ButtonPressMask | ButtonReleaseMask | PointerMotionMask,
		GrabModeAsync,GrabModeAsync,None,None,CurrentTime) ;
		
	while (gMouseIsDown) {
		XNextEvent(gDisplay,&event) ;		
		switch(event.type) {
			case ButtonRelease:
				gMouseIsDown = false ;
				break ;
			case KeyPress:
				SetModifierKeyDown(&event) ;	
				break ;
			case KeyRelease:
				SetModifierKeyUp(&event) ;
				break ;
			case MotionNotify:
				currentPoint.x = motionEvent->x ;
				currentPoint.y = motionEvent->y ;
				Draw() ;
				bottomRight.y += currentPoint.y - lastPoint.y ;
				top = (topLeft.y < bottomRight.y ? topLeft.y : bottomRight.y) ;
				bottom = (topLeft.y > bottomRight.y ? topLeft.y : bottomRight.y) ;
				if (top < boundsRect.top) top = boundsRect.top ;
				if (bottom > boundsRect.bottom) bottom = boundsRect.bottom ;
				SetRect(&mSelectionRect,left,top,right,bottom) ;
				ShowInfo() ;
				Draw() ;
				lastPoint = currentPoint ;
				break ;
			default:
				break ;
		}
	}
	XUngrabPointer(gDisplay,CurrentTime) ;
	SetActiveHandles() ;
	PenNormal() ;
}
void 	CSelection::DragRight(Point &loc) {
	Point lastPoint,currentPoint,startPoint,topLeft,bottomRight ;
	int top,left,right,bottom ;
	Rect	boundsRect ;
	
	SetRect(&boundsRect,0,0,mWindow->GetImage()->Width(),mWindow->GetImage()->Height()) ;
	mWindow->ImageToWindowRect(boundsRect) ;
	
	startPoint = loc ;
	lastPoint = startPoint ;
	top = mSelectionRect.top ;
	bottom = mSelectionRect.bottom ;
	left = mSelectionRect.left ;
	right = mSelectionRect.right ;
	topLeft.x = left ;
	topLeft.y = top ;
	bottomRight.x = right ;
	bottomRight.y = bottom ;
	PenMode(kXOrMode);
	Draw() ;		// to draw again without handles

	extern	bool	gMouseIsDown ;
	XEvent	event ;
	XMotionEvent	*motionEvent =  (XMotionEvent *)&event ;
	
	// Grab the pointer
	XGrabPointer(gDisplay,mWindow->GetXWindow(),False,ButtonPressMask | ButtonReleaseMask | PointerMotionMask,
		GrabModeAsync,GrabModeAsync,None,None,CurrentTime) ;
		
	while (gMouseIsDown) {
		XNextEvent(gDisplay,&event) ;		
		switch(event.type) {
			case ButtonRelease:
				gMouseIsDown = false ;
				break ;
			case KeyPress:
				SetModifierKeyDown(&event) ;	
				break ;
			case KeyRelease:
				SetModifierKeyUp(&event) ;
				break ;
			case MotionNotify:
				currentPoint.x = motionEvent->x ;
				currentPoint.y = motionEvent->y ;
				Draw() ;
				bottomRight.x += currentPoint.x - lastPoint.x ;
				left = (topLeft.x < bottomRight.x ? topLeft.x : bottomRight.x) ;
				right = (topLeft.x > bottomRight.x ? topLeft.x : bottomRight.x) ;
				if (left < boundsRect.left) left = boundsRect.left ;
				if (right > boundsRect.right) right = boundsRect.right ;
				SetRect(&mSelectionRect,left,top,right,bottom) ;
				ShowInfo() ;
				Draw() ;
				lastPoint = currentPoint ;
				break ;
			default:
				break ;
		}
	}
	XUngrabPointer(gDisplay,CurrentTime) ;
	SetActiveHandles() ;
	PenNormal() ;
}
void 	CSelection::Shift(Point &loc) {
	Point lastPoint,currentPoint,startPoint ;
	int dx,dy ;
	Rect	selectionRect ;
	Rect	boundsRect ;
	int 	rx,ry,wx,wy ;
	unsigned int	keyButt ;
	Window	child,root ;
	
	SetRect(&boundsRect,0,0,mWindow->GetImage()->Width(),mWindow->GetImage()->Height()) ;
	mWindow->ImageToWindowRect(boundsRect) ;
	
	startPoint = loc ;
	lastPoint = startPoint ;
	PenMode(kXOrMode);
	Draw() ;		// to draw again without handles

	extern	bool	gMouseIsDown ;
	XEvent	event ;
	
	// Grab the pointer
	XGrabPointer(gDisplay,mWindow->GetXWindow(),False,ButtonPressMask | ButtonReleaseMask | PointerMotionMask,
		GrabModeAsync,GrabModeAsync,None,None,CurrentTime) ;
		
	while (gMouseIsDown) {
		XNextEvent(gDisplay,&event) ;		
		switch(event.type) {
			case ButtonRelease:
				gMouseIsDown = false ;
				break ;
			case KeyPress:
				SetModifierKeyDown(&event) ;	
				break ;
			case KeyRelease:
				SetModifierKeyUp(&event) ;
				break ;
			case MotionNotify:
				if (XQueryPointer(gDisplay,mWindow->GetXWindow(),
						&root,&child,&rx,&ry,&wx,&wy,&keyButt)) {
					if (child == 0) {
						currentPoint.x = wx ;
						currentPoint.y = wy ;
						selectionRect = mSelectionRect ;
						Draw() ;	// Erase 
						dx = currentPoint.x - lastPoint.x ;
						dy = currentPoint.y - lastPoint.y ;
						// Check to see if we are hitting the edges of the image
						OffsetRect(&selectionRect,dx,dy) ;
						if (RectInRect(&selectionRect,&boundsRect))
							OffsetRect(&mSelectionRect,dx,dy) ;
						Draw() ;	// Draw
						lastPoint = currentPoint ;
					}
				}
				break ;
			default:
				break ;
		}
	}
	XUngrabPointer(gDisplay,CurrentTime) ;
	SetActiveHandles() ;
	PenNormal() ;
}

