// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include "types.h"
#include "TVector.h"
// Find the initial estimate for the lattice vectors U and V and the origin O
// of the 2D diffraction pattern where 3 reflections g1,g2,g3 have been marked
// at positions (x1,y1), (x2,y2) and (x3,y3) with indicies (h1,k1), (h2,k2) and (h3,k3)
// The values are stored in the vectors g1, g2 and g3
// as in g1.a[0] = x1, g1,a[1] = y1, g1.a[2] = h1 and g1.a[4] = k1 


// The equation is
//	 1 0 h1 0 k1 0 		Ox		g1x
//	 0 1 0 h1 0 k1		Oy		g1y
//	 1 0 h2 0 k2 0	*	Ux	=	g2x
//	 0 1 0 h2 0 k2		Uy		g2y
//	 1 0 h3 0 k3 0		Vx		g3x
//	 0 1 0 h3 0 k3		Vy		g3y

#define	kSingularMatrix	0
#define kNoError		1
#define kMemoryError	2

Boolean	FindInitialReciprocalLattice(TVector &g1,TVector &g2,TVector &g3) ;
int SolveLinearEquations(float **Amatrix,int Ndimension,float *BXsolution) ;

Boolean	FindInitialReciprocalLattice(TVector &g1,TVector &g2,TVector &g3) {

	int 	n = 6 ;
	float	**a ;
	float 	*b ;
	int	status ;
	int 	j;
	

	a = new float *[n] ;
	for(j=0;j<n;j++) {
		a[j] = new float[n] ;
	}
	b = new float[n] ;


	a[0][0] = a[1][1] = a[2][0] = a[3][1] = a[4][0] =  a[5][1] = 1. ;
	
	a[0][1] = a[0][3] = a[0][5] = 0. ;	
	a[1][0] = a[1][2] = a[1][4] = 0. ;
	a[2][1] = a[2][3] = a[2][5] = 0. ;
	a[3][0] = a[3][2] = a[3][4] = 0. ;
	a[4][1] = a[4][3] = a[4][5] = 0. ;
	a[5][0] = a[5][2] = a[5][4] = 0. ;
	
	a[0][2] = g1.a[2] ;	// h1
	a[0][4] = g1.a[3] ;	// k1
	a[1][3] = g1.a[2] ;	// h1
	a[1][5] = g1.a[3] ;	// k1

	a[2][2] = g2.a[2] ;	// h2
	a[2][4] = g2.a[3] ;	// k2
	a[3][3] = g2.a[2] ;	// h2
	a[3][5] = g2.a[3] ;	// k2

	a[4][2] = g3.a[2] ;	// h3
	a[4][4] = g3.a[3] ;	// k3
	a[5][3] = g3.a[2] ;	// h3
	a[5][5] = g3.a[3] ;	// k3

	b[0] = g1.a[0] ;	// g1x
	b[1] = g1.a[1] ;	// g1y
	b[2] = g2.a[0] ; 	// g2x
	b[3] = g2.a[1] ;	// g2y
	b[4] = g3.a[0] ; 	// g3x
	b[5] = g3.a[1] ;	// g3y


	status = SolveLinearEquations(a,n,b) ;
	if (status == kNoError) {
		g1.a[0] = b[0] ;
		g1.a[1] = b[1] ;
		g2.a[0] = b[2] ;
		g2.a[1] = b[3] ;
		g3.a[0] = b[4] ;
		g3.a[1] = b[5] ;
	}
	
	for(j=0;j<n;j++) {
		delete[] a[j] ;
	}
	delete[] a ;
	delete[] b  ;

	if (status != kNoError)
		return false ;
	
	return true ;
}

