//This version of MEM assume the wave is complex(1-V(x,y)**2/2,V(x,y)).
// Just deconvolute the V(x,y) and let the real part =1-V(x,y)**2/2
//
// The input file is EXP (experimental image), default (Wiener Filter Image), noise (noise image)
// and CTF (complex function)

#include <math.h>
#include <complex.h>
#include "TImageWindow.h"
#include "TRunMEMDialog.h"
#include "TWindowList.h"

extern	MEMRecord	gMEMRecord ;

#define	min(x,y) (x) < (y) ? (x) : (y)
#define	max(x,y) (x) > (y) ? (x) : (y)

void 	MWDCF(long width, long height, float FLUX, TImage &LAMBDA, TImage &SigmaImage,
			TImage &MEM, TImage &Z, TImage &DOBSImage, float MU, float OMEGA,TImage &DEFImage,TImage &Ctf) ;
void 	MWDCGRAD(long width,long height,TImage &LAMBDA,TImage &DZDLAM,TImage &PDZDLA,
			float MU,TImage &SigmaImage,TImage &DOBSImage,TImage &DPRED,float &GRAD,float &PGRAD) ;
void 	MWDCDEL(long DOINIT,long width,long height,long nconj,TImage &LAMBDA,
			TImage &DZDLAM,TImage &DELLAM,float GRAD,float PGRAD,long &iconj) ;
void 	MWDCLEN(long width,long height,TImage &DELLAM,float &Length) ;
void 	MWDCGDST(long width,long height,TImage &LAMBDA,TImage &DELLAM,float MU,
			TImage &SigmaImage,TImage &DOBSIMage,TImage &DPRED, float &GDS) ;
void 	MWDTAKST(TImage &LAMBDA,TImage &DELLAM,TImage &OUT,float scastep) ;
void 	MWDCMU(long width, long height,float OMEGA,TImage &SigmaImage,TImage &LAMBDA,float &MU) ;
void	MaximumEntropyConvolution(TImageWindow *window) ;
TImage	*GenerateCTFImage(MEMRecord &memRecord) ;
//	Function MWDCF
//  This subroutine is to Calculate the Z and MEM image
//
//	MWDCF(width,height,FLUX,LAMBDA,SigmaImage,MEM,Z, DOBSImage,MU,OMEGA)

void	FlushDisplayBuffer() ;

extern	void	WriteObjectInfo(char *text) ;
inline void	WriteProgress(char *text) { WriteObjectInfo(text) ; FlushDisplayBuffer() ;} ;
void	PutMessage(char *) ;

void MWDCF(long width, long height, float FLUX, TImage &LAMBDA, TImage &SigmaImage,
		TImage &MEM, TImage &Z, TImage &DOBSImage, float MU, float OMEGA,TImage &DEFImage,TImage &ImagCTF) {
		
	float NORM ;
	float	veryLarge,notQuiteAsLarge ;
	
	TImage Maxdr(width,height) ;
	TImage LAMTRM(width,height) ;

	// Create image of large value
	veryLarge = pow(10.,20.) ;
	notQuiteAsLarge = log10(veryLarge) ;
	Maxdr = veryLarge ;
	LAMTRM = notQuiteAsLarge ;
	
	// Convolve LAMBDA with imaginary(CTF)
	//
	// MEM = IFFT ( FFT(LAMBDA) * ImagCTF)
	TImage	temp1(LAMBDA) ;
	temp1.fft() ;		
	MEM = temp1 ;
	MEM *= ImagCTF ;		// MEM = temp1*ImagCTF ;	
	MEM.ifft() ;
	
	//	This emulates 
	//	MEM=DEFImage*Exp(Min(-MEM,LAMTRM))
	float	*p1,*p2,*p3 ;	
	p1 = MEM.GetDataArray()->GetData() ;
	p2 = DEFImage.GetDataArray()->GetData() ;
	p3 = LAMTRM.GetDataArray()->GetData() ;	
	NORM = 0. ;
	for (long i = 0; i < MEM.GetDataArray()->Length() ; i++) {
		NORM += *p1 = *p2*exp(min(-*p1,*p3)) ;
		p1++ ;
		p2++ ;
		p3++ ;
	}

	if(FLUX > 0.0) {
		MEM	*=	FLUX/NORM ;
		Z	=	-FLUX*log(NORM) ;
	}
	else {
		Z	=	FLUX*NORM ;
	}

	Z	-=	LAMBDA*DOBSImage ;

	if(MU > 0.0) {
		Z -= MU*OMEGA ;
		Z -= (LAMBDA*SigmaImage)^2/(4*MU) ;
	}
}

// Function MWDCGRAD(width,height,LAMBDA,DZDLAM,PDZDLA,MU,SigmaImage,DOBSImage,DPRED,GRAD,PGRAD)
void MWDCGRAD(long width,long height,TImage &LAMBDA,TImage &DZDLAM,TImage &PDZDLA,
		float MU,TImage &SigmaImage,TImage &DOBSImage,TImage &DPRED,float &GRAD,float &PGRAD) {

	PGRAD 	= GRAD ;
	PDZDLA = DZDLAM ;
	DZDLAM = DPRED - DOBSImage ;
	if(MU > 0.0) {
		PDZDLA	= DZDLAM ;
		DZDLAM	-= LAMBDA*(SigmaImage^2)/(2.0*MU) ;
	}

	GRAD	= DZDLAM.VectorLength() ; 	// sqrt(x-squared)
	GRAD	= GRAD*GRAD ;				// DZDLAM-squared 
	
	GRAD /= (width*height) ;
}

// Function MWDCDEL(DOINIT,width,height,nconj,LAMBDA,DZDLAM,DELLAM,GRAD,PGRAD)
void MWDCDEL(long doinit,long width,long height,long nconj,TImage &LAMBDA,
		TImage &DZDLAM,TImage &DELLAM,float GRAD,float PGRAD,long &iconj) {

	float gamma ;
	if(doinit==1) {
		iconj=0 ;
		gamma=0.0 ;
	}
	else if (iconj >= nconj) {
		iconj=0 ;
		gamma=0.0 ;
	}
	else if(PGRAD!=0.0) {  
		iconj=iconj+1 ;
		gamma=(GRAD/PGRAD) ;
		if(fabs(gamma)>1000.0) {
			iconj=0 ;
			gamma=0.0 ;
		}
	}
	else {
		iconj=0 ;
		gamma=0.0 ;
	}

	DELLAM=DZDLAM+gamma*DELLAM ;

}

//	Function MWDCLEN(width,height,DELLAM,LENGTH)
void MWDCLEN(long width,long height,TImage &DELLAM,float &Length) {
	
	Length = DELLAM.VectorLength()/(width*height) ;	// sqrt(x-squared)
}

//	Function MWDCGDST(width,height,LAMBDA,DELLAM,MU,SigmaImage,DOBSIMage,DPRED, GDS0)
void MWDCGDST(long width,long height,TImage &LAMBDA,TImage &DELLAM,float MU,
		TImage &SigmaImage,TImage &DOBSimage,TImage &DPRED, float &GDS) {
		
	TImage	temp(DELLAM) ;
	
	temp = DELLAM*(DPRED-DOBSimage) ;
		
	//GDS	= -Sum(temp) ;
	if (MU > 0.0) {
		temp = DELLAM*LAMBDA*(SigmaImage^2)/(2.0*MU) ;
		// GDS -= Sum(temp) ; PATCH HERE NEEDED
	}
	GDS /= width*height ;
}


//	Function MWDTAKST(LAMBDA,DELLAM,LAMBDA,scalestep)
void MWDTAKST(TImage &LAMBDA,TImage &DELLAM,TImage &OUT,float scastep) {
	OUT=LAMBDA+scastep*DELLAM ;
}

// FUNCTION MWDCMU(width,height,OMEGA,SigmaImage,LAMBDA,MU)
void MWDCMU(long width, long height,float OMEGA,TImage &SigmaImage,TImage &LAMBDA,float &MU) {
	float DZDMU ;
	DZDMU=0.0 ;
	TImage	temp(LAMBDA) ;
	temp *= SigmaImage ;
	temp ^= 2 ;
	// DZDMU = Sum(temp) ; PATCH NEEDED HERE
	MU=sqrt(DZDMU/(4.0*OMEGA)) ;
	MU=MU/sqrt( (float) width*height) ; //LDM: use float here ?
}


//	Function MWDONEIT
//  This is the main subroutine to solve the deconvolution
// 
void MWDONEIT(long doinit, long width, long height,long &nconj, float &FLUX,
		TImage &DOBSImage,TImage &SigmaImage, float &IOMEGA, float &TOL,TImage &MEM, TImage &Z,TImage &DPRED,
		float &GRAD,float &PGRAD, TImage &OUT,TImage &LAMBDA,TImage &DZDLAM,TImage &PDZDLA,TImage &DELLAM,TImage &DEFImage,
		long& iconj, float &scale,float& MU,TImage &ImagCTF) {
		
	float OMEGA,LENGTH,GDS0,GDS1,step,stepm,stepp,scastep ;

	TImage	temp1(MEM) ;


	if (doinit==1) {
		MWDCF(width,height,FLUX,LAMBDA,SigmaImage,MEM,Z, DOBSImage,MU,OMEGA,DEFImage,ImagCTF) ;
		temp1 = MEM ;
		temp1.fft() ;
		temp1 *= ImagCTF ;
		temp1.ifft() ;
		DPRED=1+temp1 ;
		MWDCGRAD(width,height,LAMBDA,DZDLAM,PDZDLA,MU,SigmaImage,DOBSImage,DPRED,GRAD,PGRAD) ;
	}
	else {
		OMEGA=IOMEGA ;
	}
	
	
	// Calculate Delta Lambda and the length of Lambda

	MWDCDEL(doinit,width,height,nconj,LAMBDA,DZDLAM,DELLAM,GRAD,PGRAD,iconj) ;
	MWDCLEN(width,height,DELLAM,LENGTH) ;
	if(doinit==1) {
		scale=TOL/LENGTH ;
	}
	else {
		LENGTH=scale*LENGTH ;
	}

	MWDCGDST(width,height,LAMBDA,DELLAM,MU,SigmaImage,DOBSImage,DPRED, GDS0) ;

	if(LENGTH>0) {
		step=min(1.0,TOL/LENGTH) ;
		stepm=min(10.0,max(1.0,TOL/LENGTH)) ;
	}
	else {
		step=1.0 ;
		stepm=10.0 ;
	}
	stepp=step ;
	scastep=scale*step ;

	OUT=LAMBDA ;
	MWDTAKST(LAMBDA,DELLAM,OUT,scastep) ;

	LAMBDA=OUT ;

	MWDCF(width,height,FLUX,LAMBDA,SigmaImage,MEM,Z, DOBSImage,MU,OMEGA,DEFImage,ImagCTF) ;


	temp1 = MEM ;
	temp1.fft() ;
	temp1 *= ImagCTF ;
	temp1.ifft() ;
	DPRED=1+temp1 ;
	MWDCGDST(width,height,LAMBDA,DELLAM,MU,SigmaImage,DOBSImage,DPRED, GDS1) ;

	//Calculate Optimum Step using usual Formula

	if (GDS0!=GDS1) {
		step=max(0.0,min(step*GDS0/(GDS0-GDS1),stepm)) ;
		if (step <= 0.0) {
			printf("MWDONEIT ERROR-Incorrect Curvature\n") ;
			step=1.0 ;
		}
	}
	else {
		printf("MWDONEIT ERROR-no Progress\n") ;
		step=1.0 ;
	}


	scastep=scale*(step-stepp) ;

	OUT=LAMBDA ;
	MWDTAKST(LAMBDA,DELLAM,OUT,scastep) ;

	LAMBDA=OUT ;

	// Now Update MU
	if (OMEGA > 0.0) {
		MWDCMU(width,height,OMEGA,SigmaImage,LAMBDA,MU) ;
	}

	MWDCF(width,height,FLUX,LAMBDA,SigmaImage,MEM,Z, DOBSImage,MU,OMEGA,DEFImage,ImagCTF) ;
	temp1 = MEM ;
	temp1.fft() ;
	temp1 *= ImagCTF ;
	temp1.ifft() ;
	DPRED=1+temp1 ;
	MWDCGRAD(width,height,LAMBDA,DZDLAM,PDZDLA,MU,SigmaImage,DOBSImage,DPRED,GRAD,PGRAD) ;

	// Slowly update the scaling factor
	scale=scale*(3.0+step)/4.0 ;

}


//// Function Chisq2 
void CCHISQ(TImage &DOBSImage,TImage &DPRED,TImage &SigmaImage,float &Chisq) {
	Chisq=0.0 ;
	TImage	temp(DOBSImage) ;
	temp -= DPRED ;
	temp ^= 2 ;
	// Chisq = Sum(temp) ; PATCH NEEDED HERE
}

TImage	*GenerateCTFImage(MEMRecord &memRecord) {
	float	Cs = memRecord.cs*10000000 ;
	float	freqSquared ;
	float	div = memRecord.divergence*0.001 ;
	float	delta = memRecord.delta ;
	float	focus = memRecord.defocus ;
	float	voltage = memRecord.voltage ;
	float	hsq,waveLength ;
	double	chi,pi = 2.0*asin(1.0) ;
	TArray	*data ;
	
	long	ih,ik,width,height,maxH,maxK ;
	float	calibrationSquared,calibration,coherency,chromatic,argCoherency1,argCoherency2,argChromatic ;
	Cmplx	ctfValue,*cp ;
	
	TImageWindow	*theWindow = memRecord.window ;
	TImage			*theImage = theWindow->GetImage() ;
	
	width = theImage->Width() ;
	height = theImage->Height() ;
	maxH = width/2 ;
	maxK = height/2 ;
	calibration = theImage->GetCalibration() ;
	CalibrationUnit	calibrationUnit = theImage->GetCalibrationUnit() ;
	
	if (calibrationUnit == kNanometer)
		calibration *= 10. ;
		
	calibration	= 1./(width*calibration) ;

	calibrationSquared = calibration*calibration ;
//LDM: use float here ?
    waveLength = 12.263652/sqrt((float) (voltage+9.785025e-4*voltage*voltage)*1000.) ;
	
	argChromatic = -0.5*pi*pi*waveLength*waveLength*delta*delta ;
	argCoherency1 = -pi*pi*div*div ;
	argCoherency2 = Cs*waveLength*waveLength ;
	
	TImage	*ctf = new TImage(width,height,nil,90.,k8ByteComplex) ;
	ctf->SetName("CTF") ;
	data = ctf->GetDataArray() ;
	
	for (ih = -maxH; ih <maxH; ih++) {
		hsq = ih*ih*calibrationSquared ;
		for (ik = -maxK; ik <maxK; ik++) {
			freqSquared = hsq + ik*ik*calibrationSquared ;
			chi = pi*waveLength*freqSquared*(Cs*waveLength*waveLength*freqSquared/2 + focus) ;
			// ctfValue = complex(cos(chi), sin(chi)) ; PATCH HERE
			chromatic = exp(argChromatic*freqSquared*freqSquared) ;
			coherency = exp(argCoherency1*freqSquared*(focus + argCoherency2*freqSquared)*(focus + argCoherency2*freqSquared)) ;
			ctfValue *= chromatic*coherency ;
			cp = (Cmplx *)data->GetRealHK(ih,ik) ; 
			*cp = ctfValue ;
		}
	}
	return ctf ;
}

TWindowList	*GetApplicationWindowList() ;
void	MaximumEntropyConvolution(TImageWindow *window) {
	long 	width, height,iter,niter,nconj,iconj ;
	float	GRAD,PGRAD,OMEGA,MU ;
	float 	TOL,TFLUX,TChisq,Chisq,scale ;
	TImage	*TempImage,*SigmaImage ;
	char	text[256] ;
	
	gMEMRecord.window = window ;
	TRunMEMDialog *theDialog = new TRunMEMDialog(gMEMRecord,GetApplicationWindowList()) ;
	if (!theDialog->DoModalDialog()) {
		delete theDialog ;
		return ;
	}
	theDialog->GetDialogData(&gMEMRecord) ;
	delete theDialog ; 
	

	gMEMRecord.window->Update() ;
	
	// Generate the CTF
	TImage *CTF = GenerateCTFImage(gMEMRecord) ;
	CTF->Imaginary() ;


	if (gMEMRecord.showCTFImage) {
		CTF->SetName("CTF - Sin(chi)") ;				
		CTF->Display() ;		
	}	
	
	TImage DOBSImage(*gMEMRecord.window->GetImage()) ;
	
	// Set mean to 1
	// DOBSImage /= Mean(DOBSImage) ; PATCH NEEDED HERE
	
	// Generate noise image
	TempImage = nil ;
	switch(gMEMRecord.noiseType) {
		case kUseNoiseFromWienerFilter:
			SigmaImage = WienerFilter(&DOBSImage) ;
			*SigmaImage -= DOBSImage ;
			break ;
		case kUsePercentageNoise:
			SigmaImage = new TImage(DOBSImage) ;
			*SigmaImage *= gMEMRecord.percentage/100. ;
			break ;
		case kUseNoiseImage:
			if (gMEMRecord.noiseImageWindowListIndex == -1) 
				SigmaImage = new TImage(DOBSImage.Width(),DOBSImage.Height()) ;
			else {
				TImageWindow *imageWindow = (TImageWindow *)GetApplicationWindowList()->GetWindow(gMEMRecord.noiseImageWindowListIndex) ;
				SigmaImage = new TImage(*imageWindow->GetImage()) ;
			}
			break ;
	}
	
	if (gMEMRecord.showNoiseImage) {
		SigmaImage->SetName("Noise") ;				
		SigmaImage->Display() ;		
	}	
	
	
	// Generate the DEFImage
	TImage DEFImage(DOBSImage) ;
	TempImage = new TImage(*CTF) ;	
	*TempImage ^= 2. ;
	*TempImage += 0.25 ;	
	DEFImage.fft() ;
	DEFImage *= *CTF ;
	DEFImage /= *TempImage ;
	DEFImage.ifft() ;
	
	// DEFImage -= Min(DEFImage) ; PATCHES NEEDED HERE
	// DEFImage *= 7500/Sum(DEFImage) ;
	
	
	delete TempImage ;

	niter = gMEMRecord.numIterations ;
	nconj = gMEMRecord.numConjugateSteps ;
	TFLUX = gMEMRecord.totalFlux ;
	TChisq = gMEMRecord.targetChiSQ ;
	TOL = gMEMRecord.lagrangeMultiplier ;
	
	// TFLUX = Sum(DEFImage) ; PATCH NEEDED HERE
	
	iconj	=	0 ;
	scale	=	0 ;
	OMEGA	=	1.0 ;
	MU		=	0.0 ;
	GRAD	=	0.0 ;
	
	width	= DOBSImage.Width() ;
	height	= DOBSImage.Height() ;
	
	// Generate all the work images
	TImage	ImagCTF(*CTF) ;
	TImage	Z(width,height)  ;
	TImage	OUT(width,height)  ;
	TImage	LAMBDA(width,height)  ;
	TImage	DZDLAM(width,height)  ;
	TImage	PDZDLA(width,height)  ;
	TImage	DELLAM(width,height)  ;	
	TImage	MEM(width,height) ;	
	TImage	DPRED(width,height) ;
	
	for (iter=1;iter<=niter;iter++) {

		MWDONEIT(iter,width,height,nconj,TFLUX,DOBSImage,*SigmaImage,TChisq,TOL,MEM,Z,DPRED,
				GRAD,PGRAD,OUT,LAMBDA,DZDLAM,PDZDLA,DELLAM,DEFImage,iconj,scale,MU,ImagCTF) ;


		CCHISQ(DOBSImage,DPRED,*SigmaImage,Chisq) ;
		sprintf(text,"Iter # %d, Chisq = %f",iter,Chisq) ;
		WriteProgress(text) ;
	}

	TImage	*PredictedImage = new TImage(DPRED) ;
	PredictedImage->SetName("Predicted Image From Extended Potential Map") ;
	PredictedImage->Display() ;

	TImage	*PotentialMapImage = new TImage(MEM) ;
	PotentialMapImage->SetName("Extended Potential Map") ;
	PotentialMapImage->Display() ;	
	
	delete SigmaImage ;
}

