#include "TApplication.h"
#include "TApplicationWindow.h"
// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------

#include <stdlib.h>
#include "globals.h"
#include "XDrivers.h"
#include "TBaseObject.h"
#include "config.h"

// See if we are to use forms or not
bool gUse_FORMS	= true ;


#include "EventNames.h"

XFontStruct	*gMenuFont ;
XFontStruct	*gDialogFont ;
XFontStruct	*gWindowFont ;

//#define gMenuFontName "6x10"
//#define gWindowFontName "6x10"
//#define gDialogFontName "6x10"
#define gWindowFontName "-adobe-Times-medium-r-normal-*-10-100-*-*-p-*-iso8859-1"
#define gMenuFontName "-adobe-Times-medium-r-normal-*-10-100-*-*-p-*-iso8859-1"
#define gDialogFontName "-adobe-Times-medium-r-normal-*-10-100-*-*-p-*-iso8859-1"
//#define gDialogFontName "-Adobe-Times-Medium-R-Normal--10-100-75-75-P-54-ISO8859-1"
//#define gAltFontName "-adobe-times-medium-r-normal-*-*-120-*-*-p-*-iso8859-1"
//#define gAltFontName "-Adobe-Times-Medium-R-Normal--10-100-75-75-P-54-ISO8859-1"
//#define gAltFontName1 "-Adobe-Times-Medium-R-Normal--8-80-75-75-P-44-ISO8859-1"

#define NUM_COLORS	32
short	gNumGrayLevels ;

bool	gMouseIsDown = false ;
typedef enum { PSEUDOCOLOR, TRUECOLOR } MonitorDisplayType ;
MonitorDisplayType  gMonitorDisplayType = PSEUDOCOLOR ;
unsigned long gLut[512] ;	// This takes care of all our gray level needs plus color entries
Visual	*gVisual ;
short	gImageDepth ;
Atom	gWindowManagerAtoms[2];
extern	Boolean gOptionKeyDown,gShiftKeyDown,gCommandKeyDown,gControlKey ;
Time gClickTime,gLastClickTime=0 ;
TWindow	*gFrontWindow = nil ;
void	AddWindowToWindowsMenu(TWindow *theWindow) ;

short	gReportVisuals = 0 ;
short	gDebugMode = false ;

extern	void	RemoveWindowFromWindowsMenu(TWindow *theWindow) ;
unsigned char *gMapFrom16To24BitColor = nil ;
unsigned char *gMapFrom24To16BitColor = nil ;

unsigned long gRedMask ;
unsigned long gGreenMask ;
unsigned long gBlueMask ;
short	gRedBitShift ;
short	gGreenBitShift ;
short	gBlueBitShift ;
#define	kDoubleTime	400
Boolean	gIsDoubleClick = false ;
Boolean IsDoubleClick() ;
Boolean	IsDoubleClick() { return gIsDoubleClick ; } 
long	near(float) ;
TApplication	*gApplication ;
void	TerminateApplication() ;
void	TerminateApplication() { gApplication->Terminate()  ;} 
TWindow	*GetTargetWindow() ;

short	gBitsPerPRGB ;

#define kMM		0x4D4D
#define kINTEL	0x4949

typedef	enum {kMMByteOrder = kMM, kIntelByteOrder = kINTEL} ByteOrder ;
extern	ByteOrder	gCPUByteOrder ;
extern	unsigned short ByteSwapShort(unsigned short ) ;
extern	unsigned long ByteSwapLong(unsigned long ) ;

TClipboard	*gClipboard ;

void	SetModifierKeyDown(XEvent *event) ;
void	SetModifierKeyUp(XEvent *event) ;
Boolean	ControlKeyDown() ;

Boolean	gWMFocusMove = true ;

TWindow	*GetTargetWindow() {
	return gApplication->GetTargetWindow() ;
}

bool	MouseIsDown() ;
bool	MouseIsDown() { return gMouseIsDown ; } ;
void	SetTargetWindow(TWindow *window) ;
void	SetTargetWindow(TWindow *window) {
	gApplication->SetTargetWindow(window) ;
}

extern "C" {
int myhandler (Display *display, XErrorEvent *myerr) ;
}

int myhandler (Display *display, XErrorEvent *myerr) {

	char msg[80];
	XGetErrorText(display, myerr->error_code, msg, 80);
	fprintf(stderr, "X-server error %s\n", msg);
        return 0 ;
}

void PrintUsage()  ;
void PrintUsage() {
	fprintf(stderr,"\n -------------------------------------------------------- ") ;
	fprintf(stderr,"\n Valid flags are :") ;
	fprintf(stderr,"\n -cf                     Starts up with click-to-focus    ") ;
	fprintf(stderr,"\n -t                        test: puts up a color window   ") ;
	fprintf(stderr,"\n -v                           verbose: report all visuals ") ;
	fprintf(stderr,"\n -d                        debug: report lost of x-events ") ;
	fprintf(stderr,"\n ---------------------------------------------------------") ;	
	fprintf(stderr,"\n") ;
}


TWindowList	*GetApplicationWindowList() {
	return gApplication->GetWindowList() ;
}

TApplication::TApplication(int argc,char *argv[]) {

	mWindowList = new TWindowList ;
	gClickTime = 0 ;
	mDone = false ;
	mCurrentTargetWindow = nil ;
	mApplicationWindow = nil ;
	char	*displayName = NULL ;
	if (gUse_FORMS) {
		fl_initialize(&argc,argv,"XApplication",0,0) ;
		gDisplay = fl_get_display() ;
	}
	else {
		gDisplay = XOpenDisplay(displayName) ;
		if (gDisplay == nil) {
			fprintf(stderr,"\nUnable to connect to server : %s",XDisplayName(displayName)) ;
			fflush(stderr) ;
		}	
	}
	if (gDisplay == nil) {
		fprintf(stderr,"\nFailed to create a display for default environment\n") ;
		fprintf(stderr,"Display is %s\n",XDisplayName(displayName)) ;
		exit(0) ;
	}
	//fprintf(stderr,"\nConnected to display: %s",XDisplayName(displayName)) ;
	
#ifdef	HAVE_CLICK_TO_FOCUS
gWMFocusMove = false ;
#endif

    if (argc > 1) {
	    for (int i=1;i<argc;i++) {
			if (strncmp(argv[i],"-v",2) == 0) {
				gReportVisuals = 1 ;
			}
			if (strncmp(argv[i],"-d",2) == 0) {
				gDebugMode = 1 ;
			}
			if (strncmp(argv[i],"-cf",3) == 0) {
				gWMFocusMove = false ;
			}
			if (strncmp(argv[i],"-h",2) == 0) {
				PrintUsage() ;
				exit(0) ;
			}
	    }
    }

	if (gDebugMode) {
		fprintf(stderr,"\nEDM:  pointer to focus manager  =  %d",gWMFocusMove) ;
	}
	
	// Set up the error handler and see if it works
	XSetErrorHandler(myhandler);

	Initialize() ;
	gApplication = this ;

	gCurrentPort = nil ;
	
	gClipboard = new TClipboard ;
}

TApplication::~TApplication() {
	delete mWindowList ;
	delete gClipboard ;
    delete[] gMapFrom16To24BitColor ;
    delete[] gMapFrom24To16BitColor ;
}

void	TApplication::AddWindow(TWindow *theWindow) { 
	if (gDebugMode) fprintf(stderr,"\nTApplication::AddWindow : mWindowList->AddWindow(%s)",theWindow->GetWindowTitle()) ;
	mWindowList->AddWindow(theWindow) ; 
	theWindow->SetWindowList(mWindowList) ; 
	if (theWindow->CanReceiveMenuCommands()) {
		if (gDebugMode) fprintf(stderr,"\nTApplication::AddWindow : SetTargetWindow(%s)",theWindow->GetWindowTitle()) ;
		SetTargetWindow(theWindow) ;
	}
	if (gDebugMode) fprintf(stderr,"\nTApplication::AddWindow : Leaving",theWindow->GetWindowTitle()) ;
} 
void	TApplication::RemoveWindow(TWindow *theWindow)  { 
	bool	canReceiveMenuCommands  = theWindow->CanReceiveMenuCommands() ;
	RemoveWindowFromWindowsMenu(theWindow) ;
	mWindowList->RemoveWindow(theWindow) ; 
	if (canReceiveMenuCommands) SetTargetWindow(nil) ;
} 

void TApplication::SetTargetWindow(TWindow *win) { 

	if (mCurrentTargetWindow == win) return ;
	char windowName[256] ;
	if (win != nil) {
		win->GetWindowTitle(windowName) ;

		if (gDebugMode) fprintf(stderr,"\nTApplication::SetTargetWindow : Set Target Window = %s",windowName) ;
		
		if (mCurrentTargetWindow != nil) {
			mCurrentTargetWindow->SetAsTarget(false) ;
			mLastTargetWindow = mCurrentTargetWindow ;
		}
		else {
			mLastTargetWindow = win ;
		}

		mCurrentTargetWindow = win ;
		mCurrentTargetWindow->SetAsTarget(true) ;
		BringToFront(mCurrentTargetWindow) ;
		gFrontWindow = mCurrentTargetWindow ;
		return ;
	}

	// the window is nil
	// This is because we deleted the target window
	// See if we can find a new target window in our list
	TWindow *nextWindow,*newTargetWindow = nil ;
	for(int i = 0; i < mWindowList->GetNumberOfWindows() ; i++) {
		nextWindow = mWindowList->GetWindow(i) ;
		if (nextWindow->CanReceiveMenuCommands()) {
			newTargetWindow = nextWindow ;
			if (newTargetWindow == mLastTargetWindow && newTargetWindow != nil) {
				mCurrentTargetWindow = newTargetWindow ;
				mCurrentTargetWindow->SetAsTarget(true) ;
				BringToFront(mCurrentTargetWindow) ;
				gFrontWindow = mCurrentTargetWindow ;
				mCurrentTargetWindow->GetWindowTitle(windowName) ;
				if (gDebugMode)
					fprintf(stderr,"\nTApplication::SetTargetWindow : Target Window(in Windows list) = %s",windowName) ;
				return ;
			}
		}
	}
	if (newTargetWindow != nil) {
		mCurrentTargetWindow = newTargetWindow ;
		mCurrentTargetWindow->SetAsTarget(true) ;
		BringToFront(mCurrentTargetWindow) ;
		gFrontWindow = mCurrentTargetWindow ;
		mCurrentTargetWindow->GetWindowTitle(windowName) ;
		if (gDebugMode)
			fprintf(stderr,"\nDoMouseDownInMenuBar : Target Window = %s",windowName) ;
	}
	else {	// We deleted the only targetable window that we have
		mCurrentTargetWindow = nil ;
		gFrontWindow = gControlWindow ;
		mLastTargetWindow = nil ;
		if (gDebugMode)
			fprintf(stderr,"\nTApplication::SetTargetWindow : Target Window = nil") ;
	}
}

void TApplication::Initialize() {
	char thisDisplay[256],*displayName = NULL ;
	Boolean	forcePseudoColor ;

	strcpy(thisDisplay,XDisplayName(displayName)) ;
	//fprintf(stderr,"Connected to %s !\n",thisDisplay) ; 
	//fflush(stderr) ;

	/* Get the screen size for the display	*/
	gScreen = DefaultScreen(gDisplay) ;
	gDisplayWidth = DisplayWidth(gDisplay,gScreen) ;
	gDisplayHeight = DisplayHeight(gDisplay,gScreen) ;
    
	forcePseudoColor = false ;
	gImageDepth = DefaultDepth(gDisplay,gScreen);

	short i ;
    gMapFrom16To24BitColor = new unsigned char[32] ;
    for(i=0; i < 32; i++) {
    	gMapFrom16To24BitColor[i] = near(i*255.0/31.0) ;
    }
    gMapFrom24To16BitColor = new unsigned char[256] ;
    for(i=0; i < 256; i++) {
    	gMapFrom24To16BitColor[i] = near(i*31.0/255.0) ;
    }

    if (gImageDepth == 15) {
       gOutDevice->range = 31 ;
	}

	if (gImageDepth == 8 ) {
		forcePseudoColor = true ;
	}

	GetVisual(forcePseudoColor,&gVisualInfo) ;
	gVisual = DefaultVisual(gDisplay,gScreen) ; /* New for LUT */
	gVisualInfo.visual = gVisual ;		/* New for LUT */


	gRedMask = gVisualInfo.red_mask ;
	gGreenMask = gVisualInfo.green_mask ;
	gBlueMask = gVisualInfo.blue_mask ;
	
	gBitsPerPRGB = gVisualInfo.bits_per_rgb ;
	
	// Find out how many bits to shift for each
	unsigned long bitmask ;
	unsigned long mask_mask ;
	unsigned long bit ;
	
	if (gCPUByteOrder == kMMByteOrder) {
		if (gVisualInfo.depth == 32) {
			for(gRedBitShift = 0, bitmask = 0x0000000001; gRedBitShift < 32 ; gRedBitShift++) {
				if ((gRedMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gGreenBitShift = 0, bitmask = 0x0000000001; gGreenBitShift < 32 ; gGreenBitShift++) {
				if ((gGreenMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gBlueBitShift = 0, bitmask = 0x0000000001;  gBlueBitShift < 32; gBlueBitShift++) {
				if ((gBlueMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			
			// Create the mask for the bits per pixel
			bitmask = 0 ;
			bit = 1 ;
			for(i=0; i < gBitsPerPRGB; i++) {
				bitmask |= bit ;
				bit <<= 1 ;
			}
			// We probably have 8 bits per color
			gRedMask = bitmask << gRedBitShift ;
			gGreenMask = bitmask << gGreenBitShift ;
			gBlueMask = bitmask << gBlueBitShift ;
			gRedMask &= mask_mask ;
			gGreenMask &= mask_mask ;
			gBlueMask &= mask_mask ;
		}
		else if (gVisualInfo.depth == 24) {
			mask_mask = 0x00FFFFFF ;
			gRedMask &= mask_mask ;
			gGreenMask &= mask_mask ;
			gBlueMask &= mask_mask ;
			for(gRedBitShift = 0, bitmask = 0x0000000001; gRedBitShift < 32 ; gRedBitShift++) {
				if ((gRedMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gGreenBitShift = 0, bitmask = 0x0000000001; gGreenBitShift < 32 ; gGreenBitShift++) {
				if ((gGreenMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gBlueBitShift = 0, bitmask = 0x0000000001;  gBlueBitShift < 32; gBlueBitShift++) {
				if ((gBlueMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			// Create the mask for the bits per pixel
			bitmask = 0 ;
			bit = 1 ;
			for(i=0; i < gBitsPerPRGB; i++) {
				bitmask |= bit ;
				bit <<= 1 ;
			}
			gRedMask = bitmask << gRedBitShift ;
			gGreenMask = bitmask << gGreenBitShift ;
			gBlueMask = bitmask << gBlueBitShift ;
			gRedMask &= mask_mask ;
			gGreenMask &= mask_mask ;
			gBlueMask &= mask_mask ;
		}
		else if (gVisualInfo.depth == 16 || gVisualInfo.depth == 15) {
			mask_mask = 0x0000FFFF ;
			gRedMask &= mask_mask ;
			gGreenMask &= mask_mask ;
			gBlueMask &= mask_mask ;
			for(gRedBitShift = 0, bitmask = 0x0000000001; gRedBitShift < 32 ; gRedBitShift++) {
				if ((gRedMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gGreenBitShift = 0, bitmask = 0x0000000001; gGreenBitShift < 32 ; gGreenBitShift++) {
				if ((gGreenMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gBlueBitShift = 0, bitmask = 0x0000000001;  gBlueBitShift < 32 ; gBlueBitShift++) {
				if ((gBlueMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			// Create the mask for the bits per pixel
			bitmask = 0 ;
			bit = 1 ;
			for(i=0; i < gBitsPerPRGB; i++) {
				bitmask |= bit ;
				bit <<= 1 ;
			}
			gRedMask = bitmask << gRedBitShift ;
			gGreenMask = bitmask << gGreenBitShift ;
			gBlueMask = bitmask << gBlueBitShift ;
			gRedMask &= mask_mask ;
			gGreenMask &= mask_mask ;
			gBlueMask &= mask_mask ;
		}
	}
	else {
		// Testing
		/*
		unsigned long test = 15 ;
		fprintf(stderr,"\n test %u = %0X",test,test) ;
		test = 1 ;
		fprintf(stderr,"\n test for Testing = %u = %0X",test,test) ;
		for(i = 0; i < 32; i++) {
			fprintf(stderr,"\n test shifted left %d = %u = %0X",i,test,test) ;
			test <<= 1 ;
		}
		test = 1 ;
		for(i=0; i < 32; i++) {
			fprintf(stderr,"\n test shifted right %d = %u = %0X",i,test,test) ;
			test >>= i ;
		}
		fprintf(stderr,"\n gRedMask %u = %0X",gRedMask,gRedMask) ;
		fprintf(stderr,"\n gGreenMask %u = %0X",gGreenMask,gGreenMask) ;
		fprintf(stderr,"\n gBlueMask %u = %0X",gBlueMask,gBlueMask) ;
		*/
			 
		if (gVisualInfo.depth == 32) {
			for(gRedBitShift = 0, bitmask = 0x0000000001; gRedBitShift < 32 ; gRedBitShift++) {
				if ((gRedMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gGreenBitShift = 0, bitmask = 0x0000000001; gGreenBitShift < 32 ; gGreenBitShift++) {
				if ((gGreenMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gBlueBitShift = 0, bitmask = 0x0000000001;  gBlueBitShift < 32; gBlueBitShift++) {
				if ((gBlueMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			// Create the mask for the bits per pixel
			bitmask = 0 ;
			bit = 1 ;
			for(i=0; i < gBitsPerPRGB; i++) {
				bitmask |= bit ;
				bit <<= 1 ;
			}
			gRedMask = bitmask << gRedBitShift ;
			gGreenMask = bitmask << gGreenBitShift ;
			gBlueMask = bitmask << gBlueBitShift ;
			gRedMask &= mask_mask ;
			gGreenMask &= mask_mask ;
			gBlueMask &= mask_mask ;
		}
		else if (gVisualInfo.depth == 24) {
			mask_mask = 0x00FFFFFF ;
			gRedMask &= mask_mask ;
			gGreenMask &= mask_mask ;
			gBlueMask &= mask_mask ;
			for(gRedBitShift = 0, bitmask = 0x0000000001; gRedBitShift < 32 ; gRedBitShift++) {
				if ((gRedMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gGreenBitShift = 0, bitmask = 0x0000000001; gGreenBitShift < 32 ; gGreenBitShift++) {
				if ((gGreenMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gBlueBitShift = 0, bitmask = 0x0000000001;  gBlueBitShift < 32; gBlueBitShift++) {
				if ((gBlueMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			// Create the mask for the bits per pixel
			bitmask = 0 ;
			bit = 1 ;
			for(i=0; i < gBitsPerPRGB; i++) {
				bitmask |= bit ;
				bit <<= 1 ;
			}
			gRedMask = bitmask << gRedBitShift ;
			gGreenMask = bitmask << gGreenBitShift ;
			gBlueMask = bitmask << gBlueBitShift ;
			gRedMask &= mask_mask ;
			gGreenMask &= mask_mask ;
			gBlueMask &= mask_mask ;
		}
		else if (gVisualInfo.depth == 16 || gVisualInfo.depth == 15) {
			mask_mask = 0x0000FFFF ;
			gRedMask &= mask_mask ;
			gGreenMask &= mask_mask ;
			gBlueMask &= mask_mask ;
			for(gRedBitShift = 0, bitmask = 0x0000000001; gRedBitShift < 32 ; gRedBitShift++) {
				if ((gRedMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gGreenBitShift = 0, bitmask = 0x0000000001; gGreenBitShift < 32 ; gGreenBitShift++) {
				if ((gGreenMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			for(gBlueBitShift = 0, bitmask = 0x0000000001;  gBlueBitShift < 32 ; gBlueBitShift++) {
				if ((gBlueMask & bitmask) != 0) break ;
				bitmask <<= 1 ;
			}
			// Create the mask for the bits per pixel
			bitmask = 0 ;
			bit = 1 ;
			for(i=0; i < gBitsPerPRGB; i++) {
				bitmask |= bit ;
				bit <<= 1 ;
			}
			gRedMask = bitmask << gRedBitShift ;
			gGreenMask = bitmask << gGreenBitShift ;
			gBlueMask = bitmask << gBlueBitShift ;
			gRedMask &= mask_mask ;
			gGreenMask &= mask_mask ;
			gBlueMask &= mask_mask ;
		}
	}
	if (gReportVisuals) {
		fprintf(stderr,"\n gRedBitShift = %d",gRedBitShift) ;
		fprintf(stderr,"\n gGreenBitShift = %d",gGreenBitShift) ;
		fprintf(stderr,"\n gBlueBitShift = %d",gBlueBitShift) ;
		fprintf(stderr,"\n gRedBitMask = %0X",gRedMask) ;
		fprintf(stderr,"\n gGreenBitMask = %0X",gGreenMask) ;
		fprintf(stderr,"\n gBlueBitMask = %0X",gBlueMask) ;
	}
 
	gMenuFont = XLoadQueryFont(gDisplay,gMenuFontName) ; 
	if (gMenuFont == NULL) {
		printf("\nFailed to load the font %s!\n",gMenuFontName) ;
		exit(0) ;
	}	
	gFontStructure = gMenuFont ;
	gDialogFont = XLoadQueryFont(gDisplay,gDialogFontName) ; 
	if (gDialogFont == NULL) {
		printf("\nFailed to load the font %s!\nUsing %s\n",gDialogFontName,gMenuFontName) ;
		gDialogFont = gMenuFont ;
	}	
	gWindowFont = XLoadQueryFont(gDisplay,gWindowFontName) ; 
	if (gWindowFont == NULL) {
		printf("\nFailed to load the font %s!\nUsing %s\n",gWindowFont,gMenuFontName) ;
		gWindowFont = gMenuFont ;
	}	
	gImageAttributes.background_pixel = WhitePixel(gDisplay,gScreen) ;
	gImageAttributes.border_pixel = BlackPixel(gDisplay,gScreen) ;
	gImageAttributes.backing_store = WhenMapped ;
  
	gWindowManagerAtoms[0] = XInternAtom(gDisplay,"WM_DELETE_WINDOW",false) ;
	gWindowManagerAtoms[1] = XInternAtom(gDisplay,"WM_TAKE_FOCUS",false) ;
  
	SetupColors() ;
	DefineXForms() ;
	ShowXForms() ;
	DefineWindows() ;

}

void TApplication::MainLoop() {
	while (!mDone) {
		ProcessEvents() ;
	}
	CleanUp() ;
}

void TApplication::DoKeyDown(XEvent *event)  {
	XKeyEvent *keyEvent = (XKeyEvent *)event ;
	
	// Received key down in window
	if (ControlKeyDown() || CommandKeyDown()) {
		// This might be a menu command	
		mApplicationWindow->DoMenuKey(event) ;
		return ; 
	}
	SetModifierKeyDown(event) ;	
	if (ControlKeyDown() || CommandKeyDown()) {
		return ;
	}

	// Pass it along in case we want to handle it
	TWindow *theWindow = FindWindow(keyEvent->window) ;
	if (theWindow != nil) { 
		theWindow->DoKeyDown(event) ;
	}
}

void TApplication::DoFocusIn(XEvent *event)  {
	//SetModifierKeyUp(event) ;
	XFocusChangeEvent *theEvent = (XFocusChangeEvent *)event ;
	TWindow *theWindow = FindWindow(theEvent->window) ;
	if (theWindow != nil) { 
		theWindow->DoFocusIn(theEvent) ;
	}
}

void TApplication::DoFocusOut(XEvent *event)  {
	//SetModifierKeyUp(event) ;
	XFocusChangeEvent *theEvent = (XFocusChangeEvent *)event ;
	TWindow *theWindow = FindWindow(theEvent->window) ;
	if (theWindow != nil) { 
		theWindow->DoFocusOut(theEvent) ;
	}
}


void TApplication::DoEnterNotify(XEvent *event)  {
	//SetModifierKeyUp(event) ;
	XCrossingEvent *theEvent = (XCrossingEvent *)event ;
	TWindow *theWindow = FindWindow(theEvent->window) ;
	if (theWindow != nil) { 
		theWindow->DoEnterNotify(theEvent) ;
	}
}

void TApplication::DoLeaveNotify(XEvent *event)  {
	//SetModifierKeyUp(event) ;
	XCrossingEvent *theEvent = (XCrossingEvent *)event ;
	TWindow *theWindow = FindWindow(theEvent->window) ;
	if (theWindow != nil) { 
		theWindow->DoLeaveNotify(theEvent) ;
	}
}

void TApplication::DoKeyUp(XEvent *event)  {
	SetModifierKeyUp(event) ;
	XKeyEvent *keyEvent = (XKeyEvent *)event ;
	TWindow *theWindow = FindWindow(keyEvent->window) ;
	if (theWindow != nil) { 
		theWindow->DoKeyUp(event) ;
	}
}

void TApplication::DoMapNotify(XEvent *event)  {
	XMapEvent *theEvent = (XMapEvent *)event ;
	TWindow *theWindow = FindWindow(theEvent->window) ;
	if (theWindow != nil) { 
		theWindow->DoMapNotify(event) ;
	}
}

void TApplication::DoUnmapNotify(XEvent *event)  {
	XUnmapEvent *theEvent = (XUnmapEvent *)event ;
	TWindow *theWindow = FindWindow(theEvent->window) ;
	if (theWindow != nil) { 
		theWindow->DoUnmapNotify(event) ;
	}
}

void TApplication::DoUpdate(XEvent *event)  {
	
	XExposeEvent *theEvent = (XExposeEvent *)event ;
	TWindow *theWindow ;

	// At some point it appeared that I needed to update all windows for each window update 
	// event. This does no longer appear to be true. Here is the code for both cases
	
	// CASE 1
	// This will update only the window that gets the expose event
	/*
	theWindow = FindWindow(theEvent->window) ;
	if (theWindow != nil) { 
		theWindow->Update(event) ;
	}
	*/
	
	// CASE 2
	// This will update all the windows whenever any winodw gets an expose event
	/* */
	for(int i = 0; i < mWindowList->GetNumberOfWindows() ; i++) {
		theWindow = mWindowList->GetWindow(i) ;
		theWindow->Update(event) ;
	}
	/* */

}

void TApplication::DoMouseDown(XEvent *event) {
	XButtonEvent *buttonEvent = (XButtonEvent *)event ;
	TWindow *theWindow = FindWindow(buttonEvent->window) ;
	if (theWindow != nil) {
		// Get the current Target window
		if (theWindow->CanReceiveMenuCommands()) {
			SetTargetWindow(theWindow) ;
		}
		theWindow->DoMouseDown(event) ;
	}
}

void TApplication::CleanUp() {
} 

void TApplication::SelectCursor() {
	// Find the window that is in the foreground and 
	// select the cursor accordingly
}

void TApplication::DoIdle() {
	SelectCursor() ;	// Select the correct cursor
	// Also do idle for the active window
	// for all windows do idle
} 

void TApplication::DoConfigureNotify(XEvent *event) {
	TWindow *theWindow = FindWindow(event->xexpose.window) ;
	if (theWindow != nil) 
		theWindow->DoConfigureNotify(event) ;
}

void TApplication::DoClientMessage(XEvent *event) {
	XClientMessageEvent *theMessage = (XClientMessageEvent *)event ;
	TWindow *theWindow = FindWindow(theMessage->window) ;
	if (theWindow != nil) 
		theWindow->DoClientMessage(theMessage) ;
}

void TApplication::GetVisual(Boolean forcePseudoColor,XVisualInfo *vRetInfo) {
    int numReturned, i ;
    Status retInfo ;
    XVisualInfo	*vInfo,*visPtr ;

    vInfo = XGetVisualInfo(gDisplay,VisualNoMask,NULL,&numReturned) ;
    visPtr = vInfo ;
    retInfo = 0 ;
    
    
	if (gReportVisuals) {
		// Report all visuals
		fprintf(stderr,"\nThese are the available Visuals");
		for(i=0;i<numReturned;i++,vInfo++) {
			switch(vInfo->c_class) {
				case StaticGray:
					fprintf(stderr,"\nClass = StaticGray , Bit Depth = %d",vInfo->depth) ;
					break ;
				case GrayScale:
					fprintf(stderr,"\nClass = GrayScale , Bit Depth = %d",vInfo->depth) ;
					break ;
				case StaticColor:
					fprintf(stderr,"\nClass = StaticColor , Bit Depth = %d",vInfo->depth) ;
					break ;
				case PseudoColor:
					fprintf(stderr,"\nClass = PseudoColor , Bit Depth = %d",vInfo->depth) ;
					break ;
				case TrueColor:
					fprintf(stderr,"\nClass = TrueColor , Bit Depth = %d",vInfo->depth) ;
					break ;
				case DirectColor:
					fprintf(stderr,"\nClass = DirectColor , Bit Depth = %d",vInfo->depth) ;
					break ;
			}					
		}
		vInfo = visPtr ;
	}
    
    if (forcePseudoColor) {
		if (gReportVisuals) 
	        fprintf(stderr,"Forcing pseudocolor\n");
	    // Try and find an 8-bit display
    	for(i=0;i<numReturned;i++,vInfo++) {
			if (vInfo->c_class == PseudoColor && vInfo->depth == 8) {
				retInfo = XMatchVisualInfo(gDisplay,
							DefaultScreen(gDisplay),
							vInfo->depth,
							vInfo->c_class,
							vRetInfo) ;
				gMonitorDisplayType = PSEUDOCOLOR ;
				if (gReportVisuals) 
	    			fprintf(stderr,"Got 8bit Pseudocolor\n");
				break ;
			}
		}
    }
    else {
	    // Do we have 24 bit?
        vInfo = visPtr ;
		if (gReportVisuals) 
	        fprintf(stderr,"Looing for 24bit\n");
    	for(i=0;i<numReturned;i++,vInfo++) {
			if (vInfo->c_class == TrueColor && (vInfo->depth == 24 || vInfo->depth == 32)) {
				retInfo = XMatchVisualInfo(gDisplay,
							DefaultScreen(gDisplay),
							vInfo->depth,
							vInfo->c_class,
							vRetInfo) ;
				if (gReportVisuals) 
	    			fprintf(stderr,"Got 24bit TrueColor\n");
				gMonitorDisplayType = TRUECOLOR ;
				break ;
			}
		}
    if (!retInfo) { //We have not found it yet, try 16 bit 
    
        vInfo = visPtr ;
		if (gReportVisuals) 
	        fprintf(stderr,"Looing for 16bit\n");
	    	for(i=0;i<numReturned;i++,vInfo++) {
				if (vInfo->c_class == TrueColor && (vInfo->depth == 16 || vInfo->depth == 15)) {
					retInfo = XMatchVisualInfo(gDisplay,
								DefaultScreen(gDisplay),
								vInfo->depth,
								vInfo->c_class,
								vRetInfo) ;
					if (gReportVisuals) 
		    			fprintf(stderr,"Got 16bit TrueColor\n");
					gMonitorDisplayType = TRUECOLOR ;
					break ;
				}
			}
		}		
    if (!retInfo) { //Still not found it , try 8 bit pseudocolor
        vInfo = visPtr ;
		if (gReportVisuals) 
	        fprintf(stderr,"Looing for 8bit\n");
	    	for(i=0;i<numReturned;i++,vInfo++) {
				if (vInfo->c_class == PseudoColor && vInfo->depth == 8) {
					retInfo = XMatchVisualInfo(gDisplay,
								DefaultScreen(gDisplay),
								vInfo->depth,
								vInfo->c_class,
								vRetInfo) ;
					if (gReportVisuals) 
		    			fprintf(stderr,"Got 8bit Pseudocolor\n");
					gMonitorDisplayType = TRUECOLOR ;
					break ;
				}
			}
		}		
    } 
    if (!retInfo) {
		printf("\n Unable to locate the right Visual. Exiting...\n") ;
		exit(-1) ;
    }
    
    if (gReportVisuals) {
    	// Print out the details of the visual info
    	fprintf(stderr,"\ndepth : %d",vRetInfo->depth) ;
    	fprintf(stderr,"\nred_mask: %ud %X",vRetInfo->red_mask,vRetInfo->red_mask) ;
    	fprintf(stderr,"\ngreen_mask: %ud %X",vRetInfo->green_mask,vRetInfo->green_mask) ;
    	fprintf(stderr,"\nblue_mask: %ud %X",vRetInfo->blue_mask,vRetInfo->blue_mask) ;
    	fprintf(stderr,"\ncolormap_size: %d",vRetInfo->colormap_size) ;
    	fprintf(stderr,"\nbits_per_rgb: %d",vRetInfo->bits_per_rgb) ;
    }
    XFree(visPtr) ; 
	
}

void TApplication::SetupColors() {

    int		i,base = 0 ;

    gColorMap = DefaultColormap(gDisplay,gScreen) ;
    
    if (gMonitorDisplayType == PSEUDOCOLOR) {
    	gNumGrayLevels = 128 ;
    	XAllocColorCells(gDisplay,gColorMap,0,NULL,0,gLut,
    		     	gNumGrayLevels+NUM_COLORS) ;
     	for(i=0;i < gNumGrayLevels;i++) {
    		gColors[i].pixel	= gLut[i] ;
    		base				= i*65535/(gNumGrayLevels-1) ;
    		gColors[i].red 		= base ;
    		gColors[i].green 	= base ;
    		gColors[i].blue 	= base ;
    		gColors[i].flags	= DoRed | DoGreen | DoBlue ;
    	}
   }
   else {
    	gNumGrayLevels = 256 ;
    	for(i=0;i < gNumGrayLevels;i++) {
    		gColors[i].pixel	= i ;
    		base				= i*65535/(gNumGrayLevels-1) ;
    		gColors[i].red 		= base ;
    		gColors[i].green 	= base ;
    		gColors[i].blue 	= base ;
    		gColors[i].flags	= DoRed | DoGreen | DoBlue ;
    	}
    }
    DefineColors(&gColors[gNumGrayLevels],&gLut[gNumGrayLevels],NUM_COLORS) ;

   if (gMonitorDisplayType == PSEUDOCOLOR) 
     	for(i=gNumGrayLevels;i < gNumGrayLevels+NUM_COLORS;i++) 
    		gColors[i].pixel	= gLut[i] ;
   else 
    	for(i=gNumGrayLevels;i < gNumGrayLevels+NUM_COLORS;i++) 
     		gColors[i].pixel	= i ;

    if (gMonitorDisplayType == PSEUDOCOLOR) {
   		XStoreColors(gDisplay,gColorMap,gColors,gNumGrayLevels+NUM_COLORS) ;
    }
    else {
	  	for(i=0;i < gNumGrayLevels+NUM_COLORS;i++) {
	   		gColors[i].pixel = i ;
	   		XAllocColor(gDisplay,gColorMap,&gColors[i]) ;
	   		gLut[i] =  i ;
	  	}
    }
    gImageAttributes.background_pixel = WhitePixel(gDisplay,gScreen) ;
    gImageAttributes.border_pixel = BlackPixel(gDisplay,gScreen) ;
    if (gMonitorDisplayType == PSEUDOCOLOR) {
		gImageAttributes.colormap = gColorMap ;
		gImageDepth = 8 ;	// Force the depth to 8
		gImageLineBuffer = new unsigned char[1024*4] ;
	    	gImageLine = XCreateImage(gDisplay,gVisualInfo.visual,gImageDepth,
	    			ZPixmap,0,(char *)gImageLineBuffer,4096,1,8,0) ;
    }
    else {
    	switch(gImageDepth) {
    		case 15:
    		case 16:
				gImageLineBuffer = new unsigned char[2048*4] ;
			    	gImageLine = XCreateImage(gDisplay,gVisualInfo.visual,gImageDepth,
			    			ZPixmap,0,(char *)gImageLineBuffer,4096,1,16,0) ;
    			break ;
    		case 24:
    		case 32:
				gImageLineBuffer = new unsigned char[4096*4] ;
			    	gImageLine = XCreateImage(gDisplay,gVisualInfo.visual,gImageDepth,
			    			ZPixmap,0,(char *)gImageLineBuffer,4096,1,32,0) ;
    			break ;
    	}
    }
    gBlackPixel = gColors[0].pixel ;
    gLighterGray = gColors[near(0.8*gNumGrayLevels)].pixel ;
    gGrayPixel = gColors[near(0.72*gNumGrayLevels)].pixel ;
    gDarkerGray = gColors[near(0.55*gNumGrayLevels)].pixel ;
    gDarkGray = gColors[near(0.30*gNumGrayLevels)].pixel ;
    gWhitePixel = gColors[gNumGrayLevels-1].pixel ;
    gRedPixel = gColors[gNumGrayLevels].pixel ;
    gGreenPixel = gColors[gNumGrayLevels+1].pixel ;
    gBluePixel = gColors[gNumGrayLevels+2].pixel ;

}

//short   debugProcessModalEvent = true ;
void	TApplication::ProcessModalEvent(XEvent *event) {
	// Other windows only get update and client events
	if (gDebugMode) {
		TWindow *theWindow = FindWindow(((XButtonEvent *)event)->window) ;
		if (theWindow != nil) { 
			fprintf(stderr,"\nProcessModalEvent: Incoming X-Event '%s' for window %s \n",event_names[event->type], theWindow->GetWindowTitle()) ;
		}
		else {
			fprintf(stderr,"\nProcessModalEvent: Incoming X-Event '%s' for window  NOT IN LIST %d'\n",event_names[event->type], ((XButtonEvent *)event)->window) ;
		}
	}
	switch(event->type) {
		case MapNotify:
			DoMapNotify(event) ;
			break ;
		case UnmapNotify:
			DoUnmapNotify(event) ;
			break ;
		case ConfigureNotify:
			DoConfigureNotify(event) ;
			break ;
		case EnterNotify:
			DoEnterNotify(event) ;
			break ;
		case LeaveNotify:
			DoLeaveNotify(event) ;
			break ;
		case Expose:    
			DoUpdate(event) ;
			break ;
		case FocusIn:    
			DoFocusIn(event) ;
			break ;
		case FocusOut:    
			DoFocusOut(event) ;
			break ;
			/*
		case ButtonPress:
			gIsDoubleClick = false ; 
			gMouseIsDown = true ;
			if ((buttEvent->time - gClickTime) < 400) gIsDoubleClick = true ;
				gClickTime	= buttEvent->time ;
			DoMouseDown(event) ;
			break ;
		case ButtonRelease:
			gMouseIsDown = false ;
			break ;
		case KeyPress:
			DoKeyDown(event) ;
			break ;
		case KeyRelease:
			DoKeyUp(event) ;
			break ; 
			*/
		case ClientMessage:
			DoClientMessage(event) ;
			break ;
		default:
			if (gDebugMode)
				fprintf(stderr,"Event '%s' not handled\n",event_names[event->type]) ;
			break ;
	}
}

void TApplication::ProcessEvents() {
	XEvent event ;
	FL_OBJECT	*theFormsObject ;
	XButtonEvent *buttEvent = (XButtonEvent *)&event ;

	DoIdle() ;
	if (gUse_FORMS) {
		theFormsObject = fl_do_forms() ;
		if (theFormsObject == FL_EVENT) {
			fl_XNextEvent(&event) ;
			if (gDebugMode) {
				TWindow *theWindow = FindWindow(buttEvent->window) ;
				if (theWindow != nil) { 
					fprintf(stderr,"\nIncoming X-Event '%s' for window '%s'\n",event_names[event.type],theWindow->GetWindowTitle()) ;
				}
				else {
					fprintf(stderr,"\n Incoming X-Event '%s' for window NOT IN LIST '%d'\n",event_names[event.type], buttEvent->window) ;
				}
			}
			switch(event.type) {
				case MapNotify:
					DoMapNotify(&event) ;
					break ;
				case UnmapNotify:
					DoUnmapNotify(&event) ;
					break ;
				case ConfigureNotify:
					DoConfigureNotify(&event) ;
					break ;
				case EnterNotify:
					DoEnterNotify(&event) ;
					break ;
				case LeaveNotify:
					DoLeaveNotify(&event) ;
					break ;
				case Expose:    
					DoUpdate(&event) ;
					break ;
				case FocusIn:    
					DoFocusIn(&event) ;
					break ;
				case FocusOut:    
					DoFocusOut(&event) ;
					break ;
				case ButtonPress:
					gIsDoubleClick = false ; 
					gMouseIsDown = true ;
					if ((buttEvent->time - gClickTime) < 400) gIsDoubleClick = true ;
					gClickTime	= buttEvent->time ;
					DoMouseDown(&event) ;
					break ;
				case ButtonRelease:
					gMouseIsDown = false ;
					break ;
				case KeyPress:
					DoKeyDown(&event) ;
					break ;
				case KeyRelease:
					DoKeyUp(&event) ;
					break ; 
				case ClientMessage:
					DoClientMessage(&event) ;
					break ;
				default:
					if (gDebugMode)
						fprintf(stderr,"Event '%s' not handled\n",event_names[event.type]) ;
					break ;
			}
		}
	}
	else {
		XNextEvent(gDisplay,&event) ;
		if (gDebugMode)
			fprintf(stderr,"\nIncoming X-Event '%s'\n",event_names[event.type]) ;
		switch(event.type) {
			case ConfigureNotify:
				DoConfigureNotify(&event) ;
				break ;
			case EnterNotify:
				DoEnterNotify(&event) ;
				break ;
			case LeaveNotify:
				DoLeaveNotify(&event) ;
				break ;
			case Expose:    
				DoUpdate(&event) ;
				break ;
			case FocusIn:    
				DoFocusIn(&event) ;
				break ;
			case ButtonPress:
				gIsDoubleClick = false ; 
				gMouseIsDown = true ;
				if ((buttEvent->time - gClickTime) < 400) gIsDoubleClick = true ;
				gClickTime	= buttEvent->time ;
				DoMouseDown(&event) ;
				break ;
			case ButtonRelease:
				gMouseIsDown = false ;
				break ;
			case KeyPress:
				DoKeyDown(&event) ;
				break ;
			case KeyRelease:
				DoKeyUp(&event) ;
				break ; 
			case ClientMessage:
				DoClientMessage(&event) ;
				break ;
			default:
				if (gDebugMode)
					fprintf(stderr,"Event '%s' not handled\n",event_names[event.type]) ;
				break ;
		}
	}
}

void TApplication::DefineColors(XColor *color,unsigned long *pixel,unsigned long numColors){
    XColor *thisColor ;
    unsigned long	*thisPixel ;

    thisColor = color ;
    thisPixel = pixel ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 		= 65535 ;	/* 43961 ; */
    thisColor->green 	= 0 ; // 12037 ;	/* 997 ; */
    thisColor->blue 	= 0 ; // 64675 ;	/* 65535 ; */
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 		= 0 ;
    thisColor->green 	= 65535 ;	/* 1040 ; */
    thisColor->blue 	= 0 ;	/* 46311 ; */
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 		= 0 ;
    thisColor->green 	= 0 ;	/* 602 ; */
    thisColor->blue 	= 65535 ;	/* 3105 ; */
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 65535 ;
    thisColor->green 	= 32688 ;	/* 27792 ; */
    thisColor->blue 	= 10105 ;	/* 1317 ; */
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 65535 ;
    thisColor->green 	= 61965 ;	/* 51958 ; */
    thisColor->blue 	= 10279 ;	/* 599 ; */
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 45330 ;	/* 60128 ; */
    thisColor->green 	= 65535 ;
    thisColor->blue 	= 10665 ;	/* 447 ; */
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 0 ; // 12409 ;	/* 26973 ; */
    thisColor->green 	= 65535 ;
    thisColor->blue 	= 0 ; // 11557 ;	/* 2210 ; */
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 11137 ;	/* 1080 ; */
    thisColor->green 	= 65535 ;
    thisColor->blue 	= 45654 ;	/* 19407 ; */
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 11627 ;	/* 669 ; */
    thisColor->green 	= 62405 ;	/* 65535 ; */
    thisColor->blue 	= 65535 ;	/* 45801 ; */
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 12001 ;	/* 1398 ; */
    thisColor->green 	= 30261 ; 	/* 57373 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 14117 ;	/* 1619 ; */
    thisColor->green 	= 12181 ; 	/* 35557 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;


    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 28943;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 13982;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 7643;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 33450;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 0 ; // 8050 ;	/* 1348 ; */
    thisColor->green 	= 0 ; // 2965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;
    gBluePixel = thisColor->pixel ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 4050 ;	/* 1348 ; */
    thisColor->green 	= 19965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 5535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 18050 ;	/* 1348 ; */
    thisColor->green 	= 42965 ;	/* 3561 ; */
    thisColor->blue 	= 25535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 28050 ;	/* 1348 ; */
    thisColor->green 	= 52965 ;	/* 3561 ; */
    thisColor->blue 	= 15535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 15535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 675;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 55535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 28050 ;	/* 1348 ; */
    thisColor->green 	= 52965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 40050 ;	/* 1348 ; */
    thisColor->green 	= 19965 ;	/* 3561 ; */
    thisColor->blue 	= 45535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 2965 ;	/* 3561 ; */
    thisColor->blue 	= 5535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 450 ;	/* 1348 ; */
    thisColor->green 	= 2965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 38050 ;	/* 1348 ; */
    thisColor->green 	= 32965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 41050 ;	/* 1348 ; */
    thisColor->green 	= 42965 ;	/* 3561 ; */
    thisColor->blue 	= 35535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 22050 ;	/* 1348 ; */
    thisColor->green 	= 62965 ;	/* 3561 ; */
    thisColor->blue 	= 25535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 965 ;	/* 3561 ; */
    thisColor->blue 	= 55535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 11150 ;	/* 1348 ; */
    thisColor->green 	= 32965 ;	/* 3561 ; */
    thisColor->blue 	= 44535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;

    thisColor++ ;
    thisPixel++ ;
    thisColor->pixel 	= *thisPixel ;
    thisColor->red 	= 48050 ;	/* 1348 ; */
    thisColor->green 	= 12965 ;	/* 3561 ; */
    thisColor->blue 	= 65535 ;
    thisColor->flags	= DoRed | DoGreen | DoBlue ;
}
