// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#ifndef __TCrystalWindow_h
#define __TCrystalWindow_h
#include "TButton.h"
#include "TWindow.h"
#include "TVector.h"

class   TCrystal ;

#define NUM_TOOLS	8

typedef enum { kXPointerTool=0,kXRotateTool, kXQuestionMark,
	 kXMagnifyingGlass, kXRulerTool,kXAngleTool,kXHandTool,
	 kXSizeAtomTool } KXToolType ;

typedef enum { kRunNCEMSS_Full,kRunNCEMSS_PG,kRunNCEMSS_MS,kRunNCEMSS_IM} NCEMSS_Mode ;
typedef enum { kShowNCEMSS_PG,kShowNCEMSS_MS,kShowNCEMSS_IM} NCEMSS_DisplayMode ;
typedef enum { kViewUnitCell, kViewSimulatedCell } NCEMSS_CrystalView ;

#ifndef _StructureFileType_
#define _StructureFileType_
typedef enum {kNCEMSS_FILE,kMBFIT_FILE} StructureFileType ;
#endif
class TCrystalWindow : public TWindow {

	public:
		TCrystalWindow(TCrystal *crystal,StructureFileType fileType) ;
		~TCrystalWindow() ;
		
		void	Draw() ;
		void	DoMouseDownInObject(TObject *obj,Point *pt) ;
		void	DoWMRequestDelete() ;
		void	DoMouseDownInModel(Point &loc) ;
		void	ResizeWindow(long width,long height) ;
		
		void	OnCommandHelp() { EDMHelp("TCrystalWindow") ; } ;
		
		friend	class TCrystal ;
		
	protected:

		void		SetupNCEMSSControls() ;
		void		SetupMBFITControls() ;
				
		TCrystal	*mCrystal ;
		TUserRect	*mModelFrame ;
		Rect		mCrystalRect ;
		float		mScaleFactor ;
		Boolean		mShowUCOutline ;
		
		
		TEditText	*mUField ;
		TEditText	*mVField ;
		TEditText	*mWField ;
		
		TStaticText	*mIUField ;
		TStaticText	*mIVField ;
		TStaticText	*mIWField ;
		
		TEditText	*mXCellsField ;
		TEditText	*mYCellsField ;
		TEditText	*mZCellsField ;
		
		TButton		*mEditCrystal ;
		
		TButton		*mReDisplay ;
		
		// ---------- BEGIN Controls for ncemss structure file
		
		TCrystal	*mSimulatedCrystal ;
		
		TButton		*mRunNCEMSS ;
		TButton		*mNCEMSS_ShowParams ;
		TButton		*mNCEMSS_ShowImage ;
		
		TRadioButton	*mNCEMSS_ViewNormalCell ;
		TRadioButton	*mNCEMSS_ViewSimulationCell ;
		
		TRadioButton	*mNCEMSS_RunPG ;
		TRadioButton	*mNCEMSS_RunMS ;
		TRadioButton	*mNCEMSS_RunIM ;
		TRadioButton	*mNCEMSS_RunALL ;
		
		TRadioButton	*mNCEMSS_ShowPG ;
		TRadioButton	*mNCEMSS_ShowMS ;
		TRadioButton	*mNCEMSS_ShowIM ;
		
		NCEMSS_Mode			mNCEMSS_Mode ;
		NCEMSS_DisplayMode  mNCEMSS_DisplayMode ;
		NCEMSS_CrystalView  mNCEMSS_CrystalView ;
		
		TCheckBox   *mNCEMSS_ShowResult ;
		TEditText   *mNCEMSS_nX ;
		TEditText   *mNCEMSS_nY ;
		TEditText   *mNCEMSS_nZoom ;
		
		TCheckBox   *mDisplayFourierTransform ;
		
		void		SetNCEMSSControlStates() ;
		void		SetMBFITControlStates() ;
		
		// ---------- END Controls for ncemss structure file

		// ---------- BEGIN Controls for MBFIT file
		
		TRadioButton	*mMBFIT_RunSIM ;
		TRadioButton	*mMBFIT_RunBLOCH ;
		TRadioButton	*mMBFIT_RunDS ;
		
		TButton		*mRunMBFITProgram ;
		TButton		*mMBFIT_ShowParams ;
		TButton		*mMBFIT_ShowImage ;
		
		TCheckBox   *mMBFIT_ShowResult ;
		
		// ---------- END Controls for MBFIT file
		
		StructureFileType mFileType ;

		static		short	mNumOpenCrystalWindows ;
		
		void		UpdateZoneAxis(TVector &zone) ;
		
		short		mControlAreaWidth ;

		TStaticText	*mInfo;

		void		DrawInfoWindow() ;
		void		DrawModel() ;
		void		WriteMessage(char *text) ;

		TPushButton	*mTools[NUM_TOOLS] ;
		KXToolType	mCurrentTool ;

		Str256		mMessageText ;

} ;

#endif
