#ifndef __TDiffractionDataTable_h__
#define __TDiffractionDataTable_h__
#include "TImageWindow.h"
#include "TTable.h"
#include "TFile.h"

#ifndef __SortKey
#define __SortKey
typedef enum {kSortOnG,kSortOnAmplitude} SortKey ;
#endif

typedef enum {kNoHKLError, kErrorSquareRootAmplitude, kHasHKLError } HKLErrorFlag ;
class	TRefinementSet ;
class TDiffractionDataTable : public TTable {
	public:
		TDiffractionDataTable(char *title,TDiffractionPeaks *diffractionPeaks,short planeGroup,bool friedelSymmetry) ;
		TDiffractionDataTable(FSInfo	*theFSInfo) ;
		TDiffractionDataTable(FSInfo    *theFSInfo, int ImportMode) ;
		TDiffractionDataTable(TDiffractionDataTable *s,short planeGroup,bool friedelSymmetry) ;
		~TDiffractionDataTable() ;
		
		void 	DrawCellValue(TableCellT &inCell,Rect &cellFrame) ;
		void 	DrawColLabel(Int16 colNumber) ;
		void	DrawRowLabelText(Int16 rowNumber) ;
		void	DrawColLabelText(Int16 colNumber) ;
		void 	EditCell(TableCellT &inCell) ;
		void	Save() ;
		void	SaveAsHKL() ;
		void	SortOnAmpitude() ;
		void	SortOnG() ;
		void	DoCommand(TCommand command,Boolean *hasBeenHandled) ;
		void	GetCommandStatus(TCommand,Boolean *isEnabled,Boolean *isChecked) ;
		void	GroupSymmetryRelatedReflections() ;
		void 	ClickCell(TableCellT &inCell) ;
		void	ClickColLabel(Int16 inCol) ;
		bool	CellIsEditable(TableCellT &inCell) { return (inCell.col == 4 || inCell.col == 2) ; } ;

		void	RemoveRefinementSet()  { mRefinementSet = nil ; } ;
		void	OnCommandHelp() { EDMHelp("TDiffractionDataTable") ; } ;
		
	private:
		void	GetCellText(TableCellT &inCell,char *text) ;
		void	SetCellText(TableCellT &inCell,char *text) ;
		void	DoPaste() ;
		void	DoCopy() ;
		
		friend	class	TRefinementSet ;
		TRefinementSet	*mRefinementSet ;
		
		void 	SymmetrizeHKLDataSet(TDiffractionDataTable *s,char *symm,Boolean testSigma,HKLErrorFlag errorType) ;
		
		void	CreateImageFromDiffractionData() ;
		void	SetIncludeAllReflections(bool status) ;
		void	SetUseAllPhases(bool status) ;
		long	mNumReflections ;
		float	mA ;		// Real space lattice constant
		float	mB ;		// Real space lattice constant
		float	mGamma ;	// Real space angle
		CalibrationUnit	mCalibrationUnit ;
		
		float	*mGVector ;
		float	*mAmplitude ;
		float	*mPhase ;
		short	*mH ;
		short	*mK ;
		short	*mIndex ;
		bool	*mIncludeReflection ;	// This is the same as use amplitude
		bool	*mUsePhase ;	
		short	mPlaneGroup ;
		bool	mUseAllReflections ;
		bool	mUseAllPhases ;
		bool	mAllUnknown ;
		SortKey	mSortKey ;

} ;

#endif
