#ifndef __TImageWindow
#define __TImageWindow
#include "TButton.h"
#include "TWindow.h"
#include "TImage.h"
#include "TVector.h"
#include "TLattice.h"
#include "CSelection.h"
#include "TPeakList.h" 
#include "TMask.h"
#include "TRuler.h"
#include "TDiffractionPeaks.h"

class	TImageWindow ;
class	TPeakAnnotations {
	public:
		TPeakAnnotations(TImageWindow *imageWindow,TPeakList *peakList) ;
		~TPeakAnnotations() ;
		short	mNumAnnotations ;
		TAnnotationRect	**mPeakAnnotationRects ;		
} ;

class TEDDRecord {
  public:
	TEDDRecord() ;
	TAnnotationTextRect	*g1Rect ;
	TAnnotationTextRect	*g2Rect ;
	TAnnotationTextRect	*g3Rect ;
	TVector			g1 ;
	TVector			g2 ;
	TVector			g3 ;
	Boolean			g1Defined ;
	Boolean			g2Defined ;
	Boolean			g3Defined ;
} ;

class TTEMRecord {
  public:
	TTEMRecord() ;
	TAnnotationTextRect	*g1Rect ;
	TAnnotationTextRect	*g2Rect ;
	TAnnotationCircle	*radCircle ;
	TVector			g1 ;
	TVector			g2 ;
	float			radius ;
	Boolean			g1Defined ;
	Boolean			g2Defined ;
} ;

class TImageWindow : public TWindow {
	public:
		TImageWindow(short xloc,short yloc,short width,short height,
			short depth,char *title,Boolean hasOffscreen,TImage *image = nil) ;
		~TImageWindow() ;

		virtual	void	DoCommand(TCommand, Boolean *) ;
		virtual	void	GetCommandStatus(TCommand command,Boolean *isEnabled,
							Boolean *isChecked) ;

		virtual	void	SetAsTarget(Boolean isTarget) {
								if(mCanReceiveMenuCommands) {
									mIsTarget = isTarget ;
									if (mIsTarget && mCommander != nil) 
										mCommander->ActivateListener(this) ; 
								}  
							} ;
							
		void		SetControls() {
						if(mCommander!=nil)
							mCommander->ActivateListener(this) ; } ;
			
		void		Draw() ;
		void		Close() ;		
		void		DoMouseDown(XEvent *event) ;
		void		DoKeyDown(XEvent *event) ;
		void		Update(XEvent *event = nil) ;
		void		DoWMRequestDelete() ;
		void		SetBrighness(float value) { mImage->SetBrighness(value) ; Update(nil) ; } ;
		void		SetContrast(float value) { mImage->SetContrast(value) ; Update(nil) ;  } ;
		float		GetContrastValue() { return mImage->GetContrastValue() ; } ;
		float		GetBrightnessValue() { return mImage->GetBrightnessValue() ; } ;
		void		ResizeWindow(long width,long height) ;
		
		TImage		*GetImage() { return mImage ; } ;
		
		Boolean		HasQuantifiedPeaks() { return (mDiffractionPeaks != nil) ; } ;
		Boolean 	HasSymmetrizedPeaks() { return  mHasSymmetrizedPeaks ; } ;
		
		Boolean		IsDisplayOnLogScale() { return mImage->mDisplayOnLogScale ; } ;
		void		DisplayOnLogScale(Boolean state) {  mImage->mDisplayOnLogScale = state ; mImage->Update(false) ; 
														mImage->mContrast = 1.0 ; mImage->mBrightness = 0.5 ; } ;


		TLattice	*GetTEMLattice() { return mTEMLattice ; } ;
		TLattice	*GetEDDLattice() { return mEDDLattice ; } ;
		Boolean		IsEDDLatticeVisible() { 
						return (mEDDLattice != nil && mShowEDDLattice) ; } ;
		Boolean		IsTEMLatticeVisible() { 
						return (mTEMLattice != nil && mShowTEMLattice) ; } ;
		Boolean		IsPeakAnnotationsVisisble() {
						return (mPeakList != nil && mShowPeakAnnotations) ; } ;
						
		Boolean		ImageIsRealSpace() { return mImage->IsRealSpace() ; } ;

		TPeakList	*GetPeakList() { return mPeakList ; } ;

		TEDDRecord *GetEDDRecord() { return mEDDRecord ; } ;
		TTEMRecord *GetTEMRecord() { return mTEMRecord ; } ;

		void	AddEDDLattice(TLattice *lattice) { 
					if (mEDDLattice != nil) delete mEDDLattice ;
					mEDDLattice = lattice ;  mShowEDDLattice = false ;	// Don't show yet
				} ;
		void	AddTEMLattice(TLattice *lattice) {
					if (mTEMLattice != nil) delete mTEMLattice ;
					mTEMLattice = lattice ;  mShowTEMLattice = false ;	// Don't show yet
				} ;
		void	RemoveTEMLattice() { 
					if (mTEMLattice == nil) return ; 
					delete mTEMLattice ;
					mTEMLattice = nil ; mShowTEMLattice = false ;  Update(nil) ;
				} ;
		void	RemoveEDDLattice() { 
					if (mEDDLattice == nil) return ; 
					delete mEDDLattice ;
					mEDDLattice = nil ; mShowEDDLattice = false ; Update(nil) ;
				} ;

		void	ShowEDDLattice() { if (mEDDLattice == nil) return ;
					mShowEDDLattice = true; Update(nil); } ;
		void	ShowEDDLatticeQuad(float *BX, float *BY) ;//{ if (mEDDLattice == nil) return ;
					//mShowEDDLattice = true; Update(nil); } ;
		void	HideEDDLattice() {  if (mEDDLattice == nil) return ;
					mShowEDDLattice = false; Update(nil); } ;

		void	ShowTEMLattice() { if (mTEMLattice == nil) return ;
					mShowTEMLattice = true; Update(nil); } ;
		void	HideTEMLattice() { if (mTEMLattice == nil) return ;
					mShowTEMLattice = false; Update(nil); } ;

		void	RemoveTEMAnnotations() ;	// Annotations for marking
		void	RemoveEDDAnnotations() ;	// Annotations for marking

		void	FindDiffractionPeaks() ;
		void	ShowPeakAnnotations() { if (mPeakAnnotations == nil) return ;
					mShowPeakAnnotations = true; Update(nil); } ;
		void	HidePeakAnnotations() { if (mPeakAnnotations == nil) return ;
					mShowPeakAnnotations = false; Update(nil); } ;

		void	DeleteDiffractionPeaks()  ;
		void	FindMotif() ;

		void	ExtractReflections() ;
		void	SymmetrizeReflections(short numPlaneGroup,Boolean showSymmetrizedPattern) ; 
		void	CalculateAmplitudeResidual() ;
		void	CalculateMaskedFFT() ;
		void	RefineEDDLattice() ;
		void	RefineTEMLattice() ;

		void	DrawTEMLattice() ;
		void	DrawEDDLattice() ;
		void	DrawEDDLatticeQuad(float *BX, float *BY) ;
		
		float	GetDiffractionAperture() { return mDiffractionAperture ; } ;
		float	GetCCThreshold() { return mCCThreshold ; } ;
		Boolean	GetPeakMotifVisibility(Boolean state) { return mShowPeakMotif ; } ;
		void	SetPeakMotifVisibility(Boolean state) { mShowPeakMotif = state ; } ;

		void	OnCommandHelp() { EDMHelp("TImageWindow") ; } ;

	private:
	
		void		DoHandTool(Point &location) ;
		void		DoZoom(Point &location) ;
		void		DoSelection(Point &location) ;
		void		DoRuler(Point &location) ;
		void		PasteAmplitudePhases() ;
		
		void		UpdatePeakAnnotations() ;

		void		SetCursor() ;
		void		UnsetCursor() ;
		
		TImage		*mImage ;
		TImage		*mPeakShapeImage ;
		TArray		*mPhaseResiduals[17] ;
		Point		mSymmetryOrigin[17] ;
		float		mMinimumPhaseResidual[17] ;
		TLattice	*mTEMLattice ;
		TLattice	*mEDDLattice ;
		TLattice	*mReciprocalLattice ;
		Boolean		mShowEDDLattice ;
		Boolean		mShowTEMLattice ;
		Boolean		mShowPeakAnnotations ;
		Boolean		mShowPeakMotif ;
		Boolean		mHasSymmetrizedPeaks ;
		TPeakList	*mPeakList ;
		TDiffractionPeaks	*mDiffractionPeaks ;
		TPeakAnnotations	*mPeakAnnotations ;
		TEDDRecord 	*mEDDRecord ;
		TTEMRecord *mTEMRecord ;
		CSelection	*mImageSelection ;
		float		mDiffractionAperture ;
		float		mCCThreshold ;
		TMask		*mMask ;
		TReciprocalVectors	*mReciprocalVectors ;
		TRuler		*mRuler ;

} ;

#endif
