#ifndef __TList_h
#define __TList_h
#include "TButton.h"

#include "TWindow.h"
#include "LArray.h"
#include "UMemoryMgr.h"

extern	uLong	gWhitePixel,gBlackPixel,gDarkerGray,gGrayPixel ;

#define	kDoubleClickTime	200

long	near(float) ;
typedef struct {
	short row ;
	short col ;
} ListCellT ;

class TList : public TObject {
	public:
		TList(Rect	*theListRect,short	dataSize = 256,bool hasColumnLabels = false) ;
		
		virtual			~TList() ;
		virtual void 	Draw() ;
		virtual void 	Draw(Region) ;
		
		
		ArrayIndexT		AddRow(void *item,short afterRow = -1) ;
		void			RemoveRow(short theRow) ;
		void			RemoveAllRows() { mCellData->RemoveAllItemsAfter(0) ; mRows = 0 ; Refresh() ; } ;
		
		void			Refresh()  ;
		void			SelectRow(short i) { mSelectedCell.row = i ; mSelectedCell.col = 1 ; Refresh() ; } ;
		short			GetNumberOfListItems() { return mRows ; } ;
		bool			HasSelection() { return (mSelectedCell.row > 0 && mSelectedCell.row <= mRows) ; } ;
		short			GetSelectedRow() { if (mSelectedCell.row > 0 && mSelectedCell.row <= mRows) return mSelectedCell.row ; return 0 ; } ;
		ListCellT		&GetSelectedCell() { return mSelectedCell ; } ;
		
		virtual void 	SetCellData(ListCellT &inCell,void *data) ;
		virtual bool 	FetchCellData(ListCellT &inCell,void *data) ;
		virtual void 	SetDataCells(ListCellT &inFromCell, ListCellT &inToCell,void *data) ; 
		virtual void 	FetchDataCells(ListCellT &inFromCell, ListCellT &inToCell,void *data) ; 
		virtual void 	ClickCell(ListCellT &inCell) ;
		virtual void	ClickColLabel(short inCol) ;
		virtual	Boolean	DoMouseDown(Point *) ;
		virtual	bool 	DoClickInLabels(Point &inPoint) ;
	
	protected:
	
		virtual	void 	DoClick(Point &inPoint) ;
		bool			mIsInitialized ;
	
		void			SetCanvasRect(Rect &theRect) { mCanvasRect = theRect ; } ;
		bool 			ImageRectIntersectsFrame(short left,short top,short right,short bottom) ;

		void 			ImageToLocalPoint(Point &inImagePt,Point &outLocalPt) ;
		void 			LocalToImagePoint(Point &inImagePt,Point &outLocalPt) ;
		
		virtual void	GetCellText(ListCellT &inCell,char *text) ;
		virtual void	SetCellText(ListCellT &inCell,char *text) ;

		virtual	void 	DrawCellValue(ListCellT &inCell,Rect &cellFrame) ;
		virtual void 	DrawColLabel(short colNumber) ;
		virtual void 	DrawRowLabel(short rowNumber) ;
		virtual void 	DrawCell(ListCellT &inCell) ;
		virtual void 	HighlightCell(ListCellT &inCell) ;
		virtual void	DrawRowLabelText(short rowNumber) ;
		virtual void	DrawColLabelText(short colNumber) ;
		
		void			FetchCellHitBy(const Point &inImagePt,ListCellT &outCell) ;
		bool 			IsValidCell(const ListCellT	&inCell) const ;
		bool 			FetchLocalCellFrame(const ListCellT &inCell,Rect &outCellFrame) ;

		void 			ScrollListBy(float deltaX,float deltaY) ;
		void 			ScrollListTo(float deltaX,float deltaY) ;

		double			mVScrollPos ;
		double			mHScrollPos ;
		TScrollBar		*mHorizScrollBar ;
		TScrollBar		*mVertScrollBar ;
		void			AdjustScrollBars() ;		
		
		Boolean			mHasColLabels ;
		Boolean			mHasRowLabels ;
		Boolean			mAdornRows ;
		Boolean			mAdornColumns ;
		Pattern			mAdornPattern;

		static	void	DoHScrollBar(TObject *) ;
		static	void	DoVScrollBar(TObject *) ;
		
		// these are arrays of 16 characters (max of labels)
		LArray			mRowLabels;
		LArray			mColLabels;
		
		LArray			*mCellData;
		
		short			mRows ;
		short			mCols ;
		short			mColWidth ;
		short			mRowHeight ;
		short			mCellDataSize ;
		ListCellT		mSelectedCell ;
		Rect			mCanvasRect ;
		Point			mCellClickPos ;
		FPoint			mListOrigin ;

} ;


#endif
