#ifndef __TMenu_H__
#define __TMenu_H__
#include "TWindow.h"
#include "TButton.h"

#ifndef __TCommand
#define	__TCommand
typedef unsigned long	TCommand ;
#endif

#ifndef	__TMenuKey
#define __TMenuKey
typedef KeySym	TMenuKey ;
#endif

class TMenu ;

class TMenuItem {
	public:
		TMenuItem(Rect &rect,char *title=nil,TCommand commandNum = 0,TMenuKey menuKey = 0) ;
		TMenuItem(Rect &rect,char *title,TMenu *childMenu) ;
		
		virtual	~TMenuItem() ;
		void	Draw(Boolean state) ;
		Boolean	IsEnabled() { return mEnabled ; } ;
		Boolean	IsChecked() { return mChecked ; } ;
		void	Enable(Boolean state) { mEnabled = state ; } ;
		void	Check(Boolean state) { mChecked = state ; } ;
		void	DrawCheckMark() ;
		void	DrawMenuArrow() ;
		void	SetMenuItemRect(Rect *theRect) ;
		void	AdjustWidth(short width) { mRect.right = mRect.left + width ; } ;
		friend	class	TMenu ;
		friend	class	TMenuBar ;
		friend	class	TPopupMenu ;
		friend	class	TCenteredPopupMenu ;
		friend	class TApplicationWindow ;
	private:
		Rect	mRect ;
		char	mText[256] ;
		TCommand		mCommand ;
		TMenuKey		mMenuKey ;
		TMenuItem		*mNextItem ;
		TPopUpWindow	*mWindow ;
		Boolean			mEnabled ;
		Boolean			mChecked ;
		Boolean			mIsDivider ;
		Boolean			mIsHierarchicalMenu ;
		TMenu			*mHierarchicalMenu ;
		TMenu			*mParentMenu ;
} ;


class TMenu {
	public:
		TMenu(char *title) ;
		virtual	~TMenu() ;
		void	Show() { if (mPopUpWindow != nil) { mPopUpWindow->Show(); Draw() ; } } ;
		void	Hide() { if (mPopUpWindow != nil) { mPopUpWindow->Hide() ; } } ;
		void	Draw() ;

		TCommand	GetCommand(short itemNumber)  ;
		void	AddMenuItem(char *text=nil,TCommand commandNum = 0,TMenuKey menuKey = 0) ;
		void	AddMenuItem(TMenu *hierarchicalMenu) ;
		void	RemoveMenuItem(char *text) ;
		void	AdjustMenuWindowSize() ;
		void	SetPosition(short x,short y) { mX = x ; mY = y ; mTitlePosition.x = mX + 5 ; mTitlePosition.y = mY - 4 ; } ;
		void	GetPosition(short &x,short &y) { x = mX ; y = mY ; } ;
		void	RemoveMenuItem(TCommand commandNum) ;
		void	HighlightMenuItem(short itemNumber,Boolean state) ;
		void	Enable(Boolean state) { mEnabled = state ; } ;
		void	EnableMenuItem(short itemnumber,Boolean state) ;
		void	CheckMenuItem(short itemnumber,Boolean state) ;
		void	SetMenuItemText(short itemnumber,char *text) ;
		void	GetMenuItemText(short itemnumber,char *text) ;
		void	SetParent(TPane *parent) ;
		void	SetParentMenu(TMenu *theMenu) { mParentMenu = theMenu ; } ;
		void	SetParentMenuItem(TMenuItem *theMenuItem) { mParentMenuItem = theMenuItem ; } ;
		TPane	*GetParent() { return mParentPane ; } ;
		TMenuItem	*GetMenuItem(short i) ;
		friend	class	TMenuBar ;
		friend	class TApplicationWindow ;
		
	protected:
		short	mX ;
		short	mY ;
		short	mHeightPerItem ;
		Rect	mWindowRect ;
		short	mNumMenuItems ;
		TMenuItem	*mMenuItems ;
		char	mTitle[256] ;
		Point	mTitlePosition ;
		short	mTitleWidth ;
		TPopUpWindow	*mPopUpWindow ;
		TPane	*mParentPane ;
		TMenu	*mNextMenu ;
		TMenu	*mParentMenu ;
		Boolean	mEnabled ;
		Boolean	mHighlighted ;
		TMenuItem *mParentMenuItem ;
}  ;

class	TPopupMenu : public TMenu, public TMenuButton {
	public:
		TPopupMenu(TPane *,Rect *,char *,char *label=nil,LabelPosition pos=kLabelOutsideLeft ) ;
		TPopupMenu(TPane *,short,short,short,short,char *,char *label=nil,LabelPosition pos=kLabelOutsideLeft ) ;
		TPopupMenu(TPane *,Rect *,char *text[],short choice,char *label=nil,LabelPosition pos=kLabelOutsideLeft ) ;
		TPopupMenu(TPane *,short,short,short,short,char *text[],short choice,char *label=nil,LabelPosition pos=kLabelOutsideLeft ) ;
		virtual	~TPopupMenu() ;
		short	Select(Point *where) ;
		Boolean DoMouseDown(Point *) ;
		virtual	void	DrawPopped() ;
		short	GetSelection() { return mSelectedItem ; } ;
		void	SetSelection(short item) { mSelectedItem = item ; } ;
	protected:
		short	mSelectedItem ;
		
} ;

class	TCenteredPopupMenu : public TPopupMenu {
	public:
		TCenteredPopupMenu(TPane *,Rect *,char *,char *label=nil,LabelPosition pos=kLabelOutsideLeft ) ;
		TCenteredPopupMenu(TPane *,short,short,short,short,char *,char *label=nil,LabelPosition pos=kLabelOutsideLeft ) ;
		TCenteredPopupMenu(TPane *,Rect *,char *text[],short choice,char *label=nil,LabelPosition pos=kLabelOutsideLeft ) ;
		TCenteredPopupMenu(TPane *,short,short,short,short,char *text[],short choice,char *label=nil,LabelPosition pos=kLabelOutsideLeft ) ;
		void	DrawPopped() ;
		
} ;

void	DisableMenuItem(TMenu *menu,short itemnumber) ;
void	EnableMenuItem(TMenu *menu,short itemnumber) ;
inline void EnableMenu(TMenu *menu) { menu->Enable(true) ; } 
inline void DisableMenu(TMenu *menu) { menu->Enable(false) ; }

#endif // __TMenu_H__
