// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include "TPeakList.h"
#include <math.h>

TPeak::TPeak(void) {
	pos.x = 0. ;
	pos.y = 0. ;
	value = 0. ;
	n = 0 ;
	m = 0 ;
	Mean = 0 ;
	Rfactor = 0 ;
	peakType = kFoundPeak ;
}

TPeak::TPeak(TPeak &peak) {
	pos = peak.pos ;
	value = peak.value ;
	n = peak.n ;
	m = peak.m ;
	Mean = peak.Mean ;
	Rfactor = peak.Rfactor ;
	peakType = peak.peakType ;
}
		
TPeak& TPeak::operator=(TPeak &peak) {
	pos = peak.pos ;
	n = peak.n ;
	m = peak.m ;
	value = peak.value ;
	Mean = peak.Mean ;
	Rfactor = peak.Rfactor ;
	peakType = peak.peakType ;
	return *this ;
}

TPeakEntry::TPeakEntry() {
	mPeak = nil ;
	mNextPeakEntry = nil ;
}

TPeakEntry::~TPeakEntry() {
	mNextPeakEntry = nil ;
	if (mPeak != nil)
		delete mPeak ;
}

TPeakList::TPeakList() {
	mNumPeaks = 0 ;
	mPeakEntry = nil ;
}

TPeakList::~TPeakList() {
	TPeakEntry *entry ;
	TPeakEntry *nextEntry ;
	
	// For every peak entry
	entry = mPeakEntry ;
	while(entry != nil) {
		nextEntry = entry->mNextPeakEntry ;
		delete entry ;
		entry = nextEntry ;
		mNumPeaks-- ;
	} ;
}


void TPeakList::AddPeak(TPeak &peak) {
	
	TPeakEntry *aNewPeakEntry = new TPeakEntry ;
	aNewPeakEntry->mPeak = new TPeak(peak) ;
	
	// We save time by always adding to the beginning of the list
	aNewPeakEntry->mNextPeakEntry = mPeakEntry ;
	mPeakEntry = aNewPeakEntry ;
	mNumPeaks++ ;
}

void TPeakList::AddPeak(TPeak *peak) {
	
	TPeakEntry *aNewPeakEntry = new TPeakEntry ;
	aNewPeakEntry->mPeak = peak ;
	
	// We save time by always adding to the beginning of the list
	aNewPeakEntry->mNextPeakEntry = mPeakEntry ;
	mPeakEntry = aNewPeakEntry ;
	mNumPeaks++ ;
}

void TPeakList::RemovePeak(int numberToRemove) {
	TPeakEntry* entry;
	TPeakEntry* last;
	int		currentPeakNumber ;
	
	if (numberToRemove >= mNumPeaks || numberToRemove < 0)
		return ;
	last = nil;
	currentPeakNumber = 0 ;
	for (entry = mPeakEntry ;entry != nil; entry = entry->mNextPeakEntry,currentPeakNumber++) {
	
		if ( currentPeakNumber == numberToRemove ) {// is the current peak the one to be removed?
													// yes, remove it
			// if first item in list, just set first
				if ( last == nil )				
					mPeakEntry = entry->mNextPeakEntry;
				else
					last->mNextPeakEntry = entry->mNextPeakEntry ;
			
			// free the peakEntry
				delete entry ;
				mNumPeaks-- ;
				
			// we're done; no need to continue onto the next document
				return;
		}
		else						// no, go onto the next document in the list
			last = entry;
	}
}

Boolean TPeakList::RemovePeak(Point &pos) {
	TPeakEntry* entry;
	TPeakEntry* last;
	TPeak	*currentPeak ;
	float	tolerance = 4. ;	// Within four pixels of the position
	
	last = nil;
	for (entry = mPeakEntry ;entry != nil; entry = entry->mNextPeakEntry) {
		currentPeak = entry->mPeak ;	
		if ( fabs(currentPeak->pos.x - pos.x) <= tolerance &&  fabs(currentPeak->pos.y - pos.y) <= tolerance ) {// is the current peak the one to be removed?
													// yes, remove it
			// if first item in list, just set first
				if ( last == nil )				
					mPeakEntry = entry->mNextPeakEntry;
				else
					last->mNextPeakEntry = entry->mNextPeakEntry ;
			
			// free the peakEntry
				delete entry ;
				mNumPeaks-- ;
				
			// we're done; no need to continue onto the next document
				return true;
		}
		else						// no, go onto the next document in the list
			last = entry;
	}
	return false ;
}

TPeak	*TPeakList::GetPeak(Point &pos) {
	TPeakEntry* entry;
	TPeakEntry* last;
	TPeak	*currentPeak ;
	float	tolerance = 4. ;	// Within four pixels of the position
	
	last = nil;
	for (entry = mPeakEntry ;entry != nil; entry = entry->mNextPeakEntry) {
		currentPeak = entry->mPeak ;	
		if ( fabs(currentPeak->pos.x - pos.x) <= tolerance &&  fabs(currentPeak->pos.y - pos.y) <= tolerance ) {// is the current peak the one to be returned?
				return currentPeak;
		}
		else						// no, go onto the next document in the list
			last = entry;
	}
	return nil ;

}

void TPeakList::RemovePeak(TPeak *peak) {
	TPeakEntry* entry;
	TPeakEntry* last;
	TPeak	*currentPeak ;
	
	last = nil;
	for (entry = mPeakEntry ;entry != nil; entry = entry->mNextPeakEntry) {
		currentPeak = entry->mPeak ;
		if ( currentPeak == peak ) {				// is the current peak the one to be removed?
													// yes, remove it
			// if first item in list, just set first
				if ( last == nil )				
					mPeakEntry = entry->mNextPeakEntry;
				else
					last->mNextPeakEntry = entry->mNextPeakEntry ;
			
			// free the peakEntry
				delete entry ;
				mNumPeaks-- ;
				
			// we're done; no need to continue onto the next document
				return;
		}
		else						// no, go onto the next document in the list
			last = entry;
	}
}

TPeak *TPeakList::GetPeak(int number) {
	TPeakEntry* entry;
	int i ;
	
	if (number >= mNumPeaks || number < 0)
		return nil ;
		
	for (i = 0,entry = mPeakEntry ;entry != nil; entry = entry->mNextPeakEntry,i++) {
		if (i == number) {
			return entry->mPeak ;
		}
	}
	return nil ;
}

TPeakArray::TPeakArray(TPeakList *peakList) {

	short	numActualPeaks = 0 ;
	
	TPeakEntry	*peakEntry ;
	TPeak		*peak ;
	int			i ;
	for (i = 0 , peakEntry = peakList->GetFirstPeakEntry() ;peakEntry != nil; peakEntry = peakEntry->GetNextPeakEntry(), i++) {
		peak = peakEntry->GetPeak() ;
		if (peak->peakType != kDeletedPeak) numActualPeaks++ ;
	}
	
	mNumPeaks = numActualPeaks ;
	mX = new float[mNumPeaks] ; 
	mY = new float[mNumPeaks] ; 
	mN = new short[mNumPeaks] ;
	mM = new short[mNumPeaks] ;
	mMean = new float[mNumPeaks] ;
	mRfactor = new float[mNumPeaks] ;
	mValue = new float[mNumPeaks] ; 
	
	for (i = 0 , peakEntry = peakList->GetFirstPeakEntry() ;peakEntry != nil; peakEntry = peakEntry->GetNextPeakEntry()) {
		peak = peakEntry->GetPeak() ;
		if (peak->peakType != kDeletedPeak) {
			mX[i] = peak->pos.x ;
			mY[i] = peak->pos.y ;
			mN[i] = peak->n ;
			mM[i] = peak->m ;
			mValue[i] = peak->value ;
			mMean[i] = peak->Mean ;
			mRfactor[i] = peak->Rfactor ;
			i++ ;
		}
	}
}

void TPeakArray::Sort(SortDirection direction) {
	long gap ;
	register long i,j ;
	register short itemp ;
	register float temp ;
	switch (direction) {
		case kSortAscending:
			for (gap=mNumPeaks/2;gap>0;gap/=2)
				for(i=gap;i<mNumPeaks;i++)
					for(j=i-gap;j>=0;j-=gap)
						{
							if (mValue[j] <= mValue[j+gap])
								break ;
							temp = mValue[j] ;
							mValue[j] = mValue[j+gap] ;
							mValue[j+gap] = temp ;
							
							temp = mX[j] ;
							mX[j] = mX[j+gap] ;
							mX[j+gap] = temp ;
							
							temp = mY[j] ;
							mY[j] = mY[j+gap] ;
							mY[j+gap] = temp ;
							
							itemp = mN[j] ;
							mN[j] = mN[j+gap] ;
							mN[j+gap] = itemp ;
							
							itemp = mM[j] ;
							mM[j] = mM[j+gap] ;
							mM[j+gap] = itemp ;

							temp = mMean[j] ;
							mMean[j] = mMean[j+gap] ;
							mMean[j+gap] = temp ;

							temp = mRfactor[j] ;
							mRfactor[j] = mRfactor[j+gap] ;
							mRfactor[j+gap] = temp ;

					  	}
			break ;		
		case kSortDescending:
			for (gap=mNumPeaks/2;gap>0;gap/=2)
				for(i=gap;i<mNumPeaks;i++)
					for(j=i-gap;j>=0;j-=gap)
						{
							if (mValue[j] >= mValue[j+gap])
								break ;
							temp = mValue[j] ;
							mValue[j] = mValue[j+gap] ;
							mValue[j+gap] = temp ;
							
							temp = mX[j] ;
							mX[j] = mX[j+gap] ;
							mX[j+gap] = temp ;
							
							temp = mY[j] ;
							mY[j] = mY[j+gap] ;
							mY[j+gap] = temp ;
							
							itemp = mN[j] ;
							mN[j] = mN[j+gap] ;
							mN[j+gap] = itemp ;
							
							itemp = mM[j] ;
							mM[j] = mM[j+gap] ;
							mM[j+gap] = itemp ;

							temp = mMean[j] ;
							mMean[j] = mMean[j+gap] ;
							mMean[j+gap] = temp ;

							temp = mRfactor[j] ;
							mRfactor[j] = mRfactor[j+gap] ;
							mRfactor[j+gap] = temp ;

					  	}
			break ;
	}
		
}
