#ifndef __TPotentialMapWindow_H
#define __TPotentialMapWindow_H

#include "TTable.h"
#include "TImageWindow.h"
#ifndef __SortKey
#define __SortKey
typedef enum {kSortOnG,kSortOnAmplitude} SortKey ;
#endif

#ifndef	__UnitCellDisplayRecord
#define __UnitCellDisplayRecord
typedef struct UnitCellDisplayRecord {
	short 	numCellsA ;
	short	numCellsB ;
} UnitCellDisplayRecord ;
#endif

class	TFrameRect : public TUserRect { 
	public:
		TFrameRect(Rect *rect) : TUserRect(rect) {} ;
		TFrameRect(short x1,short y1,short x2,short y2):TUserRect(x1,y1,x2,y2) {} ;
		void Draw() {
			if (!mIsVisible) return ;
			if (mParentWindow == nil) return ;
			if (!mParentWindow->IsVisible()) return ;
			FrameRect(&mRect) ;
		} ;
} ;

class TCreatePotentialMapImageDialog : public TDialog {
	public:
		TCreatePotentialMapImageDialog(UnitCellDisplayRecord *data) ;
		void	GetDialogData(void *data) { *(UnitCellDisplayRecord *)data = mData ; } ;
		void	RetrieveDialogItems() ;
		void	OnCommandHelp() { EDMHelp("TCreatePotentialMapImageDialog") ; } ;
	private:
		TEditText		*mTextNumA ;
		TEditText		*mTextNumB ;
		UnitCellDisplayRecord		mData ;		
} ;

typedef struct CTFRecord {
	float	voltage ;
	float	cs ;
	float	defocus ;
	float	delta ;
	float	div ;
} CTFRecord ;

class TPotentialDataTable : public TTable {
	public:
		TPotentialDataTable(TableAttributes *attr,TDiffractionPeaks *diffractionPeaks,short planeGroup) ;
		~TPotentialDataTable() ;
		
		void 	DrawCellValue(TableCellT &inCell,Rect &cellFrame) ;
		void 	DrawColLabel(Int16 colNumber) ;
		void	DrawColLabelText(Int16 colNumber) ;
		void	DoCommand(TCommand command,Boolean *hasBeenHandled) ;
		void	GetCommandStatus(TCommand,Boolean *isEnabled,Boolean *isChecked) ;
		void	SortOnAmpitude() ;
		void	SortOnG() ;
		void	GroupSymmetryRelatedReflections() ;
		void	ClickColLabel(Int16 inCol) ;
		void	ApplyCTFCorrections(CTFRecord &ctf,bool useScherzerLens = false,float gmax = 0.) ;
		
		void	GetCellText(TableCellT &inCell,char *text) ;
		void	Calibrate(IndexCalibrationRecord &cbr) ;
		void	GetCalibrationRecord(IndexCalibrationRecord &cbr) { cbr = mIndexCalibrationRecord ; } ;
		
		
	private:
		friend	class	TPotentialMapWindow ;
		void		DoCopy() ;
		void		DoPaste() ;
		
		TDiffractionPeaks	*mDiffractionPeaks ;
		long		mNumReflections ;
		float		*mGVector ;
		float		*mAmplitude ;
		float		*mPhase ;
		float		*mCTFCorrectedPhase ;
		short		*mH ;
		short		*mK ;
		short		*mIndex ;
		short		mPlaneGroup ;
		short		mArrayWidth ;
		short		mArrayHeight ;
		float		mGamma ;
		SortKey		mSortKey ;
		
		IndexCalibrationRecord	mIndexCalibrationRecord ;

} ;

class TPotentialMapWindow : public TWindow {
	public:
		TPotentialMapWindow(TPotentialDataTable *table) ;
		~TPotentialMapWindow() ;
		
		void	Draw() ;
		void	DoMouseDownInObject(TObject *obj,Point *pt) ;
		void	DoKeyDownInObject(TObject *obj) ;
		
		void	OnCommandHelp() { EDMHelp("TPotentialMapWindow") ; } ;
	protected:
	
		void				CreateSymmetrizedImage() ;
		void				CreatePotentialMapImage(float gmax = 0.) ;
		void				DrawSymmetrizedImage() ;
		void				DrawPotentialMapImage() ;
		void				UpdateContent() ;
		void				ExportPotentialMapImage() ;
		
		void				DoCommand(TCommand command,Boolean *hasBeenHandled) ;
		void				GetCommandStatus(TCommand,Boolean *isEnabled,
							Boolean *isChecked);
		TImage				*mSymmetrizedImage ;
		TImage				*mPotentialMapImage ;
		TFrameRect			*mImageRect ;
		TFrameRect			*mPotentialRect ;
		Rect				mActualImageRect ;
		Rect				mActualPotentialRect ;
		TTablePane			*mTablePane ;
		TPotentialDataTable	*mDataTable ;
		TRadioButton		*mIdealScherzerLens ;
		TRadioButton		*mRealLens ;		
		TEditText			*mVoltageText ;
		TEditText			*mCsText ;
		TEditText			*mDefocusText ;
		TEditText			*mDeltaText ;
		TEditText			*mDivText ;
		TButton				*mApply ;
		TButton				*mCreatePotentialMapImage ;
		CTFRecord			mCTFRecord ;

		TButton				*mCalibrateButton ;
		TEditText			*mLatticeA ;
		TEditText			*mLatticeB ;

		TEditText			*mScherzerGmaxText ;
		TButton				*mUpdateScherzerGmaxButton ;

		bool				mDataIsCalibrated ;
		
		float				*mX ;
		float				*mY ;	
		bool				mUseRealLens ;	
		bool				mSymmetrizedImageInitialized ;
		bool				mNeedToRecalculateScherzerMap ;

} ;


#endif
