#include "TRefinementSet.h"
#include "TCrystalDiffractionDataTable.h"
#include "TDiffractionDataTable.h"
#include "TRunDirectMethodsDialog.h"
#include "TFileViewDialog.h"
#include "TShowSolutionsWindow.h"

#define kUnAssignedPhase	-9999.
float ran0(int *idum) ;

TImage *CreateImageFromTiffFile(char *filename) ;
void	PutMessage(char *) ;
extern	void	AddWindow(TWindow *window) ; 
extern	class	TDMApplicationWindow *gTheControlWindow ;

extern	double	PiHalf ;
extern	double	Pi ;
extern	double	TwoPi ;

extern	bool	gDebug ; 
extern	"C" {
	extern	void equiv2(long *,long *,long *,long *,long *,long *,char *) ;
	extern	void fs2dpesub(char *filename) ;
}  

extern	char	*FS98_PATH ;
extern	char	*EDMBinDir ;

char	*sElectron = "ELECTRON" ;
char	*sXRay = "XRAY" ;
char	*sMixed = "MIXE" ;
char	*sEntropy = "ENTR" ;
char	*sEqual = "EQUA" ;
char	*sPositive = "POSI" ;
char	*sFlip = "FLIP" ;
char	*sFS2002 = "FS20" ;
char	*sOff = "OFF" ;
char	*sOn = "ON" ;
char	*sFloat	= "FLOAT" ;
char	*sVerbose = "VERB" ;
char	*sAuto = "AUTO" ;


char 	*sChemsymb[] = {"H","He","Li","Be","B","C","N","O","F","Ne","Na",
                     "Mg","Al","Si","P","S","Cl","Ar","K","Ca","Sc","Ti",
                     "V","Cr","Mn","Fe","Co","Ni","Cu","Zn","Ga","Ge","As",
                     "Se","Br","Kr","Rb","Sr","Y","Zr","Nb","Mo","Tc","Ru",
                     "Rh","Pd","Ag","Cd","In","Sn","Sb","Te","I","Xe","Cs",
                     "Ba","La","Ce","Pr","Nd","Pm","Sm","Eu","Gd","Tb","Dy",
                     "Ho","Er","Tm","Yb","Lu","Hf","Ta","W","Re","Os","Ir",
                     "Pt","Au","Hg","Tl","Pb","Bi","Po","At","Rn","Fr","Ra",
                     "Ac","Th","Pa","U"} ;

extern	char	*gPlaneGroupSymbols[]  ;

void WriteINSControlFile(TRefinementSet *refinementSet,FS98ControlRecord &theFS98ControlRecord) ;
void WriteHKLControlFile(TRefinementSet *refinementSet,FS98ControlRecord &theFS98ControlRecord) ;
void WriteEXTControlFile(TRefinementSet *refinementSet,FS98ControlRecord &theFS98ControlRecord) ;

FS98ControlRecord::FS98ControlRecord() {
	wavl	=	0.45863 ;				// Wavelength
	a		=	1.0 ;					// A
	b		=	1.0 ;					// B
	c		=	1.0 ;					// C
	al		=	90 ;					// Alpha
	be		=	90 ;					// Beta
	ga		=	90 ;					// Gamma
	symm	=	gPlaneGroupSymbols[0] ;	// symmetry group
	sfac	=	nil ;					// sfact
	unit	=	nil ;					// unit
	tfac	=	nil ;					// tfac
	aocc	=	nil ;					// aocc
	mode	=	sElectron ;				// mode 
	useElectron	=	true ;				// useElectron
	plambda		=	1.25 ;				// plambda
	lambda		=	1.5 ;				// lambda
	autoSpeed	=	false ;				// auto-speed	
	speed		=	3.0 ;				// speed
	camp		=	1. ;				// camp
	times		=	1. ;				// times
	powerOfAmplitude	= 1.0 ;			// use amplitude to power of
	width		=	1 ;					// width
	height		=	1 ;					// height
	floatBeams	=	false ;				// float beams
	niter		=	999 ;				// niter
	verbose		=	false ;				// verbose
	contentKnown	=	0 ; //false ;			// contentKnown
	numAtoms		=	0 ;				// numAtoms
	strcpy(titl,"Comment") ;			// title
	strcpy(baseName,"untitled") ;		// baseName
	showLogFile		=	false ;			// showLogFile
	showControlFile	=	false ;			// showControlFile
	cleanUpFiles	=	true ;			// cleanUpFiles
	updateINSFile	=	true ;			// Update INS FIle
	Other      =       -1 ;             // Number of lines we don't understand
	DefineOrigin = false ;				// Define the origin in divergence calculation
}

FS98ControlRecord gFS98ControlRecord  ;


float CreateRandomValue() ;
// ------------------------------------------------------
// Random value generator taken from Numerical Recipies
// ------------------------------------------------------
float CreateRandomValue() {
	float value ;
	int iseed = 3;
			
	value = ran0(&iseed) ;
    return value ;
}

float ran0(int *idum) {
	static	float  rry,rrv[98] ;
	float	dum ;
	static	int iff=0 ;
	int	j ;
	
	if (*idum < 0 || iff == 0) {	// Initializing the 98 random numbers
		iff = 1 ;
		srand(*idum) ;
		*idum=1 ;
		for (j=0;j<98;j++) dum = rand() ;
		for (j=0;j<98;j++) rrv[j] = rand() ;
		rry = rand() ;
	}
	j = (int) (97.*rry/RAND_MAX) ;
	rry = rrv[j] ;
	rrv[j] = rand() ;
	return(rry/RAND_MAX) ;
}

void WriteINSControlFile(TRefinementSet *refinementSet,FS98ControlRecord &theFS98ControlRecord) {
	char	fileName[256] ;

	// Try to write the control file 'name'.ins
	FILE 	*fp ;
	long	i,j,Other;
	
	strcpy(fileName,theFS98ControlRecord.baseName) ;
	strcat(fileName,".ins") ;
	
	fp = fopen(fileName,"w") ;
	if (fp == nil) return ;
	
	char	sfacText[256] ;
	char	unitText[256] ;
	char	tfacText[256] ;
	char	aoccText[256] ;
	char    OText[256]    ;
	
	
	fprintf(fp,"TITL\t%s\n",theFS98ControlRecord.titl) ;
	fprintf(fp,"REM\tCell size in shelx format -- not everything is used\n");
	fprintf(fp,"CELL\t%f %f %f %f %.2f %.2f %.2f\n",
		theFS98ControlRecord.wavl,theFS98ControlRecord.a,theFS98ControlRecord.b,theFS98ControlRecord.c,
		theFS98ControlRecord.al,theFS98ControlRecord.be,theFS98ControlRecord.ga) ;
        fprintf(fp,"REM\tTwo-dimensional symmetry group (lower case in general)\n");
	fprintf(fp,"SYMM\t%s\n",theFS98ControlRecord.symm) ;
	if (theFS98ControlRecord.contentKnown == 0) {
		// Set these up for the known atom types, etc.
		fprintf(fp,"REM\tKnown Atom type information \n");
		}
	else {
		if (theFS98ControlRecord.contentKnown == 1) {
			// Set these up for the known atom types, etc.
				fprintf(fp,"PUNT \n");
				fprintf(fp,"REM\tEstimated Atom type information \n");
			}
		else {
				fprintf(fp,"PUNU \n");
				fprintf(fp,"REM\tEstimated Atom type information \n");
		}
		}
	if ( theFS98ControlRecord.numAtoms == nil || theFS98ControlRecord.numAtoms < 1) {
	// Put in defaults if there is no atomic information available
		fprintf(fp,"REM\tAtom type information \n");
       	strcpy(sfacText,"Si") ;
		strcpy(unitText,"1") ;
		strcpy(tfacText,"0.25") ; //LDM
		strcpy(aoccText,"1.0") ;
		fprintf(fp,"SFAC\t%s\n",sfacText) ;
		fprintf(fp,"UNIT\t%s\n",unitText) ;
		fprintf(fp,"TFAC\t%s\n",tfacText) ;
		fprintf(fp,"AOCC\t%s\n",aoccText) ;
		fprintf(fp,"PUNT \n");
		}
	else {
		fprintf(fp,"SFAC\t") ;
		for(i = 0; i < theFS98ControlRecord.numAtoms; i++) {
			fprintf(fp,"%s\t",theFS98ControlRecord.sfac[i]) ;
		}
		fprintf(fp,"\n") ;
		fprintf(fp,"UNIT\t") ;
		for(i = 0; i < theFS98ControlRecord.numAtoms; i++) {
			fprintf(fp,"%d\t",theFS98ControlRecord.unit[i]) ;
		}
		fprintf(fp,"\n") ;
		fprintf(fp,"TFAC\t") ;
		for(i = 0; i < theFS98ControlRecord.numAtoms; i++) {
			fprintf(fp,"%.2f\t",theFS98ControlRecord.tfac[i]) ;
		}
		fprintf(fp,"\n") ;
		fprintf(fp,"AOCC\t") ;
		for(i = 0; i < theFS98ControlRecord.numAtoms; i++) {
			fprintf(fp,"%.2f\t",theFS98ControlRecord.aocc[i]) ;
		}
		fprintf(fp,"\n") ;
		}
	fprintf(fp,"REM\tIteration method used, do not change \n");
	fprintf(fp,"MODE\t%s\n",theFS98ControlRecord.mode) ;
    fprintf(fp,"REM\tType of structure factors used \n");
    if (theFS98ControlRecord.useElectron)
		fprintf(fp,"%s\n",sElectron) ;
	else
		fprintf(fp,"%s\n",sXRay) ;
    fprintf(fp,"REM\tNote, positivity term is not obsolete \n");
	fprintf(fp,"PLAMBDA\t%.2f\n",theFS98ControlRecord.plambda) ;
	fprintf(fp,"REM\tOver-relaxation \n");
	if (theFS98ControlRecord.autoSpeed) 
		fprintf(fp,"LAMBDA\t%.2f\t%s\n",theFS98ControlRecord.lambda,sAuto) ;
	else
		fprintf(fp,"LAMBDA\t%.2f\t%f\n",theFS98ControlRecord.lambda,theFS98ControlRecord.speed) ;
    fprintf(fp,"REM\tParameter to adjust number of beams in divergence calculation \n");
	fprintf(fp,"CAMP\t%.2f\n",theFS98ControlRecord.camp) ;
	fprintf(fp,"REM\tImage output control parameters \n");
	fprintf(fp,"TIMES\t%.2f\n",theFS98ControlRecord.times) ;
	fprintf(fp,"WIDTH\t%d\n",theFS98ControlRecord.width) ;
	fprintf(fp,"HEIGHT\t%d\n",theFS98ControlRecord.height) ;
	fprintf(fp,"REM\tThis term controls if F's or I's are used -- be careful \n");
	fprintf(fp,"POWER\t%.2f\n",theFS98ControlRecord.powerOfAmplitude) ;
	if (theFS98ControlRecord.floatBeams) 
		fprintf(fp,"%s\n",sFloat) ;
	if (theFS98ControlRecord.verbose) 
		fprintf(fp,"%s\n",sVerbose) ;
	if (theFS98ControlRecord.DefineOrigin)
		fprintf(fp,"ORIG\n") ;
        //Stuff EDM does not understand
    fprintf(fp,"REM\tOther content of ins file \n");
    //fprintf(stderr,"Check %d \n", theFS98ControlRecord.Other);
    Other =  theFS98ControlRecord.Other + 1 ;
    for(i=0; i < Other; i++) {
                for (j=0; j < 256 ; j++) {
                    OText[j] = theFS98ControlRecord.OtherLines[i][j] ;
                }
                fprintf(fp,"%s\n",OText);
                //fprintf(stderr,"%s\n",OText);
	}
	fprintf(fp,"END\n") ;
	fflush(fp) ;
	fclose(fp) ;
	
}

void WriteHKLControlFile(TRefinementSet *refinementSet,FS98ControlRecord &theFS98ControlRecord) {
	char	fileName[256] ;

	FILE 	*fp ;
	long	i ;
	
	// Write the control file 'name'.hkl
	strcpy(fileName,theFS98ControlRecord.baseName) ;
	strcat(fileName,".hkl") ;
	fp = fopen(fileName,"w") ;
	if (fp == nil) return ;
	
	float   UsedPhase ;
	
	for (i = 0 ;  i < refinementSet->mNumIndependentReflections; i++) {
		//LDM Trap phase here ?
		UsedPhase = 0 ;
		if(refinementSet->mRefinementReflection[i].mUsePhase) {
			UsedPhase = refinementSet->mRefinementReflection[i].mPhase[0] ;
		} 
		fprintf(fp,"%d\t%d\t0\t%f\t%f\n",refinementSet->mRefinementReflection[i].mH[0],refinementSet->mRefinementReflection[i].mK[0],
					refinementSet->mRefinementReflection[i].mAmplitude,UsedPhase) ;		
	}	
	fflush(fp) ;
	fclose(fp) ;

}

void WriteEXTControlFile(TRefinementSet *refinementSet,FS98ControlRecord &theFS98ControlRecord) {
	char	fileName[256] ;

	FILE 	*fp ;
	long	i ;
	// Update/Write the control file .ext
	strcpy(fileName,theFS98ControlRecord.baseName) ;
	strcat(fileName,".ext") ;
	fp = fopen(fileName,"w") ;
	
	if (fp == nil) return ;
	
	float   UsedPhase ;
	for (i = 0 ;  i < refinementSet->mNumIndependentReflections; i++) {
		//LDM Trap phase here ?
		UsedPhase = 0 ;
		if(refinementSet->mRefinementReflection[i].mUsePhase) {
			UsedPhase = refinementSet->mRefinementReflection[i].mPhase[0] ;
		} 
		fprintf(fp,"%d\t%d\t0\t%f\t%f\n",refinementSet->mRefinementReflection[i].mH[0],refinementSet->mRefinementReflection[i].mK[0],
					refinementSet->mRefinementReflection[i].mAmplitude,UsedPhase) ;	
	}		
	fflush(fp) ;
	fclose(fp) ;
}

TRefinementSet::TRefinementSet(TCrystalDiffractionDataTable *diffractionDataTable) {
	long 	i,j ;
	short	index,hMax,kMax,ndx ;
		
	mDataTable = diffractionDataTable ;
	mTableType = kCrystalDiffractionDataTable ;
	
	mNumReflections = diffractionDataTable->mNumReflections ;
	index = -1 ;
	mNumIndependentReflections = 0 ;
	for (i=0; i < mNumReflections;i++) {
		if (index != diffractionDataTable->mIndex[i]) {
			mNumIndependentReflections++ ;
			index = diffractionDataTable->mIndex[i] ;
		}
	}
	
	
	mRefinementReflection = nil ;	
	if (mNumIndependentReflections == 0) return ;
	
	mPlaneGroup = diffractionDataTable->mPlaneGroup ;
	mA	= diffractionDataTable->mA ;
	mB	= diffractionDataTable->mB ;
	mGamma	= diffractionDataTable->mGamma ;
	mCalibrationUnit	= diffractionDataTable->mCalibrationUnit ;
	
	mRefinementReflection = new TRefinementReflection[mNumIndependentReflections] ;
	
	index = -1 ;
	j = -1 ;
	double	degreeToRad = asin(1.0)/90. ;
	float	amplitude ;
	double	FSquareSum = 0. ;
	
	mUmax = 0. ;
	// The reflections are only defined on the upper half-plane
	// We won't convert to radians as of yet
	for (i=0; i < mNumReflections;i++) {
		amplitude = diffractionDataTable->mAmplitude[i] ;
		if (index != diffractionDataTable->mIndex[i]) {
			j++ ;
			mRefinementReflection[j].mAmplitude = amplitude ;
			//LDM
			mRefinementReflection[j].mUsePhase =  diffractionDataTable->mPhase[i] <= -9000. ? false : true ;
			//if (!diffractionDataTable->mUsePhase[i]) mRefinementReflection[j].mUsePhase = false;
			if (mRefinementReflection[j].mUsePhase) {
				mRefinementReflection[j].mPhase[0] = diffractionDataTable->mPhase[i] ;
				if (mRefinementReflection[j].mPhase[0] < 0.) mRefinementReflection[j].mPhase[0] += 360. ;	
				if (mRefinementReflection[j].mPhase[0] == 0.) mRefinementReflection[j].mPhase[0] = 360. ;	// To differentiate from unknown
			}
			else
				mRefinementReflection[j].mPhase[0] = 0. ;	// Unknown
				
			mRefinementReflection[j].mIncludeInRefinement = true ;
			mRefinementReflection[j].mH[0] = diffractionDataTable->mH[i] ;
			mRefinementReflection[j].mK[0] = diffractionDataTable->mK[i] ;
			mRefinementReflection[j].mNumEquivalentReflections = 1 ;
			index = diffractionDataTable->mIndex[i] ;
			FSquareSum += 2*amplitude*amplitude ;
			if (diffractionDataTable->mGVector[i] > mUmax)
				mUmax = diffractionDataTable->mGVector[i] ;
		}
		else {
			ndx = mRefinementReflection[j].mNumEquivalentReflections ;
			mRefinementReflection[j].mH[ndx] = diffractionDataTable->mH[i] ;
			mRefinementReflection[j].mK[ndx] = diffractionDataTable->mK[i] ;
			//LDM
			mRefinementReflection[j].mUsePhase =  diffractionDataTable->mPhase[i] <= -9000. ? false : true ;
			//if (!diffractionDataTable->mUsePhase[i]) mRefinementReflection[j].mUsePhase = false;
			if (mRefinementReflection[j].mUsePhase) {
				mRefinementReflection[j].mPhase[ndx] = diffractionDataTable->mPhase[i] ;
				if (mRefinementReflection[j].mPhase[ndx] < 0.) mRefinementReflection[j].mPhase[ndx] += 360. ;	
				if (mRefinementReflection[j].mPhase[ndx] == 0.) mRefinementReflection[j].mPhase[ndx] = 360. ;	// To differentiate from unknown
			}
			else 
				mRefinementReflection[j].mPhase[ndx] = 0. ;
				
			FSquareSum += 2*amplitude*amplitude ;
			mRefinementReflection[j].mNumEquivalentReflections++ ;
		}
	}
	
	
	// Let us calculate dimensions of the real space array to go out to a resolution of 4 pixels/ngstrm to assure 
	// appropriate seperation of atomic columns
	float	gMax = 4. ;
	if (mCalibrationUnit == kInverseAngstrom) {
		mImageWidth 	= near(2*gMax*mA) ;	
		mImageHeight	= near(2*gMax*mB) ;	
	}
	else {
		// The data is not calibrated so we don't really know where to set the cutoff
		// except for using some multiple of the max u-vector of the data
		// We will go to 4 times the max of h and k
		hMax = mUmax*mA/sin(degreeToRad*(180-mGamma)) ;
		kMax = mUmax*mB/sin(degreeToRad*(180-mGamma)) ;
		mImageWidth = 2*4*hMax ;
		mImageHeight	= 2*4*kMax ;
	}
	
	// Let us normalize the reflection data
	// We won't normalize the data set as of yet either
	/*
	double	FSum = sqrt(FSquareSum) ;
	for (i=0; i < mNumIndependentReflections;i++) {
		mRefinementReflection[i].mAmplitude /= FSum ; 
	}
	*/
}

TRefinementSet::TRefinementSet(TDiffractionDataTable *diffractionDataTable) {
	long 	i,j ;
	short	index,h0,k0,h,k,ndx ;
	short	hMax,kMax ;

	mDataTable = diffractionDataTable ;
	mTableType = kDiffractionDataTable ;
	
	mNumReflections = diffractionDataTable->mNumReflections ;
	index = -1 ;
	mNumIndependentReflections = 0 ;
	for (i=0; i < mNumReflections;i++) {
		if (diffractionDataTable->mIncludeReflection[i])
			if (index != diffractionDataTable->mIndex[i]) {
				mNumIndependentReflections++ ;
				index = diffractionDataTable->mIndex[i] ;
			}
	}
	
	
	mRefinementReflection = nil ;	
	if (mNumIndependentReflections == 0) return ;
	
	mPlaneGroup = diffractionDataTable->mPlaneGroup ;
	mA	= diffractionDataTable->mA ;
	mB	= diffractionDataTable->mB ;
	mGamma	= diffractionDataTable->mGamma ;
	mCalibrationUnit	= diffractionDataTable->mCalibrationUnit ;
	
	if (mCalibrationUnit == kPixel) mCalibrationUnit = kInversePixel ;
	
	mRefinementReflection = new TRefinementReflection[mNumIndependentReflections] ;
	
	index = -1 ;
	j = -1 ;
	double	degreeToRad = asin(1.0)/90. ;
	mUmax = 0. ;
	float	amplitude ;
	double	FSquareSum = 0. ;
	
	// The reflections are only defined on the upper half-plane
	// Leave phases in degrees
	for (i=0; i < mNumReflections;i++) {
		amplitude = diffractionDataTable->mAmplitude[i] ;
		if (diffractionDataTable->mIncludeReflection[i])
			if (index != diffractionDataTable->mIndex[i]) {
				j++ ;
				mRefinementReflection[j].mAmplitude = amplitude ;
				mRefinementReflection[j].mUsePhase =  diffractionDataTable->mPhase[i] <= -9000. ? false : true ; //LDM
			if (!diffractionDataTable->mUsePhase[i]) mRefinementReflection[j].mUsePhase = false;
				if (mRefinementReflection[j].mUsePhase) {
					mRefinementReflection[j].mPhase[0] = diffractionDataTable->mPhase[i] ;
					if (mRefinementReflection[j].mPhase[0] < 0.) mRefinementReflection[j].mPhase[0] += 360. ;	
					if (mRefinementReflection[j].mPhase[0] == 0.) mRefinementReflection[j].mPhase[0] = 360. ;	// To differentiate from unknown
				}
				else {
					mRefinementReflection[j].mPhase[0] = 0. ;
				}
				mRefinementReflection[j].mIncludeInRefinement = diffractionDataTable->mIncludeReflection[i] ;
				
				h0 = diffractionDataTable->mH[i] ;
				k0 = diffractionDataTable->mK[i] ;
				
				mRefinementReflection[j].mH[0] = diffractionDataTable->mH[i] ;
				mRefinementReflection[j].mK[0] = diffractionDataTable->mK[i] ;
				mRefinementReflection[j].mPhaseFlag[0] = 1 ;
				mRefinementReflection[j].mNumEquivalentReflections = 1 ;
				index = diffractionDataTable->mIndex[i] ;
				FSquareSum += 2*amplitude*amplitude ;
				if (diffractionDataTable->mGVector[i] > mUmax)
					mUmax = diffractionDataTable->mGVector[i] ;
			}
			else {
				h = diffractionDataTable->mH[i] ;
				k = diffractionDataTable->mK[i] ;
				ndx = mRefinementReflection[j].mNumEquivalentReflections ;
				mRefinementReflection[j].mH[ndx] = h ;
				mRefinementReflection[j].mK[ndx] = k ;
				mRefinementReflection[j].mPhaseFlag[ndx] = SetSymmetryRelatedPhaseFlag(h0,k0,h,k) ;
				
				if (!diffractionDataTable->mUsePhase[i]) mRefinementReflection[j].mUsePhase = false;
				mRefinementReflection[j].mPhase[ndx] = diffractionDataTable->mPhase[i] ;
				if (mRefinementReflection[j].mUsePhase) {
					if (mRefinementReflection[j].mPhase[ndx] < 0.) mRefinementReflection[j].mPhase[ndx] += 360. ;	
					if (mRefinementReflection[j].mPhase[ndx] == 0.) mRefinementReflection[j].mPhase[ndx] = 360. ;	// To differentiate from unknown
				}
				else {
					mRefinementReflection[j].mPhase[ndx] = 0. ;	
				}
				FSquareSum += 2*amplitude*amplitude ;
				mRefinementReflection[j].mNumEquivalentReflections++ ;
			}
	}
	
	// -------------- Code to test the flags for the phase relationship for symmetry related reflections
	/*
	fprintf(stderr,"\n") ;
	long	ih,ik,he[12],ke[12],ise[12],ne ;
	char	symm[4] ;
	for (i = 0 ;  i < mNumIndependentReflections; i++) {
		fprintf(stderr,"Reflection: [%d,%d]'s equivalent reflections:\n",mRefinementReflection[i].mH[0],mRefinementReflection[i].mK[0]) ;
		for(j=1; j < mRefinementReflection[i].mNumEquivalentReflections; j++) {
			fprintf(stderr,"\t[%d,%d]   Flag: %d\n",mRefinementReflection[i].mH[j],mRefinementReflection[i].mK[j],mRefinementReflection[i].mPhaseFlag[j]) ;
		}
		ih = mRefinementReflection[i].mH[0] ;
		ik = mRefinementReflection[i].mK[0] ;
		strncpy(symm,gPlaneGroupSymbols[mPlaneGroup-1],4) ;
		equiv2(he,ke,ise,&ih,&ik,&ne,symm) ;
		fprintf(stderr,"EQUIV: Reflection: [%d,%d]'s equivalent reflections:\n",ih,ik) ;
		for(j=0; j < ne; j++) {
			fprintf(stderr,"\t[%d,%d]   Flag: %d\n",he[j],ke[j],ise[j]) ;
		}
		
	}
	*/
	// -------------- End Code to test the flags for the phase relationship for symmetry related reflections
	
	// Let us calculate dimensions of the real space array to go out to a resolution of 4 pixels/ngstrm to assure 
	// appropriate seperation of atomic columns
	float	gMax = 4. ;
	if (mCalibrationUnit == kInverseAngstrom) {
		mImageWidth 	= near(2*gMax*mA) ;	
		mImageHeight	= near(2*gMax*mB) ;	
	}
	else {
		// The data is not calibrated so we don't really know where to set the cutoff
		// except for using some multiple of the max u-vector of the data
		// We will go to 4 times the max of h and k
		hMax = mUmax*mA/sin(degreeToRad*(180-mGamma)) ;
		kMax = mUmax*mB/sin(degreeToRad*(180-mGamma)) ;
		mImageWidth = 2*4*hMax ;
		mImageHeight	= 2*4*kMax ;
	}
	
	// Let us normalize the reflection data
	// Leave out normalization for now
	//double	FSum = sqrt(FSquareSum) ;
	//for (i=0; i < mNumIndependentReflections;i++) {
	//	mRefinementReflection[i].mAmplitude /= FSum ; 
	//}
}

// --------------------------------------------------------------------------
// return the flag showing how the phase of h1,k1 
// is symmetry related to the phase of reflection h,k
// --------------------------------------------------------------------------
long	TRefinementSet::SetSymmetryRelatedPhaseFlag(short h,short k,short h1,short k1 ) {
	long	phase = 1 ;
	// Flags are:
	//	0		:	Unrelated
	//	1		:	Same
	//	-1		:	Complex Conjugate (either because (-h,-k) or because of symmetry
	//	180		:	phase + 180
	//	-180	:	-phase + 180
	
	// We'll use 0 for unrelated
	switch (mPlaneGroup) {
		case 1:
			// Random phases 0 - 2Pi
			return 0 ;	// unrelated
			break ;
		case 2:
			// Phase 0 or 180  : (h,k)
			return 0 ;	// unrelated
			break ;
		case 3:
			// P(h,k) = P(-h,k) = -P(h,-k)
			if (h1 == -h && k1 == k) return phase ;
			return -phase ;
			break ;
		case 4:
			// P(h,k) = P(-h,k)+k*180
			if (h1 == -h && k1 == k) {
				if (k1%2 == 0)
					return phase ;
				else 
					return 180 ;
			}
			if (h1 == h && k1 == -k) {
				if (k1%2 == 0)
					return -phase ;
				else 
					return -180 ;
			}
			return -phase ;
			break ;
		case 5:
			// P(h,k) = P(-h,k) = -P(h,-k)
			if (h1 == -h && k1 == k) return phase ;
			return -phase ;
			break ;
		case 6:
			// P(h,k) = P(-h,k) = -P(h,-k)
			if (h1 == -h && k1 == k) return phase ;
			return -phase ;
			break ;
		case 7:
			// P(h,k) = P(-h,k)+k*180
			if (h1 == -h && k1 == k) {
				if (k%2 == 0)
					return phase ;
				else 
					return 180 ;
			}
			if (h1 == h && k1 == -k) {
				if (k%2 == 0)
					return -phase ;
				else 
					return -180 ;
			}
			return -phase ;
			break ;
		case 8:
			// P(h,k) = P(-h,k)+(h+k)*180
			if (h1 == -h && k1 == k) {
				if ((h+k)%2 == 0)
					return phase ;
				else 
					return 180 ;
			}
			if (h1 == -h && k1 == k) {
				if ((h+k)%2 == 0)
					return -phase ;
				else 
					return -180 ;
			}
			return -phase ;
			break ;
		case 9:
			// P(h,k) = P(-h,k) = -P(h,-k)
			if (h1 == -h && k1 == k) return phase ;
			return -phase ;
			break ;
		case 10:
			// P(h,k) = P(-k,h) = -P(k,-h)
			if (h1 == -k && k1 == h) return phase ;
			return -phase ;
			break ;
		case 11:
			// P(h,k) = P(-k,h) = P(-h,k) = -P(k,-h) = -P(h,-k)
			if (h1 == -k && k1 == h) return phase ;
			if (h1 == -h && k1 == k) return phase ;
			return -phase ;
			break ;
		case 12:
			// P(h,k) = P(-k,h) = P(-h,k)+k*180
			if (h1 == -h && k1 == k) {
				if (k%2 == 0)
					return phase ;
				else 
					return 180 ;
			}
			if (h1 == h && k1 == -k) {
				if (k%2 == 0)
					return -phase ;
				else 
					return -180 ;
			}
			if (h1 == -k && k1 == h) return phase ;
			return -phase ;
			break ;
		case 13:
			// P(h,k) = P(k,-h-k) = P(-h-k,h)
			if (h1 == k && k1 == -h-k) return phase ;
			if (h1 == -h-k && k1 == h) return phase ;
			return -phase ;
			break ;
		case 14:
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(-k,-h)
			if (h1 == k && k1 == -h-k) return phase ;
			if (h1 == -h-k && k1 == h) return phase ;
			if (h1 == -k && k1 == -h) return phase ;
			return -phase ;
			break ;
		case 15:
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(k,h)
			if (h1 == k && k1 == -h-k) return phase ;
			if (h1 == -h-k && k1 == h) return phase ;
			if (h1 == k && k1 == h) return phase ;
			return -phase ;
			break ;
		case 16:
			// P(h,k) = P(k,-h-k) = P(-h-k,h)
			if (h1 == k && k1 == -h-k) return phase ;
			if (h1 == -h-k && k1 == h) return phase ;
			return -phase ;
			break ;
		case 17:
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(k,h)
			if (h1 == k && k1 == -h-k) return phase ;
			if (h1 == -h-k && k1 == h) return phase ;
			if (h1 == k && k1 == h) return phase ;
			return -phase ;
			break ;
	}
	return phase ;
}


TRefinementSet::TRefinementSet(TRefinementSet *set) {

	mDataTable = nil ;

	this->mPlaneGroup =  set->mPlaneGroup ;
	this->mA =  set->mA ;
	this->mB =  set->mB ;
	this->mGamma =  set->mGamma ;
	this->mUmax =  set->mUmax ;
		
	this->mNumReflections =  set->mNumReflections ;
	this->mNumIndependentReflections =  set->mNumIndependentReflections ;
	
	this->mImageWidth = set->mImageWidth ;
	this->mImageHeight = set->mImageHeight ;
	this->mCalibrationUnit = set->mCalibrationUnit ;	
	
	this->mRefinementReflection = new TRefinementReflection[this->mNumIndependentReflections] ;
	for(long i = 0 ; i < this->mNumIndependentReflections; i++) 
		this->mRefinementReflection[i] = set->mRefinementReflection[i] ;
}

void	TRefinementSet::RunFS98() {
	char	*cp,fileName[256] ;
	
	// Get the lattice parameters from the dataset as well as the angle	
	if (mCalibrationUnit == kInverseAngstrom) {
		gFS98ControlRecord.a = mA ;
		gFS98ControlRecord.b = mB ;
	}
	else if (mCalibrationUnit == kInverseNanometer) {
		gFS98ControlRecord.a = 0.1*mA ;
		gFS98ControlRecord.b = 0.1*mB ;
	}
	else if (mCalibrationUnit == kInversePixel) {
		gFS98ControlRecord.a = 1.0/mA ;
		gFS98ControlRecord.b = 1.0/mB ;
	}
	gFS98ControlRecord.ga = mGamma ;
	

	// Create the base name
	mDataTable->GetTitle(fileName) ;
	cp = strrchr(fileName,'.') ;
	if (cp != NULL) *cp = '\0' ;
	// FInd any empty space
	if ((cp = strrchr(fileName,' ')) != NULL) {
		strcpy(gFS98ControlRecord.baseName,cp+1) ;
	}
	else {
		strcpy(gFS98ControlRecord.baseName,fileName) ;
	}
	
	// we only allow one refinement dialog to be active at one time
	// We could delete the old one and create a new one
	// Currently the variable gRunDirectMethodsDialog gets set in the contructor and 'nilled' in 
	// the destructor. The reason we don't have multiple dialogs and multiple refinement sets is that there
	// are only one set of names for files such as 'control', etc. and tehre would be a complete mess if we 
	// could run several refinements at the same time
	// Thus we could have
	// -----------------------------------------------------------------------------------------------------------
	//	if (gRunDirectMethodsDialog != nil) {
	//		delete gRunDirectMethodsDialog ;
	//		gRunDirectMethodsDialog = nil ;
	//	}
	//	TRunDirectMethodsDialog	*theDialog = new TRunDirectMethodsDialog(this,gFS98ControlRecord,mPlaneGroup) ;
	//	theDialog->SetCommander((TCommander *)gTheControlWindow) ;		
	//	AddWindow(theDialog) ; 
	// -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
	// If we did that, then the code in 'TDiffractionDataTable.cc ; TDiffractionDataTable::GetCommandStatus' would be
	//	case kRunFeasSet:
	//		// we only allow one refinement dialog to be active at one time
	//		// We could delete the old one and create a new one
	//		*isEnabled = mRefinementSet == nil ;
	//		break ;	
	// -----------------------------------------------------------------------------------------------------------
	
	TRunDirectMethodsDialog	*theDialog = new TRunDirectMethodsDialog(this,gFS98ControlRecord,mPlaneGroup) ;
	theDialog->SetCommander((TCommander *)gTheControlWindow) ;		
	AddWindow(theDialog) ; 
	
}

void	TRefinementSet::RunPhaseExtension(FS98ControlRecord &theFS98ControlRecord) {

	char	text[256],fileName[256] ;
	
	// Run the shell commands to execute the refinement
	sprintf(text,"%s%sfs2dpe %s",FS98_PATH,EDMBinDir,theFS98ControlRecord.baseName) ;
	fprintf(stderr,"\nRunning the command %s\n",text) ; fflush(stderr) ;
	system(text) ;
	
	// Open the solution if it exists and show it
	TImage	*solution = CreateImageFromTiffFile("EDMs.tif") ;
	if (solution != nil) {
		solution->SetName("EDMs.tif") ;
		solution->Display() ;
	}
	else {
		PutMessage("Failed to open the file 'EDMs.tif'") ;
	}
	
	TFileViewDialog *fileViewDialog ;
	if (theFS98ControlRecord.showLogFile) {
		strcpy(fileName,theFS98ControlRecord.baseName) ;
		strcat(fileName,".lst") ;

		fileViewDialog = new TFileViewDialog(fileName) ;

		if (!fileViewDialog->IsValid()) {
			delete fileViewDialog ;
			char otext[256] ;
			sprintf(otext,"Failed to load the file %s",fileName) ;
			PutMessage(otext) ;
		}
	}
	
}

void	TRefinementSet::RunConvergence(FS98ControlRecord &theFS98ControlRecord) {
	char	text[256],fileName[256] ;
	int		numBeams ;

	/* -------- Test ---------
	char	*path = getenv("PATH") ;
	char	*home = getenv("HOME") ;
	fprintf(stderr,"\nPATH = %s",path) ;
	fprintf(stderr,"\nHOME = %s",home) ;
	-------- End Test --------- */

	// Run the shell script to execute the refinement
	sprintf(text,"%s%sdiv2D.sh %s",FS98_PATH,EDMBinDir,theFS98ControlRecord.baseName) ;
	fprintf(stderr,"\nRunning the command %s\n",text) ; fflush(stderr) ;
	system(text) ;

	char otext[256] ;
	TFileViewDialog *fileViewDialog ;
	if (theFS98ControlRecord.showLogFile) {

		strcpy(fileName,"Divergence.log") ;
		fileViewDialog = new TFileViewDialog("Divergence.log") ;

		if (!fileViewDialog->IsValid()) {
			delete fileViewDialog ;
			sprintf(otext,"Failed to load the file %s",fileName) ;
			PutMessage(otext) ;
		}
	}

	if (theFS98ControlRecord.showControlFile) {
		strcpy(fileName,"control") ;
		fileViewDialog = new TFileViewDialog(fileName) ;

		if (!fileViewDialog->IsValid()) {
			delete fileViewDialog ;
			sprintf(otext,"Failed to load the file %s",fileName) ;
			PutMessage(otext) ;
		}
	}
	else {
		// Let us show how many beams have been set
		strcpy(fileName,"control") ;
		FILE *fp = fopen(fileName,"r") ;
		if (fp != nil) {
			fscanf(fp,"%d",&numBeams) ;
			sprintf(text,"Number of beams set to vary: %d",numBeams) ;
			PutMessage(text) ;
			fclose(fp) ;
		}
	}

}

void	TRefinementSet::RunGenetic(FS98ControlRecord &theFS98ControlRecord) {
	char	text[256],fileName[256] ;

	if (theFS98ControlRecord.cleanUpFiles) {
		// Run the shell script to clean up old files
		sprintf(text,"%s%sfsclean.sh %s",FS98_PATH,EDMBinDir,theFS98ControlRecord.baseName) ;
		fprintf(stderr,"\nRunning the command %s\n",text) ; fflush(stderr) ;
		system(text) ;
	} 
	
	// Run the shell script to execute the refinement
	sprintf(text,"%s%sgenal.shell2D.sh %s",FS98_PATH,EDMBinDir,theFS98ControlRecord.baseName) ;
	fprintf(stderr,"\nRunning the command %s\n",text) ; fflush(stderr) ;
	
	system(text) ;
	TFileViewDialog *fileViewDialog ;

	if (theFS98ControlRecord.showLogFile) {
		strcpy(fileName,"genetic_") ;
		strcat(fileName,theFS98ControlRecord.baseName) ;
		strcat(fileName,".log") ;

		fileViewDialog = new TFileViewDialog(fileName) ;

		if (!fileViewDialog->IsValid()) {
			delete fileViewDialog ;
			char otext[256] ;
			sprintf(otext,"Failed to load the file %s",fileName) ;
			PutMessage(otext) ;
		}
	}
}

void	TRefinementSet::ShowSolutions(FS98ControlRecord &theFS98ControlRecord) {
	
	// We need to create the ShowSolutionsWindow
	TShowSolutionsWindow *theShowSolutionsWindow = new TShowSolutionsWindow(theFS98ControlRecord) ;
	theShowSolutionsWindow->SetCommander((TCommander *)gTheControlWindow) ;		
	AddWindow(theShowSolutionsWindow) ; 

}


TArray	*TRefinementSet::RunGerschbergSaxtonIteration() {
	
	TDMIteration	*iteration = new TDMIteration(this) ;
	short	numCycles = 10 ;

	// I will test the loop for 10 cycles
	for(int i = 0; i < numCycles; i++) {
		if (i == numCycles-1) gDebug = true ;
		iteration->IterationCycle() ;
		if (i == numCycles-1) gDebug = false ;
	}
	// At the end of the cycle we are in reciprocal space
	iteration->InverseFourierTransform() ;
	TArray *theArray = new TArray(iteration->mImageArray) ;
	float	normalization = theArray->GetWidth()*theArray->GetHeight() ;
	*theArray *= normalization ;
	delete iteration ;

	return theArray ;
}

TRefinementSet::~TRefinementSet() {
	if (mRefinementReflection != nil) delete[] mRefinementReflection ;
	
	if (mDataTable != nil) {
		switch(mTableType) {
			case kCrystalDiffractionDataTable:
				((TCrystalDiffractionDataTable *)mDataTable)->RemoveRefinementSet() ;
				break ;
			case kDiffractionDataTable:
				((TDiffractionDataTable *)mDataTable)->RemoveRefinementSet() ;
				break ;
		}
	}

}

TDMIteration::TDMIteration(TRefinementSet *refinementSet) {

	mRefinementSet = new TRefinementSet(refinementSet) ;
	mImageArray = new TArray(mRefinementSet->mImageWidth,mRefinementSet->mImageHeight,kComplexArray) ;
	mReflectionIndex = new TArray(mRefinementSet->mImageWidth,mRefinementSet->mImageHeight) ;
	AssignAmplitudesAndRandomPhases() ;					// Create the starting set of known amplitudes and (partially) random phases
}

TDMIteration::~TDMIteration() {
	delete mRefinementSet ;
	if (mImageArray != nil) delete mImageArray ;
	if (mReflectionIndex != nil) delete mReflectionIndex ;
}

// --------------------------------------------------------
// Assign initial phases to the reflection set.
// If the reflection has a fixed("known")  phase, keep it
// Otherwise create a random phase consistent with the 
// plane group and assign symmetry related reflections
// accordingly
// --------------------------------------------------------
void	TDMIteration::AssignAmplitudesAndRandomPhases() {
	register long i,j ;
	long	h,k ;
	float	amp,phase,rl,im ;
	long	numIndependentReflections = mRefinementSet->mNumIndependentReflections ;
	TRefinementReflection	*refinementReflection = mRefinementSet->mRefinementReflection ;

	// This will preserve the phases that are fixed
	for (i = 0 ;  i < numIndependentReflections; i++) {
		if (!refinementReflection[i].mUsePhase) {
			refinementReflection[i].mPhase[0] = RandomPhase(refinementReflection[i].mH[0],refinementReflection[i].mK[0]) ; 
			for(j=1; j < refinementReflection[i].mNumEquivalentReflections; j++) {
				phase = SymmetryRelatedPhase(refinementReflection[i].mPhase[0],refinementReflection[i].mH[0],refinementReflection[i].mK[0],
						refinementReflection[i].mH[j],refinementReflection[i].mK[j]) ;
			}
		}
	}
	float	scaleFactor = 1. ; //mImageArray->GetWidth()*mImageArray->GetHeight() ;
	for (i = 0 ;  i < numIndependentReflections; i++) {
		amp = scaleFactor*refinementReflection[i].mAmplitude ;
		for(j=0; j < refinementReflection[i].mNumEquivalentReflections; j++) {
			phase = refinementReflection[i].mPhase[j] ;
			rl = amp*cos(phase) ;
			im = amp*sin(phase) ;
			h = refinementReflection[i].mH[j] ;
			k = refinementReflection[i].mK[j] ;
			*mImageArray->GetRealHK(h,k) = rl ;
			*mImageArray->GetImagHK(h,k) = im ;
			// Let us do the -h,-k's as well
			*mImageArray->GetRealHK(-h,-k) = rl ;
			*mImageArray->GetImagHK(-h,-k) = -im ;
		}
	}
	*mImageArray->GetRealHK(0,0) =  1.0 ;
	*mImageArray->GetImagHK(0,0) = 0.0 ;
}


// --------------------------------------------------------------------------
// Impose any support constraints that we might have 
// --------------------------------------------------------------------------
void	TDMIteration::ImposeSupportConstraints() {
}

// --------------------------------------------------------------------------
// Calculate the Figure of Merit for this data set
// --------------------------------------------------------------------------
float	TDMIteration::CalculateFigureOfMerit() {
return 0 ;
}

// --------------------------------------------------------------------------
// return the phase that is symmetry related to the phase of reflection h,k
// --------------------------------------------------------------------------
float	TDMIteration::SymmetryRelatedPhase(float phase, short h,short k,short h1,short k1 ) {
	
	float	randomValue = CreateRandomValue() ;	// 0 - 1
	switch (mRefinementSet->mPlaneGroup) {
		case 1:
			// Random phases 0 - 2Pi
			return randomValue*TwoPi ;
			break ;
		case 2:
			// Phase 0 or 180  : (h,k)
			if (randomValue < 0.5) return 0. ;
			return Pi ;
			break ;
		case 3:
			// P(h,k) = P(-h,k) = -P(h,-k)
			if (h1 == h && k1 == -k) return -phase ;
			return phase ;
			break ;
		case 4:
			// P(h,k) = P(-h,k)+k*180
			if (h1 == -h && k1 == k) return phase + k*Pi ;
			if (h1 == h && k1 == -k) return -phase - k*Pi ;
			return phase ;
			break ;
		case 5:
			// P(h,k) = P(-h,k) = -P(h,-k)
			if (h1 == h && k1 == -k) return -phase ;
			return phase ;
			break ;
		case 6:
			// P(h,k) = P(-h,k) = -P(h,-k)
			if (h1 == h && k1 == -k) return -phase ;
			return phase ;
			break ;
		case 7:
			// P(h,k) = P(-h,k)+k*180
			if (h1 == -h && k1 == k) return phase + k*Pi ;
			if (h1 == h && k1 == -k) return -phase - k*Pi ;
			return phase ;
			break ;
		case 8:
			// P(h,k) = P(-h,k)+(h+k)*180
			if (h1 == -h && k1 == k) return phase + (h+k)*Pi ;
			if (h1 == h && k1 == -k) return -phase - (h+k)*Pi ;
			return phase ;
			break ;
		case 9:
			// P(h,k) = P(-h,k) = -P(h,-k)
			if (h1 == h && k1 == -k) return -phase ;
			return phase ;
			break ;
		case 10:
			// P(h,k) = P(-k,h) = -P(k,-h)
			if (h1 == k && k1 == -h) return -phase ;
			return phase ;
			break ;
		case 11:
			// P(h,k) = P(-k,h) = P(-h,k) = -P(k,-h) = -P(h,-k)
			if (h1 == k && k1 == -h) return -phase ;
			if (h1 == h && k1 == -k) return -phase ;
			return phase ;
			break ;
		case 12:
			// P(h,k) = P(-k,h) = P(-h,k)+k*180
			if (h1 == k && k1 == -h) return -phase ;
			if (h1 == -h && k1 == k) return phase + k*Pi ;
			if (h1 == h && k1 == -k) return -phase - k*Pi ;
			return phase ;
			break ;
		case 13:
			// P(h,k) = P(k,-h-k) = P(-h-k,h)
			if (h1 == -k && k1 == h+k) return -phase ;
			if (h1 == h+k && k1 == -h) return -phase ;
			return phase ;
			break ;
		case 14:
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(-k,-h)
			if (h1 == -k && k1 == h+k) return -phase ;
			if (h1 == h+k && k1 == -h) return -phase ;
			if (h1 == k && k1 == h) return -phase ;
			return phase ;
			break ;
		case 15:
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(k,h)
			if (h1 == -k && k1 == h+k) return -phase ;
			if (h1 == h+k && k1 == -h) return -phase ;
			if (h1 == -k && k1 == -h) return -phase ;
			return phase ;
			break ;
		case 16:
			// P(h,k) = P(k,-h-k) = P(-h-k,h)
			if (h1 == -k && k1 == h+k) return -phase ;
			if (h1 == h+k && k1 == -h) return -phase ;
			return phase ;
			break ;
		case 17:
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(k,h)
			if (h1 == -k && k1 == h+k) return -phase ;
			if (h1 == h+k && k1 == -h) return -phase ;
			if (h1 == -k && k1 == -h) return -phase ;
			return phase ;
			break ;
	}
	return phase ;
}

// --------------------------------------------------------
// Create a random phase consistent with the plane group
// --------------------------------------------------------
float	TDMIteration::RandomPhase(short h,short k ) {
	// See what our choices are	
	float	randomValue = CreateRandomValue() ;	// 0 - 1
	
	switch (mRefinementSet->mPlaneGroup) {
		case 1:
		case 13:
			// Random phases 0 - 2Pi
			return randomValue*TwoPi ;
			break ;
		case 2:
		case 6:
		case 7:
		case 8:
		case 9:
		case 10:
		case 11:
		case 12:
		case 16:
		case 17:
			// Random phases 0 or Pi	for all (h,k)
			if (randomValue < 0.5) return 0. ;
			return Pi ;
			break ;
		case 3:
		case 4:
		case 5:
			// Random phases 0 or Pi	for all (h,0)
			if (k == 0) {
				if (randomValue < 0.5) return 0. ;
				return Pi ;
			}
			return randomValue*TwoPi ;
			break ;
		case 14:
			// Random phases 0 or Pi	for all (h,h), (h,-2h), (-2k,k)
			if (k == h || h == -2*k || k == -2*h) {
				if (randomValue < 0.5) return 0. ;
				return Pi ;
			}
			return randomValue*TwoPi ;
			break ;
		case 15:
			// Random phases 0 or Pi	for all (h,-h), (h,2h), (2k,k)
			if (k == -h || h == 2*k || k == 2*h) {
				if (randomValue < 0.5) return 0. ;
				return Pi ;
			}
			return randomValue*TwoPi ;
			break ;
	}
	return randomValue ;
}

// --------------------------------------------------------------------------
// Impose whatever constraints we might have in real space
// --------------------------------------------------------------------------
void	TDMIteration::ImposeRealSpaceConstraints() {
	register 	long i ;
	long		numPixels = mImageArray->GetWidth()*mImageArray->GetHeight() ;
	
	register	float	*p1 = mImageArray->GetData() ;
	
	// Positivity
	// We should introduce an adjustable parameter here
	for(i=0; i < numPixels; i++) {
		if (*p1 < 0.) *p1 = 0. ;
		p1++ ;
	}
	
	// Atomicity
	// square the array and force it ot be equal to itself
	p1 = mImageArray->GetData() ;
	float	squareValue,value ;
	for(i=0; i < numPixels; i++) {
		value = *p1 ;
		squareValue = value*value ;
		*p1++ = squareValue ;
	}
	
}

void	TDMIteration::ImposeReciprocalSpaceConstraints() {
	register long i,j ;
	long	ndx,h,k,h1,k1,hMax,kMax ; ;
	float	astar,bstar,usquare,measuredAmplitude,amplitude,normalizationConst,phase,phase0,rl,im ;
	long	numIndependentReflections = mRefinementSet->mNumIndependentReflections ;
	TRefinementReflection	*refinementReflection = mRefinementSet->mRefinementReflection ;
	
	// Force the amplitudes to be correct
	// The phase stays what is is, unless it is locked 
	*mImageArray->GetRealHK(0,0) = 1.0 ;
	*mImageArray->GetImagHK(0,0) = 0.0 ;
	
	hMax = mImageArray->GetWidth()/2 ;
	kMax = mImageArray->GetHeight()/2 ;
	
	*mReflectionIndex = -2.0 ;

	// First set all the measured amplitudes
	// The amplitudes are adjusted to the measured values
	for (i = 0 ;  i < numIndependentReflections; i++) {
		measuredAmplitude = refinementReflection[i].mAmplitude ;
		for(j=0; j < refinementReflection[i].mNumEquivalentReflections; j++) {
			h = refinementReflection[i].mH[j] ;
			k = refinementReflection[i].mK[j] ;
			//if (gDebug && j == 0) fprintf(stderr,"\n") ;
			//if (gDebug) fprintf(stderr,"Amp[%d] [%d,%d]=%.2f  ",i,h,k,measuredAmplitude) ;
			amplitude = ((Cmplx *)mImageArray->GetRealHK(h,k))->Modulus() ;
			normalizationConst = measuredAmplitude/amplitude ;
			*mImageArray->GetRealHK(h,k) *= normalizationConst ;
			*mImageArray->GetImagHK(h,k) *= normalizationConst ;
			// Let us do the -h,-k's as well
			*mImageArray->GetRealHK(-h,-k) *= normalizationConst ;
			*mImageArray->GetImagHK(-h,-k) *= normalizationConst ;
			*mReflectionIndex->GetDataHK(h,k) = i ;
			*mReflectionIndex->GetDataHK(-h,-k) = i ;
		}
	}
	
	float	aperture = 1.5 ;
	float	gMax = aperture*mRefinementSet->mUmax ;
	float	gMaxSquare = gMax*gMax ;
	double	degreeToRad = asin(1.0)/90. ;
	
	astar = 1./mRefinementSet->mA ;
	bstar = 1./mRefinementSet->mB ;
	float	csGaStar = cos(degreeToRad*(180-mRefinementSet->mGamma)) ;
	float	astSq = astar*astar ;
	float	bstSq = bstar*bstar ;
	float	ab2csga = 2*astar*bstar*csGaStar ;
	float	uhsq,uhcsga ;
	// Go through the entire array and set the phases if we have them
	for(h1=-hMax;h1<=hMax;h1++) {
		uhsq = h1*h1*astSq ;
		uhcsga = h1*ab2csga ;
		for(k1=0;k1<=kMax;k1++) {
			// First check to see if we are within the aperture imposed on the calculation
			usquare = uhsq + k1*k1*bstSq + k1*uhcsga ;
			if (usquare > gMaxSquare) {
				*mImageArray->GetRealHK(h1,k1) = 0. ;
				*mImageArray->GetImagHK(h1,k1) = 0. ;
				// Let us do the -h,-k's as well
				*mImageArray->GetRealHK(-h1,-k1) = 0. ;
				*mImageArray->GetImagHK(-h1,-k1) = 0. ;
				break ;
			}	
			ndx = *mReflectionIndex->GetDataHK(h1,k1) ;			
			if (ndx  < 0) {
				// Set this phase according to symmetry
				// We will keep the phase of this reflection, but we will ensure that
				// symmetry related reflections have the appropriate phase relationship
				// Also we will make sure that forbidden reflections are set to zero
				SetSymmetrizedAmplitudePhase(h1,k1) ;
			}
			else {
				// Set the phases of measured reflections
				h = refinementReflection[ndx].mH[0] ;
				k = refinementReflection[ndx].mK[0] ;
				amplitude = refinementReflection[ndx].mAmplitude ;
				if (refinementReflection[ndx].mUsePhase) {
					// The phase of this reflection is fixed so leave as is
					phase0 = refinementReflection[ndx].mPhase[0] ; 
				}
				else {
					// Set this phase according to symmetry
					// If we have a spacegroup, we must ensure that the phase is correct for
					// the symmetry origin
					phase0 = SetSymmetrizedPhase(ndx) ;
				}
				rl = amplitude*cos(phase0) ;
				im = amplitude*sin(phase0) ;
				*mImageArray->GetRealHK(h,k) = rl ;
				*mImageArray->GetImagHK(h,k) = im ;
				// Let us do the -h,-k's as well
				*mImageArray->GetRealHK(-h,-k) = rl ;
				*mImageArray->GetImagHK(-h,-k) = -im ;
				//if (gDebug) fprintf(stderr,"\namp[%d][%d,%d]=%.4f, phase=%.3f ",ndx,h,k,amplitude,phase0) ;
				for(j=1; j < refinementReflection[ndx].mNumEquivalentReflections; j++) {
					h = refinementReflection[ndx].mH[j] ;
					k = refinementReflection[ndx].mK[j] ;
					if (refinementReflection[ndx].mUsePhase) {
						// The phase of this reflection is fixed so leave as is
						phase = refinementReflection[ndx].mPhase[j] ; 
					}
					else {
						// Set this phase according to symmetry
						phase = SymmetryRelatedPhase(phase0, refinementReflection[ndx].mH[0], refinementReflection[ndx].mK[0],h,k) ;
					}
					//if (gDebug) fprintf(stderr,"amp[%d,%d]=%.4f, phase=%.3f ",h,k,amplitude,phase) ;
					rl = amplitude*cos(phase) ;
					im = amplitude*sin(phase) ;
					*mImageArray->GetRealHK(h,k) = rl ;
					*mImageArray->GetImagHK(h,k) = im ;
					// Let us do the -h,-k's as well
					*mImageArray->GetRealHK(-h,-k) = rl ;
					*mImageArray->GetImagHK(-h,-k) = -im ;
				}
			}
		}
	}
}

void	TDMIteration::SetSymmetrizedAmplitudePhase(long h,long k) {
	// We are trying to set phases (and amplitudes) that are not part of the measured set
	
	if (*mReflectionIndex->GetDataHK(h,k) == -1.0) return ; 	// We are already set
	float	p2,p3,p4,s,alphaSym,argY,argX ;
		
	float rl,im ;
	float amplitude = ((Cmplx *)mImageArray->GetRealHK(h,k))->Modulus() ;
	float p1 = ((Cmplx *)mImageArray->GetRealHK(h,k))->Phase() ;
	
	// Set all amplitudes and phases of symmetry related reflections. This is an unmeasured set ;
	switch (mRefinementSet->mPlaneGroup) {
		case 1:	// p1
			alphaSym = p1 ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(h,k) = rl ;
			*mImageArray->GetImagHK(h,k) = im ;
			*mImageArray->GetRealHK(-h,-k) = rl ;
			*mImageArray->GetImagHK(-h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,-k) = -1. ;
			break ;
		case 2: // p2
			// Phase 0 or 180  : (h,k)
			alphaSym = fabs(p1) < PiHalf ? 0. : Pi ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(h,k) = rl ;
			*mImageArray->GetImagHK(h,k) = im ;
			*mImageArray->GetRealHK(-h,-k) = rl ;
			*mImageArray->GetImagHK(-h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,-k) = -1. ;
			break ;
		case 3: // pm - 
			// P(h,k) = P(-h,k) = -P(h,-k)
			// Phase 0 or 180  : (h,0)
			p2 = ((Cmplx *)mImageArray->GetRealHK(-h,k))->Phase()  ;
			// Calculate symmetrized phase from P(h,k) and P(-h,k)
			argY = sin(p1)+sin(p2) ;
			argX = cos(p1)+cos(p2) ;
			alphaSym = atan2(argY,argX) ;
			if ( k == 0)	
				alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(h,k) = rl ;
			*mImageArray->GetImagHK(h,k) = im ;
			*mImageArray->GetRealHK(-h,-k) = rl ;
			*mImageArray->GetImagHK(-h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,-k) = -1. ;
			*mImageArray->GetRealHK(h,-k) = rl ;
			*mImageArray->GetImagHK(h,-k) = -im ;
			*mImageArray->GetRealHK(-h,k) = rl ;
			*mImageArray->GetImagHK(-h,k) = im ;
			*mReflectionIndex->GetDataHK(h,-k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,k) = -1. ;
			break ;
		case 4: // pg  Systematic absence (0k): k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k)+k*180
			// Phase 0 or 180  : (h,0)
			p2 = ((Cmplx *)mImageArray->GetRealHK(-h,k))->Phase()  ;
			// Calculate symmetrized phase from P(h,k) and P(-h,k)
			s = k%2 == 0 ? 1. : -1. ;
			argY = sin(p1)+s*sin(p2) ;
			argX = cos(p1)+s*cos(p2) ;
			alphaSym = atan2(argY,argX) ;
			if ( k == 0)	
				alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			if (h == 0 && k%2 == 1)
				amplitude = 0. ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(h,k) = rl ;
			*mImageArray->GetImagHK(h,k) = im ;
			*mImageArray->GetRealHK(-h,-k) = rl ;
			*mImageArray->GetImagHK(-h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,-k) = -1. ;
			alphaSym = alphaSym-k*Pi ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(-h,k) = rl ;
			*mImageArray->GetImagHK(-h,k) = im ;
			*mImageArray->GetRealHK(h,-k) = rl ;
			*mImageArray->GetImagHK(h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,-k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,k) = -1. ;
			break ;
		case 5: // cm  Systematic absence (hk): h+k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k) = -P(h,-k)
			// Phase 0 or 180  : (h,0)
			p2 = ((Cmplx *)mImageArray->GetRealHK(-h,k))->Phase()  ;
			// Calculate symmetrized phase from P(h,k) and P(-h,k)
			argY = sin(p1)+sin(p2) ;
			argX = cos(p1)+cos(p2) ;
			alphaSym = atan2(argY,argX) ;
			if ( k == 0)	
				alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			if ((h+k)%2 == 1)
				amplitude = 0. ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(h,k) = rl ;
			*mImageArray->GetImagHK(h,k) = im ;
			*mImageArray->GetRealHK(-h,-k) = rl ;
			*mImageArray->GetImagHK(-h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,-k) = -1. ;
			*mImageArray->GetRealHK(-h,k) = rl ;
			*mImageArray->GetImagHK(-h,k) = im ;
			*mImageArray->GetRealHK(h,-k) = rl ;
			*mImageArray->GetImagHK(h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,-k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,k) = -1. ;
			break ;
		case 6: // pmm  -
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k) = -P(h,-k)
			// Phase 0 or 180  : (h,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(-h,k))->Phase()  ;
			// Calculate symmetrized phase from P(h,k) and P(-h,k)
			argY = sin(p1)+sin(p2) ;
			argX = cos(p1)+cos(p2) ;
			alphaSym = atan2(argY,argX) ;
			alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(h,k) = rl ;
			*mImageArray->GetImagHK(h,k) = im ;
			*mImageArray->GetRealHK(-h,-k) = rl ;
			*mImageArray->GetImagHK(-h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,-k) = -1. ;
			*mImageArray->GetRealHK(-h,k) = rl ;
			*mImageArray->GetImagHK(-h,k) = im ;
			*mImageArray->GetRealHK(h,-k) = rl ;
			*mImageArray->GetImagHK(h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,-k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,k) = -1. ;
			break ;
		case 7: // pmg  Systematic absence (0k): k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k)+k*180
			// Phase 0 or 180  : (h,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(-h,k))->Phase()  ;
			// Calculate symmetrized phase from P(h,k) and P(-h,k)
			s = k%2 == 0 ? 1. : -1. ;
			argY = sin(p1)+s*sin(p2) ;
			argX = cos(p1)+s*cos(p2) ;
			alphaSym = atan2(argY,argX) ;
			alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			if (h == 0 && k%2 == 1)
				amplitude = 0. ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(h,k) = rl ;
			*mImageArray->GetImagHK(h,k) = im ;
			*mImageArray->GetRealHK(-h,-k) = rl ;
			*mImageArray->GetImagHK(-h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,-k) = -1. ;
			alphaSym = alphaSym-k*Pi ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(-h,k) = rl ;
			*mImageArray->GetImagHK(-h,k) = im ;
			*mImageArray->GetRealHK(h,-k) = rl ;
			*mImageArray->GetImagHK(h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,-k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,k) = -1. ;
			break ;
		case 8: // pgg  Systematic absence (h0): h = 2n+1 ; (0k): k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k)+(h+k)*180
			// Phase 0 or 180  : (h,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(-h,k))->Phase()  ;
			// Calculate symmetrized phase from P(h,k) and P(-h,k)
			s = (h+k)%2 == 0 ? 1. : -1. ;
			argY = sin(p1)+s*sin(p2) ;
			argX = cos(p1)+s*cos(p2) ;
			alphaSym = atan2(argY,argX) ;
			alphaSym = fabs(alphaSym) <PiHalf ? 0. : Pi ;
			if ((h == 0 && k%2 == 1) || (k == 0 && h%2 == 1))
				amplitude = 0. ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(h,k) = rl ;
			*mImageArray->GetImagHK(h,k) = im ;
			*mImageArray->GetRealHK(-h,-k) = rl ;
			*mImageArray->GetImagHK(-h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,-k) = -1. ;
			alphaSym = alphaSym-(h+k)*Pi ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(-h,k) = rl ;
			*mImageArray->GetImagHK(-h,k) = im ;
			*mImageArray->GetRealHK(h,-k) = rl ;
			*mImageArray->GetImagHK(h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,-k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,k) = -1. ;
			break ;
		case 9: // cmm  Systematic absence (hk): h+k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k) = -P(h,-k)
			// Phase 0 or 180  : (h,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(-h,k))->Phase()  ;
			// Calculate symmetrized phase from P(h,k) and P(-h,k)
			argY = sin(p1)+sin(p2) ;
			argX = cos(p1)+cos(p2) ;
			alphaSym = atan2(argY,argX) ;
			alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			if ((h+k)%2 == 1)
				amplitude = 0. ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(h,k) = rl ;
			*mImageArray->GetImagHK(h,k) = im ;
			*mImageArray->GetRealHK(-h,-k) = rl ;
			*mImageArray->GetImagHK(-h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,-k) = -1. ;
			*mImageArray->GetRealHK(-h,k) = rl ;
			*mImageArray->GetImagHK(-h,k) = im ;
			*mImageArray->GetRealHK(h,-k) = rl ;
			*mImageArray->GetImagHK(h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,-k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,k) = -1. ;
			break ;
		case 10: // p4
			// |F(h,k)| = |F(-k,h)|
			// P(h,k) = P(-k,h) = -P(k,-h)
			// Phase 0 or 180  : (h,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(-k,h))->Phase()  ;
			// Calculate symmetrized phase from P(h,k) and P(-h,k)
			argY = sin(p1)+sin(p2) ;
			argX = cos(p1)+cos(p2) ;
			alphaSym = atan2(argY,argX) ;
			alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(h,k) = rl ;
			*mImageArray->GetImagHK(h,k) = im ;
			*mImageArray->GetRealHK(-h,-k) = rl ;
			*mImageArray->GetImagHK(-h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,-k) = -1. ;
			*mImageArray->GetRealHK(-k,h) = rl ;
			*mImageArray->GetImagHK(-k,h) = im ;
			*mImageArray->GetRealHK(k,-h) = rl ;
			*mImageArray->GetImagHK(k,-h) = -im ;
			*mReflectionIndex->GetDataHK(-k,h) = -1. ;
			*mReflectionIndex->GetDataHK(k,-h) = -1. ;
			break ;
		case 11: // p4m
			// |F(h,k)| = |F(-k,h)| = |F(-h,k)|
			// P(h,k) = P(-k,h) = P(-h,k) = -P(k,-h) = -P(h,-k)
			// Phase 0 or 180  : (h,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(-h,k))->Phase()  ;
			p3 = ((Cmplx *)mImageArray->GetRealHK(-k,h))->Phase()  ;
			// Calculate symmetrized phase from P(h,k) and P(-h,k)
			argY = sin(p1)+sin(p2)+sin(p3) ;
			argX = cos(p1)+cos(p2)+cos(p3) ;
			alphaSym = atan2(argY,argX) ;
			alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(h,k) = rl ;
			*mImageArray->GetImagHK(h,k) = im ;
			*mImageArray->GetRealHK(-h,-k) = rl ;
			*mImageArray->GetImagHK(-h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,-k) = -1. ;
			*mImageArray->GetRealHK(-k,h) = rl ;
			*mImageArray->GetImagHK(-k,h) = im ;
			*mImageArray->GetRealHK(k,-h) = rl ;
			*mImageArray->GetImagHK(k,-h) = -im ;
			*mReflectionIndex->GetDataHK(-k,h) = -1. ;
			*mReflectionIndex->GetDataHK(k,-h) = -1. ;
			*mImageArray->GetRealHK(-h,k) = rl ;
			*mImageArray->GetImagHK(-h,k) = im ;
			*mImageArray->GetRealHK(h,-k) = rl ;
			*mImageArray->GetImagHK(h,-k) = -im ;
			*mReflectionIndex->GetDataHK(-h,k) = -1. ;
			*mReflectionIndex->GetDataHK(h,-k) = -1. ;
			break ;
		case 12: // p4g
			// |F(h,k)| = |F(-k,h)| = |F(-h,k)|
			// P(h,k) = P(-k,h) = P(-h,k)+k*180
			// Phase 0 or 180  : (h,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(-h,k))->Phase()  ;
			p3 = ((Cmplx *)mImageArray->GetRealHK(-k,h))->Phase()  ;
			// Calculate symmetrized phase from P(h,k) and P(-h,k)
			s = k%2 == 0 ? 1. : -1. ;
			argY = sin(p1)+s*sin(p2)+sin(p3) ;
			argX = cos(p1)+s*cos(p2)+cos(p3) ;
			alphaSym = atan2(argY,argX) ;
			alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			if (h == 0 && k%2 == 1)
				amplitude = 0. ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(h,k) = rl ;
			*mImageArray->GetImagHK(h,k) = im ;
			*mImageArray->GetRealHK(-h,-k) = rl ;
			*mImageArray->GetImagHK(-h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,-k) = -1. ;
			*mImageArray->GetRealHK(-k,h) = rl ;
			*mImageArray->GetImagHK(-k,h) = im ;
			*mImageArray->GetRealHK(k,-h) = rl ;
			*mImageArray->GetImagHK(k,-h) = -im ;
			*mReflectionIndex->GetDataHK(k,-k) = -1. ;
			*mReflectionIndex->GetDataHK(-k,k) = -1. ;
			alphaSym = alphaSym-k*Pi ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(-h,k) = rl ;
			*mImageArray->GetImagHK(-h,k) = im ;
			*mImageArray->GetRealHK(h,-k) = rl ;
			*mImageArray->GetImagHK(h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,-k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,k) = -1. ;
			break ;
		case 13: // p3
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
			// P(h,k) = P(k,-h-k) = P(-h-k,h)
			// Phase 0 or 180  : -
			p2 = ((Cmplx *)mImageArray->GetRealHK(k,-h-k))->Phase()  ;
			p3 = ((Cmplx *)mImageArray->GetRealHK(-h-k,h))->Phase()  ;
			// Calculate symmetrized phase
			argY = sin(p1)+sin(p2)+sin(p3) ;
			argX = cos(p1)+cos(p2)+cos(p3) ;
			alphaSym = atan2(argY,argX) ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(h,k) = rl ;
			*mImageArray->GetImagHK(h,k) = im ;
			*mImageArray->GetRealHK(-h,-k) = rl ;
			*mImageArray->GetImagHK(-h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,-k) = -1. ;
			*mImageArray->GetRealHK(k,-h-k) = rl ;
			*mImageArray->GetImagHK(k,-h-k) = im ;
			*mImageArray->GetRealHK(-k,h+k) = rl ;
			*mImageArray->GetImagHK(-k,h+k) = -im ;
			*mReflectionIndex->GetDataHK(k,-h-k) = -1. ;
			*mReflectionIndex->GetDataHK(-k,h+k) = -1. ;
			*mImageArray->GetRealHK(-h-k,h) = rl ;
			*mImageArray->GetImagHK(-h-k,h) = im ;
			*mImageArray->GetRealHK(h+k,-h) = rl ;
			*mImageArray->GetImagHK(h+k,-h) = -im ;
			*mReflectionIndex->GetDataHK(-h-k,h) = -1. ;
			*mReflectionIndex->GetDataHK(h+k,-h) = -1. ;
			break ;
		case 14: // p3m1
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(-k,-h)
			// Phase 0 or 180  : (h,h) (h,-2h) (-2k,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(k,-h-k))->Phase()  ;
			p3 = ((Cmplx *)mImageArray->GetRealHK(-h-k,h))->Phase()  ;
			p4 = ((Cmplx *)mImageArray->GetRealHK(-k,-h))->Phase()  ;
			// Calculate symmetrized phase
			argY = sin(p1)+sin(p2)+sin(p3)+sin(p4) ;
			argX = cos(p1)+cos(p2)+cos(p3)+cos(p4) ;
			alphaSym = atan2(argY,argX) ;
			if (h == k || k == -2*h || h == -2*k)
				alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(h,k) = rl ;
			*mImageArray->GetImagHK(h,k) = im ;
			*mImageArray->GetRealHK(-h,-k) = rl ;
			*mImageArray->GetImagHK(-h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,-k) = -1. ;
			*mImageArray->GetRealHK(k,-h-k) = rl ;
			*mImageArray->GetImagHK(k,-h-k) = im ;
			*mImageArray->GetRealHK(-k,h+k) = rl ;
			*mImageArray->GetImagHK(-k,h+k) = -im ;
			*mReflectionIndex->GetDataHK(k,-h-k) = -1. ;
			*mReflectionIndex->GetDataHK(-k,h+k) = -1. ;
			*mImageArray->GetRealHK(-h-k,h) = rl ;
			*mImageArray->GetImagHK(-h-k,h) = im ;
			*mImageArray->GetRealHK(h+k,-h) = rl ;
			*mImageArray->GetImagHK(h+k,-h) = -im ;
			*mReflectionIndex->GetDataHK(-h-k,h) = -1. ;
			*mReflectionIndex->GetDataHK(h+k,-h) = -1. ;
			*mImageArray->GetRealHK(-k,-h) = rl ;
			*mImageArray->GetImagHK(-k,-h) = im ;
			*mImageArray->GetRealHK(k,h) = rl ;
			*mImageArray->GetImagHK(k,h) = -im ;
			*mReflectionIndex->GetDataHK(-k,-h) = -1. ;
			*mReflectionIndex->GetDataHK(k,h) = -1. ;
			break ;
		case 15: // p31m
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(k,h)
			// Phase 0 or 180  : (h,-h) (h,2h) (2k,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(k,-h-k))->Phase()  ;
			p3 = ((Cmplx *)mImageArray->GetRealHK(-h-k,h))->Phase()  ;
			p4 = ((Cmplx *)mImageArray->GetRealHK(k,h))->Phase()  ;
			// Calculate symmetrized phase
			argY = sin(p1)+sin(p2)+sin(p3)+sin(p4) ;
			argX = cos(p1)+cos(p2)+cos(p3)+cos(p4) ;
			alphaSym = atan2(argY,argX) ;
			if (h == -k || h == 2*k || k == 2*h)
				alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(h,k) = rl ;
			*mImageArray->GetImagHK(h,k) = im ;
			*mImageArray->GetRealHK(-h,-k) = rl ;
			*mImageArray->GetImagHK(-h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,-k) = -1. ;
			*mImageArray->GetRealHK(k,-h-k) = rl ;
			*mImageArray->GetImagHK(k,-h-k) = im ;
			*mImageArray->GetRealHK(-k,h+k) = rl ;
			*mImageArray->GetImagHK(-k,h+k) = -im ;
			*mReflectionIndex->GetDataHK(k,-h-k) = -1. ;
			*mReflectionIndex->GetDataHK(-k,h+k) = -1. ;
			*mImageArray->GetRealHK(-h-k,h) = rl ;
			*mImageArray->GetImagHK(-h-k,h) = im ;
			*mImageArray->GetRealHK(h+k,-h) = rl ;
			*mImageArray->GetImagHK(h+k,-h) = -im ;
			*mReflectionIndex->GetDataHK(-h-k,h) = -1. ;
			*mReflectionIndex->GetDataHK(h+k,-h) = -1. ;
			*mImageArray->GetRealHK(-k,-h) = rl ;
			*mImageArray->GetImagHK(-k,-h) = im ;
			*mImageArray->GetRealHK(k,h) = rl ;
			*mImageArray->GetImagHK(k,h) = -im ;
			*mReflectionIndex->GetDataHK(-k,-h) = -1. ;
			*mReflectionIndex->GetDataHK(k,h) = -1. ;
			break ;
		case 16: // p6
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
			// P(h,k) = P(k,-h-k) = P(-h-k,h)
			// Phase 0 or 180  : (h,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(k,-h-k))->Phase()  ;
			p3 = ((Cmplx *)mImageArray->GetRealHK(-h-k,h))->Phase()  ;
			// Calculate symmetrized phase
			argY = sin(p1)+sin(p2)+sin(p3) ;
			argX = cos(p1)+cos(p2)+cos(p3) ;
			alphaSym = atan2(argY,argX) ;
			alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(h,k) = rl ;
			*mImageArray->GetImagHK(h,k) = im ;
			*mImageArray->GetRealHK(-h,-k) = rl ;
			*mImageArray->GetImagHK(-h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,-k) = -1. ;
			*mImageArray->GetRealHK(k,-h-k) = rl ;
			*mImageArray->GetImagHK(k,-h-k) = im ;
			*mImageArray->GetRealHK(-k,h+k) = rl ;
			*mImageArray->GetImagHK(-k,h+k) = -im ;
			*mReflectionIndex->GetDataHK(k,-h-k) = -1. ;
			*mReflectionIndex->GetDataHK(-k,h+k) = -1. ;
			*mImageArray->GetRealHK(-h-k,h) = rl ;
			*mImageArray->GetImagHK(-h-k,h) = im ;
			*mImageArray->GetRealHK(h+k,-h) = rl ;
			*mImageArray->GetImagHK(h+k,-h) = -im ;
			*mReflectionIndex->GetDataHK(-h-k,h) = -1. ;
			*mReflectionIndex->GetDataHK(h+k,-h) = -1. ;
			break ;
		case 17: // p6m
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(k,h)
			// Phase 0 or 180  : (h,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(k,-h-k))->Phase()  ;
			p3 = ((Cmplx *)mImageArray->GetRealHK(-h-k,h))->Phase()  ;
			p4 = ((Cmplx *)mImageArray->GetRealHK(k,h))->Phase()  ;
			// Calculate symmetrized phase
			argY = sin(p1)+sin(p2)+sin(p3)+sin(p4) ;
			argX = cos(p1)+cos(p2)+cos(p3)+cos(p4) ;
			alphaSym = atan2(argY,argX) ;
			alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			rl = amplitude*cos(alphaSym) ;
			im = amplitude*sin(alphaSym) ;
			*mImageArray->GetRealHK(h,k) = rl ;
			*mImageArray->GetImagHK(h,k) = im ;
			*mImageArray->GetRealHK(-h,-k) = rl ;
			*mImageArray->GetImagHK(-h,-k) = -im ;
			*mReflectionIndex->GetDataHK(h,k) = -1. ;
			*mReflectionIndex->GetDataHK(-h,-k) = -1. ;
			*mImageArray->GetRealHK(k,-h-k) = rl ;
			*mImageArray->GetImagHK(k,-h-k) = im ;
			*mImageArray->GetRealHK(-k,h+k) = rl ;
			*mImageArray->GetImagHK(-k,h+k) = -im ;
			*mReflectionIndex->GetDataHK(k,-h-k) = -1. ;
			*mReflectionIndex->GetDataHK(-k,h+k) = -1. ;
			*mImageArray->GetRealHK(-h-k,h) = rl ;
			*mImageArray->GetImagHK(-h-k,h) = im ;
			*mImageArray->GetRealHK(h+k,-h) = rl ;
			*mImageArray->GetImagHK(h+k,-h) = -im ;
			*mReflectionIndex->GetDataHK(-h-k,h) = -1. ;
			*mReflectionIndex->GetDataHK(h+k,-h) = -1. ;
			*mImageArray->GetRealHK(k,h) = rl ;
			*mImageArray->GetImagHK(k,h) = im ;
			*mImageArray->GetRealHK(-k,-h) = rl ;
			*mImageArray->GetImagHK(-k,-h) = -im ;
			*mReflectionIndex->GetDataHK(-k,-h) = -1. ;
			*mReflectionIndex->GetDataHK(k,h) = -1. ;
			break ;
	
	}		
}


float	TDMIteration::SetSymmetrizedPhase(long i) {
	float	amplitude ;
	short	numPlaneGroup = mRefinementSet->mPlaneGroup ;
	TRefinementReflection	*refinementReflection = mRefinementSet->mRefinementReflection ;
	Cmplx	*cp ;
	long	h,k ; ;
	float	s,argX,argY,p1,p2,p3,p4,alphaSym ;
	
	h = refinementReflection[i].mH[0] ;
	k = refinementReflection[i].mK[0] ;
	
	cp = (Cmplx *)mImageArray->GetRealHK(h,k) ;
	p1 = cp->Phase() ;
	amplitude = cp->Modulus() ;
	
	if (amplitude == 0.) return 0. ;
	
	// We are assuming that we had the correct origin to begin with so that we still are
	// at the symmetry origin
	switch(numPlaneGroup) {
		case 1:	// p1
			alphaSym = p1 ;
			break ;
		case 2: // p2
			// Phase 0 or 180  : (h,k)
			alphaSym = fabs(p1) < PiHalf ? 0. : Pi ;
			break ;
		case 3: // pm - 
			// P(h,k) = P(-h,k) = -P(h,-k)
			// Phase 0 or 180  : (h,0)
			p2 = ((Cmplx *)mImageArray->GetRealHK(-h,k))->Phase()  ;
			// Calculate symmetrized phase from P(h,k) and P(-h,k)
			argY = sin(p1)+sin(p2) ;
			argX = cos(p1)+cos(p2) ;
			alphaSym = atan2(argY,argX) ;
			if ( k == 0)	
				alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			break ;
		case 4: // pg  Systematic absence (0k): k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k)+k*180
			// Phase 0 or 180  : (h,0)
			p2 = ((Cmplx *)mImageArray->GetRealHK(-h,k))->Phase()  ;
			// Calculate symmetrized phase from P(h,k) and P(-h,k)
			s = k%2 == 0 ? 1. : -1. ;
			argY = sin(p1)+s*sin(p2) ;
			argX = cos(p1)+s*cos(p2) ;
			alphaSym = atan2(argY,argX) ;
			if ( k == 0)	
				alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			break ;
		case 5: // cm  Systematic absence (hk): h+k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k) = -P(h,-k)
			// Phase 0 or 180  : (h,0)
			p2 = ((Cmplx *)mImageArray->GetRealHK(-h,k))->Phase()  ;
			// Calculate symmetrized phase from P(h,k) and P(-h,k)
			argY = sin(p1)+sin(p2) ;
			argX = cos(p1)+cos(p2) ;
			alphaSym = atan2(argY,argX) ;
			if ( k == 0)	
				alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			break ;
		case 6: // pmm  -
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k) = -P(h,-k)
			// Phase 0 or 180  : (h,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(-h,k))->Phase()  ;
			// Calculate symmetrized phase from P(h,k) and P(-h,k)
			argY = sin(p1)+sin(p2) ;
			argX = cos(p1)+cos(p2) ;
			alphaSym = atan2(argY,argX) ;
			alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			break ;
		case 7: // pmg  Systematic absence (0k): k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k)+k*180
			// Phase 0 or 180  : (h,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(-h,k))->Phase()  ;
			// Calculate symmetrized phase from P(h,k) and P(-h,k)
			s = k%2 == 0 ? 1. : -1. ;
			argY = sin(p1)+s*sin(p2) ;
			argX = cos(p1)+s*cos(p2) ;
			alphaSym = atan2(argY,argX) ;
			alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			break ;
		case 8: // pgg  Systematic absence (h0): h = 2n+1 ; (0k): k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k)+(h+k)*180
			// Phase 0 or 180  : (h,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(-h,k))->Phase()  ;
			// Calculate symmetrized phase from P(h,k) and P(-h,k)
			s = (h+k)%2 == 0 ? 1. : -1. ;
			argY = sin(p1)+s*sin(p2) ;
			argX = cos(p1)+s*cos(p2) ;
			alphaSym = atan2(argY,argX) ;
			alphaSym = fabs(alphaSym) <PiHalf ? 0. : Pi ;
			break ;
		case 9: // cmm  Systematic absence (hk): h+k = 2n+1
			// |F(h,k)| = |F(-h,k)|
			// P(h,k) = P(-h,k) = -P(h,-k)
			// Phase 0 or 180  : (h,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(-h,k))->Phase()  ;
			// Calculate symmetrized phase from P(h,k) and P(-h,k)
			argY = sin(p1)+sin(p2) ;
			argX = cos(p1)+cos(p2) ;
			alphaSym = atan2(argY,argX) ;
			alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			break ;
		case 10: // p4
			// |F(h,k)| = |F(-k,h)|
			// P(h,k) = P(-k,h) = -P(k,-h)
			// Phase 0 or 180  : (h,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(-k,h))->Phase()  ;
			// Calculate symmetrized phase from P(h,k) and P(-h,k)
			argY = sin(p1)+sin(p2) ;
			argX = cos(p1)+cos(p2) ;
			alphaSym = atan2(argY,argX) ;
			alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			break ;
		case 11: // p4m
			// |F(h,k)| = |F(-k,h)| = |F(-h,k)|
			// P(h,k) = P(-k,h) = P(-h,k) = -P(k,-h) = -P(h,-k)
			// Phase 0 or 180  : (h,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(-h,k))->Phase()  ;
			p3 = ((Cmplx *)mImageArray->GetRealHK(-k,h))->Phase()  ;
			// Calculate symmetrized phase from P(h,k) and P(-h,k)
			argY = sin(p1)+sin(p2)+sin(p3) ;
			argX = cos(p1)+cos(p2)+cos(p3) ;
			alphaSym = atan2(argY,argX) ;
			alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			break ;
		case 12: // p4g
			// |F(h,k)| = |F(-k,h)| = |F(-h,k)|
			// P(h,k) = P(-k,h) = P(-h,k)+k*180
			// Phase 0 or 180  : (h,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(-h,k))->Phase()  ;
			p3 = ((Cmplx *)mImageArray->GetRealHK(-k,h))->Phase()  ;
			// Calculate symmetrized phase from P(h,k) and P(-h,k)
			s = k%2 == 0 ? 1. : -1. ;
			argY = sin(p1)+s*sin(p2)+sin(p3) ;
			argX = cos(p1)+s*cos(p2)+cos(p3) ;
			alphaSym = atan2(argY,argX) ;
			alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			break ;
		case 13: // p3
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
			// P(h,k) = P(k,-h-k) = P(-h-k,h)
			// Phase 0 or 180  : -
			p2 = ((Cmplx *)mImageArray->GetRealHK(k,-h-k))->Phase()  ;
			p3 = ((Cmplx *)mImageArray->GetRealHK(-h-k,h))->Phase()  ;
			// Calculate symmetrized phase
			argY = sin(p1)+sin(p2)+sin(p3) ;
			argX = cos(p1)+cos(p2)+cos(p3) ;
			alphaSym = atan2(argY,argX) ;
			break ;
		case 14: // p3m1
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(-k,-h)
			// Phase 0 or 180  : (h,h) (h,-2h) (-2k,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(k,-h-k))->Phase()  ;
			p3 = ((Cmplx *)mImageArray->GetRealHK(-h-k,h))->Phase()  ;
			p4 = ((Cmplx *)mImageArray->GetRealHK(-k,-h))->Phase()  ;
			// Calculate symmetrized phase
			argY = sin(p1)+sin(p2)+sin(p3)+sin(p4) ;
			argX = cos(p1)+cos(p2)+cos(p3)+cos(p4) ;
			alphaSym = atan2(argY,argX) ;
			if (h == k || k == -2*h || h == -2*k)
				alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			break ;
		case 15: // p31m
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(k,h)
			// Phase 0 or 180  : (h,-h) (h,2h) (2k,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(k,-h-k))->Phase()  ;
			p3 = ((Cmplx *)mImageArray->GetRealHK(-h-k,h))->Phase()  ;
			p4 = ((Cmplx *)mImageArray->GetRealHK(k,h))->Phase()  ;
			// Calculate symmetrized phase
			argY = sin(p1)+sin(p2)+sin(p3)+sin(p4) ;
			argX = cos(p1)+cos(p2)+cos(p3)+cos(p4) ;
			alphaSym = atan2(argY,argX) ;
			if (h == -k || h == 2*k || k == 2*h)
				alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			break ;
		case 16: // p6
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)|
			// P(h,k) = P(k,-h-k) = P(-h-k,h)
			// Phase 0 or 180  : (h,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(k,-h-k))->Phase()  ;
			p3 = ((Cmplx *)mImageArray->GetRealHK(-h-k,h))->Phase()  ;
			// Calculate symmetrized phase
			argY = sin(p1)+sin(p2)+sin(p3) ;
			argX = cos(p1)+cos(p2)+cos(p3) ;
			alphaSym = atan2(argY,argX) ;
			alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			break ;
		case 17: // p6m
			// |F(h,k)| = |F(k,-h-k)| = |F(-h-k,h)| = |F(k,h)|
			// P(h,k) = P(k,-h-k) = P(-h-k,h) = P(k,h)
			// Phase 0 or 180  : (h,k)
			p2 = ((Cmplx *)mImageArray->GetRealHK(k,-h-k))->Phase()  ;
			p3 = ((Cmplx *)mImageArray->GetRealHK(-h-k,h))->Phase()  ;
			p4 = ((Cmplx *)mImageArray->GetRealHK(k,h))->Phase()  ;
			// Calculate symmetrized phase
			argY = sin(p1)+sin(p2)+sin(p3)+sin(p4) ;
			argX = cos(p1)+cos(p2)+cos(p3)+cos(p4) ;
			alphaSym = atan2(argY,argX) ;
			alphaSym = fabs(alphaSym) < PiHalf ? 0. : Pi ;
			break ;
	
	}
	return alphaSym ;
	
}

void	TDMIteration::IterationCycle() {
	// We are starting in reciprocal space with a set of known amplitudes and
	// a set of random phases
	
	InverseFourierTransform() ;
	ImposeRealSpaceConstraints() ;
	FourierTransform() ;
	ImposeReciprocalSpaceConstraints() ;
	
}

void	TDMIteration::InverseFourierTransform() {
	mImageArray->ifft(false) ;
	
}

void	TDMIteration::FourierTransform() {
	mImageArray->fft() ;
}

