// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include <stdlib.h>
#include <string.h>
#include "globals.h"
#include "XDrivers.h"
#include "DMMenuCommands.h"
#include "TRunDirectMethodsDialog.h"
#include "TDefineContentDialog.h"
static char *scattFactChoices[] =
{ 
	"Electron",
	"Xray",
	nil
};

static char *modeChoices[] =
{ 
	"Mixed (2/3 Entropy 1/3 Positivity)",
	"Entropy Only",
	"Equal (1/2 Entropy 1/2 Positivity)",
	"Positivity Only (dangerous)",
	"Charge Flipping",
	"Atom Scaled Mode",
	nil
};

static char *floatBeamChoices[] =
{ 
	"No",
	"Yes",
	nil
};

extern	char	*sElectron ; 	//	= "ELECTRON" ;
extern	char	*sXRay ; 		//	= "XRAY" ;
extern	char	*sMixed ; 		//	= "MIXED" ;
extern	char	*sEntropy ; 	//	= "ENTROPY" ;
extern	char	*sEqual ; 		//	= "EQUAL" ;
extern	char	*sPositive ; 	//	= "POSITIVE" ;
extern	char	*sOff ; 		//	= "OFF" ;
extern	char	*sOn ; 			//	= "ON" ;
extern	char	*sFloat	; 		//	= "FLOAT" ;
extern	char	*sVerbose ; 	//	= "VERB" ;
extern	char	*sAuto ; 		//	= "AUTO" ;
extern	char	*sFlip ;		//  = "FLIP" ;
extern	char	*sFS2002 ;		//	= "FS2002" ;
char	*LFlip = "FLIP";

extern	char	*gPlaneGroupSymbols[] ; 
extern	void	SendApplicationCommand(TCommand command) ; 
extern	short	gImageDepth ; 
extern	void	PutMessage(char *) ;
extern	char	*ReadLine(char *buffer,long numChars,FILE *fp) ;

extern	char	*FS98_PATH ;
extern	char	*EDMBinDir ;

TRunDirectMethodsDialog *gRunDirectMethodsDialog = nil ;
void WriteINSControlFile(TRefinementSet *refinementSet,FS98ControlRecord &theFS98ControlRecord) ;
void WriteHKLControlFile(TRefinementSet *refinementSet,FS98ControlRecord &theFS98ControlRecord) ;
void WriteEXTControlFile(TRefinementSet *refinementSet,FS98ControlRecord &theFS98ControlRecord) ;

void	TRunDirectMethodsDialog::DefineWindowControls(FS98ControlRecord &data,short planegroup) {

	char		text[256] ;
	TGroupBox	*theBox ;
	TStaticText	*theText ;
	Rect		theRect ;
	Rect		theUCRect ;
	Rect		theParRect ;
	Rect		theContentRect ;
	Rect		theOutputRect ;
		
	if (strlen(data.baseName) == 0)
		strcpy(text,"untitled") ;
	else 
		strcpy(text,data.baseName) ;
	mBaseName = new TEditText(446,15,566,35,"Base name for data files",text);

	if (strlen(data.titl) == 0)
		strcpy(text,"") ;
	else 
		strcpy(text,data.titl) ;
	mTitle = new TEditText(146,40,566,60,"Descriptive title",text);
		
	// Specify the unit cell parameters box	
	theBox = new TGroupBox(20,70,569,195,"Unit Cell Parameters");
	AddObject(theBox) ;
	theUCRect = theBox->GetRect() ;

	mPlanegroupMenu = new TPopupMenu(this,theUCRect.left+287-80,theUCRect.top+15,theUCRect.left+287,theUCRect.top+15+20,
		gPlaneGroupSymbols,planegroup,"2-Dimensional Planegroup") ;

	Boolean angleIs90 = fabs(90.0-data.ga) < 4. ? true : false ;
	Boolean	angleIs120 = fabs(120.0-data.ga) < 4. ? true : false ;
	Boolean AequalB = fabs((data.a-data.b)/((data.a+data.b)/2.)) < 0.03 ? true : false ;
	if (angleIs90) {
		EnableMenuItem(mPlanegroupMenu,3) ;
		EnableMenuItem(mPlanegroupMenu,4) ;
		EnableMenuItem(mPlanegroupMenu,5) ;
		EnableMenuItem(mPlanegroupMenu,6) ;
		EnableMenuItem(mPlanegroupMenu,7) ;
		EnableMenuItem(mPlanegroupMenu,8) ;
		EnableMenuItem(mPlanegroupMenu,9) ;
		EnableMenuItem(mPlanegroupMenu,10) ;
		EnableMenuItem(mPlanegroupMenu,11) ;
		EnableMenuItem(mPlanegroupMenu,12) ;
		EnableMenuItem(mPlanegroupMenu,13) ;
		if (AequalB) {
			EnableMenuItem(mPlanegroupMenu,14) ;
			EnableMenuItem(mPlanegroupMenu,15) ;
			EnableMenuItem(mPlanegroupMenu,16) ;
			EnableMenuItem(mPlanegroupMenu,17) ;
		}
		else {
			DisableMenuItem(mPlanegroupMenu,14) ;
			DisableMenuItem(mPlanegroupMenu,15) ;
			DisableMenuItem(mPlanegroupMenu,16) ;
			DisableMenuItem(mPlanegroupMenu,17) ;
		}
	}
	else {
		DisableMenuItem(mPlanegroupMenu,3) ;
		DisableMenuItem(mPlanegroupMenu,4) ;
		DisableMenuItem(mPlanegroupMenu,5) ;
		DisableMenuItem(mPlanegroupMenu,6) ;
		DisableMenuItem(mPlanegroupMenu,7) ;
		DisableMenuItem(mPlanegroupMenu,8) ;
		DisableMenuItem(mPlanegroupMenu,9) ;
		DisableMenuItem(mPlanegroupMenu,10) ;
		DisableMenuItem(mPlanegroupMenu,11) ;
		DisableMenuItem(mPlanegroupMenu,12) ;
		DisableMenuItem(mPlanegroupMenu,13) ;
		DisableMenuItem(mPlanegroupMenu,14) ;
		DisableMenuItem(mPlanegroupMenu,15) ;
		DisableMenuItem(mPlanegroupMenu,16) ;
		DisableMenuItem(mPlanegroupMenu,17) ;
	}
	if (angleIs120) {
		EnableMenuItem(mPlanegroupMenu,18) ;
		EnableMenuItem(mPlanegroupMenu,19) ;
		EnableMenuItem(mPlanegroupMenu,20) ;
		EnableMenuItem(mPlanegroupMenu,21) ;
		EnableMenuItem(mPlanegroupMenu,22) ;
	}
	else {
		DisableMenuItem(mPlanegroupMenu,18) ;
		DisableMenuItem(mPlanegroupMenu,19) ;
		DisableMenuItem(mPlanegroupMenu,20) ;
		DisableMenuItem(mPlanegroupMenu,21) ;
		DisableMenuItem(mPlanegroupMenu,22) ;
	}

	AddObject(mPlanegroupMenu) ;
	
	// Check to see which planegroups can be selected based on the symmetry we have

	sprintf(text,"%f",data.a) ;
	mA = new TEditText(theUCRect.left+74,theUCRect.top+51,theUCRect.left+74+60,theUCRect.top+51+20,"A [Ang]",text);
	
	sprintf(text,"%f",data.b) ;
	mB = new TEditText(theUCRect.left+74,theUCRect.top+51+20,theUCRect.left+74+60,theUCRect.top+51+20+20,"B [Ang]",text);
	
	sprintf(text,"%f",data.c) ;
	mC = new TEditText(theUCRect.left+74,theUCRect.top+51+40,theUCRect.left+74+60,theUCRect.top+51+20+40,"C [Ang]",text);
	
	sprintf(text,"%.2f",data.al) ;
	mAlpha = new TEditText(theUCRect.left+242,theUCRect.top+51,theUCRect.left+287,theUCRect.top+51+20,"Alpha [Deg]",text);
	
	sprintf(text,"%.2f",data.be) ;
	mBeta = new TEditText(theUCRect.left+242,theUCRect.top+51+20,theUCRect.left+287,theUCRect.top+51+20+20,"Beta [Deg]",text);
	
	sprintf(text,"%.2f",data.ga) ;
	mGamma = new TEditText(theUCRect.left+242,theUCRect.top+51+40,theUCRect.left+287,theUCRect.top+51+20+40,"Gamma [Deg]",text);

	theBox = new TGroupBox(theUCRect.left+307,theUCRect.top+38,theUCRect.left+307+234,theUCRect.top+38+82,"Unit cell content");
	AddObject(theBox) ;
	theContentRect = theBox->GetRect() ;

	mUnknownCellContentU = new TRadioButton(theContentRect.left+20,theContentRect.top+8,
			theContentRect.left+20+130,theContentRect.top+23,
			"Estimated (Sum U^2=1)",data.contentKnown == 2) ;
	AddObject(mUnknownCellContentU) ;

	mUnknownCellContent = new TRadioButton(theContentRect.left+20,theContentRect.top+28,
			theContentRect.left+20+130,theContentRect.top+28+15,
			"Estimated (Sum F^2=1)",data.contentKnown == 1) ;
	AddObject(mUnknownCellContent) ;
	theRect = mUnknownCellContent->GetRect() ;

	mknownCellContent = new TRadioButton(theRect.left,theRect.bottom+10,theRect.right-30,theRect.bottom+10+18,
			"Known",data.contentKnown == 0) ;
	AddObject(mknownCellContent) ;
	theRect = mknownCellContent->GetRect() ;

	mSpecifyContent = new TButton(theContentRect.right-100,theRect.top-2,theContentRect.right-20,theRect.bottom+2,"Atom Types") ;	
	AddObject(mSpecifyContent) ;
	

	// Specify the calculation parameters box	
	theBox = new TGroupBox(theUCRect.left,theUCRect.bottom+20,theUCRect.right,theUCRect.bottom+20+100,"Calculation Parameters (Defaults are indicated)");
	AddObject(theBox) ;
	theParRect = theBox->GetRect() ;
	
	short	offset = 10 ;

	short choice ;
	choice = data.useElectron ? 1 : 2 ;
	short RshiftLDM = 35 ;
	theParRect.left -= RshiftLDM ;	
	mScattFactMenu = new TPopupMenu(this,theParRect.left+322-120,
		theParRect.top+offset,theParRect.left+322,theParRect.top+offset+20,
		scattFactChoices,choice,"Use Scattering Factors for") ;
	AddObject(mScattFactMenu) ;
	char test[4] ; size_t ModeLen = 4 ;
	sprintf(test,"%4s",&data.mode) ;
	if (strncmp ( test, sMixed, ModeLen) == 0 ) //data.mode == sMixed)
		choice = 1 ;
	else if (strncmp ( test, sEntropy, ModeLen) == 0)//(data.mode == sEntropy)
		choice = 2 ;
	else if (strncmp ( test, sEqual, ModeLen) == 0)//(data.mode == sEqual)
		choice = 3 ;
	else if (strncmp ( test, sPositive, ModeLen) == 0)//(data.mode == sPositive)
		choice = 4 ;
	else if (strncmp ( test, sFlip, ModeLen) == 0)//(data.mode == sFlip)
		choice = 5 ;
	else if (strncmp ( test, sFS2002, ModeLen) == 0)//(data.mode == sFS2002)
		choice = 6 ;

	mModeMenu = new TPopupMenu(this,theParRect.left+322-205,
		theParRect.top+offset+20,theParRect.left+322,theParRect.top+offset+20+20,
		modeChoices,choice,"Mode") ;
	AddObject(mModeMenu) ;
//	fprintf(stderr,"\n Choice = %d",choice) ;
//	fprintf(stderr,"\n TT = %s",test) ;

	choice = data.floatBeams ? 2 : 1 ;
	mFloatBeamsMenu = new TPopupMenu(this,theParRect.left+322-60,
		theParRect.top+offset+20+20,theParRect.left+322,theParRect.top+offset+20+20+20,
		floatBeamChoices,choice,"Allow all beams to vary") ;
	AddObject(mFloatBeamsMenu) ;
	
	sprintf(text,"%.2f",data.camp) ;
	mCAmp = new TEditText(theParRect.left+322-60,theParRect.top+offset+4+60,
		theParRect.left+322,theParRect.top+offset+4+60+20,"Beam definition",text);

	theRect = mCAmp->GetRect() ;
	mDefineOrigin = new TCheckBox(theRect.left-174,theRect.top,
		theRect.left-90,theRect.bottom,"Define Origin",data.DefineOrigin);
	AddObject(mDefineOrigin) ;
	
	theParRect.left += RshiftLDM - 10 ;	

	sprintf(text,"%.2f",data.plambda) ;
	mLambdaPositivity = new TEditText(theParRect.left+530-60,
	theParRect.top+offset,theParRect.left+530,theParRect.top+offset+20,"Lambda (Positivity)",text);

	sprintf(text,"%.2f",data.lambda) ;
	mLambdaPhase = new TEditText(theParRect.left+530-60,
		theParRect.top+offset+20,theParRect.left+530,theParRect.top+offset+20+20,"Lambda (Phase)",text);
	
	theRect = mFloatBeamsMenu->GetRect() ;
	theRect.right += RshiftLDM - 10 ;	
	theText = new TStaticText(theRect.right +10,
		theRect.top,theRect.right +80,
		theRect.top+20,"Turn-on speed :") ;
	AddObject(theText) ;
	
	theRect = theText->GetRect() ;
	mSpeedAdjustment = new TCheckBox(theRect.right +4 ,
		theRect.top,
		theRect.right + 80,
		theRect.top+20,"Automatic",data.autoSpeed);
	AddObject(mSpeedAdjustment) ;
	
	theRect = mSpeedAdjustment->GetRect() ;
	sprintf(text,"%.2f",data.speed) ;
		
	Rect	theRect2 =  mLambdaPhase->GetRect() ;

	mTurnOnSpeed = new TEditText(theRect.right+4,
		theRect.top,theRect2.right,theRect.top+20,"",text);
	theRect = mCAmp->GetRect() ;
	theRect2 = mTurnOnSpeed->GetRect() ;
	sprintf(text,"%.2f",data.powerOfAmplitude) ;
	mAmplitudePowerSetting = new TEditText(theRect2.left,
		theRect.top,theRect2.right,theRect.top+20,"Use Amplitude to Power:",text); 

	theParRect.left += 10 ;

	// Specify the output parameters box	
	theBox = new TGroupBox(theParRect.left,theParRect.bottom+20,theParRect.right,theParRect.bottom+20+65,"Output Parameters");
	AddObject(theBox) ;
	theOutputRect = theBox->GetRect() ;


	mVerboseOutput = new TCheckBox(theOutputRect.left+10,
		theOutputRect.top+14,
		theOutputRect.left +10 +110,
		theOutputRect.top+14+20,"Verbose Output",data.verbose);
	AddObject(mVerboseOutput) ;
 
	sprintf(text,"%.2f",data.times) ;
	mImageScaling = new TEditText(theOutputRect.left+220,theOutputRect.top+14,theOutputRect.left+220+40,theOutputRect.top+14+20,"Scaling for output",text);
	sprintf(text,"%d",data.width) ;
	mImageWidth = new TEditText(theOutputRect.left+400-40,theOutputRect.top+14,theOutputRect.left+400,theOutputRect.top+14+20,"#Cells Wide",text);
	sprintf(text,"%d",data.height) ;
	mImageHeight = new TEditText(theOutputRect.left+520-40,theOutputRect.top+14,theOutputRect.left+520,theOutputRect.top+14+20,"#Cells High",text);

	theRect = mImageHeight->GetRect() ;
	mUpdateDisplaySettings = new TButton(theRect.right-100,theRect.bottom+4,theRect.right,theRect.bottom+28,"Update Scale / Size");
	AddObject(mUpdateDisplaySettings) ;
		
	Rect	VerboseOutputRect = mVerboseOutput->GetRect() ;

	mShowLogFilesCheckBox = new TCheckBox(VerboseOutputRect.left,
		VerboseOutputRect.bottom+4,
		VerboseOutputRect.left +10 +200,
		VerboseOutputRect.bottom+4+20,"On Completion Show Log File");
	AddObject(mShowLogFilesCheckBox) ;

	mShowControlFileCheckBox = new TCheckBox(theOutputRect.left+220,
		VerboseOutputRect.bottom+4,
		theOutputRect.left+220+180,
		VerboseOutputRect.bottom+4+20,"On Completion Show Control File");
	AddObject(mShowControlFileCheckBox) ;

	Rect	ButtonRect = theOutputRect ;
	ButtonRect.left = ButtonRect.right - 110 ;
	ButtonRect.top = ButtonRect.bottom + 10 + 30 ;
	ButtonRect.bottom = ButtonRect.top + 25 ;
	
	mShowSolutionsButton = new TButton(ButtonRect.left,ButtonRect.top,ButtonRect.right,ButtonRect.bottom,"Show Solutions...");
	AddObject(mShowSolutionsButton) ;
	
	OffsetRect(&ButtonRect,-((ButtonRect.right-ButtonRect.left) + 8),0) ;
	mRunGeneticButton = new TButton(ButtonRect.left,ButtonRect.top,ButtonRect.right,ButtonRect.bottom,"Run Genetic");
	AddObject(mRunGeneticButton) ;

	OffsetRect(&ButtonRect,-((ButtonRect.right-ButtonRect.left) + 8),0) ;
	mRunConvergenceButton = new TButton(ButtonRect.left,ButtonRect.top,ButtonRect.right,ButtonRect.bottom,"Run Convergence");
	AddObject(mRunConvergenceButton) ;

	OffsetRect(&ButtonRect,-((ButtonRect.right-ButtonRect.left) + 8),0) ;
	mRunPhaseExtensionButton = new TButton(ButtonRect.left,ButtonRect.top,ButtonRect.right,ButtonRect.bottom,"Run Phase Extension");
	AddObject(mRunPhaseExtensionButton) ;
	
	mCloseButton = new TButton(theOutputRect.left,ButtonRect.top,theOutputRect.left+50,ButtonRect.bottom,"Close");
	AddObject(mCloseButton) ;

	ButtonRect = mRunPhaseExtensionButton->GetRect() ;
	OffsetRect(&ButtonRect,0,-30) ;
	mUpdateDataSet = new TButton(ButtonRect.left,ButtonRect.top,ButtonRect.right,ButtonRect.bottom,"Update From Table");
	AddObject(mUpdateDataSet) ;

	ButtonRect = mRunConvergenceButton->GetRect() ;
	OffsetRect(&ButtonRect,0,-26) ;
	mUpdateINSFile = new TCheckBox(ButtonRect.left,ButtonRect.top,ButtonRect.right,ButtonRect.bottom,"Update INS File",data.updateINSFile);
	AddObject(mUpdateINSFile) ;

	ButtonRect = mRunGeneticButton->GetRect() ;
	OffsetRect(&ButtonRect,0,-26) ;
	mRemoveOldFiles = new TCheckBox(ButtonRect.left,ButtonRect.top,ButtonRect.right,ButtonRect.bottom,"Remove Old Files",data.cleanUpFiles);
	AddObject(mRemoveOldFiles) ;

	ButtonRect = mShowSolutionsButton->GetRect() ;
	OffsetRect(&ButtonRect,0,-30) ;
	mCleanupFiles = new TButton(ButtonRect.left,ButtonRect.top,ButtonRect.right,ButtonRect.bottom,"Cleanup Files");
	AddObject(mCleanupFiles) ;

	// Add text edit objects in reverse order to get the tab order correct
	AddTextEditObject(mImageHeight) ;
	AddTextEditObject(mImageWidth) ;
	AddTextEditObject(mImageScaling) ;
	AddTextEditObject(mAmplitudePowerSetting) ;	
	AddTextEditObject(mCAmp) ;
	AddTextEditObject(mTurnOnSpeed) ;
	AddTextEditObject(mLambdaPhase) ;
	AddTextEditObject(mLambdaPositivity) ;
	AddTextEditObject(mGamma) ;
	AddTextEditObject(mBeta) ;
	AddTextEditObject(mAlpha) ;
	AddTextEditObject(mC) ;
	AddTextEditObject(mB) ;
	AddTextEditObject(mA) ;
	AddTextEditObject(mTitle) ;
	AddTextEditObject(mBaseName) ;
	
	gRunDirectMethodsDialog = this ;
}

TRunDirectMethodsDialog::TRunDirectMethodsDialog(TRefinementSet *refinementSet,FS98ControlRecord &dataRecord,short planegroup) 
	: TWindow(kDialogPosX,kDialogPosY,584,480,gImageDepth,"Direct Methods",false) {

	DefineWindowControls(dataRecord,planegroup) ;
	mAtomList = new TAtomList ;
	TContentAtom	*theAtom ;

	long	i ;

	mBkColor = gGrayPixel ;
	XSetBackground(gDisplay,mGC,mBkColor) ;
		
	mRefinementSet	= refinementSet ;

	mFS98ControlRecord = dataRecord ;
	mFS98ControlRecord.sfac = nil ;
	mFS98ControlRecord.unit = nil ;
	mFS98ControlRecord.tfac = nil ;
	mFS98ControlRecord.aocc = nil ;
	
	if (dataRecord.numAtoms > 0) {
		mFS98ControlRecord.sfac = new Str5[dataRecord.numAtoms] ;
		mFS98ControlRecord.unit = new long[dataRecord.numAtoms] ;
		mFS98ControlRecord.tfac = new float[dataRecord.numAtoms] ;
		mFS98ControlRecord.aocc = new float[dataRecord.numAtoms] ;
	}
	for(i = 0; i < dataRecord.numAtoms; i++) {
		theAtom = new TContentAtom() ;
		strcpy(theAtom->mSymbol,dataRecord.sfac[i]) ;
		theAtom->mDW = dataRecord.tfac[i] ;
		theAtom->mOccupancy = dataRecord.aocc[i] ;
		theAtom->mNumInCell = dataRecord.unit[i] ;
		mAtomList->AddAtom(theAtom) ;
		
		strcpy(mFS98ControlRecord.sfac[i],theAtom->mSymbol) ;
		mFS98ControlRecord.unit[i] = theAtom->mNumInCell ;
		mFS98ControlRecord.tfac[i] = theAtom->mDW ;
		mFS98ControlRecord.aocc[i] = theAtom->mOccupancy ;
	}
		
	mAllPhasesUnknown = true ;
	mAllPhasesSet = true ;
	for (i = 0 ;  i < mRefinementSet->mNumIndependentReflections; i++) {
		if (mRefinementSet->mRefinementReflection[i].mPhase[0] != 0.) mAllPhasesUnknown = false ;
		if (mRefinementSet->mRefinementReflection[i].mPhase[0] == 0.) mAllPhasesSet = false ;
	}			
	gRunDirectMethodsDialog = this ;
}

TRunDirectMethodsDialog::~TRunDirectMethodsDialog() {
	if (mAtomList != nil) delete mAtomList ;
	if (mFS98ControlRecord.numAtoms > 0) {
		// Clear all the old stuff
		delete[] mFS98ControlRecord.sfac ;
		delete[] mFS98ControlRecord.unit ;
		delete[] mFS98ControlRecord.tfac ;
		delete[] mFS98ControlRecord.aocc ;
	}
	delete	mRefinementSet ;
	gRunDirectMethodsDialog = nil ;
}

void	TRunDirectMethodsDialog::UpdateFromTable() {

	char	fileName[256] ;
	FILE 	*fp ;
	long	i ;
	
	TTable *theTable = mRefinementSet->GetDataTable() ;
	TableType tableType = mRefinementSet->GetTableType() ;
	if (theTable == nil) {
		PutMessage("The Table window has been closed") ;
		return ;
	}

	delete	mRefinementSet ;
	switch(tableType) {
		case kCrystalDiffractionDataTable:
			mRefinementSet = new TRefinementSet((TCrystalDiffractionDataTable *)theTable) ;
			break ;
		case kDiffractionDataTable:
			mRefinementSet = new TRefinementSet((TDiffractionDataTable *)theTable) ;
			break ;
	}
	// Rewrite the data files that go with this refinement set
	// Write the control file 'name'.ext
	strcpy(fileName,mFS98ControlRecord.baseName) ;
	strcat(fileName,".ext") ;
	fp = fopen(fileName,"w") ;
	if (fp == nil) return ;
	
	float   UsedPhase ;

	mAllPhasesUnknown = true ;
	mAllPhasesSet = true ;
	for (i = 0 ;  i < mRefinementSet->mNumIndependentReflections; i++) {
		//LDM Trap phase here ?
		UsedPhase = 0 ;
		if(mRefinementSet->mRefinementReflection[i].mUsePhase) {
			UsedPhase = mRefinementSet->mRefinementReflection[i].mPhase[0] ;
		} 
		fprintf(fp,"%d\t%d\t0\t%f\t%f\n",mRefinementSet->mRefinementReflection[i].mH[0],mRefinementSet->mRefinementReflection[i].mK[0],
					mRefinementSet->mRefinementReflection[i].mAmplitude,UsedPhase) ;	
		if (mRefinementSet->mRefinementReflection[i].mPhase[0] != 0.) mAllPhasesUnknown = false ;
		if (mRefinementSet->mRefinementReflection[i].mPhase[0] == 0.) mAllPhasesSet = false ;
	}		
	fclose(fp) ;

	// Write the control file 'name'.hkl
	strcpy(fileName,mFS98ControlRecord.baseName) ;
	strcat(fileName,".hkl") ;
	fp = fopen(fileName,"w") ;
	if (fp == nil) return ;
	
	for (i = 0 ;  i < mRefinementSet->mNumIndependentReflections; i++) {
		//LDM Trap phase here ?
		UsedPhase = 0 ;
		if(mRefinementSet->mRefinementReflection[i].mUsePhase) {
			UsedPhase = mRefinementSet->mRefinementReflection[i].mPhase[0] ;
		} 
		fprintf(fp,"%d\t%d\t0\t%f\t%f\n",mRefinementSet->mRefinementReflection[i].mH[0],mRefinementSet->mRefinementReflection[i].mK[0],
					mRefinementSet->mRefinementReflection[i].mAmplitude,UsedPhase) ;		
	}	
	fclose(fp) ;
}


void	TRunDirectMethodsDialog::UpdateDisplaySettings() {
	// Read the current ins file
	// and update scale and width/height
	char	buffer[256],text[256] ;
	FILE	*fpin,*fpout ;
	char	fileName[256] ;
	
	strcpy(fileName,gFS98ControlRecord.baseName) ;
	strcat(fileName,".ins") ;
	
	fpin = fopen(fileName,"r") ;
	if (fpin == NULL) {
		sprintf(buffer,"Failed to open the file %s",fileName) ;
		PutMessage(buffer) ;
		return ;
	}

	fpout = fopen("temp.xyz","w") ;
	for(;;) {
		if (ReadLine(buffer,256,fpin) == NULL) break ;
		else if (strstr(buffer,"TIMES") != NULL) {
			mImageScaling->GetEditText(text) ;
			mFS98ControlRecord.times = atof(text) ;
			sprintf(buffer,"TIMES\t%f",mFS98ControlRecord.times) ;
		}
		else if (strstr(buffer,"WIDTH") != NULL) {
			mImageWidth->GetEditText(text) ;
			mFS98ControlRecord.width = atoi(text) ;
			sprintf(buffer,"WIDTH\t%d",mFS98ControlRecord.width) ;
		}
		else if (strstr(buffer,"HEIGHT") != NULL) {
			mImageHeight->GetEditText(text) ;
			mFS98ControlRecord.height = atoi(text) ;
			sprintf(buffer,"HEIGHT\t%d",mFS98ControlRecord.height) ;
		}
		fprintf(fpout,"%s\n",buffer) ; 
		if (strstr(buffer,"END") != NULL) {
			break ;
		}
	}
	
	fflush(fpout) ;
	fclose(fpin) ;
	fclose(fpout) ;

	sprintf(text,"mv temp.xyz %s",fileName) ;
	system(text) ;

}

void	TRunDirectMethodsDialog::Draw() {
	
	EraseRect(&mPortRect) ;	
	TPane::Draw() ;
}

void	TRunDirectMethodsDialog::DoKeyDownInObject(TObject *obj) {
	char	text[256] ;
	short	planeGroup ;
	if (obj == mTurnOnSpeed) {
		mTurnOnSpeed->GetEditText(text) ;
		if (strlen(text) > 0) {
			mSpeedAdjustment->SetControlValue(0) ;
		}
	}
	else if (obj == mA) {
		mA->GetEditText(text) ;
		planeGroup = mPlanegroupMenu->GetSelection() ;
		switch(planeGroup) {
			case 1:	// p1
			case 2: // p2
				break ;
			case 3: // pm -
			case 4: // p11m
			case 5: // p1g1
			case 6: // p11g
			case 7: // c1m1
			case 8: // c11m
			case 9: // p2mm  -
			case 10: // p2mg
			case 11: // p2gm
			case 12: // p2gg  
			case 13: // c2mm  
				break ;
			case 14: // p4
			case 15: // p4mm
			case 16: // p4mg
			case 17: // p4gm
				mB->SetEditText(text) ;
				break ;
			case 18: // p3
			case 19: // p3m1
			case 20: // p31m
			case 21: // p6
			case 22: // p6m
				mB->SetEditText(text) ;
				break ;
		}
	}
	else if (obj == mB) {
		mB->GetEditText(text) ;
		planeGroup = mPlanegroupMenu->GetSelection() ;
		switch(planeGroup) {
			case 1:	// p1
			case 2: // p2
				break ;
			case 3: // pm -
			case 4: // p11m
			case 5: // p1g1
			case 6: // p11g
			case 7: // c1m1
			case 8: // c11m
			case 9: // p2mm  -
			case 10: // p2mg
			case 11: // p2gm
			case 12: // p2gg  
			case 13: // c2mm  
				break ;
			case 14: // p4
			case 15: // p4mm
			case 16: // p4mg
			case 17: // p4gm
				mA->SetEditText(text) ;
				break ;
			case 18: // p3
			case 19: // p3m1
			case 20: // p31m
			case 21: // p6
			case 22: // p6m
				mA->SetEditText(text) ;
				break ;
			}
		}
}

void	TRunDirectMethodsDialog::DoMouseDownInObject(TObject *obj,Point *pt) {
	char	text[256] ;
		//Try a fix for resetting of plane groups
		mGamma->GetEditText(text) ;
		float Gamma = atof(text) ;
		
		mA->GetEditText(text) ;
		float A = atof(text) ;
		
		mB->GetEditText(text) ;
		float B = atof(text) ;
		Boolean angleIs90 = fabs(90.0-Gamma) < 4. ? true : false ;
		Boolean	angleIs120 = fabs(120.0-Gamma) < 4. ? true : false ;
		Boolean AequalB = fabs((A-B)/((A+B)/2.)) < 0.03 ? true : false ;
		if (angleIs90) {
			EnableMenuItem(mPlanegroupMenu,3) ;
			EnableMenuItem(mPlanegroupMenu,4) ;
			EnableMenuItem(mPlanegroupMenu,5) ;
			EnableMenuItem(mPlanegroupMenu,6) ;
			EnableMenuItem(mPlanegroupMenu,7) ;
			EnableMenuItem(mPlanegroupMenu,8) ;
			EnableMenuItem(mPlanegroupMenu,9) ;
			EnableMenuItem(mPlanegroupMenu,10) ;
			EnableMenuItem(mPlanegroupMenu,11) ;
			EnableMenuItem(mPlanegroupMenu,12) ;
			EnableMenuItem(mPlanegroupMenu,13) ;
			if (AequalB) {
				EnableMenuItem(mPlanegroupMenu,14) ;
				EnableMenuItem(mPlanegroupMenu,15) ;
				EnableMenuItem(mPlanegroupMenu,16) ;
				EnableMenuItem(mPlanegroupMenu,17) ;
			}
			else {
				DisableMenuItem(mPlanegroupMenu,14) ;
				DisableMenuItem(mPlanegroupMenu,15) ;
				DisableMenuItem(mPlanegroupMenu,16) ;
				DisableMenuItem(mPlanegroupMenu,17) ;
			}
		}
		else {
			DisableMenuItem(mPlanegroupMenu,3) ;
			DisableMenuItem(mPlanegroupMenu,4) ;
			DisableMenuItem(mPlanegroupMenu,5) ;
			DisableMenuItem(mPlanegroupMenu,6) ;
			DisableMenuItem(mPlanegroupMenu,7) ;
			DisableMenuItem(mPlanegroupMenu,8) ;
			DisableMenuItem(mPlanegroupMenu,9) ;
			DisableMenuItem(mPlanegroupMenu,10) ;
			DisableMenuItem(mPlanegroupMenu,11) ;
			DisableMenuItem(mPlanegroupMenu,12) ;
			DisableMenuItem(mPlanegroupMenu,13) ;
			DisableMenuItem(mPlanegroupMenu,14) ;
			DisableMenuItem(mPlanegroupMenu,15) ;
			DisableMenuItem(mPlanegroupMenu,16) ;
			DisableMenuItem(mPlanegroupMenu,17) ;
		}
		if (angleIs120) {
			EnableMenuItem(mPlanegroupMenu,18) ;
			EnableMenuItem(mPlanegroupMenu,19) ;
			EnableMenuItem(mPlanegroupMenu,20) ;
			EnableMenuItem(mPlanegroupMenu,21) ;
			EnableMenuItem(mPlanegroupMenu,22) ;
		}
		else {
			DisableMenuItem(mPlanegroupMenu,18) ;
			DisableMenuItem(mPlanegroupMenu,19) ;
			DisableMenuItem(mPlanegroupMenu,20) ;
			DisableMenuItem(mPlanegroupMenu,21) ;
			DisableMenuItem(mPlanegroupMenu,22) ;
		}

	if (pt != nil)
		if (!obj->DoMouseDown(pt)) return ;
	if (obj == mPlanegroupMenu) {
		short planeGroup = mPlanegroupMenu->GetSelection() ;
		// Set lattice parameters accordingly
		mA->GetEditText(text) ;
		switch(planeGroup) {
			case 1:	// p1
			case 2: // p2
				break ;
			case 3: // pm 
			case 4: // p11m 
			case 5: // pg 
			case 6: // p11g
			case 7: // cm  
			case 8: // c11m 
			case 9: // p2mm  
			case 10: // p2mg  
			case 11: // p2gm
			case 12: // pgg
			case 13: // cmm
				mGamma->SetEditText("90.00") ;
				break ;
			case 14: // p4
			case 15: // p4mm
			case 16: // p4g
			case 17: // p4mg
				mB->SetEditText(text) ;
				mGamma->SetEditText("90.00") ;
				break ;
			case 18: // p3
			case 19: // p3m1
			case 20: // p31m
			case 21: // p6
			case 22: // p6mm
				mB->SetEditText(text) ;
				mGamma->SetEditText("120.00") ;
				break ;
		
		}
	}
	else if (obj == mRunPhaseExtensionButton) {	
		if (mAllPhasesUnknown) {
			PutMessage("Some Phases must be assigned a value") ;
			return ;
		}
		if (mAllPhasesSet  && mFloatBeamsMenu->GetSelection() == 1) {
			PutMessage("Some Phases must be allowed to vary") ;
			return ;
		}
		mRunPhaseExtensionButton->SetState(true) ;
		if (GetDialogData()) {	
			if (mFS98ControlRecord.updateINSFile) 
				WriteINSControlFile(mRefinementSet,mFS98ControlRecord) ;
			WriteEXTControlFile(mRefinementSet,mFS98ControlRecord) ;
			mRefinementSet->RunPhaseExtension(mFS98ControlRecord) ;
		}
		mRunPhaseExtensionButton->SetState(false) ;
	}
	else if (obj == mRunConvergenceButton) {
		mRunConvergenceButton->SetState(true) ;
		if (GetDialogData()) {
			if (mFS98ControlRecord.updateINSFile) 
				WriteINSControlFile(mRefinementSet,mFS98ControlRecord) ;
			WriteHKLControlFile(mRefinementSet,mFS98ControlRecord) ;
			mRefinementSet->RunConvergence(mFS98ControlRecord) ;
		}
		mRunConvergenceButton->SetState(false) ;
	}
	else if (obj == mRunGeneticButton) {
		if (mAllPhasesSet  && mFloatBeamsMenu->GetSelection() == 1) {
			PutMessage("Some Phases must be allowed to vary") ;
			return ;
		}
		mRunGeneticButton->SetState(true) ;
		if (GetDialogData()) {
			if (mFS98ControlRecord.updateINSFile) 
				WriteINSControlFile(mRefinementSet,mFS98ControlRecord) ;
			mRefinementSet->RunGenetic(mFS98ControlRecord) ;
		}
		mRunGeneticButton->SetState(false) ;
	}
	else if (obj == mShowSolutionsButton) {
		mShowSolutionsButton->SetState(true) ;
		if (GetDialogData()) {
			if (mFS98ControlRecord.updateINSFile) 
				WriteINSControlFile(mRefinementSet,mFS98ControlRecord) ;
			mRefinementSet->ShowSolutions(mFS98ControlRecord) ;
		}
		mShowSolutionsButton->SetState(false) ; 
	}
	else if (obj == mUpdateDataSet) {
		mUpdateDataSet->SetState(true) ;
		UpdateFromTable() ;
		mUpdateDataSet->SetState(false) ;
	}
	else if (obj == mUpdateDisplaySettings) {
		mUpdateDisplaySettings->SetState(true) ;
		UpdateDisplaySettings() ;
		mUpdateDisplaySettings->SetState(false) ;
	}	
	else if (obj == mCloseButton) {
		SendApplicationCommand(kClose) ;
	}
	else if (obj == mUnknownCellContent) {
		mknownCellContent->SetControlValue(0) ;
		mUnknownCellContentU->SetControlValue(0) ;
	}
	else if (obj == mknownCellContent) {
		mUnknownCellContent->SetControlValue(0) ;
		mUnknownCellContentU->SetControlValue(0) ;
	}
	else if (obj == mUnknownCellContentU) {
		mUnknownCellContent->SetControlValue(0) ;
		mknownCellContent->SetControlValue(0) ;
	}
	else if (obj == mSpecifyContent) {
		TDefineContentDialog *theDialog = new TDefineContentDialog(mAtomList) ;
		if (!theDialog->DoModalDialog()) {
			delete theDialog ;
			return  ;
		}
		delete mAtomList ;
		mAtomList = theDialog->GetAtomData() ;
		//Change, so having something does not force content flag
		//if (mAtomList != nil) {
		//	mUnknownCellContent->SetControlValue(0) ;
		//	mknownCellContent->SetControlValue(1) ; 
		//}
		//else {
		//	mUnknownCellContent->SetControlValue(1) ;
		//	mknownCellContent->SetControlValue(0) ; 
		//}
		delete theDialog ;
	}
	else if (obj == mCleanupFiles) {
		// Run the shell commands to execute the clean up
		sprintf(text,"%s%sfsclean.sh",FS98_PATH,EDMBinDir) ;
		fprintf(stderr,"\nRunning the command %s\n",text) ; fflush(stderr) ;
		system(text) ;
	}
}

bool TRunDirectMethodsDialog::GetDialogData() {

	// Get the data from the dialog
	char 	text[256] ;
	short	menuChoice,i ;
	TContentAtom	*theAtom ;
	
	mBaseName->GetEditText(mFS98ControlRecord.baseName) ;
	
	mTitle->GetEditText(mFS98ControlRecord.titl) ;

	mFS98ControlRecord.symm = gPlaneGroupSymbols[mPlanegroupMenu->GetSelection()-1] ;

	mA->GetEditText(text) ;
	mFS98ControlRecord.a = atof(text) ;
	
	mB->GetEditText(text) ;
	mFS98ControlRecord.b = atof(text) ;
	
	if (mFS98ControlRecord.a < 2.0 && mFS98ControlRecord.b < 2.0) {
		PutMessage("Lattice Parameters are too small") ;
		return false ;
	}
	
	mC->GetEditText(text) ;
	mFS98ControlRecord.c = atof(text) ;
	
	mAlpha->GetEditText(text) ;
	mFS98ControlRecord.al = atof(text) ;
	
	mBeta->GetEditText(text) ;
	mFS98ControlRecord.be = atof(text) ;
	
	mGamma->GetEditText(text) ;
	mFS98ControlRecord.ga = atof(text) ;

	//mFS98ControlRecord.contentKnown = mknownCellContent->GetControlValue() ;
	if (mknownCellContent->GetControlValue() ) {
		mFS98ControlRecord.contentKnown = 0 ;
	}
	if (mUnknownCellContent->GetControlValue() ) {
		mFS98ControlRecord.contentKnown = 1 ;
	}
	if (mUnknownCellContentU->GetControlValue() ) {
		mFS98ControlRecord.contentKnown = 2 ;
	}
	

	menuChoice = mScattFactMenu->GetSelection() ;
	mFS98ControlRecord.useElectron = menuChoice == 1 ? true : false ;
	
	menuChoice = mModeMenu->GetSelection() ;
	short ModeChoice = menuChoice ;
	if (menuChoice == 1)
		mFS98ControlRecord.mode = sMixed ;
	else if (menuChoice == 2)
		mFS98ControlRecord.mode = sEntropy ;
	else if (menuChoice == 3)
		mFS98ControlRecord.mode = sEqual ;
	else if (menuChoice == 4)
		mFS98ControlRecord.mode = sPositive ;
	else if (menuChoice == 5) {
		mFS98ControlRecord.mode = sFlip ;
		mFS98ControlRecord.plambda = 1.0 ;
		mFS98ControlRecord.lambda = 1.0 ;
	}
	else if (menuChoice == 6)
		mFS98ControlRecord.mode = sFS2002 ;

	menuChoice = mFloatBeamsMenu->GetSelection() ;
	mFS98ControlRecord.floatBeams = menuChoice == 1 ? false : true ;

	mCAmp->GetEditText(text) ;
	mFS98ControlRecord.camp = atof(text) ;

	mLambdaPositivity->GetEditText(text) ;
	mFS98ControlRecord.plambda = atof(text) ;

	mLambdaPhase->GetEditText(text) ;
	mFS98ControlRecord.lambda = atof(text) ;

	mFS98ControlRecord.autoSpeed = mSpeedAdjustment->GetControlValue() ;
	
	if (!mFS98ControlRecord.autoSpeed) {
		mTurnOnSpeed->GetEditText(text) ;
		if (strlen(text) == 0) {
			PutMessage("No value for the speed has been set") ;
			return false ;
		}
		mFS98ControlRecord.speed = atof(text) ;
	}

	mAmplitudePowerSetting->GetEditText(text) ;
	mFS98ControlRecord.powerOfAmplitude = atof(text) ;

	mFS98ControlRecord.verbose = mVerboseOutput->GetControlValue() ;
	mFS98ControlRecord.cleanUpFiles = mRemoveOldFiles->GetControlValue() ;
	mFS98ControlRecord.DefineOrigin = mDefineOrigin->GetControlValue() ;
	
	mFS98ControlRecord.showLogFile = mShowLogFilesCheckBox->GetControlValue() ;
	mFS98ControlRecord.showControlFile = mShowControlFileCheckBox->GetControlValue() ;
	
	mImageScaling->GetEditText(text) ;
	mFS98ControlRecord.times = atof(text) ;
	
	mImageWidth->GetEditText(text) ;
	mFS98ControlRecord.width = atoi(text) ;
	
	mImageHeight->GetEditText(text) ;
	mFS98ControlRecord.height = atoi(text) ;
	if (mAtomList != nil && (mFS98ControlRecord.contentKnown == 0) ) {
		if (mFS98ControlRecord.numAtoms > 0) {
			// Clear all the old stuff
			delete[] mFS98ControlRecord.sfac ;
			delete[] mFS98ControlRecord.unit ;
			delete[] mFS98ControlRecord.tfac ;
			delete[] mFS98ControlRecord.aocc ;
		}
		mFS98ControlRecord.numAtoms = mAtomList->GetNumberOfAtoms() ;
		mFS98ControlRecord.sfac = new Str5[mFS98ControlRecord.numAtoms] ;
		mFS98ControlRecord.unit = new long[mFS98ControlRecord.numAtoms] ;
		mFS98ControlRecord.tfac = new float[mFS98ControlRecord.numAtoms] ;
		mFS98ControlRecord.aocc = new float[mFS98ControlRecord.numAtoms] ;
		mFS98ControlRecord.contentKnown = 0 ; //true ;
		for(i=0; i < mFS98ControlRecord.numAtoms; i++) {
			theAtom = mAtomList->GetAtom(i) ;
			strcpy(mFS98ControlRecord.sfac[i],theAtom->mSymbol) ;
			mFS98ControlRecord.unit[i] = theAtom->mNumInCell ;
			mFS98ControlRecord.tfac[i] = theAtom->mDW ;
			mFS98ControlRecord.aocc[i] = theAtom->mOccupancy ;
		}
	}

	// Force some defaults
	if (ModeChoice == 5) {
		mFS98ControlRecord.plambda = 1.0 ;
		mFS98ControlRecord.lambda = 1.0 ;
	}
	mFS98ControlRecord.updateINSFile = mUpdateINSFile->GetControlValue() ;
	return true ;
}

