#include "TRunMEMDialog.h"
#include "TWindowList.h"
#include <stdlib.h>

MEMRecord	gMEMRecord = {nil,300.,0.65,-400.,20.,0.2,10.,0.00001,0.,1.0,0.,10,10,kUseNoiseFromWienerFilter } ;

char	*ExperimentalmageChoices[] = {"Image",nil} ;
char	*NoiseImageChoices[] = {"Image",nil} ;

#define max(x,y) (x) > (y) ? (x) : (y)

TRunMEMDialog::TRunMEMDialog(MEMRecord &memRecord,TWindowList *windowList)
	: TDialog(kDialogPosX,kDialogPosY,480,350,"Maximum Entropy Phase Extension") {

	char		text[256],windowName[256],*cp ;
	TGroupBox	*theGroupBox ;
	Rect		imageConditionsRect,noiseRect,parameterRect,optionsRect ;
	short		width,i,numImageWindows ;
	TWindow		*theWindow ;
			
	width = mPortRect.right - mPortRect.left ;

	memRecord.window->GetWindowTitle(text) ;
	if ((cp = strrchr(text,'/')) == nil) cp = text ; else cp++ ;
	strcpy(windowName,cp) ;
		
	theGroupBox = new TGroupBox(20,15,223,160,"Imaging conditions") ;
	AddObject(theGroupBox) ;
	imageConditionsRect = theGroupBox->GetRect() ;
	
	sprintf(text,"%.0f",memRecord.voltage) ;
	mVoltageText = new TEditText(imageConditionsRect.left+137,imageConditionsRect.top+10,
		imageConditionsRect.left+189,imageConditionsRect.top+30,"Voltage [kV]",text) ;
	
	sprintf(text,"%.2f",memRecord.cs) ;
	mCsText = new TEditText(imageConditionsRect.left+137,imageConditionsRect.top+36,
		imageConditionsRect.left+189,imageConditionsRect.top+56,"Cs [mm]",text) ;

	sprintf(text,"%.0f",memRecord.defocus) ;
	mDefocusText = new TEditText(imageConditionsRect.left+137,imageConditionsRect.top+62,
		imageConditionsRect.left+189,imageConditionsRect.top+82,"Defocus [A]",text) ;

	sprintf(text,"%.0f",memRecord.delta) ;
	mDeltaText = new TEditText(imageConditionsRect.left+137,imageConditionsRect.top+88,
		imageConditionsRect.left+189,imageConditionsRect.top+108,"Delta [A]",text) ;

	sprintf(text,"%.2f",memRecord.divergence) ;
	mDivergenceText = new TEditText(imageConditionsRect.left+137,imageConditionsRect.top+114,
		imageConditionsRect.left+189,imageConditionsRect.top+134,"Divergence [mrad]",text) ;
	

	theGroupBox = new TGroupBox(234,15,460,160,"Specify noise using...") ;
	AddObject(theGroupBox) ;
	noiseRect = theGroupBox->GetRect() ;
	
	mNoiseWienerFilter = new TRadioButton(noiseRect.left+8,noiseRect.top+10,
		noiseRect.left+206,noiseRect.top+30,"Wiener filtered experiment",memRecord.noiseType == kUseNoiseFromWienerFilter) ;
	AddObject(mNoiseWienerFilter) ;

	mNoisePercentage = new TRadioButton(noiseRect.left+8,noiseRect.top+36,
		noiseRect.left+206,noiseRect.top+56,"Percentage of experiment",memRecord.noiseType == kUsePercentageNoise) ;
	AddObject(mNoisePercentage) ;

	sprintf(text,"%.1f",memRecord.percentage) ;
	mPercentageText = new TEditText(noiseRect.left+148,noiseRect.top+62,
		noiseRect.left+191,noiseRect.top+82,"",text) ;

	mNoiseImage = new TRadioButton(noiseRect.left+8,noiseRect.top+88,
		noiseRect.left+206,noiseRect.top+108,"Noise image below",memRecord.noiseType == kUseNoiseImage) ;
	AddObject(mNoiseImage) ;
		
	mNoiseImageMenu = new TPopupMenu(this,noiseRect.left+59,noiseRect.top+114,
		noiseRect.left+218,noiseRect.top+134,windowName,"") ;	
		
	mNoiseImageMenu->AddMenuItem("None (Noise = 0)",999) ;
	
	numImageWindows = 0 ;
	for(i=0; i < windowList->GetNumberOfWindows() ; i++) {
		theWindow = windowList->GetWindow(i) ;
		if (theWindow->GetWindowKind() == 'iwnd')
			if (((TImageWindow *)theWindow)->GetImage()->IsRealSpace())
				numImageWindows++ ;
	}
	
	mWindowListIndex = new short[max(numImageWindows,8)] ;
	for(i=0; i < windowList->GetNumberOfWindows() ; i++) {
		theWindow = windowList->GetWindow(i) ;
		if (theWindow->GetWindowKind() == 'iwnd') {
			if (((TImageWindow *)theWindow)->GetImage()->IsRealSpace()) {
				theWindow->GetWindowTitle(text) ;
				if ((cp = strrchr(text,'/')) == nil) cp = text ; else cp++ ;
				mNoiseImageMenu->AddMenuItem(cp,999) ;
			}
		}
	}
	mNoiseImageMenuChoice = 1 ;
	mNoiseImageMenu->SetSelection(mNoiseImageMenuChoice) ;
	AddObject(mNoiseImageMenu) ;
	

	theGroupBox = new TGroupBox(20,180,460,248,"Adjustable parameters") ;
	AddObject(theGroupBox) ;
	parameterRect = theGroupBox->GetRect() ;
	
	sprintf(text,"%f",memRecord.lagrangeMultiplier) ;
	mLagrangeMultiplierText = new TEditText(parameterRect.left+344,parameterRect.top+10,
		parameterRect.left+421,parameterRect.top+30,"Size change in Langrange Multipliers",text) ;

	sprintf(text,"%d",memRecord.numIterations) ;
	mNumIterationsText = new TEditText(parameterRect.left+135,parameterRect.top+36,
		parameterRect.left+172,parameterRect.top+56,"Number of Iterations",text) ;
	
	sprintf(text,"%d",memRecord.numConjugateSteps) ;
	mNumConjugateStepsText = new TEditText(parameterRect.left+384,parameterRect.top+36,
		parameterRect.left+421,parameterRect.top+56,"Number of Successive conjugate steps",text) ;
		
		
	theGroupBox = new TGroupBox(20,parameterRect.bottom+10,460,parameterRect.bottom+50,"Options") ;
	AddObject(theGroupBox) ;
	optionsRect = theGroupBox->GetRect() ;
	
	mShowCTFImage = new TCheckBox(optionsRect.left+10,optionsRect.top+10,
		(optionsRect.right+optionsRect.left)/2,optionsRect.top+30,"Show CTF Image",memRecord.showCTFImage) ;
	AddObject(mShowCTFImage) ;
	
	mShowNoiseImage = new TCheckBox((optionsRect.right+optionsRect.left)/2,optionsRect.top+10,
		optionsRect.right,optionsRect.top+30,"Show Noise Image",memRecord.showNoiseImage) ;
	AddObject(mShowNoiseImage) ;
		
	mOKButton = new TButton(optionsRect.right-70,mPortRect.bottom-40,optionsRect.right,mPortRect.bottom-20,"Run");
	AddObject(mOKButton) ;
	mCancelButton = new TButton(optionsRect.right-150,mPortRect.bottom-40,optionsRect.right-80,mPortRect.bottom-20,"Cancel");
	AddObject(mCancelButton) ;

	// Add text edit objects in reverse order to get the tab order correct
	AddTextEditObject(mNumConjugateStepsText) ;
	AddTextEditObject(mNumIterationsText) ;
	AddTextEditObject(mLagrangeMultiplierText) ;
	AddTextEditObject(mPercentageText) ;
	AddTextEditObject(mDivergenceText) ;
	AddTextEditObject(mDeltaText) ;
	AddTextEditObject(mDefocusText) ;
	AddTextEditObject(mCsText) ;
	AddTextEditObject(mVoltageText) ;
	
	mDefaultButton = mOKButton ;
}

TRunMEMDialog::~TRunMEMDialog() {
	delete[] mWindowListIndex ;
}

void	TRunMEMDialog::DoMouseDownInObject(TObject *obj,Point *pt) {
	if (pt != nil)
		if (!obj->DoMouseDown(pt)) return ;
	mIsVisible = true ;
	if (obj == mOKButton) {
		mDone = true ; 
		mDialogReturnCode = kDialogOK ;
	}
	else if (obj == mCancelButton) {
		mDone = true ;
		mDialogReturnCode = kDialogCancel ;
	}
	else if (obj == mNoiseWienerFilter) {
		mNoisePercentage->SetControlValue(false) ;
		mNoiseImage->SetControlValue(false) ;
	}
	else if (obj == mNoisePercentage) {
		mNoiseWienerFilter->SetControlValue(false) ;
		mNoiseImage->SetControlValue(false) ;
	}
	else if (obj == mNoiseImage) {
		mNoiseWienerFilter->SetControlValue(false) ;
		mNoisePercentage->SetControlValue(false) ;
	}
}

void	TRunMEMDialog::DoKeyDownInObject(TObject *obj) {
	char	text[256] ;
	float	percentage ;
	if (obj == mPercentageText) {
		mPercentageText->GetEditText(text) ;
		percentage = atof(text) ;
		if (percentage != 0.) {
			mNoisePercentage->SetControlValue(true) ;
			mNoiseWienerFilter->SetControlValue(false) ;
			mNoiseImage->SetControlValue(false) ;
		}
	}
}

void	TRunMEMDialog::GetDialogData(void *data) {

	MEMRecord	*memRecord = (MEMRecord *)data ;
	
	// Retrieve the dialog data 
	char	text[256] ;
	
	mVoltageText->GetEditText(text) ;
	memRecord->voltage = atof(text) ;
	
	mCsText->GetEditText(text) ;
	memRecord->cs = atof(text) ;
	
	mDefocusText->GetEditText(text) ;
	memRecord->defocus = atof(text) ;
	
	mDeltaText->GetEditText(text) ;
	memRecord->delta = atof(text) ;
	
	mDivergenceText->GetEditText(text) ;
	memRecord->divergence = atof(text) ;
	
	mPercentageText->GetEditText(text) ;
	memRecord->percentage = atof(text) ;
	
	mLagrangeMultiplierText->GetEditText(text) ;
	memRecord->lagrangeMultiplier = atof(text) ;
	
	mNumIterationsText->GetEditText(text) ;
	memRecord->numIterations = atoi(text) ;
	
	mNumConjugateStepsText->GetEditText(text) ;
	memRecord->numConjugateSteps = atoi(text) ;


	memRecord->showCTFImage = mShowCTFImage->GetControlValue() ;
	memRecord->showNoiseImage = mShowNoiseImage->GetControlValue() ;
		
	
	if (mNoiseWienerFilter->GetControlValue())
		memRecord->noiseType = kUseNoiseFromWienerFilter ;
	else if (mNoisePercentage->GetControlValue())
		memRecord->noiseType = kUsePercentageNoise ;
	else if (mNoiseImage->GetControlValue()) {
		memRecord->noiseType = kUseNoiseImage ;
		// Get the name associated with the noise 
		mNoiseImageMenuChoice = mNoiseImageMenu->GetSelection() ;
		memRecord->noiseImageWindowListIndex = -1 ;
		if (mNoiseImageMenuChoice > 1) {
			memRecord->noiseImageWindowListIndex = mWindowListIndex[mNoiseImageMenuChoice-2] ;
			mNoiseImageMenu->GetMenuItemText(mNoiseImageMenuChoice,memRecord->noiseImageName) ;	
		}
	}
	
}
