// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include "TScaleDialog.h"
#include <math.h>
#include <stdlib.h>
extern	long near(float x) ;
TScaleDialog::TScaleDialog(TScaleDialogRecord *data)
		: TDialog(kDialogPosX,kDialogPosY,270,160,"Scale Image") {
	TGroupBox *grpBox ;
	mData = *data ;
	char	text[256] ;

	grpBox = new TGroupBox(20,15,250,65,"In Pixels") ;
	AddObject(grpBox) ;
	grpBox = new TGroupBox(20,75,250,125,"In Percentage") ;
	AddObject(grpBox) ;

	sprintf(text,"%d",mData.widthInPixels) ;
	mWidthPixelC = new TEditText(60,30,120,50,"Width",text) ;
	sprintf(text,"%d",mData.heightInPixels) ;
	mHeightPixelC = new TEditText(170,30,230,50,"Height",text) ;
	mWidthPercentC = new TEditText(60,90,120,110,"Width","100.00") ;
	mHeightPercentC = new TEditText(170,90,230,110,"Height","100.00") ;
	mKeepAspectRatioC = new TCheckBox(15,130,115,150,"Keep Aspect Ratio") ;	
	mKeepAspectRatio = true ;
	mKeepAspectRatioC->SetControlValue(mKeepAspectRatio) ;
	AddObject(mKeepAspectRatioC) ;

	// Add the edit fields in reverse order to get proper tab order
	AddTextEditObject(mHeightPercentC) ;
	AddTextEditObject(mWidthPercentC) ;
	AddTextEditObject(mHeightPixelC) ;
	AddTextEditObject(mWidthPixelC) ;

	mOKButton = new TButton(210,130,250,150,"OK") ;
	AddObject(mOKButton) ;
	mCancelButton = new TButton(130,130,200,150,"Cancel") ;
	AddObject(mCancelButton) ;

	mAspectRatio = ((float)mData.heightInPixels)/mData.widthInPixels ;
	mDefaultButton = mOKButton ;

}

void	TScaleDialog::DoKeyDownInObject(TObject *obj) {
	char	text[256] ;
	long	width,height ;
	float	percX,percY ;
	if (obj == mWidthPixelC) {
		mWidthPixelC->GetEditText(text) ;
		width = atoi(text) ;
		percX = 100.*width/mData.widthInPixels ;
		sprintf(text,"%.2f",percX) ;
		mWidthPercentC->SetEditText(text) ;
		if (mKeepAspectRatio) {
			height = near(mAspectRatio*width) ;
			sprintf(text,"%d",height) ;
			mHeightPixelC->SetEditText(text) ;
			percY = percX ;
			sprintf(text,"%.2f",percY) ;
			mHeightPercentC->SetEditText(text) ;
		}
		return ;
	}
	if (obj == mHeightPixelC) {
		mHeightPixelC->GetEditText(text) ;
		height = atoi(text) ;
		percY = 100.*height/mData.heightInPixels ;
		sprintf(text,"%.2f",percY) ;
		mHeightPercentC->SetEditText(text) ;
		if (mKeepAspectRatio) {
			width = near(height/mAspectRatio) ;
			sprintf(text,"%d",width) ;
			mWidthPixelC->SetEditText(text) ;
			percX = percY ;
			sprintf(text,"%.2f",percX) ;
			mWidthPercentC->SetEditText(text) ;
		}
		return ;
	}
	if (obj == mWidthPercentC) {
		mWidthPercentC->GetEditText(text) ;
		percX = atof(text) ;
		width = near(percX*mData.widthInPixels/100.) ;
		sprintf(text,"%d",width) ;
		mWidthPixelC->SetEditText(text) ;
		if (mKeepAspectRatio) {
			percY = percX ;
			sprintf(text,"%.2f",percY) ;
			mHeightPercentC->SetEditText(text) ;
			height = near(percY*mData.heightInPixels/100.) ;
			sprintf(text,"%d",height) ;
			mHeightPixelC->SetEditText(text) ;
		}
		return ;
	}
	if (obj == mHeightPercentC) {
		mHeightPercentC->GetEditText(text) ;
		percY = atof(text) ;
		height = near(percY*mData.heightInPixels/100.) ;
		sprintf(text,"%d",height) ;
		mHeightPixelC->SetEditText(text) ;
		if (mKeepAspectRatio) {
			percX = percY ;
			sprintf(text,"%.2f",percX) ;
			mWidthPercentC->SetEditText(text) ;
			width = near(percX*mData.widthInPixels/100.) ;
			sprintf(text,"%d",width) ;
			mWidthPixelC->SetEditText(text) ;
		}
		return ;
	}
}


void	TScaleDialog::DoMouseDownInObject(TObject *obj,Point *pt){
	if (pt != nil)
		if (!obj->DoMouseDown(pt)) return ;
	char	text[256] ;
	long	width,height ;
	float	percY ;
	if (obj == mOKButton) {
		mDone = true ;
		mWidthPixelC->GetEditText(text) ;
		mData.widthInPixels = atoi(text) ;
		mHeightPixelC->GetEditText(text) ;
		mData.heightInPixels = atoi(text) ;
		mWidthPercentC->GetEditText(text) ;
		mData.widthInPercent = atof(text) ;
		mHeightPercentC->GetEditText(text) ;
		mData.heightInPercent = atof(text) ;
		mDialogReturnCode = kDialogOK ;
		return ;
	}
	if (obj == mCancelButton) {
		mDone = true ;
		mDialogReturnCode = kDialogCancel ;
		return ;
	}
	if (obj == mKeepAspectRatioC) {
		mKeepAspectRatio = mKeepAspectRatioC->GetControlValue() ;
		if (mKeepAspectRatio) {
			mWidthPixelC->GetEditText(text) ;
			width = atoi(text) ;
			height = near(mAspectRatio*width) ;
			height = near(percY*mData.heightInPixels/100.) ;
			sprintf(text,"%d",height) ;
			mHeightPixelC->SetEditText(text) ;
			mWidthPercentC->GetEditText(text) ;
			mHeightPercentC->SetEditText(text) ;
		}
		return ;
	}
}
