#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <time.h>
#include "TShowSolutionsWindow.h"
#include "XDrivers.h"
#include "DMMenuCommands.h"
#include "TPeaksControlWindow.h"

TImage *CreateImageFromTiffFile(char *filename) ;
TImage *CreateImageFromTiffFileQuietly(char *filename) ;

Boolean	FileExists(char *filename) ;

extern	void	SendApplicationCommand(TCommand command) ; 
extern	short	gImageDepth ;
extern	void 	DisplayImageInRect(TImage *image, Rect *rect) ;
extern	void	PutMessage(char *) ;
extern	char	*FS98_PATH ;
extern	char	*EDMBinDir ;

// Flag for whether we are using Uniq or not
bool	UseUnique = true ;

TSolutionTable::TSolutionTable(TableAttributes *attr) 
	: TTable(attr) {
	
	mImageList = nil ;
	mNumImages = 0 ;
	mTableCellsAreSelectable = true ;
}

TSolutionTable::~TSolutionTable() {
}

void TSolutionTable::DrawCellValue(
	TableCellT &inCell, 
	Rect &cellFrame)
{
	char str[64] ;
	
	
	TextSize(0);
	Rect	theRect = cellFrame ;
	
	InsetRect(&theRect,20,4) ;
	
	if (inCell.row <= mNumImages && mImageList != nil) {
		if (mImageList[inCell.row-1] != nil) {
				MoveTo(cellFrame.left + 2, (cellFrame.bottom + cellFrame.top)/2 + 4) ;
				sprintf(str,"%3d",inCell.row-1+mStartingNumber) ;
				DrawString(str) ;
				DisplayImageInRect(mImageList[inCell.row-1], &theRect) ;
				FrameRect(&theRect) ;
		}			
	}
}


TShowSolutionsWindow::TShowSolutionsWindow(FS98ControlRecord &dataRecord) 
	: TWindow(kDialogPosX,kDialogPosY,580,400,gImageDepth,"Analyze Feasible Set Solutions",false) {

	Rect	theRect,groupBoxRect,tableRect ;
	TStaticText	*theText ;
	TGroupBox *theBox ;
	char	text[256] ;
	long	nn ;
	
	mBkColor = gGrayPixel ;
	XSetBackground(gDisplay,mGC,mBkColor) ;
	
	mFS98ControlRecord = dataRecord ;
			
	// Define Window Controls
	
	SetRect(&tableRect,10,20,210,390) ;
	
	TableAttributes	tableAtt ;
	tableAtt.nRows = 0 ;
	tableAtt.nCols = 1 ;
	tableAtt.DataSize = 4 ;
	tableAtt.ColWidth = tableRect.right-tableRect.left-16 ;
	tableAtt.RowHeight = 120 ;
	tableAtt.WindowWidth = tableRect.right-tableRect.left ;
	tableAtt.WindowHeight = tableRect.bottom-tableRect.top ;
	tableAtt.AdornColumns = false ;
	tableAtt.AdornRows = false ;
	tableAtt.HasColumnLabels = false ;
	tableAtt.HasRowLabels = false ;
	tableAtt.WindowTitle = nil ;


	if (!CountGeneralSolutions()) {
		PutMessage("Unable to find a 'solutions' file") ;
	}
	
	mNumUniqueSolutions = 0 ;
	
	mDataTable = new TSolutionTable(&tableAtt) ;
	
	mTablePane = new TTablePane(this,tableRect.left,tableRect.top,tableRect.right-tableRect.left,tableRect.bottom-tableRect.top,mDataTable) ;
	AddPane(mTablePane) ;
	mTablePane->Show() ;


	theRect = tableRect ;
	theBox = new TGroupBox(theRect.right+20,theRect.top,mPortRect.right-10,theRect.top+80,"General Solutions Set") ;
	AddObject(theBox) ;
	groupBoxRect = theBox->GetRect() ;

	theText = new TStaticText(groupBoxRect.left+10,groupBoxRect.top+10,groupBoxRect.left+140,groupBoxRect.top+30,
		"Number of General Solutions") ;
	AddObject(theText) ;
	
	theRect = theText->GetRect() ;

	sprintf(text,"%d",mNumGeneralSolutions) ;
	mNumGeneralSolutionsText = new TStaticText(theRect.right+10,theRect.top,theRect.right+50,theRect.bottom,text) ;
	AddObject(mNumGeneralSolutionsText) ;

	mDisplaySolutionsButton = new TButton(groupBoxRect.left+10,theRect.bottom+10,groupBoxRect.left+140,theRect.bottom+40,"Display Solutions") ;
	if (mNumGeneralSolutions == 0) mDisplaySolutionsButton->Disable() ;
	AddObject(mDisplaySolutionsButton) ;
	
	theRect = mDisplaySolutionsButton->GetRect() ;

	mFromText = new TEditText(theRect.right+40,theRect.top+5,theRect.right+80,theRect.bottom-5,"from","1") ;
	theRect = mFromText->GetRect() ;
	mToText = new TEditText(theRect.right+40,theRect.top,theRect.right+80,theRect.bottom,"to","1") ;

	
	theBox = new TGroupBox(groupBoxRect.left,groupBoxRect.bottom+20,mPortRect.right-10,groupBoxRect.bottom+170,"Unique Solutions Set") ;
	AddObject(theBox) ;
	groupBoxRect = theBox->GetRect() ;
		
	mCrossCorrelationToleranseText = new TEditText(groupBoxRect.right-50,groupBoxRect.top+10,groupBoxRect.right-10,groupBoxRect.top+30,
		"Use selection criteria based on cross correlation difference of","0.05") ;

	theRect = mCrossCorrelationToleranseText->GetRect() ;

	nn = mNumGeneralSolutions < 300 ? mNumGeneralSolutions : 300 ;
	sprintf(text,"%d",nn) ;
	mNumberOfTopSolutionsText = new TEditText(theRect.left-60,theRect.bottom+10,theRect.left-20,theRect.bottom+30,"from top",text) ;
	theRect = mNumberOfTopSolutionsText->GetRect() ;
	
	theText = new TStaticText(theRect.right+10,theRect.top,groupBoxRect.right-10,theRect.bottom,"solutions") ;
	AddObject(theText) ;

	theRect = theText->GetRect() ;

	mCalculateUniqueSolutionsButton	= new TButton(groupBoxRect.left+10,theRect.top-5,groupBoxRect.left+170,theRect.bottom+5,"Calculate Unique Solutions") ;
	if (mNumGeneralSolutions == 0) mCalculateUniqueSolutionsButton->Disable() ;
	AddObject(mCalculateUniqueSolutionsButton) ;
	
	theRect = mCalculateUniqueSolutionsButton->GetRect() ;

	theText = new TStaticText(theRect.left,theRect.bottom+10,theRect.left+140,theRect.bottom+30,
		"Number of Unique solutions") ;
	AddObject(theText) ;

	theRect = theText->GetRect() ;

	sprintf(text,"%d",mNumUniqueSolutions) ;
	mNumUniqueSolutionsText = new TStaticText(theRect.right+10,theRect.top,theRect.right+50,theRect.bottom,text) ;
	AddObject(mNumUniqueSolutionsText) ;
		
	mDisplayUniqueSolutionsButton = new TButton(theRect.left,theRect.bottom+10,theRect.left+160,theRect.bottom+40,"Display Unique Solutions") ;
	CountUniqueSolutions() ;
	sprintf(text,"%d",mNumUniqueSolutions) ;
	mNumUniqueSolutionsText->SetItemText(text) ;
	if (mNumUniqueSolutions == 0)
		mDisplayUniqueSolutionsButton->Disable() ;
	else 
		mDisplayUniqueSolutionsButton->Enable() ;
	AddObject(mDisplayUniqueSolutionsButton) ;
	
	theRect = mDisplayUniqueSolutionsButton->GetRect() ;

	mUniqueFromText = new TEditText(theRect.right+40,theRect.top+5,theRect.right+80,theRect.bottom-5,"from","1") ;
	theRect = mUniqueFromText->GetRect() ;
	mUniqueToText = new TEditText(theRect.right+40,theRect.top,theRect.right+80,theRect.bottom,"to","1") ;
	

	mCreateSolutionImageButton = new TButton(groupBoxRect.left,groupBoxRect.bottom+20,groupBoxRect.left+200,groupBoxRect.bottom+50,
		"Display selected solution in window") ;
	mCreateSolutionImageButton->Disable() ;
	AddObject(mCreateSolutionImageButton) ;

	mAnalyzeSolutionButton = new TButton(groupBoxRect.right-160,tableRect.bottom-26,groupBoxRect.right,tableRect.bottom+4,
		"Analyze selected solution...") ;
	mAnalyzeSolutionButton->Disable() ;
	AddObject(mAnalyzeSolutionButton) ;

	mCloseButton = new TButton(groupBoxRect.left,tableRect.bottom-26,groupBoxRect.left+100,tableRect.bottom+4,
		"Close") ;
	AddObject(mCloseButton) ;

	AddTextEditObject(mUniqueToText) ;
	AddTextEditObject(mUniqueFromText) ;
	AddTextEditObject(mNumberOfTopSolutionsText) ;
	AddTextEditObject(mCrossCorrelationToleranseText) ;
	AddTextEditObject(mToText) ;
	AddTextEditObject(mFromText) ;
	
	mImageList	= nil ;
	
		// Display the Unique solutions if they exist
		long ifrom=1 ;
		long i, imageID ;
		bool FoundUniq, AtEnd ;
		if (mNumUniqueSolutions > 0) {
			FoundUniq = true ; AtEnd = true ;
			mImageList = new TImagePtr[mNumUniqueSolutions] ;
			for(i = 0 ; i < mNumUniqueSolutions; i++) mImageList[i] = nil ;
			mNumImages = 0 ;
			for(i = 0 ; i < mNumUniqueSolutions; i++) {
				imageID = ifrom+i ;
				sprintf(text,"Uniq%d.tif",imageID) ;
				mImageList[i] = CreateImageFromTiffFileQuietly(text) ;
				if (mImageList[i] != nil) {
					mNumImages++ ;
					if ( !FoundUniq ) AtEnd = false ;//Trap where image in the middle lost
				}
				else
					FoundUniq = false ; //Trap where image is not available
			}
			if( FoundUniq || AtEnd ) { 
				mDataTable->SetImages(mImageList,mNumImages,ifrom) ;
				mTablePane->SetTableSize(mNumImages,1) ;
			}
			else {
				delete mImageList ; 
				mImageList = nil ;
				mNumImages = 0 ;
			}
		}
		if (mNumImages > 0){
			mAnalyzeSolutionButton->Enable() ;
			sprintf(text,"%d",mNumImages) ;
			mUniqueToText->SetEditText(text) ;
		}
}

TShowSolutionsWindow::~TShowSolutionsWindow() {
	if (mImageList != nil) {
		for(int i = 0 ; i < mNumImages; i++) {
			if (mImageList[i] != nil) delete mImageList[i] ;
		}
		delete[] mImageList ;
	}	
}

bool	TShowSolutionsWindow::CountGeneralSolutions() {
	// count the number of lines in the file "solutions"
	mNumGeneralSolutions = 0 ;
	FILE *fp = fopen("solutions","r") ;
	if (fp == nil) {
		return false ;
	}
	char c ;
	for(;;) {
		c = getc(fp) ;
		if ( c == EOF) break ;
		if (c == '\n' || c == '\r')
			mNumGeneralSolutions++ ;
	}
	fclose(fp) ;
	return true ;
}
bool	TShowSolutionsWindow::CountUniqueSolutions() {
	// count the number of lines in the file "xsolutions"
	mNumUniqueSolutions = 0 ;
	FILE *fp = fopen("xfoms.dat","r") ;
	if (fp == nil) {
		return false ;
	}
	char c ;
	for(;;) {
		c = getc(fp) ;
		if ( c == EOF) break ;
		if (c == '\n' || c == '\r')
			mNumUniqueSolutions++ ;
	}
	fclose(fp) ;
	return true ;
}


void	TShowSolutionsWindow::Draw() {	
	EraseRect(&mPortRect) ;	
	TPane::Draw() ;
}

void	TShowSolutionsWindow::DoMouseDown(XEvent *event) {
	TWindow::DoMouseDown(event) ;
	TableCellT selectedCell = mDataTable->GetSelectedCell() ;
	if (selectedCell.row > 0 && selectedCell.row <= mNumImages && selectedCell.col == 1) {
		mCreateSolutionImageButton->Enable() ;
		mAnalyzeSolutionButton->Enable() ; 
	}
}

void	TShowSolutionsWindow::DoKeyDownInObject(TObject *) {}

void	TShowSolutionsWindow::DoMouseDownInObject(TObject *obj,Point *pt) {
	if (pt != nil)
		if (!obj->DoMouseDown(pt)) return ;
		
	char	text[256],filename[256] ;
	long	i,numImages,ifrom,ito,fromNumRequestedSolutions,imageID ;
	float	ccc_toleranse ;
	if (obj == mCalculateUniqueSolutionsButton) {	
		mCalculateUniqueSolutionsButton->SetState(true) ;
		// Run the shell command to calculate unique solutions
		mNumberOfTopSolutionsText->GetEditText(text) ;
		fromNumRequestedSolutions = atoi(text) ;
		
		mCrossCorrelationToleranseText->GetEditText(text) ;
		ccc_toleranse = atof(text) ;
		
		sprintf(text,"%s%sxfom.sh %s %d %.2f",FS98_PATH,EDMBinDir,mFS98ControlRecord.baseName,fromNumRequestedSolutions,ccc_toleranse) ;
		system(text) ;
		
		CountUniqueSolutions() ;
		sprintf(text,"%d",mNumUniqueSolutions) ;
		mNumUniqueSolutionsText->SetItemText(text) ;
		if (mNumUniqueSolutions == 0)
			mDisplayUniqueSolutionsButton->Disable() ;
		else 
			mDisplayUniqueSolutionsButton->Enable() ;
			mCalculateUniqueSolutionsButton->SetState(false) ;
			UseUnique = true ;
	}
	else if (obj == mDisplaySolutionsButton) {
		mDisplaySolutionsButton->SetState(true) ;
		mFromText->GetEditText(text) ;
		ifrom = atoi(text) ;
		
		if (ifrom < 1) {
			PutMessage("Starting value must be greater or equal to 1") ;
			mDisplaySolutionsButton->SetState(false) ;
			return ;
		}
		
		mToText->GetEditText(text) ;
		ito = atoi(text) ;
		
		if (ito < ifrom) {
			sprintf(text,"Ending value must be greater or equal to starting value of %d",ifrom) ;
			PutMessage(text) ;
			mDisplaySolutionsButton->SetState(false) ;
			return ;
		}
		if (ito > mNumGeneralSolutions) {
			sprintf(text,"Ending value must be less or equal to %d",mNumGeneralSolutions) ;
			PutMessage(text) ;
			mDisplaySolutionsButton->SetState(false) ;
			return ;
		}
		
		// Run the shell command to Display general solutions
		sprintf(text,"%s%sfs2D.sh %s %d %d",FS98_PATH,EDMBinDir,mFS98ControlRecord.baseName,ifrom,ito) ;
		system(text) ;
		
		// Now display the solutions requested from the general set
		numImages = ito - ifrom + 1 ;
		if (numImages > 0) {
			if (mImageList != nil) {
				for(i = 0 ; i < mNumImages; i++) {
					if (mImageList[i] != nil) {
						delete mImageList[i] ;
						mImageList[i] = nil ;
					}
				}
				delete[] mImageList ;
				mImageList = nil ;
			}
			mNumImages = numImages ;
			mImageList = new TImagePtr[mNumImages] ;
			for(i = 0 ; i < mNumImages; i++) mImageList[i] = nil ;
			mNumImages = 0 ;
			for(i = 0 ; i < numImages; i++) {
				imageID = ifrom+i ;
				sprintf(filename,"Ient%d.tif",imageID) ;
				mImageList[i] = CreateImageFromTiffFile(filename) ;
				if (mImageList[i] != nil) mNumImages++ ;
			}
			mDataTable->SetImages(mImageList,mNumImages,ifrom) ;
			mTablePane->SetTableSize(mNumImages,1) ;
		}
		mDisplaySolutionsButton->SetState(false) ;
		UseUnique = false ;
	}
	else if (obj == mDisplayUniqueSolutionsButton) {
		mDisplayUniqueSolutionsButton->SetState(true) ;
		mUniqueFromText->GetEditText(text) ;
		ifrom = atoi(text) ;

		if (ifrom < 1) {
			PutMessage("Starting value must be greater or equal to 1") ;
			mDisplayUniqueSolutionsButton->SetState(false) ;
			return ;
		}
		
		mUniqueToText->GetEditText(text) ;
		ito = atoi(text) ;

		if (ito < ifrom) {
			sprintf(text,"Ending value must be greater or equal to starting value of %d",ifrom) ;
			PutMessage(text) ;
			mDisplayUniqueSolutionsButton->SetState(false) ;
			return ;
		}
		if (ito > mNumUniqueSolutions) {
			sprintf(text,"Ending value must be less or equal to %d",mNumUniqueSolutions) ;
			PutMessage(text) ;
			mDisplayUniqueSolutionsButton->SetState(false) ;
			return ;
		}
		
		// Run the shell command to Display unique solutions
		sprintf(text,"%s%sfsx2D.sh %s %d %d",FS98_PATH,EDMBinDir,mFS98ControlRecord.baseName,ifrom,ito) ;
		system(text) ;

		
		// Now display the solutions requested from the unique set
		numImages = ito - ifrom + 1 ;
		if (numImages > 0) {
			if (mImageList != nil) {
				for(i = 0 ; i < mNumImages; i++) {
					if (mImageList[i] != nil) {
						delete mImageList[i] ;
						mImageList[i] = nil ;
					}
				}
				delete[] mImageList ;
				mImageList = nil ;
			}
			mNumImages = numImages ;
			mImageList = new TImagePtr[mNumImages] ;
			for(i = 0 ; i < mNumImages; i++) mImageList[i] = nil ;
			mNumImages = 0 ;
			for(i = 0 ; i < numImages; i++) {
				imageID = ifrom+i ;
				sprintf(filename,"Uniq%d.tif",imageID) ;
				mImageList[i] = CreateImageFromTiffFile(filename) ;
				if (mImageList[i] != nil) mNumImages++ ;
			}
			mDataTable->SetImages(mImageList,mNumImages,ifrom) ;
			mTablePane->SetTableSize(mNumImages,1) ;
		}
		mDisplayUniqueSolutionsButton->SetState(false) ;
	}
	else if (obj == mCreateSolutionImageButton) {
		TableCellT selectedCell = mDataTable->GetSelectedCell() ;
		if (selectedCell.row > 0 && selectedCell.row <= mNumImages && selectedCell.col == 1) {
			if (mImageList[selectedCell.col-1] != nil) {
				TImage	*theImage = new TImage(*mImageList[selectedCell.row-1]) ;
				theImage->SetName(mImageList[selectedCell.row-1]->Name()) ;
				TableCellT theCell ;
				theCell.col = 0 ;
				theCell.row = 0 ;
				theImage->Display() ;
				mDataTable->SetSelectedCell(theCell) ;
				mAnalyzeSolutionButton->Disable() ;
				mCreateSolutionImageButton->Disable() ;
			}
		}
	}
	else if (obj == mAnalyzeSolutionButton) {
		TableCellT selectedCell = mDataTable->GetSelectedCell() ;
		if (selectedCell.row > 0 && selectedCell.row <= mNumImages && selectedCell.col == 1) {
			if (mImageList[selectedCell.col-1] != nil) {
				// Activate the peaks program
				// --------------------------
				RunPeaks(selectedCell.row) ;
				TableCellT theCell ;
				theCell.col = 0 ;
				theCell.row = 0 ;
				mDataTable->SetSelectedCell(theCell) ;
				mCreateSolutionImageButton->Disable() ;
				mAnalyzeSolutionButton->Disable() ;
			}
		}
	}
	else if (obj == mCloseButton) {
		TableCellT theCell ;
		theCell.col = 0 ;
		theCell.row = 0 ;
		mDataTable->SetSelectedCell(theCell) ;
		mAnalyzeSolutionButton->Disable() ;
		mCreateSolutionImageButton->Disable() ;
		SendApplicationCommand(kClose) ;
	}
}

void CreateFile(char *filename) ;
void CreateFile(char *filename) {
	FILE *fp = fopen(filename,"w") ;
	fclose(fp) ;
}

void	TShowSolutionsWindow::RunPeaks(short whichImage) {

	// Create the window from which everything will run
	new TPeaksControlWindow(mFS98ControlRecord,mImageList[whichImage-1],whichImage) ;
}

