// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include "TTable.h"
#include <stdio.h>
#include "globals.h"
#include "XDrivers.h"
#define	MonitorWidth	1024
#define	MonitorHeight	768 
#include "TFile.h"
#include "DMMenuCommands.h"
#include "TDMApplicationWindow.h"
// Trap here for Keypad Left/Right not being defined
#ifndef XK_KP_Left
#define XK_KP_Left XK_Left
#define XK_KP_Right XK_Right
#endif

extern	Atom	gWindowManagerAtoms[2];
typedef enum { PSEUDOCOLOR, TRUECOLOR } MonitorDisplayType ;
extern MonitorDisplayType  gMonitorDisplayType ;
extern Visual	*gVisual ;
extern	unsigned char	gStippleData[] ;
void	RemoveWindow(TWindow *) ;

extern	XFontStruct	*gMenuFont ;
extern	XFontStruct	*gDialogFont ;
extern	XFontStruct	*gWindowFont ;
extern	short	gNumGrayLevels ;

extern	Boolean	IsDoubleClick() ;
extern	GC	gGC ;
extern	Pixmap	gStipple ;

extern	TDMApplicationWindow *gTheControlWindow ;

long	near(float) ;

extern	uLong	gWhitePixel,gBlackPixel,gDarkerGray,gGrayPixel ;
extern	void	PutMessage(char *) ;
void	SetTargetWindow(TWindow *window) ;
void	SendApplicationCommand(TCommand command) ;
#define EOS '\0'

extern	short	gImageDepth ;
//SaveCode PromptSave(char *) ;

TTable::TTable(char *title) 
	: mRowLabels((Int32)(16*sizeof(char))),
	mColLabels((Int32)(16*sizeof(char)))
{
	strcpy(mTitle,title) ;
	mHasColLabels = true ;
	mHasRowLabels = true ;
	mAdornRows = true ;
	mAdornColumns = true ;
	mRows = 1 ;
	mCols = 1 ;
	mCellDataSize = 4 ;
	mColWidth = 60 ;
	mRowHeight = 20 ;
	mCellData = nil ;
	mEditField = nil ;
	mTableCellsAreSelectable = false ;
	
	mSelectedCell.row = 0 ;
	mSelectedCell.col = 0 ; 

	if (mCellDataSize > 0) {
		mCellData = new LArray((UInt32)mCellDataSize);
		mCellData->InsertItemsAt(mRows * mCols, 1, nil);
	}

	mRowLabels.InsertItemsAt(mRows, 1, nil);
	mColLabels.InsertItemsAt(mCols, 1, nil);
}

TTable::TTable(TableAttributes *att) 
	: mRowLabels((Int32)(16*sizeof(char))),
	mColLabels((Int32)(16*sizeof(char)))
{

	if (att->WindowTitle != nil)
		strcpy(mTitle,att->WindowTitle) ;
	else
		strcpy(mTitle,"") ;
	mHasColLabels = att->HasColumnLabels ;
	mHasRowLabels = att->HasRowLabels ;
	mAdornRows = att->AdornRows ;
	mAdornColumns = att->AdornColumns ;
	mRows = att->nRows ;
	mCols = att->nCols ;
	mCellDataSize = att->DataSize ;
	mColWidth = att->ColWidth ;
	mRowHeight = att->RowHeight ;
	mCellData = nil ;
	mEditField = nil ;
	mTableCellsAreSelectable = false ;
	
	mSelectedCell.row = 0 ;
	mSelectedCell.col = 0 ; 

	if (mCellDataSize > 0) {
		mCellData = new LArray((UInt32)mCellDataSize);
		mCellData->InsertItemsAt(mRows * mCols, 1, nil);
	}

	mRowLabels.InsertItemsAt(mRows, 1, nil);
	mColLabels.InsertItemsAt(mCols, 1, nil);
}


TTable::TTable(
	Int16 rows,
	Int16 cols,
	Int16 cellSize,
	Int16 colWidth,
	Int16 rowHeight,	
	char *tableTitle) 
	: mRowLabels((Int32)(16*sizeof(char))),
	mColLabels((Int32)(16*sizeof(char)))
	
{
	
	mHasColLabels = true ;
	mHasRowLabels = true ;
	mAdornRows = true ;
	mAdornColumns = true ;
	mRows = rows ;
	mCols = cols ;
	mCellDataSize = cellSize ;
	mColWidth = colWidth ;
	mRowHeight = rowHeight ;
	mCellData = nil ;
	
	strcpy(mTitle,tableTitle) ;

	mSelectedCell.row = 0 ;
	mSelectedCell.col = 0 ;

	if (mCellDataSize > 0) {
		mCellData = new LArray((UInt32)mCellDataSize);
		mCellData->InsertItemsAt(mRows * mCols, 1, nil);
	}

	mRowLabels.InsertItemsAt(mRows, 1, nil);
	mColLabels.InsertItemsAt(mCols, 1, nil);
	
	Int16 	xpos = 20 ;
	Int16 	ypos = 88 ;
				
	mTableCellsAreSelectable = false ;
	mParentPane = new TTableWindow(xpos,ypos,this) ;	
	mEditField = nil ;
}


void	
TTable::SetTableSize(Int16 nrows, Int16 ncols) {
	mRows = nrows ;
	mCols = ncols ;

	mSelectedCell.row = 0 ;
	mSelectedCell.col = 0 ; 

	if (mCellData != nil) {
		delete mCellData ;
		mCellData = new LArray((UInt32)mCellDataSize);
		mCellData->InsertItemsAt(mRows * mCols, 1, nil);
	}
	
	mRowLabels.RemoveAllItemsAfter(0) ;
	mColLabels.RemoveAllItemsAfter(0) ;

	mRowLabels.InsertItemsAt(mRows, 1, nil);
	mColLabels.InsertItemsAt(mCols, 1, nil);

}

TTable::~TTable() {
	if (mEditField != nil) delete mEditField ;
	if (mCellData != nil) delete mCellData ;
}

void	
TTable::GetCellText(TableCellT &inCell,char *text) {
	sprintf(text,"%d,%d",inCell.col,inCell.row) ;
}

void	
TTable::SetCellText(TableCellT &inCell,char *text) {
}

void	
 TTable::SetColLabel(
	Int16 inCol,
	char *title) 
{
	mColLabels.AssignItemsAt(1,inCol,(void *)title,(UInt32)(16*sizeof(char))) ;
}

void	
 TTable::SetRowLabel(
	Int16 inRow,
	char *title)
{
	mRowLabels.AssignItemsAt(1,inRow,(void *)title,(UInt32)(16*sizeof(char))) ;
}

void TTable::SetRowHeight(
	Int16 height) 
{
	mRowHeight = height ; 
}

void TTable::SetColWidth(
	Int16 width) 
{
	mColWidth = width ;
}

void TTable::Draw(RgnHandle updateRegion) 
{
	Int16 row,col ;
	Point	topLeftUpdate,botRightUpdate,topLeft,botRight;
	Rect	totalUpdateRect ;
	
	if (!mParentPane->IsVisible()) {
		fprintf(stderr,"Parent Pane not visible\n") ;
		return ;
	}
	
	FPoint	theTableOrigin = mParentPane->GetPortOrigin() ;
	
	if (updateRegion == nil) {
		totalUpdateRect = mCanvasRect ;
	}
	else {
		// Get the region associated with the update
		RegionRect(updateRegion,&totalUpdateRect) ;
		
		if (totalUpdateRect.top == 0 && totalUpdateRect.left == 0 &&
			totalUpdateRect.bottom == 0 && totalUpdateRect.right == 0) {
				return ;
		}
	}
	TStOffscreen stOffscreen(mParentPane,&mCanvasRect) ; 

	Int16 rightEdge,bottomEdge,savedLeft ;
	rightEdge = mCanvasRect.right ;
	bottomEdge = mCanvasRect.bottom ;
	
	if (totalUpdateRect.right > rightEdge) {
		totalUpdateRect.right = rightEdge ;
	}
	if (totalUpdateRect.bottom > bottomEdge) {
		totalUpdateRect.bottom = bottomEdge ;
	}
	savedLeft = totalUpdateRect.left ;
	if (mHasRowLabels) {
		DrawColLabel(0) ;
		if (totalUpdateRect.left < mColWidth) {
			totalUpdateRect.left = mColWidth ;
		}
	}

	SetClip(totalUpdateRect) ;
	
	if (mHasColLabels) {
		// Find colum label at top left of update rect
		Int16	leftColumn,rightColumn;	
		// Calculate the column intersected by the top left point
		// of the update region
		topLeft.x = totalUpdateRect.left ;
		topLeft.y = totalUpdateRect.top ;
		if (topLeft.y <= mRowHeight) {
			//need to draw  column labels
			leftColumn = (Int16)((topLeft.x-theTableOrigin.x-1)/mColWidth) ;
			if (leftColumn < 0) {
				leftColumn = 0 ;
			}
			
			botRight.x = totalUpdateRect.right ;
			rightColumn = (Int16)((botRight.x-theTableOrigin.x+1)/mColWidth) ;
			if (rightColumn > mCols) {
				rightColumn = mCols ;
			}
			if (!mHasRowLabels) {
				leftColumn++ ;
				rightColumn++ ;
			}
			
			for(col=leftColumn;col<=rightColumn;col++) {
				DrawColLabel(col) ;
			}
		}
	}
	
	if (mHasColLabels) {
		totalUpdateRect.left = savedLeft ;
		if (totalUpdateRect.top < mRowHeight)
			totalUpdateRect.top = mRowHeight ;
	}
	
	SetClip(totalUpdateRect) ;
	
	if (mHasRowLabels) {
		// Find colum label at top left of update rect
		Int16	topRow,bottomRow;	
		// Calculate the column intersected by the top left point
		// of the update region
		topLeft.x = totalUpdateRect.left ;
		topLeft.y = totalUpdateRect.top ;
		if (topLeft.x <= mColWidth) {
			// need to draw  row labels
			topRow = (Int16)((topLeft.y-theTableOrigin.y-1)/mRowHeight) ;
			if (topRow < 0) {
				topRow = 0 ;
			}
			
			botRight.y = totalUpdateRect.bottom ;
			bottomRow = (Int16)((botRight.y-theTableOrigin.y+1)/mRowHeight) ;
			if (bottomRow > mRows) {
				bottomRow = mRows ;
			}

			if (!mHasColLabels) {
				topRow++ ;
				bottomRow++ ;
			}
			for(row=topRow;row<=bottomRow;row++) {
				DrawRowLabel(row) ;
			}
		}
	}
	
	if (updateRegion == nil) {
		totalUpdateRect = mCanvasRect ;
	}
	else {
		// Get the region associated with the update
		RegionRect(updateRegion,&totalUpdateRect) ;
	}

	if (totalUpdateRect.right > rightEdge) {
		totalUpdateRect.right = rightEdge ;
	}
	if (totalUpdateRect.bottom > bottomEdge) {
		totalUpdateRect.bottom = bottomEdge ;
	}
	if (totalUpdateRect.top < mRowHeight && mHasColLabels) {
		totalUpdateRect.top = mRowHeight ;
	}
	if (totalUpdateRect.left < mColWidth  && mHasRowLabels) {
		totalUpdateRect.left = mColWidth ;
	}

	SetClip(totalUpdateRect) ;
	// ---------------------------------------
	// Draw the cells that need to be updated
	// ---------------------------------------
	
	// Find cell at top left of update rect
	TableCellT	topLeftCell;	
	// Calculate the cell intersected by the top left point
	// of the update region
	topLeft.x = totalUpdateRect.left - 1 ;
	topLeft.y = totalUpdateRect.top - 1 ;
	mParentPane->LocalToImagePoint(topLeft,topLeftUpdate) ;			
	FetchCellHitBy(topLeftUpdate, topLeftCell);
	topLeftCell.row-- ;
	topLeftCell.col-- ;
	if (topLeftCell.row < 1) {	// Lower bound is cell (1,1)
		topLeftCell.row = 1;
	}
	if (topLeftCell.col < 1) {
		topLeftCell.col = 1;
	}

								// Find cell at bottom right of update rect
	TableCellT	botRightCell;
	botRight.x = totalUpdateRect.right + 1 ;
	botRight.y = totalUpdateRect.bottom + 1 ;
	mParentPane->LocalToImagePoint(botRight, botRightUpdate);
	FetchCellHitBy(botRightUpdate, botRightCell);
	botRightCell.row++ ;
	botRightCell.col++ ;
								// Upper bound is cell (mRows,mCols)
	if (botRightCell.row > mRows) {
		botRightCell.row = mRows;
	}
	if (botRightCell.col > mCols) {
		botRightCell.col = mCols;
	}
	
								// Draw each cell within the update rect
	TableCellT	cell;
	for (cell.row = topLeftCell.row; cell.row <= botRightCell.row; cell.row++) {
		for (cell.col = topLeftCell.col; cell.col <= botRightCell.col; cell.col++) {
			DrawCell(cell);
		}
	}

}

void TTable::DoCommand(TCommand command,Boolean *hasBeenHandled) { 
	switch(command) {
		case kCopy:
			DoCopy() ;
			Refresh() ;
			*hasBeenHandled = true ;
			break ;
		case kPaste:
			DoPaste() ;
			Refresh() ;
			*hasBeenHandled = true ;
			break ;
		default:
			*hasBeenHandled = false ;
			break ;
	}
} 

void TTable::GetCommandStatus(TCommand command,Boolean *isEnabled,
							Boolean *isChecked) {
	*isChecked = false ; 
	switch(command) {
		case kCopy:
			*isEnabled = HasSelection() ;
			break ;	
		case kPaste:
			*isEnabled = HasSelection() && gTableClipBoardList != nil ;
			break ;	
		default:
			*isEnabled = false ;
			break ;
	}
} 

void TTable::DoPaste() {
}
 
void TTable::DoCopy() {
} 

void TTable::Print() 
{
	/*
	Int16 row,col ;
	RgnHandle clipRgn,oldRgn ;
	
	clipRgn = NewRgn() ;
	oldRgn = NewRgn() ;
	GetClip(oldRgn) ;
	RectRgn(clipRgn,&qd.thePort->portRect) ;
	SetClip(clipRgn) ;
	
	if (mHasColLabels) {
		DrawColLabel(0) ;
	}
	
	if (mHasColLabels) {
		for(col=1;col<=mCols;col++) {
			DrawColLabel(col) ;
		}
	}
		
	if (mHasRowLabels) {
		for(row=1;row<=mRows;row++) {
				DrawRowLabel(row) ;
		}
	}
	
	TableCellT	cell;
	for (cell.row = 1; cell.row <= mRows; cell.row++) {
		for (cell.col = 1; cell.col <=mCols; cell.col++) {
			DrawCell(cell);
		}
	}
	SetClip(oldRgn) ;
	*/
	
}

void	TTable::SetSelectedCell(TableCellT &inCell) {
	mSelectedCell = inCell ;
	mParentPane->Refresh() ;
}

void TTable::DrawCell(
	TableCellT &inCell) 
{
	Rect theRect ;
	Point CellImagePt,CellLocalPt ;
	
	bool	isSelected = mTableSelection.InSelection(inCell) ;
	
	if (mTableCellsAreSelectable) {
		isSelected = mSelectedCell.row == inCell.row && mSelectedCell.col == inCell.col ;
	}

	
	CellImagePt.x = (mHasRowLabels ? inCell.col*mColWidth : (inCell.col-1)*mColWidth) ;
	CellImagePt.y = (mHasColLabels ? inCell.row*mRowHeight : (inCell.row-1)*mRowHeight) ;
	mParentPane->ImageToLocalPoint(CellImagePt,CellLocalPt) ;
	SetRect(&theRect,CellLocalPt.x,CellLocalPt.y,CellLocalPt.x+mColWidth,
		CellLocalPt.y+mRowHeight) ;
	if (!(mSelectedCell.row == inCell.row && mSelectedCell.col == inCell.col)) {
			
		if (!isSelected) {
			SetFgColor(gGrayPixel) ;
			MoveTo(theRect.left,theRect.bottom) ;
			LineTo(theRect.left,theRect.top) ;
			LineTo(theRect.right,theRect.top) ;
			if (inCell.col == mCols) {
				LineTo(theRect.right,theRect.bottom) ;
			}
			if (inCell.row == mRows) {
				MoveTo(theRect.left,theRect.bottom) ;
				LineTo(theRect.right,theRect.bottom) ;
			}
			PenNormal() ;
			SetFgColor(gBlackPixel) ;
			DrawCellValue(inCell,theRect) ;
		}
		else {
			PenMode(kXOrMode) ;
			SetFgColor(gGrayPixel) ;
			MoveTo(theRect.left,theRect.bottom) ;
			LineTo(theRect.left,theRect.top) ;
			LineTo(theRect.right,theRect.top) ;
			if (inCell.col == mCols) {
				LineTo(theRect.right,theRect.bottom) ;
			}
			if (inCell.row == mRows) {
				MoveTo(theRect.left,theRect.bottom) ;
				LineTo(theRect.right,theRect.bottom) ;
			}
			SetFgColor(gBlackPixel) ;
			DrawCellValue(inCell,theRect) ;
			PenNormal() ;
		}
	} 
	else if (mEditField != nil) {
		
		TextSize(0);
		mEditField->SetObjectRect(theRect) ;
		mEditField->Draw() ;
		SetFgColor(gBlackPixel) ;
		
	}
	else if (mSelectedCell.row == inCell.row && mSelectedCell.col == inCell.col) {
		SetClip(mCanvasRect) ;
		PenMode(kXOrMode) ;
		SetFgColor(gGrayPixel) ;
		MoveTo(theRect.left,theRect.bottom) ;
		LineTo(theRect.left,theRect.top) ;
		LineTo(theRect.right,theRect.top) ;
		if (inCell.col == mCols) {
			LineTo(theRect.right,theRect.bottom) ;
		}
		if (inCell.row == mRows) {
			MoveTo(theRect.left,theRect.bottom) ;
			LineTo(theRect.right,theRect.bottom) ;
		}
		SetFgColor(gBlackPixel) ;
		DrawCellValue(inCell,theRect) ;
		PenMode(kXOrMode) ;
		PaintRect(&theRect) ;
		PenNormal() ;
	}
}

void TTable::DrawCellValue(
	TableCellT &inCell,
	Rect &cellFrame)
{
	char str[64] ;
	
	TextSize(0);
	MoveTo(cellFrame.left + 4, cellFrame.bottom - 4);
	sprintf(str,"%d,%d",inCell.row,inCell.col) ;
	DrawString(str) ;
}

bool
TTable::FetchLocalCellFrame(
	const TableCellT	&inCell,
	Rect				&outCellFrame)
{
								// Get Top-Left in Image coordinates
	Point	cellImage;
	
	if (mHasRowLabels)
		cellImage.x = inCell.col * mColWidth;
	else
		cellImage.x = (inCell.col - 1) * mColWidth;

	if (mHasColLabels)
		cellImage.y = inCell.row * mRowHeight;
	else
		cellImage.y = (inCell.row - 1) * mRowHeight;
	
	// Check if Cell intersects the Frame
	bool	intersectsFrame = ImageRectIntersectsFrame(
									cellImage.x, cellImage.y,
									cellImage.x + mColWidth,
									cellImage.y + mRowHeight);
	
	if (intersectsFrame) {		// Convert to Local coordinates
		mParentPane->ImageToLocalPoint(cellImage, topLeft(outCellFrame));
		outCellFrame.right = outCellFrame.left + mColWidth;
		outCellFrame.bottom = outCellFrame.top + mRowHeight;
	}
	
	return intersectsFrame;
}


bool
TTable::IsValidCell(
	const TableCellT	&inCell) const
{
	return ( (inCell.row > 0) && (inCell.row <= mRows) &&
			 (inCell.col > 0) && (inCell.col <= mCols) );
}


void
TTable::FetchCellHitBy(
	const Point		&inImagePt,
	TableCellT		&outCell)
{
	// Calculate the cell hit by the point (in image coordinates)
	Point CellPoint = inImagePt ;

	if (mHasRowLabels) {
		CellPoint.x -= mColWidth ;
	}
	if (mHasColLabels) {
		CellPoint.y -= mRowHeight ;
	}
	
	if (CellPoint.y < 0) 
		outCell.row = 0 ;
	else
		outCell.row = (CellPoint.y-1)/ mRowHeight + 1 ;
	if (CellPoint.x < 0) 
		outCell.col = 0 ;
	else
		outCell.col = (CellPoint.x-1)/ mColWidth + 1 ;
	
}

 
void TTable::SetCellData(
	TableCellT &inCell, 
	void *data) 
{
	Int32 cellIndex ;
	cellIndex = (inCell.row-1)*mCols+inCell.col ;
	mCellData->AssignItemsAt(1,cellIndex,data) ;
}

bool TTable::FetchCellData(
	TableCellT &inCell,
	void *data) 
{
	Int32 cellIndex ;
	cellIndex = (inCell.row-1)*mCols+inCell.col ;
	return(mCellData->FetchItemAt(cellIndex,data)) ;
}

void TTable::SetDataCells(
	TableCellT &inFromCell, 
	TableCellT &inToCell,
	void *data) 
{
	TableCellT cell ;
	char *dataLoc ;
	
	dataLoc = (char *)data ;
	for(cell.row=inFromCell.row;cell.row < inToCell.row;cell.row++) {
		for(cell.col=inFromCell.col;cell.col < inToCell.col;cell.col++) {
			SetCellData(cell,(void *)dataLoc) ;
			dataLoc += mCellDataSize ;
		}
	}
} 
void TTable::FetchDataCells(
	TableCellT &inFromCell, 
	TableCellT &inToCell,
	void *data) 
{
	TableCellT cell ;
	char *dataLoc ;
	
	dataLoc = (char *)data ;
	for(cell.row=inFromCell.row;cell.row < inToCell.row;cell.row++) {
		for(cell.col=inFromCell.col;cell.col < inToCell.col;cell.col++) {
			FetchCellData(cell,(void *)dataLoc) ;
			dataLoc += mCellDataSize ;
		}
	}
} 

void TTable::ClickCell(
	TableCellT &inCell) 
{
	if ((mClickTime-mLastClick) < kDoubleClickTime) {
		if (CellIsEditable(inCell)) EditCell(inCell) ;
	}
}

void	TTable::ClickColLabel(Int16 inCol) {
}

void	TTable::ClickRowLabel(Int16 inRow) {
}

void TTable::EditCell(
	TableCellT &inCell) 
{
	// We are here because the cell is editable, so create an edtit field and place it here
	Rect	theRect ;
	char	text[64] ;
	
	SetPort(mParentPane) ;
	SetFocus(mParentPane) ;
	if (!(mSelectedCell.row == inCell.row && mSelectedCell.col == inCell.col)) {
		if (mEditField != nil) {
			// Set the previous field
			mEditField->GetEditText(text) ;
			SetCellText(mSelectedCell,text) ;
			TableCellT theCell = mSelectedCell ;
			mSelectedCell.row = mSelectedCell.col = 0 ;
			delete mEditField ;
			mEditField = nil ;
			FetchLocalCellFrame(theCell,theRect) ;
			EraseRect(&theRect) ;
			DrawCell(theCell) ;
			XFlush(gDisplay) ;
		}
		
		mSelectedCell = inCell ;
		FetchLocalCellFrame(inCell,theRect) ;
		GetCellText(inCell,text) ;
		mEditField = new TTableEditField(&theRect,nil,text,(TWindow *)mParentPane) ;
		mEditField->SelectAll() ;
		mSelectedCell = inCell ;
	}
	DrawCell(inCell) ;
}

bool TTable::DoClickInLabels(
	XEvent *inEvent) 
{

	Point			theMouseLoc ;
	
	theMouseLoc.x = inEvent->xbutton.x ;
	theMouseLoc.y = inEvent->xbutton.y ;
	mClickTime 		= inEvent->xbutton.time ;
	
	TableCellT theCell ;
	
	// Se if we are in any of the column or row labels
	FetchCellHitBy(theMouseLoc,theCell) ;
	if (mHasRowLabels && theCell.col == 0) {
		if (theCell.row > 0 && theCell.row <= mRows)
			ClickRowLabel(theCell.row) ;
		return true ;
	}		
	if (mHasColLabels && theCell.row == 0) {
		if (theCell.col > 0 && theCell.col <= mCols) 
			ClickColLabel(theCell.col) ;
		return true ;
	}
	
	return false ;
}

bool 
TTable::ImageRectIntersectsFrame(
	Int16 left,
	Int16 top,
	Int16 right,
	Int16 bottom) 
{
	Point topLeft, bottomRight ;
	
	FPoint theTableOrigin = mParentPane->GetPortOrigin() ;
	
	topLeft.x = near(left + theTableOrigin.x) ;
	topLeft.y = near(top + theTableOrigin.y) ;
	
	bottomRight.x =  near(right + theTableOrigin.x) ;
	bottomRight.y =  near(bottom + theTableOrigin.y) ;
	
	Rect frame = mCanvasRect ;
	if (mHasRowLabels) {
		frame.left += mColWidth ;
	}
	if (mHasColLabels) {
		frame.top += mRowHeight ;
	}
	frame.right -= 16 ;	// Because of border possibly containing scrollbar
	frame.bottom -= 16 ; // likewise
	
	return (PtInRect(&topLeft,&frame) || PtInRect(&bottomRight,&frame)) ;
}


void TTable::DoClick(
	XEvent *inEvent) 
{

	Point	theMouseLoc ;
	GrafPtr	thePort ;
	
	theMouseLoc.x = inEvent->xbutton.x ;
	theMouseLoc.y = inEvent->xbutton.y ;
	mClickTime 		= inEvent->xbutton.time ;
	
	TableCellT theCell ;
	FetchCellHitBy(theMouseLoc,theCell) ;
	if (theCell.row > mRows || theCell.col > mCols) return ;

	if (mSelectedCell.row == theCell.row && mSelectedCell.col == theCell.col && mEditField != nil) {
		mEditField->DoMouseDown(&theMouseLoc) ;
	}
	else {		
		ClickCell(theCell) ;
		if (mTableCellsAreSelectable) {
			if (mSelectedCell.col > 0 && mSelectedCell.row > 0) {
				DrawCell(mSelectedCell) ;				
			}
			mSelectedCell = theCell ;
			DrawCell(theCell) ;
			GetPort(&thePort) ;
			XFlush(gDisplay) ;
		}
	}
		
	mLastClick = mClickTime ;
}

void	TTable::DoKeyDown(XEvent *theEvent) {
	if (mEditField == nil) return ;
	char	text[64] ;
	KeySym	theKey ;
	GrafPtr tPort ;
	GetPort(&tPort) ;
	SetPort(mParentPane) ;
	theKey = mEditField->DoKeyDown(theEvent) ;
	if (theKey == XK_Return) {
		mEditField->GetEditText(text) ;
		SetCellText(mSelectedCell,text) ;
		delete mEditField ;
		mEditField = nil ;
		mSelectedCell.row = 0 ;
		mSelectedCell.col = 0 ;
	}
	Draw(nil) ;

}

void TTable::HighlightCell(TableCellT &inCell) 
{
}


void TTable::DrawColLabel(
	Int16 colNumber) 
{
	Rect	cellFrame;
	FPoint	theTableOrigin = mParentPane->GetPortOrigin() ;
	
	if (mHasRowLabels)
		SetRect(&cellFrame,colNumber*mColWidth,0,(colNumber+1)*mColWidth,mRowHeight) ;
	else
		SetRect(&cellFrame,(colNumber-1)*mColWidth,0,colNumber*mColWidth,mRowHeight) ;
		
	if (colNumber != 0) 
		::OffsetRect(&cellFrame,near(theTableOrigin.x),0) ;
	InsetRect(&cellFrame,1,1) ;
	SetFgColor(gGrayPixel) ;
	PaintRect(&cellFrame) ;
	SetFgColor(gDarkerGray) ;
	MoveTo(cellFrame.left,cellFrame.bottom) ;
	LineTo(cellFrame.left,cellFrame.top) ;
	LineTo(cellFrame.right,cellFrame.top) ;
	SetFgColor(gBlackPixel) ;
	MoveTo(cellFrame.right,cellFrame.top) ;
	LineTo(cellFrame.right,cellFrame.bottom) ;
	LineTo(cellFrame.left,cellFrame.bottom) ;
	InsetRect(&cellFrame,-1,-1) ;
	TextSize(0);
	MoveTo(cellFrame.left + 4, cellFrame.bottom - 4);
	if (colNumber > 0) {
		DrawColLabelText(colNumber) ;
	}
}
void TTable::DrawRowLabel(
	Int16 rowNumber) 
{
	Rect	cellFrame;
	FPoint	theTableOrigin = mParentPane->GetPortOrigin() ;
	
	if (mHasColLabels)
		SetRect(&cellFrame,0,rowNumber*mRowHeight,mColWidth,(rowNumber+1)*mRowHeight) ;
	else
		SetRect(&cellFrame,0,(rowNumber-1)*mRowHeight,mColWidth,rowNumber*mRowHeight) ;
	if (rowNumber != 0)
		::OffsetRect(&cellFrame,0,near(theTableOrigin.y)) ;
	InsetRect(&cellFrame,1,1) ;
	SetFgColor(gGrayPixel) ;
	PaintRect(&cellFrame) ;
	SetFgColor(gDarkerGray) ;
	MoveTo(cellFrame.left,cellFrame.bottom) ;
	LineTo(cellFrame.left,cellFrame.top) ;
	LineTo(cellFrame.right,cellFrame.top) ;
	SetFgColor(gBlackPixel) ;
	MoveTo(cellFrame.right,cellFrame.top) ;
	LineTo(cellFrame.right,cellFrame.bottom) ;
	LineTo(cellFrame.left,cellFrame.bottom) ;
	InsetRect(&cellFrame,-1,-1) ;
	TextSize(0);
	MoveTo(cellFrame.left + 4, cellFrame.bottom - 4);
	if (rowNumber > 0) {
		DrawRowLabelText(rowNumber) ;
	}
}

void 
TTable::DrawRowLabelText(
	Int16 rowNumber) 
{
	char str[64] ;
	sprintf(str,"R%d",rowNumber) ;
	DrawString(str) ;
}

void 
TTable::DrawColLabelText(
	Int16 colNumber) 
{
	char str[64] ;
	sprintf(str,"C%d",colNumber) ;
	DrawString(str) ;
}


void TTable::Save() 
{
}

TTableWindow::TTableWindow(short xpos,short ypos,TTable *table)
	: TWindow()
{

	mTable	 = table ;
	bool	hasVertScrollBar = false ;
	bool	hasHorizScrollBar = false ;
	
	long	maxWindowWidth = ((long)((MonitorWidth-xpos-16)/mTable->mColWidth))*mTable->mColWidth ; // Integer number of columns
	long 	maxWindowHeight = ((long)((MonitorHeight-ypos-16)/mTable->mRowHeight))*mTable->mRowHeight ; // Integer number of rows
	
	long	tableWidth = mTable->mHasRowLabels ? mTable->mColWidth*(mTable->mCols+1) : mTable->mColWidth*mTable->mCols ;
	long	tableHeight = mTable->mHasColLabels ? mTable->mRowHeight*(mTable->mRows+1) : mTable->mRowHeight*mTable->mRows ;
	
	
	long	theWidth = (tableWidth > maxWindowWidth ? maxWindowWidth : tableWidth) ;
	long	theHeight = (tableHeight > maxWindowHeight ? maxWindowHeight : tableHeight) ;


	SetRect(&mCanvasRect,0,0,theWidth,theHeight) ;
	mTable->SetCanvasRect(mCanvasRect) ;
	mIsVisible = true ;

	mCanReceiveMenuCommands = true ;
	mCommander = gTheControlWindow ; 
	
	strcpy(mWindowTitle,table->mTitle) ;
	
	theHeight += 16 ;
	if (tableWidth > maxWindowWidth) {	
		hasHorizScrollBar = true ;
	}

	theWidth += 16 ;
	if (tableHeight > maxWindowHeight)	{
		hasVertScrollBar = true ;
	}
    
    SetRect(&mPortRect,0,0,theWidth,theHeight) ;
    
	XGCValues	gcv ;
	int	borderWidth = 4 ;
	mDepth = gImageDepth ;


	mXPosition = xpos ;
	mYPosition = ypos ;
    
	mFont = gWindowFont ;
 	switch (gMonitorDisplayType) { 
		case PSEUDOCOLOR :
		    mXWindow = XCreateWindow(gDisplay,RootWindow(gDisplay,gScreen),
		    				xpos,ypos,theWidth,theHeight,
		    				borderWidth,mDepth,InputOutput,
		    				gVisualInfo.visual,
		    				CWBackPixel | CWBorderPixel | CWColormap | CWBackingStore ,
		    				&gImageAttributes) ;
		    break ;
		case TRUECOLOR :
		    mXWindow = XCreateWindow(gDisplay,RootWindow(gDisplay,gScreen),
		    				xpos,ypos,theWidth,theHeight,
		    				borderWidth,mDepth,InputOutput,
		    				gVisualInfo.visual,
		    				CWBackPixel | CWBorderPixel | CWBackingStore ,
		    				&gImageAttributes) ;
		    break ;
	}

	mDrawable	= mXWindow ;

	XSizeHints sizeHints ;
	sizeHints.flags 	= USPosition | USSize | PMinSize | PMaxSize ;
	sizeHints.x	= xpos ;
	sizeHints.y	= ypos ;
	sizeHints.width	= theWidth ;
	sizeHints.height	= theHeight ;
	sizeHints.min_width 	= 2*mTable->mColWidth+16 ;
	sizeHints.min_height	=  2*mTable->mRowHeight+16 ;
	sizeHints.max_width 	= tableWidth + 16 ;
	sizeHints.max_height	=  tableHeight + 16 ;
	XSetStandardProperties(gDisplay,mXWindow,mWindowTitle,0,
    			0,0,0,&sizeHints) ;

	XWMHints theXWMHints ;
	theXWMHints.flags = StateHint ;
	theXWMHints.initial_state = NormalState ;	
	XSetWMHints(gDisplay,mXWindow,&theXWMHints) ;


	XSetWMProtocols(gDisplay,mXWindow,gWindowManagerAtoms,2) ;
	
	if (gMonitorDisplayType == PSEUDOCOLOR) {
	    XSetWindowColormap(gDisplay,mXWindow,gColorMap) ;
	    gcv.foreground = gBlackPixel ;
	    gcv.background = gWhitePixel ;
	}
	else {
	    XSetWindowColormap(gDisplay,mXWindow,gColorMap) ;
	    gcv.foreground = BlackPixel(gDisplay,gScreen) ;
	    gcv.background = WhitePixel(gDisplay,gScreen) ;
	}


	
    
	mGC = XCreateGC(gDisplay,mXWindow,GCForeground |
    			GCBackground, &gcv) ;
    
	XSetFont(gDisplay,mGC,mFont->fid) ;

	XSelectInput(gDisplay,mXWindow,ExposureMask | KeyPressMask | StructureNotifyMask |
    			FocusChangeMask | EnterWindowMask | LeaveWindowMask  |
    			KeyReleaseMask | ButtonPressMask | ButtonReleaseMask) ;
	/*
	XSelectInput(gDisplay,mXWindow,ExposureMask | KeyPressMask | StructureNotifyMask |
    			EnterWindowMask | LeaveWindowMask  |
    			KeyReleaseMask | ButtonPressMask | ButtonReleaseMask) ;
    			
    */
	if (gGC == nil) {
		gGC = mGC ;
	}

	if (gStipple == nil) {
		unsigned int rwidth,rheight ;
		XQueryBestStipple(gDisplay,mDrawable,4,4,&rwidth,&rheight) ;
		gStipple = XCreatePixmapFromBitmapData(gDisplay,mDrawable,
			(char *)gStippleData,rwidth,rwidth,
			1,0,1) ;
	}
	if (gStipple != nil) XSetStipple(gDisplay,mGC,gStipple) ;

	
	SetClipRegion(mPortRect) ;
	
	mObjectList = new TObjectList ;

	float	minVal = 0. ,maxVal ;
	
	Int16 widthOfDisplayedContent = mCanvasRect.right - mCanvasRect.left  ;
	Int16 heightOfDisplayedContent = mCanvasRect.bottom - mCanvasRect.top  ;

	if (mTable->mHasColLabels) 
		heightOfDisplayedContent -= mTable->mRowHeight ;
	if (mTable->mHasRowLabels) 
		widthOfDisplayedContent -= mTable->mColWidth ;
			
	Int16 vertScrollMax = mTable->mRowHeight*mTable->mRows - heightOfDisplayedContent ;
	Int16 horizScrollMax = mTable->mColWidth*mTable->mCols - widthOfDisplayedContent ;
	
	if (hasVertScrollBar) {
		maxVal	= vertScrollMax ;
			mVertScrollBar = new TScrollBar(mPortRect.right-16,0,mPortRect.right,mPortRect.bottom-16,kVertical,minVal,maxVal,minVal) ;
		mVertScrollBar->SetTrackingRoutine(TTableWindow::DoVScrollBar) ;		
		mVertScrollBar->SetRefCon((unsigned long)this) ;
		AddObject(mVertScrollBar) ;
	}

	if (hasHorizScrollBar) {
		maxVal	= horizScrollMax ;
		mHorizScrollBar =  new TScrollBar(0,mPortRect.bottom-16,mPortRect.right-16,mPortRect.bottom,kHorizontal,minVal,maxVal,minVal) ;
		mHorizScrollBar->SetTrackingRoutine(TTableWindow::DoHScrollBar) ;
		mHorizScrollBar->SetRefCon((unsigned long)this) ;
		AddObject(mHorizScrollBar) ;
	}
		
	RectRgn(mUpdateRegion,&mPortRect) ;
	mPortOrigin.x = 0. ;
	mPortOrigin.y = 0. ;
	mLastClick = 0 ;
	Show() ;

}

TTableWindow::~TTableWindow() 
{
	delete mTable ;
}


void	TTableWindow::Update(XEvent *event) {

	if (event->xexpose.window != mXWindow)
		return ;
	GrafPtr tPort ;
	XRectangle updateRect ;
	GetPort(&tPort) ;

	updateRect.x = event->xexpose.x ;
	updateRect.y = event->xexpose.y ;
	updateRect.width = event->xexpose.width ;
	updateRect.height = event->xexpose.height ;
	XUnionRectWithRegion(&updateRect,mUpdateRegion,mUpdateRegion) ;
	if (event->xexpose.count == 0) {
		SetPort(this) ;
		XSetRegion(gDisplay,mGC,mUpdateRegion) ;
		
		// Let us check to see if the window size has changed
		XWindowAttributes attr ;
		XGetWindowAttributes(gDisplay,mXWindow,&attr) ;
		int	width = attr.width ;
		int	height = attr.height ;
		if (width != mPortRect.right || height != mPortRect.bottom) {
			SetRect(&mPortRect,0,0,width,height) ;
			AdjustScrollBars() ;
		}

		Draw() ;
		XFlush(gDisplay) ;
		XSetClipMask(gDisplay,mGC,None) ;
		SetPort(tPort) ;
		XDestroyRegion(mUpdateRegion) ;
		mUpdateRegion = XCreateRegion() ;
	}
}

void	TTableWindow::Print() {
	mTable->Print() ;
}

void	TTableWindow::Save() {
	mTable->Save() ;
}
	
void	TTableWindow::DoMouseDown(XEvent *event) {
	Point	mouseLoc ;
	BringToFront(this) ;
	gFrontWindow = this ;
	SetPort(this) ;
	mouseLoc.x = event->xbutton.x ;
	mouseLoc.y = event->xbutton.y ;
	if (mHorizScrollBar != nil) {
		if (mHorizScrollBar->InObject(&mouseLoc)) {
			mHorizScrollBar->DoMouseDown(&mouseLoc) ;
			return ;
		}
	}
	if (mVertScrollBar != nil) {
		if (mVertScrollBar->InObject(&mouseLoc)) {
			mVertScrollBar->DoMouseDown(&mouseLoc) ;
			return ;
		}
	}
	DoClick(event) ;
}

void	TTableWindow::DoKeyDown(XEvent *event) { 
	TPane::DoKeyDown(event) ; 
	mTable->DoKeyDown(event) ; 
} 

void 	TTableWindow::DoGrow(XEvent *inEvent) {
}

bool	TTableWindow::RequestClose() { 
	/*
	SaveCode status = PromptSave("Do you want to save the table data?") ;
	switch(status) {
		case kCancelClose:
			return false ;
		case kSave:
			Save() ;
			return true ;
		case kNoSave:
			return true ;
	}
	*/
	return true ; 
} 

void	TTableWindow::DoCommand(TCommand command,Boolean *hasBeenHandled) {	
	switch(command) {
		case kSave:		
			Save() ;
			*hasBeenHandled = true ;
			break ;
		default:
			mTable->DoCommand(command,hasBeenHandled) ;
			break ;
	}
}

void	TTableWindow::GetCommandStatus(TCommand command,Boolean *isEnabled,
			Boolean *isChecked) {
	switch(command) {
		case kClose:
			*isEnabled = true ;
			break ;		
		case kSave:	
			*isEnabled = true ;
			break ;	
		default:
			mTable->GetCommandStatus(command,isEnabled,isChecked) ;
			break ;
	}
}

void	TTableWindow::DoWMRequestDelete() {
	// Send the application the close command for this window
	SetTargetWindow(this) ;
	SendApplicationCommand(kClose) ;
}

void TTableWindow::Draw() 
{
	
	SetClip(mPortRect) ;
	
	mTable->Draw(nil) ;

	Rect	theRect  ;
	SetFgColor(gGrayPixel) ;
	SetRect(&theRect,mPortRect.right-16,0,mPortRect.right,mPortRect.bottom) ;
	FillRect(&theRect) ;
	SetFgColor(gWhitePixel) ;
	DrawLine(mPortRect.right-16,0,mPortRect.right-16,mPortRect.bottom) ;
	SetFgColor(gBlackPixel) ;
	DrawLine(mPortRect.right,0,mPortRect.right,mPortRect.bottom) ;
	SetFgColor(gGrayPixel) ;
	SetRect(&theRect,0,mPortRect.bottom-16,mPortRect.right,mPortRect.bottom) ;
	FillRect(&theRect) ;
	SetFgColor(gWhitePixel) ;
	DrawLine(0,mPortRect.bottom-16,mPortRect.right,mPortRect.bottom-16) ;
	SetFgColor(gBlackPixel) ;
	DrawLine(0,mPortRect.bottom,mPortRect.right,mPortRect.bottom) ;

	SetFgColor(gWhitePixel) ;
	DrawLine(mPortRect.right-16,mPortRect.bottom-16,mPortRect.right-16,mPortRect.bottom) ;
	DrawLine(mPortRect.right-16,mPortRect.bottom-16,mPortRect.right,mPortRect.bottom-16) ;
	SetFgColor(gBlackPixel) ;
	
	if(mObjectList!=nil)
		mObjectList->DrawObjects();	
			
	SetFgColor(gBlackPixel) ;
}

void TTableWindow::ResizeWindow(long width,long height) {	
	//SetRect(&mPortRect,0,0,width,height) ;
	//AdjustScrollBars() ;

}

void TTableWindow::AdjustScrollBars() {
	
	// Check to see if we still need the scrollbar(s) 
	// If not, delete them, toherwize adjust their size and position
	long	tableWidth = mTable->mHasRowLabels ? mTable->mColWidth*(mTable->mCols+1) : mTable->mColWidth*mTable->mCols ;
	long	tableHeight = mTable->mHasColLabels ? mTable->mRowHeight*(mTable->mRows+1) : mTable->mRowHeight*mTable->mRows ;
	
	long	windowWidth = mPortRect.right - mPortRect.left ;
	long	windowHeight = mPortRect.bottom - mPortRect.top ;
	
	

	
	bool	hasVertScrollBar = tableHeight > windowHeight-16 ? true : false ;
	bool	hasHorizScrollBar = tableWidth > windowWidth-16 ? true : false ;
	
	Int16	horizWidth = windowWidth - 16 ;
	Int16	vertHeight = windowHeight - 16 ;

	Int16 vertScrollMax = tableHeight - vertHeight ;
	Int16 horizScrollMax = tableWidth - horizWidth ;
	
	float	vertScrollPos = 0. ;
	float	horizScrollPos = 0. ;
		
	if (hasHorizScrollBar) {
		// We need a horizontal scrollbar
		// Do we have one already
		if (mHorizScrollBar != nil)
			mHorizScrollBar->Resize(0,windowHeight-16,horizWidth,windowHeight,horizScrollMax) ;
		else {
			// We need to create one
			mHorizScrollBar =  new TScrollBar(0,windowHeight-16,
					horizWidth,windowHeight,kHorizontal,0,horizScrollMax,0);
			mHorizScrollBar->SetTrackingRoutine(TTableWindow::DoHScrollBar) ;
			mHorizScrollBar->SetRefCon((unsigned long)this) ;
			AddObject(mHorizScrollBar) ;
		}
		horizScrollPos	= mHorizScrollBar->GetControlValue() ;
	}
	else {
		// Do we have one already
		if (mHorizScrollBar != nil) {
			RemoveObject(mHorizScrollBar) ;
			mHorizScrollBar = nil ;
		}
	}

	if (hasVertScrollBar) {
		// We need a vertical scrollbar
		// Do we have one already
		if (mVertScrollBar != nil)
			mVertScrollBar->Resize(windowWidth-16,0,windowWidth,vertHeight,vertScrollMax) ;
		else {
			// We need to create one
			mVertScrollBar =  new TScrollBar(windowWidth-16,0,
					windowWidth,vertHeight,kVertical,0,vertScrollMax,0);
			mVertScrollBar->SetTrackingRoutine(TTableWindow::DoVScrollBar) ;		
			mVertScrollBar->SetRefCon((unsigned long)this) ;
			AddObject(mVertScrollBar) ;
		}
		vertScrollPos	= mVertScrollBar->GetControlValue() ;
	}
	else {
		// Do we have one already
		if (mVertScrollBar != nil) {
			RemoveObject(mVertScrollBar) ;
			mVertScrollBar = nil ;
		}
	}

	SetRect(&mCanvasRect,0,0,horizWidth,vertHeight) ;
	mTable->SetCanvasRect(mCanvasRect) ;
	mPortOrigin.x = -horizScrollPos ;
	if (mPortOrigin.x > 0.) mPortOrigin.x = 0. ;
	mPortOrigin.y = -vertScrollPos ;
	if (mPortOrigin.y > 0.) mPortOrigin.y = 0. ;

}



void TTableWindow::ScrollTableBy(
	float deltaX,
	float deltaY) 
{
	mPortOrigin.x -= deltaX ;
	if (mPortOrigin.x > 0.)
		mPortOrigin.x = 0. ;
	mPortOrigin.y -= deltaY ;
	if (mPortOrigin.y > 0.)
		mPortOrigin.y = 0. ;
	Refresh() ;
}

void TTableWindow::ScrollTableTo(
	float x,
	float y) 
{
	mPortOrigin.x = -x ;
	if (mPortOrigin.x > 0.)
		mPortOrigin.x = 0. ;
	mPortOrigin.y = -y ;
	if (mPortOrigin.y > 0.)
		mPortOrigin.y = 0. ;
	Refresh() ;
}

void
TTableWindow::Refresh() 
{
	GrafPtr	tPort ;
	GetPort(&tPort) ;
	SetPort(this) ;
	DisposeRgn(mUpdateRegion) ;
	mUpdateRegion = NewRgn() ;
	
	RectRgn(mUpdateRegion,&mCanvasRect) ;
	
	mTable->Draw(mUpdateRegion) ;
	
	DisposeRgn(mUpdateRegion) ;
	mUpdateRegion = NewRgn() ;
}


void 
TTableWindow::DoHScrollBar(TObject	*object) {
	float	newVal ;
	float	scrollDistance ;
	TTableWindow	*theTableWindow ;
	
	theTableWindow = (TTableWindow *)object->GetRefCon() ;
	newVal	= theTableWindow->mHorizScrollBar->GetControlValue() ;

	scrollDistance = newVal - theTableWindow->mHScrollPos ;
	if (scrollDistance != 0.) {
		theTableWindow->mHScrollPos = newVal ;
		theTableWindow->ScrollTableTo(theTableWindow->mHScrollPos,theTableWindow->mVScrollPos) ;
	}
}

void 
TTableWindow::DoVScrollBar(TObject	*object) {
	float	newVal ;
	float	scrollDistance ;
	TTableWindow	*theTableWindow ;
	
	theTableWindow = (TTableWindow *)object->GetRefCon() ;
	newVal	= theTableWindow->mVertScrollBar->GetControlValue() ;

	scrollDistance = newVal - theTableWindow->mVScrollPos ;
	if (scrollDistance != 0) {
		theTableWindow->mVScrollPos = newVal ;
		theTableWindow->ScrollTableTo(theTableWindow->mHScrollPos,theTableWindow->mVScrollPos) ;
	}
}	

void TTableWindow::DoClick(
	XEvent *inEvent) 
{

	Point			theMouseLoc ;
	
	theMouseLoc.x = inEvent->xbutton.x ;
	theMouseLoc.y = inEvent->xbutton.y ;
	mClickTime 		= inEvent->xbutton.time ;
	
	if (!mTable->DoClickInLabels(inEvent)) {
		Point imagePoint ;
		LocalToImagePoint(theMouseLoc,imagePoint) ;
		inEvent->xbutton.x = imagePoint.x ;
		inEvent->xbutton.y = imagePoint.y ;
		mTable->DoClick(inEvent) ;
	}		
	mLastClick = mClickTime ;
}

Boolean TTableEditField::DoMouseDown(Point *pt) {
	Rect	tRect,clipRect ;
	int 	x,y ;
	int	EndStrXPos,numChars ;
	int 	xBegPos,yBegPos,xCurrPos,yCurrPos ;

	XFontStruct *font = (gCurrentPort == nil) ? gFontStructure : gCurrentPort->GetFont() ;

	xBegPos = mRect.left + 5 ; 	// Beginning point of the string 
	yBegPos = mRect.bottom - 5 ;// Geginning point of the string
	numChars = strlen(mText) ;	// Current number of characters	
	EndStrXPos = xBegPos + XTextWidth(font,mText,numChars) ;
	xCurrPos = mInsertionPt.x ;		// Current Point to put character
	yCurrPos = mInsertionPt.y ;
	
	if (IsDoubleClick()) {
		SelectAll() ;
		Draw() ;
		return true ;
	}

	RgnHandle	savedClip = NewRgn() ;
	GetClip(savedClip) ;
	
	clipRect	= mRect ;
	InsetRect(&clipRect,2,2) ;
	SetClip(clipRect) ;
		
	int textPlacement = (mRect.top+mRect.bottom)/2 + font->ascent/2 ;
	if (mNumSelChars == numChars) {
		// The selection must be cleared
		// We do this whether we have 
		// selection or not	
		mFirstSelChar = 0 ;
		mNumSelChars = 0 ;
		tRect = mRect ;
		InsetRect(&tRect,2,2) ;
		SetFgColor(gColors[near(0.8125*gNumGrayLevels)].pixel) ;
		FillRect(&tRect) ;
		SetFgColor(gBlackPixel) ;
		DrawText(xBegPos,textPlacement,mText) ;
	}
	if (numChars == 0) {	// There are no characters in this field yet.
		mInsertionPt.x = xBegPos ;
		mInsertionPt.y = yBegPos ;
		SetClip(savedClip) ;
		DisposeRgn(savedClip) ;
		return true ;
	}
	// Only need to find the insertion point 
	x = pt->x ;
	y = pt->y ;

	Point	currentPt=*pt,lastPt = *pt,startPt = *pt ;
	Rect	selRect ;
	SetRect(&selRect,x,mRect.top+3,x,mRect.bottom-2) ;
	int right,left ;
	SetFgColor(BlackPixel(gDisplay,gScreen)) ;
	SetBkColor(WhitePixel(gDisplay,gScreen)) ;
	XSetFunction(gDisplay,gCurrentPort->GetGC(),GXinvert) ;
	while(Button()) {
		GetMouse(&currentPt) ;
		if (currentPt.x != lastPt.x) {
			left = (currentPt.x < lastPt.x ? currentPt.x : lastPt.x) ;
			right = (currentPt.x > lastPt.x ? currentPt.x : lastPt.x) ;
			SetRect(&selRect,left,selRect.top,right,selRect.bottom) ;
			FillRect(&selRect) ;
			lastPt = currentPt ;
		}
	}
	PenNormal() ;
	SetFgColor(gBlackPixel) ;
	SetBkColor(gGrayPixel) ;
	SetClip(savedClip) ;
	DisposeRgn(savedClip) ;

	left = (currentPt.x < startPt.x ? currentPt.x : startPt.x) ;
	right = (currentPt.x > startPt.x ? currentPt.x : startPt.x) ;
	if (left != right) {
		mFirstSelChar = FindCharacterToRight(left) ;
		int	lastSelChar = FindCharacterToLeft(right) ;
		mNumSelChars = lastSelChar - mFirstSelChar + 1 ;
		if (mNumSelChars == 0) 
			mFirstSelChar = 0 ;
	}

	x = left ;
	if (x <= xBegPos) {	// Change the x current position 
		xCurrPos = xBegPos ;
	}
	else if (x > xBegPos && x < EndStrXPos) { // Change x pos.
		xCurrPos = xBegPos + XTextWidth(font,
			mText,FindCharacterToRight(x)) ;
	}
	else {
		xCurrPos = EndStrXPos ; // Append at the end 
	}
	mInsertionPt.x = xCurrPos ;
	mInsertionPt.y = yBegPos ;
	Draw() ;
	SetFgColor(gBlackPixel) ;
	SetBkColor(gWhitePixel) ;
	return true ;
}
KeySym 	TTableEditField::DoKeyDown(XEvent *theEvent) {
	Rect	tRect,clipRect ;
	int 	i,x,y,height,width,descent,nInChars ;
	int		inString=false,EndStrXPos,numPrevChar,numChars ;
	int 	xBegPos,yBegPos,xCurrPos,yCurrPos,count ;
	KeySym 	keysym ;
	XComposeStatus compose ;
	extern	GrafPtr gCurrentPort ;
	char 	*cpst,InputBuffer[80],lastCharString[2],
		savestring[80],endString[80] ;
		
	if (!mIsActive) return 0 ;
		
	XFontStruct *font = (gCurrentPort == nil) ? gFontStructure : gCurrentPort->GetFont() ;

	xBegPos = mRect.left + 5 ; 	// Beginning point of the string 
	yBegPos = mRect.bottom - 5 ;// Geginning point of the string
	numChars = strlen(mText) ;	// Current number of characters	
	EndStrXPos = xBegPos + 
    		XTextWidth(font,mText,numChars) ;
	xCurrPos = mInsertionPt.x ;		// Current Point to put character
	yCurrPos = mInsertionPt.y ;
	height = font->ascent ;	// Height of text
	descent = font->descent ;	// Descent below baseline

	// Obviously this control is already the active
	// edit text field, so process characters

	int textPlacement = (mRect.top+mRect.bottom)/2 + font->ascent/2 ;

	// What is the character
	count = XLookupString((XKeyEvent *)theEvent,InputBuffer,80,&keysym,&compose) ; 
	
	//fprintf(stderr,"keysym = %c , Integer %d , Hex: %X\n",keysym,keysym,keysym) ;
 	// Do we have a current selection
	clipRect = mRect ;
	InsetRect(&clipRect,2,2) ;
	
	if (mNumSelChars > 0 || mNumSelChars == numChars) {
		if (keysym == XK_Tab) {
			// The selection must be cleared */
			mFirstSelChar = 0 ;
			mNumSelChars = 0 ;
			tRect = mRect ;
			InsetRect(&tRect,3,3) ;
			SetFgColor(gColors[near(0.8125*gNumGrayLevels)].pixel) ;
			FillRect(&tRect) ;
			SetFgColor(gBlackPixel) ;
			SetClip(clipRect) ;
			DrawText(xBegPos,textPlacement,mText) ;
			SetClip(gCurrentPort->GetPortRect()) ;
		}
		else if (keysym == XK_Return || 
				keysym == XK_KP_Enter || keysym == XK_Linefeed) {
				// don't do anything here
		}
// Added by LDM
		else if (keysym == XK_Left || keysym == XK_Right ||
			keysym == XK_KP_Left || keysym == XK_KP_Right	) {
				// don't do anything here
		}
		
		else { // Erase the selection and change the
    			// Insertion point */
			if (mNumSelChars == numChars) {	// The whole string is selected
				tRect = mRect ;
				InsetRect(&tRect,3,3) ;
				SetClip(clipRect) ;
				SetFgColor(gColors[near(0.8125*gNumGrayLevels)].pixel) ;
				FillRect(&tRect) ;
 				SetFgColor(gBlackPixel) ;
  				mText[0] = EOS ;
				numChars = strlen(mText)  ;
				EndStrXPos = xBegPos + XTextWidth(font,mText,numChars) ;
				mInsertionPt.x = xBegPos ;
				mInsertionPt.y = yBegPos ;
				mFirstSelChar = 0 ;
				mNumSelChars = 0 ;
				SetClip(gCurrentPort->GetPortRect()) ;
			}
			else {	// The selection is not the entire string
				if (mFirstSelChar > 0) {
					strncpy(savestring,mText,mFirstSelChar) ;
					savestring[mFirstSelChar] = EOS ;
				}
				if (mFirstSelChar+mNumSelChars < numChars) {	// Save the end
					strcpy(endString,mText+mFirstSelChar+mNumSelChars) ;
					strcpy(mText,savestring) ;
					strcat(mText,endString) ;
				}
				else {
					strcpy(mText,savestring) ;
				}
				SetClip(clipRect) ;
				x = xBegPos ;
				width = XTextWidth(font,mText,strlen(mText)) ;
				y = yBegPos ;
				SetFgColor(gColors[near(0.8125*gNumGrayLevels)].pixel) ;
				FillRect(&clipRect) ;
 				SetFgColor(gBlackPixel) ;
				mInsertionPt.x = x + XTextWidth(font,mText,
					mFirstSelChar) ;
				DrawText(xBegPos,textPlacement,mText) ;
				SetClip(gCurrentPort->GetPortRect()) ;
				mInsertionPt.y = yBegPos ;
				numChars = strlen(mText) ;
				EndStrXPos = xBegPos + XTextWidth(font,mText,numChars) ;
				mFirstSelChar = 0 ;
				mNumSelChars = 0 ;
			}
			if (keysym == XK_Delete || keysym == XK_BackSpace) {
				return(true)  ; // We are done 
			}
		}
	}
	// Is the insertion point within the string
	xCurrPos = mInsertionPt.x ;
	yCurrPos = mInsertionPt.y ;
	numPrevChar = -1 ; // Initialize to beginning of str
	if (xCurrPos == xBegPos) { // Beginning of string 
		inString = true ;
		numPrevChar = -1 ;
	}
	else if (xCurrPos < EndStrXPos) { /* Inside the string */
		for (i=0;i<numChars;i++) {
			x = xBegPos + XTextWidth(font,mText,i+1) ;
			if (x >= xCurrPos)
				break ;
		}
		inString = true ;
		// The number of the prev char (starting at 0)
		numPrevChar = i ; 
	}
	else {
		numPrevChar = numChars - 1;
	}

	if (keysym == XK_Tab || keysym == XK_Return || 
			keysym == XK_KP_Enter || keysym == XK_Linefeed) {
		// These characters are special, and we	
		// return the symbol 
		return(keysym) ;
	}
// Added by LDM
		else if (keysym == XK_Left || keysym == XK_Right ||
			keysym == XK_KP_Left || keysym == XK_KP_Right	) {
			return(keysym) ;
		}
	else if (keysym == XK_Delete || keysym == XK_BackSpace) {
		if (numPrevChar >= 0) { // Erase a character
			lastCharString[0] = mText[numPrevChar];
    			lastCharString[1] = EOS ;
			SetClip(clipRect) ;
			if (inString) {
				// We must contract the string
				// The character to erase is # numPrevChar
				cpst = mText+numPrevChar ;
				x = xBegPos + XTextWidth(font,mText,numPrevChar) ;
				width = XTextWidth(font,cpst,numChars-numPrevChar) ;
				y = yBegPos ;
				SetRect(&tRect,x,y-height,x+width,y+descent) ;
				SetFgColor(gColors[near(0.8125*gNumGrayLevels)].pixel) ;
				FillRect(&tRect) ;
 				SetFgColor(gBlackPixel) ;
				*cpst = EOS ; // Erases the char
				// Append the rest of the string
				strcat(mText,cpst+1) ;
				DrawText(xBegPos,textPlacement,mText);
   			}
			else {
				width = XTextWidth(font,lastCharString,1) ;
				mText[numPrevChar] = EOS ;
				x = EndStrXPos - width ;
				y = yBegPos ;
				SetRect(&tRect,x,y-height,x+width,y+descent) ;
				SetFgColor(gColors[near(0.8125*gNumGrayLevels)].pixel) ;
				FillRect(&tRect) ;
 				SetFgColor(gBlackPixel) ;
			}
			SetClip(gCurrentPort->GetPortRect()) ;
			mInsertionPt.x = x ;
			mInsertionPt.y = y ;
		}
	}
	else if (keysym >= XK_Shift_L && keysym <= XK_Hyper_R) {
		// Nothing, Modifier keys
	}
	else if (keysym >= XK_F1 && keysym <= XK_F35) {
		// Nothing, function keys 
	}
// Added by LDM
		else if (keysym == XK_Left || keysym == XK_Right ||
			keysym == XK_KP_Left || keysym == XK_KP_Right	) {
				// don't do anything here
		}
	else {
		InputBuffer[1] = '\0' ;
		if (inString) {
			// We must expand the string
			SetClip(clipRect) ;
			nInChars = 1 ; // always just one character // strlen(InputBuffer) ;
			cpst = mText+numPrevChar+1 ;
			strcpy(savestring,cpst) ; // Save the end
			if (numChars > 0) {
				x = xBegPos + XTextWidth(font,mText,numPrevChar+1) ;
				width = XTextWidth(font,cpst,numChars-numPrevChar-1) ;
				y = yBegPos ;
				SetRect(&tRect,x,y-height,x+width,y+descent) ;
				SetFgColor(gColors[near(0.8125*gNumGrayLevels)].pixel) ;
				FillRect(&tRect) ;
				SetFgColor(gBlackPixel) ;
			}
			*cpst = EOS ;
			// Append the input string
			strcat(mText,InputBuffer) ;
			strcat(mText,savestring) ;
			SetClip(clipRect) ;
			DrawText(xBegPos,textPlacement,mText);
			SetClip(gCurrentPort->GetPortRect()) ;
			x = xBegPos + XTextWidth(font,mText,
				numPrevChar+1+nInChars) ;
				mInsertionPt.x = x ;
		}
		else {
			nInChars = 1 ; // always just one character // strlen(InputBuffer) ;
			cpst = mText + numPrevChar + 1 ;
			SetClip(clipRect) ;
			for(i=0;i<nInChars;i++) {
				*cpst++ = InputBuffer[i] ;
				*cpst++ = EOS ;
				DrawChar(xCurrPos,textPlacement,InputBuffer[i]) ;
				xCurrPos = xBegPos + XTextWidth(font,mText,
				strlen(mText)) ;
			}
			SetClip(gCurrentPort->GetPortRect()) ;
			mInsertionPt.x = xCurrPos ;
		}
	}
	if (mActionRoutine != nil) {
		(*mActionRoutine)(this) ;
	}		
	Draw() ;
	return(true) ;
}


void 	TTableEditField::Draw() {
	Rect	tRect,clipRect,selectionRect ;
	unsigned long oldFg ;
	extern	uLong gBlackPixel,gWhitePixel ;
	int		numChars ;
	int 	textPlacement ;

	
	oldFg = GetFgColor() ;
	tRect = mRect ;
	clipRect = mRect ;
	InsetRect(&tRect,2,2) ;
	InsetRect(&clipRect,2,2) ;
	EraseRect(&tRect) ;
	

	RgnHandle clipRgn = NewRgn() ;
	GetClip(clipRgn) ;

	XFontStruct *font = (gCurrentPort == nil) ? gFontStructure : gCurrentPort->GetFont() ;
	numChars = strlen(mText) ;
	textPlacement = (mRect.top+mRect.bottom)/2 + font->ascent/2 ;

	SetFgColor(gBlackPixel) ;
	DrawLine(tRect.left,tRect.bottom,tRect.left,tRect.top) ;
	DrawLine(tRect.left,tRect.top,tRect.right,tRect.top) ;
	SetFgColor(gWhitePixel) ;
	DrawLine(tRect.right,tRect.top,tRect.right,tRect.bottom) ;
	DrawLine(tRect.right,tRect.bottom,tRect.left,tRect.bottom) ;
	PenSize(1,1) ;
	SetFgColor(gColors[near(0.8125*gNumGrayLevels)].pixel) ;
	FillRect(&tRect) ;
	SetFgColor(gBlackPixel) ;
	// If there is no selection we will draw a caret to indicate 
	// the insertion point
	SetClip(clipRect) ;	// Set clip so we don't draw outside our area
	if (mNumSelChars == 0) {
		DrawText(mRect.left+5,textPlacement,mText) ; 
		DrawLine(mInsertionPt.x,mRect.top+2,mInsertionPt.x,mRect.bottom-2) ;
	}
	else {
		// Invert the area that is selected
		int	x,y,left,right ;
		x = mRect.left+5 ;
		y = textPlacement ;
		DrawText(x,y,mText) ;
		SetFgColor(gBlackPixel) ;
		SetBkColor(gWhitePixel) ;
		XSetFunction(gDisplay,gCurrentPort->GetGC(),GXinvert) ;
		left = mRect.left+5+
				XTextWidth(font,mText,mFirstSelChar) ;			
		right = left+XTextWidth(font,mText+mFirstSelChar,
					mNumSelChars) ;
		if (mFirstSelChar == 0 && mNumSelChars == numChars) 
			SetRect(&selectionRect,mRect.left+2,mRect.top+3,
				mRect.right-2,mRect.bottom-2) ;
		else 
			SetRect(&selectionRect,left,mRect.top+3,right,mRect.bottom-2) ;
		FillRect(&selectionRect) ;
	}
	PenNormal() ;
	SetFgColor(gBlackPixel) ;
	SetBkColor(gWhitePixel) ;
	SetClip(clipRgn) ;
	DisposeRgn(clipRgn) ;
	XFlush(gDisplay) ;

}

TTablePane::TTablePane(TPane *parent,short xpos,short ypos,short width,short height,TTable *table)
	: TPane(parent,xpos,ypos,width,height)
{

	mTable	 = table ;
	bool	hasVertScrollBar = false ;
	bool	hasHorizScrollBar = false ;
	
	
	long	tableWidth = mTable->mHasRowLabels ? mTable->mColWidth*(mTable->mCols+1) : mTable->mColWidth*mTable->mCols ;
	long	tableHeight = mTable->mHasColLabels ? mTable->mRowHeight*(mTable->mRows+1) : mTable->mRowHeight*mTable->mRows ;
	
	mTable->mParentPane = this ;

	SetRect(&mCanvasRect,0,0,width-16,height-16) ;
	mTable->SetCanvasRect(mCanvasRect) ;

	long	thePaneWidth = width ;
	long	thePaneHeight = height ;
	
	
	if (tableWidth > width-16) {	
		hasHorizScrollBar = true ;
	}

	if (tableHeight > height-16)	{
		hasVertScrollBar = true ;
	}
    
    SetRect(&mPortRect,0,0,thePaneWidth,thePaneHeight) ;
	mObjectList = new TObjectList ;

	float	minVal = 0. ,maxVal ;
	
	Int16 widthOfDisplayedContent = mCanvasRect.right - mCanvasRect.left  ;
	Int16 heightOfDisplayedContent = mCanvasRect.bottom - mCanvasRect.top  ;

	if (mTable->mHasColLabels) 
		heightOfDisplayedContent -= mTable->mRowHeight ;
	if (mTable->mHasRowLabels) 
		widthOfDisplayedContent -= mTable->mColWidth ;
			
	Int16 vertScrollMax = mTable->mRowHeight*mTable->mRows - heightOfDisplayedContent ;
	Int16 horizScrollMax = mTable->mColWidth*mTable->mCols - widthOfDisplayedContent ;
	
	if (hasVertScrollBar) {
		maxVal	= vertScrollMax ;
		if (hasHorizScrollBar) 
			mVertScrollBar = new TScrollBar(mPortRect.right-16,0,mPortRect.right,
				mPortRect.bottom-16,kVertical,minVal,maxVal,minVal);
		else
			mVertScrollBar = new TScrollBar(mPortRect.right-16,0,mPortRect.right,
				mPortRect.bottom,kVertical,minVal,maxVal,minVal);
		mVertScrollBar->SetTrackingRoutine(TTablePane::DoVScrollBar) ;		
		mVertScrollBar->SetRefCon((unsigned long)this) ;
		AddObject(mVertScrollBar) ;
	}

	if (hasHorizScrollBar) {
		maxVal	= horizScrollMax ;
		if (hasVertScrollBar) 
			mHorizScrollBar =  new TScrollBar(0,mPortRect.bottom-16,
				mPortRect.right-16,mPortRect.bottom,kHorizontal,minVal,maxVal,minVal);
		else
			mHorizScrollBar =  new TScrollBar(0,mPortRect.bottom-16,
				mPortRect.right,mPortRect.bottom,kHorizontal,minVal,maxVal,minVal);
		mHorizScrollBar->SetTrackingRoutine(TTablePane::DoHScrollBar) ;
		mHorizScrollBar->SetRefCon((unsigned long)this) ;
		AddObject(mHorizScrollBar) ;
	}
		
	RectRgn(mUpdateRegion,&mPortRect) ;
	mPortOrigin.x = 0. ;
	mPortOrigin.y = 0. ;
	mLastClick = 0 ;
	mIsVisible = true ;
	

}

TTablePane::~TTablePane() 
{
	delete mTable ;
}

void	TTablePane::Update(XEvent *event) {

	if (event->xexpose.window != mXWindow)
		return ;
	GrafPtr tPort ;
	XRectangle updateRect ;
	GetPort(&tPort) ;

	updateRect.x = event->xexpose.x ;
	updateRect.y = event->xexpose.y ;
	updateRect.width = event->xexpose.width ;
	updateRect.height = event->xexpose.height ;
	XUnionRectWithRegion(&updateRect,mUpdateRegion,mUpdateRegion) ;
	if (event->xexpose.count == 0) {
		SetPort(this) ;
		XSetRegion(gDisplay,mGC,mUpdateRegion) ;
		
		// Let us check to see if the window size has changed
		XWindowAttributes attr ;
		XGetWindowAttributes(gDisplay,mXWindow,&attr) ;
		int	width = attr.width ;
		int	height = attr.height ;
		if (width != mPortRect.right || height != mPortRect.bottom) {
			SetRect(&mPortRect,0,0,width,height) ;
			AdjustScrollBars() ;
		}

		Draw() ;
		XFlush(gDisplay) ;
		XSetClipMask(gDisplay,mGC,None) ;
		SetPort(tPort) ;
		XDestroyRegion(mUpdateRegion) ;
		mUpdateRegion = XCreateRegion() ;
	}
}
	
void	TTablePane::SetTableSize(Int16 nrows, Int16 ncols) { 
	mTable->SetTableSize(nrows,ncols) ; 
	AdjustScrollBars() ; 
	Refresh() ;
} 

void	TTablePane::DoMouseDown(XEvent *event) {
	TPane::DoMouseDown(event) ;
	DoClick(event) ;
}

void	TTablePane::DoKeyDown(XEvent *event) { 
	TPane::DoKeyDown(event) ; 
	mTable->DoKeyDown(event) ; 
} 


void TTablePane::Draw() 
{
	GrafPtr	tPort ;
	GetPort(&tPort) ;
	SetPort(this) ;
	
	SetClip(mPortRect) ;
	mTable->Draw(nil) ;

	SetClip(mPortRect) ;
	Rect	theRect  ;
	SetFgColor(gGrayPixel) ;
	SetRect(&theRect,mPortRect.right-16,0,mPortRect.right,mPortRect.bottom) ;
	FillRect(&theRect) ;
	SetFgColor(gWhitePixel) ;
	DrawLine(mPortRect.right-16,0,mPortRect.right-16,mPortRect.bottom) ;
	//SetFgColor(gGrayPixel) ;
	SetFgColor(gGrayPixel) ;
	SetRect(&theRect,0,mPortRect.bottom-16,mPortRect.right,mPortRect.bottom) ;
	FillRect(&theRect) ;
	SetFgColor(gWhitePixel) ;
	DrawLine(0,mPortRect.bottom-16,mPortRect.right,mPortRect.bottom-16) ;
	DrawLine(mPortRect.right-1,0,mPortRect.right-1,mPortRect.bottom) ;
	DrawLine(0,mPortRect.bottom-1,mPortRect.right,mPortRect.bottom-1) ;

	SetFgColor(gWhitePixel) ;
	DrawLine(mPortRect.right-16,mPortRect.bottom-16,mPortRect.right-16,mPortRect.bottom) ;
	DrawLine(mPortRect.right-16,mPortRect.bottom-16,mPortRect.right,mPortRect.bottom-16) ;
	SetFgColor(gBlackPixel) ;
	
	if(mObjectList!=nil)
		mObjectList->DrawObjects();	
			
	SetFgColor(gBlackPixel) ;
	SetPort(tPort) ;
}

void TTablePane::AdjustScrollBars() {
	
	// Check to see if we still need the scrollbar(s) 
	// If not, delete them, toherwize adjust their size and position
	long	tableWidth = mTable->mHasRowLabels ? mTable->mColWidth*(mTable->mCols+1) : mTable->mColWidth*mTable->mCols ;
	long	tableHeight = mTable->mHasColLabels ? mTable->mRowHeight*(mTable->mRows+1) : mTable->mRowHeight*mTable->mRows ;
	
	long	windowWidth = mPortRect.right - mPortRect.left ;
	long	windowHeight = mPortRect.bottom - mPortRect.top ;
		
	bool	hasVertScrollBar = tableHeight > windowHeight-16 ? true : false ;
	bool	hasHorizScrollBar = tableWidth > windowWidth-16 ? true : false ;
	
	Int16	horizWidth = windowWidth - 16 ;
	Int16	vertHeight = windowHeight - 16 ;

	Int16 vertScrollMax = tableHeight - vertHeight ;
	Int16 horizScrollMax = tableWidth - horizWidth ;
	
	float	vertScrollPos = 0. ;
	float	horizScrollPos = 0. ;
		
	if (hasHorizScrollBar) {
		// We need a horizontal scrollbar
		// Do we have one already
		if (mHorizScrollBar != nil)
			mHorizScrollBar->Resize(0,windowHeight-16,horizWidth,windowHeight,horizScrollMax) ;
		else {
			// We need to create one
			mHorizScrollBar =  new TScrollBar(0,windowHeight-16,
					horizWidth,windowHeight,kHorizontal,0,horizScrollMax,0);
			mHorizScrollBar->SetTrackingRoutine(TTablePane::DoHScrollBar) ;
			mHorizScrollBar->SetRefCon((unsigned long)this) ;
			AddObject(mHorizScrollBar) ;
		}
		horizScrollPos	= mHorizScrollBar->GetControlValue() ;
	}
	else {
		// Do we have one already
		if (mHorizScrollBar != nil) {
			RemoveObject(mHorizScrollBar) ;
			mHorizScrollBar = nil ;
		}
	}

	if (hasVertScrollBar) {
		// We need a vertical scrollbar
		// Do we have one already
		if (mVertScrollBar != nil)
			mVertScrollBar->Resize(windowWidth-16,0,windowWidth,vertHeight,vertScrollMax) ;
		else {
			// We need to create one
			mVertScrollBar =  new TScrollBar(windowWidth-16,0,
					windowWidth,vertHeight,kVertical,0,vertScrollMax,0);
			mVertScrollBar->SetTrackingRoutine(TTablePane::DoVScrollBar) ;		
			mVertScrollBar->SetRefCon((unsigned long)this) ;
			AddObject(mVertScrollBar) ;
		}
		vertScrollPos	= mVertScrollBar->GetControlValue() ;
	}
	else {
		// Do we have one already
		if (mVertScrollBar != nil) {
			RemoveObject(mVertScrollBar) ;
			mVertScrollBar = nil ;
		}
	}

	SetRect(&mCanvasRect,0,0,horizWidth,vertHeight) ;
	mTable->SetCanvasRect(mCanvasRect) ;
	mPortOrigin.x = -horizScrollPos ;
	if (mPortOrigin.x > 0.) mPortOrigin.x = 0. ;
	mPortOrigin.y = -vertScrollPos ;
	if (mPortOrigin.y > 0.) mPortOrigin.y = 0. ;
	
	Refresh() ;

}



void TTablePane::ScrollTableBy(
	float deltaX,
	float deltaY) 
{
	mPortOrigin.x -= deltaX ;
	if (mPortOrigin.x > 0.)
		mPortOrigin.x = 0. ;
	mPortOrigin.y -= deltaY ;
	if (mPortOrigin.y > 0.)
		mPortOrigin.y = 0. ;
	Refresh() ;
}

void TTablePane::ScrollTableTo(
	float x,
	float y) 
{
	mPortOrigin.x = -x ;
	if (mPortOrigin.x > 0.)
		mPortOrigin.x = 0. ;
	mPortOrigin.y = -y ;
	if (mPortOrigin.y > 0.)
		mPortOrigin.y = 0. ;
	Refresh() ;
}

void
TTablePane::Refresh() 
{
	GrafPtr	tPort ;
	GetPort(&tPort) ;
	SetPort(this) ;
	DisposeRgn(mUpdateRegion) ;
	mUpdateRegion = NewRgn() ;
	
	RectRgn(mUpdateRegion,&mCanvasRect) ;
	
	mTable->Draw(mUpdateRegion) ;
	
	TPane::Draw() ;
	
	DisposeRgn(mUpdateRegion) ;
	mUpdateRegion = NewRgn() ;
}


void 
TTablePane::DoHScrollBar(TObject	*object) {
	float	newVal ;
	float	scrollDistance ;
	TTablePane	*theTablePane ;
	
	theTablePane = (TTablePane *)object->GetRefCon() ;
	newVal	= theTablePane->mHorizScrollBar->GetControlValue() ;

	scrollDistance = newVal - theTablePane->mHScrollPos ;
	if (scrollDistance != 0.) {
		theTablePane->mHScrollPos = newVal ;
		theTablePane->ScrollTableTo(theTablePane->mHScrollPos,theTablePane->mVScrollPos) ;
	}
}

void 
TTablePane::DoVScrollBar(TObject	*object) {
	float	newVal ;
	float	scrollDistance ;
	TTablePane	*theTablePane ;
	
	theTablePane = (TTablePane *)object->GetRefCon() ;
	newVal	= theTablePane->mVertScrollBar->GetControlValue() ;

	scrollDistance = newVal - theTablePane->mVScrollPos ;
	if (scrollDistance != 0) {
		theTablePane->mVScrollPos = newVal ;
		theTablePane->ScrollTableTo(theTablePane->mHScrollPos,theTablePane->mVScrollPos) ;
	}
}	

void TTablePane::DoClick(
	XEvent *inEvent) 
{

	Point			theMouseLoc ;
	
	theMouseLoc.x = inEvent->xbutton.x ;
	theMouseLoc.y = inEvent->xbutton.y ;
	mClickTime 		= inEvent->xbutton.time ;
	
	if (!mTable->DoClickInLabels(inEvent)) {
		Point imagePoint ;
		LocalToImagePoint(theMouseLoc,imagePoint) ;
		inEvent->xbutton.x = imagePoint.x ;
		inEvent->xbutton.y = imagePoint.y ;
		mTable->DoClick(inEvent) ;
	}		
	mLastClick = mClickTime ;
}
