// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include <math.h>
#include "TVector.h"

TVector::TVector(float a1,float a2,float a3,float a4) {
	a[0] = a1 ;
	a[1] = a2 ;
	a[2] = a3 ;
	a[3] = a4 ;
}

float TVector::length() {
	float len ;
	int i ;
	
	len = 0. ;
	for(i=0;i<3;i++) {
		len += a[i]*a[i] ;
	}
	
	len = sqrt(len) ;
		
	return len ;
}

/* brian's addition 3/25/96 */



SimpleDirection 
TVector::GetSimpleD()
{
	if(a[0] == 0.0)
	{
		if(a[1] == 0.0 && a[2] == 1.0)
			return k001;
		if(a[1] == 1.0){
			if(a[2] == 0.0)
				return k010;
			if(a[2] == 1.0)
				return k011;
		}
		if(a[1] == -1.0 && a[2] == 1.0)
			return k0_11;
	}
	if(a[0] == 1.0)
	{
		if(a[1] == 0.0){
			if(a[2] == 0.0)
				return k100;
			if(a[2] == 1.0)
				return k101;
		}
		if(a[1] == 1.0){
			if(a[2] == 0.0)
				return k110;
			if(a[2] == 1.0)
				return k111;
		}
		if(a[1] == -1.0){
			if(a[2] == 0.0)
				return k1_10;
			if(a[2] == 1.0)
				return k1_11;
		}
		if(a[1] == 2.0 && a[2] == 0.0)
			return k120;
	}
	if(a[0] == -1.0)
	{
		if(a[1] == 0.0 && a[2] == 1.0)
			return k_101;
		if(a[1] == 1.0){
			if(a[2] == 0.0)
				return k_110;
			if(a[2] == 1.0)
				return k_111;
		}
		if(a[1] == 2.0 && a[2] == 0.0)
			return k_120;
		if(a[1] == -1.0 && a[2] == 1.0)
			return k_1_11;
	}
	if(a[0] == 2.0)
	{
		if(a[1] == 1.0 && a[2] == 0.0)
			return k210;
		if(a[1] == -1.0 && a[2] == 0.0)
			return k2_10;
	}
	
	return kNotSimple;
}


TVector operator*(const TVector& vector,const float factor) {
	TVector tvec ;
	int i ;
	
    for(i=0;i<4;i++) {
    	tvec.a[i] *= factor ;
    }
    return tvec ;
}

TVector operator/(const TVector& vector,const float factor) {
	TVector tvec ;
	int i ;
	
    for(i=0;i<4;i++) {
    	tvec.a[i] /= factor ;
    }
    return tvec ;
}

TVector operator+(const TVector& vector,const float factor) {
	TVector tvec ;
	int i ;
	
    for(i=0;i<4;i++) {
    	tvec.a[i] += factor ;
    }
    return tvec ;
}

TVector operator-(const TVector& vector,const float factor) {
	TVector tvec ;
	int i ;
	
    for(i=0;i<4;i++) {
    	tvec.a[i] -= factor ;
    }
    return tvec ;
}

TVector operator*(const TVector& v1,const TVector& v2) {
	TVector v3 ;
	
    v3.a[0] = v1.a[1]*v2.a[2] - v2.a[1]*v1.a[2] ;
    v3.a[1] = v1.a[2]*v2.a[0] - v1.a[0]*v2.a[2] ;
    v3.a[2] = v1.a[0]*v2.a[1] - v2.a[0]*v1.a[1] ;
	v3.a[3] = 0. ;
	return v3 ;
}
TVector operator+(const TVector& v1,const TVector& v2) {
	TVector v3 ;
	
	for(int i = 0;i<4;i++) {
		v3.a[i] = v1.a[i]+v2.a[i] ;
	}
	return v3 ;
}

TVector operator-(const TVector& v1,const TVector& v2) {
	TVector v3 ;
	
	for(int i = 0;i<4;i++) {
		v3.a[i] = v1.a[i]-v2.a[i] ;
	}
	return v3 ;
}
float   operator^(const TVector& v1,const TVector& v2) {
	// We assume 3D vector with v1.a[3] = 0. , v2.a[3] = 0.
	float dot = v1.a[0]*v2.a[0] + v1.a[1]*v2.a[1] + v1.a[2]*v2.a[2] + v1.a[3]*v2.a[3] ;
	return dot ;
}

void NormalizeVector(TVector& v1) {
	// We assume 3D vector with v1.a[3] = 0
	
	float fact = 1./v1.length() ;
	for(int i=0;i<3;i++) {
		v1.a[i] *= fact ;
	}
}

TVector& 
TVector::operator*=(const float factor) {
	int i ;
	
    for(i=0;i<4;i++) {
    	this->a[i] *= factor ;
    }
    return *this ;
}

TVector& 
TVector::operator/=(const float factor) {
	int i ;
	
    for(i=0;i<4;i++) {
    	this->a[i] /= factor ;
    }
    return *this ;
}
TVector& 
TVector::operator+=(const float factor) {
	int i ;
	
    for(i=0;i<4;i++) {
    	this->a[i] += factor ;
    }
    return *this ;
}
TVector& 
TVector::operator-=(const float factor) {
	int i ;
	
    for(i=0;i<4;i++) {
    	this->a[i] -= factor ;
    }
    return *this ;
}

