// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------


#include <stdio.h>
#include <stdlib.h>
#include "TWindowList.h"
#include "TWindow.h"
#include "globals.h"
#include "XDrivers.h"
#include "TFileViewDialog.h"
#include "config.h"

typedef enum { PSEUDOCOLOR, TRUECOLOR } MonitorDisplayType ;
extern MonitorDisplayType  gMonitorDisplayType ;
extern Visual	*gVisual ;

void	RemoveWindow(TWindow *) ;
extern	long	near(float) ;

extern	XFontStruct	*gMenuFont ;
extern	XFontStruct	*gDialogFont ;
extern	XFontStruct	*gWindowFont ;

extern	short	gDebugMode ;
extern	Atom	gWindowManagerAtoms[2];

extern	TWindow	*GetTargetWindow() ;

void	AddWindowToWIndowsMenu(TWindow *theWindow) ;
void	RemoveWindowFromWindowsMenu(TWindow *theWindow) ;
void	SetTargetWindow(TWindow *window) ;
void	SendApplicationCommand(TCommand command) ; 

extern	char	*FS98_PATH ;
extern	char	*EDMDocDir ;
extern	void PutMessage(char *) ;

extern	Boolean	gWMFocusMove ;
Boolean	gFrontWindowIsModal = false ;

GC	gGC = nil ;
Pixmap	gStipple = nil ;
unsigned char	gStippleData[] =
				{	0x3F,0xF3,0xFF,0xFF,
					0xFF,0xFF,0xFF,0xFF,
					0xFF,0xFF,0xFF,0xFF,
					0x00,0x00,0xFF,0xFF } ;

TGrafPort::TGrafPort() {
	mClipRegion = NewRgn() ;
	mFont = gFontStructure ;
	mPortOrigin.x = mPortOrigin.y = 0. ;
	mGC = nil ;
	mDepth = 8 ;
	mFgColor = 0 ;
	mBkColor = 255 ;
	mDrawable = nil ;
	SetRect(&mPortRect,0,0,0,0) ;
	mPortOrigin.x = 0 ;
	mPortOrigin.y = 0 ;
}

TGrafPort::TGrafPort(short width,short height,short depth) {
	SetRect(&mPortRect,0,0,width,height) ;
	mDepth = depth ;
	mClipRegion = NewRgn() ;
	RectRgn(mClipRegion,&mPortRect) ;
	mFont = gFontStructure ;
	mPortOrigin.x = mPortOrigin.y = 0. ;
	mGC = nil ;
	mDrawable = nil ;
	mFgColor = 0 ;
	mBkColor = 255 ;
}

TGrafPort::TGrafPort(FL_OBJECT *object) {
	SetRect(&mPortRect,0,0,object->w,object->h) ;
	OffsetRect(&mPortRect,object->x,object->y) ;
	mDepth = fl_get_canvas_depth(object) ;
	mDrawable = FL_ObjWin(object) ;
	mGC = fl_state[fl_get_vclass()].gc[0] ;
	mPortOrigin.x = mPortOrigin.y = 0. ;

	mFont = gFontStructure ;

	mClipRegion = NewRgn() ;
	RectRgn(mClipRegion,&mPortRect) ;

	// If this is a forms window, we don't want the closing of the
	// window through the window manager to close down the program
	XSetWMProtocols(gDisplay,mDrawable,gWindowManagerAtoms,2) ;
}

void 	
TGrafPort::ImageToLocalPoint(
	Point &inImagePt,
	Point &outLocalPt) 
{
	outLocalPt.x =  near(inImagePt.x + mPortOrigin.x) ;
	outLocalPt.y =  near(inImagePt.y + mPortOrigin.y) ;
}

void 	
TGrafPort::LocalToImagePoint(
	Point &inLocalPt,
	Point &outImagePt) 
{
	outImagePt.x =  near(inLocalPt.x - mPortOrigin.x) ;
	outImagePt.y =  near(inLocalPt.y - mPortOrigin.y) ;
}

void TGrafPort::SetClipRegion(RgnHandle rgn) {
	DisposeRgn(mClipRegion) ;
	mClipRegion = NewRgn() ;
	Rect	theRect ;
	RegionRect(rgn,&theRect) ;
	RectRgn(mClipRegion,&theRect) ;
}

void TGrafPort::SetClipRegion(Rect &rect) {
	DisposeRgn(mClipRegion) ;
	mClipRegion = NewRgn() ;
	RectRgn(mClipRegion,&rect) ;
}

TGrafPort::~TGrafPort() {
	DisposeRgn(mClipRegion) ;
}

void TGrafPort::SetBkColor(uLong color) { 
	mBkColor = color ; 
	XSetBackground(gDisplay,mGC,color) ;
} 

void TGrafPort::SetFgColor(uLong color) { 
	mFgColor = color ; 
	XSetForeground(gDisplay,mGC,color) ;
} 

TPopUpWindow::TPopUpWindow(TPane *parent,short x,short y,short width,short height)
		: TGrafPort(width,height,parent->GetDepth()){

	XSetWindowAttributes attr ;
	mParentPane = parent ;

	mXWindow = XCreateSimpleWindow(gDisplay,RootWindow(gDisplay,gScreen),x,y,
		width,height,1,parent->GetFgColor(),parent->GetBkColor()) ;
	mParentIsRoot = true ; 
	attr.override_redirect = true ;
	attr.save_under = true ;


	XChangeWindowAttributes(gDisplay,mXWindow,CWOverrideRedirect | CWSaveUnder,&attr) ;
	mDrawable = mXWindow ;
	mGC = parent->GetGC() ;
	mFont = parent->GetFont() ;
}

TPopUpWindow::~TPopUpWindow() {
	if (mParentIsRoot == true) 
		XDestroyWindow(gDisplay,mXWindow) ; 
}

TPane::TPane(TPane *parent,short xloc,short yloc,short width,short height) 
 : TGrafPort(width,height) {
	mParentPane = parent ;
	mUpdateRegion	= XCreateRegion() ;
	
	mPointerInsideWindow = false ;

	// We will associate this font with the menu font
	if (parent != nil) {
		mDepth = parent->GetDepth() ;
		mXWindow = XCreateSimpleWindow(gDisplay,parent->GetXWindow(),xloc,yloc,
			width,height,1,
			parent->GetFgColor(),parent->GetBkColor()) ;
		mDrawable = mXWindow ;
		SetRect(&mPaneRect,xloc,yloc,xloc+width,yloc+height) ;
		XGCValues	gcv ;
		gcv.foreground = parent->GetFgColor() ;
		gcv.background = parent->GetBkColor() ;
		mGC = XCreateGC(gDisplay,mXWindow,GCForeground |
    			GCBackground, &gcv) ;
		XSelectInput(gDisplay,mXWindow,ExposureMask) ;
		mFont = gMenuFont ;
		XSetFont(gDisplay,mGC,mFont->fid) ;
	}
	else {
		SetRect(&mPaneRect,0,0,width,height) ;
		// If parent == nil, we are instancing a TWindow class
	}
	mObjectList = nil ;
	mEditTextList = nil ;
	mCurrentTextEditField = nil ;
	mIsVisible = false ;
	mTakeFocus = false ;
	mCrossingInEventTime = 0 ;
	mCrossingOutEventTime = 0 ;

	mVScrollPos = 0. ;
	mHScrollPos = 0. ;
	mVertScrollBar 	= nil ;
	mHorizScrollBar = nil ;

}


TPane::TPane(TPane *parent,Rect *rect) 
 : TGrafPort(rect->right-rect->left,rect->bottom-rect->top,parent->GetDepth()) {
	mParentPane = parent ;
	mPaneRect = *rect ;
	mUpdateRegion	= XCreateRegion() ;

	mPointerInsideWindow = false ;

	if (parent != nil) {
		mXWindow = XCreateSimpleWindow(gDisplay,parent->GetXWindow(),rect->left,rect->top,
			rect->right-rect->left,rect->bottom-rect->top,1,
			parent->GetFgColor(),parent->GetBkColor()) ;
		mDrawable = mXWindow ;
		XGCValues	gcv ;
		gcv.foreground = parent->GetFgColor() ;
		gcv.background = parent->GetBkColor() ;
		mGC = XCreateGC(gDisplay,mXWindow,GCForeground |
    			GCBackground, &gcv) ;
		XSelectInput(gDisplay,mXWindow,ExposureMask) ;
		mFont = gMenuFont ;
		XSetFont(gDisplay,mGC,mFont->fid) ;
	}
	else {
		OffsetRect(&mPaneRect,-mPaneRect.left,-mPaneRect.top) ;
		// If parent == nil, we are instancing a TWindow class
	}
	mObjectList = nil ;
	mEditTextList = nil ;
	mCurrentTextEditField = nil ;
	mIsVisible = false ;
	mTakeFocus = false ;
	mCrossingInEventTime = 0 ;
	mCrossingOutEventTime = 0 ;

	mVScrollPos = 0. ;
	mHScrollPos = 0. ;
	mVertScrollBar 	= nil ;
	mHorizScrollBar = nil ;
}

void	TPane::AddTextEditObject(TEditText *obj) {
	if (mObjectList==nil) mObjectList = new TObjectList;
	if (mEditTextList==nil) mEditTextList = new TEditTextList;
	mObjectList->AddObject(obj) ; 
	obj->SetParentWindow(this) ;
	mEditTextList->AddObject(obj);
	mCurrentTextEditField = obj ;
}

void	TPane::Select(TEditText *theText) {
	if (mCurrentTextEditField != theText) {
		mCurrentTextEditField->SetFocus(false) ;
	}
	mCurrentTextEditField = theText ;
	mCurrentTextEditField->SetFocus(true) ;
}

void	TPane::SelectAll(TEditText *theText) {
	if (mCurrentTextEditField != theText) {
		mCurrentTextEditField->SetFocus(false) ;
	}
	mCurrentTextEditField = theText ;
	mCurrentTextEditField->SelectAll() ;
	mCurrentTextEditField->SetFocus(true) ;
}

void	TPane::DoMouseDown(XEvent *theEvent) {

	if (mObjectList == nil) return ;
	Point thePoint ;
	TObject *theObject = nil ;
	TEditText	*editText ;
	
	SetPort(this) ;

	thePoint.x = theEvent->xbutton.x ;
	thePoint.y = theEvent->xbutton.y ;
	// are the coordinates relative to us, or the parent window
	if (mParentPane != nil) {
		if (theEvent->xbutton.window == mParentPane->GetXWindow()) {
			thePoint.x -= mPaneRect.left ;
			thePoint.y -= mPaneRect.top ;
		}
	}
	// Make sure the event point is relative to us
	theEvent->xbutton.x = thePoint.x ;
	theEvent->xbutton.y = thePoint.y ;
	if (mObjectList == nil) return ;
	if (mObjectList->GetNumberOfObjects() > 0) {
		theObject = FindObject(&thePoint) ;
		if (theObject != nil) {
			if (mEditTextList != nil) {
				if ((editText=(TEditText *)mEditTextList->GetObject(&thePoint)) != nil) {
					if (mCurrentTextEditField != editText) {
						mCurrentTextEditField->SetFocus(false) ;
					}
					mCurrentTextEditField = editText ;
					mCurrentTextEditField->SetFocus(true) ;
				}
			}
			DoMouseDownInObject(theObject,&thePoint) ;
		}
	}
}

void	TPane::DoEnterNotify(XCrossingEvent *event) { 
	mPointerInsideWindow = true ; 
	mCrossingInEventTime = event->time ;
	SetCursor() ; 
	if (gDebugMode) {
		fprintf(stderr,"\nDoEnterNotify focus = %d",event->focus) ;
		fprintf(stderr,"\nDoEnterNotify:  Enter window time = %u",mCrossingInEventTime) ;
		
	}
} 

void	TPane::DoLeaveNotify(XCrossingEvent *event) { 
	mPointerInsideWindow = false ;  
	mCrossingOutEventTime = event->time ;
	UnsetCursor() ; 
	if (gDebugMode) {
		fprintf(stderr,"DoLeaveNotify focus = %d",event->focus) ;
		fprintf(stderr,"\nDoLeaveNotify:  Leave window time = %u",mCrossingOutEventTime) ;
		
	}
} 

void	TPane::DoKeyDown(XEvent *theEvent) {
	if (mObjectList == nil) return ;

	if (mCurrentTextEditField != nil) {
		KeySym	theKey ;
		GrafPtr tPort ;
		GetPort(&tPort) ;
		SetPort(this) ;
		theKey = mCurrentTextEditField->DoKeyDown(theEvent) ;
		DoKeyDownInObject(mCurrentTextEditField) ;
		if (theKey == XK_Tab) {
			mCurrentTextEditField->SetFocus(false) ;
			mCurrentTextEditField = 
					mEditTextList->GetNextField(mCurrentTextEditField) ;
			mCurrentTextEditField->SelectAll() ;
		}
		SetPort(tPort) ;
	}
}

void TPane::ScrollBits(
	Int16		inLeftDelta,			// Pixels to scroll horizontally
	Int16		inTopDelta)				// Pixels to scroll vertically
{
										// Get Frame in Port coords

}

void	
TPane::InvalPortRgn(
	RgnHandle inUpdateRgn)
{
	//CopyRgn(inUpdateRgn,mUpdateRegion) ;
	//InvalRgn(inUpdateRgn) ;
}

void TPane::Update(XEvent *event) {

	// If the update event does not belong to us and
	// since a pane (at least for now) does not have children
	// we just get out of here
	if (event->xexpose.window != mXWindow)
		return ;

	GrafPtr tPort ;
	XRectangle updateRect ;
	GetPort(&tPort) ;

	updateRect.x = event->xexpose.x ;
	updateRect.y = event->xexpose.y ;
	updateRect.width = event->xexpose.width ;
	updateRect.height = event->xexpose.height ;
	XUnionRectWithRegion(&updateRect,mUpdateRegion,mUpdateRegion) ;
	if (event->xexpose.count == 0) {
		SetPort(this) ;
		XSetRegion(gDisplay,mGC,mUpdateRegion) ;
		Draw() ;
		XSetClipMask(gDisplay,mGC,None) ;
		SetPort(tPort) ;
		XDestroyRegion(mUpdateRegion) ;
		mUpdateRegion = XCreateRegion() ;
	}
	XFlush(gDisplay) ;
}

TPane::~TPane() {
	if (mObjectList != nil)		delete 	mObjectList ; 
	if (mEditTextList != nil)	delete	mEditTextList ;
	if (mUpdateRegion != nil) XDestroyRegion(mUpdateRegion) ;
	XDestroySubwindows(gDisplay,mXWindow) ;
	XUnmapWindow(gDisplay,mXWindow) ;
	XDestroyWindow(gDisplay,mXWindow) ;
	XFreeGC(gDisplay,mGC) ;
}

void	TWindow::SetWindowTitle(char *title) {
	XStoreName(gDisplay,mXWindow,title) ;
	strcpy(mWindowTitle,title) ;
}

void	TWindow::GetWindowTitle(char *title) {
	strcpy(title,mWindowTitle) ;
}

TWindow::TWindow() 
	: TPane(nil,0,0,0,0) {

	mIsTarget = false ;

	mCanReceiveMenuCommands = true ;	// By default
	mWindowType = kNormalWindow ; // So that we can be the target for menu operations

	mMagnification = 1. ;
	mPortOrigin.x = 0. ;
	mPortOrigin.y = 0. ;

	mIsValid = false ;
	mIsVisible = false ;
	mHasFocus = false ;

	mXPosition = 0 ;
	mYPosition = 0 ;

	mOffscreen = nil ;
	mCommander = nil ;
	mFgColor = gBlackPixel ;
	mBkColor = gWhitePixel ;
	mPaneList = nil ;
	
	mWindowKind = 'wind' ;
	mWindowContent = 'gnrc' ;
}

TWindow::TWindow(short xloc,short yloc,short width,short height,short depth,char *title,
		Boolean hasOffscreen,Boolean resizable) : TPane(nil,xloc,yloc,width,height) {

	int		winX,winY ;
	int		borderWidth ;
    
	XSizeHints sizeHints ;
	XGCValues	gcv ;

	mIsTarget = false ;
	
	mCanReceiveMenuCommands = true ;	// By default
	mWindowType = kNormalWindow ; // So that we can be the target for menu operations

	mIsValid = false ;
	mIsVisible = false ;

	mOffscreen = nil ;
	mCommander = nil ;

	mHasFocus = false ;
	mFocusEventTime = 0 ;

	borderWidth = 4 ;
	mXPosition = winX = xloc ;
	mYPosition = winY = yloc ;

	mMagnification = 1. ;
	mPortOrigin.x = 0. ;
	mPortOrigin.y = 0. ;
	    
	mFont = gWindowFont ;
 	switch (gMonitorDisplayType) { 
		case PSEUDOCOLOR :
		    mXWindow = XCreateWindow(gDisplay,RootWindow(gDisplay,gScreen),
		    				winX,winY,width,height,
		    				borderWidth,depth,InputOutput,
		    				gVisualInfo.visual,
		    				CWBackPixel | CWBorderPixel | CWColormap | CWBackingStore ,
		    				&gImageAttributes) ;
		    break ;
		case TRUECOLOR :
		    mXWindow = XCreateWindow(gDisplay,RootWindow(gDisplay,gScreen),
		    				winX,winY,width,height,
		    				borderWidth,depth,InputOutput,
		    				gVisualInfo.visual,
		    				CWBackPixel | CWBorderPixel | CWBackingStore ,
		    				&gImageAttributes) ;
		    break ;
	}

	mDrawable	= mXWindow ;
	
	if (resizable) {
		sizeHints.flags 	= USPosition | USSize  ;
		sizeHints.x	= winX ;
		sizeHints.y	= winY ;
		sizeHints.width	= width ;
		sizeHints.height	= height ;
		XSetStandardProperties(gDisplay,mXWindow,title,0,
	    			0,0,0,&sizeHints) ;
	}
	else {
		sizeHints.flags 	= USPosition | USSize | PMinSize | PMaxSize ;
		sizeHints.x	= winX ;
		sizeHints.y	= winY ;
		sizeHints.width	= width ;
		sizeHints.height	= height ;
		sizeHints.min_width 	= width ;
		sizeHints.min_height	= height ;
		sizeHints.max_width 	= width ;
		sizeHints.max_height	= height ;
		XSetStandardProperties(gDisplay,mXWindow,title,0,
	    			0,0,0,&sizeHints) ;
	}

	if (title == nil)
		strcpy(mWindowTitle,"") ;
	else
		strcpy(mWindowTitle,title) ;
	
	XWMHints theXWMHints ;
	theXWMHints.flags = StateHint | InputHint ;
	theXWMHints.input = false ;
	theXWMHints.initial_state = NormalState ;	
	XSetWMHints(gDisplay,mXWindow,&theXWMHints) ;
		
	XSetWMProtocols(gDisplay,mXWindow,gWindowManagerAtoms,2) ;

	if (gMonitorDisplayType == PSEUDOCOLOR) {
	    // If the color map exists, set it up
	    XSetWindowColormap(gDisplay,mXWindow,gColorMap) ;
	    gcv.foreground = gBlackPixel ;
	    gcv.background = gWhitePixel ;
	}
	else {
	    XSetWindowColormap(gDisplay,mXWindow,gColorMap) ;
	    gcv.foreground = BlackPixel(gDisplay,gScreen) ;
	    gcv.background = WhitePixel(gDisplay,gScreen) ;
	}


	mFgColor = gBlackPixel ;
	mBkColor = gWhitePixel ;
	mDepth = depth ;
	
	mWindowKind = 'wind' ;
	mWindowContent = 'gnrc' ;
    
	mGC = XCreateGC(gDisplay,mXWindow,GCForeground |
    			GCBackground, &gcv) ;
    
	XSetFont(gDisplay,mGC,mFont->fid) ;


	/*
	XSetWindowAttributes attr ;
	attr.save_under = true ;
	XChangeWindowAttributes(gDisplay,mXWindow,CWSaveUnder,&attr) ;
	*/

	XSelectInput(gDisplay,mXWindow,ExposureMask | KeyPressMask | StructureNotifyMask |
    			FocusChangeMask | EnterWindowMask | LeaveWindowMask  |
    			KeyReleaseMask | ButtonPressMask | ButtonReleaseMask | OwnerGrabButtonMask) ;
	
    			
	/*
	XSelectInput(gDisplay,mXWindow,ExposureMask | KeyPressMask | StructureNotifyMask | SubstructureNotifyMask |
    			EnterWindowMask | LeaveWindowMask  |
    			KeyReleaseMask | ButtonPressMask | ButtonReleaseMask | OwnerGrabButtonMask) ;
	XSelectInput(gDisplay,mXWindow,ExposureMask | KeyPressMask | StructureNotifyMask |
    			FocusChangeMask | EnterWindowMask | LeaveWindowMask  |
    			KeyReleaseMask | ButtonPressMask | ButtonReleaseMask | OwnerGrabButtonMask) ;
    */
	if (hasOffscreen) {
		if ((mOffscreen = new TOffscreen(this)) == nil) {
			printf("\n Unable to set up an offscreen environment") ;
			exit(-1) ;
		}
	}


	if (gGC == nil) {
		gGC = mGC ;
	}

	if (gStipple == nil) {
		unsigned int rwidth,rheight ;
		XQueryBestStipple(gDisplay,mDrawable,4,4,&rwidth,&rheight) ;
		gStipple = XCreatePixmapFromBitmapData(gDisplay,mDrawable,
			(char *)gStippleData,rwidth,rwidth,
			1,0,1) ;
	}
	if (gStipple != nil) XSetStipple(gDisplay,mGC,gStipple) ;

	mPaneList = nil ;
}

TWindow::~TWindow() {

	if (mOffscreen != nil) delete	mOffscreen ;
	if (mPaneList != nil) delete mPaneList ;
	

}

void	TWindow::AddPane(TPane *pane) {
	if (mPaneList == nil) mPaneList = new TPaneList ;
	mPaneList->AddPane(pane) ;
}

void	TWindow::DeletePane(TPane *pane) {
	mPaneList->RemovePane(pane) ;
}

void TWindow::Erase(Rect *rect) {
	if (!mIsVisible) return ;
	Rect theRect ;
	
	theRect = (rect != nil ? *rect : mPortRect) ;
	EraseRect(&theRect) ;
}


void TWindow::DoIdle() {
}

void TWindow::Update(XEvent *event) {

	// If the update event does not belong to our main window
	// it might still belong to one of our children
	// so we send the update event to each one
	// we just get out of here
	if (event->xexpose.window != mXWindow) {
		TPane *pane ;
		if (mPaneList != nil) {
			for(int i = 0 ; i < mPaneList->GetNumberOfPanes(); i++) {
				pane = mPaneList->GetPane(i) ;
				pane->Update(event) ;
			}
		}	
		return ;
	}

	GrafPtr tPort ;
	XRectangle updateRect ;
	GetPort(&tPort) ;

	updateRect.x = event->xexpose.x ;
	updateRect.y = event->xexpose.y ;
	updateRect.width = event->xexpose.width ;
	updateRect.height = event->xexpose.height ;
	XUnionRectWithRegion(&updateRect,mUpdateRegion,mUpdateRegion) ;
	if (event->xexpose.count == 0) {
		SetPort(this) ;
		XSetRegion(gDisplay,mGC,mUpdateRegion) ;
		Draw() ;
		XSetClipMask(gDisplay,mGC,None) ;
		SetPort(tPort) ;
		XDestroyRegion(mUpdateRegion) ;
		mUpdateRegion = XCreateRegion() ;
	}
	XFlush(gDisplay) ;
}

void	TWindow::DoWMRequestDelete() {
	if (strcmp(mWindowTitle,"edm") == 0) {
		PutMessage("Please exit the program from the menu") ;
		return ;
	}
	// Send the application the close command for this window
	SetTargetWindow(this) ;
	SendApplicationCommand(kClose) ;
}

void TWindow::DoClientMessage(XClientMessageEvent *theMessage) {
	// Check to see what kind of message this is
	if (gDebugMode)
		fprintf(stderr,"\n DoClientMessage for %s\n",mWindowTitle) ;
		
	if (gFrontWindowIsModal) return ;
	  
	if (theMessage->type == XA_WINDOW) {
	
		if (theMessage->data.l[0] == gWindowManagerAtoms[0]) {
			if (gDebugMode)
				fprintf(stderr,"\n DELETE Message Atom for %s\n",mWindowTitle) ;
			DoWMRequestDelete() ;
			return ;
		}
		
		if (theMessage->data.l[0] == gWindowManagerAtoms[1]) {
			if (gDebugMode)
				fprintf(stderr,"\n FOCUS Message Atom for %s\n",mWindowTitle) ;
			// Do we want to take focus or not
			mFocusEventTime = (Time)theMessage->data.l[1] ;
			if (gWMFocusMove) {
				// If we have a pointer to focus manager then we get a focus message as soon as the 
				// mouse enters the window or the title bar
				// We also get a focus message when we leave the window
				// even though we haven't yet received a leave notify event
				if (mPointerInsideWindow) {
					// It is a crossing IN / focus IN event
					if ((mFocusEventTime - mCrossingInEventTime) > 10) {
						mTakeFocus = true ;
						if (gDebugMode) {
							fprintf(stderr,"\nDoClientMessage:  mPointerInsideWindow : mFocusEventTime-mCrossingInEventTime = %u  mTakeFocus = TRUE",mFocusEventTime-mCrossingInEventTime) ;
							fprintf(stderr,"\nDoClientMessage:  mPointerInsideWindow : mFocusEventTime-mCrossingOutEventTime = %u  mTakeFocus = TRUE",mFocusEventTime-mCrossingOutEventTime) ;
						}
					}
					else if ((mFocusEventTime - mCrossingOutEventTime) < 10) {
						mTakeFocus = false ;
						mHasFocus = false ;
					}
					else {
						mTakeFocus = false ;
					}
				}
				else {
					if (mHasFocus) {
							// Check to see if this is a focus out event
							if (gDebugMode) {
								fprintf(stderr,"\nDoClientMessage:  Pointer OUTSIDE : mFocusEventTime-mCrossingInEventTime = %u  mTakeFocus = TRUE",mFocusEventTime-mCrossingInEventTime) ;
								fprintf(stderr,"\nDoClientMessage:  Pointer OUTSIDE : mFocusEventTime-mCrossingOutEventTime = %u  mTakeFocus = TRUE",mFocusEventTime-mCrossingOutEventTime) ;
							}
							if ((mFocusEventTime - mCrossingOutEventTime) < 10) {
								// It is a focus out event for sure
								mTakeFocus = false ;
								mHasFocus = false ;
								if (gDebugMode) {
									fprintf(stderr,"\nDoClientMessage:  We already have focus. Assuming this is a Focus OUT") ;
								}
							}
							else {
								// It is a focus IN and not a focus OUT event
								if ((mFocusEventTime - mCrossingInEventTime) > 10) {
									if (gDebugMode) {
										fprintf(stderr,"\nDoClientMessage:  We already have focus. Assuming this is a Focus IN") ;
									}
									if ( mCanReceiveMenuCommands) { 
										// Determine if we are ta
										if (GetTargetWindow() != this  && GetTargetWindow() != nil) {
											SetTargetWindow(this) ; 
											BringToFront(this) ;
											gFrontWindow = this ;
											if (gDebugMode) {
												fprintf(stderr,"\n nDoClientMessage SET TARGET WINDOW ") ;
											}
											mTakeFocus = false ;
										}
									}
								}
								else {
									if (gDebugMode) {
										fprintf(stderr,"\nDoClientMessage:  mFocusEventTime - mCrossingInEventTime) > 10 Not sure what to do") ;
									}
								}
							}
						}	
						mHasFocus = false ;				
				}
			}
			else {
				// We always take focus on click to focus manager
				mTakeFocus = true ;
			}
			XSetInputFocus(gDisplay,mXWindow,RevertToPointerRoot ,mFocusEventTime)  ;
			return ;
		}
	}
}

void TWindow::DoCommand(TCommand command,Boolean *hasBeenHandled) {
	switch(command) {
	
		case kHelp:
			OnCommandHelp() ;		
			*hasBeenHandled = true ;
			break ;
		case kClose:	
			*hasBeenHandled = false ;	// We let ourselves be closed				
										// by the application
			break ;
		default:
			break ;
	}
}
void TWindow::GetCommandStatus(TCommand command,Boolean *isEnabled,Boolean *isChecked) {
	switch(command) {
		case kHelp:
		case kClose:
			*isEnabled = true ;	
			break ;
		default:
			break ;
	}
} 

void TWindow::Close() {
	if (mWindowList != nil) {
		mWindowList->RemoveWindow(this) ;	// We are removed by the owner
	}
	else {
		delete this ;
	}
}

void TWindow::DoMouseDown(XEvent *event) {

	Point thePoint ;
	TPane *pane ;
	
	SetPort(this) ;
	
	// First find out if we are in one of the panes
	thePoint.x = event->xbutton.x ;
	thePoint.y = event->xbutton.y ;
	// The point is relative to the application window
	// and is not in local coordinates of the pane
	if (gFrontWindow != this && !gFrontWindowIsModal) {
		BringToFront(this) ;
		gFrontWindow = this ;
		SetFocus(this) ;
	}
	if (mPaneList != nil) {
		for(int i = 0 ; i < mPaneList->GetNumberOfPanes(); i++) {
			pane = mPaneList->GetPane(i) ;
			if (pane->InPane(&thePoint)) {				
				SetPort(pane) ;
				pane->DoMouseDown(event) ;
				return ;
			}
		}
	}	

	TPane::DoMouseDown(event) ;	
	
}

void TWindow::Show() {
	XMapRaised(gDisplay,mXWindow) ;
 }


void TWindow::DoConfigureNotify(XEvent *event) {
	XConfigureEvent *theEvent = (XConfigureEvent *)event ;
	
	// See if it is a resize event or changing stacking order event
	// See if we are changing the height and width of the window
	long width = mPortRect.right - mPortRect.left ;
	long height = mPortRect.bottom - mPortRect.top ;

	if (theEvent->width != width || theEvent->height != height) {
		ResizeWindow(theEvent->width,theEvent->height) ;
	}
	else if (theEvent->x != mXPosition || theEvent->y != mYPosition) {
		// We are just changing position and possibly stacking order
		mXPosition = theEvent->x ;
		mYPosition = theEvent->y ;
	}
	else {
		if (gDebugMode) {
			fprintf(stderr,"\n DoConfigureNotify Stacking order for %s",mWindowTitle) ;
		}
	}
		
}

void	TWindow::DoFocusIn(XFocusChangeEvent *event) { 
	if (!gWMFocusMove) mTakeFocus = true ;
	mHasFocus = true ; 
	if (mCanReceiveMenuCommands) { 

		// Let us check the details of this event
		if (gDebugMode) {
			fprintf(stderr,"\n%s : Focus IN:  mTakeFocus = %d",mWindowTitle,mTakeFocus) ;
			switch(event->mode) {
				case NotifyNormal:
					fprintf(stderr,"\n XFocusChangeEvent Mode = NotifyNormal") ;
					break ;
				case NotifyGrab:
					fprintf(stderr,"\n XFocusChangeEvent Mode = NotifyNormal") ;
					break ;
				case NotifyUngrab:
					fprintf(stderr,"\n XFocusChangeEvent Mode = NotifyNormal") ;
					break ;
			}
			switch(event->detail) {
				case NotifyAncestor:
					fprintf(stderr,"\n XFocusChangeEvent detail = NotifyAncestor") ;
					break ;
				case NotifyInferior:
					fprintf(stderr,"\n XFocusChangeEvent detail = NotifyInferior") ;
					break ;
				case NotifyNonlinear:
					fprintf(stderr,"\n XFocusChangeEvent detail = NotifyNonlinear") ;
					break ;
				case NotifyNonlinearVirtual:
					fprintf(stderr,"\n XFocusChangeEvent detail = NotifyNonlinearVirtual") ;
					break ;
				case NotifyPointer:
					fprintf(stderr,"\n XFocusChangeEvent detail = NotifyPointer") ;
					break ;
				case NotifyPointerRoot:
					fprintf(stderr,"\n XFocusChangeEvent detail = NotifyPointerRoot") ;
					break ;
			}
		}

		if (mTakeFocus  && GetTargetWindow() != this) {
			SetTargetWindow(this) ; 
			gFrontWindow = this ; 
		}
		
	}
	mTakeFocus = false ;
} 

void	TWindow::DoFocusOut(XFocusChangeEvent *event) {
	mHasFocus = false ; 
	mTakeFocus = false ;
	if (mCanReceiveMenuCommands) { 
		// Let us check the details of this event
		if (gDebugMode) {
			fprintf(stderr,"\n Focus OUT") ;
			switch(event->mode) {
				case NotifyNormal:
					fprintf(stderr,"\n XFocusChangeEvent Mode = NotifyNormal") ;
					break ;
				case NotifyGrab:
					fprintf(stderr,"\n XFocusChangeEvent Mode = NotifyNormal") ;
					break ;
				case NotifyUngrab:
					fprintf(stderr,"\n XFocusChangeEvent Mode = NotifyNormal") ;
					break ;
			}
			switch(event->detail) {
				case NotifyAncestor:
					fprintf(stderr,"\n XFocusChangeEvent detail = NotifyAncestor") ;
					break ;
				case NotifyInferior:
					fprintf(stderr,"\n XFocusChangeEvent detail = NotifyInferior") ;
					break ;
				case NotifyNonlinear:
					fprintf(stderr,"\n XFocusChangeEvent detail = NotifyNonlinear") ;
					break ;
				case NotifyNonlinearVirtual:
					fprintf(stderr,"\n XFocusChangeEvent detail = NotifyNonlinearVirtual") ;
					break ;
				case NotifyPointer:
					fprintf(stderr,"\n XFocusChangeEvent detail = NotifyPointer") ;
					break ;
				case NotifyPointerRoot:
					fprintf(stderr,"\n XFocusChangeEvent detail = NotifyPointerRoot") ;
					break ;
			}
		}
	}
}			

void TWindow::DoMapNotify(XEvent *event) {
	mIsVisible = true ;
	TPane *pane ;
	if (mPaneList != nil) {
		for(int i = 0 ; i < mPaneList->GetNumberOfPanes(); i++) {
			pane = mPaneList->GetPane(i) ;
			pane->SetVisible(true)  ;
		}
	}
	// For some reason, Map Notify events get received, but not
	// expose events for the window under Linux
	// Thus I am putting in a Draw for the window in here as well
	SetPort(this) ;
	Draw() ;		// I shouldn't have to do this here
	SetFocus(this) ;
}

void TWindow::DoUnmapNotify(XEvent *event) {
	mIsVisible = false ;
	TPane *pane ;
	if (mPaneList != nil) {
		for(int i = 0 ; i < mPaneList->GetNumberOfPanes(); i++) {
			pane = mPaneList->GetPane(i) ;
			pane->SetVisible(false)  ;
		}
	}	
}

void TWindow::ResizeWindow(long width,long height) {
	GrafPtr	tPort ;
	GetPort(&tPort) ;
	SetPort(this) ;
	SetRect(&mPortRect,0,0,width,height) ;
	SetClipRegion(mPortRect) ;
	if (mOffscreen != nil) {
		delete mOffscreen ;	
		if ((mOffscreen = new TOffscreen(this)) == nil) {
			printf("\n Unable to set up an offscreen environment") ;
			exit(-1) ;
		}
	}
	SetPort(tPort) ;
}

void TWindow::DoKeyDown(XEvent *event) {
	TPane *pane ;

	// Check the panes
	if (mPaneList != nil) {
		for(int i = 0 ; i < mPaneList->GetNumberOfPanes(); i++) {
			pane = mPaneList->GetPane(i) ;
			pane->DoKeyDown(event) ;
		}
	}	
	TPane::DoKeyDown(event) ;
}

void TWindow::DoKeyUp(XEvent *event) {
	TPane *pane ;

	// Check the panes
	if (mPaneList != nil) {
		for(int i = 0 ; i < mPaneList->GetNumberOfPanes(); i++) {
			pane = mPaneList->GetPane(i) ;
			pane->DoKeyDown(event) ;
		}
	}	
}

void TWindow::SelectCursor() {
}

void TWindow::Draw() {
	if (!mIsVisible) return ;
	GrafPtr	thePort ;
	GetPort(&thePort) ;
	SetPort(this) ;
	if (mOffscreen != nil) {
		Rect theRect = mPortRect ;
		CopyBits(mOffscreen,this,&theRect,&theRect,srcCopy) ;
	}
	else {	
		TPane *pane ;
		if (mPaneList != nil) {
			for(int i = 0 ; i < mPaneList->GetNumberOfPanes(); i++) {
				pane = mPaneList->GetPane(i) ;
				pane->Draw() ;
			}
		}	
	}
	SetPort(thePort) ;
}

void TWindow::WindowToImagePoint(Point &thePoint) {
	thePoint.x = near(thePoint.x/mMagnification + mPortOrigin.x) ;
	thePoint.y = near(thePoint.y/mMagnification + mPortOrigin.y) ;
}

void TWindow::ImageToWindowPoint(Point &thePoint) {
	thePoint.x = near((thePoint.x-mPortOrigin.x)*mMagnification) ;
	thePoint.y = near((thePoint.y-mPortOrigin.y)*mMagnification) ;	
}

void TWindow::WindowToImageRect(Rect &theRect) {
	theRect.left = near(theRect.left/mMagnification + mPortOrigin.x) ;
	theRect.right = near(theRect.right/mMagnification + mPortOrigin.x) ;
	theRect.top = near(theRect.top/mMagnification + mPortOrigin.y) ;
	theRect.bottom = near(theRect.bottom/mMagnification + mPortOrigin.y) ;
}

void TWindow::ImageToWindowRect(Rect &theRect) {
	theRect.left = near((theRect.left-mPortOrigin.x)*mMagnification) ;
	theRect.right = near((theRect.right-mPortOrigin.x)*mMagnification) ;
	theRect.top = near((theRect.top-mPortOrigin.y)*mMagnification) ;
	theRect.bottom = near((theRect.bottom-mPortOrigin.y)*mMagnification) ;
}

void TWindow::WindowToImagePoint(FPoint &thePoint) {
	thePoint.x = thePoint.x/mMagnification + mPortOrigin.x ;
	thePoint.y = thePoint.y/mMagnification + mPortOrigin.y ;
}

void TWindow::ImageToWindowPoint(FPoint &thePoint) {
	thePoint.x = (thePoint.x-mPortOrigin.x)*mMagnification ;
	thePoint.y = (thePoint.y-mPortOrigin.y)*mMagnification ;	
}

void TWindow::WindowToImageRect(FRect &theRect) {
	theRect.left = theRect.left/mMagnification + mPortOrigin.x ;
	theRect.right = theRect.right/mMagnification + mPortOrigin.x ;
	theRect.top = theRect.top/mMagnification + mPortOrigin.y ;
	theRect.bottom = theRect.bottom/mMagnification + mPortOrigin.y ;
}

void TWindow::ImageToWindowRect(FRect &theRect) {
	theRect.left = (theRect.left-mPortOrigin.x)*mMagnification ;
	theRect.right = (theRect.right-mPortOrigin.x)*mMagnification ;
	theRect.top = (theRect.top-mPortOrigin.y)*mMagnification ;
	theRect.bottom = (theRect.bottom-mPortOrigin.y)*mMagnification ;
}


TFrameWindow::TFrameWindow(short xloc,short yloc,short width,short height,
			short depth,char *title) 
		: TWindow(xloc,yloc,width,height,depth,title,false) {
}

TFrameWindow::~TFrameWindow() {
}

 
void	TFrameWindow::Draw() {
	mIsVisible = true ;
}

void	TFrameWindow::DoMouseDown(XEvent *event) {
}


// --------- TOffscreen
TOffscreen::TOffscreen(TGrafPort *thePort,Rect *rect) : 
	TGrafPort(thePort->GetPortRect().right-thePort->GetPortRect().left,
		thePort->GetPortRect().bottom-thePort->GetPortRect().top) {

	XGCValues	gcv ;
	int		width,height ;
	Rect	thePortRect ;

	if (rect == nil) {
		thePortRect = thePort->GetPortRect() ;
	}
	else {
		thePortRect = *rect ;
	}

	width = thePortRect.right - thePortRect.left ;
	height = thePortRect.bottom - thePortRect.top ;

	mGC = nil ;
	mPixmap = nil ;

	SetRect(&mPortRect,0,0,width,height) ;
	if ((mPixmap = XCreatePixmap(gDisplay,thePort->GetDrawable(),
		width,height,thePort->GetDepth())) == 0) {
		return ;
	}
	
	mDrawable	= mPixmap ;

	if (gMonitorDisplayType == PSEUDOCOLOR) {
		gcv.foreground = gBlackPixel ;
		gcv.background = gWhitePixel ;
	}
	else {
		gcv.foreground = BlackPixel(gDisplay,gScreen) ;
		gcv.background = WhitePixel(gDisplay,gScreen) ;
	}
	mFgColor = gBlackPixel ;
	mBkColor = gWhitePixel ;
	mDepth = thePort->GetDepth() ;
	mGC = XCreateGC(gDisplay,mPixmap,GCForeground |
    			GCBackground, &gcv) ;

	mFont = thePort->GetFont() ;

	if (mGC != nil) {
		XCopyGC(gDisplay,thePort->GetGC(),GCFont,mGC) ;
		XSetForeground(gDisplay,mGC,gWhitePixel) ;
    
		XFillRectangle(gDisplay,mPixmap,mGC,0,0,
    			width,height) ;
		XSetForeground(gDisplay,mGC,gBlackPixel) ;
	}
}

TOffscreen::~TOffscreen(){
	if (mPixmap != nil) XFreePixmap(gDisplay,mPixmap) ;
	if (mGC != nil) XFreeGC(gDisplay,mGC) ;
}
// --------- TStOffscreen
TStOffscreen::TStOffscreen(TGrafPort *currentPort,Rect *rect) : 
	TGrafPort(currentPort->GetPortRect().right-currentPort->GetPortRect().left,
		currentPort->GetPortRect().bottom-currentPort->GetPortRect().top) {

	XGCValues	gcv ;
	int		width,height ;
	Rect	thePortRect ;

	if (rect == nil)
		thePortRect = currentPort->GetPortRect() ;
	else
		thePortRect = *rect ;

	width = thePortRect.right - thePortRect.left ;
	height = thePortRect.bottom - thePortRect.top ;

	SetRect(&mPortRect,0,0,width,height) ;
	mRect = thePortRect ;

	mGC = nil ;
	mPixmap = nil ;
	
	mCurrentPort = currentPort ;

	if ((mPixmap = XCreatePixmap(gDisplay,currentPort->GetDrawable(),
		width,height,currentPort->GetDepth())) == 0) {
		fprintf(stderr,"Failed to ceate pixmap\n") ;
		return ;
	}
	

	mDrawable	= mPixmap ;

	SetPort(this) ;

//
	if (gMonitorDisplayType == PSEUDOCOLOR) {
	    // If the color map exists, set it up
	    //XSetWindowColormap(gDisplay,mDrawable,gColorMap) ;
	    gcv.foreground = gBlackPixel ;
	    gcv.background = gWhitePixel ;
	}
	else {
	    //XSetWindowColormap(gDisplay,mDrawable,gColorMap) ;
	    gcv.foreground = BlackPixel(gDisplay,gScreen) ;
	    gcv.background = WhitePixel(gDisplay,gScreen) ;
	}
//


	//GetPort(&mCurrentPort) ;
	mFgColor = gBlackPixel ;
	mBkColor = gWhitePixel ;
	mDepth = currentPort->GetDepth() ;
	
	mGC = XCreateGC(gDisplay,mPixmap,GCForeground |
    			GCBackground, &gcv) ;

	if (mGC != nil) {
		mHasLocalGC = true ;
		XCopyGC(gDisplay,currentPort->GetGC(),GCFont,mGC) ;
		XSetForeground(gDisplay,mGC,gWhitePixel) ;
    
		XFillRectangle(gDisplay,mPixmap,mGC,0,0,
    			width,height) ;
		XSetForeground(gDisplay,mGC,gBlackPixel) ;
	}
	else {
		mHasLocalGC = false ;
		mGC = currentPort->GetGC() ;
	}
	SetPort(this) ;
}

TStOffscreen::~TStOffscreen(){
	if (mPixmap != nil) {
		SetPort(mCurrentPort) ;
		Rect	clipRect ;
		GetClip(&clipRect) ;
		SetClip(mRect) ;
		XCopyArea(gDisplay,mPixmap,mCurrentPort->GetDrawable(),
			mCurrentPort->GetGC(),0,
    			0,mRect.right-mRect.left,
    			mRect.bottom-mRect.top,mRect.left,mRect.top) ;
	
		XFreePixmap(gDisplay,mPixmap) ;
		if (mGC != nil && mHasLocalGC) XFreeGC(gDisplay,mGC) ;
		SetClip(clipRect) ;
		XFlush(gDisplay) ;
	}
}

void SetFocus(TPane *pane) {
	XSetInputFocus(gDisplay,pane->GetXWindow(),RevertToPointerRoot,CurrentTime) ;
}

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

extern	char	*EDMBrowser ;

void	EDMHelp(char *text) {
	char filename[256] ;
	char basename[256] ;
	char command[256] ;
	
	int	status = 0 ;
	sprintf(basename,"%s%s%s",FS98_PATH,EDMDocDir,text) ;
	
	// Check to see if we have the browser
	#ifdef HAVE_SYS_STAT_H
		struct stat sb ;
//Ignore check for cygwin
    #ifndef HAVE_SYS_STAT_H
		status = stat(EDMBrowser,&sb) ;
		if (status == -1) {
			// We failed to verify the existence of the browser
			// Load in the equivalent text files and show in a forms file browser
			sprintf(filename,"%s.txt",basename) ;			
			TFileViewDialog *fileViewDialog = new TFileViewDialog(filename) ;
			//fprintf(stderr,"At one, %s.txt\n",basename) ;
			if (!fileViewDialog->IsValid()) {
				delete fileViewDialog ;
				char otext[256] ;
				sprintf(otext,"Failed to load the file %s",filename) ;
				//fprintf(stderr,"Failed to load the file %s\n",filename) ;
				PutMessage(otext) ;
			}
		}
		else 
	#endif	
		{
			// Try to open the browser using the path and name of the browser	
			sprintf(filename,"%s.htm",basename) ;
			//fprintf(stderr,"At 2, %s.htm\n",basename) ;
			sprintf(command,"%s %s &",EDMBrowser,filename) ;
			//fprintf(stderr,"At 3, %s %s &\n",EDMBrowser,filename) ;
			status = system(command) ;
			
			// If we didn't get a browser at this moment, we are without on-line help
		}
	#else
		// We don't have stat to verify the existence of the browser
		// Load in the equivalent text files and show in a forms file browser
		sprintf(filename,"%s.txt",basename) ;			
		//fprintf(stderr,"At 4 %s.txt\n",basename) ;
		TFileViewDialog *fileViewDialog = new TFileViewDialog(filename) ;

		if (!fileViewDialog->IsValid()) {
			delete fileViewDialog ;
			char otext[256] ;
			sprintf(otext,"Failed to load the file %s",filename) ;
//			fprintf(stderr,"A Failed to load the file %s",filename) ;
			PutMessage(otext) ;
		}
	#endif
	
}
