      subroutine GBLDM(Y,S,SY,SS,F,STEP,MAXMIX,MEMORY,DMIX,IFAIL,DELTA)
           implicit real*4 (a-h,o-z)
           dimension STEP(MAXMIX),F(MAXMIX)
           dimension Y(MAXMIX,MEMORY),S(MAXMIX,MEMORY)
           dimension SS(MEMORY,MEMORY),SY(MEMORY,MEMORY)
           real*4 YYINV(MEMORY,MEMORY),YPROJ(MEMORY),YTG(MEMORY),
     X           YT(MEMORY,MEMORY)
!           real*4,allocatable :: Undef(:), Def(:)
!--------------------------------------------------------------------
!          Good Broyden according to LDM
!
!          Input Parameters
!          MAXMIX:         Length of the variable vector
!          MEMORY:         Total number of memory values to use
!                          Most recent is last
!          Y,S:            Conventional Y and S arrays
!          YY:             Matrix of Y*Y values
!          SY:             Matrix of S*Y values
!          F:              -Grad(MAXMIX) at the current point (residue)
!          DMIX:           Scaler for initial matrix
!
!          Output
!          STEP            Multi-Secant Step
!
!
!--------------------------------------------------------------------
!        allocate (YYINV(MEMORY,MEMORY),YPROJ(MEMORY),YTG(MEMORY))
!        allocate (YT(MEMORY,MEMORY) )
!       Invert YY
        ninverts=0
!       
        DO N=1,MEMORY
         DO M=1,MEMORY
           YT(M,N)=SY(M,N)
         ENDDO
        ENDDO
!       Difference between BB and Multi-Secant, from Byrd Nocedal Schnabel 6.2
        DO I=2,MEMORY
                DO J=1,I-1
                        YT(I,J)=YT(I,J)-SS(I,J)/dmix
                ENDDO
        ENDDO

!
!       Note, Add is negative since matrix is negative
        ADD=-DELTA
1000    continue
        DO N=1,MEMORY
         DO M=1,MEMORY
           YYINV(M,N)=YT(M,N)
         ENDDO
        ENDDO
        call minverse(YYINV,MEMORY,MEMORY,IFAIL)
!
!       If the matrix is singular, add a diagonal component to it
        if(ifail .lt. 0)then
                  Write(21,*)':DIRE Singular, adding ',add
                  ninverts=ninverts+1
                  do J=1,MEMORY
                     YT(J,J)=YT(J,J)+ADD
                  enddo
                  ADD=ADD*2.D0
                  if(ninverts .lt. 10)goto 1000
!                 Bomb to Pratt
                  IFAIL=-1
                  goto 100
        endif
!
        ntries=0
2000    continue
        DO N=1,MEMORY
!       Dot product Y_T . F
                YTG(N)=0
                DO K=1,MAXMIX
                   YTG(N)=YTG(N)+S(K,N)*F(K)
                ENDDO
        ENDDO
!       Multiply by YYINV
        DO N=1,MEMORY
           YPROJ(N)=0
           DO K=1,MEMORY
              YPROJ(N)=YPROJ(N)+YTG(K)*YYINV(N,K)
           ENDDO
        ENDDO
!       Create the step
        DO J=1,MAXMIX
!          New Position
           ST = 0
           SP = F(J)
           DO N=1,MEMORY
              ST = ST - S(J,N)*YPROJ(N)
              SP = SP - Y(J,N)*YPROJ(N)
           ENDDO
           STEP(J)  =ST+DMIX*SP
        ENDDO

        IFAIL=0
100     continue !deallocate (YYINV, YTG, YPROJ )
        RETURN
        END

