C  This program is free software; you can redistribute it and/or modify
C  it under the terms of the GNU General Public License as published by
C  the Free Software Foundation; either version 2 of the License, or
C  (at your option) any later version.
C
C  This program is distributed in the hope that it will be useful,
C  but WITHOUT ANY WARRANTY; without even the implied warranty of
C  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
C  GNU General Public License for more details.
C
C  You should have received a copy of the GNU General Public License
C  along with this program; if not, write to the Free Software
C  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
C
C
C program sconf generates the start configuration for a genet algorithm
 
       PROGRAM SCONF
 
C  mparam: maximal number of parameters
C  mnbits: maximal number of bits allowed to describe one parameter
C  mindiv: maximal number of chromosomes per population
C         PARAMETER (MPARAM = 80, MNBITS = 4, MINDIV = 2000)
        INCLUDE 'CrossPars.h'
C
C  Current limits
C     mparam to 30 - may need to be increased
C     mnbits to 10 - Increased from 8 to 10 for multislice use
C     Number of members 100 - Decreased from 120 to 100 to avoid
C                 memory problems after increasing mnbits
 
C  nparam: number of parameters
C  nbits(i): number of bits describing parameter i
C  nndiv: number of chromosomes in one population
C  mbits: maximal number of bits describing one parameter
C  totbts: total number of bits describing all parameters in one chromosome
C  pow2(i): powers of 2
C  bit(i,j): bit no. j in parameter i of a chromosome
C  brange(i), erange(i): beginning and end of parameter range of parameter i
C  r1: number to initialize random number generation
C  randi(i): random number
C  actval , crit: help variables
         INTEGER NPARAM, NBITS(MPARAM), NNDIV, MBITS, POW2(0:3)
         INTEGER*2 BIT(MPARAM,4)
         INTEGER TOTBTS
         REAL BRANGE(MPARAM), ERANGE(MPARAM)
         REAL R1, RANDI(MPARAM), ACTVAL, CRIT
         REAL R(97)
         INTEGER IDUM, IX1, IX2, IX3, IFF
         INTEGER II1, II2
         REAL RR1
 
         OPEN (1, FILE='control', STATUS='old')
         OPEN (2, FILE='population', STATUS='unknown')
         OPEN (10, FILE='error', STATUS='unknown')
         OPEN (11, FILE='bithelp', STATUS='unknown')
         OPEN (12, FILE='random.hold')
 
   10    FORMAT(I4,2F12.8)
   20    FORMAT(17A1,200I1)
 
C======================================================================
C readin needed information from file 'control'
C======================================================================
 
C number of parameters?
         READ (1, 10) NPARAM
         IF (NPARAM .GT. 80) THEN
           WRITE (6, *) 'maximal number of parameters exceeded'
           WRITE (10, '(a5)') 'error'
           GO TO 1000
         END IF
 
C number of bits, range of every parameter?
         MBITS = 0
         TOTBTS = 0
         DO 100 I=1,NPARAM
           READ (1, 10) NBITS(I), BRANGE(I), ERANGE(I)
           IF (NBITS(I) .GT. MBITS) THEN
             MBITS = NBITS(I)
           END IF
           TOTBTS = TOTBTS + NBITS(I)
 
  100      CONTINUE
 
         WRITE (11, 10) TOTBTS
 
         IF (MBITS .GT. 4) THEN
           WRITE (6, *) 'maximal number of bits per parameter exceeded'
           WRITE (10, '(a5)') 'error'
           GO TO 1000
         END IF
 
 
C number of chromosomes in one population?
         READ (1, 10) NNDIV
 
C initial value for random number generation
         READ (1, *) IDUM
C skip a line, then read number of children - overwrite nndiv with this here
         READ (1, *) JUNK
         READ (1, *) NCHILD
 
C indicate initialization of random number generator is needed
         IFF = 0
 
C======================================================================
C create population
C======================================================================
 
C calculate necessary powers of 2
         CALL POWER2 (POW2,4,MBITS)
 
C run loop over different chromosomes
         DO 200 J=1,NCHILD
 
C       create a random number for every parameter
C       and translate the resulting parameter into a digital code
 
           DO 210 II1=1,NPARAM
C               call srand(r1)
             CALL RANC1 (IX1,IX2,IX3,IDUM,IFF,R,R1)
             RANDI(II1) = R1
C                create next seed value for random number generation
C                 r1 = 2. * r1
 
C           calculate actual parameter value
             ACTVAL = (ERANGE(II1) - BRANGE(II1)) * RANDI(II1)
             RR1 = ERANGE(II1) - BRANGE(II1)
 
C           convert into 0-1-chain
 
             DO 220 II2=1,NBITS(II1)
               CRIT = RR1 / (2.D0 ** II2)
               IF (ACTVAL .GE. CRIT) THEN
                 BIT(II1,II2) = 1
                 ACTVAL = ACTVAL - CRIT
               ELSE
                 BIT(II1,II2) = 0
               END IF
  220          CONTINUE
 
  210        CONTINUE
 
C     write bit chain for present chromosome to file 'population'
           WRITE (2, 20) (' ', L=1,17), ((BIT(I,K), K=1,NBITS(I)), I=1,
     X        NPARAM)
 
  200      CONTINUE
 
C     Save parameters for random # generator (ix1,ix2,ix3,idum,iff,r(97))
         WRITE (12, *) IX1, IX2, IX3, IDUM, IFF
         DO   I=0,31
           WRITE (12, *) (R(J), J=(I*3+1),(I*3+3))
         END DO
         WRITE (12, *) R(97)
 
 
C======================================================================
 
 
 1000    CONTINUE
         WRITE (10, '(a8)') 'no error'
       END
 
 
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
CCCCCCCCC   SUBROUTINES   CCCCCCCCC
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
 
C subroutine power2 calculates necessary powers of 2
       SUBROUTINE POWER2 (POW2, MNBITS, MBITS )
         INTEGER POW2(0:MNBITS-1)
         POW2(0) = 1
         DO 100 I=1,MBITS-1
  100      POW2(I) = POW2(I-1) * 2
         RETURN
       END
 
C subroutine ranc1 is a random number generator care Numerical Recepies Book
C      double precision function ranc1(idum)
       SUBROUTINE RANC1 (IX1, IX2, IX3, IDUM, IFF, R, RANDOM )
         REAL R(97), RANDOM
         INTEGER IX1, IX2, IX3
         PARAMETER (M1 = 259200, IA1 = 7141, IC1 = 54773, RM1 = 1.D0 /
     X      259200)
         PARAMETER (M2 = 134456, IA2 = 8121, IC2 = 28411, RM2 = 1.D0 /
     X      134456)
         PARAMETER (M3 = 243000, IA3 = 4561, IC3 = 51349)
         IF (IDUM .LT. 0 .OR. IFF .EQ. 0) THEN
           IFF = 1
           IX1 = MOD (MOD (54773 - IDUM, 259200) * 7141 + 54773, 259200)
           IX2 = MOD (IX1, 134456)
           IX1 = MOD (IX1 * 7141 + 54773, 259200)
           IX3 = MOD (IX1, 243000)
           DO   J=1,97
             IX1 = MOD (IX1 * 7141 + 54773, 259200)
             IX2 = MOD (IX2 * 8121 + 28411, 134456)
             R(J) = (FLOAT (IX1) + FLOAT (IX2) * RM2) * RM1
           END DO
           IDUM = 1
         END IF
         IX1 = MOD (IX1 * 7141 + 54773, 259200)
         IX2 = MOD (IX2 * 8121 + 28411, 134456)
         IX3 = MOD (IX3 * 4561 + 51349, 243000)
         J = (IX3 * 97) / 243000 + 1
         IF (J .GT. 97 .OR. J .LT. 1) THEN
           WRITE (6, *) 'J is out of range, value ', J
C          PAUSE
         END IF
         RANDOM = R(J)
         IF (R(J) .GT. 1.D0) RANDOM = R(J) - INT (R(J))
         R(J) = (FLOAT (IX1) + FLOAT (IX2) * RM2) * RM1
         RETURN
       END
 
