/*
 * mbfunc.h (for mbfunc.c)
 *
 *  Please include "complex.h", "cmatutil.h", "strfact.h" and "lattice.h"
 *   before this header file.
 */

typedef struct {
  /* calc_intensity */
  double	*Sg, *beta_g, *beta_g_sel;
  complex	**a_matrix, **b_matrix, **eigen_vec, *eigen_value, *boundary_cond;
  int		*nth_branch, *nth_g_sel, *selinfo_nth_reflex;
  /* For calc_perturb */
  int		num_perturb;
  complex	**eigen_vec_inv, **p_vg, **pa_mat, **pb_mat, **p_eigen_vec,
  *p_eigen_val, *p_boundary_cond;
} WORK_CAL_INT;


typedef struct {
  int  ih, ig, flag;  /* ih, ig: pixels, flag: calc. or no calc. */
  double  H, G;       /* H & G: position */
} HGPIX;


void vg_to_pvg( int ct_perturb, int nr, complex **vg, complex ***perturb_vg );
void pvg_to_vg( int ct_perturb, int nr, complex ***perturb_vg, complex **p_vg );
void potential_matrix_csd(
  int num_exact_rf, int num_bethe_rf, HKL *reflex,
  CRYSTAL_STRCT_DATA *csd, double kvolts, CELL_PAR bv, complex **vg
  );


void vgv_fxgv_csd(
  int num_exact_rf, int num_bethe_rf, HKL *reflex,
  CRYSTAL_STRCT_DATA *csd, double kvolts,
  CELL_PAR bv, complex *vgv, complex *zgv
  );
void xrd_fxgv_csd(
  int num_exact_rf, int num_bethe_rf, HKL *reflex,
  CRYSTAL_STRCT_DATA *csd, double kvolts, CELL_PAR bv, complex *fxgv
  );
void ncgv_csd(
  int num_exact_rf, int num_bethe_rf, HKL *reflex,
  CRYSTAL_STRCT_DATA *csd, double kvolts, CELL_PAR bv, complex *ncgv
  );


double calc_abs_K( CRYSTAL_STRCT_DATA *csd, double kvolts, CELL_PAR bv );
UVW downward(UVW inc);
void incident_directions(
  int calc_type, int num_points,
  double H_begin, double H_end, double G_begin, double G_end,
  int *num_cal_points, int *h_pos, int *g_pos, R_HKL *k_xy,
  UVW incidence, R_HKL zolzbase1, CELL_PAR bv
  );
double exct_err(
  double k_z, double abs_incidence, double abs_k,
  HKL g, R_HKL k_xy, UVW incidence, CELL_PAR bv
  );
double calc_excitation_error(
  R_HKL K_xy, double K_z, double abs_K, HKL g, double g_z, CELL_PAR bv
  );

void pre_calc_intensity(
  double accel_kvol, UVW incidence, UVW surf_norm, CELL_PAR bv,
  double *abs_incidence, double *abs_surf_norm, double *cosfai_inc_surf,
  double *wave_length, double *ev_dash
  );
void alloc_work_cal_int(
  int nr, int num_cal_g, int num_perturb, WORK_CAL_INT *wci
  );
void free_work_cal_int(WORK_CAL_INT *wci);

void calc_intensity_with_perturb(
	R_HKL k_xy, BEAM_SEARCH_PAR beam_search,
	int ne, int nb, HKL *reflex, complex **vg,
	int num_perturb, complex ***perturb_vgs,
	int num_cal_g, int nth_g[], int num_cal_thick, double thickness[],
	double accel_kvol, UVW incidence, UVW surf_norm, CELL_PAR bv,
	double abs_incidence, double abs_surf_norm, double cosfai_inc_surf,
	double wave_length, double ev_dash,
	double **intensity,	double ***p_intensity, WORK_CAL_INT *wci
);
void calc_intensity_with_perturb_ug(
	R_HKL K_xy, double abs_K,
	BEAM_SEARCH_PAR beam_search,
	int ne, int nb, HKL *reflex, complex **vg,
	int num_perturb, complex ***perturb_vgs,
	int num_cal_g, int *nth_g, int num_cal_thick, double *thickness,
	double accel_kvol, UVW incidence, UVW surf_norm, CELL_PAR bv,
	double abs_incidence, double abs_surf_norm, double cosfai_inc_surf,
	double **intensity,	double ***p_intensity, WORK_CAL_INT *wci
);
void calc_intensity_ug_spcell(
	R_HKL K_xy, double abs_K, complex *probe,
	BEAM_SEARCH_PAR beam_search,
	int ne, int nb, HKL *reflex, complex **vg,
	int num_perturb, complex ***perturb_vgs,
	int num_cal_g, int *nth_g, int num_cal_thick, double *thickness,
	double accel_kvol, UVW incidence, UVW surf_norm, CELL_PAR bv,
	double abs_incidence, double abs_surf_norm, double cosfai_inc_surf,
	double **intensity,	double ***p_intensity, WORK_CAL_INT *wci
);
void calc_intensity_ug_coherent(
	R_HKL K_xy, double abs_K, complex initial_phase,
	BEAM_SEARCH_PAR beam_search,
	int ne, int nb, HKL *reflex, complex **vg,
	int num_perturb, complex ***perturb_vgs,
	int num_cal_g, int *nth_g, int num_cal_thick, double *thickness,
	double accel_kvol, UVW incidence, UVW surf_norm, CELL_PAR bv,
	double abs_incidence, double abs_surf_norm, double cosfai_inc_surf,
	complex **amplitude, complex ***p_amp, WORK_CAL_INT *wci
);
void calc_disp_surf_ug(
  int ne, int nb, int num_cal_points, double accel_kvol,
  LAT_PAR lp, CELL_PAR bv, double *H, double *G, 
  R_HKL *k_xy, double abs_K, UVW incidence, UVW surf_norm,
  HKL *reflex, complex **vg,
  int branch_begin, int branch_end, int *rf_amp, int num_rf_amp,
  BEAM_SEARCH_PAR beam_search,
  int num_thick, double *thickness, char *infile
  );
complex calc_projpot( R_UVW rx, int nr, HKL reflex[], complex **vg );

complex calc_inv_fourier( R_UVW rx, int nr, HKL reflex[], complex *vg );

/*void calc_bloch_state(
  double accel_kvol, LAT_PAR lp, CELL_PAR bv,
  R_HKL k_xy, UVW incidence, UVW surf_norm, 
  int num_exact, int num_bethe, HKL reflex[], complex **vg,
  double x_begin, double x_end, double y_begin, double y_end,
  R_UVW x1, R_UVW x2,
  int num_points_x, int num_points_y, int branch_begin, int branch_end,
  char *infile
  );*/
void calc_bloch_state_ug(
  double accel_kvol, LAT_PAR lp, CELL_PAR bv, R_HKL k_xy, double abs_K,
  UVW incidence, UVW surf_norm, 
  int num_exact, int num_bethe, HKL reflex[], complex **vg,
  BEAM_SEARCH_PAR beam_search,
  double x_begin, double x_end, double y_begin, double y_end, R_UVW x1, R_UVW x2,
  int num_points_x, int num_points_y, int branch_begin, int branch_end,
  char *infile, FILE *fo_log
  );

void calc_tfringe_ug(
  double accel_kvol, LAT_PAR lp, CELL_PAR bv, R_HKL K_xy, 
  double abs_K, UVW incidence, UVW surf_norm, 
  int ne, int nb, HKL *reflex, complex **vg,
  BEAM_SEARCH_PAR beam_search,
  int num_cal_g, int *nth_g, 
  int num_thick, double *thickness, char *infile
  );
