/* -----------------------------------------------------------------------------
   Copyright (c) 2001 Roar Kilaas
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
  ----------------------------------------------------------------------------- */


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "tiff.h"
#include "config.h"

#ifdef IBM
#define MM 
#endif
#ifdef WORDS_BIGENDIAN
#define BYTEORDER 0x4D4D
#else 
#define BYTEORDER 0x4949
#endif

#define kMM		0x4D4D
#define kINTEL	0x4949
#define max(x,y) (x) > (y) ? (x) : (y)
TiffRecord gTiff = {0,0,8,1,1,1,0,0,0,0,0,0,0,0,0,0} ;

typedef	enum {kMMByteOrder = kMM, kIntelByteOrder = kINTEL} ByteOrder ;
ByteOrder	gByteOrder = (ByteOrder)BYTEORDER ;

mktiff(filen,data,n,m)
	int n[1], m[1], data[];
	char filen[];
{
	tiff(filen,data,n,m);
}
mktiff_(filen,data,n,m)
	int n[1], m[1], data[];
	char filen[];
{
	tiff(filen,data,n,m);
}


Boolean WriteImageDataToByteTiff(char *filename,float *data,int ncols,int nrows,Boolean flipImage) {

	Boolean status = false ;
	float 	*p1,minVal,maxVal,fval,scale ;
	unsigned	char	*address,*c1,*byteData  ;
	unsigned	char	val ;
	int 	nBytes,i ;
	FILE	*fp ;

	// Find min and max so we can scale
	p1 = data ;
	minVal = 1.0e+20 ;
	maxVal = -minVal ;
	for(i=0;i<nrows*ncols;i++) {
		if (*p1 > maxVal)
			maxVal = *p1 ;
		if (*p1 < minVal)
			minVal = *p1 ;
		p1++ ;
	}
	scale = 255./(maxVal-minVal) ; 
	byteData = malloc(nrows*ncols) ;

	c1 = byteData ;
	p1 = data ;
	for(i=0;i<nrows*ncols;i++) {
		fval = (*p1++-minVal)*scale ;
		*c1++ = (unsigned char)fval ;
	}
		
	if ((fp = CreateTiffFile(filename)) != NULL) {
		WriteTiffDirectory(fp,kGrayScale,ncols,nrows) ;
		nBytes = ncols*sizeof(unsigned char) ;
		if (flipImage) {
			address = byteData+(nrows-1)*ncols ;
			for (i=0;i<nrows;i++,address -= ncols) {
				WriteTiffStrip(fp,address,nBytes) ;
			}
		}
		else {
			for (i=0,address=byteData;i<nrows;i++,address += ncols) {
				WriteTiffStrip(fp,address,nBytes) ;
			}
		}
		CloseTiffFile(fp) ;
		status = true ;
	}
	free(byteData) ;
	return status ;
}


tiff(filename,data,n,m)
	char *filename ;
	char *data ;
	int *n,*m ;
{

	unsigned char *address ;
	long 	nBytes,i,ncols,nrows ;
	FILE	*fp ;
	
	ncols = *n ;
	nrows = *m ;
	
	if ((fp = CreateTiffFile(filename)) != NULL) {
		WriteTiffDirectory(fp,kGrayScale,ncols,nrows) ;
		nBytes = ncols*sizeof(char) ;
		for (i=0,address= (unsigned char *)data;i<nrows;i++,address += ncols) {
			WriteTiffStrip(fp,address,nBytes) ;
		}
		CloseTiffFile(fp) ;
	}
	else {
		printf("\nFailed to create the file %s\n",filename) ;
	}
}

tiff_(filename,data,n,m)
	char *filename ;
	char *data ;
	int *n,*m ;
{

	unsigned char *address ;
	long 	nBytes,i,ncols,nrows ;
	FILE	*fp ;
	
	ncols = *n ;
	nrows = *m ;
	
	if ((fp = CreateTiffFile(filename)) != NULL) {
		WriteTiffDirectory(fp,kGrayScale,ncols,nrows) ;
		nBytes = ncols*sizeof(char) ;
		for (i=0,address= (unsigned char *)data;i<nrows;i++,address += ncols) {
			WriteTiffStrip(fp,address,nBytes) ;
		}
		CloseTiffFile(fp) ;
	}
	else {
		printf("\nFailed to create the file %s\n",filename) ;
	}
}

FILE *CreateTiffFile(char * filename) {
	FILE *fp ;
	
	fp = fopen(filename,"w") ;
	return fp ;
}

void CloseTiffFile(FILE *fp) {
	fclose(fp) ;
}

PIFD CreateTiffIFD(short numDE) {
	PIFD pifd ;
	
	pifd = (PIFD) malloc(sizeof(IFD)) ;
	pifd->numEntries = 0 ;
	pifd->entry = (PDIRENTRY)malloc(numDE*sizeof(DIRENTRY)) ;
	pifd->nextIFD = 0 ;
	return(pifd) ;
}

void WriteTiffIFD(FILE *fp,PIFD pifd){
	int numBytes ;
	
	numBytes = sizeof(short) ;
	fwrite(&pifd->numEntries,numBytes,1,fp) ;
	numBytes = pifd->numEntries*sizeof(DIRENTRY) ;
	fwrite((char *)pifd->entry,numBytes,1,fp) ;
	pifd->nextIFD = 0 ;
	numBytes = sizeof(int) ; /* long on some machines */
	fwrite((char *)&pifd->nextIFD,numBytes,1,fp) ;
}
void AddTiffTag(PIFD ifd,short TagID,short TagType,int numValues,int value) {
	
	ifd->entry[ifd->numEntries].deTag = TagID ;
	ifd->entry[ifd->numEntries].deType = TagType ;
	ifd->entry[ifd->numEntries].deLength = numValues ;
#ifdef WORDS_BIGENDIAN	
	/* We need to do this for Motorola Byte order */
	if (TagType == TIFFSHORT && numValues == 1) 
		value <<= 16 ;
#endif
	ifd->entry[ifd->numEntries].deVal = value ;
	ifd->numEntries++ ;
}
void WriteTiffDirectory(FILE *fp,TiffImageType imageType,int width,int height) {
	PIFD pifd ;
	int rational[4],numBytes ;
	int	value ;
	int tgxoffset,tgyoffset,tgstripoffset,tgpaletteoffset ;
	
	switch(imageType) {
		case kBiLevel:
			break ;
		case kGrayScale:
			tgxoffset = sizeof(TIFFHDR)+sizeof(short)+NEGRAYSCALE*sizeof(DIRENTRY)+sizeof(int)  ;
			tgyoffset = tgxoffset + 2*sizeof(int) ;
			tgstripoffset = tgyoffset + 2*sizeof(int) ;
			WriteTiffHeader(fp) ;
			pifd = CreateTiffIFD(NEGRAYSCALE) ;
			AddTiffTag(pifd,TGNEWSUBFILETYPE,TIFFLONG,1,0) ;
			AddTiffTag(pifd,TGIMAGEWIDTH,TIFFLONG,1,width) ;
			AddTiffTag(pifd,TGIMAGELENGTH,TIFFLONG,1,height) ;
			AddTiffTag(pifd,TGBITSPERSAMPLE,TIFFSHORT,1,8) ;
			AddTiffTag(pifd,TGCOMPRESSION,TIFFSHORT,1,1) ;
			AddTiffTag(pifd,TGPHOTOMETRICINTERPRETATION,TIFFSHORT,1,1) ;
			AddTiffTag(pifd,TGSTRIPOFFSETS,TIFFLONG,1,tgstripoffset) ;
			AddTiffTag(pifd,TGROWSPERSTRIP,TIFFLONG,1,height) ;
			AddTiffTag(pifd,TGSTRIPBYTECOUNTS,TIFFLONG,1,width*height) ;
			AddTiffTag(pifd,TGXRESOLUTION,TIFFRATIONAL,1,tgxoffset) ;
			AddTiffTag(pifd,TGYRESOLUTION,TIFFRATIONAL,1,tgyoffset) ;
			AddTiffTag(pifd,TGRESOLUTIONUNIT,TIFFSHORT,1,2) ;
			WriteTiffIFD(fp,pifd) ;
			rational[0] = 72 ;
			rational[1] = 1 ;
			rational[2] = 72 ;
			rational[3] = 1 ;
			numBytes = 4*sizeof(int) ;
			fwrite((char *)rational,numBytes,1,fp) ;
			free(pifd->entry) ;
			free(pifd) ;
			break ;
		case kPalette:
			tgxoffset = sizeof(TIFFHDR)+sizeof(short)+NEPALETTE*sizeof(DIRENTRY)+sizeof(int) ;
			tgyoffset = tgxoffset + 2*sizeof(int) ;
			tgpaletteoffset = tgyoffset + 2*sizeof(int) ;
			tgstripoffset = tgpaletteoffset + 3*256*sizeof(short) ;	/* The colortable	*/
			WriteTiffHeader(fp) ;
			pifd = CreateTiffIFD(NEPALETTE) ;
			AddTiffTag(pifd,TGNEWSUBFILETYPE,TIFFLONG,1,0) ;
			AddTiffTag(pifd,TGIMAGEWIDTH,TIFFLONG,1,width) ;
			AddTiffTag(pifd,TGIMAGELENGTH,TIFFLONG,1,height) ;
			AddTiffTag(pifd,TGBITSPERSAMPLE,TIFFSHORT,1,8) ;
			AddTiffTag(pifd,TGCOMPRESSION,TIFFSHORT,1,1) ;
			AddTiffTag(pifd,TGPHOTOMETRICINTERPRETATION,TIFFSHORT,1,3) ;
			AddTiffTag(pifd,TGSTRIPOFFSETS,TIFFLONG,1,tgstripoffset) ;
			AddTiffTag(pifd,TGROWSPERSTRIP,TIFFLONG,1,height) ;
			AddTiffTag(pifd,TGSTRIPBYTECOUNTS,TIFFLONG,1,width*height) ;
			AddTiffTag(pifd,TGXRESOLUTION,TIFFRATIONAL,1,tgxoffset) ;
			AddTiffTag(pifd,TGYRESOLUTION,TIFFRATIONAL,1,tgyoffset) ;
			AddTiffTag(pifd,TGRESOLUTIONUNIT,TIFFSHORT,1,2) ;
			AddTiffTag(pifd,TGCOLORMAP,TIFFSHORT,768,tgpaletteoffset) ;
			WriteTiffIFD(fp,pifd) ;
			rational[0] = 72 ;
			rational[1] = 1 ;
			rational[2] = 72 ;
			rational[3] = 1 ;
			numBytes = 4*sizeof(int) ;
			fwrite((char *)rational,numBytes,1,fp) ;
			free(pifd->entry) ;
			free(pifd) ;
			break ;
		case kRGB:
			tgxoffset = sizeof(TIFFHDR)+sizeof(short)+NERGB*sizeof(DIRENTRY)+sizeof(int) ;
			tgyoffset = tgxoffset + 2*sizeof(int) ;
			tgstripoffset = tgyoffset + 2*sizeof(int) ;
			WriteTiffHeader(fp) ;
			pifd = CreateTiffIFD(NERGB) ;
			AddTiffTag(pifd,TGNEWSUBFILETYPE,TIFFLONG,1,0) ;
			AddTiffTag(pifd,TGIMAGEWIDTH,TIFFLONG,1,width) ;
			AddTiffTag(pifd,TGIMAGELENGTH,TIFFLONG,1,height) ;
			AddTiffTag(pifd,TGBITSPERSAMPLE,TIFFSHORT,1,8) ;
			AddTiffTag(pifd,TGCOMPRESSION,TIFFSHORT,1,1) ;
			AddTiffTag(pifd,TGPHOTOMETRICINTERPRETATION,TIFFSHORT,1,2) ;
			AddTiffTag(pifd,TGSAMPLESPERPIXEL,TIFFSHORT,1,3) ;
			AddTiffTag(pifd,TGSTRIPOFFSETS,TIFFLONG,1,tgstripoffset) ;
			AddTiffTag(pifd,TGROWSPERSTRIP,TIFFLONG,1,height) ;
			AddTiffTag(pifd,TGSTRIPBYTECOUNTS,TIFFLONG,1,3*width*height) ;
			AddTiffTag(pifd,TGXRESOLUTION,TIFFRATIONAL,1,tgxoffset) ;
			AddTiffTag(pifd,TGYRESOLUTION,TIFFRATIONAL,1,tgyoffset) ;
			AddTiffTag(pifd,TGRESOLUTIONUNIT,TIFFSHORT,1,2) ;
			WriteTiffIFD(fp,pifd) ;
			rational[0] = 72 ;
			rational[1] = 1 ;
			rational[2] = 72 ;
			rational[3] = 1 ;
			numBytes = 4*sizeof(int) ;
			fwrite((char *)rational,numBytes,1,fp) ;
			free(pifd->entry) ;
			free(pifd) ;
			break ;
		case kFloatingPoint:
			tgxoffset = sizeof(TIFFHDR)+sizeof(short)+NEFLPOINT*sizeof(DIRENTRY)+sizeof(int)  ;
			tgyoffset = tgxoffset + 2*sizeof(int) ;
			tgstripoffset = tgyoffset + 2*sizeof(int) ;
			WriteTiffHeader(fp) ;
			pifd = CreateTiffIFD(NEFLPOINT) ;
			AddTiffTag(pifd,TGNEWSUBFILETYPE,TIFFLONG,1,0) ;
			AddTiffTag(pifd,TGIMAGEWIDTH,TIFFLONG,1,width) ;
			AddTiffTag(pifd,TGIMAGELENGTH,TIFFLONG,1,height) ;
			AddTiffTag(pifd,TGBITSPERSAMPLE,TIFFSHORT,1,8*sizeof(float)) ;
			AddTiffTag(pifd,TGCOMPRESSION,TIFFSHORT,1,1) ;
			AddTiffTag(pifd,TGPHOTOMETRICINTERPRETATION,TIFFSHORT,1,1) ;
			AddTiffTag(pifd,TGSAMPLEFORMAT,TIFFSHORT,1,3) ;
			AddTiffTag(pifd,TGSAMPLESPERPIXEL,TIFFSHORT,1,1) ;
			AddTiffTag(pifd,TGSTRIPOFFSETS,TIFFLONG,1,tgstripoffset) ;
			AddTiffTag(pifd,TGROWSPERSTRIP,TIFFLONG,1,height) ;
			AddTiffTag(pifd,TGSTRIPBYTECOUNTS,TIFFLONG,1,width*height*sizeof(float)) ;
			AddTiffTag(pifd,TGXRESOLUTION,TIFFRATIONAL,1,tgxoffset) ;
			AddTiffTag(pifd,TGYRESOLUTION,TIFFRATIONAL,1,tgyoffset) ;
			AddTiffTag(pifd,TGRESOLUTIONUNIT,TIFFSHORT,1,2) ;
			WriteTiffIFD(fp,pifd) ;
			rational[0] = 72 ;
			rational[1] = 1 ;
			rational[2] = 72 ;
			rational[3] = 1 ;
			numBytes = 4*sizeof(int) ;
			fwrite((char *)rational,numBytes,1,fp) ;
			free(pifd->entry) ;
			free(pifd) ;
			break ;
		case kComplexFloatingPoint:
			tgxoffset = sizeof(TIFFHDR)+sizeof(short)+NECFLPOINT*sizeof(DIRENTRY)+sizeof(int)  ;
			tgyoffset = tgxoffset + 2*sizeof(int) ;
			tgstripoffset = tgyoffset + 2*sizeof(int) ;
			WriteTiffHeader(fp) ;
			pifd = CreateTiffIFD(NECFLPOINT) ;
			AddTiffTag(pifd,TGNEWSUBFILETYPE,TIFFLONG,1,0) ;
			AddTiffTag(pifd,TGIMAGEWIDTH,TIFFLONG,1,width) ;
			AddTiffTag(pifd,TGIMAGELENGTH,TIFFLONG,1,height) ;
			value = 8*sizeof(float) ; value << 16 ; value +=  8*sizeof(float) ;
			AddTiffTag(pifd,TGBITSPERSAMPLE,TIFFSHORT,1,value) ;
			AddTiffTag(pifd,TGCOMPRESSION,TIFFSHORT,1,1) ;
			AddTiffTag(pifd,TGPHOTOMETRICINTERPRETATION,TIFFSHORT,1,1) ;
			AddTiffTag(pifd,TGSAMPLEFORMAT,TIFFSHORT,1,3) ;
			AddTiffTag(pifd,TGSAMPLESPERPIXEL,TIFFSHORT,1,2) ;
			AddTiffTag(pifd,TGSTRIPOFFSETS,TIFFLONG,1,tgstripoffset) ;
			AddTiffTag(pifd,TGROWSPERSTRIP,TIFFLONG,1,height) ;
			AddTiffTag(pifd,TGSTRIPBYTECOUNTS,TIFFLONG,1,2*width*height*sizeof(float)) ;
			AddTiffTag(pifd,TGXRESOLUTION,TIFFRATIONAL,1,tgxoffset) ;
			AddTiffTag(pifd,TGYRESOLUTION,TIFFRATIONAL,1,tgyoffset) ;
			AddTiffTag(pifd,TGRESOLUTIONUNIT,TIFFSHORT,1,2) ;
			WriteTiffIFD(fp,pifd) ;
			rational[0] = 72 ;
			rational[1] = 1 ;
			rational[2] = 72 ;
			rational[3] = 1 ;
			numBytes = 4*sizeof(int) ;
			fwrite((char *)rational,numBytes,1,fp) ;
			free(pifd->entry) ;
			free(pifd) ;
			break ;
	}
}
void WriteTiffPalette( FILE *fp, PaletteColor *colorTable, int numEntries) {
	short shortZero;
	int numBytes,i ;
	
	numBytes = sizeof(short) ;
	shortZero = 0 ;
	for(i=0;i<256;i++) {
		if (i < numEntries)
			fwrite(&colorTable[i].red,numBytes,1,fp) ;
		else 
			fwrite(&shortZero,numBytes,1,fp) ;
	}
	for(i=0;i<256;i++) {
		if (i < numEntries)
			fwrite(&colorTable[i].green,numBytes,1,fp) ;
		else 
			fwrite(&shortZero,numBytes,1,fp) ;
	}
	for(i=0;i<256;i++) {
		if (i < numEntries)
			fwrite(&colorTable[i].blue,numBytes,1,fp) ;
		else 
			fwrite(&shortZero,numBytes,1,fp) ;
	}
}
void WriteTiffStrip( FILE *fp, Ptr data,int nBytes) {
	int numBytes ;
	
	/* Write the data	*/
	numBytes = nBytes ;
	fwrite(data,numBytes,1,fp) ;
}
void WriteTiffHeader(FILE *fp) {
	TIFFHDR header ;
	int numBytes ;
	
	header.thByteOrder = BYTEORDER; 	
	header.thVersion = 42 ;
	header.thIfdOffset = 8 ;
	numBytes = sizeof(TIFFHDR) ;
	fwrite(&header,numBytes,1,fp) ;
}
void WriteTiffTag( short TagID, short TagType, int numValues,int  offset) {
	DIRENTRY ifd ;
	
	ifd.deTag = TagID ;
	ifd.deType = TagType ;
	ifd.deLength = numValues ;
	ifd.deVal = offset ;
}


Boolean ReadTiffDirectory(FILE *fp,TIFFHDR *thdr) {
	PDIRENTRY dirEntry,nextEntry ;
	long	nextIFD ;
	short	numEntries,i,j,*sp,bytesPerSample ;
	int		currFilePosition,*lp ;
	unsigned long	numBytes,numDataBytes ;
	unsigned char *address ;
	Boolean	InternalDataStorage = false ;
	
	/* Move to the position of the IFD	*/
	currFilePosition = ftell(fp) ;
	fseek(fp,thdr->thIfdOffset,0) ;
	/* Read the number of entries	*/
	if (fread(&numEntries,sizeof(short),1,fp) != 1) {
		printf("\nError reading number of entries\n") ; fflush(stdout) ;
		return false;
	}

	if (thdr->thByteOrder != BYTEORDER) {
		numEntries = ByteSwapShort(numEntries) ;
	}
    
	dirEntry = (PDIRENTRY)malloc(numEntries*sizeof(DIRENTRY)) ;
	if (dirEntry == 0) {
		printf("\nError allocating %d number of entries\n",numEntries) ; 
		fflush(stdout) ;
		return false ;
	}
	
	gTiff.samplesPerPixel = 1 ; /* In case it does not get set later */
	gTiff.sampleFormat = 1 ;

	
	/* Read the directory entries 	*/
	if (fread(dirEntry,numEntries*sizeof(DIRENTRY),1,fp) != 1) {
		free(dirEntry) ;
		printf("\nError reading entries\n") ; fflush(stdout) ;
		return false ;
	}	
	/*	Read the pointer to the next directory	*/	
	if (fread((char *)&nextIFD,sizeof(int),1,fp) != 1) {
		printf("\nError reading next IFD\n") ; fflush(stdout) ;
		free(dirEntry) ;
		return false ;
	}
	/* Process the tags	*/
	nextEntry = dirEntry ;
	for(i=0;i<numEntries;i++) {
		ProcessTiffDirEntry(fp,thdr,nextEntry) ;
		nextEntry++ ;
	} 
	
	/* Check to see if this is a kind of TIFF file that we support	*/
	switch(gTiff.photoMetricInterpretation) {
		case 0:	/* WhiteIsZero	*/
		case 1: /* WhiteIsBlack	*/
			break ;	
		case 2:	/* RGB	*/
			break ;
		case 3: /* RGB Palette */
			break ;
		case 4:	/* Transparency Mask */
			/* Unsupported	*/
			printf("\nApplication does not support Transparency Mask Tiff Files\n") ; fflush(stdout) ;
			free(dirEntry) ;
			return false ;
			break ;
		case 5:	/* CMYK */
			/* Unsupported	*/
			printf("\nApplication does not support CMYK Tiff Files\n") ; fflush(stdout) ;
			free(dirEntry) ;
			return false ;
			break ;
		case 6:	/* YCbCr */
			/* Unsupported	*/
			printf("\nApplication does not support YCbCr Tiff Files\n") ; fflush(stdout) ;
			free(dirEntry) ;
			return false ;
			break ;
		case 8:	/* CIELab */
			/* Unsupported	*/
			printf("\nApplication does not support CIELab Tiff Files\n") ; fflush(stdout) ;
			dirEntry ; /* free(dirEntry) ; */
			return false ;
			break ;
	}
	switch(gTiff.compression) {
		case 1:		/* Uncompressed */
			break ;
		default:	/* Compressed -- Unsupported */
			printf("\nApplication does not support compressed Tiff Files\n") ; 
			fflush(stdout) ;
			free(dirEntry) ;
			return false ;
			break ;
	}
			
	/* Read the colormap if one exists	*/
	if (gTiff.photoMetricInterpretation == 3 && gTiff.numPaletteColors != 0) {	
		/* Palette Image */
		currFilePosition = ftell(fp) ;
		fseek(fp,gTiff.paletteOffset,0) ;
		if (gTiff.palette != nil) {
			free(gTiff.palette) ;
		}
		numBytes = gTiff.numPaletteColors*sizeof(PaletteColor) ; 
		gTiff.palette = (PaletteColor *)malloc(numBytes) ;
		if (gTiff.palette != nil) {
			fread(gTiff.palette,sizeof(PaletteColor),gTiff.numPaletteColors,fp) ;
			/* Arrange the palette into its regular order. 
			   It was read in as RRRR..GGGG...BBB
			   while it should be organized into RGBRGBRGB
			   Also it may have to be byte-swapped */
			free(gTiff.palette) ;
			gTiff.palette = NULL ;
		}
		fseek(fp,currFilePosition,0) ;
	}
	
	/* Read the Image data	*/
	/* Move to the correct position	*/
	currFilePosition = ftell(fp) ;
	
	if (gTiff.numberOfStrips > 1) {
		/* Get the appropriate offsets and bytecounts	*/
		fseek(fp,*gTiff.stripOffset,0) ;
		for(i=0;i<gTiff.numberOfStrips;i++) {
			fread(gTiff.stripOffset+i,sizeof(int),1,fp) ;
		    if (thdr->thByteOrder != BYTEORDER) {
		    	*(gTiff.stripOffset+i) = ByteSwapLong(*(gTiff.stripOffset+i)) ;
		    }
			if (gTiff.stripOffsetType == TIFFSHORT && thdr->thByteOrder == kMM) {
				*(gTiff.stripOffset+i) >> 16 ;
			}
		}
		fseek(fp,*gTiff.stripByteCount,0) ;
		numDataBytes = 0 ;
		for(i=0;i<gTiff.numberOfStrips;i++) {
			fread(gTiff.stripByteCount+i,sizeof(int),1,fp) ;
		    if (thdr->thByteOrder != BYTEORDER) {
		    	*(gTiff.stripByteCount+i) = ByteSwapLong(*(gTiff.stripByteCount+i)) ;
		    }
			if (gTiff.stripByteCountType == TIFFSHORT && thdr->thByteOrder == kMM) {
				*(gTiff.stripByteCount+i) >> 16 ;
			}
			numDataBytes += *(gTiff.stripByteCount+i) ;
		}
	}
	else {
		if (gTiff.stripByteCount == nil) {
			/* I don't know why CRISP does not set this tag */
			gTiff.stripByteCount = (int *)malloc(32*sizeof(int)) ;
			numDataBytes = gTiff.width*gTiff.height*gTiff.bitsPerSample/8 ;
			gTiff.stripByteCount[0] = numDataBytes ;
		}
		else {
			numDataBytes = gTiff.stripByteCount[0] ;
		}
	}				
			
	gTiff.data = (unsigned char *)malloc(numDataBytes) ;
	if (gTiff.data == NULL) {
		fprintf(stderr,"\nMemory allocation error") ; fflush(stderr) ;
		return false ;
	}
		
	address = gTiff.data ;
	for(i=0;i<gTiff.numberOfStrips;i++) {
		fseek(fp,gTiff.stripOffset[i],0) ;
		/* Read the values	*/
		fread(address,sizeof(char),gTiff.stripByteCount[i],fp) ;
		address += gTiff.stripByteCount[i] ;
	}
	
	free(gTiff.stripOffset) ;
	free(gTiff.stripByteCount) ;
	
	if (thdr->thByteOrder != BYTEORDER) {
		SwapDataBytes(gTiff.data,gTiff.width,gTiff.height,
    			gTiff.sampleFormat,gTiff.bitsPerSample,gTiff.samplesPerPixel) ;
	}
	
	fseek(fp,currFilePosition,0) ;
	free(dirEntry) ;
	return true ;
}

void SwapDataBytes(unsigned char *data,int width,int height,
		short sampleFormat,short bitsPerSample,short samplesPerPixel) {
	int bytesPerSample,i,numValues ;
	unsigned short	*sp ;
	unsigned int	*lp ;
	float	*fp ;

	bytesPerSample = bitsPerSample/sizeof(char) ;
	numValues = samplesPerPixel*width*height ;
	
	switch(sampleFormat) {
		case 1:	/* Unsigned integer data */
		case 2:	/* Twos's complement signed integer data	*/
		case 4:	/* Undefined, treat as unsigned integer data	*/
			switch(bytesPerSample) {
				case 2:
					sp = (unsigned short *)data ;
					for(i=0;i<numValues;i++) {
						*sp = ByteSwapShort(*sp) ;
						sp++ ;
					}
					break ;
				case 4:
					lp = (unsigned int *)gTiff.data ;
					for(i=0;i<numValues;i++) {
						*lp = ByteSwapLong(*lp) ;
						lp++ ;
					}
					break ;
				default:
					break ;
			}
			break ;
		case 3:	/* IEEE floating point data [IEEE]	*/
			lp = (unsigned int *)data ;
			fp = (float *)data ;
			for(i=0;i<numValues;i++) {
				*fp = ByteSwapFloat(*lp) ;
				lp++ ;
				fp++ ;
			}
			break ;
	}

}

Boolean ProcessTiffDirEntry(FILE *fp,TIFFHDR *thdr,PDIRENTRY dirEntry){
	short	TagID,TagType,svalue ;
	int		numValues,value ;
	int		debug = 0 ;

    if (thdr->thByteOrder != BYTEORDER) {
    	TagID = ByteSwapShort(dirEntry->deTag) ;
    	TagType = ByteSwapShort(dirEntry->deType) ;
    	numValues = ByteSwapLong(dirEntry->deLength) ;
    	value = ByteSwapLong(dirEntry->deVal) ;
    	if (TagType == TIFFSHORT && thdr->thByteOrder == kMM && numValues == 1 ) 
			value >>= 16 ;
    }
	else {
		TagID = dirEntry->deTag ;
		TagType = dirEntry->deType ;
		numValues = dirEntry->deLength ;
		value = dirEntry->deVal ;
    	if (TagType == TIFFSHORT && thdr->thByteOrder == kMM && numValues == 1) 
			value >>= 16 ;
	}
		
	switch(TagID) {
		case TGNEWSUBFILETYPE:
		case TGOLDSUBFILETYPE:
			break ;
		case TGIMAGEWIDTH:
			gTiff.width = value ;
			break ;
		case TGIMAGELENGTH:
			gTiff.height = value ;
			break ;
		case TGBITSPERSAMPLE:
			gTiff.bitsPerSample = value ;
			break ;
		case TGCOMPRESSION:
			gTiff.compression = value ;
			break ;
		case TGPHOTOMETRICINTERPRETATION:
			gTiff.photoMetricInterpretation = value ;
			break ;
		case TGSTRIPOFFSETS:
			gTiff.numberOfStrips = numValues ;
			gTiff.stripOffset = (int *)malloc(numValues*sizeof(int)) ;
			*gTiff.stripOffset = value ;
			break ;
		case TGSAMPLESPERPIXEL:
			gTiff.samplesPerPixel = value ;
			break ;
		case TGROWSPERSTRIP:
			gTiff.rowPerStrip = value ;
			break ;
		case TGSTRIPBYTECOUNTS:
			gTiff.numberOfStrips = numValues ;
			gTiff.stripByteCount = (int *)malloc(numValues*sizeof(int)) ;
			*gTiff.stripByteCount = value ;
			break ;
		case TGXRESOLUTION:
			gTiff.tgxOffset = value ;
			break ;
		case TGYRESOLUTION:
			gTiff.tgyOffset = value ;
			break ;
		case TGRESOLUTIONUNIT:
			gTiff.resolutionUnit = value ;
			break ;
		case TGCOLORMAP:
			gTiff.numPaletteColors = numValues/3 ;	/* RGB */
			gTiff.paletteOffset = value ;
			break ;
		case TGSAMPLEFORMAT:
			gTiff.sampleFormat = value ;
			break ;
		case TGORIENTATION:
			/* We don't process	*/
			break ;
		default:
			break ;
	}
	return true ;
}

Boolean ReadTiffHeader(FILE *fp,TIFFHDR *header) {
	int numBytes ;
	
	numBytes = sizeof(TIFFHDR) ;
	if (fread(header,numBytes,1,fp) != 1) {
        	return false;
    	}
    	/* If we need to byteswap, do so ! */
   	if (header->thByteOrder != BYTEORDER) {
    		header->thVersion = ByteSwapShort(header->thVersion) ;
    		header->thIfdOffset = ByteSwapLong(header->thIfdOffset) ;
    	}
    	return true ;
}

unsigned short ByteSwapShort(unsigned short value) {
	unsigned short leftByte,rightByte ;
	unsigned short retValue ;
	
	leftByte = (value & 0xFF00) >> 8 ;
	rightByte = (value & 0x00FF) << 8 ;
	
	retValue = leftByte | rightByte ;	
	return retValue ;
}


unsigned long ByteSwapLong(unsigned long value) {
	unsigned long leftByte,rightByte,leftMiddleByte,rightMiddleByte ;
	unsigned long retValue ;
	
	leftByte = (value & 0xFF000000) >> 24 ;
	leftMiddleByte = (value & 0x00FF0000) >> 8 ;
	rightByte = (value & 0x000000FF) << 24 ;
	rightMiddleByte = (value & 0x0000FF00) << 8 ;
	
	retValue = leftByte | rightByte | leftMiddleByte | rightMiddleByte ;	
	return retValue ;
}

float ByteSwapFloat(unsigned long value) {
	unsigned long leftByte,rightByte,leftMiddleByte,rightMiddleByte,ivalue ;
	float retValue,*fp;
		
	leftByte = (value & 0xFF000000) >> 24 ;
	leftMiddleByte = (value & 0x00FF0000) >> 8 ;
	rightByte = (value & 0x000000FF) << 24 ;
	rightMiddleByte = (value & 0x0000FF00) << 8 ;
		
	ivalue = leftByte | rightByte | leftMiddleByte | rightMiddleByte ;	
	
	fp = (float *)&ivalue ;
	retValue = *fp ;
	return retValue ;
}

