// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include "storage.h"
#include "ProjCellRecord.h"
#include "funcdefs.h"
#define fabs(x) 	((x) < 0. ? -(x) : (x))
// -----------------------------------------------------------
// Straighten the unit cell in case GAMMA is different from 90
// ----------------------------------------------------------- 
void AdjustGamma(long *np,long *mp,long nx,long ny) {
	float a_axis,b_axis,del_x,del_y,shift_i_cnst,
	shift_i,dleft,ri ;
	long jmax,ihi,ilo,numBytes ;
	register float *rp1,*rp2 ;
	register long j,i ;

	if (fabs(gProjCell->ga - 90.) < 0.005) return ;  // Nothing to do

	jmax = *np*(*mp) ;

	a_axis = gProjCell->a*nx ;			// Total length of A axis
	b_axis = gProjCell->b*ny ;			// Total length of B axis
	del_x = a_axis/(*np) ;			// Delta x
	del_y = b_axis/(*mp) ;			// Delta y
	shift_i_cnst = - gProjCell->csga*
	           del_y/del_x  ; 		// Distance from upright y-axis to
	                          		// the old skewed y-axis when multiplied
	                          		// by (j-1). The distance has been scaled
	                          		// by dividing by DEL_X to give offset in
	                          		// index.				 
	rp1 = gWorkSpace ;
	rp2 = gWorkSpace + jmax ;		// Place for temporary row
	numBytes = (*np)*sizeof(float) ;
	for (j=0;j<*mp;j++)	{			// LOOP OVER ALL THE ROWS
		memcpy((char *)rp2,(char *)rp1,numBytes) ;	// Temporary store the row

		shift_i = shift_i_cnst*j ;	// Offset in index i to get the total
			       					// index i into the old skewed system.
		dleft = shift_i - (long)shift_i ;	// Fractional distance from point
											// to the left.
		if (dleft > 0.999) {
			dleft = 1. ;
			shift_i = (long)shift_i ;
		}
		for(i=0;i<*np;i++)	{		// LOOP OVER ALL THE COLUMNS
			ri = i + shift_i ;		// Old index I of this point. It is real
									// because it may lay between two indices.
			ilo = ri ;				// Idex of point to the left
			ihi = ilo + 1 ;			// I index of nearest point to the right
			ilo = ilo%(*np) ;		// Wrap around the index if necessary
			ihi = ihi%(*np) ;
			*rp1++ = (1.-dleft)*(*(rp2+ilo)) + dleft*(*(rp2+ihi)) ;
		}
	}
}
