// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include "XDrivers.h"
#include "CSelection.h"

extern	GrafPtr	gCurrentPort ;
#define	HSIZE	3
CSelection::CSelection(Rect *clipRect)
{
	SetRect(&mSelectionRect,0,0,0,0) ;
	SetRect(&mTopLeftCorner,0,0,0,0) ;
	SetRect(&mTopRightCorner,0,0,0,0) ;
	SetRect(&mBottomLeftCorner,0,0,0,0) ;
	SetRect(&mBottomRightCorner,0,0,0,0) ;
	SetRect(&mTopMiddle,0,0,0,0) ;
	SetRect(&mLeftMiddle,0,0,0,0) ;
	SetRect(&mBottomMiddle,0,0,0,0) ;
	SetRect(&mRightMiddle,0,0,0,0) ;

	mClipRgn = nil ;
	if (clipRect != nil) {
		mClipRgn = NewRgn() ;
		RectRgn(mClipRgn,clipRect) ;
	}
}

CSelection::~CSelection() {
	if (mClipRgn != nil) 
		SetClip(mClipRgn) ;
	Draw(true) ; 
	if (mClipRgn != nil) {
		DisposeRgn(mClipRgn) ;
		SetClip(gCurrentPort->GetPortRect()) ;
	}
} ;

void CSelection::Modify(Point &loc)
{
	Point where = loc ;
	if (mClipRgn != nil) 
		SetClip(mClipRgn) ;
	
	if (PtInRect(&where,&mTopLeftCorner)) {
		DragTopLeft(where) ;
		return ;
	}
	if (PtInRect(&where,&mTopRightCorner)) {
		DragTopRight(where) ;
		return ;
	}
	if (PtInRect(&where,&mBottomLeftCorner)) {
		DragBottomLeft(where) ;
		return ;
	}
	if (PtInRect(&where,&mBottomRightCorner)) {
		DragBottomRight(where) ;
		return ;
	}
	if (PtInRect(&where,&mTopMiddle)) {
		DragTop(where) ;
		return ;
	}
	if (PtInRect(&where,&mLeftMiddle)) {
		DragLeft(where) ;
		return ;
	}
	if (PtInRect(&where,&mBottomMiddle)) {
		DragBottom(where) ;
		return ;
	}
	if (PtInRect(&where,&mRightMiddle)) {
		DragRight(where) ;
		return ;
	}
	if (PtInRect(&where,&mSelectionRect)) {
		Shift(where) ;
		return ;
	}
	
	SetClip(gCurrentPort->GetPortRect()) ;
	// We are outside an already existing region
	// Do nothing
}

void CSelection::DoMouseDown(Point &loc)
{
	Point where = loc ;

	if (mClipRgn != nil) 
		SetClip(mClipRgn) ;
	
	if (PtInRect(&where,&mTopLeftCorner)) {
		DragTopLeft(where) ;
		return ;
	}
	if (PtInRect(&where,&mTopRightCorner)) {
		DragTopRight(where) ;
		return ;
	}
	if (PtInRect(&where,&mBottomLeftCorner)) {
		DragBottomLeft(where) ;
		return ;
	}
	if (PtInRect(&where,&mBottomRightCorner)) {
		DragBottomRight(where) ;
		return ;
	}
	if (PtInRect(&where,&mTopMiddle)) {
		DragTop(where) ;
		return ;
	}
	if (PtInRect(&where,&mLeftMiddle)) {
		DragLeft(where) ;
		return ;
	}
	if (PtInRect(&where,&mBottomMiddle)) {
		DragBottom(where) ;
		return ;
	}
	if (PtInRect(&where,&mRightMiddle)) {
		DragRight(where) ;
		return ;
	}
	if (PtInRect(&where,&mSelectionRect)) {
		Shift(where) ;
		return ;
	}
	
	// We are outside an already existing region
	// We start a new one . 
	// It will be erase in the Drag routine
	Drag(where) ;	
	SetClip(gCurrentPort->GetPortRect()) ;
}

void CSelection::Zoom(float factor,short x,short y) {
	short width,height,x0,y0 ;
	
	width = mSelectionRect.right - mSelectionRect.left ;
	height = mSelectionRect.bottom - mSelectionRect.top ;
	width = (long) (width*factor + 0.5) ;
	height = (long) (height*factor + 0.5) ;
	x0 = (mSelectionRect.right + mSelectionRect.left)/2 + x ;
	y0 = (mSelectionRect.bottom + mSelectionRect.top)/2 + y ;
	mSelectionRect.left = x0 - width/2 ;
	mSelectionRect.top = y0 - height/2 ;
	mSelectionRect.right = mSelectionRect.left + width ;
	mSelectionRect.bottom = mSelectionRect.top + height ;
	SetActiveHandles() ;
}

void CSelection::Update(Rect *clipRect) {
	RgnHandle	clipRgn,savedClip ;
	GrafPtr	currentPort ;
	
	GetPort(&currentPort) ;
	savedClip = NewRgn() ;
	clipRgn = NewRgn() ;
	RectRgn(savedClip,&currentPort->GetPortRect()) ;
	InsetRect(clipRect,1,1) ;
	RectRgn(clipRgn,clipRect) ;
	SetClip(clipRgn) ;
	Draw(true) ;
	SetClip(savedClip) ;
	DisposeRgn(clipRgn) ;
	DisposeRgn(savedClip) ;
	InsetRect(clipRect,-1,-1) ;
}
void CSelection::Pan(short dx,short dy) {
	OffsetRect(&mSelectionRect,dx,dy) ;
	SetActiveHandles() ;
}

void CSelection::DrawHandles() {
	PenMode(kXOr);
	FrameRect(&mTopLeftCorner) ;
	FrameRect(&mTopRightCorner) ;
	FrameRect(&mBottomLeftCorner) ;
	FrameRect(&mBottomRightCorner) ;
	FrameRect(&mTopMiddle) ;
	FrameRect(&mLeftMiddle) ;
	FrameRect(&mBottomMiddle) ;
	FrameRect(&mRightMiddle) ;	
	PenNormal() ;
}
void CSelection::Draw(Boolean drawHandles) {
	PenMode(kXOr);
	//PenPat(&qd.gray);
	FrameRect(&mSelectionRect) ;	
	PenNormal() ;
	if (drawHandles) {
		DrawHandles() ;
	}
}
void 	CSelection::Drag(Point &loc) {
	Point lastPoint,currentPoint,startPoint ;
	long top,left,right,bottom ;
	
	currentPoint = loc ; ;
	lastPoint = currentPoint ;
	startPoint = currentPoint ;
	top = lastPoint.y ;
	bottom = lastPoint.y ;
	left = lastPoint.x ;
	right = lastPoint.x ;
	Draw(true) ;	// erase old
	SetRect(&mSelectionRect,left,top,right,bottom) ;
	Draw() ;		// draw (actually empty drawing)
	while(Button()) {
		GetMouse(&currentPoint) ;
		if (currentPoint.x != lastPoint.x || currentPoint.y != lastPoint.y) {
			Draw() ;
			left = (currentPoint.x < startPoint.x ? currentPoint.x : startPoint.x) ;
			right = (currentPoint.x > startPoint.x ? currentPoint.x : startPoint.x) ;
			top = (currentPoint.y < startPoint.y ? currentPoint.y : startPoint.y) ;
			bottom = (currentPoint.y > startPoint.y ? currentPoint.y : startPoint.y) ;
			SetRect(&mSelectionRect,left,top,right,bottom) ;
			Draw() ;
			lastPoint = currentPoint ;
		}
	}
	SetActiveHandles() ;
	DrawHandles() ;
}

void 	CSelection::SetActiveHandles() {
	if ((mSelectionRect.left != mSelectionRect.right) && (mSelectionRect.top != mSelectionRect.bottom)) {
		SetRect(&mTopLeftCorner,mSelectionRect.left-HSIZE,mSelectionRect.top-HSIZE,mSelectionRect.left+HSIZE,mSelectionRect.top+HSIZE) ;
		SetRect(&mTopRightCorner,mSelectionRect.right-HSIZE,mSelectionRect.top-HSIZE,mSelectionRect.right+HSIZE,mSelectionRect.top+HSIZE) ;
		SetRect(&mBottomLeftCorner,mSelectionRect.left-HSIZE,mSelectionRect.bottom-HSIZE,mSelectionRect.left+HSIZE,mSelectionRect.bottom+HSIZE) ;
		SetRect(&mBottomRightCorner,mSelectionRect.right-HSIZE,mSelectionRect.bottom-HSIZE,mSelectionRect.right+HSIZE,mSelectionRect.bottom+HSIZE) ;
		SetRect(&mTopMiddle,(mSelectionRect.left+mSelectionRect.right)/2-HSIZE,mSelectionRect.top-HSIZE,
				(mSelectionRect.left+mSelectionRect.right)/2+HSIZE,mSelectionRect.top+HSIZE) ;
		SetRect(&mLeftMiddle,mSelectionRect.left-HSIZE,(mSelectionRect.top+mSelectionRect.bottom)/2-HSIZE,
				mSelectionRect.left+HSIZE,(mSelectionRect.top+mSelectionRect.bottom)/2+HSIZE) ;
		SetRect(&mBottomMiddle,(mSelectionRect.left+mSelectionRect.right)/2-HSIZE,mSelectionRect.bottom-HSIZE,
				(mSelectionRect.left+mSelectionRect.right)/2+HSIZE,mSelectionRect.bottom+HSIZE) ;
		SetRect(&mRightMiddle,mSelectionRect.right-HSIZE,(mSelectionRect.top+mSelectionRect.bottom)/2-HSIZE,
				mSelectionRect.right+HSIZE,(mSelectionRect.top+mSelectionRect.bottom)/2+HSIZE) ;	
	}
	else {
		SetRect(&mTopLeftCorner,0,0,0,0) ;
		SetRect(&mTopRightCorner,0,0,0,0) ;
		SetRect(&mBottomLeftCorner,0,0,0,0) ;
		SetRect(&mBottomRightCorner,0,0,0,0) ;
		SetRect(&mTopMiddle,0,0,0,0) ;
		SetRect(&mLeftMiddle,0,0,0,0) ;
		SetRect(&mBottomMiddle,0,0,0,0) ;
		SetRect(&mRightMiddle,0,0,0,0) ;
	}	
}

void 	CSelection::DragTopLeft(Point &loc) {
	Point lastPoint,currentPoint,startPoint,topLeft,bottomRight ;
	long top,left,right,bottom ;
	
	startPoint = loc ;
	lastPoint = startPoint ;
	top = mSelectionRect.top ;
	bottom = mSelectionRect.bottom ;
	left = mSelectionRect.left ;
	right = mSelectionRect.right ;
	topLeft.x = left ;
	topLeft.y = top ;
	bottomRight.x = right ;
	bottomRight.y = bottom ;
	Draw(true) ;	// To erase ourself
	Draw() ;		// to draw without handles
	while(Button()) {
		GetMouse(&currentPoint) ;
		if (currentPoint.x != lastPoint.x || currentPoint.y != lastPoint.y) {
			Draw() ;
			topLeft.x += currentPoint.x - lastPoint.x ;
			topLeft.y += currentPoint.y - lastPoint.y ;
			left = (topLeft.x < bottomRight.x ? topLeft.x : bottomRight.x) ;
			right = (topLeft.x > bottomRight.x ? topLeft.x : bottomRight.x) ;
			top = (topLeft.y < bottomRight.y ? topLeft.y : bottomRight.y) ;
			bottom = (topLeft.y > bottomRight.y ? topLeft.y : bottomRight.y) ;
			SetRect(&mSelectionRect,left,top,right,bottom) ;
			Draw() ;
			lastPoint = currentPoint ;
		}
	}
	SetActiveHandles() ;
	DrawHandles() ;
}
void 	CSelection::DragTopRight(Point &loc) {
	Point lastPoint,currentPoint,startPoint,topLeft,bottomRight ;
	long top,left,right,bottom ;
	
	startPoint = loc ;
	lastPoint = startPoint ;
	top = mSelectionRect.top ;
	bottom = mSelectionRect.bottom ;
	left = mSelectionRect.left ;
	right = mSelectionRect.right ;
	topLeft.x = left ;
	topLeft.y = top ;
	bottomRight.x = right ;
	bottomRight.y = bottom ;
	Draw(true) ;	// To erase ourself
	Draw() ;		// to draw without handles
	while(Button()) {
		GetMouse(&currentPoint) ;
		if (currentPoint.x != lastPoint.x || currentPoint.y != lastPoint.y) {
			Draw() ;
			bottomRight.x += currentPoint.x - lastPoint.x ;
			topLeft.y += currentPoint.y - lastPoint.y ;
			left = (topLeft.x < bottomRight.x ? topLeft.x : bottomRight.x) ;
			right = (topLeft.x > bottomRight.x ? topLeft.x : bottomRight.x) ;
			top = (topLeft.y < bottomRight.y ? topLeft.y : bottomRight.y) ;
			bottom = (topLeft.y > bottomRight.y ? topLeft.y : bottomRight.y) ;
			SetRect(&mSelectionRect,left,top,right,bottom) ;
			Draw() ;
			lastPoint = currentPoint ;
		}
	}
	SetActiveHandles() ;
	DrawHandles() ;
}
void 	CSelection::DragBottomLeft(Point &loc) {
	Point lastPoint,currentPoint,startPoint,topLeft,bottomRight ;
	long top,left,right,bottom ;
	
	startPoint = loc ;
	lastPoint = startPoint ;
	top = mSelectionRect.top ;
	bottom = mSelectionRect.bottom ;
	left = mSelectionRect.left ;
	right = mSelectionRect.right ;
	topLeft.x = left ;
	topLeft.y = top ;
	bottomRight.x = right ;
	bottomRight.y = bottom ;
	Draw(true) ;	// To erase ourself
	Draw() ;		// to draw without handles
	while(Button()) {
		GetMouse(&currentPoint) ;
		if (currentPoint.x != lastPoint.x || currentPoint.y != lastPoint.y) {
			Draw() ;
			topLeft.x += currentPoint.x - lastPoint.x ;
			bottomRight.y += currentPoint.y - lastPoint.y ;
			left = (topLeft.x < bottomRight.x ? topLeft.x : bottomRight.x) ;
			right = (topLeft.x > bottomRight.x ? topLeft.x : bottomRight.x) ;
			top = (topLeft.y < bottomRight.y ? topLeft.y : bottomRight.y) ;
			bottom = (topLeft.y > bottomRight.y ? topLeft.y : bottomRight.y) ;
			SetRect(&mSelectionRect,left,top,right,bottom) ;
			Draw() ;
			lastPoint = currentPoint ;
		}
	}
	SetActiveHandles() ;
	DrawHandles() ;
}
void 	CSelection::DragBottomRight(Point &loc) {
	Point lastPoint,currentPoint,startPoint,topLeft,bottomRight ;
	long top,left,right,bottom ;
	
	startPoint = loc ;
	lastPoint = startPoint ;
	top = mSelectionRect.top ;
	bottom = mSelectionRect.bottom ;
	left = mSelectionRect.left ;
	right = mSelectionRect.right ;
	topLeft.x = left ;
	topLeft.y = top ;
	bottomRight.x = right ;
	bottomRight.y = bottom ;
	Draw(true) ;	// To erase ourself
	Draw() ;		// to draw without handles
	while(Button()) {
		GetMouse(&currentPoint) ;
		if (currentPoint.x != lastPoint.x || currentPoint.y != lastPoint.y) {
			Draw() ;
			bottomRight.x += currentPoint.x - lastPoint.x ;
			bottomRight.y += currentPoint.y - lastPoint.y ;
			left = (topLeft.x < bottomRight.x ? topLeft.x : bottomRight.x) ;
			right = (topLeft.x > bottomRight.x ? topLeft.x : bottomRight.x) ;
			top = (topLeft.y < bottomRight.y ? topLeft.y : bottomRight.y) ;
			bottom = (topLeft.y > bottomRight.y ? topLeft.y : bottomRight.y) ;
			SetRect(&mSelectionRect,left,top,right,bottom) ;
			Draw() ;
			lastPoint = currentPoint ;
		}
	}
	SetActiveHandles() ;
	DrawHandles() ;
}
void 	CSelection::DragTop(Point &loc) {
	Point lastPoint,currentPoint,startPoint,topLeft,bottomRight ;
	long top,left,right,bottom ;
	
	startPoint = loc ;
	lastPoint = startPoint ;
	top = mSelectionRect.top ;
	bottom = mSelectionRect.bottom ;
	left = mSelectionRect.left ;
	right = mSelectionRect.right ;
	topLeft.x = left ;
	topLeft.y = top ;
	bottomRight.x = right ;
	bottomRight.y = bottom ;
	Draw(true) ;	// To erase ourself
	Draw() ;		// to draw without handles
	while(Button()) {
		GetMouse(&currentPoint) ;
		if (currentPoint.y != lastPoint.y) {
			Draw() ;
			topLeft.y += currentPoint.y - lastPoint.y ;
			top = (topLeft.y < bottomRight.y ? topLeft.y : bottomRight.y) ;
			bottom = (topLeft.y > bottomRight.y ? topLeft.y : bottomRight.y) ;
			SetRect(&mSelectionRect,left,top,right,bottom) ;
			Draw() ;
			lastPoint = currentPoint ;
		}
	}
	SetActiveHandles() ;
	DrawHandles() ;
}
void 	CSelection::DragLeft(Point &loc) {
	Point lastPoint,currentPoint,startPoint,topLeft,bottomRight ;
	long top,left,right,bottom ;
	
	startPoint = loc ;
	lastPoint = startPoint ;
	top = mSelectionRect.top ;
	bottom = mSelectionRect.bottom ;
	left = mSelectionRect.left ;
	right = mSelectionRect.right ;
	topLeft.x = left ;
	topLeft.y = top ;
	bottomRight.x = right ;
	bottomRight.y = bottom ;
	Draw(true) ;	// to erase the previous
	Draw() ;		// to draw again without handles
	while(Button()) {
		GetMouse(&currentPoint) ;
		if (currentPoint.x != lastPoint.x) {
			Draw() ;
			topLeft.x += currentPoint.x - lastPoint.x ;
			left = (topLeft.x < bottomRight.x ? topLeft.x : bottomRight.x) ;
			right = (topLeft.x > bottomRight.x ? topLeft.x : bottomRight.x) ;
			SetRect(&mSelectionRect,left,top,right,bottom) ;
			Draw() ;
			lastPoint = currentPoint ;
		}
	}
	SetActiveHandles() ;
	DrawHandles() ;
}
void 	CSelection::DragBottom(Point &loc) {
	Point lastPoint,currentPoint,startPoint,topLeft,bottomRight ;
	long top,left,right,bottom ;
	
	startPoint = loc ;
	lastPoint = startPoint ;
	top = mSelectionRect.top ;
	bottom = mSelectionRect.bottom ;
	left = mSelectionRect.left ;
	right = mSelectionRect.right ;
	topLeft.x = left ;
	topLeft.y = top ;
	bottomRight.x = right ;
	bottomRight.y = bottom ;
	Draw(true) ;	// to erase the previous
	Draw() ;		// to draw again without handles
	while(Button()) {
		GetMouse(&currentPoint) ;
		if (currentPoint.y != lastPoint.y) {
			Draw() ;
			bottomRight.y += currentPoint.y - lastPoint.y ;
			top = (topLeft.y < bottomRight.y ? topLeft.y : bottomRight.y) ;
			bottom = (topLeft.y > bottomRight.y ? topLeft.y : bottomRight.y) ;
			SetRect(&mSelectionRect,left,top,right,bottom) ;
			Draw() ;
			lastPoint = currentPoint ;
		}
	}
	SetActiveHandles() ;
	DrawHandles() ;
}
void 	CSelection::DragRight(Point &loc) {
	Point lastPoint,currentPoint,startPoint,topLeft,bottomRight ;
	long top,left,right,bottom ;
	
	startPoint = loc ;
	lastPoint = startPoint ;
	top = mSelectionRect.top ;
	bottom = mSelectionRect.bottom ;
	left = mSelectionRect.left ;
	right = mSelectionRect.right ;
	topLeft.x = left ;
	topLeft.y = top ;
	bottomRight.x = right ;
	bottomRight.y = bottom ;
	Draw(true) ;	// to erase the previous
	Draw() ;		// to draw again without handles
	while(Button()) {
		GetMouse(&currentPoint) ;
		if (currentPoint.x != lastPoint.x) {
			Draw() ;
			bottomRight.x += currentPoint.x - lastPoint.x ;
			left = (topLeft.x < bottomRight.x ? topLeft.x : bottomRight.x) ;
			right = (topLeft.x > bottomRight.x ? topLeft.x : bottomRight.x) ;
			SetRect(&mSelectionRect,left,top,right,bottom) ;
			Draw() ;
			lastPoint = currentPoint ;
		}
	}
	SetActiveHandles() ;
	DrawHandles() ;
}
void 	CSelection::Shift(Point &loc) {
	Point lastPoint,currentPoint,startPoint ;
	long dx,dy ;
	
	startPoint = loc ;
	lastPoint = startPoint ;
	Draw(true) ;	// to erase the previous
	Draw() ;		// to draw again without handles
	while(Button()) {
		GetMouse(&currentPoint) ;
		if (currentPoint.x != lastPoint.x || currentPoint.y != lastPoint.y) {
			Draw() ;	// Erase 
			dx = currentPoint.x - lastPoint.x ;
			dy = currentPoint.y - lastPoint.y ;
			OffsetRect(&mSelectionRect,dx,dy) ;
			Draw() ;	// Draw
			lastPoint = currentPoint ;
		}
	}
	SetActiveHandles() ;
	DrawHandles() ;
}

