// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>

#include "TImage.h"
#include "TAtom.h"
#include "TMatrix.h"
#include "TCrystal.h"
#include "TFile.h"

#include "funcdefs.h"	
#include "PhasegratingRecord.h"
#include "ProjCellRecord.h"
#include "displ.h"
#include "MslicRecord.h"
#include "storage.h"
#include "types.h"
#include "files.h"
#include "filenames.h"
#include "xtal.h"
#include "errors.h"
#include "globals.h"
#include "dfpdspl.h"
#include "symops.h"
#include "numbers.h"

#define	EOS	'\0'

extern	char 	AtomNames[MAXTYPES][8] ;
extern	long	noz[] ;
extern struct { short ipot,icmplx ; } flag ;

#include "TSetActiveElementsForm.h"
#include "TChemicalConstraintsDialog.h"
#include "TBondValenceDialog.h"

#include "TRefineParameterProgressPage.h"
extern	TRefineParameterProgressPage *gRefineParameterProgressPage ;
extern long gCentroSymmetric ;
extern short gSliceThreeD ;

#include "TStructureRefinementProgressPage.h"
TStructureRefinementProgressPage *gStructureRefinementProgressPage = nil ;

void	SetPage(TPage *) ;
void	PutMessage(char *) ;
void	ExitFromParameterRefinementProgress() ;

typedef struct	ElementRecord {
	short number ;
	short index[20] ;
} ElementRecord,*ElementPtr ;

void FrontWindowToNamedPictFile(char *PictFileName) ;
void Convert2DhklTo3Dhkl(float h2,float k2,float l2,float *h3,float *k3,float *l3) ;

#define	BlockMove(x,y,n)	memcpy((y),(x),(n))
#define	NewPtr(n)			malloc(n)
#define DisposePtr(ptr)		free(ptr)
typedef	void	*Ptr ;

void EncodeSymop(char *string,TMatrix *sym_op) ;
void RestoreParameterRefinementData() ;

void	ExitFromStructureAnnealing() ;
void	ExitFromStructureRefinementProgress() ;

extern 	void SetupCalculationParameters() ;
extern struct { long status ; } calculation ;

TMatrix	*gSymOps = nil ;	

Boolean	gModifyOneByOne = false ;
unsigned char gRefinementFlag ;

typedef struct {
	short	firstAtomNdx ;
	short	secondAtomNdx ;
	float	maxBondLength ;
	float	valenceSum ;
	float	bFactor ;
	float	weight ;
} BondStrength ;

typedef struct {
	short	numberOfBonds ;
	float	totalWeight ;
	BondStrength bondStrength[20] ;
} BondStrengthData ;

BondStrengthData gBondStrengthData 	= 	{ 0, 0.01 } ;	

extern	Boolean		gCalculateBondStrengths ;
extern	float		gBondValenceSumWeight ;
extern	Boolean		gUseSymmetryElements ;
extern 	FRect		gComparisonRect ;
extern 	short		gNumActiveElements ;
extern 	short 		gActiveElementNumber[20] ;	// Maximum 
extern	Boolean		gSaveLogFile ;
extern	float 		gAspectRatio ;
extern	Boolean 	gCalculateExactFTXCF ;
extern	Boolean 	gEstimateImageShift ;
extern	long 		gNumSymops ;
extern	RMatrix 	gTMatrix ; 
extern	char		micname[],gStructureName[] ;

void 	GetBondStrengthCalculationSettings() ;
float 	OldBondValenceEnergy() ;
float 	FourierSpaceXCoeffCompare(TImage *image1,TImage *image2,long &shiftN,long &shiftM);
void	CopyImage(TImage *image1, TImage *image2) ;
Boolean	NotInGroup(short index) ;

extern	long 	gShiftN,gShiftM ;
char	DateString[64] ;

void	WriteStatus(char *) ;

Boolean AdjustImageSize(TImage **image,long width,long height) ;
Boolean HasExperimentalImage() ;
Boolean HasStandardDeviationImage() ;
float	ChemicalEnergy() ;
float	RealSpaceXCoeffCompare(TImage *image1,TImage *image2) ;
extern 	TImage	*gExperimentalImage ;
extern 	TImage	*gStandardDeviationImage ;
void	DisplayImageInWindow(TImage *, TWindow *, Rect *,Boolean ) ;
float	CalculateMatchingEnergy(TImage *,TImage *,TImage *) ;
void	ApplyLens(float zphi[],float ihk[],float,float ,short ,short ,long ) ;
void 	SetActiveElements() ;
void 	SetActiveAreaAndAtoms() ;
void	SaveRefinedParameters() ;
void	CompareAndRefineParameters() ;
void	CompareAndRefineStructure() ;
Boolean GetFittingParameters() ;
void	ModifyParameter(long *groupIndex) ;
Boolean AcceptNewParameter(float *,TImage *,TImage *,TImage *) ;
void	TestingAnnealing() ;
Boolean	CalculateInitialPotential(float *p0) ;
void	ReplaceStructureWithNew(float *,float *,long ,float ,float ,float , float , float ) ;
Boolean	CalculateNewPotential(float *,float *,long ,float ,float ,float , float , float ) ;
void	ModifyStructure(long *,float *,float *,float *, float *, float *) ;

extern	float gImageComparisonWeight ;
extern	float gBondValenceEnergy ;
extern	float gLastBondValenceEnergy ;

float BondValenceEnergy() ;
float BondValenceSum(short ndx,short index) ;

Boolean	gUseSymmetryElements	= false ;
Boolean	gCalculateBondStrengths	= false ;
float	gBondValenceSumWeight = 1. ;
typedef struct {
	short numBasisAtoms ;
	TAtom *atom ;
} TBasis ;

TBasis gBasis = {0,nil} ;
TBasis sBasis = {0,nil} ;

#define min(x,y)	((x) < (y) ? (x) : (y))
extern char 	stext[] ;
extern struct 	{ float wavel ; short nslice ; float sigma ; } gMSRec ;
extern short 	gUseElScattFactors ;
extern char 	*chemsymb[] ;
extern long 	gSpacegroupNumber ;
extern Boolean	gIncludeImageComparison ;
extern Boolean	gIncludeBondDistances ;
extern Boolean	gIncludeBondValenceSum ;

extern long 	gCalculationIsChanged,gPGisChanged,gMSisChanged,gIMisChanged ;
extern long 	LAYERED ;

#define WPO		1
#define PHSGRT	2
#define MSLICE	4
#define IMAGE	8

// #define fabs(x) ((x) < 0. ? -(x) : (x))

float	CreateRandomDisplacement(float maxAmplitude) ;
float	CreateRandomValue(float maxValue) ;
float	ran0(long *idum) ;
Boolean NotEquivalentAtom(float x,float y,float z,long icount,long ndx) ;
void	ApplySymopsToBasis() ;
void	CalculateNewExitWaveFunction(float *potential) ;
void	CalculateNewImage(float *exitwavefunction,TImage *calculatedImage) ;
Boolean AcceptNewStructure(float *,TImage *,TImage *,TImage *) ;
float	RandomProbability() ;

#define kRefineX	0x01
#define kRefineY	0x02
#define kRefineZ	0x04
#define kRefineDW	0x08
#define kRefineOCC	0x10			

RGBColor RGBred 	= 	{65535,0,0} ;
RGBColor RGBgreen	=	{0,65535,0} ;
RGBColor RGBblue	=	{0,0,65535} ;


extern	float 	gChemicalEnergy ;
extern	float   gChangeInChemicalEnergy ;
extern	float	gMatchingEnergy ;

#ifndef __RefinementGroup
#define __RefinementGroup
typedef struct RefinementGroup {
	long numInGroup ;
	long *atomIndex ;
	unsigned char	*refine ;

} RefinementGroup ;

typedef struct RefinementList {
	long numGroups ;
	RefinementGroup *groupPtr ;
} RefinementList ;
#endif 

typedef enum {kCrossCorrelation,kChiSquared,kFTAmplCrossCorrelation } FittingMethod ;

extern	RefinementList gRefinementList ;
#ifdef __VMS
#pragma message disable typnolnkvarlnk
#endif
extern 	ElementPtr MyElements ;
#ifdef __VMS
#pragma message enable typnolnkvarlnk
#endif

typedef struct {
	float	maxAmplitude ;	// In Angstrom
	Boolean	useActiveRect ;
	FRect	activeRect ;
} RelaxationAtomRecord ;

typedef struct {
	float	start ;			// Starting Temperature
	float	end ;			// Ending Temperature
	float	current ;		// current temperature
	float	dTFraction ;	// Temperature decrement
	long		maxNumTriesPerTemperature ;
	long		maxNumGoodTries ;
} RelaxationTemperatureRecord ;

typedef struct {
	unsigned long	ID ;
	float	minValue ;
	float	maxValue ;
	float	startingValue[2] ;
	float	maxChangeInValue ;
	float	currentValue[3] ;	// We may need to store a magnitude and a direction or possible magnitude and components
} ParameterRecord ;

typedef struct {
	short	numFittingParameters ;
	ParameterRecord parameter[10] ;	// only space for 10 different parameters to fit
	RelaxationTemperatureRecord temperature ;
	float previousEnergy ;
	float thisEnergy ;
	float minEnergy ;
	long  numTriesThisTemperature ;
	FittingMethod fittingMethod ;
} ParameterFittingRecord ;

typedef struct {
	RelaxationAtomRecord		atom ;	// In Angstrom
	RelaxationTemperatureRecord	temperature ;
	float previousEnergy ;
	float thisEnergy ;
	float minEnergy ;
	long  numTriesThisTemperature ;
	FittingMethod fittingMethod ;
} RelaxationParamRecord ;

extern	RelaxationParamRecord gRelaxationParamRecord ;
extern	ParameterFittingRecord gParameterFittingRecord ;

typedef struct { 
	float ax[4*MAXTYPES] ;			/* Only allow up to MAXTYPES different atoms */
	float bx[4*MAXTYPES] ;			/* Only allow up to MAXTYPES different atoms */
	float cx[MAXTYPES] ; 
} SFParamFitRecord,*SFParamFitPtr ;

float *PP0=nil ;

typedef struct {
	short 	typeAtom1 ;
	short 	typeAtom2 ;
	float 	optimumDistance ;
	float	weight ;
	Boolean	use3Ddistance ;
	Boolean	isActive ;
} ChemicalBondRecord ;

typedef struct {
	short numberOfBonds ;
	ChemicalBondRecord bond[20] ;
} ChemicalBondData ;

ChemicalBondData gChemicalBondData = {0} ;
long CreateRandomAtomIndex(RefinementList *list) ;

extern	long	gIndex ;
float	gDx,gDy,gDz,gDdw,gDocc ;

void	LoadExperimentalImage() ;
void	CompareWithSimulation() ;
void	CompareAndRefine() ;
void	SimulateAnnealing() ;
void	GetChemicalBondData() ;
void	GetBondValenceParameters() ;

extern	char	gStructureName[256] ;

void	run_p_pg(),run_p_ms(),run_p_im() ;
float	AtomicFormFactor(float ,long noz[],long ,float *,SFParamFitPtr ) ;

void 	CrossTransferFunction(float h1,float k1,float h2,float k2,float u1sq,float u2sq,
		float cxa1,float cxa2,float cxd,float astsq,float bstsq,
		float abga,float *xcoeff) ;
float	CalculateInitialImageMatch(float *,TImage *,TImage *,TImage *) ;
float	RealSpaceXCoeffCompareFromArea(TImage *image1,TImage *image2,FRect *rect) ;
Boolean GetAnnealingParameters() ;
void 	DeleteRefinementList(RefinementList *list) ;
void	SaveRefinedStructureFile(char *file_name) ;
void	SaveRefinedStructure() ;

void	SaveBasis() ;
void	RestoreBasis() ;
void	DeleteSavedBasis() ;

static	float	gCalculatedEnergy ;
void SaveBasis(){
	// First make sure we have allocated space for the saved basis
	if (gBasis.numBasisAtoms != 0) {
		if (sBasis.numBasisAtoms == 0) {
			sBasis.atom = new TAtom[gBasis.numBasisAtoms] ;
			if (sBasis.atom != nil) {
				sBasis.numBasisAtoms = gBasis.numBasisAtoms ;
			}
		}
	}
	for(long i = 0;i<sBasis.numBasisAtoms;i++) {
		*(sBasis.atom+i) = *(gBasis.atom+i) ;
	}
}

void RestoreBasis() {

	for(long i = 0;i<sBasis.numBasisAtoms;i++) {
		*(gBasis.atom+i) = *(sBasis.atom+i) ;
	}
}

void DeleteSavedBasis() {
	if (sBasis.atom != nil) {
		delete[] sBasis.atom ;
		sBasis.numBasisAtoms = 0 ;
	}
}


void	RetrieveActiveElements(long arg1,long arg2) {
	TSetActiveElementsForm *elementForm = (TSetActiveElementsForm *)arg1 ;
	if (arg2 == 1) {
		elementForm->GetActiveElements() ;
	}
	delete elementForm ;
}

void 	SetActiveElements() {
	
	TSetActiveElementsForm *elementForm ;
	elementForm = new TSetActiveElementsForm(RetrieveActiveElements) ;
	if (elementForm != nil) {
		elementForm->Show() ;
	}

}


void DeleteRefinementList(RefinementList *list) {
	long i ;
	
	if (list->numGroups != 0) {
		for(i=0;i<list->numGroups;i++) {
			delete (list->groupPtr+i)->atomIndex ;
			delete (list->groupPtr+i)->refine ;
		}
		delete list->groupPtr ;
	}
	list->numGroups = 0 ; 
}

#include "TAtomRefinementPage.h"
TAtomRefinementPage *gAtomRefinementPage = nil ;
void 	SetActiveAreaAndAtoms() {
	if (gAtomRefinementPage == nil)
		gAtomRefinementPage = 
			new TAtomRefinementPage(gControlWindow,gProjCell) ;
	SetPage(gAtomRefinementPage) ;
	gAtomRefinementPage->Show() ;
}

void GetChemicalBondData() {

	TChemicalConstraintsDialog *theDialog = new TChemicalConstraintsDialog ;
	theDialog->SetFormData((void *)&gChemicalBondData) ;
	if (theDialog->DoModal()) {
		theDialog->GetFormData((void *)&gChemicalBondData) ;
	}
	delete theDialog ;
}


float CalculateMatchingEnergy(TImage *experimentalImage,TImage *errorImage,TImage *calculatedImage) {
	float	matchingEnergy ;
	FRect   activeRectangle ;
	
	if (gRelaxationParamRecord.atom.useActiveRect) {
		activeRectangle = gRelaxationParamRecord.atom.activeRect ;
	}
	else {
		activeRectangle.top = 0. ;
		activeRectangle.left = 0. ;
		activeRectangle.bottom = 1. ;
		activeRectangle.right = 1. ;
	}
	
	// Compare
	matchingEnergy = 0. ;
	if (gIncludeImageComparison) {
		switch(gRelaxationParamRecord.fittingMethod) {
			case kCrossCorrelation:
				matchingEnergy = 1.0 - RealSpaceXCoeffCompareFromArea(experimentalImage,
					calculatedImage,&activeRectangle) ;
				break ;
			case kChiSquared:
				matchingEnergy = ComputeImageChiSquare(experimentalImage,
								calculatedImage,errorImage,
								&activeRectangle,2*gRefinementList.numGroups) ;
				break ;
		}
	}
	
	gRelaxationParamRecord.thisEnergy = matchingEnergy ;
	
	if (gIncludeBondDistances && gChemicalBondData.numberOfBonds > 0) {
		gRelaxationParamRecord.thisEnergy += ChemicalEnergy() ;
	}
	
	if (gIncludeBondValenceSum && gBondStrengthData.numberOfBonds > 0) {
		gRelaxationParamRecord.thisEnergy += BondValenceEnergy() ;
	}
	return (gRelaxationParamRecord.thisEnergy) ;
}

Boolean AcceptNewStructure(float *potential,TImage *experimentalImage,
		TImage *errorImage,TImage *calculatedImage) {
	// We must do the multislice calculation and test the image with
	// the experimental image. If the new image is a better match, we
	// return true. Otherwise we test if we should acccept the new structure
	// and return true, otherwise false.
	// We also set the flag to see if we should terminate the entire search
	Boolean accept ;
	char	text[256] ;
	long	index ;
	float	*exitWaveFunction ;
	float	dE,probabilityOfTransition,randomProbability,matchingEnergy ;
	FRect   activeRectangle ;
	static	float previousMatchingEnergy = -1. ;
	static	float previousChemicalEnergy = -1. ;
	//static	long	count = 0 ;
	
	if (gIncludeImageComparison) {
		// Calculate the exit wavefunction
		CalculateNewExitWaveFunction(potential) ;
		if(gAbortStatus) {
			gAbortStatus = false ;
			return false ;
		}	
		
		// Calculate the Image
		exitWaveFunction = gWorkSpace ;
		CalculateNewImage(exitWaveFunction,calculatedImage) ;
		if(gAbortStatus) {
			gAbortStatus = false ;
			return false ;
		}	
	}
	
	accept = false ;
	
	if (gRelaxationParamRecord.atom.useActiveRect) {
		activeRectangle = gRelaxationParamRecord.atom.activeRect ;
	}
	else {
		activeRectangle.top = 0. ;
		activeRectangle.left = 0. ;
		activeRectangle.bottom = 1. ;
		activeRectangle.right = 1. ;
	}
	
	// Compare
	matchingEnergy = 0. ;
	if (gIncludeImageComparison) {
		switch(gRelaxationParamRecord.fittingMethod) {
			case kCrossCorrelation:
				matchingEnergy = 1.0 - 
					RealSpaceXCoeffCompareFromArea(experimentalImage,
					calculatedImage,&activeRectangle) ;
				dE = (gRelaxationParamRecord.thisEnergy - 
					gRelaxationParamRecord.previousEnergy) ;
				//if (count%10 == 0) 
				//{
				//fprintf(stderr,"\nE (%.2f, %.2f, %.2f, %.2f) = %f",
				//	activeRectangle.left,activeRectangle.top,
				//	activeRectangle.right,activeRectangle.bottom,
				//	1.0-matchingEnergy) ;
				//}
				//count++ ;
				break ;
			case kChiSquared:
				matchingEnergy = ComputeImageChiSquare(experimentalImage,
							calculatedImage,errorImage,
							&activeRectangle,2*gRefinementList.numGroups) ;
				break ;
		}
	}
	
	gRelaxationParamRecord.thisEnergy = matchingEnergy ;
	
	if (gIncludeBondDistances && gChemicalBondData.numberOfBonds > 0) {
		float	chemicalContribution = ChemicalEnergy() ;
		gRelaxationParamRecord.thisEnergy += chemicalContribution ;
	}
	
	if (gIncludeBondValenceSum && gBondStrengthData.numberOfBonds > 0) {
		float	valenceSumContribution = BondValenceEnergy() ;
		gRelaxationParamRecord.thisEnergy += valenceSumContribution ;
	}
	
	dE = (gRelaxationParamRecord.thisEnergy - gRelaxationParamRecord.previousEnergy) ;	
	if (dE < 0.) 
		accept = true ;
	else {
		probabilityOfTransition = exp(-dE/gRelaxationParamRecord.temperature.current) ;
		randomProbability = RandomProbability() ;
		if (randomProbability < probabilityOfTransition) {
			accept = true ;
		}
	}
	
	if (accept) {
		gRelaxationParamRecord.previousEnergy = gRelaxationParamRecord.thisEnergy ;
		gChemicalEnergy += gChangeInChemicalEnergy ;
		gMatchingEnergy = matchingEnergy ;
		previousMatchingEnergy = gMatchingEnergy ;
		previousChemicalEnergy = gChemicalEnergy ;
		gLastBondValenceEnergy = gBondValenceEnergy ;

		 // Take the position of the first atom in the group
		index = *(gRefinementList.groupPtr+gIndex)->atomIndex ;
		sprintf(stext,"T=%.5f ",gRelaxationParamRecord.temperature.current) ;
		if (gUseSymmetryElements) {
			// All the modifications have already been stored in gBasis. If the basis is not
			// accepted it gets restored to saved state upon return
			if ((gRefinementFlag & kRefineX) != 0) {	// Modified x
				sprintf(text,"x:%.3f ",(gBasis.atom+index)->xc) ;
				strcat(stext,text) ;
			}
			if ((gRefinementFlag & kRefineY) != 0) {	// Modified y
				sprintf(text,"y:%.3f ",(gBasis.atom+index)->yc) ;
				strcat(stext,text) ;
			}
			if ((gRefinementFlag & kRefineZ) != 0) {	// Modified z
				sprintf(text,"y:%.3f ",(gBasis.atom+index)->zc) ;
				strcat(stext,text) ;
			}
			if ((gRefinementFlag & kRefineDW) != 0) {	// Modified dw
				sprintf(text,"dw:%.3f ",(gBasis.atom+index)->dw) ;
				strcat(stext,text) ;
			}
			if ((gRefinementFlag & kRefineOCC) != 0) {	// Modified occ
				sprintf(text,"oc:%.3f ",(gBasis.atom+index)->occ) ;
				strcat(stext,text) ;
			}
		}
		else {
			if ((gRefinementFlag & kRefineX) != 0) {	// Modified x
				sprintf(text,"x:%.3f ",*(gProjCell->x+index)+gDx) ;
				strcat(stext,text) ;
			}
			if ((gRefinementFlag & kRefineY) != 0) {	// Modified y
				sprintf(text,"y:%.3f ",*(gProjCell->y+index)+gDy) ;
				strcat(stext,text) ;
			}
			if ((gRefinementFlag & kRefineZ) != 0) {	// Modified z
				sprintf(text,"z:%.3f ",*(gProjCell->y+index)+gDz) ;
				strcat(stext,text) ;
			}
			if ((gRefinementFlag & kRefineDW) != 0) {	// Modified dw
				sprintf(text,"dw:%.3f ",*(gProjCell->dw+index)+gDdw) ;
				strcat(stext,text) ;
			}
			if ((gRefinementFlag & kRefineOCC) != 0) {	// Modified occ
				sprintf(text,"oc:%.3f ",*(gProjCell->occ+index)+gDocc) ;
				strcat(stext,text) ;
			}
		}
		sprintf(text,",E=%.4f",gRelaxationParamRecord.thisEnergy) ;
		strcat(stext,text) ;
		//WriteStatus(stext) ;
	}
	else {
		// restore the energy to the previous accepted energy
		gRelaxationParamRecord.thisEnergy = gRelaxationParamRecord.previousEnergy ;
		if (previousMatchingEnergy != -1.) {
			gMatchingEnergy = previousMatchingEnergy ;
		}
		if (previousChemicalEnergy != -1.) {
			gChemicalEnergy = previousChemicalEnergy ;
		}
		if (gLastBondValenceEnergy != -1.) {
			gBondValenceEnergy = gLastBondValenceEnergy ;
		}
	}
	return accept ;
}


void ReplaceStructureWithNew(float *pOld,float *pNew,long groupIndex,float dx,
		float dy,float dz, float ddw, float docc) {
	long jmax,i,index ;
	unsigned char refinementFlag ;
	
	
	if (gIncludeImageComparison) {
		jmax = gPhsgrtRec->nbeam*gPhsgrtRec->mbeam ;
		BlockMove((const void *)pNew,(void *)pOld,jmax*sizeof(float)) ;
	}
	if (!gUseSymmetryElements) {	// We need to update the gProjCell structure
		for( i=0; i < (gRefinementList.groupPtr+groupIndex)->numInGroup ; i++) {
			index = *((gRefinementList.groupPtr+groupIndex)->atomIndex+i) ;
			refinementFlag = *((gRefinementList.groupPtr+groupIndex)->refine+i) ;
			
			//fprintf(stderr,"\nAtom[%d] (%.3f,%.3f,%.3f) -> ",index,
			//	*(gProjCell->x+index),*(gProjCell->y+index),*(gProjCell->z+index)) ;
			if ((refinementFlag & kRefineX) != 0) {
				*(gProjCell->x+index) += dx ;
			}
			if ((refinementFlag & kRefineY) != 0) {
				*(gProjCell->y+index) += dy ;
			}
			if ((refinementFlag & kRefineZ) != 0) {
				*(gProjCell->z+index) += dz ;
			}
			if ((refinementFlag & kRefineDW) != 0) {
				*(gProjCell->dw+index) += ddw ;
			}
			if ((refinementFlag & kRefineOCC) != 0) {
				*(gProjCell->occ+index) += docc ;
			}

			
			if (*(gProjCell->x+index) < 0. ) {
				*(gProjCell->x+index) += 1.0 ;
			}
			if (*(gProjCell->x+index) > 1. ) {
				*(gProjCell->x+index) -= 1.0 ;
			}
			if (*(gProjCell->y+index) < 0. ) {
				*(gProjCell->y+index) += 1.0 ;
			}
			if (*(gProjCell->y+index) > 1. ) {
				*(gProjCell->y+index) -= 1.0 ;
			}
			if (*(gProjCell->z+index) < 0. ) {
				*(gProjCell->z+index) += 1.0 ;
			}
			if (*(gProjCell->z+index) > 1. ) {
				*(gProjCell->z+index) -= 1.0 ;
			}
			if (*(gProjCell->dw+index) < 0. ) {
				*(gProjCell->dw+index) = 0. ;
			}
			if (*(gProjCell->occ+index) < 0. ) {
				*(gProjCell->occ+index) = 0. ;
			}
			//fprintf(stderr,"(%.3f,%.3f,%.3f)",
			//	*(gProjCell->x+index),*(gProjCell->y+index),*(gProjCell->z+index)) ;
		}
	} 
	else {
		// The gProjCell structure has been replaced already
		// We just have to make sure that it gets restored if we didn't accept the 
		// new structure and that is controlled from SimulateAnnealing
	}
}

void ModifyStructure(long *groupIndex,float *dx,float *dy,float *dz, float *ddw, float *docc) {

	// We need a random number generator for determining which atom
	// should be moved and also the amount of movement
	// The possible atom to be moved can be one of all or be in a specified
	// selection of atoms
	long i,encountered ;
	float ddx,ddy,displacement,angle ;
	long index,indexlist[5],numVariables,ndx ;
	float	amplitude,factor  ;
	unsigned char refinementFlag ;
	
	// Now decide what kind of change we need
	// This depends on what it is that we are refining
	// If we are refining positions x,y or maybe also dw and occ
	// Find a random index for the atom to move

	if (gRefinementList.numGroups == 1)
		index = 0 ;
	else
		index = CreateRandomAtomIndex(&gRefinementList) ;	
	// We need to change the logic of how this is used


	*dx = 0. ;
	*dy = 0. ;
	*dz = 0. ;
	*ddw = 0. ;
	*docc = 0. ;
	
	factor = 0.2 + 0.8*(gRelaxationParamRecord.temperature.current - 
			gRelaxationParamRecord.temperature.end) / 
			(gRelaxationParamRecord.temperature.start-gRelaxationParamRecord.temperature.end) ;
	// At beginning temperature = AMP, at end temp = 0.2*AMP
	amplitude = gRelaxationParamRecord.atom.maxAmplitude*factor ;	
	
	if (gModifyOneByOne) {
		// Find out which parameter we should vary. I don't know if 
		// it is better to move all at once or
		// to try each one individually. I think there are times it 
		// works better to try each one individually
		// in case one of the chosen optimizing parameters should not be altered
		refinementFlag = *((gRefinementList.groupPtr+index)->refine)  ;
		numVariables = 0 ;
		for(i=0;i<5;i++)
			indexlist[i] = 0 ;
		if ((refinementFlag & kRefineX) != 0) {
			indexlist[0] = 1 ;
			numVariables++ ;
		}
		if ((refinementFlag & kRefineY) != 0) {
			indexlist[1] = 1 ;
			numVariables++ ;
		}
		if ((refinementFlag & kRefineZ) != 0) {
			indexlist[2] = 1 ;
			numVariables++ ;
		}
		if ((refinementFlag & kRefineDW) != 0) {
			indexlist[3] = 1 ;
			numVariables++ ;
		}
		if ((refinementFlag & kRefineOCC) != 0) {
			indexlist[4] = 1 ;
			numVariables++ ;
		}
		ndx = (long) CreateRandomValue(numVariables) ;
		if (ndx == numVariables) ndx-- ;
		for(i=0,encountered = 0;i<5;i++) {
			if (indexlist[i] != 0) {
				 if (ndx == encountered) {
				 	switch(i) {
				 		case 0:	// X
							displacement = CreateRandomValue(amplitude) ;
							*dx = displacement/gProjCell->a ;
							break ;
				 		case 1:	// Y
							displacement = CreateRandomValue(amplitude) ;
							*dy = displacement/gProjCell->b ;
							break ;
				 		case 2:	// Z
							displacement = CreateRandomValue(amplitude) ;
							*dz = displacement/gProjCell->c ;
							break ;
						case 3: // DW
							*ddw = 0.2 - CreateRandomValue(0.4) ;	// Value between -0.2 - 0.2
							break ;
						case 4: // OCC
							*docc = 0.2 - CreateRandomValue(0.4) ;	// Value between -0.2 - 0.2
							break ;
				 	}
				 	break ;
				 }
				 encountered++ ;
			}
		}

	}
	else {		
		// Create a displacement for this group of atoms
		// Generate the movement
		displacement = CreateRandomValue(amplitude) ;
		angle = CreateRandomValue(4*asin(1.0)) ;
		float	theta = CreateRandomValue(asin(1.0)) ;

		refinementFlag = *((gRefinementList.groupPtr+index)->refine)  ;
		
		//Scale in unit cell parameters
		if ((refinementFlag & kRefineZ) != 0) {
			*dz = displacement*cos(theta)/gProjCell->c ;
			displacement *= sin(theta) ;
		}

		if ((refinementFlag & kRefineX) != 0) {
			ddx = displacement*cos(angle)/gProjCell->a ;
			*dx = ddx ;
		}
		if ((refinementFlag & kRefineY) != 0) {
			ddy = displacement*sin(angle)/gProjCell->b ;
			*dy = ddy ;
		}
		if ((refinementFlag & kRefineDW) != 0) 
			*ddw = 0.2 - CreateRandomValue(0.4) ;	// Value  -0.2 - 0.2
		if ((refinementFlag & kRefineOCC) != 0) 
			*docc = 0.2 - CreateRandomValue(0.4) ;	// Value  -0.2 - 0.2
	}
	*groupIndex = index ;
}

void ModifyParameter(long *groupIndex) {

	float 	ddx,ddy,change,angle,x,y,value ;
	long 	index,n ;
	float	amplitude  ;
	double 	pi ;
	
	pi = 2.0*asin(1.0) ;
	// Find a random index for the parameter to change
	amplitude = gParameterFittingRecord.numFittingParameters ;  // 0 - gParameterFittingRecord.numFittingParameters
	index = (long)CreateRandomValue(amplitude) ;
	if (index == gParameterFittingRecord.numFittingParameters) index = index - 1 ;

	switch (gParameterFittingRecord.parameter[index].ID) {
		case 'THIC':
			// We should only go in increments of unit cell dimensions
			n = near(CreateRandomDisplacement(gParameterFittingRecord.parameter[index].maxChangeInValue)) ;
			if (n == 0) {
				*groupIndex = -1 ;
				return ;
			}
			change = n*gPhsgrtRec->slcth ;
			//change = CreateRandomDisplacement(gParameterFittingRecord.parameter[index].maxChangeInValue) ;
			gMsliceRec.thick += change ;
			if (gMsliceRec.thick <= gParameterFittingRecord.parameter[index].minValue) 
				gMsliceRec.thick = gPhsgrtRec->slcth ; ;
			if (gMsliceRec.thick > gParameterFittingRecord.parameter[index].maxValue) 
				gMsliceRec.thick -= change ;
			gMsliceRec.t1 = gMsliceRec.thick ;
			gParameterFittingRecord.parameter[index].currentValue[0] = gMsliceRec.thick ;
			break ;
		case 'DEFO':
			change = CreateRandomDisplacement(gParameterFittingRecord.parameter[index].maxChangeInValue) ;
			gDisplayRec.dfocus += change ;
			if (gDisplayRec.dfocus < gParameterFittingRecord.parameter[index].minValue) 
				gDisplayRec.dfocus -= change ;
			if (gDisplayRec.dfocus > gParameterFittingRecord.parameter[index].maxValue) 
				gDisplayRec.dfocus -= change ;
			gDisplayRec.f1 = gDisplayRec.dfocus ;
			gParameterFittingRecord.parameter[index].currentValue[0] = gDisplayRec.dfocus ;
			break ;
		case 'CS  ':
			change = CreateRandomDisplacement(gParameterFittingRecord.parameter[index].maxChangeInValue) ;
			gDisplayRec.cs += change ;
			if (gDisplayRec.cs < gParameterFittingRecord.parameter[index].minValue) 
				gDisplayRec.cs -= change ;
			if (gDisplayRec.cs > gParameterFittingRecord.parameter[index].maxValue) 
				gDisplayRec.cs -= change ;
			gParameterFittingRecord.parameter[index].currentValue[0] = gDisplayRec.cs ;
			break ;
		case 'DIV ':
			change = CreateRandomDisplacement(gParameterFittingRecord.parameter[index].maxChangeInValue) ;
			gDisplayRec.div += change ;
			if (gDisplayRec.div < gParameterFittingRecord.parameter[index].minValue) 
				gDisplayRec.div -= change ;
			if (gDisplayRec.div > gParameterFittingRecord.parameter[index].maxValue) 
				gDisplayRec.div -= change ;
			gParameterFittingRecord.parameter[index].currentValue[0] = gDisplayRec.div ;
			break ;
		case 'DEL ':
			change = CreateRandomDisplacement(gParameterFittingRecord.parameter[index].maxChangeInValue) ;
			gDisplayRec.del += change ;
			if (gDisplayRec.del < gParameterFittingRecord.parameter[index].minValue) 
				gDisplayRec.del -= change ;
			if (gDisplayRec.del > gParameterFittingRecord.parameter[index].maxValue) 
				gDisplayRec.del -= change ;
			gParameterFittingRecord.parameter[index].currentValue[0] = gDisplayRec.del ;
			break ;
		case 'CTLT':			
			double tiltx,tilty ;
			tiltx = gMsliceRec.kxin/gProjCell->a ;
			tilty = -gMsliceRec.kxin*gProjCell->ast*gProjCell->csga + gMsliceRec.kyin*gProjCell->bst ;
			//tiltx = gMsliceRec.kxin*gProjCell->ast*gProjCell->siga ;
			//tilty = gMsliceRec.kyin*gProjCell->bst - gMsliceRec.kxin*gProjCell->ast*gProjCell->csga/gProjCell->siga ;
			
			// Make a change in the tiltangle which is a change in the vector positioning the Laue circle
			amplitude = 0.001*CreateRandomValue(gParameterFittingRecord.parameter[index].maxChangeInValue)/gMSRec.wavel ;
			angle = CreateRandomValue(2.0*pi) ;
			ddx = amplitude*cos(angle) ;
			ddy = amplitude*sin(angle) ;
			tiltx += ddx ;
			tilty += ddy ;
			
			// Now calculate the new values for the tilt
			gMsliceRec.kxin = tiltx*gProjCell->a ;
			gMsliceRec.kyin = (tilty + tiltx*gProjCell->a*gProjCell->ast*gProjCell->csga) / gProjCell->bst ;
			gParameterFittingRecord.parameter[index].currentValue[0] = 1000.0*asin(sqrt(tiltx*tiltx+tilty*tilty)*gMSRec.wavel) ; // mrad
			gParameterFittingRecord.parameter[index].currentValue[1] = gMsliceRec.kxin ;
			gParameterFittingRecord.parameter[index].currentValue[2] = gMsliceRec.kyin ;
			break ;
		case 'ASTI':
			x = gDisplayRec.TwoFoldAstigMag*cos(gDisplayRec.TwoFoldAstigAngle*pi/180.) ;
			y = gDisplayRec.TwoFoldAstigMag*sin(gDisplayRec.TwoFoldAstigAngle*pi/180.) ;
			amplitude = CreateRandomValue(gParameterFittingRecord.parameter[index].maxChangeInValue) ;
			angle = CreateRandomValue(2.0*pi) ;
			ddx = amplitude*cos(angle) ;
			ddy = amplitude*sin(angle) ;
			x += ddx ;
			y += ddy ;
			value = sqrt(x*x+y*y) ;
			if (value > gParameterFittingRecord.parameter[index].maxValue) {
				x -= 2*ddx ;
				y -= 2*ddy ;
				value = sqrt(x*x+y*y) ;
			}
			gDisplayRec.TwoFoldAstigMag = value ;
			if (x != 0.) {
				angle = atan(y/x) ;
			}
			else {
				if (y != 0.)
					angle = pi/2 ;
				else
					angle = 0. ;
			}
			if (x < 0.)
				angle += pi ;
			gDisplayRec.TwoFoldAstigAngle = angle*180./pi ;
			gParameterFittingRecord.parameter[index].currentValue[0] = gDisplayRec.TwoFoldAstigMag ;
			gParameterFittingRecord.parameter[index].currentValue[1] = gDisplayRec.TwoFoldAstigAngle ;
			break ;
		case 'AST3':
			x = gDisplayRec.ThreeFoldAstigMag*cos(gDisplayRec.ThreeFoldAstigAngle*pi/180.) ;
			y = gDisplayRec.ThreeFoldAstigMag*sin(gDisplayRec.ThreeFoldAstigAngle*pi/180.) ;
			amplitude = CreateRandomValue(gParameterFittingRecord.parameter[index].maxChangeInValue) ;
			angle = CreateRandomValue(2.0*pi) ;
			ddx = amplitude*cos(angle) ;
			ddy = amplitude*sin(angle) ;
			x += ddx ;
			y += ddy ;
			value = sqrt(x*x+y*y) ;
			if (value > gParameterFittingRecord.parameter[index].maxValue) {
				x -= 2*ddx ;
				y -= 2*ddy ;
				value = sqrt(x*x+y*y) ;
			}
			gDisplayRec.ThreeFoldAstigMag = value ;
			if (x != 0.) {
				angle = atan(y/x) ;
			}
			else {
				if (y != 0.)
					angle = pi/2 ;
				else
					angle = 0. ;
			}
			if (x < 0.)
				angle += pi ;
			gDisplayRec.ThreeFoldAstigAngle = angle*180./pi ;
			gParameterFittingRecord.parameter[index].currentValue[0] = gDisplayRec.ThreeFoldAstigMag ;
			gParameterFittingRecord.parameter[index].currentValue[1] = gDisplayRec.ThreeFoldAstigAngle ;
			break ;
	}
	
	*groupIndex = index ;
}

float	CalculateInitialImageMatch(float *exitWaveFunction,TImage *experimentalImage,TImage *errorImage,TImage *calculatedImage) {
	// We must do the multislice calculation and test the image with
	// the experimental image. 
	float	matchingEnergy ;
	FRect   activeRectangle ;
	long	shiftN,shiftM ;
	TImage	*FTcalculatedImage ;
	
	
	// Calculate the Image
	CalculateNewImage(exitWaveFunction,calculatedImage) ;
	
	// We should test the part of the image that is the active comparson rectangle
	// at the moment
	activeRectangle = gComparisonRect ;
	
	// Compare
	switch(gParameterFittingRecord.fittingMethod) {
		case kCrossCorrelation:
			matchingEnergy = 1.0 - RealSpaceXCoeffCompareFromArea(experimentalImage,
				calculatedImage,&activeRectangle) ;
			break ;
		case kFTAmplCrossCorrelation:
			FTcalculatedImage = CopyImage(calculatedImage) ;
			FourierTransformImage(FTcalculatedImage) ;
			matchingEnergy = 1.0 - FourierSpaceXCoeffCompare(experimentalImage,
				FTcalculatedImage,shiftN,shiftM) ;
			break ; 
		case kChiSquared:
			matchingEnergy = ComputeImageChiSquare(experimentalImage,
				calculatedImage,errorImage,
				&activeRectangle,gParameterFittingRecord.numFittingParameters) ;
			break ;
	}
	return matchingEnergy ;
}

Boolean AcceptNewParameter(float *exitWaveFunction,TImage *experimentalImage,TImage *errorImage,TImage *calculatedImage) {
	// We must do the multislice calculation and test the image with
	// the experimental image. If the new image is a better match, we
	// return true. Otherwise we test if we should acccept the new structure
	// and return true, otherwise false.
	// We also set the flag to see if we should terminate the entire search
	Boolean accept ;
	float	dE,probabilityOfTransition,randomProbability,matchingEnergy ;
	FRect   activeRectangle ;
	long	shiftN,shiftM ;
	TImage	*FTcalculatedImage ;
	
	
	// Calculate the Image
	CalculateNewImage(exitWaveFunction,calculatedImage) ;
	if(gAbortStatus) {
		return false ;
	}	
	
	accept = false ;
	
	// We should test the part of the imag that is the active comparson rectangle
	// at the moment
	activeRectangle = gComparisonRect ;

	
	// Compare
	switch(gParameterFittingRecord.fittingMethod) {
		case kCrossCorrelation:
			matchingEnergy = 1.0 - RealSpaceXCoeffCompareFromArea(experimentalImage,
				calculatedImage,&activeRectangle) ;
			gParameterFittingRecord.thisEnergy = matchingEnergy ;
			dE = gParameterFittingRecord.thisEnergy - 
				gParameterFittingRecord.previousEnergy ;
			break ;
		case kFTAmplCrossCorrelation:
			FTcalculatedImage = CopyImage(calculatedImage) ;
			FourierTransformImage(FTcalculatedImage) ;
			matchingEnergy = 1.0 - FourierSpaceXCoeffCompare(experimentalImage,
				FTcalculatedImage,shiftN,shiftM) ;
			gParameterFittingRecord.thisEnergy = matchingEnergy ;
			dE = gParameterFittingRecord.thisEnergy - 
				gParameterFittingRecord.previousEnergy ;
			DeleteImage(FTcalculatedImage) ;
			gShiftN = shiftN ;
			gShiftM = shiftM ;
			break ; 
		case kChiSquared:
			matchingEnergy = ComputeImageChiSquare(experimentalImage,
				calculatedImage,errorImage,
				&activeRectangle,gParameterFittingRecord.numFittingParameters) ;
			gParameterFittingRecord.thisEnergy = matchingEnergy;
			dE = gParameterFittingRecord.thisEnergy - 
				gParameterFittingRecord.previousEnergy ;
			break ;
	}
	
	if (dE < 0.) 
		accept = true ;
	else {
		probabilityOfTransition = exp(-dE/gParameterFittingRecord.temperature.current) ;
		randomProbability = RandomProbability() ;
		if (randomProbability < probabilityOfTransition) {
			accept = true ;
		}
	}
	gCalculatedEnergy = gParameterFittingRecord.thisEnergy ;
	if (accept) {
		gParameterFittingRecord.previousEnergy = gParameterFittingRecord.thisEnergy ;
		gMatchingEnergy = matchingEnergy ;
		sprintf(stext,"T=%.5f ,E=%.4f",
			gParameterFittingRecord.temperature.current,
			gParameterFittingRecord.thisEnergy) ;
		WriteStatus(stext) ;
	}
	else {
		gParameterFittingRecord.thisEnergy = gParameterFittingRecord.previousEnergy ;
	}
	return accept ;
}

void RestoreParameterRefinementData() {
	switch (gParameterFittingRecord.parameter[gIndex].ID) {
		case 'THIC':
			gParameterFittingRecord.parameter[gIndex].currentValue[0] = 
								gMsliceRec.thick ;
			break ;
		case 'DEFO':
			gParameterFittingRecord.parameter[gIndex].currentValue[0] = 
								gDisplayRec.dfocus ;
			break ;
		case 'CS  ':
			gParameterFittingRecord.parameter[gIndex].currentValue[0] = 
								gDisplayRec.cs ;
			break ;
		case 'DIV ':
			gParameterFittingRecord.parameter[gIndex].currentValue[0] = 
								gDisplayRec.div ;
			break ;
		case 'DEL ':
			gParameterFittingRecord.parameter[gIndex].currentValue[0] = 
								gDisplayRec.del ;
			break ;
		case 'CTLT':			
			double tiltx,tilty ;
			tiltx = gMsliceRec.kxin/gProjCell->a ;
			tilty = -gMsliceRec.kxin*gProjCell->ast*gProjCell->csga + 
						gMsliceRec.kyin*gProjCell->bst ;
			gParameterFittingRecord.parameter[gIndex].currentValue[0] = 
						1000.0*asin(sqrt(tiltx*tiltx+tilty*tilty)*gMSRec.wavel) ;
			gParameterFittingRecord.parameter[gIndex].currentValue[1] = 
						gMsliceRec.kxin ;
			gParameterFittingRecord.parameter[gIndex].currentValue[2] = 
						gMsliceRec.kyin ;
			break ;
		case 'ASTI':
			gParameterFittingRecord.parameter[gIndex].currentValue[0] = 
								gDisplayRec.TwoFoldAstigMag ;
			gParameterFittingRecord.parameter[gIndex].currentValue[1] = 
								gDisplayRec.TwoFoldAstigAngle ;
			break ;
		case 'AST3':
			gParameterFittingRecord.parameter[gIndex].currentValue[0] = 
					gDisplayRec.ThreeFoldAstigMag ;
			gParameterFittingRecord.parameter[gIndex].currentValue[1] = 
								gDisplayRec.ThreeFoldAstigAngle ;
			break ;
	}
}

void CompareAndRefineParameters() {
	long jmax,groupIndex ;
	long	numSuccesfulTries ;
	char filename[32],buffer[256],*cp,text[256],PictFileName[256] ;
	time_t	secs ;
	long	maxTriesPerTemp,maxSuccessfulTries,transitionIndex,numAttempt,numDisplayIndex ;
	long	ierr ;
	float	dx,dy, maxValueOfEnergy, minValueOfEnergy, scaleFactor ;
	short	ix,iy ;
	TImage	*experimentalImage,*calculatedImage,*sigmaImage,
		*FTexperimentalImage,*lastAcceptedImage ;
	time_t 	lastOutputTime,currentTime,startTime,timeBetweenOutput=1 ;
	short	count = 0 ;
	float 	*MNew,*MOld,*exitWaveFunction ;
	FILE	*fp ;
	FRect	activeRectangle ;	
	short 	itemNumber,itemType ;
	Rect	itemRect,evolutionRect ;
	Handle	itemHandle ;
	DisplRecord	oldDisplRec ;
	MslicRecord	oldMslicRec ;
	Boolean	newExitWavefunction,acceptedValue ;
	float tilt,tiltangle,tiltx,tilty,theta,pi,h3,k3,l3 ;
	unsigned char date[256],timeString[256] ;
	

	float ewavel,evolt ;
	
	// Calculate the wavelength in centimeters if we aren't yet set
	evolt = gMsliceRec.volt ;
	ewavel = 1.2263652e-7/sqrt((evolt+9.785025e-4*evolt*evolt)*1000.) ;
	gMSRec.sigma = 2.08886e+6*ewavel*sqrt(1.0+5.886579e-20/(ewavel*ewavel)) ;
	gMSRec.wavel = 1.0e+8*ewavel ;  // Wavelength in angstroms
	
	setup_trnsf() ;
	SetPhasegratingParameters(AtomNames,noz) ;

	jmax = gPhsgrtRec->nbeam*gPhsgrtRec->mbeam ;

	// This is the aspect ratio of the height to the width for the
	// experimental image
	dx = gExperimentalImage->mWidth ;
	dy = gExperimentalImage->mHeight ;
	gAspectRatio = dy/dx ;

	experimentalImage = CopyImage(gExperimentalImage) ;
	// Instead of flipping the experimental image, we will flip the calculated images
	
	// Let us see if the image dimensions are different from the simulation sampling.
	// If one of the dimensions is different, adjust to fit the calculated image
	if (experimentalImage->mWidth != gPhsgrtRec->nbeam || 
			experimentalImage->mHeight != gPhsgrtRec->mbeam) {
		if (AdjustImageSize(&experimentalImage, gPhsgrtRec->nbeam,gPhsgrtRec->mbeam))
			experimentalImage->mSimulationSampling = true ;
	}
	
	sigmaImage = nil ;
	if (HasStandardDeviationImage()) {
		sigmaImage = CopyImage(gStandardDeviationImage) ;
		if (sigmaImage->mWidth != gPhsgrtRec->nbeam || 
				sigmaImage->mHeight != gPhsgrtRec->mbeam) {
			if (AdjustImageSize(&sigmaImage, gPhsgrtRec->nbeam,gPhsgrtRec->mbeam))
				sigmaImage->mSimulationSampling = true ;
		}
	}
	
	if (PP0 != nil) 
		free(PP0) ;
	PP0 = (float *)malloc(2*jmax*sizeof(float)) ;
	if (PP0 == nil) {
		PutMessage("Unable to allocate space for the initial potential") ;
		return ;
	}

	pi = 2.0*asin(1.0) ;
		
	// Set the parameters according to the starting values given by the user
  	for(long i=0;i<gParameterFittingRecord.numFittingParameters;i++) {
  		switch(gParameterFittingRecord.parameter[i].ID) {
  			case 'THIC':
				gMsliceRec.thick = gMsliceRec.t1 = gMsliceRec.t2 = 
					gParameterFittingRecord.parameter[i].startingValue[0];
  				break ;
  			case 'DEFO':
 				gDisplayRec.dfocus = gDisplayRec.f1 = gDisplayRec.f2 = 
					gParameterFittingRecord.parameter[i].startingValue[0];
  				break ;
  			case 'CS  ':
  				gDisplayRec.cs = gParameterFittingRecord.parameter[i].startingValue[0];
  				break ;
  			case 'DIV ':
  				gDisplayRec.div = 
					gParameterFittingRecord.parameter[i].startingValue[0];
  				break ;
  			case 'DEL ':
  				gDisplayRec.del = gDisplayRec.f1 = gDisplayRec.f2 = 
					gParameterFittingRecord.parameter[i].startingValue[0];
  				break ;
  			case 'CTLT':
				tiltangle = 
					0.001*gParameterFittingRecord.parameter[i].startingValue[0] ;
				
				// Now calculate the new values for the tilt
				tilt = sin(tiltangle)/gMSRec.wavel ;
				theta = gParameterFittingRecord.parameter[i].startingValue[1]*pi/180. ;
				tiltx = tilt*cos(theta) ;
				tilty = tilt*sin(theta) ;
				gMsliceRec.kxin = tiltx*gProjCell->a ;
				gMsliceRec.kyin = (tilty + tiltx*gProjCell->a*gProjCell->ast*
					gProjCell->csga) / gProjCell->bst ;
  				break ;
  			case 'ASTI':
				gDisplayRec.TwoFoldAstigMag = 
					gParameterFittingRecord.parameter[i].startingValue[0] ;
				gDisplayRec.TwoFoldAstigAngle = 
					gParameterFittingRecord.parameter[i].startingValue[1] ;
  				break ;
  			case 'AST3':
				gDisplayRec.ThreeFoldAstigMag = 
					gParameterFittingRecord.parameter[i].startingValue[0] ;
				gDisplayRec.ThreeFoldAstigAngle = 
					gParameterFittingRecord.parameter[i].startingValue[1] ;
  				break ;
  		}
	}
	
	MOld = (float *)malloc(2*jmax*sizeof(float)) ;
	if (MOld == nil) {
		PutMessage("Unable to allocate space for the initial exit wavefunction") ;
		return ;
	}

	MNew = (float *)malloc(2*jmax*sizeof(float)) ;
	if (MNew == nil) {
		PutMessage("Unable to allocate space for the new exit wavefunction") ;
		return ;
	}
	
	calculatedImage = CreateNewImage(experimentalImage->mWidth,
		experimentalImage->mHeight) ;
	if (calculatedImage == nil) {
		PutMessage("Unable to allocate space for the calculated image") ;
		return ;
	}
	lastAcceptedImage = CreateNewImage(experimentalImage->mWidth,
		experimentalImage->mHeight) ;
	if (lastAcceptedImage == nil) {
		PutMessage("Unable to allocate space for the last accepted image") ;
		return ;
	}
	SetImageName(lastAcceptedImage,"Final") ;

	activeRectangle.top = 0. ;
	activeRectangle.left = 0. ;
	activeRectangle.bottom = 1. ;
	activeRectangle.right = 1. ;
	
	maxTriesPerTemp = 2*gParameterFittingRecord.numFittingParameters*
		gParameterFittingRecord.temperature.maxNumTriesPerTemperature ;
	maxSuccessfulTries = 2*gParameterFittingRecord.numFittingParameters*
		gParameterFittingRecord.temperature.maxNumGoodTries ;

	groupIndex = 0 ;
	dx = 0. ;
	dy = 0. ;

	switch(gParameterFittingRecord.fittingMethod) {
		case kCrossCorrelation:
			gParameterFittingRecord.minEnergy = 0.0001 ; // 0.001 ;
			break ;
		case kFTAmplCrossCorrelation:
			gParameterFittingRecord.minEnergy = 0.0001 ; // 0.001 ;
			break ;
		case kChiSquared:
			if (HasStandardDeviationImage()) {			
				gParameterFittingRecord.minEnergy = 1.0 ;
			}
			else {	// probably testing, so set the min to 0.1
				gParameterFittingRecord.minEnergy = 0.1 ;
			}
			break ;
	}
	
	gParameterFittingRecord.thisEnergy = gParameterFittingRecord.minEnergy + 100. ;
	gParameterFittingRecord.previousEnergy = 10000000. ; // so to set it high
	gMatchingEnergy = gParameterFittingRecord.previousEnergy ;	
	

	if (!CalculateInitialPotential(PP0)) {
		PutMessage("Error calculating initial projected potential") ;
		return ;
	}

	transitionIndex = 0 ;	
	gChemicalEnergy = -1. ; // At start of a run
	groupIndex = 0 ;
		
	Boolean sCalculateExactFTXCF = gCalculateExactFTXCF ;
	Boolean sEstimateImageShift = gEstimateImageShift ;
	gCalculateExactFTXCF = false ;
	gEstimateImageShift = true ;

	if (gParameterFittingRecord.fittingMethod == kFTAmplCrossCorrelation) {
		FTexperimentalImage = CopyImage(experimentalImage) ;
		FourierTransformImage(FTexperimentalImage) ;
	}
	else {
		FTexperimentalImage = experimentalImage ; 	// Set them to be the same
	}

	// Calculate the exit wavefunction from the potential
	CalculateNewExitWaveFunction(PP0) ;

	exitWaveFunction = gWorkSpace ;

	// Store the exit wavefunction
	jmax = gPhsgrtRec->nbeam*gPhsgrtRec->mbeam ;
	BlockMove((const void *)exitWaveFunction,(void *)MOld,2*jmax*sizeof(float)) ;
	BlockMove((const void *)exitWaveFunction,(void *)MNew,2*jmax*sizeof(float)) ;	

	float intialEnergy = 
		CalculateInitialImageMatch(exitWaveFunction,
						FTexperimentalImage,sigmaImage,calculatedImage) ;
	gParameterFittingRecord.previousEnergy	= intialEnergy ;
	gMatchingEnergy = intialEnergy ;
	
	BlockMove((const void *)MOld,(void *)exitWaveFunction,2*jmax*sizeof(float)) ;
		
	time(&secs) ;	
	sprintf(DateString,"_%d",secs) ;
	if (gSaveLogFile) {
		strcpy(filename,gStructureName) ;
		strcat(filename,DateString) ;
		strcat(filename,".log") ;
		fp = fopen(filename,"w") ;
	}
	
	
	numAttempt = 0 ;

	gParameterFittingRecord.temperature.current = gParameterFittingRecord.temperature.start ;
	gRefineParameterProgressPage = 
			new TRefineParameterProgressPage(gControlWindow) ;
	SetPage(gRefineParameterProgressPage) ;

	if (gSaveLogFile)
		gRefineParameterProgressPage->SetLogFileName(filename) ;

	gRefineParameterProgressPage->Show() ;

	for(gParameterFittingRecord.temperature.current=
			gParameterFittingRecord.temperature.start;
			gMatchingEnergy > gParameterFittingRecord.minEnergy && 
			gParameterFittingRecord.temperature.current > 
			gParameterFittingRecord.temperature.end;
			gParameterFittingRecord.temperature.current *= 
			gParameterFittingRecord.temperature.dTFraction) {
		
		numSuccesfulTries = 0 ;
		for (gParameterFittingRecord.numTriesThisTemperature=0;
				gMatchingEnergy > gParameterFittingRecord.minEnergy && 
				gParameterFittingRecord.numTriesThisTemperature < maxTriesPerTemp ; 
				gParameterFittingRecord.numTriesThisTemperature++) {
			
			// Modify one of the parameters
			oldDisplRec = gDisplayRec ;
			oldMslicRec = gMsliceRec ;
			newExitWavefunction = false ;
			if (numAttempt != 0) {
				ModifyParameter(&groupIndex) ;	
				if (groupIndex != -1) { //Something valid has changed
				// Calculate new exit wavefunction if needed
					if (gParameterFittingRecord.parameter[groupIndex].ID == 'THIC' || 
							gParameterFittingRecord.parameter[groupIndex].ID == 'CTLT') {
									
						BlockMove((const void *)MNew,(void *)MOld,
							2*jmax*sizeof(float)) ;	// So we can resore it if needed
						CalculateNewExitWaveFunction(PP0) ;
						exitWaveFunction = gWorkSpace ;
						BlockMove((const void *)exitWaveFunction,
							(void *)MNew,2*jmax*sizeof(float)) ;	// Save it 
						newExitWavefunction = true ;
					}
					else {
						exitWaveFunction = gWorkSpace ;
						BlockMove((const void *)MNew,(void *)exitWaveFunction,
							2*jmax*sizeof(float)) ;
					}
				}
				else {
					//PutMessage("Received Invalid change index") ;
				}
			}
			else {
				time(&lastOutputTime) ;	
				startTime = lastOutputTime ;
				gRefineParameterProgressPage->UpdateCurrentParameters();
				gRefineParameterProgressPage->UpdateComputedImage(calculatedImage);
				// We don't have to do anything. The exitwavefunction 
				// is at gWorkSpace and MOld and MNew is set
			}
			gIndex = groupIndex ;
			if (gAbortStatus) {
				gAbortStatus = false ;
				// restore to last accepted value
				gDisplayRec = oldDisplRec ;
				gMsliceRec = oldMslicRec ;
				if (newExitWavefunction) {
					BlockMove((const void *)MOld,(void *)MNew,2*jmax*sizeof(float)) ;
				}
				if (gIndex != -1) {
					// Restore the previous data
					void RestoreParameterRefinementData() ;
				}
				break ;
			}
			acceptedValue = false ;
			if (groupIndex != -1) {
				// FTexperimentalImage = experimentalImage for 
				// everything but kFTCrossCorrelation
				if (acceptedValue=AcceptNewParameter(exitWaveFunction,
						FTexperimentalImage,sigmaImage,calculatedImage)) {	
					numSuccesfulTries++ ;
					CopyImage(calculatedImage,lastAcceptedImage) ;
					if (gSaveLogFile) {
						transitionIndex++ ;
						fprintf(fp,"%d\t%d\t%f\n",transitionIndex,
							numSuccesfulTries,gParameterFittingRecord.thisEnergy) ;
					}
					if (numSuccesfulTries > maxSuccessfulTries)
						break ;
				}
				else {
					// restore previous data
					gDisplayRec = oldDisplRec ;
					gMsliceRec = oldMslicRec ;
					if (newExitWavefunction) {
						BlockMove((const void *)MOld,(void *)MNew,2*jmax*sizeof(float)) ;
					}
					// Restore the previous data
					void RestoreParameterRefinementData() ;
				}
			}
			if (acceptedValue) {
				time(&currentTime) ;	
				if ((currentTime-lastOutputTime) > timeBetweenOutput) {
					// Update our progress dialog
					gRefineParameterProgressPage->UpdateComputedImage(calculatedImage);
					
					// Write current energy to the window
					gRefineParameterProgressPage->UpdateCurrentParameters();
					lastOutputTime = currentTime ; 
				}
			}
			numAttempt++ ;
		}
		if (abandn()) 
			break ;
		if (numSuccesfulTries == 0) {
			break ;
		}
			
	}
	if (gSaveLogFile) {
		fclose(fp) ;
	}

	gCalculateExactFTXCF = sCalculateExactFTXCF ;
	gEstimateImageShift = sEstimateImageShift ;
	if (gParameterFittingRecord.fittingMethod == kFTAmplCrossCorrelation) {
		DeleteImage(FTexperimentalImage) ;		
	}
	
	gRefineParameterProgressPage->DisplayLatestParameters() ;
	if (lastAcceptedImage != nil)
		gRefineParameterProgressPage->UpdateComputedImage(lastAcceptedImage) ;
	
	if (gAbortStatus) 
		gAbortStatus = false ;
	
	free(PP0) ;
	free(MNew) ;
	free(MOld) ;
	PP0 = nil ;

	if (experimentalImage != nil) DeleteImage(experimentalImage) ;
	if (sigmaImage != nil) DeleteImage(sigmaImage) ;
	if (calculatedImage != nil) DeleteImage(calculatedImage) ;
	if (lastAcceptedImage != nil) DeleteImage(lastAcceptedImage) ;

}


void CompareAndRefineStructure() {
	SimulateAnnealing() ;
}

void SimulateAnnealing() {
	long jmax,groupIndex ;
	long	numSuccesfulTries ;
	char 	filename[32],*cp ;
	time_t	secs,lastOutputTime,currentTime,timeBetweenOutput=1 ;
	long	maxTriesPerTemp,maxSuccessfulTries,transitionIndex,
		numAttempt,numDisplayIndex ;
	long 	ierr ;
	float	dx,dy, maxValueOfEnergy, minValueOfEnergy, scaleFactor,dz,ddw,docc ;
	short	ix,iy ;
	TImage	*experimentalImage=nil,*calculatedImage=nil,
		*sigmaImage=nil,*lastAcceptedImage=nil ;
	short	count = 0 ;
	float 	*pNew ;
	FILE	*fp ;
	FRect	activeRectangle ;
	TCrystal	*theCrystal ;
	unsigned char date[256] ;
	Boolean	goAhead ;

	float ewavel,evolt ;
	
	// Calculate the wavelength in centimeters if we aren't yet set
	evolt = gMsliceRec.volt ;
	ewavel = 1.2263652e-7/sqrt((evolt+9.785025e-4*evolt*evolt)*1000.) ;
	gMSRec.sigma = 2.08886e+6*ewavel*sqrt(1.0+5.886579e-20/(ewavel*ewavel)) ;
	gMSRec.wavel = 1.0e+8*ewavel ;  // Wavelength in angstroms
	
	setup_trnsf() ;
	SetPhasegratingParameters(AtomNames,noz) ;
	
	jmax = gPhsgrtRec->nbeam*gPhsgrtRec->mbeam ;
	
	// This is the aspect ratio of the height to the width for the
	// experimental image
	dx = gExperimentalImage->mWidth ;
	dy = gExperimentalImage->mHeight ;
	gAspectRatio = dy/dx ;

	experimentalImage = CopyImage(gExperimentalImage) ;
	// Instead of flipping the experimental image, we will flip the calculated images
	
	// Let us see if the image dimensions are different from the simulation sampling.
	// If one of the dimensions is different, adjust to fit the calculated image
	if (experimentalImage->mWidth != gPhsgrtRec->nbeam || 
			experimentalImage->mHeight != gPhsgrtRec->mbeam) {
		if (AdjustImageSize(&experimentalImage, gPhsgrtRec->nbeam,gPhsgrtRec->mbeam))
			experimentalImage->mSimulationSampling = true ;
	}
	
	sigmaImage = nil ;
	if (HasStandardDeviationImage()) {
		sigmaImage = CopyImage(gStandardDeviationImage) ;
		if (sigmaImage->mWidth != gPhsgrtRec->nbeam || 
				sigmaImage->mHeight != gPhsgrtRec->mbeam) {
			if (AdjustImageSize(&sigmaImage, gPhsgrtRec->nbeam,gPhsgrtRec->mbeam))
				sigmaImage->mSimulationSampling = true ;
		}
	}

	if (gIncludeImageComparison) {
		if (PP0 != nil) 
			free(PP0) ;
		PP0 = (float *)malloc(2*jmax*sizeof(float)) ;
		if (PP0 == nil) {
			PutMessage("Unable to allocate space for the initial potential") ;
			if (experimentalImage != nil) DeleteImage(experimentalImage) ;
			if (sigmaImage != nil) DeleteImage(sigmaImage) ;
			if (calculatedImage != nil) DeleteImage(calculatedImage) ;
			if (lastAcceptedImage != nil) DeleteImage(lastAcceptedImage) ;
			ExitFromStructureAnnealing() ;
			return ;
		}
		pNew = (float *)malloc(2*jmax*sizeof(float)) ;
		if (pNew == nil) {
			PutMessage("Unable to allocate space for the initial potential") ;
			if (experimentalImage != nil) DeleteImage(experimentalImage) ;
			if (sigmaImage != nil) DeleteImage(sigmaImage) ;
			if (calculatedImage != nil) DeleteImage(calculatedImage) ;
			if (lastAcceptedImage != nil) DeleteImage(lastAcceptedImage) ;
			ExitFromStructureAnnealing() ;
			return ;
		}
		
		calculatedImage = CreateNewImage(experimentalImage->mWidth,
			experimentalImage->mHeight) ;
		if (calculatedImage == nil) {
			PutMessage("Unable to allocate space for the calculated image") ;
			if (experimentalImage != nil) DeleteImage(experimentalImage) ;
			if (sigmaImage != nil) DeleteImage(sigmaImage) ;
			if (calculatedImage != nil) DeleteImage(calculatedImage) ;
			if (lastAcceptedImage != nil) DeleteImage(lastAcceptedImage) ;
			ExitFromStructureAnnealing() ;
			return ;
		}
		lastAcceptedImage = CreateNewImage(experimentalImage->mWidth,
			experimentalImage->mHeight) ;
		if (lastAcceptedImage == nil) {
			PutMessage("Unable to allocate space for the last accepted image") ;
			if (experimentalImage != nil) DeleteImage(experimentalImage) ;
			if (sigmaImage != nil) DeleteImage(sigmaImage) ;
			if (calculatedImage != nil) DeleteImage(calculatedImage) ;
			if (lastAcceptedImage != nil) DeleteImage(lastAcceptedImage) ;
			ExitFromStructureAnnealing() ;
			return ;
		}
		SetImageName(lastAcceptedImage,"Final") ;
	}

	if (gRelaxationParamRecord.atom.useActiveRect) {
		activeRectangle = gRelaxationParamRecord.atom.activeRect ;
	}
	else {
		activeRectangle.top = 0. ;
		activeRectangle.left = 0. ;
		activeRectangle.bottom = 1. ;
		activeRectangle.right = 1. ;
	}
	
	maxTriesPerTemp = 2*gRefinementList.numGroups*
		gRelaxationParamRecord.temperature.maxNumTriesPerTemperature ;
	maxSuccessfulTries = 2*gRefinementList.numGroups*
		gRelaxationParamRecord.temperature.maxNumGoodTries ;

	groupIndex = 0 ;
	dx = 0. ;
	dy = 0. ;
	dz = 0. ;
	ddw = 0. ;
	docc = 0. ;

	switch(gRelaxationParamRecord.fittingMethod) {
		case kCrossCorrelation:
		case kFTAmplCrossCorrelation:
			gRelaxationParamRecord.minEnergy = 0.0001 ; // 0.001 ;
			break ;
		case kChiSquared:
			if (HasStandardDeviationImage()) {			
				gRelaxationParamRecord.minEnergy = 1.0 ;
			}
			else {	// probably testing, so set the min to 0.1
				gRelaxationParamRecord.minEnergy = 0.1 ;
			}
			break ;
	}
	
	gRelaxationParamRecord.thisEnergy = gRelaxationParamRecord.minEnergy + 1. ;
	gRelaxationParamRecord.previousEnergy = 10000000. ; // so to set it high
	gMatchingEnergy = gRelaxationParamRecord.previousEnergy ;	
	
	time(&secs) ;	
	sprintf(DateString,"_%d",secs) ;	
	if (gSaveLogFile) {
		strcpy(filename,gStructureName) ;
		strcat(filename,DateString) ;
		strcat(filename,".log") ;
		fp = fopen(filename,"w") ;
	}
	
	transitionIndex = 0 ;	
	gChemicalEnergy = -1. ; // At start of a run
	gBondValenceEnergy = -1. ; // At start of a run
		
			
	goAhead = false ;	
	if (gIncludeImageComparison) {
		float	*exitWaveFunction ;
		goAhead = CalculateInitialPotential(PP0) ;
		CalculateNewExitWaveFunction(PP0) ;
		// Calculate the Image
		exitWaveFunction = gWorkSpace ;
		CalculateNewImage(exitWaveFunction,calculatedImage) ;
		CopyImage(calculatedImage,lastAcceptedImage) ;
	}
	else {
		goAhead = ((gIncludeBondDistances&& gChemicalBondData.numberOfBonds > 0)
			  || (gIncludeBondValenceSum && gBondStrengthData.numberOfBonds > 0)) ;
	}

	if (!goAhead) {
		if (gIncludeImageComparison) {
			PutMessage("Error in CalculateNewPotential") ;
		}
		else {
			PutMessage("No refinement criteria have been specified") ;
		}
		ExitFromStructureAnnealing() ;
		if (gIncludeImageComparison) {
			free(PP0) ;
			free(pNew) ;
		}

		if (experimentalImage != nil) DeleteImage(experimentalImage) ;
		if (sigmaImage != nil) DeleteImage(sigmaImage) ;
		if (calculatedImage != nil) DeleteImage(calculatedImage) ;
		if (lastAcceptedImage != nil) DeleteImage(lastAcceptedImage) ;

		PP0 = nil ;
	
		if (gSymOps != nil) {
			delete[] gSymOps ;
			gSymOps = nil ;
		}
		return ;
	}

	
	gRelaxationParamRecord.temperature.current=gRelaxationParamRecord.temperature.start ;

	gStructureRefinementProgressPage = 
			new TStructureRefinementProgressPage(gControlWindow,gProjCell) ;
	SetPage(gStructureRefinementProgressPage) ;

	theCrystal = gStructureRefinementProgressPage->GetCrystal() ;
	if (gUseSymmetryElements) {
		theCrystal->SetSymmops(gNumSymops,symop,&gTMatrix) ;
		theCrystal->UseSymmetries(gUseSymmetryElements) ;	
	}

	//gRelaxationParamRecord.thisEnergy = CalculateMatchingEnergy(experimentalImage,
	//	sigmaImage,calculatedImage) ;
	//gMatchingEnergy = gRelaxationParamRecord.thisEnergy ;
	//gRelaxationParamRecord.previousEnergy = gMatchingEnergy ;

	gStructureRefinementProgressPage->Show() ;
						
	if (gIncludeImageComparison) {
		gStructureRefinementProgressPage->UpdateComputedImage(calculatedImage);
	}
	if (gSaveLogFile) {
		gStructureRefinementProgressPage->SetLogFileName(filename) ;
	}
	
		
	for(gRelaxationParamRecord.temperature.current=gRelaxationParamRecord.temperature.start;
			gMatchingEnergy >= gImageComparisonWeight*gRelaxationParamRecord.minEnergy && 
			gRelaxationParamRecord.thisEnergy > gRelaxationParamRecord.minEnergy &&
			gRelaxationParamRecord.temperature.current > gRelaxationParamRecord.temperature.end;
			gRelaxationParamRecord.temperature.current *= 
				gRelaxationParamRecord.temperature.dTFraction) {
			
		numSuccesfulTries = 0 ;
		for (gRelaxationParamRecord.numTriesThisTemperature=0;
				gMatchingEnergy >= gImageComparisonWeight*gRelaxationParamRecord.minEnergy && 
				gRelaxationParamRecord.thisEnergy > gRelaxationParamRecord.minEnergy &&
				gRelaxationParamRecord.numTriesThisTemperature < maxTriesPerTemp ; 
				gRelaxationParamRecord.numTriesThisTemperature++) {
				
			if (gUseSymmetryElements) {
				SaveBasis() ;
			}
			ModifyStructure(&groupIndex,&dx,&dy,&dz,&ddw,&docc) ;	
			if (gIncludeImageComparison) {
				CalculateNewPotential(PP0,pNew,groupIndex,dx,dy,dz,ddw,docc) ;
			}

			if (abandn()) {
				RestoreBasis() ;
				break ;
			}
			gIndex = groupIndex ;
			gDx = dx ;
			gDy = dy ;
			gDz = dz ;
			gDdw = ddw ;
			gDocc = docc ;
			if (AcceptNewStructure(pNew,experimentalImage,
						sigmaImage,calculatedImage)) {
				ReplaceStructureWithNew(PP0,pNew,groupIndex,dx,dy,dz,ddw,docc) ;
				if (gIncludeImageComparison) {
					CopyImage(calculatedImage,lastAcceptedImage) ;
				}
				numSuccesfulTries++ ;
				time(&currentTime) ;
				if ((currentTime-lastOutputTime) > timeBetweenOutput) {
						
					// Update the model window
					theCrystal->RedisplayShiftedAtoms() ;
						
					if (gIncludeImageComparison) {
						gStructureRefinementProgressPage->UpdateComputedImage(calculatedImage);
					} 
						
					// Write current energy to the window
					gStructureRefinementProgressPage->UpdateCurrentParameters() ;
					lastOutputTime = currentTime ;
				}

				if (gSaveLogFile) {
					transitionIndex++ ;
					fprintf(fp,"%d\t%d\t%f\n",transitionIndex,numSuccesfulTries,
						gRelaxationParamRecord.thisEnergy) ;
				}
				if (numSuccesfulTries > maxSuccessfulTries)
					break ;
			}
			else {
				if (gUseSymmetryElements) {
					RestoreBasis() ;
				}
			}
		}
		if (abandn()) {
			RestoreBasis() ;
			break ;
		}
		if (numSuccesfulTries == 0)
			break ;
				
	}

	if (gUseSymmetryElements) {
		DeleteSavedBasis() ;
	}
	if (gSaveLogFile) {
		fclose(fp) ;
	}

	if (gIncludeImageComparison) {
		theCrystal->RedisplayShiftedAtoms() ;
		gStructureRefinementProgressPage->UpdateComputedImage(lastAcceptedImage);
	}
	gStructureRefinementProgressPage->DisplayLatestParameters() ;
		
	if (gAbortStatus) {
		gAbortStatus = false ;
	}


	if (gIncludeImageComparison) {
		free(PP0) ; PP0 = nil ;
		free(pNew) ; pNew = nil ;
	}

	if (experimentalImage != nil) DeleteImage(experimentalImage) ;
	if (sigmaImage != nil) DeleteImage(sigmaImage) ;
	if (calculatedImage != nil) DeleteImage(calculatedImage) ;
	if (lastAcceptedImage != nil) DeleteImage(lastAcceptedImage) ;
	
}

float BondValenceEnergy() {
	long 	i,j,k,n,ndx,zno1,zno2,index ;
	float	lowerTolerance,upperTolerance ;
	float	dx,dy,dz,r,x1,y1,z1,tsq,drsq,rsq,bondValenceSum,changeInEnergy,UpperValenceSum,LowerValenceSum,
			penalty ;
	Boolean	calculateStrength ;

	if (!gCalculateBondStrengths) {
		return 0. ;
	}
	changeInEnergy = 0. ; 
	lowerTolerance = .98 ;
	upperTolerance = 1.02 ;
	if (gBondValenceEnergy == -1.) {	// This is at the start of the run
		// We need to calculate the bond valence sum for all the specified atoms that
		// have been specified as movable, or the ones that have a bond with a movable
		// atom
		gBondValenceEnergy = 0. ;
		for(j=0; j < gRefinementList.numGroups;j++) {
			for(k=0; k < (gRefinementList.groupPtr+j)->numInGroup ; k++) {
				ndx = *((gRefinementList.groupPtr+j)->atomIndex+k) ;// This is the index into gProjCell
				calculateStrength = false ;
				zno1 = *(gProjCell->atno+ndx) ;
				// Is it the first atom in a group
				for (i=0;i<gBondStrengthData.numberOfBonds;i++) {
					if (zno1 == (gBondStrengthData.bondStrength[i].firstAtomNdx+1)) {
						calculateStrength = true ;
						index = i ;
						r = gBondStrengthData.bondStrength[i].maxBondLength ;
						UpperValenceSum = 
							upperTolerance*gBondStrengthData.bondStrength[i].valenceSum ;
						LowerValenceSum = 
							lowerTolerance*gBondStrengthData.bondStrength[i].valenceSum ;
						rsq = r*r ;
						break ;
					}
				}
				if (calculateStrength) {
					bondValenceSum = BondValenceSum(ndx,index) ;
					// Now compare the bondValenceSum to what we should have
					penalty = 0. ;
					if (bondValenceSum > UpperValenceSum) {
						penalty = bondValenceSum - UpperValenceSum ;
						penalty = penalty*penalty ;
					}
					else if (bondValenceSum < LowerValenceSum) {
						penalty = bondValenceSum - LowerValenceSum ;
						penalty = penalty*penalty ;
					}
					gBondValenceEnergy += gBondStrengthData.bondStrength[index].weight*penalty ;
				}
				
				// Se if we also should include some neighbors to the atoms we are moving
				calculateStrength = false ;
				zno2 = zno1 ;
				for (i=0;i<gBondStrengthData.numberOfBonds;i++) {
					if (zno2 == (gBondStrengthData.bondStrength[i].secondAtomNdx+1)) {
						calculateStrength = true ;
						index = i ;
						r = gBondStrengthData.bondStrength[i].maxBondLength ;
						UpperValenceSum = upperTolerance*gBondStrengthData.bondStrength[i].valenceSum ;
						LowerValenceSum = lowerTolerance*gBondStrengthData.bondStrength[i].valenceSum ;
						rsq = r*r ;
						break ;
					}
				}
				x1 = *(gProjCell->x+ndx) ;
				y1 = *(gProjCell->y+ndx) ;
				z1 = *(gProjCell->z+ndx) ;
				if (calculateStrength) {	// It is the second atom in a bond-strength calculation	
					for(i=0;i<gProjCell->natom;i++) {	
						// We need to compare every atom that may be effected by moving this atom
						zno1 = *(gProjCell->atno+i) ;
						if (zno1 == (gBondStrengthData.bondStrength[index].firstAtomNdx+1)) {
							if (NotInGroup(i)) {
								// This is a bond to add to the sum if we are within bond distance
								dx = (*(gProjCell->x+i) - x1)*gProjCell->a ;
								if (fabs(dx) <= r) {
									dy = (*(gProjCell->y+i) - y1)*gProjCell->b ;
									if (fabs(dy) <= r) {
										tsq = dx*dx+dy*dy ;
										if (tsq <= rsq) {
											dz = (*(gProjCell->z+i) - z1)*gProjCell->c ;
											drsq = tsq+dz*dz ;
											if (drsq <= rsq) {
												// bond valence strength
												bondValenceSum = BondValenceSum(i,index) ;
												// Now compare the bondValenceSum to what we should have
												penalty = 0. ;
												if (bondValenceSum > UpperValenceSum) {
													penalty = bondValenceSum - UpperValenceSum ;
													penalty = penalty*penalty ;
												}
												else if (bondValenceSum < LowerValenceSum) {
													penalty = bondValenceSum - LowerValenceSum ;
													penalty = penalty*penalty ;
												}
												gBondValenceEnergy += 
													gBondStrengthData.bondStrength[index].weight*penalty ;	
											}
										}
									}
								}
							}
						}
					}					
				}

			}
		}
		gLastBondValenceEnergy = gBondValenceEnergy ;
	}
	else {
		//Move the atoms to the new positions
		for(n=0; n< (gRefinementList.groupPtr+gIndex)->numInGroup;n++) {
			ndx = *((gRefinementList.groupPtr+gIndex)->atomIndex+n) ;
			// Calculate the energy for the position of the atom being moved before
			// and after it is moved
			*(gProjCell->x+ndx) += gDx ;	
			*(gProjCell->y+ndx) += gDy ;	
			*(gProjCell->z+ndx) += gDz ;	
			*(gProjCell->dw+ndx) += gDdw ;	
			*(gProjCell->occ+ndx) += gDocc ;	
		}
		gBondValenceEnergy = 0. ;
		for(j=0; j < gRefinementList.numGroups;j++) {
			for(k=0; k < (gRefinementList.groupPtr+j)->numInGroup ; k++) {
				ndx = *((gRefinementList.groupPtr+j)->atomIndex+k) ;// This is the index into gProjCell
				calculateStrength = false ;
				zno1 = *(gProjCell->atno+ndx) ;
				for (i=0;i<gBondStrengthData.numberOfBonds;i++) {
					if (zno1 == (gBondStrengthData.bondStrength[i].firstAtomNdx+1)) {
						calculateStrength = true ;
						index = i ;
						r = gBondStrengthData.bondStrength[i].maxBondLength ;
						UpperValenceSum = upperTolerance*gBondStrengthData.bondStrength[i].valenceSum ;
						LowerValenceSum = lowerTolerance*gBondStrengthData.bondStrength[i].valenceSum ;
						rsq = r*r ;
						break ;
					}
				}
				if (calculateStrength) {
					bondValenceSum = BondValenceSum(ndx,index) ;
					// Now compare the bondValenceSum to what we should have
					penalty = 0. ;
					if (bondValenceSum > UpperValenceSum) {
						penalty = bondValenceSum - UpperValenceSum ;
						penalty = penalty*penalty ;
					}
					else if (bondValenceSum < LowerValenceSum) {
						penalty = bondValenceSum - LowerValenceSum ;
						penalty = penalty*penalty ;
					}
					gBondValenceEnergy += gBondStrengthData.bondStrength[index].weight*penalty ;			
				}

				// See if we also should include some neighbors to the atoms we are moving
				calculateStrength = false ;
				for (i=0;i<gBondStrengthData.numberOfBonds;i++) {
					if (zno1 == (gBondStrengthData.bondStrength[i].secondAtomNdx+1)) {
						calculateStrength = true ;
						index = i ;
						r = gBondStrengthData.bondStrength[i].maxBondLength ;
						UpperValenceSum = upperTolerance*gBondStrengthData.bondStrength[i].valenceSum ;
						LowerValenceSum = lowerTolerance*gBondStrengthData.bondStrength[i].valenceSum ;
						rsq = r*r ;
						break ;
					}
				}
				x1 = *(gProjCell->x+ndx) ;
				y1 = *(gProjCell->y+ndx) ;
				z1 = *(gProjCell->z+ndx) ;
				if (calculateStrength) {	// It is the second atom in a bond-strength calculation	
					for(i=0;i<gProjCell->natom;i++) {	// We need to compare every atom that may be effected by moving this atom
						if (i != ndx) {
							// Check bonds with the atom we are considering
							// We will only allow the first atom to have bond to one other atom
							zno1 = *(gProjCell->atno+i) ;
							if (zno1 == (gBondStrengthData.bondStrength[index].firstAtomNdx+1)) {
								// This is a bond to add to the sum if we are within bond distance
								dx = (*(gProjCell->x+i) - x1)*gProjCell->a ;
								if (fabs(dx) <= r) {
									dy = (*(gProjCell->y+i) - y1)*gProjCell->b ;
									if (fabs(dy) <= r) {
										tsq = dx*dx+dy*dy ;
										if (tsq <= rsq) {
											dz = (*(gProjCell->z+i) - z1)*gProjCell->c ;
											drsq = tsq+dz*dz ;
											if (drsq <= rsq) {
												// bond valence strength
												bondValenceSum = BondValenceSum(i,index) ;
												// Now compare the bondValenceSum to what we should have
												penalty = 0. ;
												if (bondValenceSum > UpperValenceSum) {
													penalty = bondValenceSum - UpperValenceSum ;
													penalty = penalty*penalty ;
												}
												else if (bondValenceSum < LowerValenceSum) {
													penalty = bondValenceSum - LowerValenceSum ;
													penalty = penalty*penalty ;
												}
												gBondValenceEnergy += gBondStrengthData.bondStrength[index].weight*penalty ;	
											}
										}
									}
								}
							}
						}
					}					
				}


			}
		}
		// Move the atoms back
		for(n=0; n< (gRefinementList.groupPtr+gIndex)->numInGroup;n++) {
			ndx = *((gRefinementList.groupPtr+gIndex)->atomIndex+n) ;
			// Calculate the energy for the position of the atom being moved before
			// and after it is moved
			*(gProjCell->x+ndx) -= gDx ;	
			*(gProjCell->y+ndx) -= gDy ;	
			*(gProjCell->z+ndx) -= gDz ;	
			*(gProjCell->dw+ndx) -= gDdw ;	
			*(gProjCell->occ+ndx) -= gDocc ;	
		}
	}
	return gBondValenceSumWeight*gBondValenceEnergy ;
}

float OldBondValenceEnergy() {
	long i,j,k,ndx,zno1,index ;
	float	dx,dy,dz,r,x1,y1,z1,tsq,drsq,rsq,bondValenceSum,changeInEnergy,UpperValenceSum,LowerValenceSum,
			newEnergyForThisAtom,previousEnergyForThisAtom,penalty ;
	Boolean	calculateStrength ;

	if (!gCalculateBondStrengths) {
		return 0. ;
	}
	changeInEnergy = 0. ; 
	if (gBondValenceEnergy == -1.) {	// This is at the start of the run
		// We need to calculate the bond valence sum for all the specified atoms that
		// have been specified as movable, or the ones that have a bond with a movable
		// atom
		gBondValenceEnergy = 0. ;
		for(j=0; j < gRefinementList.numGroups;j++) {
			for(k=0; k < (gRefinementList.groupPtr+j)->numInGroup ; k++) {
				ndx = *((gRefinementList.groupPtr+j)->atomIndex+k) ;			// This is the index into gProjCell
				calculateStrength = false ;
				zno1 = *(gProjCell->atno+ndx) ;
				for (i=0;i<gBondStrengthData.numberOfBonds;i++) {
					if (zno1 == (gBondStrengthData.bondStrength[i].firstAtomNdx+1)) {
						calculateStrength = true ;
						index = i ;
						r = gBondStrengthData.bondStrength[i].maxBondLength ;
						UpperValenceSum = 1.05*gBondStrengthData.bondStrength[i].valenceSum ;
						LowerValenceSum = 0.8*gBondStrengthData.bondStrength[i].valenceSum ;
						rsq = r*r ;
						break ;
					}
				}
				if (calculateStrength) {
					bondValenceSum = BondValenceSum(ndx,index) ;
					// Now compare the bondValenceSum to what we should have
					penalty = 0. ;
					if (bondValenceSum > UpperValenceSum) {
						penalty = bondValenceSum - UpperValenceSum ;
						penalty = penalty*penalty ;
					}
					else if (bondValenceSum < LowerValenceSum) {
						penalty = bondValenceSum - LowerValenceSum ;
						penalty = penalty*penalty ;
					}
					gBondValenceEnergy += gBondStrengthData.bondStrength[index].weight*penalty ;			
				}
			}
		}
		gLastBondValenceEnergy = gBondValenceEnergy ;
	}
	else {
		// Calculate the change in bond valence "energy" associated with moving this atom
		changeInEnergy = 0. ;
		for(long n=0; n< (gRefinementList.groupPtr+gIndex)->numInGroup;n++) {
			ndx = *((gRefinementList.groupPtr+gIndex)->atomIndex+n) ;
			// Calculate the energy for the position of the atom being moved before
			// and after it is moved
			previousEnergyForThisAtom = 0. ;
			newEnergyForThisAtom = 0. ;
			
			calculateStrength = false ;
			zno1 = *(gProjCell->atno+ndx) ;
			// Check to see if is the first atom in a bond-valence sum calculation
			for (i=0;i<gBondStrengthData.numberOfBonds;i++) {
				if (zno1 == (gBondStrengthData.bondStrength[i].firstAtomNdx+1)) {
					calculateStrength = true ;
					UpperValenceSum = 1.05*gBondStrengthData.bondStrength[i].valenceSum ;
					LowerValenceSum = 0.8*gBondStrengthData.bondStrength[i].valenceSum ;
					index = i ;
					r = gBondStrengthData.bondStrength[i].maxBondLength ;
					rsq = r*r ;
					break ;
				}
			}
			if (calculateStrength) {		
				bondValenceSum = BondValenceSum(ndx,index) ;
				// Now compare the bondValenceSum to what we should have
				penalty = 0. ;
				if (bondValenceSum > UpperValenceSum) {
					penalty = bondValenceSum - UpperValenceSum ;
					penalty = penalty*penalty ;
				}
				else if (bondValenceSum < LowerValenceSum) {
					penalty = bondValenceSum - LowerValenceSum ;
					penalty = penalty*penalty ;
				}
				previousEnergyForThisAtom += gBondStrengthData.bondStrength[index].weight*penalty ;	
				
				// Calculate the new energy for this atom	
				*(gProjCell->x+ndx) += gDx ;	
				*(gProjCell->y+ndx) += gDy ;	
				*(gProjCell->z+ndx) += gDz ;	
				bondValenceSum = BondValenceSum(ndx,index) ;
				*(gProjCell->x+ndx) -= gDx ;	
				*(gProjCell->y+ndx) -= gDy ;
				*(gProjCell->z+ndx) -= gDz ;
				// Now compare the bondValenceSum to what we should have
				penalty = 0. ;
				if (bondValenceSum > UpperValenceSum) {
					penalty = bondValenceSum - UpperValenceSum ;
					penalty = penalty*penalty ;
				}
				else if (bondValenceSum < LowerValenceSum) {
					penalty = bondValenceSum - LowerValenceSum ;
					penalty = penalty*penalty ;
				}
				newEnergyForThisAtom += gBondStrengthData.bondStrength[index].weight*penalty ;	
			}
			
			
			calculateStrength = false ;
			// Check to see if is the second atom in a bond-valence sum calculation
			for (i=0;i<gBondStrengthData.numberOfBonds;i++) {
				if (zno1 == (gBondStrengthData.bondStrength[i].secondAtomNdx+1)) {
					calculateStrength = true ;
					UpperValenceSum = 1.05*gBondStrengthData.bondStrength[i].valenceSum ;
					LowerValenceSum = 0.8*gBondStrengthData.bondStrength[i].valenceSum ;
					index = i ;
					r = gBondStrengthData.bondStrength[i].maxBondLength ;
					rsq = r*r ;
					break ;
				}
			}
			x1 = *(gProjCell->x+ndx) ;
			y1 = *(gProjCell->y+ndx) ;
			z1 = *(gProjCell->z+ndx) ;
			if (calculateStrength) {	// It is the second atom in a bond-strength calculation	
				for(i=0;i<gProjCell->natom;i++) {	// We need to compare every atom that may be effected by moving this atom
					if (i != ndx) {
						// Check bonds with the atom we are considering
						// We will only allow the first atom to have bond to one other atom
						zno1 = *(gProjCell->atno+i) ;
						if (zno1 == (gBondStrengthData.bondStrength[index].firstAtomNdx+1)) {
							// This is a bond to add to the sum if we are within bond distance
							dx = (*(gProjCell->x+n) - x1)*gProjCell->a ;
							if (fabs(dx) <= r) {
								dy = (*(gProjCell->y+n) - y1)*gProjCell->b ;
								if (fabs(dy) <= r) {
									tsq = dx*dx+dy*dy ;
									if (tsq <= rsq) {
										dz = (*(gProjCell->z+n) - z1)*gProjCell->c ;
										drsq = tsq+dz*dz ;
										if (drsq <= rsq) {
											// bond valence strength
											bondValenceSum = BondValenceSum(i,index) ;
											// Now compare the bondValenceSum to what we should have
											penalty = 0. ;
											if (bondValenceSum > UpperValenceSum) {
												penalty = bondValenceSum - UpperValenceSum ;
												penalty = penalty*penalty ;
											}
											else if (bondValenceSum < LowerValenceSum) {
												penalty = bondValenceSum - LowerValenceSum ;
												penalty = penalty*penalty ;
											}
											previousEnergyForThisAtom += gBondStrengthData.bondStrength[index].weight*penalty ;	
											
											// Now we should calculate the bond valence sum for this atom after we move the one in question
											// This is a bond to add to the sum if we are within bond distance
											*(gProjCell->x+ndx) += gDx ;	
											*(gProjCell->y+ndx) += gDy ;	
											*(gProjCell->z+ndx) += gDy ;	
											bondValenceSum = BondValenceSum(i,index) ;
											*(gProjCell->x+ndx) -= gDx ;	
											*(gProjCell->y+ndx) -= gDy ;
											*(gProjCell->z+ndx) -= gDz ;	
											// Now compare the bondValenceSum to what we should have
											penalty = 0. ;
											if (bondValenceSum > UpperValenceSum) {
												penalty = bondValenceSum - UpperValenceSum ;
												penalty = penalty*penalty ;
											}
											else if (bondValenceSum < LowerValenceSum) {
												penalty = bondValenceSum - LowerValenceSum ;
												penalty = penalty*penalty ;
											}
											newEnergyForThisAtom += gBondStrengthData.bondStrength[index].weight*penalty ;	
										}
									}
								}
							}
						}
					}
				}					
			}
		}
		changeInEnergy = newEnergyForThisAtom - previousEnergyForThisAtom ;	
	}
	gBondValenceEnergy += changeInEnergy ;
	return gBondValenceSumWeight*gBondValenceEnergy ;
}

float ChemicalEnergy() {
	float changeInEnergy ;
	float	previousEnergyForThisAtom,newEnergyForThisAtom ;
	double	dx,dy,dz,drsq,argsq,r0sq,drmax,r0,weight,arg8 ;
	double	cutoff = 4.0 ;
	long i,j,index,ndx ;
	Boolean is3D ;
	
	if (gChemicalBondData.numberOfBonds == 0) {
		if (gChemicalEnergy == -1. ) {
			gChemicalEnergy = 0. ;
			gChangeInChemicalEnergy = 0. ;
		}
		return 0. ;
	}
	

	//sumWeights = 0. ;
	Boolean hasChemicalEnergy ;
	short	zno1,zno2 ;
	// The energy we are using is W = -5/(r/r0)**8 + 4/(r/r0)**10
	// Let us try W = -2.5/(r/r0)**8 + 2/(r/r0)**10 + 0.5exp(((r/r0 - 1)**2)/0.2)
	// Let us try W = -0.25/(r/r0)**8 + 0.2/(r/r0)**10
	if (gChemicalEnergy == -1. ) {
		// We need to calculate the total energy of the system
		gChemicalEnergy = 0. ;
		for(j=0;j<gProjCell->natom;j++) {
			// See if this atom has a bond length associated with it
			hasChemicalEnergy = false ;
			zno1 = *(gProjCell->atno+j) ;
			for(long k=0;k<gChemicalBondData.numberOfBonds;k++) {
				if (zno1 == gChemicalBondData.bond[k].typeAtom1 || zno1 == gChemicalBondData.bond[k].typeAtom2) {
					hasChemicalEnergy = true ;
					break ;
				}
			}
			if (hasChemicalEnergy) {
				for(i=0;i<gProjCell->natom;i++) {
					if (i != j) {
						hasChemicalEnergy = false ;
						zno2 = *(gProjCell->atno+i) ;
						for(long l=0;l<gChemicalBondData.numberOfBonds;l++) {
							if (zno1 == gChemicalBondData.bond[l].typeAtom1 && zno2 == gChemicalBondData.bond[l].typeAtom2) {
								hasChemicalEnergy = true ;
								r0 = gChemicalBondData.bond[l].optimumDistance ;
								r0sq = r0*r0 ;
								drmax = 4*r0 ;
								weight = gChemicalBondData.bond[l].weight ;
								is3D = gChemicalBondData.bond[l].use3Ddistance ;
								break ;
							}
							else if (zno2 == gChemicalBondData.bond[l].typeAtom1 && zno1 == gChemicalBondData.bond[l].typeAtom2) {
								hasChemicalEnergy = true ;
								r0 = gChemicalBondData.bond[l].optimumDistance ;
								r0sq = r0*r0 ;
								drmax = 4*r0 ;
								weight = gChemicalBondData.bond[l].weight ;
								is3D = gChemicalBondData.bond[l].use3Ddistance ;
								break ;
							}
						}
						if (hasChemicalEnergy) {
							dx = (*(gProjCell->x+i) - *(gProjCell->x+j))*gProjCell->a ;
							dy = (*(gProjCell->y+i) - *(gProjCell->y+j))*gProjCell->b ;
							if (is3D) {
								dz = (*(gProjCell->z+i) - *(gProjCell->z+j))*gProjCell->c ;
								drsq = dx*dx+dy*dy+dz*dz ;
							}
							else {
								drsq = dx*dx+dy*dy ;
							}
							argsq = drsq/r0sq ;
							if (argsq < cutoff && argsq > 0.000001) {
								arg8 = argsq*argsq*argsq*argsq ;
								//arg = sqrt(argsq) ;
								gChemicalEnergy += weight*(0.2/(argsq*arg8) - 0.25/arg8);
								//sumWeights += weight ;	// Find out how many we have
							}
						}
					}
				}
			}
		}
		gChemicalEnergy /= 2. ;
		//gChemicalEnergy += sumWeights/2. ;	
	}
	
	changeInEnergy = 0. ;
	for(long n=0; n< (gRefinementList.groupPtr+gIndex)->numInGroup;n++) {
		index = *((gRefinementList.groupPtr+gIndex)->atomIndex+n) ;
		// Calculate the energy for the position of the atom being moved before
		// and after it is moved
		previousEnergyForThisAtom = 0. ;
		newEnergyForThisAtom = 0. ;
		
		hasChemicalEnergy = false ;
		zno1 = *(gProjCell->atno+index) ;
		for(long k=0;k<gChemicalBondData.numberOfBonds;k++) {
			if (zno1 == gChemicalBondData.bond[k].typeAtom1 || zno1 == gChemicalBondData.bond[k].typeAtom2) {
				hasChemicalEnergy = true ;	// Possibly
				break ;
			}
		}
		if (hasChemicalEnergy) {
			for(i=0;i<gProjCell->natom;i++) {
				if (i != index) {
					hasChemicalEnergy = false ;
					zno2 = *(gProjCell->atno+i) ;
					// First we must ensure that this atom is not part of the same group that we are moving
					Boolean inTheGroup = false ;
					for(long m=0; m< (gRefinementList.groupPtr+gIndex)->numInGroup;m++) {
						ndx = *((gRefinementList.groupPtr+gIndex)->atomIndex+m) ;
						if (i == ndx) {
							// We are in the same group
							inTheGroup = true ;
							break ;
						}
					}
					if (!inTheGroup) {
						for(long l=0;l<gChemicalBondData.numberOfBonds;l++) {
							if (zno1 == gChemicalBondData.bond[l].typeAtom1 && zno2 == gChemicalBondData.bond[l].typeAtom2) {
								hasChemicalEnergy = true ;
								r0 = gChemicalBondData.bond[l].optimumDistance ;
								r0sq = r0*r0 ;
								drmax = 4*r0 ;
								weight = gChemicalBondData.bond[l].weight ;
								is3D = gChemicalBondData.bond[l].use3Ddistance ;
								break ;
							}
							else if (zno2 == gChemicalBondData.bond[l].typeAtom1 && zno1 == gChemicalBondData.bond[l].typeAtom2) {
								hasChemicalEnergy = true ;
								r0 = gChemicalBondData.bond[l].optimumDistance ;
								r0sq = r0*r0 ;
								drmax = 4*r0 ;
								weight = gChemicalBondData.bond[l].weight ;
								is3D = gChemicalBondData.bond[l].use3Ddistance ;
								break ;
							}
						}
						if (hasChemicalEnergy) {
							dx = (*(gProjCell->x+i) - *(gProjCell->x+index))*gProjCell->a ;
							if (dx < drmax) {
								dy = (*(gProjCell->y+i) - *(gProjCell->y+index))*gProjCell->b ;
								if (dy < drmax) {
									if (is3D) {
										dz = (*(gProjCell->z+i) - *(gProjCell->z+index))*gProjCell->c ;
										if (dz < drmax) {
											drsq = dx*dx+dy*dy +dz*dz ;
											argsq = drsq/r0sq ;
											if (argsq < cutoff && argsq > 0.000001) {
												//arg = sqrt(argsq) ;
												arg8 = argsq*argsq*argsq*argsq ; // weight*(4./(argsq*arg8) - 5./arg8) ;
												previousEnergyForThisAtom += weight*(0.2/(argsq*arg8) - 0.25/arg8);
												dx -= gDx*gProjCell->a ;
												dy -= gDy*gProjCell->b ;
												dz -= gDz*gProjCell->c ;
												drsq = dx*dx+dy*dy+dz*dz ;
												argsq = drsq/r0sq ;
												//arg = sqrt(argsq) ;
												arg8 = argsq*argsq*argsq*argsq ;
												newEnergyForThisAtom += weight*(0.2/(argsq*arg8) - 0.25/arg8);
											}
										}
									}
									else {
										drsq = dx*dx+dy*dy ;
										argsq = drsq/r0sq ;
										if (argsq < cutoff && argsq > 0.000001) {
											//arg = sqrt(argsq) ;
											arg8 = argsq*argsq*argsq*argsq ;
											previousEnergyForThisAtom += weight*(0.2/(argsq*arg8)-0.25/arg8);
											dx -= gDx*gProjCell->a ;
											dy -= gDy*gProjCell->b ;
											drsq = dx*dx+dy*dy ;
											argsq = drsq/r0sq ;
											//arg = sqrt(argsq) ;
											arg8 = argsq*argsq*argsq*argsq ;
											newEnergyForThisAtom += weight*(0.2/(argsq*arg8)-0.25/arg8);
										}
									}
								}
							}
						}
					}
				}
			}
		}
		changeInEnergy += newEnergyForThisAtom - previousEnergyForThisAtom ;
	}
	gChangeInChemicalEnergy = changeInEnergy ;
	return (gChemicalEnergy+gChangeInChemicalEnergy) ;
}

Boolean CalculateInitialPotential(float *p0) {
	long ierr ;
	long numFloats ;
	
	numFloats = gPhsgrtRec->nbeam*gPhsgrtRec->mbeam ;
	if (gPGisChanged) {		/* The Phsgrt needs to be run 		*/
		setup_trnsf() ;
		SetWatchCursor() ;
		CalculateProjectedPotential(gWorkSpace,noz,numFloats) ;
		InitCursor() ;
		if (gAbortStatus) {
			gAbortStatus = false ;
			return false  ;
		}
		
		if (gIOError) {
			err_io(gIOError) ;
			return false  ;
		}
		if (ierr == -1)
		  	return false  ;
		gPGisChanged = false ;
		gMSisChanged = true ;
		calculation.status &= ~MSLICE ;
		calculation.status |= PHSGRT ;
		BlockMove((const void *)gWorkSpace,(void *)p0,numFloats*sizeof(float)) ;
		rewrite_file(flname.pin) ;
	}
	else {
		if (!open_input_file(flname.pout)) 
			return false ;
		flag.ipot = 1 ;
		flag.icmplx = 0 ;
		if (!read_input_file(1,(char *)p0,numFloats)) {
			close_input_file() ;
			return false ;
		}
		close_input_file() ;
		gPGisChanged = false ;
		gMSisChanged = true ;
		calculation.status &= ~MSLICE ;
		calculation.status |= PHSGRT ;
		BlockMove((const void *)p0,(void *)gWorkSpace,numFloats*sizeof(float)) ;
	}
	return true ;
}

extern struct { float ax[MAXTYPES][4],bx[MAXTYPES][4],cx[MAXTYPES] ; } xrf ;
Boolean CalculateNewPotential(float *pOld,float *pNew,long groupIndex,float dx,float dy,float dz, float ddw, float docc) {	
	long hmax,kmax,num_elements,index ;
	long nmax,mmax,nh ;
	float umax,umax1,umax2,elapsedTime,dw,dw1,tfact,tfact1 ;
	long h,k,i,j,ndx ;	/* register */
	short type,lastType ;
	float dh,dk,hksq,rm2csga,dhsq,dhga,zf,zf1,*rpp,*rpn ;
	float *px,*py,*pocc,*pdw ;
	short *ptype ;
	float x,y,z,occ,occ1 ; 
	float pi=3.141592654,cnst,fact,usqmax ;
	float arg,pi2 ;
	SFParamFitPtr	paramFit ;
	long jmax ;
	float Vr0,Vr1,Vi0,Vi1 ;
	unsigned char refinementFlag ;

	nmax = gPhsgrtRec->nbeam ;
	mmax = gPhsgrtRec->mbeam ;
	jmax = nmax*mmax ;
	
	paramFit 	= (SFParamFitPtr)malloc(sizeof(SFParamFitRecord)) ;
	if (paramFit == nil) {
		PutMessage("Out of Memory !") ;
		return false ;
	}
	for(j=0;j<gPhsgrtRec->iatom;j++) {		
		paramFit->cx[j] = xrf.cx[j] ;		// Unused if we use electron scattering factors
		for(i=0;i<4;i++) {
			paramFit->ax[j*4+i] = xrf.ax[j][i] ;
			paramFit->bx[j*4+i] = xrf.bx[j][i] ;
		}
	}
	//  the numbers nmax and mmax corresponds to the number of intervals
	//  along the unit cell axes.                                  
    
	cnst = 47.874/gProjCell->vol ;
	hmax = nmax/2 ;
	kmax = mmax/2 ;
        
    
	umax1 = (hmax-1)*(gProjCell->ast)*(gProjCell->siga) ;
	umax2 = (kmax-1)*(gProjCell->bst)*(gProjCell->siga) ;
	umax = min(umax1,umax2) ;
	usqmax = umax*umax ;   

	pi2 = 2*pi ;
	fact = pi ;

	ndx = 0 ;
	
	lastType = -1 ;

	// Zero all the coefficients 
	num_elements = 2*jmax ;
	for(i=0,rpp=pNew;i<num_elements;i++) {
		*rpp++ = 0. ;
	}
	
	rm2csga = -2.*gProjCell->csga ;

	gRefinementFlag = *((gRefinementList.groupPtr+groupIndex)->refine) ;	
		

	//If we are using symmetry elements then the basis need to
	// be modified, symmetry elements applied and the data put into the gProjCell structure
	// This is alot of unneccesary work, but I don't have an elegant solution right now
	if (!gUseSymmetryElements) {
		// Loop over the half plane -hmax+1 < h < hmax; k >= 0 
		// and fill in the bottom half as we go along          
		for (h=0;h<nmax;h++)  { 			// Go along the h-axis          
			abandn() ;
			if (gAbortStatus) {
				free(paramFit) ;
				return false ;
			}
			nh = h ;           				// nh represents the index h 
			if (nh > hmax)
				nh -= nmax ;   				// The index is negative

			dh = nh*(gProjCell->ast) ;
			dhsq = dh*dh ;
			dhga = dh*rm2csga ; 
			for (k=0;k<= kmax;k++)  { 		// Loop over k >= 0
				dk = k*(gProjCell->bst) ;
				hksq = dhsq+dk*dk+dk*dhga ;  // the length of g-square 
				if (hksq <= usqmax)  {       // Do not bother if too large
					lastType = -1 ;
					for(i=0; i< (gRefinementList.groupPtr+groupIndex)->numInGroup;i++) {
						index = *((gRefinementList.groupPtr+groupIndex)->atomIndex+i) ;
						refinementFlag = *((gRefinementList.groupPtr+groupIndex)->refine+i) ;
						// The index is either into the gProjCell when we are not using symmetry 
						// elements -- the normal way ---- or into the structure gBasis when we are
						// using symmetry elements.  
						// Point to the data
						x 	= *(gProjCell->x + index) ;
						y 	= *(gProjCell->y + index) ;
						z 	= *(gProjCell->z + index) ;
						occ = *(gProjCell->occ + index) ;
						dw 	= *(gProjCell->dw + index) ;
						type = *(gProjCell->ntype + index) ;
						if (type != lastType) {
							zf = AtomicFormFactor(hksq,noz,type-1,&dw,paramFit) ;
							lastType = type ;
						}
						if ((refinementFlag & kRefineX) == 0) {
							dx = 0. ;
						}
						if ((refinementFlag & kRefineY) == 0) {
							dy = 0. ;
						}
						if ((refinementFlag & kRefineZ) == 0) {
							dz = 0. ;
						}
						if ((refinementFlag & kRefineDW) != 0) {
							dw1 = dw + ddw ;
							if (dw1 < 0.) dw = 0. ;
							zf1 = AtomicFormFactor(hksq,noz,type-1,&dw1,paramFit) ;
						}
						else {
							zf1 = zf ;
						}
						if ((refinementFlag & kRefineOCC) != 0) {
							occ1 = occ + docc ;
							if (occ1 < 0.) occ1 = 0. ;
						}
						else {
							occ1 = occ ;
						}
						arg = nh*x+k*y ;
						arg *= pi2 ;
						
						Vr0 = cos(arg) ;
						Vi0 = -sin(arg) ;
											
						arg = nh*(x+dx)+k*(y+dy) ;
						arg *= pi2 ;
						
						Vr1 = cos(arg) ;
						Vi1 = -sin(arg) ;
						
						tfact = zf*occ ;
						tfact1 = zf1*occ1 ;
		                rpp = pNew + 2*(k*nmax+h) ;
		                *rpp += cnst*(tfact1*Vr1-tfact*Vr0) ; 		// Keep adding, we can 
		                *(rpp+1) += cnst*(tfact1*Vi1-tfact*Vi0) ;	// move more than one atom
		                   
		                // Set the bottom half of the diffraction plane
		                if (k > 0) {
							if (nh <= 0) {  		// Negative h's, positive k's
								rpn = pNew - 2*nh + 2*(mmax-k)*nmax ;
								*rpn = *rpp ;
								*(rpn+1) = - *(rpp+1) ;
							}
		                    else if (nh > 0) { 
								rpn = pNew + 2*(nmax-nh) + 
										   2*(mmax-k)*nmax ;
								*rpn = *rpp ;
								*(rpn+1) = - *(rpp+1) ;
							}
						}
					}
				}
			}
			//sprintf(stext,"PHSGRT: Calculated %d fcoeff",2*(h+1)*kmax) ;
		  	//WriteStatus(stext) ;
		}
		mmffti(pNew,nmax,mmax,&elapsedTime) ;
		condense(pNew,jmax) ;
		for(i=0,rpp = pOld,rpn = pNew;i<jmax;i++,rpp++,rpn++) {
			*rpn += *rpp ;
		}
	}
	else {
		for(i=0; i< (gRefinementList.groupPtr+groupIndex)->numInGroup;i++) {
			index = *((gRefinementList.groupPtr+groupIndex)->atomIndex+i) ;
			refinementFlag = *((gRefinementList.groupPtr+groupIndex)->refine+i) ;
			// The index is into the structure gBasis when we are
			// using symmetry elements. 
			if ((refinementFlag & kRefineX) != 0) {
				(gBasis.atom + index)->xc += dx ;
			}
			if ((refinementFlag & kRefineY) != 0) {
				(gBasis.atom + index)->yc += dy ;
			}
			if ((refinementFlag & kRefineZ) != 0) {
				(gBasis.atom + index)->zc += dz ;
			}
			if ((refinementFlag & kRefineDW) != 0) {
				(gBasis.atom + index)->dw += ddw ;
				if ((gBasis.atom + index)->dw < 0.) (gBasis.atom + index)->dw = 0. ;
			}
			if ((refinementFlag & kRefineOCC) != 0) {
				(gBasis.atom + index)->occ += docc ;
				if ((gBasis.atom + index)->occ < 0.) (gBasis.atom + index)->occ = 0. ;
			}
		}
		// Apply the symmetry operators to the basis and stuff the result into gProjCell
		// Calculate the new potential
		ApplySymopsToBasis() ;
		float sumr,sumi ;
		long last_type ;
		long  imax ;
		
		imax = gProjCell->natom ;
		for (h=0;h<nmax;h++)  { 			// Go along the h-axis          
			nh = h ;           				// nh represents the index h 
            		if (nh > hmax)
                		nh -= nmax ;   				// The index is negative

            		dh = nh*(gProjCell->ast) ;
            		dhsq = dh*dh ;
            		dhga = dh*rm2csga ;
            		for (k=0;k<= kmax;k++)  { 		// Loop over k >= 0
                		dk = k*(gProjCell->bst) ;
                		hksq = dhsq+dk*dk+dk*dhga ;  // the length of g-square 
                		if (hksq <= usqmax)  {       // Do not bother if too large
				  	sumr = 0. ;
					sumi = 0. ;
                    		last_type = -1 ;
	
                    		for (i=0,px=gProjCell->x,py=gProjCell->y,pocc=gProjCell->occ,
							pdw=gProjCell->dw,ptype=gProjCell->ntype;
                    				i<imax; px++,py++,pocc++,pdw++,ptype++,i++) { 	// Sum over the atoms
						j = *ptype ;
						if (j != last_type)    // Get the scattering factor
							zf = AtomicFormFactor(hksq,noz,j-1,pdw,paramFit) ;
						arg = nh*(*px)+k*(*py) ;
						arg *= pi2 ;
						tfact = zf*(*pocc) ;
						sumr += tfact*cos(arg) ;
						sumi -= tfact*sin(arg) ;
						last_type = j ;
					}
                    		rpp = pNew + 2*(k*nmax+h) ;
                    		*rpp = cnst*sumr ;
                    		if (!gCentroSymmetric)       // Set the imaginary part
                       		*(rpp+1) = cnst*sumi ;
                    		// Set the bottom half of the diffraction plane
                    		if (k > 0) {
						if (nh <= 0) {  		// Negative h's, positive k's
							rpn = pNew - 2*nh + 2*(mmax-k)*nmax ;
							*rpn = *rpp ;
							*(rpn+1) = - *(rpp+1) ;
						}
						else if (nh > 0) { 
							rpn = pNew + 2*(nmax-nh) + 
									   2*(mmax-k)*nmax ;
							*rpn = *rpp ;
							*(rpn+1) = - *(rpp+1) ;
						}
					}
				}
			}
		}

		mmffti(pNew,nmax,mmax,&elapsedTime) ;
		condense(pNew,jmax) ;
	}
	free(paramFit) ;
	return true;
}

void ApplySymopsToBasis() {
  	// ------------------------------------------------------------	
  	// Applies the symmetry operators on the atoms in the basis 
  	// ------------------------------------------------------------	
    
	register long i,j,ndx,icount ;
	float xc,yc,zc ;
	float xnew,ynew,znew ;


	ndx = 0 ;
	for (j=0;j<gBasis.numBasisAtoms;j++) {
		icount = 0 ;
		xc = (gBasis.atom+j)->xc ;
		yc = (gBasis.atom+j)->yc ;
		zc = (gBasis.atom+j)->zc ;
		for (i=0;i<gNumSymops;i++) {
			xnew = gSymOps[i].m[0][0]*xc + gSymOps[i].m[0][1]*yc +
				   gSymOps[i].m[0][2]*zc + gSymOps[i].m[0][3] ;
			ynew = gSymOps[i].m[1][0]*xc + gSymOps[i].m[1][1]*yc +
				   gSymOps[i].m[1][2]*zc + gSymOps[i].m[1][3] ;
			znew = gSymOps[i].m[2][0]*xc + gSymOps[i].m[2][1]*yc +
				   gSymOps[i].m[2][2]*zc + gSymOps[i].m[2][3] ;
			xnew = one(xnew) ;
			ynew = one(ynew) ;
			znew = one(znew) ;
			if (NotEquivalentAtom(xnew,ynew,znew,icount,ndx)) {
				gProjCell->x[ndx] = xnew ;
				gProjCell->y[ndx] = ynew ;
				gProjCell->z[ndx] = znew ;
				gProjCell->dw[ndx] = (gBasis.atom+j)->dw ;
				gProjCell->occ[ndx] = (gBasis.atom+j)->occ ;
				gProjCell->ntype[ndx] =(gBasis.atom+j)->type ;
				gProjCell->atno[ndx] = (gBasis.atom+j)->no ;
				ndx++ ;
				icount++ ;
			}
		}
	}
	gProjCell->natom = ndx ;
}

Boolean NotEquivalentAtom(float x,float y,float z,long icount,long ndx) {

	register long k ;
	long xtest,ytest,ztest ;
	float xdiff,ydiff,zdiff ; 

	xtest = false ; 
	ytest = false ; 
	ztest = false ;
	for (k=ndx-icount;k<ndx;k++) {
		xdiff = (x-gProjCell->x[k])*gProjCell->a ;
		ydiff = (y-gProjCell->y[k])*gProjCell->b ;
		zdiff = (z-gProjCell->z[k])*gProjCell->c ;      
		if (fabs(xdiff) < 0.1) xtest = true ;
		if (fabs(ydiff) < 0.1) ytest = true ;
		if (fabs(zdiff) < 0.1) ztest = true ;
		if (xtest && ytest && ztest) {
			return(false) ;			// The atom already exists 
		}
		xtest = false ; 
		ytest = false ; 
		ztest = false ;
	}
	return(true) ;                   // New atom
}

void CalculateNewExitWaveFunction(float *potential) {
	register long i ;
	UInt32	numBytes ;
	long	jmax ;
	long 	numSlices ;
	float	*zphsgr,*zprop,*zphi,sumInt ;
	register float *rp1 ;
	
	jmax 	= gPhsgrtRec->nbeam*gPhsgrtRec->mbeam ;
	zphi 	= gWorkSpace ;
	zphsgr 	= zphi +  2*jmax ;
	zprop 	= zphsgr + 2*jmax ;
	
	for(i=0,rp1=zphi;i<jmax;i++) {	// Normalized to 1 for every image point
								// Normalizing such that SUM( PSI sq)= 1 
		*rp1++ = 1. ;       	// introduces an undesirable dependence on
		*rp1++ = 0. ;       	// the number of sampling points   
	}
	
	numSlices = (long)(gMsliceRec.thick/gPhsgrtRec->slcth+0.5) ;
	gMsliceRec.thick = numSlices*gPhsgrtRec->slcth ;
	gMsliceRec.t1 = gMsliceRec.thick ;
	numBytes = jmax*sizeof(float) ;
	BlockMove((const void *)potential,(void *)zphsgr,numBytes) ;
	SetPhasegrating(zphsgr,gPhsgrtRec->slcth,gMSRec.sigma,jmax) ;
   	CalculatePropagator(zprop,gPhsgrtRec->slcth) ;
	
	for(i=0;i<numSlices;i++) {
		ApplyPhasegrating(zphi,zphsgr,jmax) ;
		ApplyPropagator(zphi,zprop,gPhsgrtRec->nbeam,gPhsgrtRec->mbeam,jmax,0,i);
	}	
}

void CalculateNewImage(float *exitwavefunction,TImage *calculatedImage) {
  	float *imageData,cs,div,elapsedTime ;
	long jmax ;
	float defocus,aperture ;
	 
	/* ----------------------------------------------- */
	/* Set the parameters for the lens and the display */
	/* ----------------------------------------------- */
	SetLensParameters(&gDisplayRec.cs,&gDisplayRec.div,&cs,&div,
		&gPhsgrtRec->gmax,&gPhsgrtRec->nbeam,&gPhsgrtRec->mbeam,&gPhsgrtRec->nz) ;
	
	jmax = gPhsgrtRec->nbeam*gPhsgrtRec->mbeam ;
	imageData = exitwavefunction + 4*jmax ;
		
	gDisplayRec.f1 = gDisplayRec.dfocus ;
	defocus = gDisplayRec.f1 ;
	aperture = gDisplayRec.aprtrd ;
			
	mmfft(exitwavefunction,gPhsgrtRec->nbeam,gPhsgrtRec->mbeam,&elapsedTime) ;  // Store  diffraction pattern 
	ApplyLens(exitwavefunction,imageData,cs,div,gPhsgrtRec->nbeam,gPhsgrtRec->mbeam,jmax) ;
	if (gAbortStatus) {		
		return ;
	}
	// Set the image intensity 
	SetImageIntensities(imageData,jmax) ;
	// Stuff the image values into the calculated image	'calculatedImage'
	SetImageData(calculatedImage,imageData,k4ByteReal) ;		
	FlipImage(calculatedImage) ; // so that everything is referred to in normal image coordinates
}

void ApplyLens(float zphi[],float ihk[],float cs,float theta,short nbeam,short mbeam,long jmax) {
  	float aprtsq,abga,testint,astsq,bstsq,test,th,tk,*rp1,*rp2,*tp1,*tp2,zrl,zim ;
	float hcapert,kcapert,hoaxis,koaxis,one_ov_jmax,*zfi,zrl1,zrl2,zim1,zim2 ;
	float hpr,kpr,hprh,kprk,u1,u2,u1sq,u2sq,u1frth,u2frth,arg1,arg2,h1,k1,h2,k2 ;
	long nh,h,nk,k,i,j,hmax,kmax,maxBeams,numBeamsInAperture ;
	short *ip1,*ip2,*ih,*ik ;
	float pi=3.14159265,c02,c04,cxa1,cxa2,cxd,xcoeff,elapsedTime,vib,vibcnst,usq ;
	float ux,uy,cosVib,sinVib ;
	Boolean isDirectionalVibration = false ;
	float	c2a,c3a,c2aAngle,c3aAngle,angle,arg ;
	
	aprtsq = gDisplayRec.aprtrd*gDisplayRec.aprtrd ;
	abga = -2.*gProjCell->ast*gProjCell->bst*gProjCell->csga ;
	c02 = pi*gMSRec.wavel*gDisplayRec.dfocus ;
	c04 = 0.5*pi*cs*gMSRec.wavel*gMSRec.wavel*gMSRec.wavel ;
	cxd = -0.5*pi*pi*gMSRec.wavel*gMSRec.wavel*gDisplayRec.del*gDisplayRec.del ;
	cxa1 = gDisplayRec.dfocus*pi*theta*0.77 ;
	cxa2 = cs*gMSRec.wavel*gMSRec.wavel*pi*theta*0.77 ;
	c2a = 0.5*pi*gMSRec.wavel*gDisplayRec.TwoFoldAstigMag ;					// 1/2LA2
	c3a = 2./3.*pi*gMSRec.wavel*gMSRec.wavel*gDisplayRec.ThreeFoldAstigMag ;	// 2/3LLA3 
	c2aAngle = gDisplayRec.TwoFoldAstigAngle*pi/180. ;
	c3aAngle = gDisplayRec.ThreeFoldAstigAngle*pi/180. ;
	
	testint = pow(10.,-dfpdspl.min_int) ;	 
	
	hmax = nbeam/2 ; 
	kmax = mbeam/2 ;
	maxBeams = hmax*kmax ;	/* Maximum # of beams that can be in the aperture */
	ih 	= (short *)NewPtr(sizeof(short)*maxBeams) ;
	if (ih == (short *)0) {
		PutMessage("Out of Memory") ;
		return ;
	}
	ik 	= (short *)NewPtr(sizeof(short)*maxBeams) ;
	if (ik == (short *)0) {
		PutMessage("Out of Memory") ;
		DisposePtr((Ptr)ih) ;
		return ;
	}
	zfi = (float *)NewPtr(sizeof(float)*2*maxBeams) ;
	if (zfi == (float *)0) {
		PutMessage("Out of Memory") ;
		DisposePtr((Ptr)ih) ;
		DisposePtr((Ptr)ik) ;
		return ;
	}

	astsq = gProjCell->ast*gProjCell->ast ;
	bstsq = gProjCell->bst*gProjCell->bst ;
	hcapert = gDisplayRec.hc ;
	kcapert = gDisplayRec.kc ;
	hoaxis = gDisplayRec.ho ;
	koaxis = gDisplayRec.ko ;
	one_ov_jmax = 1./jmax ;
	vibcnst = 0. ;
	if (gDisplayRec.vibrationX != gDisplayRec.vibrationY) {		
		isDirectionalVibration = true ;
		vibcnst = pi*pi ;
		cosVib = cos(gDisplayRec.vibrationAngle*pi/180.) ;
		sinVib = sin(gDisplayRec.vibrationAngle*pi/180.) ;
	}
	else {
		vibcnst = pi*pi*gDisplayRec.vibrationX*gDisplayRec.vibrationX ;
	}
	tp1 = &zrl ;
	tp2 = &zim ;
	
	// --------------------------------------------------------------------------- 
	// Fill the array zfi with intensities > testint that lies within the aperture
	// --------------------------------------------------------------------------- 
	rp2 = zfi ;
	ip1 = ih ;
	ip2 = ik ;
	numBeamsInAperture = 0 ;
	for(nh=0;nh<nbeam;nh++) {
		h = nh ;
		if (h>hmax) h -= nbeam ;
		th = h - hcapert ;
		for(nk=0;nk<mbeam;nk++) {
			k = nk ;
			if (k > kmax) k -= mbeam ;
			tk = k - kcapert ;
			test = th*th*astsq+tk*tk*bstsq+th*tk*abga ;
			if (test <= aprtsq) {
				rp1 = zphi + 2*(nk*nbeam+nh) ;
				*tp1 = *rp1*one_ov_jmax ;
				*tp2 = *(rp1+1)*one_ov_jmax ;
				if (((*tp1)*(*tp1)+(*tp2)*(*tp2)) > testint) {
					*rp2++ = *tp1 ;
					*rp2++ = *tp2 ;
					*ip1++ = h ;
					*ip2++ = k ;
					numBeamsInAperture++ ;
				}
			}
		}
	}
	
	/* Set the image intensities to zero */
	for(i=0;i<2*jmax;ihk[i++]=0.)
		;
		
	for(i=0;i<numBeamsInAperture;i++) {
		hpr = *(ih+i) ;
		kpr = *(ik+i) ;
		h1 = hpr - hoaxis ;
		k1 = kpr - koaxis ;
		u1sq = h1*h1*astsq + k1*k1*bstsq + h1*k1*abga ;
		u1frth = u1sq*u1sq ;
		arg1 = c02*u1sq + c04*u1frth ;
		if (gDisplayRec.TwoFoldAstigMag != 0. || gDisplayRec.ThreeFoldAstigMag != 0.) {
			if (h1 != 0. && k1 != 0.) {
				u1 = sqrt(u1sq) ;
				angle = acos(h1/(u1*gProjCell->a)) ;
				if (k1 < 0.) angle = -angle ;
				if (gDisplayRec.TwoFoldAstigMag != 0.) {
					arg = 2.*(angle-c2aAngle) ;
					arg1 += c2a*u1sq*cos(arg) ;	// + c2a * u-square
				}
				if (gDisplayRec.ThreeFoldAstigMag != 0.) {
					arg = 3.*(angle-c3aAngle) ;
					arg1 += c3a*u1sq*u1*cos(arg) ;	// + c3a * u-third
				}
			}
		}
		*tp1 = cos(arg1) ;
		*tp2 = -sin(arg1) ;
		rp1 = zfi + 2*i ;
		zrl1 = *rp1*(*tp1) - *(rp1+1)*(*tp2) ;
		zim1 = *rp1*(*tp2) + *(rp1+1)*(*tp1) ;
		for(j=0;j<numBeamsInAperture;j++) {
			hprh = *(ih+j) ;
			kprk = *(ik+j) ;
			h = (long) (hpr - hprh) ;
			k = (long) (kpr - kprk) ;
			if (h < hmax && h > -hmax && k < kmax && k > -kmax) {
				h2 = hprh-hoaxis ;
				k2 = kprk-koaxis ;
				u2sq = h2*h2*astsq + k2*k2*bstsq + h2*k2*abga ;
				u2frth = u2sq*u2sq ;
				arg2 = c02*u2sq + c04*u2frth ;
				if (gDisplayRec.TwoFoldAstigMag != 0. || gDisplayRec.ThreeFoldAstigMag != 0.) {
					if (h2 != 0. && k2 != 0.) {
						u2 = sqrt(u2sq) ;
						angle = acos(h2/(u2*gProjCell->a)) ;
						if (k2 < 0.) angle = -angle ;
						if (gDisplayRec.TwoFoldAstigMag != 0.) {
							arg = 2.*(angle-c2aAngle) ;
							arg2 += c2a*u2sq*cos(arg) ;	// + c2a * u-square
						}
						if (gDisplayRec.ThreeFoldAstigMag != 0.) {
							arg = 3.*(angle-c3aAngle) ;
							arg2 += c3a*u2sq*u2*cos(arg) ;	// + c3a * u-third
						}
					}
				}
				*tp1 = cos(arg2) ;
				*tp2 = -sin(arg2) ;
				rp1 = zfi + 2*j ;
				zrl2 = *rp1*(*tp1) - *(rp1+1)*(*tp2) ;
				zim2 = *rp1*(*tp2) + *(rp1+1)*(*tp1) ;
				nh = h ;
				nk = k ;
				if (nh < 0) nh += nbeam ;
				if (nk < 0) nk += mbeam ;
				zrl = zrl1*zrl2+zim1*zim2 ;
				zim = zim1*zrl2-zrl1*zim2 ;
				CrossTransferFunction(h1,k1,h2,k2,u1sq,u2sq,cxa1,cxa2,cxd,astsq,bstsq,abga,&xcoeff) ;
				rp2 = ihk + 2*(nk*nbeam+nh) ;
				if (vibcnst != 0.) {
					if (isDirectionalVibration) {
						ux = h*gProjCell->ast + k*gProjCell->bst*gProjCell->csga ;
						uy = k*gProjCell->bst*gProjCell->siga ;
						vib = vibcnst*(
							gDisplayRec.vibrationX*gDisplayRec.vibrationX*(ux*cosVib-uy*sinVib)*(ux*cosVib-uy*sinVib) +
							gDisplayRec.vibrationY*gDisplayRec.vibrationY*(ux*sinVib+uy*cosVib)*(ux*sinVib+uy*cosVib)
							) ;
						if (vib <= 9.2)
							xcoeff = xcoeff*exp(-vib) ;
						else
							xcoeff = 0. ;
					}
					else {
						usq = h*h*astsq + k*k*bstsq + h*k*abga ;
						vib = vibcnst*usq ;
						if (vib <= 9.2)
							xcoeff = xcoeff*exp(-vib) ;
						else
							xcoeff = 0. ;
					}
				}
				*rp2 += zrl*xcoeff ;
				*(rp2+1) += zim*xcoeff ; 
			}
		}
	}
	mmffti(ihk,nbeam,mbeam,&elapsedTime) ;
	DisposePtr((Ptr)ih) ;
	DisposePtr((Ptr)ik) ;
	DisposePtr((Ptr)zfi) ;
}

float RandomProbability() {
	long iseed = 3 ;
	float probability ;
	
	probability = ran0(&iseed) ;
	return probability ;
}

long CreateRandomAtomIndex(RefinementList *list) {
	long index ;
	float rndom ;
	long iseed = 3 ;
	
	rndom = (list->numGroups-1)*ran0(&iseed) ;	// In the range 0 - the number of atoms that can be moved
	index = near(rndom) ;						// The number of the group to move
	return index ;
}

float CreateRandomDisplacement(float maxAmplitude) {
	float displacement ;
	long iseed = 3;
	
		
	displacement = 2*ran0(&iseed) - 1. ;
	displacement *= maxAmplitude ; /* 'Random' number >=-ampl <= ampl */
	return displacement ;
}

float CreateRandomValue(float maxValue) {
	float value ;
	long iseed = 3;
	
		
	value = ran0(&iseed) ;
	value *= maxValue ; /* 'Random' number >=0  <= maxValue */
	return value ;
}

float ran0(long *idum) {
	static	float  rry,rrv[98] ;
	float	dum ;
	static	long iff=0 ;
	long	j ;
	
	if (*idum < 0 || iff == 0) {	// Initializing the 98 random numbers
		iff = 1 ;
		srand(*idum) ;
		*idum=1 ;
		for (j=0;j<98;j++) dum = rand() ;
		for (j=0;j<98;j++) rrv[j] = rand() ;
		rry = rand() ;
	}
	j = (long) (97.*rry/RAND_MAX) ;
	rry = rrv[j] ;
	rrv[j] = rand() ;
	return(rry/RAND_MAX) ;
}

void SaveRefinedStructure() {
	// ------------------------------------------------------------------ 
	// Put up the standard file dialog Box to determine the file name 
	// ------------------------------------------------------------------ 
	FSInfo 	theFInfo ;
	char 	filename[256],baseName[256] ;
	char	promptName[256] ;
	
	sprintf(theFInfo.mFileName,"Refined_%s",gStructureName) ;
	ChooseFile("Save refined structure as:","*",theFInfo) ;
	
	if (theFInfo.mReply) {
		// The file name also has the directory as part of the name
		char *cp ;
		if ((cp = strrchr(theFInfo.mFileName,'/')) == NULL)
			cp = theFInfo.mFileName ;
		else 
			cp++ ;
		strcpy(baseName,cp) ;
		strcpy(filename,baseName) ;
		// If there is a . in the name, we need to strip it
		if ((cp=strpbrk(filename,".")) != NULL) {
			*cp = '\0' ;
		}

		calculation.status = 1 ;	// Nothing done for this file
		SaveRefinedStructureFile(baseName) ;
		r_sngle_structure(baseName,"old") ;
		gCalculationIsChanged = true ;
		gPGisChanged = true ;
		read_fls(0) ;
	}
}

void SaveRefinedParameters() {
	// ------------------------------------------------------------------ 
	// Put up the standard file dialog Box to determine the file name 
	// ------------------------------------------------------------------ 
	FSInfo 	theFInfo ;
	char 	filename[256],baseName[256] ;
	char	promptName[256] ;
	
	sprintf(theFInfo.mFileName,"RefinedPar_%s",gStructureName) ;
	ChooseFile("Save refined parameters as:","*",theFInfo) ;
	
	if (theFInfo.mReply) {
		// The file name also has the directory as part of the name
		char *cp ;
		if ((cp = strrchr(theFInfo.mFileName,'/')) == NULL)
			cp = theFInfo.mFileName ;
		else 
			cp++ ;
		strcpy(baseName,cp) ;
		strcpy(filename,baseName) ;
		// If there is a . in the name, we need to strip it
		if ((cp=strpbrk(filename,".")) != NULL) {
			*cp = '\0' ;
		}
		strcat(filename,".at") ;

		calculation.status = 1 ;	// Nothing done for this file
		rewrite_file(filename) ;
		r_sngle_structure(baseName,"old") ;
		gCalculationIsChanged = true ;
		gPGisChanged = true ;
		read_fls(0) ;
	}

}

extern struct TranslationRec {
	float value ;
	char  *symbol ;
} translations[] ;
static char symb[] = {'x','y','z'} ;

void EncodeSymop(char *string,TMatrix *sym_op) {

	long i,j,k ;
	float test ;
	char *cp ;
	Boolean	first ;

	cp = string ;
	for (i=0;i<3;i++) {
		first = true ;
		for (j=0;j<3;j++) {
			test = sym_op->m[i][j] ;
			if (test != 0.) {
				if (first) {
					if (test < 0.) sprintf(cp++,"-") ;
						sprintf(cp++,"%c",symb[j]) ;
					first = false ;
				}
				else {
					if (test < 0.) 
						sprintf(cp++,"-%c",symb[j]) ;
					else
						sprintf(cp++,"+%c",symb[j]) ;
					cp++ ;
				}
			}
		}
		if (sym_op->m[i][3] != 0.) {
			j = 0 ;
			do {
				test = sym_op->m[i][3] - translations[j++].value ;
				test = (test >= 0. ? test : -test) ;
			}
			while ( test > 0.001 && j < 11) ;
			sym_op->m[i][3] = translations[--j].value ;
			sprintf(cp++,"+%s",translations[j].symbol) ;
			cp += 3 ;
		}
		if (i!=2) sprintf(cp++,",") ;
	}
	*cp = EOS ;

}

void SaveRefinedStructureFile(char *file_name) {
	FILE *fp ;
	char filename[40] ;
	long j ;
	
	// Convert to the correct filename
	strcpy(filename,file_name) ;
	strcat(filename,".at") ;
 

	fp = fopen(filename,"w") ;
	if (fp == nil) {
		PutMessage("Failed to create the structure file") ;
		return ;
	}
	fprintf(fp,"%s\n","Refined Structure") ;
	if (gUseSymmetryElements) {
		fprintf(fp,"%d\n",gSpacegroupNumber) ;
	}
	else {
		fprintf(fp,"1\n") ;
	}
	fprintf(fp,"%f %f %f %f %f %f\n",gProjCell->a,gProjCell->b,gProjCell->c
                                    ,gProjCell->al,gProjCell->be,gProjCell->ga) ;
	fprintf(fp,"%f\n",gPhsgrtRec->gmax) ; // GMAX 
	fprintf(fp,"0 0 1\n") ;
	if (gUseSymmetryElements) {
		fprintf(fp,"%d 1\n",gNumSymops) ;
		fprintf(fp,"%d %d\n",gBasis.numBasisAtoms,gPhsgrtRec->iatom) ;
		for (j=0;j<gBasis.numBasisAtoms;j++) {
			fprintf(fp,"%d %s %f %f %f %f %f\n",
		       	gBasis.atom[j].type,chemsymb[gBasis.atom[j].no-1],gBasis.atom[j].xc,
		       	gBasis.atom[j].yc,gBasis.atom[j].zc,gBasis.atom[j].dw,
		       	gBasis.atom[j].occ) ;
		}
	}
	else {
    		fprintf(fp,"1 1\n") ;
		fprintf(fp,"%d %d\n",gProjCell->natom,gPhsgrtRec->iatom) ;
		for (j=0;j<gProjCell->natom;j++) {
			fprintf(fp,"%d %s %f %f %f %f %f\n",
		       	gProjCell->ntype[j],chemsymb[gProjCell->atno[j]-1],gProjCell->x[j],
		       	gProjCell->y[j],gProjCell->z[j],gProjCell->dw[j],
		       	gProjCell->occ[j]) ;
		}
	}
	fprintf(fp,"%s %f %f %f\n",micname,gDisplayRec.cs,gDisplayRec.del,gDisplayRec.div) ;
	fprintf(fp,"%f\n",gMsliceRec.volt) ;
	fprintf(fp,"0.00000 0.00000\n") ;
   	fprintf(fp,"%f\n",gMsliceRec.thick) ;
	fprintf(fp,"NO\n") ;
   	fprintf(fp,"%f\n",gDisplayRec.f1) ;
	fprintf(fp,"%f\n %f %f %f %f\n",gDisplayRec.aprtrd,
            gDisplayRec.hc,gDisplayRec.kc,gDisplayRec.ho,gDisplayRec.ko) ;
	
	if (gUseSymmetryElements) {
		char symoptext[64] ;
		for (j=0;j<gNumSymops;j++) {
			EncodeSymop(symoptext,&gSymOps[j]) ;		
    			fprintf(fp,"%s\n",symoptext) ;
		}
   	}
   	else {
    		fprintf(fp,"%s\n",txtptr_symop[0]) ;
	}
	
	fprintf(fp,"%d\n",1) ;	// Nothing has been done
	fprintf(fp,"%f %f %f\n",gDisplayRec.vibrationX,gDisplayRec.vibrationY,
		gDisplayRec.vibrationAngle) ;
	
	fclose(fp) ;
}

void	GetBondValenceParameters() {
	TBondValenceDialog *theDialog = new TBondValenceDialog() ;
	theDialog->SetFormData((void *)&gBondStrengthData) ;
	if (theDialog->DoModal()) {
		theDialog->GetFormData((void *)&gBondStrengthData) ;
	}
	delete theDialog ;
}

float BondValenceSum(short indx,short index) {	 
	register long n ;
	float	x1,y1,z1,x2,y2,z2,dx,dy,dz,dxsq,dysq,dzsq,drsq,rsq,r,d ;
	float	x2pr,y2pr,z2pr ;
	short	mina,minb,minc,maxa,maxb,maxc,na,nb,nc ;
	float	bondValenceSum,maxBondLength ;
	
	
	// Calculate the bond valence sum for the atom "indx" associated with a bond "index"
	
	r = gBondStrengthData.bondStrength[index].maxBondLength ;
	rsq = r*r ;
	maxBondLength = r ;
		
	x1 =  	gProjCell->a*(*(gProjCell->x+indx)) ;
	y1 =  	gProjCell->b*(*(gProjCell->y+indx)) ;
	z1 =  	gProjCell->c*(*(gProjCell->z+indx)) ;

	mina = (x1 < maxBondLength ? -1 : 0) ;
	minb = (y1 < maxBondLength ? -1 : 0) ;
	minc = (z1 < maxBondLength ? -1 : 0) ;
	maxa = (gProjCell->a-x1 <  maxBondLength ? 1 : 0) ;
	maxb = (gProjCell->b-y1 <  maxBondLength ? 1 : 0) ;
	maxc = (gProjCell->c-z1 <  maxBondLength ? 1 : 0) ;
	
	bondValenceSum = 0. ;
	for (n = 0; n < gProjCell->natom;n++) {
		// Unit cell view
		if (n != indx) {
			x2 =  	gProjCell->a*(*(gProjCell->x+n)) ;
			y2 = 	gProjCell->b*(*(gProjCell->y+n)) ;
			z2 = 	gProjCell->c*(*(gProjCell->z+n)) ;
			for (nc = minc;nc <=maxc;nc++) {
				z2pr = z2 + nc*gProjCell->c ;
				dz = z2pr - z1 ;
				dzsq = dz*dz ;
				if (dzsq >= rsq)
					continue ;
				for (nb = minb; nb <= maxb; nb++) {
					y2pr = y2 + nb*gProjCell->b ;
					dy = y2pr - y1 ;
					dysq = dy*dy ;
					if (dysq >= rsq)
						continue ;
					for (na = mina; na <= maxa; na++) {
						x2pr = x2 + na*gProjCell->a ;
						dx = x2pr - x1 ;
						// See if we are within the min and max bondlengths to be
						// considered
						dxsq = dx*dx ;
						if (dxsq > rsq)
							continue ;
						drsq = dxsq + dysq + dzsq ;
						if (drsq <= rsq) { 
							// bond to be considered and added to bondsum
							d = sqrt(drsq) ;
							bondValenceSum += exp((gBondStrengthData.bondStrength[index].bFactor-d)/0.37) ;	
						}
					}
				}
			}
		
		}
	}
	if (bondValenceSum == 0.) {
		bondValenceSum = gBondStrengthData.bondStrength[index].valenceSum ;
	}	
	return bondValenceSum ;		
}

Boolean	NotInGroup(short index) {
	long j,k,ndx ;
	Boolean notInGroup = true ;
	for(j=0; j < gRefinementList.numGroups;j++) {
		for(k=0; k < (gRefinementList.groupPtr+j)->numInGroup ; k++) {
			ndx = *((gRefinementList.groupPtr+j)->atomIndex+k) ;// This is the index into gProjCell
			if (ndx == index) {
				notInGroup = false ;
				return notInGroup ;
			}
		}
	}
	return notInGroup ;
}

float AtomicFormFactor(float usq,long noz[],long j,float *dw,SFParamFitPtr paramFit) {
	// -----------------------------------------------------------
	// This subroutine calculates the electron scattering factors 
	// presently using coefficients by Doyle and Turner.         
	// -----------------------------------------------------------
	
	register float xrff,elff ;
	register long i ;
	float ssq,cnst=0.0239340,fdw,elfrmf ; 
	
	ssq = usq/4. ;
	fdw = *dw ;
	
	if (gUseElScattFactors) {
		elfrmf = 0. ;
		for(i=0;i<4;i++)
			elfrmf +=  paramFit->ax[j*4+i]*exp(-paramFit->bx[j*4+i]*ssq) ;
		elfrmf *= exp(-fdw*ssq) ;
		return(elfrmf) ;
	}
	else {
		if (ssq == 0.) {
			elff = 0. ;
			for(i=0;i<4;i++)
				elff =  elff + paramFit->ax[j*4+i]*paramFit->bx[j*4+i] ;
			elff += paramFit->cx[j] ;
			elfrmf = cnst*elff ;
		}
		else {
			xrff = 0. ;
			for(i=0;i<4;i++)
				xrff = xrff + paramFit->ax[j*4+i]*exp(-paramFit->bx[j*4+i]*ssq) ;
			xrff += paramFit->cx[j] ;
			elfrmf = cnst*(*(noz+j)-xrff)/ssq*exp(-fdw*ssq) ;
		}
		return(elfrmf) ;
	}	
}

void CrossTransferFunction(float h1,float k1,float h2,float k2,float u1sq,float u2sq,
	float cxa1,float cxa2,float cxd,float astsq,float bstsq,float abga,float *xcoeff) {
    float u,v,arg1,arg2 ;
	
    u = cxa1*(h1-h2) + cxa2*(u1sq*h1-u2sq*h2) ;
    v = cxa1*(k1-k2) + cxa2*(u1sq*k1-u2sq*k2) ;
    arg1 = cxd*(u1sq-u2sq)*(u1sq-u2sq) ;
    arg2 = u*u*astsq + v*v*bstsq + u*v*abga ;
    *xcoeff = exp(arg1-arg2) ;
}
