// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <math.h>
#include "TMatrix.h"

TMatrix::TMatrix() {
}

TMatrix::TMatrix(float diag) {
	short i,j ;
	
	for(i=0;i<4;i++)
		for(j=0;j<4;j++) {
			if (i == j)
				m[i][j] = diag ;
			else 
				m[i][j] = 0. ;
		}
}

TMatrix& TMatrix::operator=(const TMatrix& m1) {
	register long i,j ;
	
	for (j=0;j<4;j++)
		for(i=0;i<4;i++)
			m[i][j] = m1.m[i][j] ;
			
	return *this ;
}

TMatrix& 
TMatrix::operator=(const float value) {
	register long i,j ;
	
	for (j=0;j<4;j++) {
		for(i=0;i<4;i++) {
			if (i == j) 
				m[i][i] = value ;
			else
				m[i][j] = 0. ;
		}
	}			
	return *this ;
}

TMatrix operator*(const TMatrix& m1,const TMatrix& m2) {
	TMatrix temp ;
	register long i,j,k ;
	
	for(i=0;i<4;i++)
    	for(j=0;j<4;j++) {
    		temp.m[i][j] = 0. ;
    		for(k=0;k<4;k++) {
    			temp.m[i][j] += m1.m[i][k]*m2.m[k][j] ;
    		}
    	}
    	
    return temp ;
}

TMatrix operator+(const TMatrix& m1,const TMatrix& m2) {
	TMatrix temp ;
	register long i,j ;
	
	for(i=0;i<4;i++)
    	for(j=0;j<4;j++) {
    		temp.m[i][j] = m1.m[i][j] + m2.m[i][j] ;
    	}
    	
    return temp ;
}

TMatrix operator-(const TMatrix& m1,const TMatrix& m2) {
	TMatrix temp ;
	register long i,j ;
	
	for(i=0;i<4;i++)
    	for(j=0;j<4;j++) {
    		temp.m[i][j] = m1.m[i][j] - m2.m[i][j] ;
    	}
    	
    return temp ;
}

TVector operator*(const TMatrix& mat,const TVector& vec) {
	TVector tvec ;
	long i,j ;
	
    for(i=0;i<4;i++) {
    	tvec.a[i] = 0. ;
    	for(j=0;j<4;j++) {
    		tvec.a[i] += mat.m[i][j]*vec.a[j] ;
    	}
    }
    return tvec ;
}

void InvertMatrix(const TMatrix& m1,TMatrix& m2) {
	float det ;
    long i,j ;

    m2.m[0][0] = m1.m[1][1]* m1.m[2][2]- m1.m[1][2]* m1.m[2][1] ;
    m2.m[1][0] =  m1.m[1][2]* m1.m[2][0]- m1.m[1][0]* m1.m[2][2] ;
    m2.m[2][0] =  m1.m[1][0]* m1.m[2][1]- m1.m[1][1]* m1.m[2][0] ;
    m2.m[0][1] =  m1.m[2][1]* m1.m[0][2]- m1.m[0][1]* m1.m[2][2] ;
    m2.m[1][1] =  m1.m[0][0]* m1.m[2][2]- m1.m[2][0]* m1.m[0][2] ;
    m2.m[2][1] =  m1.m[2][0]* m1.m[0][1]- m1.m[0][0]* m1.m[2][1] ;
    m2.m[0][2] =  m1.m[0][1]* m1.m[1][2]- m1.m[1][1]* m1.m[0][2] ;
    m2.m[1][2] =  m1.m[1][0]* m1.m[0][2]- m1.m[0][0]* m1.m[1][2] ;
    m2.m[2][2] =  m1.m[0][0]* m1.m[1][1]- m1.m[1][0]* m1.m[0][1] ;

    det =  m1.m[0][0]*m2.m[0][0]+ m1.m[0][1]*m2.m[1][0]+ m1.m[0][2]*m2.m[2][0] ;

    for(i=0;i<3;i++)
    	for(j=0;j<3;j++) {
    		m2.m[i][j] = m2.m[i][j]/det ;
    	}

    for (i=0;i<3;i++) {
    	m2.m[i][3] = 0. ;
    	m2.m[3][i] = 0. ;
    }
    m2.m[3][3] = 1. ;
}

void TransposeMatrix(const TMatrix& m1,TMatrix& m2) {
	 long i,j ;

    for(i=0;i<4;i++) 
    	for(j=0;j<4;j++)
    		m2.m[i][j] = m1.m[j][i] ;
}
