// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <stdio.h>
#include "TMenu.h"
#include "globals.h"
#include "XDrivers.h"
#include "TPage.h"

extern	UInt32	gLighterGray ;

#define ON 1 

// -------------------------------
// TMenu
// -------------------------------

TMenu::TMenu(char *heading) {

	strcpy(mTitle,heading) ;
	mTitleWidth = StringWidth(heading)+10 ;
	mHeightPerItem = 20 ;
	mNumMenuItems = 0 ;
	mMenuItems = nil ;
	SetRect(&mWindowRect,0,0,0,0) ;
	mNextMenu = nil ;
	mPopUpWindow = nil ;
	mParentPane = nil ;
}

TMenu::~TMenu() {
	TMenuItem *menuItem,*nextMenuItem ;
	for( menuItem = mMenuItems; menuItem != nil;) {
    		nextMenuItem = menuItem->mNextItem ;
		delete menuItem ;
		menuItem = nextMenuItem ;
	}

	if (mPopUpWindow != nil) delete mPopUpWindow ;
}

void TMenu::Draw() {
	long i,x,y,numChars ;
	TMenuItem *menuItem ;
	extern Display *gDisplay ;

	// The menu draws into its own window, so the menu itself starts at	
	// 0,0 and the menuitems start with respect to this origin.	
	if (mPopUpWindow == nil) {
		long	width = mWindowRect.right - mWindowRect.left ;
		long	height = mWindowRect.bottom - mWindowRect.top ;
		mPopUpWindow = new TPopUpWindow(mParentPane,mX,mY,width,height) ;
		for( menuItem = mMenuItems; menuItem != nil;menuItem = menuItem->mNextItem) {
    			menuItem->mWindow = mPopUpWindow ;
		}
	}
	mPopUpWindow->Show() ;
	XFlush(gDisplay) ;
	x = 10 ;
	menuItem = mMenuItems ;
	for (i=0;i<mNumMenuItems;i++) {
		y = (i+1)*mHeightPerItem - 5 ;
		numChars = strlen(menuItem->mText) ;
		if (numChars > 0) {
			XDrawString(gDisplay,mPopUpWindow->GetDrawable(),
				mPopUpWindow->GetGC(),x,y,menuItem->mText,numChars) ;
		}
		menuItem = menuItem->mNextItem ;
	}
}

void TMenu::AddMenuItem(char *text) {
	long i ;
	TMenuItem *menuItem ;
	long width,lWidth ;
	Rect	rect ;

	lWidth = mWindowRect.right ;
	width = StringWidth(text) + 20 ;
	lWidth = (lWidth > width ? lWidth : width) ;
	mWindowRect.right = lWidth ;
	SetRect(&rect,0,mWindowRect.bottom,mWindowRect.right,
		mWindowRect.bottom+mHeightPerItem) ;
	mWindowRect.bottom += mHeightPerItem ;
	// Find the location of the previous menuitem and add a link to this
	if (mNumMenuItems == 0) {
		mMenuItems = new TMenuItem(rect,text) ;
		menuItem = mMenuItems ;
	}
	else {
		for(i=0,menuItem=mMenuItems;menuItem->mNextItem != 0;i++) {
			menuItem->AdjustWidth(lWidth) ;
			menuItem = menuItem->mNextItem ;
		}
		menuItem->mNextItem = new TMenuItem(rect,text) ;
	}
	if (mPopUpWindow != nil) menuItem->mWindow = mPopUpWindow ;
	mNumMenuItems++ ;
}

void TMenu::HighlightMenuItem(short itemNumber,Boolean state) {
	short i ;
	TMenuItem *menuItem ;
	// item numbers start at 1
	menuItem = mMenuItems ;
	for(i=0;i<itemNumber-1 && menuItem != nil ;i++) {
		menuItem = menuItem->mNextItem ;
	}
	if (menuItem != nil) menuItem->Draw(state) ;
}

void TMenu::SetMenuItemText(short itemnumber,char *text) {
    short i ;
    TMenuItem *menuItem ;

    /* Number starts at 1 */
    menuItem = mMenuItems ;
    for(i=0;i<itemnumber-1 && menuItem != nil;i++) {
    	menuItem = menuItem->mNextItem ;
    }
    if (menuItem != nil) strcpy(menuItem->mText,text) ;
}

void TMenu::GetMenuItemText(short itemnumber,char *text) {
    short i ;
    TMenuItem *menuItem ;

    /* Number starts at 1 */
    menuItem = mMenuItems ;
    for(i=0;i<itemnumber-1 && menuItem != nil;i++) {
    	menuItem = menuItem->mNextItem ;
    }
    if (menuItem != nil) strcpy(text,menuItem->mText) ;
}

// -------------------------------
// TPopupMenu
// -------------------------------

TPopupMenu::TPopupMenu(TPane *parent,short x1,short y1,short x2, short y2,char *text,
	char *label,LabelPosition pos) 
		: TMenu(text),TMenuButton(x1,y1,x2,y2,text,label,pos) {
	mParentPane = parent ;
	mSelectedItem = 1 ;	// The menu starts at 1
}

TPopupMenu::TPopupMenu(TPane *parent,Rect *rect,char *text,char *label,LabelPosition pos) 
		: TMenu(text),TMenuButton(rect,text,label,pos) {
	mParentPane = parent ;
	mSelectedItem = 1 ;	// The menu starts at 1
}

TPopupMenu::TPopupMenu(TPane *parent,Rect *rect,char *text[],short choice,char *label,
	LabelPosition pos) 
		: TMenu(text[choice-1]),TMenuButton(rect,text[choice-1],label,pos) {
	mParentPane = parent ;
	mSelectedItem = choice ;	// The menu starts at 1

	for(long i=0;text[i] != nil; i++) {
		AddMenuItem(text[i]) ;
	}
}

TPopupMenu::TPopupMenu(TPane *parent,short x1,short y1,short x2, short y2,char *text[],
	short choice,char *label,LabelPosition pos) 
		: TMenu(text[choice-1]),TMenuButton(x1,y1,x2,y2,text[choice-1],label,pos) {
	mParentPane = parent ;
	mSelectedItem = choice ;	// The menu starts at 1

	for(long i=0;text[i] != nil; i++) {
		AddMenuItem(text[i]) ;
	}
}

TPopupMenu::~TPopupMenu() {
}

Boolean TPopupMenu::DoMouseDown(Point *where) {

	if (!mIsActive) return false ;
	if (mState) return false ;

	mState = ON ;
	TMenuButton::Draw() ;

	Point	menuLocation ;
	Point	globalPosition ;
	GetGlobalMouse(&globalPosition) ;
	long dx = globalPosition.x - where->x ;
	long dy = globalPosition.y - where->y ;
	menuLocation.x = where->x ;
	menuLocation.y = (mRect.top+mRect.bottom)/2 - (2*mSelectedItem-1)*mHeightPerItem/2 ;
	menuLocation.x += dx ;
	menuLocation.y += dy ;
	short	selection = Select(&menuLocation) ;
	Boolean	status = false ;
	if (selection != mSelectedItem) status = true ;
	mState = OFF ;
	TMenuButton::Draw() ;
	return status ;
} 

short	TPopupMenu::Select(Point *where) {
	mX = where->x ;
	mY = where->y ;

	TMenuItem	*menuItem ;
    int 	rx,ry,wx,wy ;
	Rect 	tRect ;
	unsigned int	keyButt ;
	Window	child,root ;
	short	lastmenuitem,selectedItem,j ;
	Point	currentPt ;
	extern	Display	*gDisplay ;
        short   inNewMenuItem = 0 ;

	// The menu draws into its own window, so the menu itself starts at	
	// 0,0 and the menuitems start with respect to this origin.	
	
	if (mPopUpWindow == nil) {
		long	width = mWindowRect.right - mWindowRect.left ;
		long	height = mWindowRect.bottom - mWindowRect.top ;
		mPopUpWindow = new TPopUpWindow(mParentPane,
				mX,mY,width,height) ;
		mPopUpWindow->SetFgColor(gBlackPixel) ;
		for( menuItem = mMenuItems; menuItem != nil;menuItem = menuItem->mNextItem) {
    			menuItem->mWindow = mPopUpWindow ;
		}
	}
	XMoveWindow(gDisplay,mPopUpWindow->GetXWindow(),mX,mY) ;

	selectedItem	= mSelectedItem ;
	lastmenuitem	= mSelectedItem ;

	mPopUpWindow->Show() ;
	XFlush(gDisplay) ;

	DrawPopped() ;

	while (Button()) {
		XQueryPointer(gDisplay,mPopUpWindow->GetDrawable(),
				&root,&child,&rx,&ry,&wx,&wy,&keyButt) ;
		if (child == 0)	{	// we are in the menu
			currentPt.x = wx ;
			currentPt.y = wy ;
			menuItem = mMenuItems ;
			for (j=0;j<mNumMenuItems;j++) {
				tRect = menuItem->mRect ;
				if (PtInRect(&currentPt,&tRect)) {
					selectedItem = j+1 ;
					if (selectedItem != lastmenuitem) {
						if (lastmenuitem != 0) {
							/* Unhighlight old */
							HighlightMenuItem(lastmenuitem,false) ; 
						}
						lastmenuitem = selectedItem ;
						/* Highlight new */
						HighlightMenuItem(j+1,true) ;
					}
					break ;
				}
				menuItem = menuItem->mNextItem ;
			}
		}
	}
	if (mSelectedItem != lastmenuitem) {
		char text[256] ;
		GetMenuItemText(lastmenuitem,text) ;
		SetMenuText(text) ;
                inNewMenuItem = 1 ;
	}

	mSelectedItem = lastmenuitem ;
	mPopUpWindow->Hide() ;
        return inNewMenuItem ;
}

void	TPopupMenu::DrawPopped() {
	TMenuItem *menuItem ;

	menuItem = mMenuItems ;
	for (long i=0;i<mNumMenuItems;i++) {
		if (mSelectedItem == i+1)
			menuItem->Draw(true) ;
		else
			menuItem->Draw(false) ;
		menuItem = menuItem->mNextItem ;
	}
}


// -------------------------------
// TMenuItem
// -------------------------------
TMenuItem::TMenuItem(Rect &rect,char *text) {
	mRect = rect ;
	strcpy(mText,text) ;
	mNextItem = nil ;
}

TMenuItem::~TMenuItem() {
}

void TMenuItem::Draw(Boolean state) {
	Rect	tRect = mRect ;
	short	x = 10 ;
	short	y = mRect.bottom-5 ;

	if (state == ON) {
		mWindow->SetFgColor(gLighterGray) ;
		XFillRectangle(gDisplay,mWindow->GetDrawable(),mWindow->GetGC(),
			mRect.left,mRect.top,mRect.right-mRect.left,
			mRect.bottom-mRect.top) ;
		InsetRect(&tRect,1,1) ;
		mWindow->SetFgColor(gWhitePixel) ;
		XDrawLine(gDisplay,mWindow->GetDrawable(),
			mWindow->GetGC(),tRect.left,tRect.bottom,tRect.left,tRect.top) ;
		XDrawLine(gDisplay,mWindow->GetDrawable(),
			mWindow->GetGC(),tRect.left,tRect.top,tRect.right,tRect.top) ;
		InsetRect(&tRect,1,2) ;
		mWindow->SetFgColor(gDarkerGray) ;
		XDrawLine(gDisplay,mWindow->GetDrawable(),
			mWindow->GetGC(),tRect.right,tRect.top,tRect.right,tRect.bottom) ;
		XDrawLine(gDisplay,mWindow->GetDrawable(),
			mWindow->GetGC(),tRect.right,tRect.bottom,tRect.left,tRect.bottom) ;
		InsetRect(&tRect,-1,-1) ;
		mWindow->SetFgColor(gBlackPixel) ;
		XDrawLine(gDisplay,mWindow->GetDrawable(),
			mWindow->GetGC(),tRect.right,tRect.top,tRect.right,tRect.bottom) ;
		XDrawLine(gDisplay,mWindow->GetDrawable(),
			mWindow->GetGC(),tRect.right,tRect.bottom,tRect.left,tRect.bottom) ;
		XDrawString(gDisplay,mWindow->GetDrawable(),
			mWindow->GetGC(),x,y,mText,strlen(mText)) ;
	}
	else {
		XClearArea(gDisplay,mWindow->GetDrawable(),mRect.left,mRect.top,
			mRect.right-mRect.left,mRect.bottom-tRect.top,0) ;
		XDrawString(gDisplay,mWindow->GetDrawable(),
			mWindow->GetGC(),x,y,mText,strlen(mText)) ;
	}
}


