// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include "TMenuBar.h"
#include "globals.h"
#include "XDrivers.h"
#include <stdio.h>

TMenuBar::TMenuBar(TPane *pane,short width,short height) {
	mNumMenus = 0 ;
	mMenus = nil ;
	mWidth = width ;
	mHeight = height ;
	mWindowPane = pane ;
}

void	TMenuBar::AddMenu(TMenu *menu) {
	short	x ;
	TMenu *theMenu ;

	menu->mTitlePosition.y = mHeight-5 ;
	x = 0 ;
	menu->mY = mHeight+1 ;
	if (mNumMenus == 0) {
		menu->mTitlePosition.x = 5 ;
		menu->mX = x ;
		mMenus = menu ;
	}
	else {
		// We must link this menu to the ones we have
		for(theMenu = mMenus; theMenu->mNextMenu != nil; 
				theMenu = theMenu->mNextMenu) {
			x += theMenu->mTitleWidth ;
		}
		menu->mX = x ;
		menu->mTitlePosition.x = x + 5 ;
		theMenu->mNextMenu = menu ;
	}
	menu->SetParent(mWindowPane) ;
	mNumMenus++ ;			
}

void	TMenuBar::Draw() {
	long i ;
	TMenu *theMenu ;
	Rect tRect ;
	GrafPtr tPort ;

	GetPort(&tPort) ;
	SetPort(mWindowPane) ;
	SetFgColor(mBkColor) ;
	SetRect(&tRect,0,0,mWidth,mHeight) ;
	FillRect(&tRect) ;
	SetFgColor(mFgColor) ;

	theMenu = mMenus ;
	for (i=0;i<mNumMenus;i++) {
		DrawText(theMenu->mTitlePosition.x,theMenu->mTitlePosition.y,
			theMenu->mTitle) ;
		theMenu = theMenu->mNextMenu ;
	}
	SetPort(tPort) ;
}

void	TMenuBar::DoMouseDown(Point *where) {
	// Search through the list of sub-menus to find out in
	// which we had a mousedown	
	int 	i,x,y,rx,ry,wx,wy ;
	TMenu 	*theMenu,*lastMenu ;
	TMenuItem	*menuItem ;
	Rect 	tRect ;
	unsigned int	keyButt ;
	Window	child,root ;
	Boolean	menuIsDrawn,intheMenu,inMenuTitle,first ;
	short	lastmenu,lastmenuitem,selectedMenu,
    		selectedItem,j ;
	Point	currentPt ;
	GrafPtr	tPort ;
	extern	Display	*gDisplay ;

	selectedMenu	= -1 ;
	selectedItem	= -1 ;
	lastmenu		= -1 ;
	lastmenuitem	= -1 ;
	menuIsDrawn		= false ;
	intheMenu		= false ;
	inMenuTitle		= false ;
	first			= false ;

	GetPort(&tPort) ;
	SetPort(mWindowPane) ;
	theMenu = mMenus ;
	while (Button()) {
		if (XQueryPointer(gDisplay,mWindowPane->GetXWindow(),
				&root,&child,&rx,&ry,&wx,&wy,&keyButt)) {
			if (child != 0) {
				/* We are in a child */
				intheMenu = true ;
			}
			else {
				intheMenu = false ;
			}
		}
		currentPt.x = wx ;
		currentPt.y = wy ;
		SetRect(&tRect,0,0,mWidth,mHeight) ;
		if (PtInRect(&currentPt,&tRect)) {	/* Within the menubar	*/
			inMenuTitle = false ;
			theMenu = mMenus ;
			for (i=0;i<mNumMenus;i++) {
				x = theMenu->mX ;
				y = 0 ;
				SetRect(&tRect,x,y,x+theMenu->mTitleWidth,mHeight) ;
				if (PtInRect(&currentPt,&tRect)) { 
					inMenuTitle = true ;
					first = false ;
					selectedMenu = i ;
					if (lastmenu != selectedMenu) {
						lastmenuitem = -1 ;
						selectedItem = -1 ;
						intheMenu = false ;
						if (menuIsDrawn) {
							lastMenu->Hide() ;
    					    	//gUpdateBehindMenu(lastMenu) ;
    					    	// Unhighligh Old title */
    					    	HighlightMenuTitle(lastMenu,false) ;
						}
						lastmenu = i ;
						lastMenu = theMenu ;
						HighlightMenuTitle(theMenu,true) ;
						theMenu->Show() ;
						menuIsDrawn = true ;
						break ;
					}
					else { /* Same menu */
						if (intheMenu) {
							/* Unhighlight lastmenuitem */
							theMenu->HighlightMenuItem(lastmenuitem,false) ;
							selectedItem = -1 ;
							lastmenuitem = -1 ;
							intheMenu = false ;
						}
					}
					break ;
				}
				theMenu = theMenu->mNextMenu ;
			}
			if (!inMenuTitle && menuIsDrawn) {
				lastMenu->Hide() ;
    				/* Update the exposed region */
    				//gUpdateBehindMenu(lastMenu) ;
    				selectedMenu = -1 ;
    				selectedItem = -1 ;
    				lastmenu = -1 ;
    				lastmenuitem = -1 ;
    				menuIsDrawn = false ;
    			}
		}
		if (menuIsDrawn) {
			intheMenu = false ;
			XQueryPointer(gDisplay,theMenu->mPopUpWindow->GetDrawable(),
				&root,&child,&rx,&ry,&wx,&wy,&keyButt) ;
			currentPt.x = wx ;
			currentPt.y = wy ;
			menuItem = theMenu->mMenuItems ;
			for (j=0;j<theMenu->mNumMenuItems;j++) {
				tRect = menuItem->mRect ;
				if (PtInRect(&currentPt,&tRect)) {
					intheMenu = true ;
					selectedItem = j ;
					if (selectedItem != lastmenuitem) {
						if (lastmenuitem != -1) {
							/* Unhighlight old */
							theMenu->HighlightMenuItem(lastmenuitem,false) ; 
						}
						lastmenuitem = j ;
						/* Highlight new */
						theMenu->HighlightMenuItem(j,true) ;
					}
					break ;
				}
				menuItem = menuItem->mNextItem ;
			}
			if (!intheMenu && lastmenuitem != -1) {
				/* Unhighlight this item */
				theMenu->HighlightMenuItem(lastmenuitem,false) ; 
				selectedItem = -1 ;
				lastmenuitem = -1 ;
			}
		}
	}
	if (selectedItem != -1) {
		/* Flash the Item */
	}
	if (lastmenu != -1) {
		/* Hide the menu */
		lastMenu->Hide() ;
		/* Update the exposed region */
		//gUpdateBehindMenu(lastMenu) ;
		/* Unhighlight the title */
		HighlightMenuTitle(lastMenu,false) ;
	}
	where->x = selectedMenu ;
	where->y = selectedItem ;
	SetPort(tPort) ;
} 

void TMenuBar::HighlightMenuTitle(TMenu *menu,Boolean state) {
	extern Display *gDisplay ;
	extern uLong gBlackPixel,gWhitePixel ;
	Rect tRect ;
	uLong oldFgColor ;

	SetRect(&tRect,menu->mX,0,
		menu->mX+menu->mTitleWidth,mHeight) ;
	InsetRect(&tRect,1,2) ;
	oldFgColor = mWindowPane->GetFgColor() ;
	if (state) {
		mWindowPane->SetFgColor(gBlackPixel) ;
        	XFillRectangle(gDisplay,mWindowPane->GetXWindow(),mWindowPane->GetGC(),
        		tRect.left,tRect.top,tRect.right-tRect.left,
        		tRect.bottom-tRect.top) ;
		mWindowPane->SetFgColor(gWhitePixel) ;
		XDrawString(gDisplay,mWindowPane->GetXWindow(),
			mWindowPane->GetGC(),menu->mTitlePosition.x,menu->mTitlePosition.y,
    			menu->mTitle,strlen(menu->mTitle)) ;
	}
	else {
        	XClearArea(gDisplay,mWindowPane->GetXWindow(),tRect.left,tRect.top,
    			tRect.right-tRect.left,tRect.bottom-tRect.top,0) ;
		XDrawString(gDisplay,mWindowPane->GetXWindow(),mWindowPane->GetGC(),
    			menu->mTitlePosition.x,menu->mTitlePosition.y,
    			menu->mTitle,strlen(menu->mTitle)) ;
	}
	mWindowPane->SetFgColor(oldFgColor) ;
}

/*
void gUpdateBehindMenu(menu) 
	GMenu *menu ;
{
    XEvent 	event ;
    XRectangle 	clipRect,*rectangles,totalRect ;
    Rect	cRect,tRect ;
    extern	Display	*gDisplay ;
    extern	GrafPtr gWindow ;
    GrafPtr	tPort ;

    GetPort(&tPort) ;
    SetPort(menu->parent) ;
    XCheckMaskEvent(gDisplay,ExposureMask,&event) ;
    cRect.left = event.xexpose.x ;
    cRect.top = event.xexpose.y;
    cRect.right = cRect.left + event.xexpose.width ;
    cRect.bottom = cRect.top + event.xexpose.height ;
    tRect = cRect ; 
    while (XCheckMaskEvent(gDisplay,ExposureMask,&event)) {
	cRect.left = event.xexpose.x ;
	cRect.top = event.xexpose.y,
	cRect.right = cRect.left + event.xexpose.width ;
	cRect.bottom = cRect.top + event.xexpose.height ;
    	UnionRect(&tRect,&cRect,&tRect) ;
    }
    totalRect.x = cRect.left ;
    totalRect.y = cRect.top ;
    totalRect.width = cRect.right - cRect.left ;
    totalRect.height = cRect.bottom - cRect.top ;
    rectangles = &totalRect ;
    XSetClipRectangles(gDisplay,(menu->parent)->gc,0,0,
		rectangles,1,Unsorted) ;
    UpdateMyWindow(rectangles) ;
    clipRect.x = 0 ;
    clipRect.y = 0 ;
    clipRect.width = (menu->parent)->portRect.right - 
		(menu->parent)->portRect.left ;
    clipRect.height = (menu->parent)->portRect.bottom - 
		(menu->parent)->portRect.top ;
    rectangles = &clipRect ;
    XSetClipRectangles(gDisplay,(menu->parent)->gc,0,0,rectangles,
		1,Unsorted) ;
    SetPort(tPort) ;
}

Boolean InMenuBar(Point *pt,TMenuBar *menubar) {
	SetRect(&tRect,0,0,menubar->GetWIdth(),menubar->GetHeight()) ;
	if (PtInRect(pt,&tRect)) {
		return true ;
	}
	return false ;
}
*/
