// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include "TSetActiveElementsForm.h"
typedef struct	{
	short number ;
	short index[20] ;
} ElementRecord,*ElementPtr ;
extern	ElementPtr MyElements ;

extern	short	gNumActiveElements ;
extern	short	gActiveElementNumber[20] ;
extern	char	*chemsymb[] ;

#include "funcdefs.h"
TSetActiveElementsForm::TSetActiveElementsForm(FPVII callback) {

	FL_OBJECT *obj;
	mXFormsRec = (XFormsRecord *) fl_calloc(1, sizeof(XFormsRecord));
	
	FL_Coord	width	= 330 ;
	FL_Coord	height	= 225 ;

	mXFormsRec->mForm = fl_bgn_form(FL_NO_BOX, width, height);
	obj = fl_add_box(FL_UP_BOX,0,0,width,height,"");
	mOK = fl_add_button(FL_NORMAL_BUTTON,240,190,80,25,"OK");
	mCancel = fl_add_button(FL_NORMAL_BUTTON,130,190,90,25,"Cancel");
	mInactiveList = obj = fl_add_browser(FL_HOLD_BROWSER,10,40,100,130,"Inactive");
	fl_set_object_lalign(obj,FL_ALIGN_TOP);
	mActiveList = obj = fl_add_browser(FL_HOLD_BROWSER,220,40,100,130,"Active");
	fl_set_object_lalign(obj,FL_ALIGN_TOP);
	mMakeActive = fl_add_button(FL_NORMAL_BUTTON,120,80,90,20,"Make Active");
	mMakeInactive = fl_add_button(FL_NORMAL_BUTTON,120,50,90,20,"Make Inactive");
	obj = fl_add_text(FL_NORMAL_TEXT,50,5,260,20,
		"Active elements for the refinement procedure");
	fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);
	fl_end_form();

	fl_set_browser_fontstyle(mActiveList,FL_BOLD_STYLE) ;
	fl_set_browser_fontstyle(mInactiveList,FL_BOLD_STYLE) ;

	DeactivateButton(mMakeActive) ;
	DeactivateButton(mMakeInactive) ;

	// Create the call back
	fl_set_object_callback(mOK,TSetActiveElementsForm::DoClick,(long)this) ;
	fl_set_object_callback(mCancel,TSetActiveElementsForm::DoClick,(long)this) ;
	fl_set_object_callback(mInactiveList,TSetActiveElementsForm::DoClick,(long)this) ;
	fl_set_object_callback(mActiveList,TSetActiveElementsForm::DoClick,(long)this) ;
	fl_set_object_callback(mMakeActive,TSetActiveElementsForm::DoClick,(long)this) ;
	fl_set_object_callback(mMakeInactive,TSetActiveElementsForm::DoClick,(long)this) ;

	mXFormsRec->mForm->fdui = mXFormsRec;
	mCallback	= callback ;

	fl_set_clipping(0,0,width,height) ;
	Initialize() ;

}

void	TSetActiveElementsForm::Initialize() {
 	short 		i,j,ndx,ndx2 ;
	Boolean		isActive,hasInactiveElement ;
	hasInactiveElement = false ;

	fl_freeze_form(mXFormsRec->mForm);
	fl_clear_browser(mInactiveList);
	fl_clear_browser(mActiveList);

	for (i=0;i<MyElements->number;i++) {
		isActive = false ;
		ndx = MyElements->index[i] ;
		ndx++ ;
		for (j=0;j<gNumActiveElements;j++) {
			ndx2 = gActiveElementNumber[j] ;
			if (ndx2 == ndx) {
				isActive = true ;
				break ;
			}
		}
		if (!isActive) {			
			fl_add_browser_line(mInactiveList, chemsymb[ndx-1]) ;
			hasInactiveElement = true ; 
		}
	}
	
	// Set any elements that are Active
	for (i=0;i<gNumActiveElements;i++) {
		ndx = gActiveElementNumber[i] ;
		fl_add_browser_line(mActiveList, chemsymb[ndx-1]) ;
	}
	fl_unfreeze_form(mXFormsRec->mForm);
}

void 	TSetActiveElementsForm::Show() {
	fl_show_form(mXFormsRec->mForm,FL_PLACE_MOUSE,FL_FULLBORDER,"Set Active Elements") ;
	fl_redraw_form(mXFormsRec->mForm) ;
}

void	TSetActiveElementsForm::GetActiveElements() {
	gNumActiveElements = fl_get_browser_maxline(mActiveList);
	for (long i=1; i <= gNumActiveElements; i++) {
		gActiveElementNumber[i-1] = atomno((char *)fl_get_browser_line(mActiveList,i)) ;
	}
}

void 	TSetActiveElementsForm::DoClick(FL_OBJECT *obj,long arg) {
	long	lineSelected ;
	TSetActiveElementsForm	*theForm = (TSetActiveElementsForm *)arg ;

	if (obj == theForm->mOK) {
		(*theForm->mCallback)(arg,1) ;
		return ;
	}
	if (obj == theForm->mCancel) {
		(*theForm->mCallback)(arg,0) ;
		return ;
	}
	if (obj == theForm->mMakeActive) {
		lineSelected = fl_get_browser(theForm->mInactiveList) ;
		if (lineSelected > 0) {
			fl_add_browser_line(theForm->mActiveList,
				fl_get_browser_line(theForm->mInactiveList,lineSelected)) ;	
			fl_delete_browser_line(theForm->mInactiveList,lineSelected) ;
			DeactivateButton(theForm->mMakeActive) ;
		}
		return ;
	}
	if (obj == theForm->mMakeInactive) {
		lineSelected = fl_get_browser(theForm->mActiveList) ;
		if (lineSelected > 0) {
			fl_add_browser_line(theForm->mInactiveList,
				fl_get_browser_line(theForm->mActiveList,lineSelected)) ;		
			fl_delete_browser_line(theForm->mActiveList,lineSelected) ;
			DeactivateButton(theForm->mMakeInactive) ;
		}
		return ;
	}
	if (obj == theForm->mInactiveList) {
		lineSelected = fl_get_browser(theForm->mInactiveList) ;
		if (lineSelected > 0) {
			ActivateButton(theForm->mMakeActive) ;

		}
		return ;
	}
	if (obj == theForm->mActiveList) {
		lineSelected = fl_get_browser(theForm->mActiveList) ;
		if (lineSelected > 0) {
			ActivateButton(theForm->mMakeInactive) ;
		}
		return ;
	}
}

TSetActiveElementsForm::~TSetActiveElementsForm() {
	fl_hide_form(mXFormsRec->mForm) ;
	fl_free_form(mXFormsRec->mForm) ;
}

void	ActivateButton(FL_OBJECT *button) {
	fl_activate_object(button) ;
	fl_set_object_lcol(button,FL_BLACK) ;
}

void	DeactivateButton(FL_OBJECT *button) {
	fl_deactivate_object(button) ;
	fl_set_object_lcol(button,FL_INACTIVE_COL) ;
}
