// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include "TWindowList.h"
#include "TWindow.h"
#include "globals.h"
#include "XDrivers.h"
#include "funcdefs.h"
typedef enum { PSEUDOCOLOR, TRUECOLOR } MonitorDisplayType ;
extern MonitorDisplayType  gMonitorDisplayType ;
extern Visual	*gVisual ;

void	SetModifierKeyDown(XEvent *event) ;
void	SetModifierKeyUp(XEvent *event) ;
void	RemoveWindow(TWindow *) ;

extern	Atom	gDeleteWindowAtom ;

GC	gGC = nil ;
Pixmap	gStipple = nil ;
unsigned char	gStippleData[] =
				{	0x3F,0xF3,0xFF,0xFF,
					0xFF,0xFF,0xFF,0xFF,
					0xFF,0xFF,0xFF,0xFF,
					0x00,0x00,0xFF,0xFF } ;

TGrafPort::TGrafPort() {
	mClipRegion = NewRgn() ;
}

TGrafPort::TGrafPort(short width,short height,short depth) {
	SetRect(&mPortRect,0,0,width,height) ;
	mDepth = depth ;
	mClipRegion = NewRgn() ;
	RectRgn(mClipRegion,&mPortRect) ;
}

TGrafPort::TGrafPort(FL_OBJECT *object) {
	SetRect(&mPortRect,0,0,object->w,object->h) ;
	OffsetRect(&mPortRect,object->x,object->y) ;
	mDepth = fl_get_canvas_depth(object) ;
	mDrawable = FL_ObjWin(object) ;
	mGC = fl_state[fl_get_vclass()].gc[0] ;

	mClipRegion = NewRgn() ;
	RectRgn(mClipRegion,&mPortRect) ;

	// If this is a forms window, we don't want the closing of the
	// window through the window manager to close down the program
	XSetWMProtocols(gDisplay,mDrawable,&gDeleteWindowAtom,1) ;
}

void TGrafPort::SetClipRegion(RgnHandle rgn) {
	DisposeRgn(mClipRegion) ;
	mClipRegion = NewRgn() ;
	Rect	theRect ;
	RegionRect(rgn,&theRect) ;
	RectRgn(mClipRegion,&theRect) ;
}

void TGrafPort::SetClipRegion(Rect &rect) {
	DisposeRgn(mClipRegion) ;
	mClipRegion = NewRgn() ;
	RectRgn(mClipRegion,&rect) ;
}

TGrafPort::~TGrafPort() {
	DisposeRgn(mClipRegion) ;
}

void TGrafPort::SetBkColor(uLong color) { 
	mBkColor = color ; 
	XSetBackground(gDisplay,mGC,color) ;
} 

void TGrafPort::SetFgColor(uLong color) { 
	mFgColor = color ; 
	XSetForeground(gDisplay,mGC,color) ;
} 

TPopUpWindow::TPopUpWindow(TPane *parent,short x,short y,short width,short height)
		: TGrafPort(width,height,parent->GetDepth()){

	XSetWindowAttributes attr ;
	mParentPane = parent ;

	//mXWindow = XCreateSimpleWindow(gDisplay,parent->GetXWindow(),x,y,
	//	width,height,1,parent->GetFgColor(),parent->GetBkColor()) ;
	// mParentIsRoot = false ; 

	mXWindow = XCreateSimpleWindow(gDisplay,RootWindow(gDisplay,gScreen),x,y,
		width,height,1,parent->GetFgColor(),parent->GetBkColor()) ;
	mParentIsRoot = true ; 
	attr.override_redirect = true ;


	XChangeWindowAttributes(gDisplay,mXWindow,CWOverrideRedirect,&attr) ;
	mDrawable = mXWindow ;
	mGC = parent->GetGC() ;
}

TPopUpWindow::~TPopUpWindow() {
	if (mParentIsRoot == true) 
		XDestroyWindow(gDisplay,mXWindow) ; 
}

TPane::TPane(TPane *parent,short xloc,short yloc,short width,short height) 
 : TGrafPort(width,height) {
	mParentPane = parent ;
	if (parent != nil) {
		mXWindow = XCreateSimpleWindow(gDisplay,parent->GetXWindow(),xloc,yloc,
			width,height,1,
			parent->GetFgColor(),parent->GetBkColor()) ;
		mDrawable = mXWindow ;

		XGCValues	gcv ;
		gcv.foreground = parent->GetFgColor() ;
		gcv.background = parent->GetBkColor() ;
		mGC = XCreateGC(gDisplay,mXWindow,GCForeground |
    			GCBackground, &gcv) ;
	}
	// If parent == nil, we are instancing a TWindow class
}


TPane::TPane(TPane *parent,Rect *rect) 
 : TGrafPort(rect->right-rect->left,rect->bottom-rect->top,parent->GetDepth()) {
	mParentPane = parent ;
	if (parent != nil) {
		mXWindow = XCreateSimpleWindow(gDisplay,parent->GetXWindow(),rect->left,rect->top,
			rect->right-rect->left,rect->bottom-rect->top,1,
			parent->GetFgColor(),parent->GetBkColor()) ;
		mDrawable = mXWindow ;

		XGCValues	gcv ;
		gcv.foreground = parent->GetFgColor() ;
		gcv.background = parent->GetBkColor() ;
		mGC = XCreateGC(gDisplay,mXWindow,GCForeground |
    			GCBackground, &gcv) ;
	}
	// If parent == nil, we are instancing a TWindow class
}

TPane::~TPane() {
	XDestroySubwindows(gDisplay,mXWindow) ;
	XUnmapWindow(gDisplay,mXWindow) ;
	XDestroyWindow(gDisplay,mXWindow) ;
	XFreeGC(gDisplay,mGC) ;
}

void	TWindow::SetWindowTitle(char *title) {
	XStoreName(gDisplay,mXWindow,title) ;
	strcpy(mWindowTitle,title) ;
}

void	TWindow::GetWindowTitle(char *title) {
	strcpy(title,mWindowTitle) ;
}

TWindow::TWindow(short xloc,short yloc,short width,short height,short depth,char *title,
		Boolean hasOffscreen) : TPane(nil,xloc,yloc,width,height) {

	long	winX,winY ;
	long	borderWidth ;
    
	XSizeHints sizeHints ;
	XGCValues	gcv ;

	mIsValid = false ;
	mUpdateRegion	= XCreateRegion() ;

	mOffscreen = nil ;

	borderWidth = 4 ;
	winX = xloc ;
	winY = yloc ;
    
 	switch (gMonitorDisplayType) { 
		case PSEUDOCOLOR :
		    mXWindow = XCreateWindow(gDisplay,RootWindow(gDisplay,gScreen),
		    				winX,winY,width,height,
		    				borderWidth,depth,InputOutput,
		    				gVisualInfo.visual,
		    				CWBackPixel | CWBorderPixel | CWColormap,
		    				&gImageAttributes) ;
		    break ;
		case TRUECOLOR :
		    mXWindow = XCreateWindow(gDisplay,RootWindow(gDisplay,gScreen),
		    				winX,winY,width,height,
		    				borderWidth,depth,InputOutput,
		    				gVisualInfo.visual,
		    				CWBackPixel | CWBorderPixel,
		    				&gImageAttributes) ;
		    break ;
	}

	mDrawable	= mXWindow ;

	sizeHints.flags 	= USPosition | USSize | PMinSize | PMaxSize ;
	sizeHints.x	= winX ;
	sizeHints.y	= winY ;
	sizeHints.width	= width ;
	sizeHints.height	= height ;
	sizeHints.min_width 	= width ;
	sizeHints.min_height	= height ;
	sizeHints.max_width 	= width ;
	sizeHints.max_height	= height ;
	XSetStandardProperties(gDisplay,mXWindow,title,0,
    			0,0,0,&sizeHints) ;

	XSetWMProtocols(gDisplay,mXWindow,&gDeleteWindowAtom,1) ;
	strcpy(mWindowTitle,title) ;

	if (gMonitorDisplayType == PSEUDOCOLOR) {
	    // If the color map exists, set it up
	    XSetWindowColormap(gDisplay,mXWindow,gColorMap) ;
	    gcv.foreground = gBlackPixel ;
	    gcv.background = gWhitePixel ;
	}
	else {
	    XSetWindowColormap(gDisplay,mXWindow,gColorMap) ;
	    gcv.foreground = BlackPixel(gDisplay,gScreen) ;
	    gcv.background = WhitePixel(gDisplay,gScreen) ;
	}


	mFgColor = gBlackPixel ;
	mBkColor = gWhitePixel ;
	mDepth = depth ;
    
	mGC = XCreateGC(gDisplay,mXWindow,GCForeground |
    			GCBackground, &gcv) ;
    
	XSetFont(gDisplay,mGC,gFontStructure->fid) ;
	XSelectInput(gDisplay,mXWindow,ExposureMask | KeyPressMask | StructureNotifyMask |
    			KeyReleaseMask | ButtonPressMask | ButtonReleaseMask) ;
	if (hasOffscreen) {
		if ((mOffscreen = new TOffscreen(this)) == nil) {
			printf("\n Unable to set up an offscreen environment") ;
			exit(-1) ;
		}
	}

	/*
	Atom	*atomList ;
	long 	num_prop ;
	atomList	= XListProperties(gDisplay,mXWindow,&num_prop) ;
	for(long prop=0; prop < num_prop; prop++) {
		fprintf(stderr,"\nAtom[%d] = %s",
			prop,XGetAtomName(gDisplay,atomList[prop])) ;
	}
	*/

	XMapWindow(gDisplay,mXWindow) ;

	if (gGC == nil) {
		gGC = mGC ;
	}

	if (gStipple == nil) {
		unsigned int rwidth,rheight ;
		XQueryBestStipple(gDisplay,mDrawable,4,4,&rwidth,&rheight) ;
		gStipple = XCreatePixmapFromBitmapData(gDisplay,mDrawable,
			(char *)gStippleData,rwidth,rwidth,
			1,0,1) ;
	}
	if (gStipple != nil) XSetStipple(gDisplay,mGC,gStipple) ;

}

TWindow::~TWindow() {

	XDestroyRegion(mUpdateRegion) ;
	if (mOffscreen != nil) delete	mOffscreen ;

}


void TWindow::Erase(Rect *rect) {
	Rect theRect ;
	
	theRect = (rect != nil ? *rect : mPortRect) ;
	EraseRect(&theRect) ;
}


void TWindow::DoIdle() {
}

void TWindow::DoClientMessage(XClientMessageEvent *theMessage) {
	// Check to see what kind of message this is
	if (theMessage->type == XA_WINDOW) {
		if (theMessage->data.l[0] == gDeleteWindowAtom) {
			DoWMRequestDelete() ;
		}
	}
}

void TWindow::Close() {
	if (mWindowList != nil) {
		mWindowList->RemoveWindow(this) ;	// We are removed by the owner
	}
	else {
		delete this ;
	}
}

void TWindow::DoMouseDown(XEvent *event) {
	Window 	focusWindow ;
	int	revert_to ;
	XGetInputFocus(gDisplay,&focusWindow,&revert_to) ;
	if (focusWindow != mXWindow)
		XSetInputFocus(gDisplay,mXWindow,RevertToParent,CurrentTime) ;

	TPane::DoMouseDown(event) ;	// to set the time
}

void TWindow::Show() {
	XMapRaised(gDisplay,mXWindow) ;
}

void TWindow::Update(XEvent *event) {
	
	GrafPtr tPort ;
	XRectangle updateRect,clipRect ;
	GetPort(&tPort) ;

	updateRect.x = event->xexpose.x ;
	updateRect.y = event->xexpose.y ;
	updateRect.width = event->xexpose.width ;
	updateRect.height = event->xexpose.height ;
	XUnionRectWithRegion(&updateRect,mUpdateRegion,mUpdateRegion) ;
	if (event->xexpose.count == 0) {
		SetPort(this) ;
		XSetRegion(gDisplay,mGC,mUpdateRegion) ;
		Draw() ;
		clipRect.x = mPortRect.left ;
		clipRect.y = mPortRect.top ;
		clipRect.width = mPortRect.right ;
		clipRect.height = mPortRect.bottom ;
		XSetClipRectangles(gDisplay,mGC,0,0,
				&clipRect,1,Unsorted) ;
		SetPort(tPort) ;
		XDestroyRegion(mUpdateRegion) ;
		mUpdateRegion = XCreateRegion() ;
	}
}

void TWindow::Resize(XEvent *event) {
}

void TWindow::DoKeyDown(XEvent *event) {
	SetModifierKeyDown(event) ;
}

void TWindow::DoKeyUp(XEvent *event) {
	SetModifierKeyUp(event) ;
}

void TWindow::SelectCursor() {
}

void TWindow::Draw() {
	if (mOffscreen != nil) {
		Rect theRect = mPortRect ;
		CopyBits(mOffscreen,this,&theRect,&theRect,srcCopy) ;
	}
	else {	
	}
}


TFrameWindow::TFrameWindow(short xloc,short yloc,short width,short height,
			short depth,char *title) 
		: TWindow(xloc,yloc,width,height,depth,title,false) {
	mPaneList = nil ;
}

TFrameWindow::~TFrameWindow() {
	if (mPaneList != nil) delete mPaneList ;
}

void	TFrameWindow::AddPane(TPane *pane) {
	if (mPaneList == nil) mPaneList = new TPaneList ;
	mPaneList->AddPane(pane) ;
}

void	TFrameWindow::DeletePane(TPane *pane) {
	mPaneList->RemovePane(pane) ;
}
 
void	TFrameWindow::Draw() {
}

void	TFrameWindow::DoMouseDown(XEvent *event) {
}


// --------- TOffscreen
TOffscreen::TOffscreen(TWindow *window,Rect *rect) : 
	TGrafPort(window->GetPortRect().right-window->GetPortRect().left,
		window->GetPortRect().bottom-window->GetPortRect().top) {

	XGCValues	gcv ;
	long		width,height ;
	Rect	thePortRect ;

	if (rect == nil) {
		thePortRect = window->GetPortRect() ;
	}
	else {
		thePortRect = *rect ;
	}

	width = thePortRect.right - thePortRect.left ;
	height = thePortRect.bottom - thePortRect.top ;

	mGC = nil ;
	mPixmap = nil ;

	SetRect(&mPortRect,0,0,width,height) ;
	if ((mPixmap = XCreatePixmap(gDisplay,window->GetXWindow(),
		width,height,window->GetDepth())) == 0) {
		return ;
	}
	
	mDrawable	= mPixmap ;

	if (gMonitorDisplayType == PSEUDOCOLOR) {
		gcv.foreground = gBlackPixel ;
		gcv.background = gWhitePixel ;
	}
	else {
		gcv.foreground = BlackPixel(gDisplay,gScreen) ;
		gcv.background = WhitePixel(gDisplay,gScreen) ;
	}
	mFgColor = gBlackPixel ;
	mBkColor = gWhitePixel ;
	mDepth = window->GetDepth() ;
	mGC = XCreateGC(gDisplay,mPixmap,GCForeground |
    			GCBackground, &gcv) ;

	if (mGC != nil) {
		XCopyGC(gDisplay,window->GetGC(),GCFont,mGC) ;
		XSetForeground(gDisplay,mGC,gWhitePixel) ;
    
		XFillRectangle(gDisplay,mPixmap,mGC,0,0,
    			width,height) ;
		XSetForeground(gDisplay,mGC,gBlackPixel) ;
	}
}

TOffscreen::~TOffscreen(){
	if (mPixmap != nil) XFreePixmap(gDisplay,mPixmap) ;
	if (mGC != nil) XFreeGC(gDisplay,mGC) ;
}

// --------- TStOffscreen
TStOffscreen::TStOffscreen(TWindow *window,Rect *rect) : 
	TGrafPort(window->GetPortRect().right-window->GetPortRect().left,
		window->GetPortRect().bottom-window->GetPortRect().top) {

	XGCValues	gcv ;
	long		width,height ;
	Rect	thePortRect ;

	if (rect == nil)
		thePortRect = window->GetPortRect() ;
	else
		thePortRect = *rect ;

	width = thePortRect.right - thePortRect.left ;
	height = thePortRect.bottom - thePortRect.top ;

	SetRect(&mPortRect,0,0,width,height) ;
	mRect = thePortRect ;

	mGC = nil ;
	mPixmap = nil ;

	if ((mPixmap = XCreatePixmap(gDisplay,window->GetXWindow(),
		width,height,window->GetDepth())) == 0) {
		return ;
	}
	

	mDrawable	= mPixmap ;
	GetPort(&mCurrentPort) ;
	SetPort(this) ;
	if (gMonitorDisplayType == PSEUDOCOLOR) {
		gcv.foreground = gBlackPixel ;
		gcv.background = gWhitePixel ;
	}
	else {
		gcv.foreground = BlackPixel(gDisplay,gScreen) ;
		gcv.background = WhitePixel(gDisplay,gScreen) ;
	}
	mFgColor = gBlackPixel ;
	mBkColor = gWhitePixel ;
	mDepth = window->GetDepth() ;
	mGC = XCreateGC(gDisplay,mPixmap,GCForeground |
    			GCBackground, &gcv) ;

	if (mGC != nil) {
		mHasLocalGC = true ;
		XCopyGC(gDisplay,window->GetGC(),GCFont,mGC) ;
		XSetForeground(gDisplay,mGC,gWhitePixel) ;
    
		XFillRectangle(gDisplay,mPixmap,mGC,0,0,
    			width,height) ;
		XSetForeground(gDisplay,mGC,gBlackPixel) ;
	}
	else {
		mHasLocalGC = false ;
		mGC = window->GetGC() ;
	}
}

TStOffscreen::~TStOffscreen(){
	if (mPixmap != nil) {
		SetPort(mCurrentPort) ;
		SetClip(mRect) ;
		XCopyArea(gDisplay,mPixmap,mCurrentPort->GetDrawable(),
			mCurrentPort->GetGC(),0,
    			0,mRect.right-mRect.left,
    			mRect.bottom-mRect.top,mRect.left,mRect.top) ;
	
		XFreePixmap(gDisplay,mPixmap) ;
		if (mGC != nil && mHasLocalGC) XFreeGC(gDisplay,mGC) ;
		SetClip(mCurrentPort->GetPortRect()) ;
	}
}
