// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include "header.h"
#include "funcdefs.h"
#include "menu_index.h"
extern float CNTRST,BRITE ;
extern char stext[] ;
static long xbl,xbh,ybl,ybh,xcl,xch,ycl,ych ;

// ---------------------------------------------------------------
// function to adjust contrast and brightness       
// ---------------------------------------------------------------
void r_adj_cntrst() {

	xbl = 100 ; ybl = 100 ; xbh = 700 ; ybh = 135 ;
	xcl = 920 ; ycl = 200 ; xch = 1000 ; ych = 650 ;
	set_menu(SET_BR_CNTRST) ;
	CLEAR_DISPLAY ;
	draw_cntrst_curve() ;
	draw_brite_bar(xbl,ybl,xbh,ybh) ;
	lbl_brite_bar(xbl,ybl,xbh,ybh) ;
	draw_contr_bar(xcl,ycl,xch,ych) ;
	lbl_contr_bar(xcl,ycl,xch,ych) ;
	displ_menu() ;
}

// ---------------------------------------------------------------
// Label the bar-graph showing the brightness level            
// ---------------------------------------------------------------
void lbl_brite_bar(long xl,long yl,long xh,long yh) {

    temdtx(xh-135,yl-25,"BRIGHTNESS") ;
    temdtx(xl,yh+10,"0") ;
    temdtx((xh+xl)/2-15,yh+10,"0.5") ;
    temdtx(xh,yh+10,"1") ;
}

// ---------------------------------------------------------------
// Label the bar-graph showing the contrast level            
// ---------------------------------------------------------------
void lbl_contr_bar(long xl,long yl,long xh,long yh) {

    temdtx(xl-50,yh+50,"CONTRAST") ;
    temdtx(xl-40,yl,"-4") ;
    temdtx(xl-40,(yh+yl)/2,"0") ;
    temdtx(xl-40,yh," 4") ;
}

// ---------------------------------------------------------------
// Draw the brightness bar-graph                      
// ---------------------------------------------------------------
void draw_brite_bar(long xl,long yl,long xh,long yh){

    long xb ;
    xb = (xh-xl)*BRITE + xl ;
    boks(xl,yl,xh,yh) ;
    fill_area(xl,yl,xb,yh) ;
    sprintf(stext,"%.2f",BRITE) ;
    temdtx(xb+10,yl+10,stext) ;
}

// ---------------------------------------------------------------
// Draw the contrast bar-graph                           
// ---------------------------------------------------------------
void draw_contr_bar(long xl,long yl,long xh,long yh) {

    long yb ;
    float level ;
    level = MIN(4.,CNTRST) ;
    level = MAX(level,-4.) ;
    yb = (yh-yl)*(level+4.)/8. + yl ;
    boks(xl,yl,xh,yh) ;
    fill_area(xl,yl,xh,yb) ;
    sprintf(stext,"%.1f",CNTRST) ;
    temdtx(xl+25,yb+10,stext) ;
} 
 
// ---------------------------------------------------------------
// Draw the input-to-output mapping curve for the image intensity
// complete with labels and frame                               
// ---------------------------------------------------------------
void draw_cntrst_curve() {

    long xo,yo,xm,ym ;
    xo = 100 ; xm = 700 ; yo = 200 ; ym = 650 ;
    boks(xo,yo,xm,ym) ;
    plot_cntrst(xo,yo,xm,ym) ;
    temdtx(xo-30,yo-30,"BLACK") ;
    temdtx(xm-30,yo-30,"WHITE") ;
    temdtx(xo-100,yo,"BLACK") ;
    temdtx((xo+xm)/2-40,yo-30,"INPUT") ;
    temdtx(xo-100,ym-10,"WHITE") ;
}  

// ---------------------------------------------------------------
// Determine the end-points of the intensity mapping line and draw 
// the line.                                                      
// ---------------------------------------------------------------
void plot_cntrst(long xo,long yo,long xm,long ym) {

    long ix1,iy1,ix2,iy2 ;
    float x1,x2,y1,y2 ;

    if (CNTRST > 0.) {
		y1 = (1.+CNTRST)*BRITE -CNTRST ; 
		x1 = 0. ;
		if (y1 < 0.) {
			y1 = 0. ;
			x1 = - ((1.+CNTRST)*BRITE - CNTRST)/CNTRST ;
		}
		x2 = 1. ;
		y2 = (1.+CNTRST)*BRITE  ;
		if (y2 > 1.) {
			y2 = 1. ;
			x2 = (1.-(1.+CNTRST)*BRITE +CNTRST)/CNTRST ;
		}
	}
    else if (CNTRST < 0.) {
		y1 = (1.-CNTRST)*BRITE  ;
		x1 = 0. ;
		if (y1 > 1.) {
			y1 = 1. ;
			x1 = (1.-(1.-CNTRST)*BRITE)/CNTRST ;
		}
		x2 = 1. ;
		y2 = CNTRST + (1.-CNTRST)*BRITE ;
		if (y2 < 0.) {
			y2 = 0. ;
			x2 = -(1.-CNTRST)*BRITE/CNTRST ;
		}
	}
    else {
		x1 = 0. ;
		x2 = 1. ;
		y1 = BRITE ;
		y2 = BRITE ;
	}
    ix1 = x1*(xm-xo) + xo ;
    ix2 = x2*(xm-xo) + xo ;
    iy1 = y1*(ym-yo) + yo ;
    iy2 = y2*(ym-yo) + yo ;
    temdln(ix1,iy1,ix2,iy2) ;
}

// ---------------------------------------------------------------
// Plot a new intensity mapping curve. Invoked by the commands to 
//  alter the brightness or contrast or setting of black-white  
// ---------------------------------------------------------------
void modify_cntrst(long i) {

    long xo,yo,xm,ym ;

    xo = 100 ; xm = 700 ; yo = 200 ; ym = 650 ;
    eraser(0,0,1000,32) ;
    OFF_VIDEO ;          // Erase the curve 
    plot_cntrst(xo,yo,xm,ym) ;
    NORMAL_VIDEO ;
    switch(i) {
		case 0 :
			OFF_VIDEO ;
			draw_brite_bar(xbl,ybl,xbh,ybh) ;
			NORMAL_VIDEO ;
			BRITE -= .02 ;
			if (BRITE < 0.) BRITE = 0. ;
			draw_brite_bar(xbl,ybl,xbh,ybh) ;
			plot_cntrst(xo,yo,xm,ym) ;
			break ;
		case 1 :
			OFF_VIDEO ;
			draw_brite_bar(xbl,ybl,xbh,ybh) ;
			NORMAL_VIDEO ;
			BRITE += .02 ;
			if (BRITE > 1.) BRITE = 1. ;
			draw_brite_bar(xbl,ybl,xbh,ybh) ;
			plot_cntrst(xo,yo,xm,ym) ;
			break ;
		case 2 :
			OFF_VIDEO ;
			draw_contr_bar(xcl,ycl,xch,ych) ;
			NORMAL_VIDEO ;
			CNTRST -= .2 ;
			draw_contr_bar(xcl,ycl,xch,ych) ;
			plot_cntrst(xo,yo,xm,ym) ;
			break ;
		case 3 :                                  
			OFF_VIDEO ;
			draw_contr_bar(xcl,ycl,xch,ych) ;
			NORMAL_VIDEO ;
			CNTRST += .2 ;
			draw_contr_bar(xcl,ycl,xch,ych) ;
			plot_cntrst(xo,yo,xm,ym) ;
			break ;
		case 4 : 
			OFF_VIDEO ;
			draw_brite_bar(xbl,ybl,xbh,ybh) ;
			draw_contr_bar(xcl,ycl,xch,ych) ;
			NORMAL_VIDEO ;
			CNTRST = -4. ;
			BRITE = 0. ;
			draw_brite_bar(xbl,ybl,xbh,ybh) ;
			draw_contr_bar(xcl,ycl,xch,ych) ;
			set_blkwht(xo,yo,xm,ym) ;
			draw_brite_bar(xbl,ybl,xbh,ybh) ;
			draw_contr_bar(xcl,ycl,xch,ych) ;
			plot_cntrst(xo,yo,xm,ym) ;
			break ;
	}
}

