// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include "header.h"
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include "menudef.h"
#include "menu_index.h"
#include "dfpdspl.h"
#include "params.h"
#include "atoms.h"
#include "PhasegratingRecord.h"
#include "MslicRecord.h"
#include "microscopes.h"
#include "displ.h"
#include "filenames.h"
#include "layers.h"
#include "funcdefs.h"
#define IDY	5
extern long gNumSymops,gNumBasisAtoms,gSpacegroupNumber,gPGisChanged ;
extern char title[],micname[],plots[],stext[] ;
extern struct	{ long status ; } calculation ;

// ---------------------------------------------------------------
// Display the parameters and prompt for input data 
// ---------------------------------------------------------------
void create_file() {

    char stemp[80],buffer[80],*cp ;
    long k,iz,ntype,natom,TTFS,i,j,x,y,mic_number ;
    float ftemp ;
    extern long LAYERED,NUM_DEFOC,NUM_THICK ;
    extern long gSecondSetting,gAlternateSetting ;

    set_menu(CH_MAIN_PAR) ;
    gCurrentButton = (*gCurrentMenu) ;
    gCurrentButton++ ; 		// to offset because of the CHANGE button
    
    // First display all the parameters to be requested
    if (!LAYERED) { // Display everything 
		for (i=0;i<gNumInputParameters;i++)  {
			x = parameter[i].xl ;
			y = parameter[i].yl+IDY ;
			temdtx(x,y,parameter[i].txt) ;
		}
	}
    else { // Layered structure, and we don't have all the parameters 
		for (i=0;i<gNumInputParameters;i++) {
			x = parameter[i].xl ;
			y = parameter[i].yl+IDY ;
			if (i!=1 && i!=4 && i!=12 && i!=13) { // Empty fields
				if (i==10) 
					temdtx(x,y,"Numb. of Layers : ") ;
				else if (i==11) 
					temdtx(x,y,"Numb. of Slices : ") ;
				else
					temdtx(x,y,parameter[i].txt) ;
				}
			}
	}
	
    // Next request data while outputting an
    // instructional message at the bottom 
    for (k=0;k<gNumInputParameters;k++) {
         x = parameter[k].xl ;
         y = parameter[k].yl+IDY ;
         switch(k)  {
            case 0:   // Get the title
				displ_msg(k) ;         
				REVERSE_VIDEO ;   
				temdtx(x,y,parameter[k].txt) ;
				ReadString(title,40) ;
				gInputParameter[0].cv = title ;
				NORMAL_VIDEO ;
				write_par(k) ;
				break ;
			case 1:   // Get the space-group
				if (!LAYERED) {
					displ_msg(k) ;         
					REVERSE_VIDEO ;   
					temdtx(x,y,parameter[k].txt) ;
					gSpacegroupNumber = get_spcgrp_number() ;
					gInputParameter[1].iv = gSpacegroupNumber ;
					NORMAL_VIDEO ;
					write_par(k) ;
				}
				else { // LAYERED case. Space group is # 1				
					gSpacegroupNumber = 1 ;
					gInputParameter[1].iv = gSpacegroupNumber ;
				}
				break ;
			case 2:   // Length of the A-axis 
				displ_msg(k) ;
				REVERSE_VIDEO ;
				temdtx(x,y,parameter[k].txt) ;
				gInputParameter[k].fv = get_float(3.) ;
				NORMAL_VIDEO ;
				write_par(k) ;
				break ;
			case 3:   // Length of the B-axis 
				if (gSpacegroupNumber >= 75) {
					gInputParameter[k].fv = gInputParameter[2].fv ;
					write_par(k) ;
					break ;
				}
				displ_msg(k) ;
				REVERSE_VIDEO ;
				temdtx(x,y,parameter[k].txt) ;
				gInputParameter[k].fv = get_float(3.) ;
				NORMAL_VIDEO ;
				write_par(k) ;
				break ;
			case 4:   // Length of the C-axis
				if (!LAYERED) {
					/* Cubic system */
					if (gSpacegroupNumber >= 195 ) {
						gInputParameter[k].fv = gInputParameter[2].fv ;
						write_par(k) ;
						break ;
					}
					/* Trigonal with Rhombohedral axes */
					else if (gSpacegroupNumber <= 167 && 
					gSpacegroupNumber >= 143) {
						if (gSecondSetting) {
							gInputParameter[k].fv = gInputParameter[2].fv ;
							write_par(k) ;
							break ;
						}
					}
					displ_msg(k) ;
					REVERSE_VIDEO ;
					temdtx(x,y,parameter[k].txt) ;
					gInputParameter[k].fv = get_float(3.) ;
					NORMAL_VIDEO ;
					write_par(k) ;
				}
				else
					gInputParameter[k].fv = 3. ; /* Dummy length */
				break ;
			case 5:   // ALPHA in degrees 
				if (!LAYERED) {
					if (gSpacegroupNumber <= 2 /* Triclinic */
							|| (gSpacegroupNumber >= 147 &&
							gSpacegroupNumber <= 167 && 
							gSecondSetting) /* Trigonal (rhomb)*/) {
						REVERSE_VIDEO ; 
						displ_msg(k) ;
						temdtx(x,y,parameter[k].txt) ;
						gInputParameter[k].fv = get_float(90.) ;
						NORMAL_VIDEO ; 
					}
					else {
						gInputParameter[k].fv = 90. ;
						write_par(k) ;
						break ;
					}
				}
				else
					gInputParameter[k].fv = 90. ;
				write_par(k) ;
				break ;
			case 6:   // BETA in degrees  
				if (!LAYERED) {
					if (gSpacegroupNumber >= 16 /* Not Monoclinic */) {
						gInputParameter[k].fv = gInputParameter[k-1].fv ;
						write_par(k) ;
						break ;
					}
					displ_msg(k) ;
					REVERSE_VIDEO ; 
					temdtx(x,y,parameter[k].txt) ;
					gInputParameter[k].fv = get_float(90.) ;
					NORMAL_VIDEO ; 
				}
				else
					gInputParameter[k].fv = 90. ;
				write_par(k) ;
				break ;
			case 7:   // GAMMA in degrees 
				// Cubic,Orthorombic or Tetragonal spacegroup 
				if (gSpacegroupNumber >= 195 ||
						(gSpacegroupNumber > 15 && gSpacegroupNumber < 143)) {
					gInputParameter[k].fv = 90. ;
					write_par(k) ;
					break ;
				}
				// Hexagonal setting
				else if (gSpacegroupNumber > 167 && 
						gSpacegroupNumber < 195) {
					gInputParameter[k].fv = 120. ;
					write_par(k) ;
					break ;
				}
				// Trigonal system
				else if (gSpacegroupNumber <= 167 && 
						gSpacegroupNumber >= 143) {
					if (!gSecondSetting) {
						gInputParameter[k].fv = 120. ;
						write_par(k) ;
						break ;
					}
					else {
						gInputParameter[k].fv = gInputParameter[5].fv ;
						write_par(k) ;
						break ;
					}
				}
				displ_msg(k) ;
				REVERSE_VIDEO ; 
				temdtx(x,y,parameter[k].txt) ;
				gInputParameter[k].fv = get_float(90.) ;
				NORMAL_VIDEO ; 
				write_par(k) ;
				break ;
			case 8:   // GMAX in rec. Angstr 
				// Use default and set to 2.0 
				gInputParameter[k].fv = 2.0 ;
				write_par(k) ;
				break ;
			case 9:   /* Get the zone-axis */
				if (!LAYERED) {
					displ_msg(k) ;         
					REVERSE_VIDEO ;   
					temdtx(x,y,parameter[k].txt) ;
					gPhsgrtRec->iuvw[0] = get_int(0) ;
					gPhsgrtRec->iuvw[1] = get_int(0) ;
					gPhsgrtRec->iuvw[2] = get_int(1) ;
					NORMAL_VIDEO ;
				}
				else {
					gPhsgrtRec->iuvw[0] = 0 ;
					gPhsgrtRec->iuvw[1] = 0 ;
					gPhsgrtRec->iuvw[2] = 1 ; 
				}
				write_par(k) ;
				break ;
			case 10:  
				if (!LAYERED) { // Set the number of symmetry operators 
					displ_msg(k) ;         
					REVERSE_VIDEO ;   
					temdtx(x,y,parameter[k].txt) ;
					if (gSpacegroupNumber == 0) {
						gNumSymops = get_int(1) ;
						NORMAL_VIDEO ;
						gInputParameter[k].iv = gNumSymops ;
						get_symop() ;
					}
					else {
						assm_spcgrp(gSpacegroupNumber) ;
						gInputParameter[k].iv = gNumSymops ;
					}
					NORMAL_VIDEO ;
					write_par(k) ;
				}
				else {
					// Treat the LAYERED case
					gLayerRec.num_pg = 0 ;
					gInputParameter[k].iv = gLayerRec.num_pg ;
					write_par(k) ;
				}
				break ;
			case 11:  
				if (!LAYERED) {
					// The number of slices per unit cell
					// We will set this to 1 upon creation
					gInputParameter[k].iv = 1 ;
					write_par(k) ;
				}
				else {
					// Number of slices are set later, so set to 0 
					gLayerRec.num_slc = 0 ;
					gInputParameter[k].iv = gLayerRec.num_slc ;
					write_par(k) ;
				}
				break ;
			case 12:  
				if (!LAYERED) {
					displ_msg(k) ;         
					// REVERSE_VIDEO ; 
					temdtx(x,y,parameter[k].txt) ;
					gInputParameter[k].iv = get_int(1) ;
					gNumBasisAtoms = gInputParameter[k].iv ;
					write_par(k) ;
					get_atoms() ;
					gInputParameter[k+1].iv = 
					(ntype = types(basis,gNumBasisAtoms)) ;
					sort(basis,gNumBasisAtoms) ;
					ApplySymmetries() ;
					NORMAL_VIDEO ;
					write_par(k) ;
				}
				break ;
			case 13:  
				if (!LAYERED) {
					// Number of different types of atoms 
					// It has been set by case 12         
					write_par(k) ;
				}
				break ;
			case 14:  // Get the microscope name and set par. 
				displ_msg(k) ;         
				REVERSE_VIDEO ;   
				temdtx(x,y,parameter[k].txt) ;
				ReadString(micname,10) ; 
				if (strlen(micname) == 0)
					strcpy(micname,"ARM") ;
				NORMAL_VIDEO ;
				gInputParameter[k].cv = micname ;
				// Find which microscope
				mic_number = set_microscope(micname) ;
				write_par(k) ;
				if (mic_number != -1) {
					gCurrentButton++ ;
					gInputParameter[++k].fv = 
					microscope[mic_number].sph_ab ;
					write_par(k) ;
					gCurrentButton++ ;
					gInputParameter[++k].fv = 
					microscope[mic_number].delta ;
					write_par(k) ;
					gCurrentButton++ ;
					gInputParameter[++k].fv = 
					microscope[mic_number].theta ;
					dfpdspl.theta = gInputParameter[k].fv ;
					write_par(k) ;
					gCurrentButton++ ;
					gInputParameter[++k].fv = 
					microscope[mic_number].voltage ;
					write_par(k) ;
				}
				break ;
			case 15:  // Spherical abberation in mm 
				displ_msg(k) ;
				REVERSE_VIDEO ;
				temdtx(x,y,parameter[k].txt) ;
				gInputParameter[k].fv = get_float(2.) ;
				NORMAL_VIDEO ;
				write_par(k) ;
				break ;
			case 16:  // Spread of defocus in Angst 
				displ_msg(k) ;
				REVERSE_VIDEO ;
				temdtx(x,y,parameter[k].txt) ;
				gInputParameter[k].fv = get_float(80.) ;
				NORMAL_VIDEO ;
				write_par(k) ;
				break ;
			case 17:  // Divergence angle (half) mrad 
				displ_msg(k) ;
				REVERSE_VIDEO ;
				temdtx(x,y,parameter[k].txt) ;
				gInputParameter[k].fv = get_float(0.5) ;
				NORMAL_VIDEO ;
				dfpdspl.theta = gInputParameter[k].fv ;
				write_par(k) ;
				break ;
			case 18:  // Microscope voltage in kV 
				displ_msg(k) ;
				REVERSE_VIDEO ;
				temdtx(x,y,parameter[k].txt) ;
				gInputParameter[k].fv = get_float(200.) ;
				NORMAL_VIDEO ;
				write_par(k) ;
				break ;
			case 19:  // Center of the LAUE circle in units of H
			case 20:  // Center of the LAUE circle in units of K
				displ_msg(k) ;
				REVERSE_VIDEO ;
				temdtx(x,y,parameter[k].txt) ;
				gInputParameter[k].fv = get_float(0.) ;
				NORMAL_VIDEO ;
				write_par(k) ;
				break ;
			case 21:  // Possible thru-thickness series
				displ_msg(k) ;         
				REVERSE_VIDEO ;   
				temdtx(x,y,parameter[k].txt) ;
				cp = buffer ;
				TTFS = FALSE ;
				ReadString(buffer,40) ;
				if (strchr(buffer,',') != 0 ||
					strchr(buffer,'(') != 0) TTFS = TRUE ;
				if (!TTFS) {
					if (strlen(buffer)==0)
						gInputParameter[k].fv = 20. ;
					else
						gInputParameter[k].fv = atof(buffer) ;
					gMsliceRec.t1 = gInputParameter[k].fv ;
					gMsliceRec.t2 = gMsliceRec.t1  ;
					gMsliceRec.dt = 0. ;
					gMsliceRec.tts = FALSE ;
					NUM_THICK = 1 ;
				}
				else {
					get_ttfs(buffer,&gMsliceRec.t1,&gMsliceRec.t2,&gMsliceRec.dt) ;
					gMsliceRec.tts = TRUE ;
					gInputParameter[k].fv = gMsliceRec.t2 ;
					NUM_THICK = (gMsliceRec.t2-gMsliceRec.t1)/gMsliceRec.dt + 1 ;
				}
				NORMAL_VIDEO ;
				write_par(k) ;
				break ;
			case 22:  /* Set amplitude output YES/NO */
				displ_msg(k) ;         
				REVERSE_VIDEO ;   
				temdtx(x,y,parameter[k].txt) ;
				gMsliceRec.iplot = FALSE ;
				ReadString(plots,4) ; 
				gInputParameter[k].cv = plots ;
				NORMAL_VIDEO ;
				eraser(0,0,1020,40) ;
				eraser(520,270,1020,420) ;
				if (plots[0]=='y' || plots[0]=='Y') {
					gMsliceRec.iplot = TRUE ;
					temdtx(0,130,
					"Enter the number of reflections to be output -> ");
					gMsliceRec.nw = get_int(1) ;
					eraser(0,120,1020,160) ;
					for(j=0;j<gMsliceRec.nw;j++) {
						sprintf(stext,
						"Enter index h k l for reflection # %d : ",j+1);
						temdtx(0,130,stext) ;
						gMsliceRec.ihw[j] = get_int(0) ;
						gMsliceRec.ikw[j] = get_int(0) ;
						gMsliceRec.ilw[j] = get_int(0) ;
						while(!in_zone(gMsliceRec.ihw[j],gMsliceRec.ikw[j],
								gMsliceRec.ilw[j],gPhsgrtRec->iuvw)) {
							eraser(0,120,1020,160) ;
							temdtx(0,130,
							"Reflection not in zone, re-enter -> ") ;
							gMsliceRec.ihw[j] = get_int(0) ;
							gMsliceRec.ikw[j] = get_int(0) ;
							gMsliceRec.ilw[j] = get_int(0) ;
						} ;
						eraser(0,120,1020,160) ;
					}
				}
				else
					strcpy(plots,"NO") ;
				NORMAL_VIDEO ;
				write_par(k) ;
				break ;
			case 23:  /* Possible thru-defocus */
				displ_msg(k) ;         
				REVERSE_VIDEO ;   
				temdtx(x,y,parameter[k].txt) ;
				cp = buffer ;
				TTFS = FALSE ;
				ReadString(buffer,40) ;
				if (strchr(buffer,',') != 0 ||
					strchr(buffer,'(') != 0) TTFS = TRUE ;
				if (!TTFS) {
					if (strlen(buffer)==0)
						gInputParameter[k].fv = -600. ;
					else
						gInputParameter[k].fv = atof(buffer) ;
					gDisplayRec.f1 = gInputParameter[k].fv ;
					gDisplayRec.f2 = gDisplayRec.f1 ;
					gDisplayRec.df = 0. ;
					gDisplayRec.tfs = FALSE ;
					NUM_DEFOC = 1 ;
				}
				else {
					get_ttfs(buffer,
					&gDisplayRec.f1,&gDisplayRec.f2,&gDisplayRec.df) ;
					gDisplayRec.tfs = TRUE ;
					gInputParameter[k].fv = gDisplayRec.f2 ;
					ftemp = (gDisplayRec.f2-gDisplayRec.f1)/gDisplayRec.df ;
					ftemp = (ftemp > 0. ? ftemp : -ftemp) ;
					NUM_DEFOC = (long)ftemp + 1 ;
				}
				NORMAL_VIDEO ;
				write_par(k) ;
				break ;
			case 24:  /* Objective lens aperture in rec. Angstrom    */
				displ_msg(k) ;
				REVERSE_VIDEO ;
				temdtx(x,y,parameter[k].txt) ;
				gInputParameter[k].fv = get_float(0.7) ;
				NORMAL_VIDEO ;
				write_par(k) ;
			break ;
			case 25:  /* Center of the objective apert in units of H */
			case 26:  /* Center of the objective apert in units of K */
			case 27:  /* Center of the optic axis in units of H */
			case 28:  /* Center of the optic axis in units of K */
				displ_msg(k) ;         
				REVERSE_VIDEO ;   
				temdtx(x,y,parameter[k].txt) ;
				gInputParameter[k].fv = get_float(0.) ;
				NORMAL_VIDEO ;
				write_par(k) ;
				break ;
		}
		eraser(0,0,1020,40) ;
		gCurrentButton++ ;
	}
	
	if (!LAYERED) {
		set_nz() ;
		gPGisChanged = TRUE ;
	}
	calculation.status = 0 ;	// The status of a new file is
							// one with nothing calculated
	setup_trnsf() ;		// so that we transfer the input data to the appropriate
					// structures so that the file is written properly
	rewrite_file(flname.pin) ;
	if (!LAYERED)
		set_menu(MAIN_SNGL_STR) ;
	else
		set_menu(MAIN_LAY_STR) ;
}
