// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
// -----------------------------------------------------------
// Routine to display the diffraction pattern on the screen
// The indexing of the diffraction pattern should not be used	
// It is a hang over from some old fortran code.	
// -----------------------------------------------------------
#include <stdio.h>
#include <math.h>
#include "globals.h"
#include "header.h"
#include "menudef.h"
#include "PhasegratingRecord.h"
#include "MslicRecord.h"
#include "ProjCellRecord.h"
#include "displ.h"
#include "dfpdspl.h"
#include "storage.h"
#include "devices.h"
#include "funcdefs.h"
extern  short gShadedSpheres ;
extern 	long gObjectiveAperture ;
extern 	struct { long xpos,ypos ; } cursor ;
extern 	struct { long ipot,icmplx ; } flag ;
extern 	long ERASE ;
extern  char	POSTSCRIPT_FILE[],gFileName[] ;
#define fabs(x) ((x) < 0. ? -(x) : (x))
extern	short gPrintValues ;

// -------------------------------------------------------------------
// Routine to display the diffraction pattern. The diffr. 
// beams get sorted and displayed according to their     
// intensity and the angle of divergence. The pattern can
// be displayed on a framestore, laserprinter or the CRT
// If the CRT is used, the diffraction pattern is drawn
// as circles whose radius is proportional to the spot
// intensity and the scaling is only such that the spots
// do not overlap. If the Framestore is used, the scaling
// is according to the cameralength.	
// -------------------------------------------------------------------
void dispfft(long *pnp,long *pmp) {

	static long ic ;
	float zrl,zim,ampl,phase,*rp,*xhk,*yhk,*f,*g,*r,gsq ;
	float arg,arg1,astsq,bstsq,abcs,one_ov_jmax,rlintmin ;
	float wavel,camera_const,scale,sigma,sigmasq,xp,yp,aprtsq ;
	float fact,rint,rat,gx,gy,dx,dy,cnst,dh,dk,dst,r0,tstint ;
	short *ihk,*i16p ;
	long   xmin,xmax,ymin,ymax,maxh,maxk,itolsq,n,nh,m,nk,nl,
			jh,jk,jl,ix,iy ;
	long   xpos,ypos,num,iapert_rad,ix0,iy0 ;
	long   mag_fact,jmax,ixc,iyc,i,intens ;
	unsigned char bb[512] ;
	long   np,mp ;
	float pi ;
	extern char stext[] ;
	FILE *fp ;

	np = *pnp ;
	mp = *pmp ;
	f = gWorkSpace+gMaxArraySize/2 ;
	r = f ;
	g = r+gMaxArraySize/8 ;
	xhk = g ;
	yhk = xhk+gMaxArraySize/8 ;
	ihk = (short *)(yhk+gMaxArraySize/8) ; 

	jmax = np*mp ;
	pi = 4.*atan2(1.,1.) ;
	one_ov_jmax = 1./jmax ;
	maxk = mp/2 ;
	maxh = np/2 ;
	astsq = gProjCell->ast*gProjCell->ast ;
	bstsq = gProjCell->bst*gProjCell->bst ;
	abcs = gProjCell->ast*gProjCell->bst*gProjCell->csga ;
	rlintmin = pow(10.,-dfpdspl.cutoff) ;
	wavel = 1.2263652e-7/sqrt((gMsliceRec.volt+9.785025e-4*
	gMsliceRec.volt*gMsliceRec.volt)*1000.) ;
	wavel *= 1.0e+8 ;
	xpos = cursor.xpos ;
	ypos = cursor.ypos ;
	ixc = dfpdspl.xcenter ;  /* Center of the Diffraction Pattern */
	iyc = dfpdspl.ycenter ;  /* Center of the diffraction pattern */

	/* We will set up the camera length to mimick the 200cx blown up by */
	/* a factor of MAG_FACT                                             */
	mag_fact = 8 ;
	/* if (flag.ipot) mag_fact = 1 ;*/
	/* -------- Camera length = 520 mm Default -------------- */
	camera_const = dfpdspl.camera_length*wavel*mag_fact ;
	scale = 50. ;   /* Saturate the film by scaling up */
	sigma = dfpdspl.camera_length*dfpdspl.theta
			*0.001*mag_fact ; /* size of D. spots */
	itolsq = (sigma+2.)*(sigma+2.) ;
	sigmasq = sigma*sigma ;
	aprtsq = gDisplayRec.aprtrd*gDisplayRec.aprtrd ;
	
	if (flag.ipot) {
		fact = 1. ;
	}
	else {
		fact = pow(10.,dfpdspl.cutoff) ;
	}
	
	if (gPrintValues && !dfpdspl.index) {
		fp = fopen(gFileName,"w") ;
		fprintf(fp,"\n   H\t  K\t  L\t    AMPL.\t    PHASE") ;
		for(m=0;m<mp;m++) {
			nk = m ;
			if(nk > maxk) nk -= mp ;
			for(n=0;n<np;n++) {
				nh = n ;
				if (nh > maxh) nh -= np ;
				rp = gWorkSpace + 2*(m*np+n) ;
				zrl = *rp*one_ov_jmax ;
				zim = *(rp+1)*one_ov_jmax ; /* scaling due to having done */
				rint = zrl*zrl+zim*zim ;    /* the FFT. Intensities go    */
				if (rint > rlintmin) {
					gsq = nh*nh*astsq+nk*nk*bstsq-2.*nh*nk*abcs ;
					if (gsq <= aprtsq) {
						ampl = sqrt(rint) ;
						phase = atan2(zim,zrl)/pi ;
						Find3DHKLIndicies(nh,nk,0,&jh,&jk,&jl,gProjCell->hkltr) ;
						fprintf(fp,"\n %3d\t%3d\t%3d\t%10.6f\t%10.6f",
						jh,jk,jl,ampl,phase) ;
					}
				}
			}
		}
		fprintf(fp,"\n") ;
		fclose(fp) ;
		return ;
	}
	if(!dfpdspl.index) {
		ic = 0 ;
		for(m=0;m<mp;m++) {
			nk = m ;
			if(nk > maxk) nk -= mp ;
			for(n=0;n<np;n++) {
				nh = n ;
				if (nh > maxh) nh -= np ;
				rp = gWorkSpace + 2*(m*np+n) ;
				zrl = *rp*one_ov_jmax ;
				zim = *(rp+1)*one_ov_jmax ; /* scaling due to having done */
				rint = zrl*zrl+zim*zim ;    /* the FFT. Intensities go    */
				                /* from 0-1 for the WF. For   */
				                /* FFT of the POT they go as  */
				                /* the structure factor sq.   */
				rint *= fact ;  /* Ensure that I> min_fft_int gets */
				    /* displayed                       */
				if (rint != 0.) {
					if(flag.ipot) {
						rint = sqrt(rint) ; /* Struct.fact(0-tens ev) */
						rint = sqrt(rint) ; /* Bring to scale of D.P  */
					}
					else {
						rint = log10(rint); /* 0 - min_fft_int (~6) */
					}
					if ((flag.ipot && rint >=1.) ||
							(!flag.ipot && rint > 0.)) {      /* Display */
						i16p = ihk + 2*ic ;
						*(f+ic) = rint ;
						*i16p = nh ;
						*(i16p+1) = nk ;
						*(g+ic) = nh*nh*astsq+ nk*nk*bstsq-2.*nh*nk*abcs ;
						ic++ ;
					}
				}
			}
		}
	}
	/* Sort the reflections according to their g-vector */
	SortGVectors(f,g,ihk,ic) ;

	/* Plot the diffraction pattern */
	rat = gProjCell->bst/gProjCell->ast ;
	if (ic >= 1) {
		cnst = log(10.) ; /* So that we are looking at Ln(I) */
		for( i=0;i<ic;i++) {
			*(f+ic) *= cnst ;
			nh = (long)(*(ihk+2*i)) ;
			nk = (long)(*(ihk+2*i+1)) ;
			gy = nk*gProjCell->bst - nh*gProjCell->ast*gProjCell->csga ;
			gx = nh*gProjCell->ast*gProjCell->siga ;
			ix0 = dfpdspl.xcenter + camera_const*gx ;
			iy0 = dfpdspl.ycenter - camera_const*gy ;
			arg1 = *(f+i)*sigmasq ;
			dy = sqrt(arg1) ;
			ymin = iy0 - dy ;
			ymax = iy0 + dy ; 
			if(dfpdspl.index) {
				if(((xpos-ix0)*(xpos-ix0)+(ypos-iy0)*(ypos-iy0))
						< itolsq) {
					/* Found the right spot */
					/* Write out the hkl    */
					nl = 0 ;
					Find3DHKLIndicies(nh,nk,nl,&jh,&jk,&jl,gProjCell->hkltr) ;
					sprintf(stext,"%d %d %d",jh,jk,jl) ;
					gDrawText(ix0-StringWidth(stext)/2,iy0,stext) ;
					return ;
				}
			}
			else {
				if (gShadedSpheres) {
					for(iy=ymin;iy<=ymax;iy++) {
						if (iy > 0 && iy < OutDevice->yres) {
							dy = iy-iy0 ;
							arg = arg1 - dy*dy ;
							if(arg > 0.) {
								dx = sqrt(arg) ;
								xmin = ix0-dx ;
								xmax = ix0+dx ;
								if (xmin > 0 && xmax < OutDevice->xres) {
									num = 0 ;
									for(ix=xmin;ix<=xmax;ix++) {
										dx = ix-ix0 ;
										intens = scale*
												(arg1-dx*dx-dy*dy)/sigmasq ;
										if (intens>OutDevice->range) 
											intens=OutDevice->range;
										if (intens < 0) intens=0 ;
										/* Invert */
										intens = OutDevice->range-intens ;
										if (OutDevice->type == WS) 
											bb[num++] = gLut[intens] ;
										else
											bb[num++] = intens ;
									}
									gDrawImageLine(xmin,iy,num,bb) ;
								}
							}
						}
					}
				}
				else {
					gFillCircle(ix0,iy0,near(.8*dy)) ;
				}
			}
		}
		if(gObjectiveAperture) {
			/* Find the center of the circle */
			gy = gDisplayRec.kc*gProjCell->bst - gDisplayRec.hc*gProjCell->ast*gProjCell->csga ;
			gx = gDisplayRec.hc*gProjCell->ast*gProjCell->siga ;
			ix0 = dfpdspl.xcenter + camera_const*gx ;
			iy0 = dfpdspl.ycenter - camera_const*gy ;
			iapert_rad = camera_const*gDisplayRec.aprtrd ;
			gDrawCircle(ix0,iy0,iapert_rad) ;
		}
	}
}
