// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <stdio.h>
#include "types.h"
#include "menudef.h"
#include "funcdefs.h"
#include "devices.h"
#include "globals.h"
#define ON	1
#define OFF	0

// --------------------------------------------------------------------------
// Subroutine to display a particular item within the menu
// --------------------------------------------------------------------------

void display(COMMAND *item) { 
	DrawControl(item) ;
}

void DrawControl(COMMAND *control) {
    Rect	tRect ;
    short 	x,y,width,height ;
    extern	unsigned long	gBlackPixel,gDarkerGray,gGrayPixel ;

    set_video_normal() ;
    width = (control->xh - control->xl)/2 ;
    height = (control->yh - control->yl)/2 ;
    x = control->xl/2 ;
    y = Terminal.yres - control->yh ;
    y /= 2 ;
    SetFgColor(GetBkColor()) ;
    SetRect(&tRect,x,y,x+width,y+height) ;
    InsetRect(&tRect,1,1) ;
    EraseRect(&tRect) ;
    PenSize(2,2) ;
    SetFgColor(gWhitePixel) ;
    DrawLine(tRect.left,tRect.bottom,tRect.left,tRect.top) ;
    DrawLine(tRect.left,tRect.top,tRect.right,tRect.top) ;
    if (!control->disabled)
    	SetFgColor(gBlackPixel) ;
    else 
    	SetFgColor(gDarkerGray) ;
    DrawLine(tRect.right,tRect.top,tRect.right,tRect.bottom) ;
    DrawLine(tRect.right,tRect.bottom,tRect.left,tRect.bottom) ;
    PenSize(1,1) ;
    width = StringWidth(control->txt) ;
    x = (tRect.left+tRect.right)/2-width/2 ;
    y = (tRect.top+tRect.bottom)/2 + 5 ;
    DrawText(x,y,control->txt) ;
    SetFgColor(gBlackPixel) ;
    XFlush(gDisplay) ;
}

void HighlightControl(COMMAND *control) {
	Rect	tRect ;
	short 	x,y,width,height ;
	extern	uLong	gWhitePixel,gBlackPixel,gDarkerGray,gGrayPixel ;

	set_video_normal() ;
	width = (control->xh - control->xl)/2 ;
	height = (control->yh - control->yl)/2 ;
	control->status = ON ;
	x = control->xl/2 ;
	y = Terminal.yres - control->yh ;
	y /= 2 ;
	SetRect(&tRect,x,y,x+width,y+height) ;
	InsetRect(&tRect,1,1) ;
	SetFgColor(gDarkerGray) ;
	FillRect(&tRect) ;
	PenSize(2,2) ;
	SetFgColor(gBlackPixel) ;
	DrawLine(tRect.left,tRect.bottom,tRect.left,tRect.top) ;
	DrawLine(tRect.left,tRect.top,tRect.right,tRect.top) ;
	SetFgColor(gWhitePixel) ;
   	if (control->disabled) SetFgColor(gGrayPixel) ;
	DrawLine(tRect.right,tRect.top,tRect.right,tRect.bottom) ;
	DrawLine(tRect.right,tRect.bottom,tRect.left,tRect.bottom) ;
	PenSize(1,1) ;
	width = StringWidth(control->txt) ;
	x = (tRect.left+tRect.right)/2-width/2 ;
	y = (tRect.top+tRect.bottom)/2 + 5 ;
	DrawText(x,y,control->txt) ;
	SetFgColor(gBlackPixel) ;
	XFlush(gDisplay) ;
}

void UnHighlightControl(COMMAND *control) {

    control->status = OFF ;
    DrawControl(control) ;
}

void EnableControl(COMMAND *control) {
	control->disabled = false ;
	DrawControl(control) ;
}

void DisableControl(COMMAND *control) {
	control->disabled = true ;
	DrawControl(control) ;
}