// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <stdio.h>
#include "header.h"
#include "funcdefs.h"
#include "graphics.h"
#include "char_symbol.h"
#include "ttypes.h"
#include "menudef.h"
#include "devices.h"
#include "globals.h"
extern FILE *psfile ;
extern Visual *gVisual ;

// ----------------------------------------------------------------
// Write a point on the display. Presently only 3 outputs 
// are defined   WorkStation=0  FrameStore =1 LaserPrinter=2 
// All the routines in here expect the coordinate systems (0,0) 
// to be at the upper left corner. Thus any transformation to the 
// local coordinate system takes place here.	
// ----------------------------------------------------------------
void gDrawPixel(long x,long y) {
    switch(OutDevice->type) {
		case WS:
			DrawPixel(x,y) ;
			break;
		case FS :
			semdpt(&x,&y) ;
			break ;
		case LP :
			y = OutDevice->yres -y - 1 ;
			/* Generate Postscript Code */
			break ;
    }
} 

// ------------------------------------------------------- 
// Routine to draw a line in physical Device coordinates 
// ------------------------------------------------------- 
void gDrawLine(long x1,long y1,long x2,long y2) {
    switch(OutDevice->type) {
		case WS:
			DrawLine(x1,y1,x2,y2) ;
			break;
		case FS :
			semdln(&x1,&y1,&x2,&y2) ;
			break ;
		case LP :
			y1 = OutDevice->yres -y1 - 1 ;
			y2 = OutDevice->yres -y2 - 1 ;
			/* Generate Postscript Code */
			fprintf(psfile,
				"\nnewpath %d %d moveto %d %d lineto stroke",
				x1,y1,x2,y2) ;
			break ;
    }
}

// ----------------------------------------------------------
// Move the "pen" to a point in physical device coordinates 
// ----------------------------------------------------------
void gMoveTo(long ix,long iy) {
    switch(OutDevice->type) {
		case WS:
    			gCurrentTextPos.x = ix ;
    			gCurrentTextPos.y = iy ;
			break;
		case FS :
			break ;
		case LP :
			iy = OutDevice->yres - iy - 1 ;
			/* Generate Postscript Code */
			fprintf(psfile,"\n%d %d moveto",ix,iy) ;
			break ;
    }
}

// ----------------------------------------------------
// Draw a character on the display with attributes     
// ---------------------------------------------------- 
void gDrawChar(long x,long y,char c) {
    switch(OutDevice->type) {
		case WS:
			DrawChar(x,y,c) ;
			break;
		case FS :
			break ;
		case LP :
			y = OutDevice->yres -y - 1 ;
			/* Generate Postscript Code */
			fprintf(psfile,"\n%d %d moveto",x,y) ;
			fprintf(psfile,"\n{%c} show",c) ;
			break ;
    }
}

// -----------------------------------------------
// Routine to draw a circle in PDC units   
// -----------------------------------------------
void gDrawCircle(long x,long y,long rad) {
		
    long linewidth ;
    long ilw,y1 ;

    linewidth = GetLineWidth() ;
    ilw = linewidth ;
    switch(OutDevice->type) {
		case WS:
			SetLineWidth(ilw) ;
			DrawCircle(x,y,rad) ;
			SetLineWidth(linewidth) ;	
			break ;
		case  FS:
			break ;
		case LP :
			y1 = OutDevice->yres - y - 1 ;
			/* Generate Postscript Code */
			fprintf(psfile,"\nnewpath") ;
			fprintf(psfile,"\n%d %d %d 0 360 arc",x,y1,rad) ;
			fprintf(psfile,"\n%d setlinewidth",ilw) ;
                        fprintf(psfile,"\nstroke") ;
			fprintf(psfile,"\n%d setlinewidth",linewidth) ;
			break ;
    	}
}

void gSetFgColor(long value) {
    float gray ;
    switch(OutDevice->type) {
		case WS:
			SetFgColor(value) ;
			break ;
		case FS:
			break ;
		case LP:
			gray = value/127. ;
			fprintf(psfile,"\n%f setgray",gray) ;
			break ;
	    }
}

// -----------------------------------------------
// Routine to fill a circle in PDC units
// -----------------------------------------------
void gFillCircle(long x,long y,long rad) {
		
    long linewidth ;
    long ilw,y1 ;

    linewidth = GetLineWidth() ;
    ilw = linewidth ;
    switch(OutDevice->type) {
		case WS:
			SetLineWidth(ilw) ;
			FillCircle(x,y,rad) ;
			SetLineWidth(linewidth) ;	
			break ;
		case  FS:
			break ;
		case LP :
			y1 = OutDevice->yres - y - 1 ;
			/* Generate Postscript Code */
			fprintf(psfile,"\nnewpath") ;
			fprintf(psfile,"\n%d %d %d 0 360 arc",x,y1,rad) ;
			fprintf(psfile,"\n%d setlinewidth",ilw) ;
                        fprintf(psfile,"\nfill") ;
			fprintf(psfile,"\n%d setlinewidth",linewidth) ;
			break ;
    	}
}

// ----------------------------------------------- 
// Routine to draw a rectangle in PDC units  
// ----------------------------------------------- 
void gDrawRect(long xlow,long ylow,long xhi,long yhi) {
    long linewidth ;
    Rect tRect ;
    long i,ilw,y1,y2 ;

    linewidth = GetLineWidth() ;
    ilw = linewidth ;
    switch(OutDevice->type) {
		case WS:
			SetLineWidth(ilw) ;
			SetRect(&tRect,xlow,ylow,xhi,yhi) ;
			FrameRect(&tRect) ;
			SetLineWidth(linewidth) ;	
			break ;
		case  FS:
			for(i=0;i<ilw;i++) {
				semdln(&xlow,&ylow,&xhi,&ylow) ;
				semdln(&xhi,&ylow,&xhi,&yhi) ;
				semdln(&xhi,&yhi,&xlow,&yhi) ;
				semdln(&xlow,&yhi,&xlow,&ylow) ;
				xlow++ ;
				ylow++ ;
				xhi-- ;
				yhi-- ;
			}
			break ;
		case LP :
			ylow = OutDevice->yres -ylow - 1 ;
			yhi = OutDevice->yres -yhi - 1 ;
			/* Generate Postscript Code */
			fprintf(psfile,"\nnewpath") ;
			fprintf(psfile,"\n%d %d moveto",xlow,ylow) ;
			fprintf(psfile,"\n%d %d lineto",xhi,ylow) ;
			fprintf(psfile,"\n%d %d lineto",xhi,yhi) ;
			fprintf(psfile,"\n%d %d lineto",xlow,yhi) ;
			fprintf(psfile,"\nclosepath") ;
			fprintf(psfile,"\n%d setlinewidth",ilw) ;
                        fprintf(psfile,"\nstroke") ;
			break ;
    	}
}

// -----------------------------------------------
// Routine to write text at x,y in PDC units
// -----------------------------------------------
void gDrawText(long x,long y,char *text) {
    switch(OutDevice->type) {
		case WS:
    		    DrawText(x,y,text) ;
			break ;
		case  FS:
    			break ;
		case LP :
			y = OutDevice->yres -y - 1 ;
			/* Generate Postscript Code */
			fprintf(psfile,"\n%d %d moveto",x,y) ;
			fprintf(psfile,"\n(%s) show",text) ;
			break ;
    }
}

void gDrawImageLine(long x,long y,long n,unsigned char *imageline) {
    long i,j,ix,iy ;
   
    switch(OutDevice->type) {
		case WS:
			DrawImageLine(x,y,n,imageline) ;
    			break ;
		case  FS:
    			break ;
		case LP :
			iy = OutDevice->yres-y-1 ;
			ix = x ;
			/* Generate Postscript Code */
    			fprintf(psfile,"\n%d 1 %d %d DrawImageLine",n,ix,iy) ;
    			for (i=0;i<n;i++) {
    				if (i%40==0)
    					fprintf(psfile,"\n") ;
    				fprintf(psfile,"%02x",imageline[i]) ;
			}
			break ;
    }
}

long gReadImageLine(long x,long y,long n, unsigned char *imageline) {   
    switch(OutDevice->type) {
		case WS:
			ReadImageLine(x,y,n, imageline) ;
    		break ;
		case  FS:
    		break ;
		case LP :
			y = OutDevice->yres -y - 1 ;
			break ;
    }
    return(1) ;
}

long gDrawImage(unsigned char *image,long x,long y,long width,long height) {
    switch(OutDevice->type) {
		case WS:
			DrawImage(image,x,y,width,height) ;
			break ;
		case  FS:
    		break ;
		case LP :
			y = OutDevice->yres -y - 1 ;
			/* Generate Postscript Code */
			break ;
    }
    return(1) ;
}


long gReadImage( unsigned char *image,long x,long y,long width,long height) {

    switch(OutDevice->type) {
		case WS:
			// We will allow for the display to be both 
			// in PseudoColor and TrueColor			
			ReadRGBImage(image,x,y,width,height) ;
			break ;
		case  FS:
    			break ;
		case LP :
			/* Generate Postscript Code */
			y = OutDevice->yres -y - 1 ;
			break ;
    }
    return(1) ;
}

void LandScape() {
  long it ;
  switch(OutDevice->type) 
	{
	case WS:
	case FS:
		break ;
	case LP:
		it = OutDevice->xres ;
		fprintf(psfile,"\nLandscape") ;
		OutDevice->xres = OutDevice->yres ;
		OutDevice->yres = it ;
		break ;
	}
  }
