// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <stdlib.h>
#include <math.h>

#ifdef VMS
#include <decw$include:keysym.h>
#include <decw$include/Xlib.h>
#include <decw$include/Xutil.h>
#else
#include <X11/keysym.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#endif
#include "header.h"
#include "menudef.h"
#include "globals.h"
#include "funcdefs.h"

extern Display *gDisplay;
extern Point 	gCurrentTextPos;
extern XFontStruct *gFontStructure;
extern char gCurrentText[];
extern char buffer[80];

KeySym	GetKey(XEvent *theEvent) ;

#define YES 1
#define NO 0
#define CR '\015'
#define DEL '\177'
#define BS '\010'
#define COMMA ','
#define SPACE ' '
#define PERIOD '.'

char buffer[80] ;

// ----------------------------------------------------------
// Return a floating point number prompted from the screen
// ----------------------------------------------------------
float get_float(float defaultValue) {
	char *GetNumber(),*retString ;

	retString = GetNumber() ;
	if (strlen(retString)==0)
		return (defaultValue) ;
	else
		return (atof(retString)) ;
}

// ------------------------------------------------------
// Return a integer number prompted from the screen    
// ------------------------------------------------------
long get_int(long defaultValue) {
    char *GetNumber(),*retString ;
   
    retString = GetNumber() ;
    if (strlen(retString)==0)
		return(defaultValue) ;
    else
		return (atoi(retString)) ;
}

KeySym	GetKey(XEvent *theEvent) {
	KeySym 	keysym ;
	XComposeStatus compose ;
	char 	InputBuffer[80] ;

	long count = XLookupString((XKeyEvent *)theEvent,
			InputBuffer,80,&keysym,&compose) ;
	return	keysym ;
}

// ------------------------------------------------------
// Get a floating/long number from the terminal. Saves   
// only valid characers and echoes only valid characters 
// Invalid characters are not echoed and the bell will  
//  sound                                                 
// ------------------------------------------------------
char * GetNumber() {
	Rect	tRect,pRect ;
	long 	i,ichar ;
	long 	x,y,height,width,descent ;
	long 	xBegPos,yBegPos,count ;
	XEvent 	theEvent ;
	KeySym 	keysym ;
	XComposeStatus compose ;
	char 	InputBuffer[80],lastCharString[2] ;
	char 	oldCurrentText[132] ;

	// Fill the buffer with the mask
	for (ichar = 0 ; ichar<80 ; buffer[ichar++] = EOS) ;

	strcpy(oldCurrentText,gCurrentText) ;

	// Locate the current text position
	x = gCurrentTextPos.x ;
	y = gCurrentTextPos.y ;
	height = gFontStructure->ascent ;
	descent = gFontStructure->descent ;
	xBegPos = x ;
	yBegPos = y ;
	SetRect(&pRect,gCurrentTextPos.x,gCurrentTextPos.y-height+2,
	gCurrentTextPos.x+6,gCurrentTextPos.y+1) ;
	FillRect(&pRect) ;

	compose.compose_ptr = (char *)0 ;
	compose.chars_matched = 0 ;
	
	for (ichar = 0 ;ichar < 80; ) {
		XNextEvent(gDisplay,&theEvent) ;
		if (theEvent.type == KeyPress) {
			count = XLookupString((XKeyEvent *)&theEvent,
			InputBuffer,80,
			&keysym,&compose) ;
			InputBuffer[count] = EOS ;
			if ((keysym <= XK_9 && 
					keysym >= XK_0) || 
					(keysym <= XK_KP_9 &&
					keysym >= XK_KP_0) ||
					keysym == XK_period || 
						keysym == XK_minus) {
				strcat(buffer,InputBuffer) ;
				EraseRect(&pRect) ;
				for(i=0;i<count;i++) {
					EchoChar(InputBuffer[i]) ;
					ichar++ ;
					x = xBegPos + 
					XTextWidth(gFontStructure,buffer,ichar) ;
					gCurrentTextPos.x = x ;
				}
				strcpy(gCurrentText,oldCurrentText) ;
				strcat(gCurrentText,buffer) ;
				SetRect(&pRect,gCurrentTextPos.x,
				gCurrentTextPos.y-height+2,
				gCurrentTextPos.x+6,
				gCurrentTextPos.y+1) ;
				FillRect(&pRect) ;
			}        
			else {
				switch(keysym) {
					case XK_Return:
					case XK_KP_Enter:
					case XK_Linefeed:
						EraseRect(&pRect) ;
						buffer[ichar] = EOS ;
						return (buffer) ;
						break ;
					case XK_space:
					case XK_comma:
						strcat(buffer,InputBuffer) ;
						EraseRect(&pRect) ;
						EchoChar(buffer[ichar++]) ;
						buffer[ichar] = EOS ;
						x = xBegPos + 
						XTextWidth(gFontStructure,buffer,ichar) ;
						gCurrentTextPos.x = x ;
						ichar-- ;
						if (ichar > 0) {
						buffer[ichar] = EOS ;
						strcpy(gCurrentText,oldCurrentText) ;
						strcat(gCurrentText,buffer) ;
						return (buffer) ;
						}
						SetRect(&pRect,gCurrentTextPos.x,
						gCurrentTextPos.y-height+2,
						gCurrentTextPos.x+6,
						gCurrentTextPos.y+1) ;
						FillRect(&pRect) ;
						strcpy(gCurrentText,oldCurrentText) ;
						strcat(gCurrentText,buffer) ;
						break ;
					case XK_Delete:
					case XK_BackSpace  :
						EraseRect(&pRect) ;
						lastCharString[0] = buffer[ichar-1] ;
						lastCharString[1] = EOS ;
						buffer[ichar] = EOS ;
						width = XTextWidth(gFontStructure,
						lastCharString,1) ;
						if (ichar > 0) {
							buffer[--ichar] = EOS ;
							x = xBegPos + 
							XTextWidth(gFontStructure,buffer,ichar) ;
							SetRect(&tRect,x,y-height,
							x+width,y+descent) ;
							gCurrentTextPos.x = x ;
							EraseRect(&tRect) ;
						}
						SetRect(&pRect,gCurrentTextPos.x,
						gCurrentTextPos.y-height+2,
						gCurrentTextPos.x+6,
						gCurrentTextPos.y+1) ;
						FillRect(&pRect) ;
						strcpy(gCurrentText,oldCurrentText) ;
						strcat(gCurrentText,buffer) ;
						break ;
					default:
						break ;
					}
				}
			}
	}
	return (buffer) ;
}

// ------------------------------------------
// Echo a character on the virtual display  
// ------------------------------------------
void EchoChar(char c) {
	char st[2] ;

	sprintf(st,"%c",c) ;
	XDrawString(gDisplay,gCurrentPort->GetDrawable(),gCurrentPort->GetGC(),
			gCurrentTextPos.x,gCurrentTextPos.y,st,strlen(st)) ;
	XFlush(gDisplay) ;
}

// -----------------------------------------------------------
// Routine to read a string from the display. The only problem 
// at this time is that spaces will not terminate the string  
// ----------------------------------------------------------- 
long ReadString(char *buffer,long numChar) {

	Rect	tRect,pRect ;
	long 	ichar ;
	long 	x,y,height,width,descent ;
	long 	xBegPos,yBegPos,count,i,numChars ;
	XEvent 	theEvent ;
	KeySym 	keysym ;
	XComposeStatus compose ;
	char 	InputBuffer[80],lastCharString[2] ;
	char 	oldCurrentText[132] ;

	/* Zero the string before filling it */
	for (ichar=0;ichar<numChar;buffer[ichar++] = EOS) ;

	strcpy(oldCurrentText,gCurrentText) ;

	/* Locate the current text position */
	x = gCurrentTextPos.x ;
	y = gCurrentTextPos.y ;
	height = gFontStructure->ascent ;
	descent = gFontStructure->descent ;
	xBegPos = x ;
	yBegPos = y ;
	SetRect(&pRect,gCurrentTextPos.x,
	gCurrentTextPos.y-height+2,
	gCurrentTextPos.x+6,
	gCurrentTextPos.y+1) ;
	FillRect(&pRect) ;

	compose.compose_ptr = (char *)0 ;
	compose.chars_matched = 0 ;
	for (ichar = 0 ;ichar < numChar; ) {
		XNextEvent(gDisplay,&theEvent) ;
		if (theEvent.type == KeyPress) {
			count = XLookupString((XKeyEvent *)&theEvent,
			InputBuffer,80,
			&keysym,&compose) ;
			InputBuffer[count] = EOS ;
			if (keysym == XK_Return || 
					keysym == XK_KP_Enter || 
					keysym == XK_Linefeed) {
				buffer[ichar] = EOS ;
				EraseRect(&pRect) ;
				return (ichar) ;
			}
			else if (keysym == XK_Delete ||
					keysym == XK_BackSpace) {
				EraseRect(&pRect) ;
				lastCharString[0] = buffer[ichar-1] ;
				lastCharString[1] = EOS ;
				buffer[ichar] = EOS ;
				width = XTextWidth(gFontStructure,
				lastCharString,1) ;
				if (ichar > 0) {
					buffer[--ichar] = EOS ;
					x = xBegPos + 
					XTextWidth(gFontStructure,buffer,ichar) ;
					SetRect(&tRect,x,y-height,
					x+width,y+descent) ;
					gCurrentTextPos.x = x ;
					EraseRect(&tRect) ;
				}
				SetRect(&pRect,gCurrentTextPos.x,
				gCurrentTextPos.y-height+2,
				gCurrentTextPos.x+6,
				gCurrentTextPos.y+1) ;
				FillRect(&pRect) ;
				strcpy(gCurrentText,oldCurrentText) ;
				strcat(gCurrentText,buffer) ;
			}
			else if ((keysym >= XK_Shift_L) &&
					(keysym <= XK_Hyper_R))
				; // Nothing, Modifier key 
			else if ((keysym >= XK_F1) && (keysym <= XK_F35))
				; // Nothing, function key
			else {
				EraseRect(&pRect) ;
				numChars = count ;
				strcat(buffer,InputBuffer) ;
				for(i=0;i<numChars;i++) {
					EchoChar(InputBuffer[i]) ;
					ichar++ ;
					x = xBegPos + 
					XTextWidth(gFontStructure,buffer,ichar) ;
					gCurrentTextPos.x = x ;
				}
				SetRect(&pRect,gCurrentTextPos.x,
				gCurrentTextPos.y-height+2,
				gCurrentTextPos.x+6,
				gCurrentTextPos.y+1) ;
				FillRect(&pRect) ;
				strcpy(gCurrentText,oldCurrentText) ;
				strcat(gCurrentText,buffer) ;
			}
		}
	}
	return (ichar) ;
}

char *get_snumber() {

	Rect tRect;
	long ichar;
	long x,y,height,width,descent;
	long xBegPos,yBegPos,count;
	XEvent theEvent;
	KeySym keysym;
	XComposeStatus compose;
	char InputBuffer[80],lastCharString[2];
	char oldCurrentText[132];

	for (ichar=0; ichar<80; buffer[ichar++] = EOS);
	strcpy(oldCurrentText,gCurrentText);

	x = gCurrentTextPos.x;
	y = gCurrentTextPos.y;
	height = gFontStructure->ascent;
	descent = gFontStructure->descent;
	xBegPos = x;
	yBegPos = y;

	for (ichar=0; ichar<80; ) {
		XNextEvent(gDisplay,&theEvent);
		if (theEvent.type == KeyPress) {
			count = XLookupString((XKeyEvent *)&theEvent,InputBuffer,80,
					&keysym,&compose);
			buffer[ichar]=InputBuffer[0];
			if (buffer[ichar] ==COMMA)  buffer[ichar] =SPACE;
			if ((buffer[ichar]<='9' && buffer[ichar] >='0')
					|| buffer[ichar] == PERIOD || buffer[ichar] =='-' || 
					buffer[ichar] == SPACE) {
				EchoChar(buffer[ichar++]);
				buffer[ichar] = EOS;
				x = xBegPos+ XTextWidth(gFontStructure,buffer,ichar);
				gCurrentTextPos.x = x;
				strcpy(gCurrentText,oldCurrentText);
				strcat(gCurrentText,buffer);
			}
			else {
				switch(buffer[ichar]) {
					case CR : 
						buffer[ichar] = '\0';
						return(buffer); 
						break;
					case DEL :
					case BS :
						lastCharString[0] = buffer[ichar-1];
						lastCharString[1] = EOS;
						buffer[ichar] = EOS;
						width = XTextWidth(gFontStructure,lastCharString,1);
						if (ichar>0) {
							buffer[--ichar] = EOS;
							x = xBegPos + XTextWidth(gFontStructure,buffer,ichar);
							SetRect(&tRect,x,y-height,x+width,y+descent);
							gCurrentTextPos.x=x;
							EraseRect(&tRect);
						}
						break;
					default:
						buffer[ichar] = '\0';
						break;
				}
			}
		}
	}
	return(buffer);
}
     
