// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <stdlib.h>
#include <math.h>
#include "funcdefs.h"
#include "header.h"
#include "menudef.h"
#include "atoms.h"
#include "params.h"
#include "PhasegratingRecord.h"
#include "displ.h"
#include "MslicRecord.h"
#include "microscopes.h"
#define IDY 5
extern long LAYERED,gSpacegroupNumber,gNumSymops,gNumUnitCellAtoms,gNumBasisAtoms,
           gCalculationIsChanged,gPGisChanged,gMSisChanged,gIMisChanged,gSymopsAreChanged,NUM_THICK,
           NUM_DEFOC,gSecondSetting ;
extern char title[],micname[],plots[],stext[] ;
#define PG_MASK	1
#define MS_MASK	2
#define IM_MASK	4
extern struct { long status ; } calculation ;
// ------------------------------------------------------------------
// Function to modify a chosen input parameter in menu1 - data in 
// input  --.at file                                             
// ------------------------------------------------------------------
void modify_par(long i) {

	long j,TTFS,index ;
	char c,buffer[40],*cp ;
	float pv,ftemp ; 

	i-- ; /* set back 1 because of CHANGE */
	if (LAYERED) {
		switch(i) {
			case 1 : /* Space group number */
			case 4 : /* C-axis             */
			case 5 : /* Alpha              */
			case 6 : /* BETA               */
			case 9 : /* ZONE - axis        */
			case 10: /* Number of diff. Layers */
			case 11: /* Number of slices total */
			case 12: /* No assignment          */
			case 13: /* No assignment          */
				return ;
				break ;
			default:
				break ;  /* Just continue */
		}
	}
	eraser((*gCurrentButton)->xl,(*gCurrentButton)->yl-4,(*gCurrentButton)->xh,(*gCurrentButton)->yh) ;
	REVERSE_VIDEO ;
	temdtx((*gCurrentButton)->xl,(*gCurrentButton)->yl+IDY,(*gCurrentButton)->txt) ;
	switch(i) {
		case 0 :   /* TITLE OF THE STRUCTURE */
			ReadString(title,40) ;
			gInputParameter[0].cv = title ;
			NORMAL_VIDEO ;
			break ;
		case 1 :  /* SPACE GROUP NUMBER */
			gSpacegroupNumber = get_spcgrp_number() ;
			NORMAL_VIDEO ;
			gInputParameter[i].iv = gSpacegroupNumber ;
			if ( gSpacegroupNumber != 0 )
				assm_spcgrp(gSpacegroupNumber) ;
			else {
				temdtx(0,130,
				" Enter the number of symmetry operators -> ") ;
				gNumSymops = get_int(1) ;
				get_symop() ;
				NORMAL_VIDEO ;
			}
			gInputParameter[10].iv = gNumSymops ;
			/* rewrite the # of symmetry operators */
			gCurrentButton += 9 ;
			write_par(10) ;
			gCurrentButton -= 9 ;
			gSymopsAreChanged = TRUE ;
			ApplySymmetries() ;
			gPGisChanged = TRUE ;
			/* Change B if appropriate */
			if (gSpacegroupNumber >= 75) {
				gCurrentButton += 2 ;
				gInputParameter[i+2].fv = gInputParameter[i+1].fv ;
				write_par(i+2) ;
				gCurrentButton -= 2 ;
			}
			/* And also C if appropriate */
			/* Cubic system */
			if (gSpacegroupNumber >= 195 ) {
				gCurrentButton += 3 ;
				gInputParameter[i+3].fv = gInputParameter[i+1].fv ;
				write_par(i+3) ;
				gCurrentButton -= 3 ;
			}
			/* Trigonal with Rhombohedral axes */
			else if (gSpacegroupNumber <= 167 && gSpacegroupNumber >= 143) {
				if (gSecondSetting /* Rhombohedral */) {
					gInputParameter[i+3].fv = gInputParameter[i+1].fv ;
					gCurrentButton += 3 ;
					write_par(i+3) ;
					gCurrentButton -= 3 ;
				}
			}
			/* ... and ALPHA and BETA  */
			/* Everything but monoclinic and triclinic */
			if (gSpacegroupNumber >= 16) {
				gInputParameter[5].fv = 90. ;
				gCurrentButton += 4 ;
				write_par(5) ;
				gCurrentButton += 1 ;
				gInputParameter[6].fv = 90. ;
				write_par(6) ;
				gCurrentButton -= 5 ;
			}
			/* ... and possibly GAMMA */
			/* Cubic,Orthorombic or Tetragonal spacegroup */
			if (gSpacegroupNumber >= 195 ||
			(gSpacegroupNumber > 15 && gSpacegroupNumber < 143)) {
				gInputParameter[7].fv = 90. ;
				gCurrentButton += 6 ;
				write_par(7) ;
				gCurrentButton -= 6 ;
			}
			/* Hexagonal setting */
			else if (gSpacegroupNumber > 167 && gSpacegroupNumber < 195) {
				gInputParameter[7].fv = 120. ;
				gCurrentButton += 6 ;
				write_par(7) ;
				gCurrentButton -= 6 ;
			}
			/* Trigonal system */
			else if (gSpacegroupNumber <= 167 && gSpacegroupNumber >= 143) {
				if (!gSecondSetting /* Hexagonal */) {
					gInputParameter[7].fv = 120. ;
					gCurrentButton += 6 ;
					write_par(7) ;
					gCurrentButton -= 6 ;
				}
				else /* Rhombohedal */ {
					gInputParameter[7].fv = gInputParameter[5].fv ;
					gCurrentButton += 6 ;
					write_par(7) ;
					gCurrentButton -= 6 ;
				}
			}
			ViewNewUnitCell() ;
			break ;
		case 2 :  /* LENGTH OF THE A - AXIS */
			gInputParameter[i].fv = get_float(3.) ;
			NORMAL_VIDEO ;
			/* Change also B if appropriate */
			if (gSpacegroupNumber >= 75) {
				gInputParameter[i+1].fv = gInputParameter[i].fv ;
				gCurrentButton += 1 ;      
				write_par(i+1) ;
				gCurrentButton -= 1 ;      
			}
			/* And also C if appropriate */
			/* Cubic system */
			if (gSpacegroupNumber >= 195 ) {
				gInputParameter[i+2].fv = gInputParameter[i].fv ;
				gCurrentButton += 2 ;      
				write_par(i+2) ;
				gCurrentButton -= 2 ;      
			}
			/* Trigonal with Rhombohedral axes */
			else if (gSpacegroupNumber <= 167 && gSpacegroupNumber >= 143) {
				if (gSecondSetting /* Rhombohedral */) {
					gInputParameter[i+2].fv = gInputParameter[i].fv ;
					gCurrentButton += 2 ;      
					write_par(i+2) ;
					gCurrentButton -= 2 ;      
				}
			}
			if (!LAYERED) {
				gPGisChanged = TRUE ;
				ViewNewUnitCell() ;
			}
			break ;
		case 3 :  /* LENGTH OF THE B - AXIS */
			gInputParameter[i].fv = get_float(gInputParameter[2].fv) ;
			NORMAL_VIDEO ;
			if (!LAYERED) {
				gPGisChanged = TRUE ;
				ViewNewUnitCell() ;
			}
			break ;
		case 4 :  /* LENGTH OF THE C - AXIS */
			gInputParameter[i].fv = get_float(gInputParameter[2].fv) ;
			NORMAL_VIDEO ;
			/* Reset the number of slices/cell */
			set_nz() ;
			gCurrentButton += 7 ;
			write_par(11) ;
			if (gInputParameter[11].iv != 1) BELL ;
			gCurrentButton -= 7 ;
			gPGisChanged = TRUE ;
			ViewNewUnitCell() ;
			break ;
		case 5 :  /* ALPHA IN DEGREES   */
			gInputParameter[i].fv = get_float(90.) ;
			if (gSpacegroupNumber <= 167 && gSpacegroupNumber >= 143 && gSecondSetting) {
				gInputParameter[i+1].fv = gInputParameter[i].fv ;
				gCurrentButton++ ;
				write_par(i+1) ;
				gCurrentButton++ ;
				gInputParameter[i+2].fv = gInputParameter[i].fv ;
				write_par(i+2) ;
				gCurrentButton -= 2 ;
			}
			NORMAL_VIDEO ;
			if (!LAYERED) {
				gPGisChanged = TRUE ;
				ViewNewUnitCell() ;
			}
			break ;
		case 6 :  /* BETA IN DEGREES    */
		case 7 :  /* GAMMA IN DEGREES   */
			gInputParameter[i].fv = get_float(90.) ;
			NORMAL_VIDEO ;
			if (!LAYERED) {
				gPGisChanged = TRUE ;
				ViewNewUnitCell() ;
			}
			break ;
		case 8 :  /* GMAX; MAXIMUM RECIPR. LATT. VECT. */
			gInputParameter[i].fv = get_float(2.) ;
			NORMAL_VIDEO ;
			if (!LAYERED) {
				set_nz() ;
				gCurrentButton += 3 ;
				write_par(11) ;
				if (gInputParameter[11].iv != 1) BELL ;
				gCurrentButton -= 3 ;
				gPGisChanged = TRUE ;
			}
			break ;
		case 9  :        /* ZONE AXIS                         */
			gPhsgrtRec->iuvw[0] = get_int(0) ;
			gPhsgrtRec->iuvw[1] = get_int(0) ;
			gPhsgrtRec->iuvw[2] = get_int(1) ;
			NORMAL_VIDEO ;
			set_nz() ;
			gCurrentButton += 2 ;
			write_par(11) ;
			if (gInputParameter[11].iv != 1) BELL ;
			gCurrentButton -= 2 ;
			gPGisChanged = TRUE ;
			ViewNewUnitCell() ;
			break ;
		case 10 :        /* NUMBER OF SYMMETRY OPERATORS      */
			NORMAL_VIDEO ;
			break ;
		case 11 :        /* NUMBER OF SLICES PER UNIT CELL    */
			gInputParameter[i].iv = get_int(1) ;
			NORMAL_VIDEO ;
			break ;
		case 12 :        /* NUMBER OF ATOMS IN THE BASIS      */
		case 13 :        /* NUMBER OF DIFFERENT BASIS ATOMS   */
			NORMAL_VIDEO ;
			break ;
		case 14 : /* MICROSCOPE NAME */
			ReadString(micname,10) ;
			if (strlen(micname) == 0)
				strcpy(micname,"ARM") ;
			NORMAL_VIDEO ;
			gInputParameter[i].cv = micname ;
			index = set_microscope(micname) ;
			if (index != -1)   /* This is a "known" microscope */ {				 
				/* Set the associated parameters */
				gInputParameter[i+1].fv = microscope[index].sph_ab ;
				gInputParameter[i+2].fv = microscope[index].delta ;
				gInputParameter[i+3].fv = microscope[index].theta ;
				gInputParameter[i+4].fv = microscope[index].voltage ;
				gCurrentButton++ ; write_par(i+1) ;
				gCurrentButton++ ; write_par(i+2) ;
				gCurrentButton++ ; write_par(i+3) ;
				gCurrentButton++ ; write_par(i+4) ;
				gCurrentButton -= 4 ;
				if (!LAYERED) /* Then also set number of slices */ {
					set_nz() ;
					gCurrentButton -= 3 ;
					write_par(11) ;
					if (gInputParameter[11].iv != 1) BELL ;
					gCurrentButton += 3 ;
				}
			}
			gMSisChanged = TRUE ;
			break ;
		case 15 : /* SPHERICAL ABBERATION IN MM     */
		case 16 : /* SPREAD OF DEFOCUS IN ANGSTR.   */
		case 17 : /* BEAM DIVERGENCE (HALF) IN MRAD */
			gInputParameter[i].fv = get_float(0.) ;
			NORMAL_VIDEO ;
			gIMisChanged = TRUE ;
			break ;
		case 18 :         /* MICROSCOPE VOLTAGE                */
			gInputParameter[i].fv = get_float(200.) ;
			NORMAL_VIDEO ;
			if (!LAYERED) {
				set_nz() ;
				gCurrentButton -= 7 ;
				write_par(11) ;
				if (gInputParameter[11].iv != 1) BELL ;
				gCurrentButton += 7 ;
			}
			gMSisChanged = TRUE ;
			break ;
		case 19 : /* CENTER OF LAUE CIRCLE (IN UNITS OF H) */
		case 20 : /* CENTER OF LAUE CIRCLE (IN UNITS OF K) */
			gInputParameter[i].fv = get_float(0.) ;
			NORMAL_VIDEO ;
			gMSisChanged = TRUE ;
			break ;
		case 22 : /* AMPLITUDES YES/NO */
			gMsliceRec.iplot = FALSE ;
			ReadString(plots,4) ;
			NORMAL_VIDEO ;
			gInputParameter[i].cv = plots ;
			eraser(520,270,1020,470) ;
			if (plots[0]=='y' || plots[0]=='Y') {
				REVERSE_VIDEO ;
				gMsliceRec.iplot = TRUE ;
				temdtx(0,130,
					" Enter the number of reflections to be output -> ") ;
				gMsliceRec.nw = get_int(1) ;
				eraser(0,120,1020,160) ;
				for(j=0;j<gMsliceRec.nw;j++) {
					sprintf(stext,
						"Enter index h k l for reflection # %d : ",
						j+1);
					temdtx(0,130,stext) ;
					gMsliceRec.ihw[j] = get_int(0) ;
					gMsliceRec.ikw[j] = get_int(0) ;
					gMsliceRec.ilw[j] = get_int(0) ;
					while(!in_zone(gMsliceRec.ihw[j],gMsliceRec.ikw[j],
							gMsliceRec.ilw[j],gPhsgrtRec->iuvw)) {
						eraser(0,120,1020,160) ;
						temdtx(0,130,
						"Reflection not in zone, re-enter -> ") ;
						gMsliceRec.ihw[j] = get_int(0) ;
						gMsliceRec.ikw[j] = get_int(0) ;
						gMsliceRec.ilw[j] = get_int(0) ;
					} ;
					eraser(0,120,1020,160) ;
				}
				NORMAL_VIDEO ;
				temdtx(520,450,
					"The indices are :     h    k    l"); 
				for(j=0;j<gMsliceRec.nw;j++) {
					sprintf(stext,
						"%3d %3d %3d",
						gMsliceRec.ihw[j],gMsliceRec.ikw[j],gMsliceRec.ilw[j]);
					temdtx(664,420-j*24,stext) ;
				}
			}
			gMSisChanged = TRUE ;
			break ;
		case 21 :       /* FOIL THICKNESS                    */
		case 23 :       /* OBJECTIVE LENS DEFOCUS            */
			cp = buffer ;
			TTFS = FALSE ;
			ReadString(buffer,40) ;
			NORMAL_VIDEO ;
			if (strchr(buffer,'(') != 0 || strchr(buffer,',') != 0) 
				TTFS = TRUE ;
			if (!TTFS) {
				gInputParameter[i].fv = atof(buffer) ;
				if (i==21) {
					gMsliceRec.t1 = gInputParameter[i].fv ;
					gMsliceRec.t2 = gMsliceRec.t1  ;
					gMsliceRec.dt = 0. ;
					gMsliceRec.tts = FALSE ;
					NUM_THICK = 1 ;
					gMSisChanged = TRUE ;
				}
				else {
					gDisplayRec.f1 = gInputParameter[i].fv ;
					gDisplayRec.f2 = gDisplayRec.f1 ;
					gDisplayRec.df = 0. ;
					gDisplayRec.tfs = FALSE ;
					NUM_DEFOC = 1 ;
					gIMisChanged = TRUE ;
				}
			}
			else {
				if (i==21) {
					get_ttfs(buffer,
					&gMsliceRec.t1,&gMsliceRec.t2,&gMsliceRec.dt) ;
					gMsliceRec.tts = TRUE ;
					gInputParameter[i].fv = gMsliceRec.t2 ;
					NUM_THICK = (gMsliceRec.t2-gMsliceRec.t1)/gMsliceRec.dt + 1 ;
					gMSisChanged = TRUE ;
				}
				else {
					get_ttfs(buffer,&gDisplayRec.f1,&gDisplayRec.f2,&gDisplayRec.df) ;
					gDisplayRec.tfs = TRUE ;
					gInputParameter[i].fv = gDisplayRec.f2 ;
					ftemp = (gDisplayRec.f2-gDisplayRec.f1)/gDisplayRec.df ;
					ftemp = (ftemp > 0. ? ftemp : -ftemp) ;
					NUM_DEFOC = (long)ftemp + 1 ;
					gIMisChanged = TRUE ;
				}
			}
			break ;
		case 24 : /* SIZE OF THE OBJECTIVE LENS APERTURE IN REC. A */
		case 25 : /* CENTER OF OBJECTIVE APERTURE IN UNITS OF H */
		case 26 : /* CENTER OF OBJECTIVE APERTURE IN UNITS OF K */
		case 27 : /* CENTER OF OPTIC AXIS IN UNITS OF H */
		case 28 : /* CENTER OF OPTIC AXIS IN UNITS OF K */
			gInputParameter[i].fv = get_float(0.) ;
			NORMAL_VIDEO ;
			gIMisChanged = TRUE ;
			break ;
	}
	write_par(i) ;
	if (gPGisChanged)
		calculation.status &= ~PG_MASK ;/* Phasegrating needs to be run */
	else if (gMSisChanged)
		calculation.status &= ~MS_MASK ;/* Multislice needs to be run	*/
	else if (gIMisChanged)
		calculation.status &= ~IM_MASK ;/* Image needs to be run	*/
	gCalculationIsChanged = TRUE ;
}
