// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include "header.h"
#include "funcdefs.h"
#include "menudef.h"
#include "errors.h"
#include "filenames.h"
#define PG_MASK	1
#define MS_MASK	2
#define IM_MASK	4
extern struct { long status ; } calculation ;
extern long gCalculationIsChanged,gPGisChanged,gMSisChanged,gIMisChanged ;
extern long LAYERED ;
extern char stext[] ;
extern float RANGE ;
void run_p_ms() ;
Boolean	UpdateCalculation() ;
Boolean	RerunImageCalculation() ;
Boolean	RerunMultisliceCalculation() ;
Boolean	FileExists(char *filename) ;
void	UpdateCalculationStatus() ;

void	UpdateCalculationStatus() {
	if (!FileExists(flname.pout)) {
		gPGisChanged = true ;
	}
	else if (!FileExists(flname.mout)) {
		gMSisChanged = true ;
	}
	else if (!FileExists(flname.iout)) {
		gIMisChanged = true ;
	}
}
Boolean	UpdateCalculation() {
	long ierr ;
	setup_trnsf() ;
	if (!FileExists(flname.pout)) {
		gPGisChanged = true ;
	}
	if (!FileExists(flname.mout)) {
		gMSisChanged = true ;
	}
	if (!FileExists(flname.iout)) {
		gIMisChanged = true ;
	}
	if (gPGisChanged)  {
		if (!LAYERED) {
			gPGisChanged = FALSE ;
			SetWatchCursor() ;
			CalculatePhasegrating(&ierr) ;
			InitCursor() ;
			if (ierr == -1) return false ;
			if (ierr == -5) {
				/* Exceeded the array space	*/
				REVERSE_VIDEO ;
				temdtx(0,130,
				"Maximum array size exceeded. Reduce Gmax or the unit cell !") ;
				NORMAL_VIDEO ;
				BELL ;
				return false ;
			}

			if (gAbortStatus) {
				gAbortStatus = FALSE ;
				REVERSE_VIDEO ;
				temdtx(0,130,"Cancelled !") ;
				NORMAL_VIDEO ;
				return false ;
			}
			if (gIOError) {
				err_io(gIOError) ;
				return false ;
			}
			calculation.status |= PG_MASK ;
			gMSisChanged = TRUE ;
			return RerunMultisliceCalculation() ;
		}
		else {
			err_seq(8) ;
			return false ;
		}
	}
	return RerunMultisliceCalculation() ;	
}

Boolean	RerunMultisliceCalculation() {
	if(gMSisChanged) {
		gMSisChanged = FALSE ;
		SetWatchCursor() ;
		DoMultisliceCalculation() ;
		InitCursor() ;
		if (gAbortStatus) {
			REVERSE_VIDEO ;
			temdtx(0,130,"Cancelled !") ;
			NORMAL_VIDEO ;
			gAbortStatus = FALSE ;
			return false ;
		}
		if (gIOError) {
			err_io(gIOError) ;
			return false ;
		}
		calculation.status |= MS_MASK ;
		gIMisChanged = TRUE ;
	}
	return RerunImageCalculation() ;    
}

Boolean	RerunImageCalculation() {

	long nobeam,ierr ;
	float sum ;

	if (gIMisChanged) {
		gIMisChanged = FALSE ;
		SetWatchCursor() ;
		CalculateImages(&nobeam,&sum,&RANGE,&ierr) ;
		InitCursor() ;
		if (gAbortStatus) {
			REVERSE_VIDEO ;
			temdtx(0,130,"Cancelled !") ;
			NORMAL_VIDEO ;
			gAbortStatus = FALSE ;
			return false ;
		}
		if (gIOError) {
			err_io(gIOError) ;
			return false ;
		}
		if (ierr == -1)
			err_seq(6) ;
		else {
			calculation.status |= IM_MASK ;
			return true ;
		}
	}
	return true ;
}

// --------------------------------------------------------------------------
// Start-up routine for the multislice programs. Activated by the "RUN" 
//  command in the control-menu                                    
// --------------------------------------------------------------------------
void run_prog() {
	eraser(0,120,1020,160) ;
	on_highlight() ;
	gCurrentCommand = gCurrentButton ;
	off_highlight() ;
	gCurrentButton -= 4 ;
	run_p_pg() ;
}

// --------------------------------------------------------------------------
// Control routine for execution of the phasegrating routine when
//  initiated through the "RUN" command in the control menu    
// --------------------------------------------------------------------------
void run_p_pg() {
	long ierr ;
	if (gPGisChanged)  {
		if (!LAYERED) {
			gCurrentCommand = gCurrentButton ;
			on_highlight() ;
			setup_trnsf() ;
			gPGisChanged = FALSE ;
			CalculatePhasegrating(&ierr) ;
			off_highlight() ;
			if (ierr == -1) return ;
			if (ierr == -5) {
				/* Exceeded the array space	*/
				REVERSE_VIDEO ;
				temdtx(0,130,
				"Maximum array size exceeded. Reduce Gmax or the unit cell !") ;
				NORMAL_VIDEO ;
				BELL ;
				InitCursor() ;
				return ;
			}

			if (gAbortStatus) {
				gAbortStatus = FALSE ;
				REVERSE_VIDEO ;
				temdtx(0,130,"Cancelled !") ;
				NORMAL_VIDEO ;
				return ;
			}
			if (gIOError) {
				err_io(gIOError) ;
				return ;
			}
			calculation.status |= PG_MASK ;
			gMSisChanged = TRUE ;
			gCurrentButton += 6 ;
			run_p_ms() ;
		}
		else {
			err_seq(8) ;
			return ;
		}
	}
	else {
		gCurrentButton += 6 ;
		run_p_ms() ;
	}
}

// --------------------------------------------------------------------------
// Function to start up the phasegrating routine when activated 
//  through the command "PHSGRT" in the control menu        
// --------------------------------------------------------------------------
void run_phsgrt() {
	long ierr ;

	on_highlight() ;
	gCurrentCommand = gCurrentButton ;
	if (!LAYERED) {
		if (gPGisChanged) {
			setup_trnsf() ;
			gPGisChanged = FALSE ;
			CalculatePhasegrating(&ierr) ;
			off_highlight() ;
			if (ierr == -5) {
				/* Exceeded the array space	*/
				REVERSE_VIDEO ;
				temdtx(0,130,
				"Maximum array size exceeded. Reduce Gmax or the unit cell !") ;
				NORMAL_VIDEO ;
				BELL ;
				InitCursor() ;
				return ;
			}
			if (gAbortStatus) {
				gAbortStatus = FALSE ;
				REVERSE_VIDEO ;
				temdtx(0,130,"Cancelled !") ;
				NORMAL_VIDEO ;
				return ;
			}
			if (ierr == -1)
				return ;
			if (gIOError) {
				err_io(gIOError) ;
				return ;
			}
			calculation.status |= PG_MASK ;
			gMSisChanged = TRUE ;
		}
		else  {
			off_highlight() ;
			err_seq(0) ;
		}
	}
	else {
		off_highlight() ;
		if (LAYERED)
		err_seq(8) ;
		else 
		err_seq(10) ;
	}
	InitCursor() ;
}

// --------------------------------------------------------------------------
// Runs the msliceroutine when invoked by the "MSLICE" command  
// --------------------------------------------------------------------------
void run_mslice() {
	on_highlight() ;
	gCurrentCommand = gCurrentButton ;
	if(gMSisChanged && !gPGisChanged) {
		eraser(0,120,1020,160) ;
		setup_trnsf() ;
		gMSisChanged = FALSE ;
		DoMultisliceCalculation() ;
		off_highlight() ;
		if (gAbortStatus) {
			REVERSE_VIDEO ;
			temdtx(0,130,"Cancelled !") ;
			NORMAL_VIDEO ;
			gAbortStatus = FALSE ;
			return ;
		}
		if (gIOError) {
			err_io(gIOError) ;
			return ;
		}
		calculation.status |= MS_MASK ;
		gIMisChanged = TRUE ;
	}
	else if (gMSisChanged && gPGisChanged) {
		off_highlight() ;
		if (LAYERED)
			err_seq(8) ;
		else 
			err_seq(10) ;
	}
	else {
		off_highlight() ;
		err_seq(1) ;
	}
	InitCursor() ;
}
  
// --------------------------------------------------------------------------
// runs the msliceroutine when invoked through the "RUN" command
// --------------------------------------------------------------------------
void run_p_ms() {

	if(gMSisChanged) {
		gCurrentCommand = gCurrentButton ;
		on_highlight() ;
		setup_trnsf() ;
		gMSisChanged = FALSE ;
		DoMultisliceCalculation() ;
		off_highlight() ;
		if (gAbortStatus) {
			REVERSE_VIDEO ;
			temdtx(0,130,"Cancelled !") ;
			NORMAL_VIDEO ;
			gAbortStatus = FALSE ;
			return ;
		}
		if (gIOError) {
			err_io(gIOError) ;
			return ;
		}
		calculation.status |= MS_MASK ;
		gIMisChanged = TRUE ;
	}
	gCurrentButton -= 5 ;
	run_p_im() ;    
}

// --------------------------------------------------------------------------
// Runs the image routine when invoked by the "RUN" command  
// --------------------------------------------------------------------------
void run_p_im() {

	long nobeam,ierr ;
	float sum ;

	if (gIMisChanged) {
		gCurrentCommand = gCurrentButton ;
		on_highlight() ;
		setup_trnsf() ;
		gIMisChanged = FALSE ;
		CalculateImages(&nobeam,&sum,&RANGE,&ierr) ;
		off_highlight() ;
		if (gAbortStatus) {
			REVERSE_VIDEO ;
			temdtx(0,130,"Cancelled !") ;
			NORMAL_VIDEO ;
			gAbortStatus = FALSE ;
			return ;
		}
		if (gIOError) {
			err_io(gIOError) ;
			return ;
		}
		if (ierr == -1)
			err_seq(6) ;
		else {
			REVERSE_VIDEO ;
			sprintf(stext,
			" INTEGRATED INTENSITY FOR %d BEAMS IS %f ",nobeam,sum) ;
			temdtx(0,130,stext) ;
			NORMAL_VIDEO ;
			calculation.status |= IM_MASK ;
		}
	}
	else err_seq(5) ;
}

// --------------------------------------------------------------------------
// Runs the routine Image when invoked by the "IMAGE" command    
// --------------------------------------------------------------------------
void run_image() {
	long nobeam,ierr ;
	float sum ;

	on_highlight() ;
	gCurrentCommand = gCurrentButton ;
	if (gIMisChanged && !gPGisChanged && !gMSisChanged) {
		eraser(0,120,1020,160) ;
		setup_trnsf() ;
		gIMisChanged = FALSE ;
		CalculateImages(&nobeam,&sum,&RANGE,&ierr) ;
		off_highlight() ;
		if (gIOError)  {
			err_io(gIOError) ;
			return ;
		}
		if (ierr == -1)
			err_seq(6) ;
		else {
			REVERSE_VIDEO ;
			sprintf(stext," INTEGRATED INTENSITY FOR %d BEAMS IS %f ",nobeam,sum) ;
			temdtx(0,130,stext) ;
			NORMAL_VIDEO ;
			calculation.status |= IM_MASK ;
		}
	}
	else if (gIMisChanged && gMSisChanged && !gPGisChanged) {
		off_highlight() ;
		err_seq(11) ;
	}
	else if (gIMisChanged && gPGisChanged) {
		off_highlight() ;
		if (LAYERED)
			err_seq(8) ;
		else 
			err_seq(10) ;
	}
	else {
		off_highlight() ;
		err_seq(2) ;
	}
	InitCursor() ;
}

void ReportFileOpenError(char *filename) {
	char message[256] ;
	extern Boolean gNoGUI ;

	sprintf(message,"Unable to topen the file '%s' !",filename) ;
	if (gNoGUI) {
		write_message(message) ;
		return ;
	}
	REVERSE_VIDEO ;
	write_message(message) ;
	NORMAL_VIDEO ;
	BELL ;
}
