// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include "menudef.h"
#include "MslicRecord.h"
#include "displ.h"
#include <math.h>
#include "ttypes.h"
#include "devices.h"
#include "funcdefs.h"
#include "graphics.h"
extern GDevice *OutDevice ;
extern struct { float wl ; long nslice ; float sigma ; } gMSRec ;
extern struct { long ctf_autoscale ; float ctf_umax ; } ctf ;

void ctfplot() {
	float u[500],yctf[500],ydiv[500],ctftot[500],ychrm[500],ydamp[500],
	ydampn[500],zrl,zim,*rp1,*rp2,pi=3.14159265,cs,th,df,wl,tst ;
	float xmax,xmin,ymax,ymin,umax,usq,del,deltau,arg,arg1,x,y,phase,size ;
	char ics[40],idf[40],ivolt[40],ith[40],idel[40] ;
	char buffer[132] ;
	long i,j,npoints,npt,mark_int ;

	cs = gDisplayRec.cs*1.e+7 ;
	th = gDisplayRec.div*1.e-3 ;
	th = 0.77*th ;			/* Gaussian th = 0.77*bessel */
	df = gDisplayRec.dfocus ;
	del = gDisplayRec.del ;
	wl = 12.263652/sqrt((gMsliceRec.volt+9.785025e-4*
	gMsliceRec.volt*gMsliceRec.volt)*1000.) ;

	npoints = 0 ;
	npt = 500 ;
	if (ctf.ctf_autoscale)
		umax = 1. ;
	else
		umax = ctf.ctf_umax ;

	deltau = umax/npt ;
	for(i=0;i<npt;i++) {
		u[i] = i*deltau ;
		usq = u[i]*u[i] ;
		arg = pi*wl*usq*(df+0.5*cs*wl*wl*usq) ;
		yctf[i] = sin(arg) ;
		arg1 = pi*wl*del*usq ;
		arg1 = -0.5*arg1*arg1 ; 
		ychrm[i] = exp(arg1) ;
		x = pi*th*u[i]*(df + cs*wl*wl*usq) ;
		y = pi*th*u[i]*pi*wl*del*del*usq ;
		zrl = x*x - y*y ;
		zim = 2.*x*y ;
		tst = zrl ;
		if (!ctf.ctf_autoscale || tst < 4.) {
			npoints++ ;
			zrl = exp(-zrl) ;            /*  Z = exp(zlr+izim) */
			zrl = zrl*cos(zim) ;
			zim = -zrl*sin(zim) ;
			ydiv[i] = sqrt(zrl*zrl+zim*zim) ;
			if (ydiv[i] != 0.) 
				phase = atan2(zim,zrl) ;
			else
				phase = 0. ;
			phase += arg ;
			ydamp[i] = ychrm[i]*ydiv[i] ;
			ydampn[i] = -ydamp[i] ;
			ctftot[i] = sin(phase)*ychrm[i]*ydiv[i] ;
		} 
	}             
	if (ctf.ctf_autoscale)
		umax = u[npoints-1] ;
	else
		umax = ctf.ctf_umax ;

	SetLineWidth(1) ;
	physor(0.15,0.30) ;
	area2d(0.7,0.6) ;
	frame() ;
	size = 1.5 ;
	strcpy(buffer,"Contrast Transfer Function") ;
	headin(buffer,size) ;
	size = 1.4 ;
	strcpy(buffer,"CTF") ;
	yname(buffer,size) ;
	eushft('!') ;
	instrct('%') ;
	sprintf(ics,"Spher. Abber.  = %3.1f mm",gDisplayRec.cs) ;
	sprintf(idf,"Obj. Lens Def.  = %6.0f !D",df) ;
	sprintf(ivolt,"Accel.Voltage  = %6.1f kV",gMsliceRec.volt) ;
	sprintf(ith,"Div. Angle [Half]  = %4.2f mrad",gDisplayRec.div) ;
	sprintf(idel,"Delta  = %4.0f !D",gDisplayRec.del) ;
	size = 1.4 ;
	strcpy(buffer,"Scattering Vector [!D%EH.7-1%EXHX]") ;
	xname(buffer,size) ;
	xmin = 0. ;
	xmax = umax ;
	ymin = -1.01 ;
	ymax = 1.01 ;
	graph2d(xmin,xmax,ymin,ymax) ;
	line(0.,0.,xmax,0.) ;
	mark_int = 0 ;
	SetLineStyle(SOLID) ;
	curve(u,yctf,npoints,mark_int) ;
	curve(u,ydamp,npoints,mark_int) ;
	curve(u,ydampn,npoints,mark_int) ;
	SetLineWidth(3) ;
	curve(u,ctftot,npoints,mark_int) ;
	SetLineWidth(1) ;
	messag(0.15,0.18,1.5,0.,ivolt) ;
	messag(0.15,0.15,1.5,0.,ics) ;
	messag(0.15,0.12,1.5,0.,idf) ;
	messag(0.15,0.09,1.5,0.,idel) ;
	messag(0.15,0.06,1.5,0.,ith) ;
}
