// -----------------------------------------------------------------------------
//  Copyright (c) 2001 Roar Kilaas
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// -----------------------------------------------------------------------------
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
//#include <stream.h> Temporary hack

#include "graphics.h"
#include "TCrystal.h"
#include "globals.h"
#include "numbers.h"
#include "TCrystalWindow.h" 

#define PG_MASK	1
#define MS_MASK	2
#define IM_MASK	4

#define NUM_COLORS	32

float covrad[] = {
		.32,.93,1.23,.9,.82,.77,.75,.73,.72,.71,
     		1.54,1.36,1.18,1.11,1.06,1.02,.99,.98,
     		2.03,1.74,1.44,1.32,1.22,1.18,1.17,1.17,1.16,
     		1.15,1.17,1.25,1.26,1.22,1.20,1.16,1.14,1.12,
     		2.16,1.91,1.62,1.45,1.34,1.3,1.27,1.25,1.25,
     		1.28,1.34,1.48,1.44,1.41,1.4,1.36,1.33,1.31,
     		2.35,1.98,1.69,
     		1.65,1.65,1.64,1.63,1.62,1.85,1.61,1.59,1.59,
     		1.58,1.57,1.56,1.56,1.56,
     		1.44,1.34,1.30,1.28,1.26,1.27,1.30,
     		1.34,1.49,1.48,1.47,1.46,1.46,1.45,1.5,
     		2.55,2.1,1.65,1.65,1.5,1.42,1.5,1.5,1.5,1.5,1.5,
		1.5,1.5,1.5} ;

char *chemsymb[] = {"H","He","Li","Be","B","C","N","O","F","Ne","Na",
                     "Mg","Al","Si","P","S","Cl","Ar","K","Ca","Sc","Ti",
                     "V","Cr","Mn","Fe","Co","Ni","Cu","Zn","Ga","Ge","As",
                     "Se","Br","Kr","Rb","Sr","Y","Zr","Nb","Mo","Tc","Ru",
                     "Rh","Pd","Ag","Cd","In","Sn","Sb","Te","I","Xe","Cs",
                     "Ba","La","Ce","Pr","Nd","Pm","Sm","Eu","Gd","Tb","Dy",
                     "Ho","Er","Tm","Yb","Lu","Hf","Ta","W","Re","Os","Ir",
                     "Pt","Au","Hg","Tl","Pb","Bi","Po","At","Rn","Fr","Ra",
                     "Ac","Th","Pa","U"} ;

extern  void SetZone(TVector &zone) ;
extern  void SetNearestZone(float *x,float *y,float *z) ;

extern float one(float x);
extern void encode_symop(char *string,struct GENERATOR *sym_op) ;
extern	uLong	gWhitePixel,gBlackPixel,gDarkerGray,gGrayPixel,gGreenPixel,gRedPixel,gBluePixel ;

extern	char 	*ReadLine(char *buffer,long numChars,FILE *fp) ;
extern	void	PutMessage(char *) ;

long gSecondSetting,gAlternateSetting,gCentroSymmetric ;
long generators[] = 			{ 040201,0442201,0440211,042211,010402,
                            0114402,014422,0110422,020104,0220144,
                            0221104,021144,020411,012204,0440102,
                            0224411,0112244,0441122,0220401,010244,
                            041102,024401,0110204,040122,0442211,
                            040211,042201,0440201,0114422,010422,
                            0110402,014402,0221144,021104,020144,
                            0220104,0224401,0110244,041122,020401,
                            010204,040102,024411,0112204,0440122,
                            0220411,012244,0441102,0222601,0464401,
                            0260401,024601,0262211,0444611,0460211,
                            042611,024611,0260411,0464411,0222611,
                            0460201,042601,0262201,0444601} ;
struct translations {
	float value ;
	char  *symbol ;
} translations[] = { 
	{1./2.,"1/2"},{1./3.,"1/3"},
	{2./3.,"2/3"},{1./4.,"1/4"},
	{3./4.,"3/4"},{1./6.,"1/6"},
	{5./6.,"5/6"},{1./8.,"1/8"},
	{3./8.,"3/8"},{5./8.,"5/8"},
	{7./8.,"7/8"}
} ;

char spacegroup[132] ;
static char symb[] = {'x','y','z'} ;

GENERATOR gen[20],temp_sym,temp_gen ;
GENERATOR symop[200] ;
char symop_txt[200][64] ;
char *txtptr_symop[200] ;

GENERATOR inv_gen = {-1.,0.,0.,0.,0.,-1.,0.,0.,0.,0.,-1.,0.,
                            0.,0.,0.,1.} ;
GENERATOR identity_matr = {1.,0.,0.,0.,0.,1.,0.,0.,0.,0.,1.,0.,
                                  0.,0.,0.,1.} ;
GENERATOR p_matr,q_matr ;

typedef struct Generators {
				long generators[64] ;
				} Generators,*GenPtr ;
#ifdef __VMS
#pragma message disable typnolnkvarlnk
#endif
GenPtr GP = (GenPtr)generators ;
#ifdef __VMS
#pragma message enable typnolnkvarlnk
#endif

#define MINUS '-'
#define BLANK ' '
#define COMMA ','
#define	EOS	'\0'
#define	NEWLINE	'\n'

extern char 	*chemsymb[] ;
extern short	gNumGrayLevels ;

RMatrix gTMatrix ; 
short	gNumSymops ;


#define BlackColor 193
#define WhiteColor 0

void PenPat(Pattern *pattern=nil) ;

RGBColor RGBwhite = {65535,65535,65535} ;
RGBColor RGBblack = {0,0,0} ;
RGBColor RGBred = {65535,0,0} ;
RGBColor RGBgreen = {0,65535,0} ;
RGBColor RGBblue = {0,0,65535} ;

long get_spcgrp_number() ;
void assm_spcgrp(long spcgrp_number) ;
void generate_spcgrp(char  *spacegroup) ;
void encode_symop(char *string,struct GENERATOR *sym_op) ;
void apply_gen(struct GENERATOR *oper,struct GENERATOR *source,struct GENERATOR *dest) ;
long ident(struct GENERATOR *p_sym) ;
long new_symop(struct GENERATOR *p_symop,long num_symop) ;
void decode_spcgrp(char *string,long *pnum_gen,long *pinversion) ;
void get_inverse(struct GENERATOR *source,struct GENERATOR *dest) ;
void decode_gen(long rot,long trnsl,struct GENERATOR *sym_op) ;
void decode_symop(long j) ; 
long atomno(char *cp) ;

float	CheckZeroOne(float x) ;
inline  long SIGN(float x) { return (x < 0. ? -1   : 1) ; }
inline  float min(float x,float y) { return (x < y ? x : y) ; }
inline  long min(long x,long y) { return (x < y ? x : y) ; }
inline  float max(float x,float y) { return (x > y ? x : y) ; }
inline  long max(long x,long y) { return (x > y ? x : y) ; }

extern	Boolean	gDebugMode ;

char *gSpacegroupStrings[] = 
	{
		"1 0 1 1 0 0",
		"2 1 1 1 0 0",
		"3 0 1 3 0 0",
		"4 0 1 3 1200 0",
		"5 0 2 1 121200 3 0 0",
		"6 0 1 27 0 0",
		"7 0 1 27 12 0",
		"8 0 2 1 121200 27 0 0",
		"9 0 2 1 121200 27 12 0",
		"10 1 1 3 0 0",
		"11 1 1 3 1200 0",
		"12 1 2 1 121200 3 0 0",
		"13 1 1 3 12 0",
		"14 1 1 3 1212 0",
		"15 1 2 1 121200 3 12 0",
		"16 0 2 2 0 3 0 0",
		"17 0 2 2 12 3 12 0",
		"18 0 2 2 0 3 121200 0",
		"19 0 2 2 120012 3 1212 0",
		"20 0 3 1 121200 2 12 3 12 0",
		"21 0 3 1 121200 2 0 3 0 0",
		"22 0 4 1 1212 1 120012 2 0 3 0 0",
		"23 0 3 1 121212 2 0 3 0 0",
		"24 0 3 1 121212 2 120012 3 1212 0",
		"25 0 2 2 0 27 0 0",
		"26 0 2 2 12 27 12 0",
		"27 0 2 2 0 27 12 0",
		"28 0 2 2 0 27 120000 0",
		"29 0 2 2 12 27 120000 0",
		"30 0 2 2 0 27 1212 0",
		"31 0 2 2 120012 27 120012 0",
		"32 0 2 2 0 27 121200 0",
		"33 0 2 2 12 27 121200 0",
		"34 0 2 2 0 27 121212 0",
		"35 0 3 1 121200 2 0 27 0 0",
		"36 0 3 1 121200 2 12 27 12 0",
		"37 0 3 1 121200 2 0 27 12 0",
		"38 0 3 1 1212 2 0 27 0 0",
		"39 0 3 1 1212 2 0 27 1200 0",
		"40 0 3 1 1212 2 0 27 120000 0",
		"41 0 3 1 1212 2 0 27 121200 0",
		"42 0 4 1 1212 1 120012 2 0 27 0 0",
		"43 0 4 1 1212 1 120012 2 0 27 141414 0",
		"44 0 3 1 121212 2 0 27 0 0",
		"45 0 3 1 121212 2 0 27 121200 0",
		"46 0 3 1 121212 2 0 27 120000 0",
		"47 1 2 2 0 3 0 0",
		"48 0 3 2 0 3 0 25 121212 1 40201 141414 ",
		"49 1 2 2 0 3 12 0",
		"50 0 3 2 0 3 0 25 121200 1 40201 141400 ",
		"51 1 2 2 120000 3 0 0",
		"52 1 2 2 120000 3 121212 0",
		"53 1 2 2 120012 3 120012 0",
		"54 1 2 2 120000 3 12 0",
		"55 1 2 2 0 3 121200 0",
		"56 1 2 2 121200 3 1212 0",
		"57 1 2 2 12 3 1212 0",
		"58 1 2 2 0 3 121212 0",
		"59 0 3 2 0 3 121200 25 121200 1 40201 141400 ",
		"60 1 2 2 121212 3 12 0",
		"61 1 2 2 120012 3 1212 0",
		"62 1 2 2 120012 3 1200 0",
		"63 1 3 1 121200 2 12 3 12 0",
		"64 1 3 1 121200 2 1212 3 1212 0",
		"65 1 3 1 121200 2 0 3 0 0",
		"66 1 3 1 121200 2 0 3 12 0",
		"67 1 3 1 121200 2 1200 3 1200 0",
		"68 0 4 1 121200 2 121200 3 0 25 1212 1 40201 1414 ",
		"69 1 4 1 1212 1 120012 2 0 3 0 0",
		"70 0 5 1 1212 1 120012 2 0 3 0 25 141414 1 40201 989898 ",
		"71 1 3 1 121212 2 0 3 0 0",
		"72 1 3 1 121212 2 0 3 121200 0",
		"73 1 3 1 121212 2 120012 3 1212 0",
		"74 1 3 1 121212 2 1200 3 1200 0",
		"75 0 2 2 0 19 0 0",
		"76 0 2 2 12 19 14 0",
		"77 0 2 2 0 19 12 0",
		"78 0 2 2 12 19 34 0",
		"79 0 3 1 121212 2 0 19 0 0",
		"80 0 3 1 121212 2 121212 19 1214 0",
		"81 0 2 2 0 43 0 0",
		"82 0 3 1 121212 2 0 43 0 0",
		"83 1 2 2 0 19 0 0",
		"84 1 2 2 0 19 12 0",
		"85 0 3 2 0 19 121200 25 121200 1 40201 941400 ",
		"86 0 3 2 0 19 121212 25 121212 1 40201 949494 ",
		"87 1 3 1 121212 2 0 19 0 0",
		"88 0 4 1 121212 2 121212 19 1214 25 1214 1 40201 9498 ",
		"89 0 3 2 0 19 0 3 0 0",
		"90 0 3 2 0 19 121200 3 121200 0",
		"91 0 3 2 12 19 14 3 0 0",
		"92 0 3 2 12 19 121214 3 121214 0",
		"93 0 3 2 0 19 12 3 0 0",
		"94 0 3 2 0 19 121212 3 121212 0",
		"95 0 3 2 12 19 34 3 0 0",
		"96 0 3 2 12 19 121234 3 121234 0",
		"97 0 4 1 121212 2 0 19 0 3 0 0",
		"98 0 4 1 121212 2 121212 19 1214 3 120034 0",
		"99 0 3 2 0 19 0 27 0 0",
		"100 0 3 2 0 19 0 27 121200 0",
		"101 0 3 2 0 19 12 27 12 0",
		"102 0 3 2 0 19 121212 27 121212 0",
		"103 0 3 2 0 19 0 27 12 0",
		"104 0 3 2 0 19 0 27 121212 0",
		"105 0 3 2 0 19 12 27 0 0",
		"106 0 3 2 0 19 12 27 121200 0",
		"107 0 4 1 121212 2 0 19 0 27 0 0",
		"108 0 4 1 121212 2 0 19 0 27 12 0",
		"109 0 4 1 121212 2 121212 19 1214 27 0 0",
		"110 0 4 1 121212 2 121212 19 1214 27 12 0",
		"111 0 3 2 0 43 0 3 0 0",
		"112 0 3 2 0 43 0 3 12 0",
		"113 0 3 2 0 43 0 3 121200 0",
		"114 0 3 2 0 43 0 3 121212 0",
		"115 0 3 2 0 43 0 27 0 0",
		"116 0 3 2 0 43 0 27 12 0",
		"117 0 3 2 0 43 0 27 121200 0",
		"118 0 3 2 0 43 0 27 121212 0",
		"119 0 4 1 121212 2 0 43 0 27 0 0",
		"120 0 4 1 121212 2 0 43 0 27 12 0",
		"121 0 4 1 121212 2 0 43 0 3 0 0",
		"122 0 4 1 121212 2 0 43 0 3 120034 0",
		"123 1 3 2 0 19 0 3 0 0",
		"124 1 3 2 0 19 0 3 12 0",
		"125 0 4 2 0 19 0 3 0 25 121200 1 40201 949400 ",
		"126 0 4 2 0 19 0 3 0 25 121212 1 40201 949494 ",
		"127 1 3 2 0 19 0 3 121200 0",
		"128 1 3 2 0 19 0 3 121212 0",
		"129 0 4 2 0 19 121200 3 121200 25 121200 1 40201 941400 ",
		"130 0 4 2 0 19 121200 3 121212 25 121200 1 40201 941400 ",
		"131 1 3 2 0 19 12 3 0 0",
		"132 1 3 2 0 19 12 3 12 0",
		"133 0 4 2 0 19 121212 3 12 25 121212 1 40201 941494 ",
		"134 0 4 2 0 19 121212 3 0 25 121212 1 40201 941494 ",
		"135 1 3 2 0 19 12 3 121200 0",
		"136 1 3 2 0 19 121212 3 121212 0",
		"137 0 4 2 0 19 121212 3 121212 25 121212 1 40201 941494 ",
		"138 0 4 2 0 19 121212 3 121200 25 121212 1 40201 941494 ",
		"139 1 4 1 121212 2 0 19 0 3 0 0",
		"140 1 4 1 121212 2 0 19 0 3 12 0",
		"141 0 5 1 121212 2 121212 19 1214 3 120034 25 1214 1 40201 1498 ",
		"142 0 5 1 121212 2 121212 19 1214 3 120014 25 1214 1 40201 1498 ",
		"143 0 1 49 0 0",
		"144 0 1 49 13 0",
		"145 0 1 49 23 0",
		"146 0 2 1 231313 49 0 1 54723 0 ",
		"147 1 1 49 0 0",
		"148 1 2 1 231313 49 0 1 54723 0",
		"149 0 2 49 0 16 0 0",
		"150 0 2 49 0 13 0 0",
		"151 0 2 49 13 16 23 0",
		"152 0 2 49 13 13 0 0",
		"153 0 2 49 23 16 13 0",
		"154 0 2 49 23 13 0 0",
		"155 0 3 1 231313 49 0 13 0 1 54723 0",
		"156 0 2 49 0 37 0 0",
		"157 0 2 49 0 40 0 0",
		"158 0 2 49 0 37 12 0",
		"159 0 2 49 0 40 12 0",
		"160 0 3 1 231313 49 0 37 0 1 54723 0",
		"161 0 3 1 231313 49 0 37 12 1 54723 0",
		"162 1 2 49 0 16 0 0 0",
		"163 1 2 49 0 16 12 0",
		"164 1 2 49 0 13 0 0",
		"165 1 2 49 0 13 12 0",
		"166 1 3 1 231313 49 0 13 0 1 54723 0",
		"167 1 3 1 231313 49 0 13 12 1 54723 0",
		"168 0 2 49 0 2 0 0",
		"169 0 2 49 13 2 12 0",
		"170 0 2 49 23 2 12 0",
		"171 0 2 49 23 2 0 0",
		"172 0 2 49 13 2 0 0",
		"173 0 2 49 0 2 12 0",
		"174 0 2 49 0 26 0 0",
		"175 1 2 49 0 2 0 0",
		"176 1 2 49 0 2 12 0",
		"177 0 3 49 0 2 0 13 0 0",
		"178 0 3 49 13 2 12 13 13 0",
		"179 0 3 49 23 2 12 13 23 0",
		"180 0 3 49 23 2 0 13 23 0",
		"181 0 3 49 13 2 0 13 13 0",
		"182 0 3 49 0 2 12 13 0 0",
		"183 0 3 49 0 2 0 37 0 0",
		"184 0 3 49 0 2 0 37 12 0",
		"185 0 3 49 0 2 12 37 12 0",
		"186 0 3 49 0 2 12 37 0 0",
		"187 0 3 49 0 26 0 37 0 0",
		"188 0 3 49 0 26 12 37 12 0",
		"189 0 3 49 0 26 0 13 0 0",
		"190 0 3 49 0 26 12 13 0 0",
		"191 1 3 49 0 2 0 13 0 0",
		"192 1 3 49 0 2 0 13 12 0",
		"193 1 3 49 0 2 12 13 12 0",
		"194 1 3 49 0 2 12 13 0 0",
		"195 0 3 2 0 3 0 5 0 0",
		"196 0 5 1 1212 1 120012 2 0 3 0 5 0 0",
		"197 0 4 1 121212 2 0 3 0 5 0 0",
		"198 0 3 2 120012 3 1212 5 0 0",
		"199 0 4 1 121212 2 120012 3 1212 5 0 0",
		"200 1 3 2 0 3 0 5 0 0",
		"201 0 4 2 0 3 0 5 0 25 121212 1 40201 949494",
		"202 1 5 1 1212 1 120012 2 0 3 0 5 0 0",
		"203 0 6 1 1212 1 120012 2 0 3 0 5 0 25 141414 1 40201 989898 ",
		"204 1 4 1 121212 2 0 3 0 5 0 0",
		"205 1 3 2 120012 3 1212 5 0 0",
		"206 1 4 1 121212 2 120012 3 1212 5 0 0",
		"207 0 4 2 0 3 0 5 0 13 0 0",
		"208 0 4 2 0 3 0 5 0 13 121212 0",
		"209 0 6 1 1212 1 120012 2 0 3 0 5 0 13 0 0 0",
		"210 0 6 1 1212 1 120012 2 1212 3 121200 5 0 13 341434 0",
		"211 0 5 1 121212 2 0 3 0 5 0 13 0 0",
		"212 0 4 2 120012 3 1212 5 0 13 143434 0",
		"213 0 4 2 120012 3 1212 5 0 13 341414 0",
		"214 0 5 1 121212 2 120012 3 1212 5 0 13 341414 0",
		"215 0 4 2 0 3 0 5 0 40 0 0",
		"216 0 6 1 1212 1 120012 2 0 3 0 5 0 40 0 0",
		"217 0 5 1 121212 2 0 3 0 5 0 40 0 0",
		"218 0 4 2 0 3 0 5 0 40 121212 0",
		"219 0 6 1 1212 1 120012 2 0 3 0 5 0 40 121212 0",
		"220 0 5 1 121212 2 120012 3 1212 5 0 40 141414 0",
		"221 1 4 2 0 3 0 5 0 13 0 0",
		"222 0 5 2 0 3 0 5 0 13 0 25 121212 1 40201 949494 ",
		"223 1 4 2 0 3 0 5 0 13 121212 0",
		"224 0 5 2 0 3 0 5 0 13 121212 25 121212 1 40201 949494",
		"225 1 6 1 1212 1 120012 2 0 3 0 5 0 13 0 0",
		"226 1 6 1 1212 1 120012 2 0 3 0 5 0 13 121212 0",
		"227 0 7 1 1212 1 120012 2 1212 3 121200 5 0 13 341434 25 141414 1 40201 989898",
		"228 0 7 1 1212 1 120012 2 1212 3 121200 5 0 13 341434 25 343434 1 40201 b8b8b8",
		"229 1 5 1 121212 2 0 3 0 5 0 13 0 0",
		"230 1 5 1 121212 2 120012 3 1212 5 0 13 341414 0"
	} ;
// ---------------------------------------------------------
// Set nearest integer zone
// ---------------------------------------------------------
void SetNearestZone(float *x,float *y,float *z) {
	float dx,dy,dz,fx,fy,fz,minNum=1000.,maxNum=-1000.,kFact,commFact ;
	register int i ;
	int	isInt ;
	
	if (fabs(*x) < 0.001) *x = 0. ;
	if (fabs(*y) < 0.001) *y = 0. ;
	if (fabs(*z) < 0.001) *z = 0. ;
	fx = fabs(*x) ;
	fy = fabs(*y) ;
	fz = fabs(*z) ;
	if (fx != 0) minNum = min(minNum,fx) ;
	if (fy != 0) minNum = min(minNum,fy) ;
	if (fz != 0) minNum = min(minNum,fz) ;
	if (fx != 0) maxNum = max(maxNum,fx) ;
	if (fy != 0) maxNum = max(maxNum,fy) ;
	if (fz != 0) maxNum = max(maxNum,fz) ;
	if ((maxNum/minNum) > 10.) {
		/* Check if we should possible divide by the next largest number	*/
		if (fx > minNum && fx < maxNum) {	/* we have a middle number	*/
			if ((fx/minNum) > 10.) {
				minNum = fx ;
			}
		}
		else if (fy > minNum && fy < maxNum) {	/* we have a middle number	*/
			if ((fy/minNum) > 10.) {
				minNum = fy ;
			}
		}
		else if (fz > minNum && fz < maxNum) {	/* we have a middle number	*/
			if ((fz/minNum) > 10.) {
				minNum = fz ;
			}
		}
		else { /* No middle number */
			minNum = maxNum ;
		}
	}
	fx /= minNum ;
	fy /= minNum ;
	fz /= minNum ;
	
	/* At this point we have reduced the smallest number to a 1.  However, the next largest	*/
	/* number could be something of the sort 10/6, so that the zone could be 10 6 0			*/
	/* See if the fractional part is close to zero.											*/
	isInt = false ;
	for (i=1;i <= 20;i++) {
		fx *= i ;
		fy *= i ;
		fz *= i ;
		dx = fabs(fx-near(fx)) ;
		dy = fabs(fy-near(fy)) ;
		dz = fabs(fz-near(fz)) ;
		if (dx < 0.1 && dy < 0.1 && dz < 0.1) {
			isInt = true ;
			kFact = i ;
			break ;
		}
	}
	if (isInt) {
		commFact = kFact/minNum ;
	}
	else {
		commFact = 1./minNum ;
	}
	*x *= commFact ;
	*y *= commFact ;
	*z *= commFact ;
	*x = near(*x) ;
	*y = near(*y) ;
	*z = near(*z) ;
}

// ---------------------------------------------------------
// Set nearest real zone
// ---------------------------------------------------------
void SetZone(TVector &zone) {
	float fx,fy,fz,minNum=1000.,maxNum=-1000.,kFact ;
	long iFact ;
	
	fx = fabs(zone.a[0]) ;
	fy = fabs(zone.a[1]) ;
	fz = fabs(zone.a[2]) ;
	
	if (fx != 0.) minNum = min(minNum,fx) ;
	if (fy != 0.) minNum = min(minNum,fy) ;
	if (fz != 0.) minNum = min(minNum,fz) ;
	
	if (fx != 0.) maxNum = max(maxNum,fx) ;
	if (fy != 0.) maxNum = max(maxNum,fy) ;
	if (fz != 0.) maxNum = max(maxNum,fz) ;
	
	kFact = maxNum ;
	
	// Divide by the smallest number different from zero
	fx /= minNum ;
	fy /= minNum ;
	fz /= minNum ;
	
	// Check for the largest value after division
	maxNum = -1000 ;
	if (fx != 0.) maxNum = max(maxNum,fx) ;
	if (fy != 0.) maxNum = max(maxNum,fy) ;
	if (fz != 0.) maxNum = max(maxNum,fz) ;
	
	iFact = (long)maxNum ;
	if (iFact >= 10) {
		fx /= iFact ;
		fy /= iFact ;
		fz /= iFact ;
	}

	zone.a[0] = SIGN(zone.a[0])*fx ;
	zone.a[1] = SIGN(zone.a[1])*fy ;
	zone.a[2] = SIGN(zone.a[2])*fz ;
}

extern char *chemsymb[] ;
long atomno(char *cp) {
	long i,iflag ;
	char symb[4],id[4],*symbp,*symbpt,*idp ;

	idp = id ;
	while((*idp++ = toupper(*cp++))!='\0') ;
	for (i=0;i<92;i++) {
		symbp = symb ;
		symbpt = chemsymb[i] ;
		while((*symbp++ = toupper(*symbpt++))!='\0')  ;
		if ((iflag = strncmp(id,symb,2))==0) break ;
	}
	if (iflag==0)
		return(i+1) ;
	else
		return(-1) ;
}

void assm_spcgrp(long spcgrp_number) {
	strcpy(spacegroup,gSpacegroupStrings[spcgrp_number-1]) ;
	generate_spcgrp(spacegroup) ;
}

void generate_spcgrp(char  *spacegroup) {
	long i,j,k,num_gen ;
	long inversion,num_sym ;
	long ione ;

	symop[0] = identity_matr ;
	gNumSymops = 1 ;
	decode_spcgrp(spacegroup,&num_gen,&inversion) ;
	for (i=0;i<num_gen;i++) {
		num_sym = gNumSymops ;
		temp_gen = gen[i] ;
		do {
			for (j=0;j< num_sym;j++) {
				apply_gen(&temp_gen,&symop[j],&temp_sym) ;
				if (new_symop(&temp_sym,gNumSymops))
					symop[gNumSymops++] = temp_sym ;
			}
			apply_gen(&gen[i],&temp_gen,&temp_sym) ;
			if (!(ione=ident(&temp_sym)))
				temp_gen = temp_sym ;
		} while (!ione) ;
	}
	num_sym = gNumSymops ;
	gCentroSymmetric = FALSE ;
	if (inversion) {
		gCentroSymmetric = TRUE ;
		for (j=0;j< num_sym;j++) {
			apply_gen(&inv_gen,&symop[j],&temp_sym) ;
			symop[gNumSymops++] = temp_sym ;
		}
	}
	for (k=0;k< gNumSymops;k++) {
		encode_symop(&symop_txt[k][0],&symop[k]) ;
		txtptr_symop[k] = &symop_txt[k][0] ;
	}
}

void encode_symop(char *string,struct GENERATOR *sym_op) {
	long first,i,j ;
	float test ;
	char *cp ;

	cp = string ;
	for (i=0;i<3;i++) {
		first = TRUE ;
		for (j=0;j<3;j++) {
			test = sym_op->matr[i][j] ;
			if (test != 0.) {
				if (first) {
					if (test < 0.) sprintf(cp++,"-") ;
						sprintf(cp++,"%c",symb[j]) ;
					first = FALSE ;
				}
				else {
					if (test < 0.) 
						sprintf(cp++,"-%c",symb[j]) ;
					else
						sprintf(cp++,"+%c",symb[j]) ;
					cp++ ;
				}
			}
		}
		if (sym_op->matr[i][3] != 0.) {
			j = 0 ;
			do {
				test = sym_op->matr[i][3] - translations[j++].value ;
				test = (test >= 0. ? test : -test) ;
			}
			while ( test > 0.001 && j < 11) ;
			sym_op->matr[i][3] = translations[--j].value ;
			sprintf(cp++,"+%s",translations[j].symbol) ;
			cp += 3 ;
		}
		if (i!=2) sprintf(cp++,",") ;
	}
	*cp = EOS ;
}

void apply_gen(struct GENERATOR *oper,struct GENERATOR *source,struct GENERATOR *dest) {
	long i,j,k ;


	for (i=0;i<4;i++)
		for (j=0;j<4;j++) {
			dest->matr[i][j] = 0. ;
			for (k=0;k<4;k++)
				dest->matr[i][j] += (oper->matr[i][k]*source->matr[k][j]) ;
		}
	for (k=0;k<3;k++) {
		dest->matr[k][3] = one(dest->matr[k][3]) ;
		if (dest->matr[k][3] < 0.001 || dest->matr[k][3] > 0.999)
			dest->matr[k][3] = 0. ;
	}
}

long ident(struct GENERATOR *p_sym) {
	long i ; 
	float *p1,*p2 ;

	p1 = &p_sym->matr[0][0] ;
	p2 = &symop[0].matr[0][0] ;
	for (i=0;i<16 && *p1++ == *p2++;i++) ;
	if (i==16)
		return(TRUE) ;
	else 
		return(FALSE) ;
}   

long new_symop(struct GENERATOR *p_symop,long num_symop) {
	long i,j,new_gen ;
	float *p_sym,*p_old_sym,test ;

	new_gen = TRUE ;
	for (i=0;i< num_symop;i++) {
		p_sym = &p_symop->matr[0][0] ;
		p_old_sym = &symop[i].matr[0][0] ;
		for(j=0; j < 16 ;j++) {
			test = *p_sym - *p_old_sym ;
			test = (test >= 0. ? test : -test) ;
			if (test > 0.001) 
				break ;
			p_sym++ ;
			p_old_sym++ ;
		}
		if (j==16) {
			new_gen = FALSE ;
			break ;
		}
	}
	return(new_gen) ;
}

#define BLANK ' '
#include <math.h>
void decode_spcgrp(char *string,long *pnum_gen,long *pinversion) {
	long num_gen,inversion ;
	long i,j,k,spcgrp_number,p_transl,p_rot ;
	long rot[20],trnsl[20] ;
	char *cp,buf[80] ;

	cp = string ;
	/* skip first blanks */
	while (*cp==BLANK) cp++ ;
	i=0 ;
	while(*cp != BLANK)
		buf[i++] = *cp++ ;
	buf[i] = EOS ;
	spcgrp_number = atoi(buf) ;
	/* skip first blanks */
	while (*cp==BLANK) cp++ ;
	i=0 ;
	while(*cp != BLANK)
		buf[i++] = *cp++ ;
	buf[i] = EOS ;
	inversion = atoi(buf) ;
	/* skip first blanks */
	while (*cp==BLANK) cp++ ;
	i=0 ;
	while(*cp != BLANK)
		buf[i++] = *cp++ ;
	buf[i] = EOS ;
	num_gen = atoi(buf) ;

	for (k=0;k<num_gen;k++) 
		for (j=0;j<2;j++) {
			/* skip first blanks */
			while (*cp==BLANK) cp++ ;
			i=0 ;
			while(*cp != BLANK)
				buf[i++] = *cp++ ;
			buf[i] = EOS ;
			if (j==0)
				rot[k] = atoi(buf) ;
			else
				trnsl[k] = atox(buf) ;
		}               

	/* Check to see if there is a second setting */
	/* skip first blanks */
	while (*cp==BLANK) cp++ ;
	i=0 ;
	while(*cp != BLANK && *cp != EOS)
		buf[i++] = *cp++ ;
	buf[i] = EOS ;
	gAlternateSetting = atoi(buf) ;

	/* If there is a second setting, then get the P and the Q matrix */
	if (gAlternateSetting) {
		for (j=0;j<2;j++) {
			/* skip first blanks */
			while (*cp==BLANK) cp++ ;
			i=0 ;
			while(*cp != BLANK && *cp != EOS && i < 132)
				buf[i++] = *cp++ ;
			buf[i] = EOS ;
			if (j)
				p_transl = atox(buf) ;
			else
				p_rot = atoo(buf) ;
		}
		/* Now we have to set the values in the p and q matrices */
		decode_gen(p_rot,p_transl,&p_matr) ;
		get_inverse(&p_matr,&q_matr) ;
	}
	for (i=0;i<num_gen;i++)
		decode_gen(generators[rot[i]-1],trnsl[i],&gen[i]) ;
	*pnum_gen = num_gen ;
	*pinversion = inversion ;
}

void get_inverse(struct GENERATOR *source,struct GENERATOR *dest) {
	long i ;
	float det ;

	det = source->matr[0][0]*(source->matr[1][1]*source->matr[2][2] -
	                  source->matr[1][2]*source->matr[2][1]) +
	source->matr[0][1]*(source->matr[2][0]*source->matr[1][2] -
	                  source->matr[1][0]*source->matr[2][2]) +
	source->matr[0][2]*(source->matr[1][0]*source->matr[2][1] -
	                  source->matr[2][0]*source->matr[1][1]) ;
	dest->matr[0][0] = (source->matr[1][1]*source->matr[2][2] -
	            source->matr[2][1]*source->matr[1][2])/det ;
	dest->matr[0][1] = (source->matr[2][1]*source->matr[0][2] -
	            source->matr[0][1]*source->matr[2][2])/det ;
	dest->matr[0][2] = (source->matr[0][1]*source->matr[1][2] -
	            source->matr[1][1]*source->matr[0][2])/det ;
	dest->matr[1][0] = (source->matr[2][0]*source->matr[1][2] -
	            source->matr[1][0]*source->matr[2][2])/det ;
	dest->matr[1][1] = (source->matr[0][0]*source->matr[2][2] -
	            source->matr[2][0]*source->matr[0][2])/det ;
	dest->matr[1][2] = (source->matr[1][0]*source->matr[0][2] -
	            source->matr[0][0]*source->matr[1][2])/det ;
	dest->matr[2][0] = (source->matr[1][0]*source->matr[2][1] -
	            source->matr[2][0]*source->matr[1][1])/det ;
	dest->matr[2][1] = (source->matr[2][0]*source->matr[0][1] -
	            source->matr[0][0]*source->matr[2][1])/det ;
	dest->matr[2][2] = (source->matr[0][0]*source->matr[1][1] -
	            source->matr[1][0]*source->matr[0][1])/det ;
	for (i=0;i<3;i++) {
		dest->matr[i][3] = - source->matr[i][3] ;
		dest->matr[3][i] = source->matr[3][i] ;
	}
	dest->matr[3][3] = source->matr[3][3] ;
}

void decode_gen(long rot,long trnsl,struct GENERATOR *sym_op) {
	long i,j,k,ndx,den[3],num[3] ;
	float t[16],*p_float ;

	for (i=0;i<16;t[i++]=0.) ;
	/* decode the rotation part */
	for (k=0;k<3;k++)
		for (j=0;j<2;j++)
			for (i=0,ndx=10-k*4;i<3;i++,ndx--) {
				if (rot%2) {
					if (j)
						t[ndx] = - t[ndx] ;
					else
						t[ndx] = 1. ;
				}
				rot >>= 1 ;
			}
	/* decode the translation part */
	den[2] = trnsl & 0x0000000F ;
	num[2] = (trnsl & 0x000000f0) >> 4 ;
	if (num[2] > 8) num[2] = 8 - num[2] ;
	den[1] = (trnsl & 0x00000f00) >> 8 ;
	num[1] = (trnsl & 0x0000f000) >> 12 ;
	if (num[1] > 8) num[1] = 8 - num[1] ;
	den[0] = (trnsl & 0x000f0000) >> 16 ;
	num[0] = (trnsl & 0x00f00000) >> 20 ;
	if (num[0] > 8) num[0] = 8 - num[0] ;
	for (i=0;i<3;i++)
		if (den[i] != 0)
			t[3+i*4] = (float)num[i]/den[i] ;
	t[15] = 1. ;
	p_float = &sym_op->matr[0][0] ;
	for (i=0;i<16;*p_float++=t[i++]) ;
	if (sym_op != &p_matr) {
		if (gSecondSetting && gAlternateSetting) {
			apply_gen(sym_op,&q_matr,&temp_sym) ;
			apply_gen(&p_matr,&temp_sym,sym_op) ;
		}
	}
}

void decode_symop(long j) ;
// -------------------------------------------------------------
// Routine to decode the symmetry operator stored as an 
// ascii string.                                   
// -------------------------------------------------------------
void decode_symop(long j) {
	char buffer[3][20],*cp,*cp1,c ;
	long  i,k,n,number_1,number_2 ;

	k = 0 ;
	i = 0 ;
	n = 0 ;
	cp = txtptr_symop[j] ;
	
	do {
		c = *cp++ ;
		if (c != EOS) {
			if (c==COMMA) {
				buffer[k++][i] = EOS ;
				i = 0 ;
			}
			else if (c != BLANK)
				buffer[k][i++] = c ;
		}
	}
	while (c != EOS) ;
	
	buffer[k][i] = EOS ;
	for (k=0;k<3;k++)
		{
		if ((cp=strpbrk(&buffer[k][0],"xX")) != NULL) {
			if (cp == &buffer[k][0])
				symop[j].matr[k][0] = 1. ;
			else {
				if (*(--cp) == MINUS)
					symop[j].matr[k][0] = -1. ;
				else
					symop[j].matr[k][0] = 1. ;
			}
		}
		else
			symop[j].matr[k][0] = 0. ;
			
		if ((cp=strpbrk(&buffer[k][0],"yY")) != NULL) {
			if (cp == &buffer[k][0])
				symop[j].matr[k][1] = 1. ;
			else {
				if (*(--cp) == MINUS)
					symop[j].matr[k][1] = -1. ;
				else
					symop[j].matr[k][1] = 1. ;
			}
		}
		else
			symop[j].matr[k][1] = 0. ;
			
		if ((cp=strpbrk(&buffer[k][0],"zZ")) != NULL) {
			if (cp == &buffer[k][0])
				symop[j].matr[k][2] = 1. ;
			else {
				if (*(--cp) == MINUS)
					symop[j].matr[k][2] = -1. ;
				else
					symop[j].matr[k][2] = 1. ;
			}
		}
		else
			symop[j].matr[k][2] = 0. ;
	}
	/* sort out the translation vectors */
	for (k=0;k<3;k++) {
		number_1 = 0 ;
		number_2 = 0 ;
		if ((cp=strpbrk(&buffer[k][0],"1234567890/")) != NULL) {
			/* there is a translation vector */
			/* Check to see if it is written as a fraction */
			if ((cp1=strpbrk(&buffer[k][0],"/")) != NULL) {  /* It is */
				number_1 = *(--cp1) - '0' ;
				/* check for the sign */
				if (cp1 != &buffer[k][0])
					if ((c = *(cp1-1)) == MINUS)
						number_1 = -number_1 ;
				cp1 += 2 ;
				while (*cp1 >= '0' && *cp1 <= '9') {
					number_2 = number_2*10 + *cp1 - '0' ;
					cp1++ ;
				}
				symop[j].matr[k][3] = (float)number_1/number_2 ;
			}
			else {  /* It is written with just the denominator */
				number_1 = 1 ;
				/* First get the sign */
				if ( cp != &buffer[k][0])
					if (( c = *(cp-1)) == MINUS)
						number_1 = -number_1 ;
				while (*cp >= '0' && *cp <= '9') {
					number_2 = number_2*10 + *cp - '0' ;
					cp++ ;
				}
				symop[j].matr[k][3] = (float)number_1/number_2 ;
			}
		}
		else
			symop[j].matr[k][3] = 0. ;
	}
	for (k=0;k<2;k++)
		symop[j].matr[3][k] = 0. ;
	symop[j].matr[3][3] = 1. ;
}

HRTEMImageSimulationRecord::HRTEMImageSimulationRecord() {
	mGMax = 2.0 ;
	mZoneAxis.a[0] = 0 ; mZoneAxis.a[1] = 0 ; mZoneAxis.a[2] = 1 ; 
	m3DSlicing = false ;
	mNumSubSlices = 1 ;
	mStartThickness = 100. ;
	mEndThickness = 100. ;
	mDeltaThickness = 0. ;
	mStoreReflections = false ;
	mNumReflectionsToStore = 0 ;
	mReflectionsToStore = nil ;
	
	mIsModified = false ;
	
}

void	HRTEMImageSimulationRecord::ShowParameters() {
	THRTEMSimulationParameterDialog *theDialog = new THRTEMSimulationParameterDialog(this) ;
	theDialog->DoModalDialog() ;
	delete theDialog ;
}

Boolean	HRTEMImageSimulationRecord::SetCalculationSize(TCrystal *crystal) {
	// Find the dimensions of the width of the calculation from the crystal and the zone-axis
	// We will do this by reading in the files that ncemss created for this structure
	
	char	baseName[256],fileName[256] ;
	char	latticeLine[256],recipLatticeLine[256],gammaLine[256],widthLine[256],heightLine[256] ;
	float   z ;
	Boolean hasA=false,hasAStar=false,hasGamma=false,hasWidth=false,hasHeight=false ;
	
	strcpy(baseName,crystal->mFSInfo.mFileName) ;

	mCrystal = crystal ;

	char *cp = strstr(baseName,".at") ;
	if (cp == nil) return false ;
	
	*cp = '\0' ;
	
	// Open the phase-grating print file
	sprintf(fileName,"%s.p_prnt",baseName) ;
	FILE *fp ;
	fp = fopen(fileName,"r") ;
	if (fp == nil) return false ;
	
	// Read until we get to the line containing the calculation dimensions
	char buffer[256] ;
	Boolean done = false ;
	while (!done) {
		if (ReadLine(buffer,256,fp) == nil)
			break ;
		if (!hasA) {
			cp = strstr(buffer,"A =") ;
			if (cp != nil) {
				strcpy(latticeLine,cp) ;
				hasA = true ;
			}
		}
		if (!hasGamma) {
			cp = strstr(buffer,"GAMMA =") ;
			if (cp != nil) {
				strcpy(gammaLine,cp) ;
				hasGamma = true ;
			}
		}
		if (!hasAStar) {
			cp = strstr(buffer,"AST =") ;
			if (cp != nil) {
				strcpy(recipLatticeLine,cp) ;
				hasAStar = true ;
			}
		}
		if (!hasWidth) {
			cp = strstr(buffer,"The x-axis") ;
			if (cp != nil) {
				strcpy(widthLine,cp) ;
				hasWidth = true ;
			}
		}
		if (!hasHeight) {
			cp = strstr(buffer,"The y-axis") ;
			if (cp != nil) {
				strcpy(heightLine,cp) ;
				hasHeight = true ;
			}
		}
		if (hasA && hasGamma && hasAStar && hasWidth && hasHeight) done = true ;
	}
	fclose(fp) ;
	if (cp == nil) return false ;

	// mTEMRecord.mSimulatedCrystal = new TCrystal ; 

	sscanf(latticeLine,"A = %f ,   B = %f ,   C = %f",&mA,&mB,&z) ;
	sscanf(gammaLine,"GAMMA = %f Degrees",&mGamma) ;
	sscanf(recipLatticeLine,"AST = %f, BST = %f",&mAStar,&mBStar) ;
	sscanf(widthLine,"The x-axis is divided into %d",&mCalculationWidth) ;
	sscanf(heightLine,"The y-axis is divided into %d",&mCalculationHeight) ;

	return true ;

}

HRTEMImageSimulationRecord::~HRTEMImageSimulationRecord() {
	if (mReflectionsToStore != nil) delete[] mReflectionsToStore ;
}

THRTEMSimulationParameterDialog::THRTEMSimulationParameterDialog(HRTEMImageSimulationRecord *simRec) 
: TDialog(200,200,396,320,"Simulation Parameters") {
	mHRTEMImageSimulationRecord = simRec ;
	
	Rect	theRect,theRect1,theRect2,theRect3 ;
	char	text[256] ;
	
	long	width = mPortRect.right - mPortRect.left ;
	long	height = mPortRect.bottom - mPortRect.top ;
	
	TGroupBox   *theGroupBox = new TGroupBox(16,16,width-16,16+114,"Specimen") ;
	AddObject(theGroupBox) ;
	theRect1 = theRect = theGroupBox->GetRect() ;
	theGroupBox = new TGroupBox(theRect.left+10,theRect.top+16,theRect.left+10+150,theRect.top+16+54,"Zone Axis") ;
	AddObject(theGroupBox) ;
	theRect2 = theRect = theGroupBox->GetRect() ;
	
	// Define and set the controls 
	sprintf(text,"%d",near(simRec->mZoneAxis.a[0])) ;
	mZoneU = new TEditText(theRect.left+10,theRect.bottom-10-20,theRect.left+10+39,theRect.bottom-10,"U",text);
	theRect = mZoneU->GetRect() ;
	OffsetRect(&theRect,theRect.right-theRect.left+8,0) ;
	sprintf(text,"%d",near(simRec->mZoneAxis.a[1])) ;
	mZoneV = new TEditText(&theRect,"V",text);
	theRect = mZoneV->GetRect() ;
	OffsetRect(&theRect,theRect.right-theRect.left+8,0) ;
	sprintf(text,"%d",near(simRec->mZoneAxis.a[2])) ;
	mZoneW = new TEditText(&theRect,"W",text);
	
	theRect = theRect2 ;
	OffsetRect(&theRect,theRect.right-theRect.left+8,0) ;
	theGroupBox = new TGroupBox(theRect.left,theRect.top,theRect1.right-10,theRect.bottom,"Thickness [A]") ;
	AddObject(theGroupBox) ;
	theRect3 = theRect = theGroupBox->GetRect() ;

	sprintf(text,"%.2f",simRec->mStartThickness) ;
	mThicknessStart = new TEditText(theRect.left+10,theRect.bottom-10-20,theRect.left+10+50,theRect.bottom-10,"Start",text);
	theRect = mThicknessStart->GetRect() ;
	OffsetRect(&theRect,theRect.right-theRect.left+8,0) ;
	sprintf(text,"%.2f",simRec->mDeltaThickness) ;
	mThicknessInc = new TEditText(&theRect,"Increment",text);
	theRect = mThicknessInc->GetRect() ;
	OffsetRect(&theRect,theRect.right-theRect.left+8,0) ;
	sprintf(text,"%.2f",simRec->mEndThickness) ;
	mThicknessEnd = new TEditText(&theRect,"End",text);
	
	theRect = theRect2 ;
	OffsetRect(&theRect,0,theRect2.bottom-theRect2.top+14) ;
	theRect.left = theRect.right-40 ;
	theRect.bottom = theRect.top+20 ;
	sprintf(text,"%.2f",simRec->mGMax) ;
	mGMax = new TEditText(&theRect,"Gmax [1/A]",text) ;

	
	mZoneU->SetLabelPosition(kLabelOutsideTop) ;
	mZoneV->SetLabelPosition(kLabelOutsideTop) ;
	mZoneW->SetLabelPosition(kLabelOutsideTop) ;
	mThicknessStart->SetLabelPosition(kLabelOutsideTop) ;
	mThicknessInc->SetLabelPosition(kLabelOutsideTop) ;
	mThicknessEnd->SetLabelPosition(kLabelOutsideTop) ;
	
	theRect = theRect1 ;
	OffsetRect(&theRect,0,theRect1.bottom-theRect1.top+16) ;
	theRect.bottom += 2 ;
	theGroupBox = new TGroupBox(&theRect,"Microscope parameters") ;
	AddObject(theGroupBox) ;
	theRect = theRect1 = theGroupBox->GetRect() ;
	
	theRect2 = mZoneW->GetRect() ;
	theRect.top = theRect1.top+14 ;
	theRect.right = theRect2.right ;
	theRect.bottom = theRect.top + 20 ;
	theRect.left = theRect.right - 40 ;
	
	sprintf(text,"%.3f",simRec->mMicroscope.mCs) ;
	mCs = new TEditText(&theRect,"Cs [mm] ",text);
	OffsetRect(&theRect,0,24) ;
	sprintf(text,"%.3f",simRec->mMicroscope.mConvergence) ;
	mDiv = new TEditText(&theRect,"Divergence [mrad] ",text);
	OffsetRect(&theRect,0,24) ;
	sprintf(text,"%.0f",simRec->mMicroscope.mSpreadOfFocus) ;
	mDelta = new TEditText(&theRect,"Spread of Focus [A] ",text);
	OffsetRect(&theRect,0,24) ;
	sprintf(text,"%.0f",simRec->mMicroscope.mVoltage) ;
	mVoltage = new TEditText(&theRect,"Voltage [kV] ",text);

	theRect = theRect3 ;
	theRect.top = theRect1.top + 16 ;
	theRect.bottom = theRect.top + 60 ;

	theGroupBox = new TGroupBox(&theRect,"Objective Lens Defocus [A]") ;
	AddObject(theGroupBox) ;
	theRect3 = theRect = theGroupBox->GetRect() ;
	
	sprintf(text,"%.0f",simRec->mMicroscope.mStartFocus) ;
	mFocusStart = new TEditText(theRect.left+10,theRect.bottom-10-20,theRect.left+10+50,theRect.bottom-10,"Start",text);
	theRect = mFocusStart->GetRect() ;
	OffsetRect(&theRect,theRect.right-theRect.left+8,0) ;
	sprintf(text,"%.0f",simRec->mMicroscope.mDeltaFocus) ;
	mFocusInc = new TEditText(&theRect,"Increment",text);
	theRect = mFocusInc->GetRect() ;
	OffsetRect(&theRect,theRect.right-theRect.left+8,0) ;
	sprintf(text,"%.0f",simRec->mMicroscope.mEndFocus) ;
	mFocusEnd = new TEditText(&theRect,"End",text);

	mFocusStart->SetLabelPosition(kLabelOutsideTop) ;
	mFocusInc->SetLabelPosition(kLabelOutsideTop) ;
	mFocusEnd->SetLabelPosition(kLabelOutsideTop) ;
	
	AddTextEditObject(mFocusEnd) ;
	AddTextEditObject(mFocusInc) ;
	AddTextEditObject(mFocusStart) ;
	
	AddTextEditObject(mVoltage) ;
	AddTextEditObject(mDelta) ;
	AddTextEditObject(mDiv) ;
	AddTextEditObject(mCs) ;	
	
	AddTextEditObject(mGMax) ;
	AddTextEditObject(mThicknessEnd) ;
	AddTextEditObject(mThicknessInc) ;
	AddTextEditObject(mThicknessStart) ;
	AddTextEditObject(mZoneW) ;
	AddTextEditObject(mZoneV) ;
	AddTextEditObject(mZoneU) ;
	
	mOKButton = new TButton(width-16-70,height-16-20,width-16,height-16,"OK") ;
	AddObject(mOKButton) ;
	theRect = mOKButton->GetRect() ;
	OffsetRect(&theRect,-(theRect.right-theRect.left+10),0) ;
	mCancelButton = new TButton(&theRect,"Cancel") ;
	AddObject(mCancelButton) ;
	SetDefaultButton(mOKButton) ;
	
}

void	THRTEMSimulationParameterDialog::RetrieveDialogItems() {
	// Read back the prameters in the dialog box and update the simulation record with the
	// new parameters
	char	text[256] ;
	float   val,tolerance=0.000001 ;
	
	mZoneU->GetEditText(text) ;
	val = mHRTEMImageSimulationRecord->mZoneAxis.a[0] ;
	mHRTEMImageSimulationRecord->mZoneAxis.a[0] = atof(text) ;
	if (fabs(val-mHRTEMImageSimulationRecord->mZoneAxis.a[0]) > tolerance) {
		mHRTEMImageSimulationRecord->mIsModified = true ;
		mHRTEMImageSimulationRecord->mCalculationStatus &= ~PG_MASK ;	// Phasegrating needs to be run
	}

	mZoneV->GetEditText(text) ;
	val = mHRTEMImageSimulationRecord->mZoneAxis.a[1] ;
	mHRTEMImageSimulationRecord->mZoneAxis.a[1] = atof(text) ;
	if (fabs(val-mHRTEMImageSimulationRecord->mZoneAxis.a[1]) > tolerance) {
		mHRTEMImageSimulationRecord->mIsModified = true ;
		mHRTEMImageSimulationRecord->mCalculationStatus &= ~PG_MASK ;	// Phasegrating needs to be run
	}
	
	mZoneW->GetEditText(text) ;
	val = mHRTEMImageSimulationRecord->mZoneAxis.a[2] ;
	mHRTEMImageSimulationRecord->mZoneAxis.a[2] = atof(text) ;
	if (fabs(val-mHRTEMImageSimulationRecord->mZoneAxis.a[2]) > tolerance) {
		mHRTEMImageSimulationRecord->mIsModified = true ;
		mHRTEMImageSimulationRecord->mCalculationStatus &= ~PG_MASK ;	// Phasegrating needs to be run
	}	

	mThicknessStart->GetEditText(text) ;
	val = mHRTEMImageSimulationRecord->mStartThickness ;
	mHRTEMImageSimulationRecord->mStartThickness = atof(text) ;
	if (fabs(val-mHRTEMImageSimulationRecord->mStartThickness) > tolerance) {
		mHRTEMImageSimulationRecord->mIsModified = true ;
		mHRTEMImageSimulationRecord->mCalculationStatus &= ~MS_MASK ;	// Multislice needs to be run
	}
	
	mThicknessEnd->GetEditText(text) ;
	val = mHRTEMImageSimulationRecord->mEndThickness ;
	mHRTEMImageSimulationRecord->mEndThickness = atof(text) ;
	if (fabs(val-mHRTEMImageSimulationRecord->mEndThickness) > tolerance) {
		mHRTEMImageSimulationRecord->mIsModified = true ;
		mHRTEMImageSimulationRecord->mCalculationStatus &= ~MS_MASK ;	// Multislice needs to be run
	}
	
	mThicknessInc->GetEditText(text) ;
	val = mHRTEMImageSimulationRecord->mDeltaThickness ;
	mHRTEMImageSimulationRecord->mDeltaThickness = atof(text) ;
	if (fabs(val-mHRTEMImageSimulationRecord->mDeltaThickness) > tolerance) {
		mHRTEMImageSimulationRecord->mIsModified = true ;
		mHRTEMImageSimulationRecord->mCalculationStatus &= ~MS_MASK ;	// Multislice needs to be run
	}
	
	mGMax->GetEditText(text) ;
	val = mHRTEMImageSimulationRecord->mGMax ;
	mHRTEMImageSimulationRecord->mGMax = atof(text) ;
	if (fabs(val-mHRTEMImageSimulationRecord->mGMax) > tolerance) {
		mHRTEMImageSimulationRecord->mIsModified = true ;
		mHRTEMImageSimulationRecord->mCalculationStatus &= ~PG_MASK ;	// Phasegrating needs to be run
	}
	
	mCs->GetEditText(text) ;
	val = mHRTEMImageSimulationRecord->mMicroscope.mCs ;
	mHRTEMImageSimulationRecord->mMicroscope.mCs = atof(text) ;
	if (fabs(val-mHRTEMImageSimulationRecord->mMicroscope.mCs) > tolerance) {
		mHRTEMImageSimulationRecord->mIsModified = true ;
		mHRTEMImageSimulationRecord->mCalculationStatus &= ~IM_MASK ;	// Image needs to be run
	}
	
	mDiv->GetEditText(text) ;
	val = mHRTEMImageSimulationRecord->mMicroscope.mConvergence ;
	mHRTEMImageSimulationRecord->mMicroscope.mConvergence = atof(text) ;
	if (fabs(val-mHRTEMImageSimulationRecord->mMicroscope.mConvergence) > tolerance) {
		mHRTEMImageSimulationRecord->mIsModified = true ;
		mHRTEMImageSimulationRecord->mCalculationStatus &= ~IM_MASK ;	// Image needs to be run
	}
	
	mDelta->GetEditText(text) ;
	val = mHRTEMImageSimulationRecord->mMicroscope.mSpreadOfFocus ;
	mHRTEMImageSimulationRecord->mMicroscope.mSpreadOfFocus = atof(text) ;
	if (fabs(val-mHRTEMImageSimulationRecord->mMicroscope.mSpreadOfFocus) > tolerance) {
		mHRTEMImageSimulationRecord->mIsModified = true ;
		mHRTEMImageSimulationRecord->mCalculationStatus &= ~IM_MASK ;	// Image needs to be run
	}
	
	mVoltage->GetEditText(text) ;
	val = mHRTEMImageSimulationRecord->mMicroscope.mVoltage ;
	mHRTEMImageSimulationRecord->mMicroscope.mVoltage = atof(text) ;
	if (fabs(val-mHRTEMImageSimulationRecord->mMicroscope.mVoltage) > tolerance) {
		mHRTEMImageSimulationRecord->mIsModified = true ;
		mHRTEMImageSimulationRecord->mCalculationStatus &= ~IM_MASK ;	// Image needs to be run
	}
	
	mFocusStart->GetEditText(text) ;
	val = mHRTEMImageSimulationRecord->mMicroscope.mStartFocus ;
	mHRTEMImageSimulationRecord->mMicroscope.mStartFocus = atof(text) ;
	if (fabs(val-mHRTEMImageSimulationRecord->mMicroscope.mStartFocus) > tolerance) {
		mHRTEMImageSimulationRecord->mIsModified = true ;
		mHRTEMImageSimulationRecord->mCalculationStatus &= ~IM_MASK ;	// Image needs to be run
	}
	
	mFocusInc->GetEditText(text) ;
	val = mHRTEMImageSimulationRecord->mMicroscope.mDeltaFocus ;
	mHRTEMImageSimulationRecord->mMicroscope.mDeltaFocus = atof(text) ;
	if (fabs(val-mHRTEMImageSimulationRecord->mMicroscope.mDeltaFocus) > tolerance) {
		mHRTEMImageSimulationRecord->mIsModified = true ;
		mHRTEMImageSimulationRecord->mCalculationStatus &= ~IM_MASK ;	// Image needs to be run
	}
	
	mFocusEnd->GetEditText(text) ;
	val = mHRTEMImageSimulationRecord->mMicroscope.mEndFocus ;
	mHRTEMImageSimulationRecord->mMicroscope.mEndFocus = atof(text) ;
	if (fabs(val-mHRTEMImageSimulationRecord->mMicroscope.mEndFocus) > tolerance) {
		mHRTEMImageSimulationRecord->mIsModified = true ;
		mHRTEMImageSimulationRecord->mCalculationStatus &= ~IM_MASK ;	// Image needs to be run
	}

    if (mHRTEMImageSimulationRecord->mDeltaThickness != 0.) {
		mHRTEMImageSimulationRecord->mNumThicknesses = 
			(mHRTEMImageSimulationRecord->mEndThickness-mHRTEMImageSimulationRecord->mStartThickness)/mHRTEMImageSimulationRecord->mDeltaThickness + 1 ;
	}
    else {
		mHRTEMImageSimulationRecord->mNumThicknesses = 1 ;
	}
	
    if (mHRTEMImageSimulationRecord->mMicroscope.mDeltaFocus != 0.) {
		mHRTEMImageSimulationRecord->mNumFocii = 
			(mHRTEMImageSimulationRecord->mMicroscope.mEndFocus-mHRTEMImageSimulationRecord->mMicroscope.mStartFocus)/mHRTEMImageSimulationRecord->mMicroscope.mDeltaFocus + 1 ;
	}
    else {
		mHRTEMImageSimulationRecord->mNumFocii = 1 ;
	}
	
}

MBFITSimulationRecord::MBFITSimulationRecord() {
	mZoneAxis.a[0] = 1 ; mZoneAxis.a[1] = 1 ; mZoneAxis.a[2] = 1 ; 
	mSurfaceNormal = mZoneAxis ;
	
	mZoneBaseVector.a[0] = 1 ; mZoneBaseVector.a[1] = -2 ; mZoneBaseVector.a[2] = 0 ; 
	mTiltH = 0. ;
	mTiltG = 0. ;
	
	mIsModified = false ;
	
	mMicroscope.mVoltage = 100. ;
	mCalcType = kMBFIT_CALC_KXY_RANGE_Disk ;
	
	mNumCalculationPoints = 101 ;
	mMinH   = -1.0 ;
	mMaxH   = 1.0 ;
	mMinG   = -1.0 ;
	mMaxG   = 1.0 ;
	
	mNumCalculationPointsX = 501 ;
	mNumCalculationPointsY = 501 ;
	mMinX   = -2.0 ;
	mMaxX   = 2.0 ;
	mMinY   = -2.0 ;
	mMaxY   = 2.0 ;
	
	
	mBeamSearchAuto.num_g_limit = 200 ;
	mBeamSearchAuto.g_max = 4.0 ;
	mBeamSearchAuto.lz_min = 0 ;
	mBeamSearchAuto.lz_max = 0 ;
	mBeamSearchAuto.sg_exact_max = 0.02 ;
	mBeamSearchAuto.sg_bethe_max = 0.03 ;
	mBeamSearchAuto.xi_g_max = 1e+11 ;

	mNumThicknesses = 0 ;
	mThicknesses = nil ;
	mNumLowOrderG = 0 ;
	mLowOrderG = nil ;
	mNumExact = 0 ;
	mLowOrderG = nil ;
	mExact = 0 ;
	mBethe = nil ;
	
	branchBegin = 1 ;
	branchEnd	= 7 ;
	
	mNumReflectionsToOutput = 0 ; // Means all
	mOutputReflection = nil ;
}

MBFITSimulationRecord::~MBFITSimulationRecord() {
	if (mThicknesses != nil) delete[] mThicknesses ;
	if (mOutputReflection != nil) delete[] mOutputReflection ;
	if (mLowOrderG != nil) delete[] mLowOrderG ;
	if (mExact != nil) delete[] mExact ;
	if (mBethe != nil) delete[] mBethe ;
}

void	MBFITSimulationRecord::ShowParameters() {
	TMBFITSimulationParameterDialog *theDialog = new TMBFITSimulationParameterDialog(this) ;
	theDialog->DoModalDialog() ;
	delete theDialog ;
}

TMBFITSimulationParameterDialog::TMBFITSimulationParameterDialog(MBFITSimulationRecord *simRec) 
: TDialog(100,100,676,480,"MBFIT Parameters") {
	mMBFITSimulationRecord = simRec ;
	
	long	i ;
	Rect	theRect,theRect1,theRect2,theRect3,theOrientationRect ;
	TStaticText	*theText ;
	char	text[256] ;
	
	long	width = mPortRect.right - mPortRect.left ;
	long	height = mPortRect.bottom - mPortRect.top ;
	
	TGroupBox   *theGroupBox = new TGroupBox(16,16,16+234,16+110,"Orientation") ;
	AddObject(theGroupBox) ;
	theOrientationRect = theRect1 = theRect = theGroupBox->GetRect() ;
	
	// Define and set the controls 
	theText = new TStaticText(theRect.left+10,theRect.top+10,theRect.left+124,theRect.top+10+20,"Zone Axis [uvw]") ;
	AddObject(theText) ;
	theRect2 = theText->GetRect() ;
	sprintf(text,"%d",near(simRec->mZoneAxis.a[0])) ;
	mZoneU = new TEditText(theRect2.right+10,theRect2.top,theRect2.right+10+25,theRect2.bottom,"",text);
	theRect = mZoneU->GetRect() ;
	OffsetRect(&theRect,theRect.right-theRect.left+8,0) ;
	sprintf(text,"%d",near(simRec->mZoneAxis.a[1])) ;
	mZoneV = new TEditText(&theRect,"",text);
	theRect = mZoneV->GetRect() ;
	OffsetRect(&theRect,theRect.right-theRect.left+8,0) ;
	sprintf(text,"%d",near(simRec->mZoneAxis.a[2])) ;
	mZoneW = new TEditText(&theRect,"",text);
	
	theRect = theGroupBox->GetRect() ;
	OffsetRect(&theRect,0,24) ;
	theText = new TStaticText(theRect.left+10,theRect.top+10,theRect.left+124,theRect.top+10+20,"Surface Normal [uvw]") ;
	AddObject(theText) ;
	theRect2 = theText->GetRect() ;
	sprintf(text,"%d",near(simRec->mSurfaceNormal.a[0])) ;
	mSurfaceNormalU = new TEditText(theRect2.right+10,theRect2.top,theRect2.right+10+25,theRect2.bottom,"",text);
	theRect = mSurfaceNormalU->GetRect() ;
	OffsetRect(&theRect,theRect.right-theRect.left+8,0) ;
	sprintf(text,"%d",near(simRec->mSurfaceNormal.a[1])) ;
	mSurfaceNormalV = new TEditText(&theRect,"",text);
	theRect = mSurfaceNormalV->GetRect() ;
	OffsetRect(&theRect,theRect.right-theRect.left+8,0) ;
	sprintf(text,"%d",near(simRec->mSurfaceNormal.a[2])) ;
	mSurfaceNormalW = new TEditText(&theRect,"",text);
	
	
	theRect = theGroupBox->GetRect() ;
	OffsetRect(&theRect,0,48) ;
	theText = new TStaticText(theRect.left+10,theRect.top+10,theRect.left+124,theRect.top+10+20,"Surface Normal [uvw]") ;
	AddObject(theText) ;
	theRect2 = theText->GetRect() ;
	sprintf(text,"%d",near(simRec->mZoneBaseVector.a[0])) ;
	mBaseVectorU = new TEditText(theRect2.right+10,theRect2.top,theRect2.right+10+25,theRect2.bottom,"",text);
	theRect = mBaseVectorU->GetRect() ;
	OffsetRect(&theRect,theRect.right-theRect.left+8,0) ;
	sprintf(text,"%d",near(simRec->mZoneBaseVector.a[1])) ;
	mBaseVectorV = new TEditText(&theRect,"",text);
	theRect = mBaseVectorV->GetRect() ;
	OffsetRect(&theRect,theRect.right-theRect.left+8,0) ;
	sprintf(text,"%d",near(simRec->mZoneBaseVector.a[2])) ;
	mBaseVectorW = new TEditText(&theRect,"",text);

	theRect = theGroupBox->GetRect() ;
	OffsetRect(&theRect,0,72) ;
	theText = new TStaticText(theRect.left+10,theRect.top+10,theRect.left+124,theRect.top+10+20,"Tilt") ;
	AddObject(theText) ;
	theRect2 = theText->GetRect() ;
	sprintf(text,"%d",near(simRec->mTiltH)) ;
	mTiltH = new TEditText(theRect2.right+10,theRect2.top,theRect2.right+10+41,theRect2.bottom,"",text);
	theRect = mTiltH->GetRect() ;
	OffsetRect(&theRect,theRect.right-theRect.left+8,0) ;
	sprintf(text,"%d",near(simRec->mTiltG)) ;
	mTiltK = new TEditText(&theRect,"",text);

	theRect = theGroupBox->GetRect() ;
	OffsetRect(&theRect,0,theRect.bottom-theRect.top+16) ;
	theRect.bottom += 24 ;
	theGroupBox = new TGroupBox(&theRect,"Calculation Renge") ;
	AddObject(theGroupBox) ;
	theRect2 = theRect = theGroupBox->GetRect() ;

	
	OffsetRect(&theRect,16,16) ;
	theRect.right = (theRect2.right+theRect2.left)/2 - 10 ;
	theRect.bottom = theRect.top + 20 ;
	m2D_Disk = new TRadioButton(&theRect,"2D Disk",false) ;
	AddObject(m2D_Disk) ;
	OffsetRect(&theRect,theRect.right-theRect.left,0) ;
	
	m2D_Square = new TRadioButton(&theRect,"2D Square",false) ;
	AddObject(m2D_Square) ;
	
	theRect = m2D_Disk->GetRect() ;
	OffsetRect(&theRect,0,20) ;
	m1D = new TRadioButton(&theRect,"1D",false) ;
	AddObject(m1D) ;
	
	theRect = m2D_Square->GetRect() ;
	OffsetRect(&theRect,0,20) ;
	m0D = new TRadioButton(&theRect,"0-D",false) ;
	AddObject(m0D) ;
	
	if (simRec->mCalcType == kMBFIT_CALC_KXY_RANGE_0D) m0D->SetState(true) ;
	else if (simRec->mCalcType == kMBFIT_CALC_KXY_RANGE_1D) m1D->SetState(true) ;
	else if (simRec->mCalcType == kMBFIT_CALC_KXY_RANGE_Disk) m2D_Disk->SetState(true) ;
	else if (simRec->mCalcType == kMBFIT_CALC_KXY_RANGE_Square) m2D_Square->SetState(true) ;
	
	theRect = m0D->GetRect() ;
	OffsetRect(&theRect,0,20) ;
	sprintf(text,"%d",simRec->mNumCalculationPoints) ;
	mPixelsAcrossDisk = new TEditText(theRect2.right - 16 -40,theRect.top,theRect2.right - 16,theRect.bottom,"Numbers of pixels across the disk",text);
	
	OffsetRect(&theRect,0,24) ;
	theRect.right = theRect2.left + (theRect2.right+theRect2.left)/2 ;
	theRect.left = theRect.right - 40 ;
	sprintf(text,"%.2f",simRec->mMinH) ;
	mH_Min = new TEditText(theRect.right-40,theRect.top,theRect.right,theRect.bottom,"h - min",text);
	theRect = mH_Min->GetRect() ;
	theRect.right = theRect2.right - 16 ;
	sprintf(text,"%.2f",simRec->mMaxH) ;
	mH_Max = new TEditText(theRect.right-40,theRect.top,theRect.right,theRect.bottom,"h - max",text);
	
	theRect = mH_Min->GetRect() ;
	OffsetRect(&theRect,0,24) ;
	sprintf(text,"%.2f",simRec->mMinG) ;
	mG_Min = new TEditText(&theRect,"k - min",text);
	
	theRect = mH_Max->GetRect() ;
	OffsetRect(&theRect,0,24) ;
	sprintf(text,"%.2f",simRec->mMaxG) ;
	mG_Max = new TEditText(&theRect,"k - max",text);

	theRect2 = theRect = theGroupBox->GetRect() ;
	OffsetRect(&theRect,theRect.right-theRect.left+16,0) ;
	theRect.right = width - 10 ;
	theRect.bottom -= 48 ;
	theGroupBox = new TGroupBox(&theRect,"X-Y Calculation Range") ;
	AddObject(theGroupBox) ;
	theRect1 = theGroupBox->GetRect() ;
	
	SetRect(&theRect,(theRect1.left+theRect1.right)/2 - 60, theRect1.top+10,(theRect1.left+theRect1.right)/2-4,theRect1.top+10+20) ;
	sprintf(text,"%d",simRec->mNumCalculationPointsX) ;
	mXYRange_XPoints = new TEditText(&theRect,"X (odd)",text);

	theRect = mXYRange_XPoints->GetRect() ;
	OffsetRect(&theRect,0,24) ;
	sprintf(text,"%.2f",simRec->mMinX) ;
	mXYRange_XBegin = new TEditText(&theRect,"X Begin",text);

	theRect = mXYRange_XBegin->GetRect() ;
	OffsetRect(&theRect,0,24) ;
	sprintf(text,"%.2f",simRec->mMaxX) ;
	mXYRange_XEnd = new TEditText(&theRect,"X End",text);

	SetRect(&theRect,theRect1.right-16 - 60, theRect1.top+10,theRect1.right-16,theRect1.top+10+20) ;
	sprintf(text,"%d",simRec->mNumCalculationPointsY) ;
	mXYRange_YPoints = new TEditText(&theRect,"Y (odd)",text);

	theRect = mXYRange_YPoints->GetRect() ;
	OffsetRect(&theRect,0,24) ;
	sprintf(text,"%.2f",simRec->mMinY) ;
	mXYRange_YBegin = new TEditText(&theRect,"Y Begin",text);

	theRect = mXYRange_YBegin->GetRect() ;
	OffsetRect(&theRect,0,24) ;
	sprintf(text,"%.2f",simRec->mMaxY) ;
	mXYRange_YEnd = new TEditText(&theRect,"Y End",text);
	
	
	theRect1 = theRect = theGroupBox->GetRect() ;
	OffsetRect(&theRect,0,theRect.bottom-theRect.top+16) ;
	theRect.bottom = theRect2.bottom ;
	theGroupBox = new TGroupBox(&theRect,"Output Branch") ;
	AddObject(theGroupBox) ;
	
	theRect1 = theGroupBox->GetRect() ;
	
	theRect.right = (theRect.right + theRect.left)/2 - 20 ;
	theRect.left = theRect.right - 30 ; 
	theRect.top += 7 ;
	theRect.bottom = theRect.top + 20 ;
	sprintf(text,"%d",simRec->branchBegin) ;
	mBranchStart = new TEditText(&theRect,"Starting Branch",text) ;
	OffsetRect(&theRect,(theRect1.right-theRect1.left)/2,0) ;
	sprintf(text,"%d",simRec->branchEnd) ;
	mBranchEnd = new TEditText(&theRect,"Ending Branch",text) ;
			
	theRect = theRect2 ;
	theRect.right = theRect1.right ;
	OffsetRect(&theRect,0,theRect.bottom-theRect.top+16) ;
	theGroupBox = new TGroupBox(&theRect,"Beam Search Method") ;
	AddObject(theGroupBox) ;
	theRect3 = theRect1 = theGroupBox->GetRect() ;
	theRect1.left += 16 ;
	theRect1.top += 10 ;
	theRect1.bottom = theRect1.top + 20 ;
	theRect1.right = theRect1.left + 100 ;
	
	mAutomaticBeamSearch = new TRadioButton(&theRect1,"Automatic",simRec->mBeamSearchType == kMBFIT_AutoBeamSearch_SG) ;
	AddObject(mAutomaticBeamSearch) ;
	
	OffsetRect(&theRect1,(theRect.right-theRect.left)/2,0) ;
	mManualBeamSearch = new TRadioButton(&theRect1,"Manual",simRec->mBeamSearchType == kMBFIT_ManualBeamSearch) ;
	AddObject(mManualBeamSearch) ;
	
	theRect1 = mAutomaticBeamSearch->GetRect() ;
	OffsetRect(&theRect1,100,24) ;
	theRect1.right = theRect1.left + 50 ;
	sprintf(text,"%d",simRec->mBeamSearchAuto.num_g_limit) ;
	mAutoMaxReflections = new TEditText(&theRect1,"Max # Reflections ",text);
	
	OffsetRect(&theRect1,0,24) ;
	sprintf(text,"%.2f",simRec->mBeamSearchAuto.g_max) ;
	mAutoGMax = new TEditText(&theRect1,"gmax [1/A] ",text);

	OffsetRect(&theRect1,0,24) ;
	sprintf(text,"%.2f",simRec->mBeamSearchAuto.sg_exact_max) ;
	mAutoSG_Exact_Max = new TEditText(&theRect1,"sg_exact_max ",text);

	OffsetRect(&theRect1,0,24) ;
	sprintf(text,"%.2e",simRec->mBeamSearchAuto.xi_g_max) ;
	mAutoX1_G_Max = new TEditText(&theRect1,"xi_g_max ",text);

	theRect1 = mAutoMaxReflections->GetRect() ;
	OffsetRect(&theRect1,140,0) ;
	sprintf(text,"%d",simRec->mBeamSearchAuto.lz_min) ;
	mAutoZ_Min = new TEditText(&theRect1,"z_min ",text);

	OffsetRect(&theRect1,0,24) ;
	sprintf(text,"%d",simRec->mBeamSearchAuto.lz_max) ;
	mAutoZ_Max = new TEditText(&theRect1,"z_max ",text);

	OffsetRect(&theRect1,0,24) ;
	sprintf(text,"%.2f",simRec->mBeamSearchAuto.sg_bethe_max) ;
	mAutoSG_Bethe_Max = new TEditText(&theRect1,"sg_bethe_max ",text);

	theRect1 = mManualBeamSearch->GetRect() ;
	OffsetRect(&theRect1,10,40) ;
	theRect1.right = theRect1.left + (theRect3.right - theRect1.left - 10)/2 - 60 ;
	theRect1.bottom = theRect3.bottom - 10 ;
	mExactList = new TList(&theRect1) ;
	AddObject(mExactList) ;
	theRect = mExactList->GetRect() ;

	for(i=0; i < simRec->mNumExact; i++) {
		sprintf(text,"%3d %3d %3d",simRec->mExact[i].a[0],simRec->mExact[i].a[1],simRec->mExact[i].a[2]) ;
		mExactList->AddRow(text) ;
	}
	
	theRect1 = theRect ;
	theRect1.bottom = theRect.top ;
	theRect1.top = theRect1.bottom-20 ;
	theText = new TStaticText(&theRect1,"# Exact") ;
	AddObject(theText) ;

	OffsetRect(&theRect,theRect.right-theRect.left+5,5) ;
	theRect.bottom = theRect.top+20 ;
	theRect.right = theRect.left+50 ;
	mExact = new TEditText(&theRect,"","") ;
	if (simRec->mBeamSearchType != kMBFIT_ManualBeamSearch) mExact->Disable() ;
	OffsetRect(&theRect,0,25) ;
	mAddExact = new TButton(&theRect,"Add") ;
	mAddExact->Disable() ;
	AddObject(mAddExact) ;
	OffsetRect(&theRect,0,25) ;
	mDeleteExact = new TButton(&theRect,"Delete") ;
	mDeleteExact->Disable() ;
	AddObject(mDeleteExact) ;
	
	theRect = mExactList->GetRect() ;
	OffsetRect(&theRect,(theRect.right - theRect.left)+65,0) ;
	mBetheList = new TList(&theRect) ;
	AddObject(mBetheList) ;
	theRect = mBetheList->GetRect() ;

	for(i=0; i < simRec->mNumBethe; i++) {
		sprintf(text,"%3d %3d %3d",simRec->mBethe[i].a[0],simRec->mBethe[i].a[1],simRec->mBethe[i].a[2]) ;
		mBetheList->AddRow(text) ;
	}
	
	theRect1 = theRect ;
	theRect1.bottom = theRect.top ;
	theRect1.top = theRect1.bottom-20 ;
	theText = new TStaticText(&theRect1,"# Bethe") ;
	AddObject(theText) ;

	OffsetRect(&theRect,theRect.right-theRect.left+5,5) ;
	theRect.bottom = theRect.top+20 ;
	theRect.right = theRect.left+50 ;
	mBethe = new TEditText(&theRect,"","") ;
	if (simRec->mBeamSearchType != kMBFIT_ManualBeamSearch) mBethe->Disable() ;
	OffsetRect(&theRect,0,25) ;
	mAddBethe = new TButton(&theRect,"Add") ;
	mAddBethe->Disable() ;
	AddObject(mAddBethe) ;
	OffsetRect(&theRect,0,25) ;
	mDeleteBethe = new TButton(&theRect,"Delete") ;
	mDeleteBethe->Disable() ;
	AddObject(mDeleteBethe) ;
	
	theRect.left = theOrientationRect.right + 16 ; 
	theRect.right = theRect.left + 120 ;
	theRect.bottom = theOrientationRect.bottom ;
	theRect.top = theRect.bottom - 70 ;
	
	mThicknessList = new TList(&theRect) ;
	AddObject(mThicknessList) ;
	theRect = mThicknessList->GetRect() ;

	for(i=0; i < simRec->mNumThicknesses; i++) {
		sprintf(text,"%.0f",simRec->mThicknesses[i]) ;
		mThicknessList->AddRow(text) ;
	}
	
	theRect1 = theRect ;
	theRect1.bottom = theRect.top ;
	theRect1.top = theRect1.bottom-20 ;
	theText = new TStaticText(&theRect1,"Output Thicknesses") ;
	AddObject(theText) ;

	
	theRect = mThicknessList->GetRect() ;
	OffsetRect(&theRect,theRect.right-theRect.left+10,0) ;
	theRect.bottom = theRect.top+20 ;
	theRect.right = theRect.left+60 ;
	mThickness = new TEditText(&theRect,"","") ;
	OffsetRect(&theRect,0,25) ;
	mAddThickness = new TButton(&theRect,"Add") ;
	mAddThickness->Disable() ;
	AddObject(mAddThickness) ;
	OffsetRect(&theRect,0,25) ;
	mDeleteThickness = new TButton(&theRect,"Delete") ;
	mDeleteThickness->Disable() ;
	AddObject(mDeleteThickness) ;

	theRect = mAddThickness->GetRect() ;
	theRect.top = theOrientationRect.top ;
	theRect.bottom = theRect.top + 20 ;
	sprintf(text,"%.0f",simRec->mMicroscope.mVoltage) ;
	mVoltage = new TEditText(theRect.right-40,theRect.top,theRect.right,theRect.bottom,"Microscope Voltage [kV]  ",text);
	
	theRect = mThicknessList->GetRect() ;
	OffsetRect(&theRect,theRect.right-theRect.left+90,0) ;
	mBeamList = new TList(&theRect) ;
	AddObject(mBeamList) ;
	theRect = mBeamList->GetRect() ;

	theRect1 = theRect ;
	theRect1.bottom = theRect.top ;
	theRect1.top = theRect1.bottom-20 ;
	mOutputSelectedBeams = new TRadioButton(&theRect1,"Beams to Output") ;
	AddObject(mOutputSelectedBeams) ;
	OffsetRect(&theRect1,0,-20) ;
	mOutputAllBeams = new TRadioButton(&theRect1,"Output All Beams") ;
	AddObject(mOutputAllBeams) ;
	
	if (simRec->mNumReflectionsToOutput == 0) {
		mOutputAllBeams->SetState(true) ;
		mBeamList->Disable() ;
	}
	else {
		mOutputSelectedBeams->SetState(true) ;
		for(i=0; i < simRec->mNumReflectionsToOutput; i++) {
			sprintf(text,"%3d %3d %3d",simRec->mOutputReflection[i].a[0],simRec->mOutputReflection[i].a[1],simRec->mOutputReflection[i].a[2]) ;
			mBeamList->AddRow(text) ;
		}
	}
	
	OffsetRect(&theRect,theRect.right-theRect.left+10,0) ;
	theRect.bottom = theRect.top+20 ;
	theRect.right = theRect.left+60 ;
	mBeam = new TEditText(&theRect,"","") ;
	if (simRec->mNumReflectionsToOutput == 0) mBeam->Disable() ;
	OffsetRect(&theRect,0,25) ;
	mAddBeam = new TButton(&theRect,"Add") ;
	mAddBeam->Disable() ;
	AddObject(mAddBeam) ;
	OffsetRect(&theRect,0,25) ;
	mDeleteBeam = new TButton(&theRect,"Delete") ;
	mDeleteBeam->Disable() ;
	AddObject(mDeleteBeam) ;
	
	AddTextEditObject(mBethe) ;
	AddTextEditObject(mExact) ;

	AddTextEditObject(mAutoSG_Bethe_Max) ;

	AddTextEditObject(mAutoZ_Max) ;
	AddTextEditObject(mAutoZ_Min) ;

	AddTextEditObject(mAutoX1_G_Max) ;

	AddTextEditObject(mAutoSG_Exact_Max) ;
	AddTextEditObject(mAutoGMax) ;

	AddTextEditObject(mAutoMaxReflections) ;
	
	AddTextEditObject(mBranchEnd) ;
	AddTextEditObject(mBranchStart) ;

	AddTextEditObject(mXYRange_YEnd) ;
	AddTextEditObject(mXYRange_YBegin) ;

	AddTextEditObject(mXYRange_XEnd) ;
	AddTextEditObject(mXYRange_XBegin) ;

	AddTextEditObject(mXYRange_YPoints) ;
	AddTextEditObject(mXYRange_XPoints) ;

	AddTextEditObject(mG_Max) ;
	AddTextEditObject(mG_Min) ;
	AddTextEditObject(mH_Max) ;
	AddTextEditObject(mH_Min) ;
	AddTextEditObject(mPixelsAcrossDisk) ;
	
	AddTextEditObject(mBeam) ;
	AddTextEditObject(mThickness) ;
	AddTextEditObject(mVoltage) ;
	
	AddTextEditObject(mTiltK) ;
	AddTextEditObject(mTiltH) ;

	AddTextEditObject(mBaseVectorW) ;
	AddTextEditObject(mBaseVectorV) ;
	AddTextEditObject(mBaseVectorU) ;

	AddTextEditObject(mSurfaceNormalW) ;
	AddTextEditObject(mSurfaceNormalV) ;
	AddTextEditObject(mSurfaceNormalU) ;

	AddTextEditObject(mZoneW) ;
	AddTextEditObject(mZoneV) ;
	AddTextEditObject(mZoneU) ;
	
	mOKButton = new TButton(width-16-70,height-16-20,width-16,height-16,"OK") ;
	AddObject(mOKButton) ;
	theRect = mOKButton->GetRect() ;
	OffsetRect(&theRect,-(theRect.right-theRect.left+10),0) ;
	mCancelButton = new TButton(&theRect,"Cancel") ;
	AddObject(mCancelButton) ;
	SetDefaultButton(mOKButton) ;

	if (simRec->mMode == kMBFIT_SIM_MODE) {
		mBranchStart->Disable() ;
		mBranchEnd->Disable() ;
		m0D->Disable() ;

		mXYRange_XPoints->Disable() ;
		mXYRange_YPoints->Disable() ;
		mXYRange_XBegin->Disable() ;
		mXYRange_XEnd->Disable() ;
		mXYRange_YBegin->Disable() ;
		mXYRange_YEnd->Disable() ;
	}
	else if (simRec->mMode == kMBFIT_DISPERSION_MODE) {
		m2D_Disk->Disable() ;
		m2D_Square->Disable() ;
		m0D->Disable() ;
		m0D->Disable() ;
		mOutputAllBeams->Disable() ;
		mXYRange_XPoints->Disable() ;
		mXYRange_YPoints->Disable() ;
		mXYRange_XBegin->Disable() ;
		mXYRange_XEnd->Disable() ;
		mXYRange_YBegin->Disable() ;
		mXYRange_YEnd->Disable() ;
	}
	else if (simRec->mMode == kMBFIT_BLOCH_MODE) {
		m2D_Disk->Disable() ;
		m2D_Square->Disable() ;
		m1D->Disable() ;
		mThickness->Disable() ;
		mThicknessList->Disable() ;
	}

	if (simRec->mBeamSearchType == kMBFIT_AutoBeamSearch_SG) {
		mManualBeamSearch->SetState(false) ;
		mExactList->Disable() ;
		mBetheList->Disable() ;
	}
	else if (simRec->mBeamSearchType == kMBFIT_ManualBeamSearch) {
		mAutomaticBeamSearch->SetState(false) ;
		mAutoMaxReflections->Disable() ;
		mAutoZ_Min->Disable() ;
		mAutoZ_Max->Disable() ;
		mAutoGMax->Disable() ;
		mAutoSG_Exact_Max->Disable() ;
		mAutoSG_Bethe_Max->Disable() ;
		mAutoX1_G_Max->Disable() ;
	}
	
}

void	TMBFITSimulationParameterDialog::DoKeyDownInObject(TObject *obj) {
	char text[256] ;
	if (obj == mThickness) {
		mThickness->GetEditText(text) ;
		if (strlen(text) > 0) {
			if (!mAddThickness->IsEnabled()) mAddThickness->Enable() ;
		}
	}
	else if (obj == mBeam) {
		mBeam->GetEditText(text) ;
		if (strlen(text) > 0) {
			if (!mAddBeam->IsEnabled()) mAddBeam->Enable() ;
		}
	}
	else if (obj == mExact) {
		mExact->GetEditText(text) ;
		if (strlen(text) > 0) {
			if (!mAddExact->IsEnabled()) mAddExact->Enable() ;
		}
	}
	else if (obj == mBethe) {
		mBethe->GetEditText(text) ;
		if (strlen(text) > 0) {
			if (!mAddBethe->IsEnabled()) mAddBethe->Enable() ;
		}
	}
}

void	TMBFITSimulationParameterDialog::DoMouseDownInObject(TObject *obj,Point *pt) {
	char	text[256] ;
	short	theSelectedRow ;
	ListCellT	theCell ;
	
	theCell.col = 1 ;
	
	
	if (pt != nil)
		if (!obj->DoMouseDown(pt)) return ;
	if (obj == mOKButton) {
		mDone = true ;
		mDialogReturnCode = kDialogOK ;
	}
	else if (obj == mCancelButton) {
		mDone = true ;
		mDialogReturnCode = kDialogCancel ;
	}
	else if (obj == m2D_Disk) {
		mMBFITSimulationRecord->mIsModified = true ;
		m2D_Square->SetState(false) ;
		m1D->SetState(false) ;
		m0D->SetState(false) ;
	}
	else if (obj == m2D_Square) {
		mMBFITSimulationRecord->mIsModified = true ;
		m2D_Disk->SetState(false) ;
		m1D->SetState(false) ;
		m0D->SetState(false) ;
	}
	else if (obj == m1D) {
		mMBFITSimulationRecord->mIsModified = true ;
		m2D_Square->SetState(false) ;
		m2D_Disk->SetState(false) ;
		m0D->SetState(false) ;
	}
	else if (obj == m0D) {
		mMBFITSimulationRecord->mIsModified = true ;
		m2D_Square->SetState(false) ;
		m1D->SetState(false) ;
		m2D_Disk->SetState(false) ;
	}
	else if (obj == mAutomaticBeamSearch) {
		mMBFITSimulationRecord->mIsModified = true ;
		mManualBeamSearch->SetState(false) ;
		mAutoMaxReflections->Enable() ;
		mAutoZ_Min->Enable() ;
		mAutoZ_Max->Enable() ;
		mAutoGMax->Enable() ;
		mAutoSG_Exact_Max->Enable() ;
		mAutoSG_Bethe_Max->Enable() ;
		mAutoX1_G_Max->Enable() ;
		mExactList->Disable() ;
		mBetheList->Disable() ;
		mExact->Disable() ;
		mBethe->Disable() ;
	}
	else if (obj == mManualBeamSearch) {
		mMBFITSimulationRecord->mIsModified = true ;
		mAutomaticBeamSearch->SetState(false) ;
		mAutoMaxReflections->Disable() ;
		mAutoZ_Min->Disable() ;
		mAutoZ_Max->Disable() ;
		mAutoGMax->Disable() ;
		mAutoSG_Exact_Max->Disable() ;
		mAutoSG_Bethe_Max->Disable() ;
		mAutoX1_G_Max->Disable() ;
		mExactList->Enable() ;
		mBetheList->Enable() ;
		mExact->Enable() ;
		mBethe->Enable() ;
	}
	else if (obj == mOutputAllBeams) {
		mMBFITSimulationRecord->mIsModified = true ;
		mOutputSelectedBeams->SetState(false) ;
		mBeamList->Disable() ;
		mBeam->Disable() ;
		mAddBeam->Disable() ;
		mDeleteBeam->Disable() ;
		
	}
	else if (obj == mOutputSelectedBeams) {
		mOutputAllBeams->SetState(false) ;
		mBeamList->Enable() ;
		mBeam->Enable() ;
		if(mBeamList->GetSelectedRow() != 0) mDeleteBeam->Enable() ;
	}
	else if (obj == mThicknessList) {
		if(mThicknessList->GetSelectedRow() != 0) mDeleteThickness->Enable() ;	
		else mDeleteThickness->Disable() ;
	}
	else if (obj == mBeamList) {
		if(mBeamList->GetSelectedRow() != 0) mDeleteBeam->Enable() ;		
		else mDeleteBeam->Disable() ;
	}
	else if (obj == mExactList) {
		if(mExactList->GetSelectedRow() != 0) mDeleteExact->Enable() ;		
		else mDeleteExact->Disable() ;
	}
	else if (obj == mBetheList) {
		if(mBetheList->GetSelectedRow() != 0) mDeleteBethe->Enable() ;		
		else mDeleteBethe->Disable() ;
	}
	else if (obj == mAddThickness) {
		mThickness->GetEditText(text) ;
		if (strlen(text) > 0) {
			mMBFITSimulationRecord->mIsModified = true ;
			mThicknessList->AddRow(text) ;
			mThickness->SetEditText("") ;
			mAddThickness->Disable() ;
		}
	}
	else if (obj == mAddBeam) {
		mBeam->GetEditText(text) ;
		if (strlen(text) > 0) {
			mMBFITSimulationRecord->mIsModified = true ;
			mBeamList->AddRow(text) ;
			mBeam->SetEditText("") ;
			mAddBeam->Disable() ;
		}
	}
	else if (obj == mAddExact) {
		mExact->GetEditText(text) ;
		if (strlen(text) > 0) {
			mMBFITSimulationRecord->mIsModified = true ;
			mExactList->AddRow(text) ;
			mExact->SetEditText("") ;
			mAddExact->Disable() ;
		}
	}
	else if (obj == mAddBethe) {
		mBethe->GetEditText(text) ;
		if (strlen(text) > 0) {
			mMBFITSimulationRecord->mIsModified = true ;
			mBetheList->AddRow(text) ;
			mBethe->SetEditText("") ;
			mAddBethe->Disable() ;
		}
	}
	else if (obj == mDeleteThickness) {
		mThicknessList->RemoveRow(mThicknessList->GetSelectedRow()) ;		
		mMBFITSimulationRecord->mIsModified = true ;
	}
	else if (obj == mDeleteBeam) {
		mBeamList->RemoveRow(mBeamList->GetSelectedRow()) ;		
		mMBFITSimulationRecord->mIsModified = true ;
	}
	else if (obj == mDeleteExact) {
		mExactList->RemoveRow(mExactList->GetSelectedRow()) ;		
		mMBFITSimulationRecord->mIsModified = true ;
	}
	else if (obj == mDeleteBethe) {
		mBetheList->RemoveRow(mBetheList->GetSelectedRow()) ;		
		mMBFITSimulationRecord->mIsModified = true ;
	}

}

void	TMBFITSimulationParameterDialog::RetrieveDialogItems() {
	// Read back the prameters in the dialog box and update the simulation record with the
	// new parameters
	char	text[256] ;
	float   val,tolerance=0.000001 ;
	long	ival,i ;
	ListCellT	theCell ;
	
	theCell.col = 1 ;
	
	mZoneU->GetEditText(text) ;
	val = atof(text) ;
	if (fabs(val-mMBFITSimulationRecord->mZoneAxis.a[0]) > tolerance) {
		mMBFITSimulationRecord->mZoneAxis.a[0] = val ;
		mMBFITSimulationRecord->mIsModified = true ;
	}

	mZoneV->GetEditText(text) ;
	val = atof(text) ;
	if (fabs(val-mMBFITSimulationRecord->mZoneAxis.a[1]) > tolerance) {
		mMBFITSimulationRecord->mZoneAxis.a[1] = val ;
		mMBFITSimulationRecord->mIsModified = true ;
	}
	
	mZoneW->GetEditText(text) ;
	val = atof(text) ;
	if (fabs(val-mMBFITSimulationRecord->mZoneAxis.a[2]) > tolerance) {
		mMBFITSimulationRecord->mZoneAxis.a[2] = val ;
		mMBFITSimulationRecord->mIsModified = true ;
	}	

	// Surface Normal
	mSurfaceNormalU->GetEditText(text) ;
	val = atof(text) ;
	if (fabs(val-mMBFITSimulationRecord->mSurfaceNormal.a[0]) > tolerance) {
		mMBFITSimulationRecord->mSurfaceNormal.a[0] = val ;
		mMBFITSimulationRecord->mIsModified = true ;
	}

	mSurfaceNormalV->GetEditText(text) ;
	val = atof(text) ;
	if (fabs(val-mMBFITSimulationRecord->mSurfaceNormal.a[1]) > tolerance) {
		mMBFITSimulationRecord->mSurfaceNormal.a[1] = val ;
		mMBFITSimulationRecord->mIsModified = true ;
	}
	
	mSurfaceNormalW->GetEditText(text) ;
	val = atof(text) ;
	if (fabs(val-mMBFITSimulationRecord->mSurfaceNormal.a[2]) > tolerance) {
		mMBFITSimulationRecord->mSurfaceNormal.a[2] = val ;
		mMBFITSimulationRecord->mIsModified = true ;
	}	

	// Zone Base Vector
	mBaseVectorU->GetEditText(text) ;
	val = atof(text) ;
	if (fabs(val-mMBFITSimulationRecord->mZoneBaseVector.a[0]) > tolerance) {
		mMBFITSimulationRecord->mZoneBaseVector.a[0] = val ;
		mMBFITSimulationRecord->mIsModified = true ;
	}

	mBaseVectorV->GetEditText(text) ;
	val = atof(text) ;
	if (fabs(val-mMBFITSimulationRecord->mZoneBaseVector.a[1]) > tolerance) {
		mMBFITSimulationRecord->mZoneBaseVector.a[1] = val ;
		mMBFITSimulationRecord->mIsModified = true ;
	}
	
	mBaseVectorW->GetEditText(text) ;
	val = atof(text) ;
	if (fabs(val-mMBFITSimulationRecord->mZoneBaseVector.a[2]) > tolerance) {
		mMBFITSimulationRecord->mZoneBaseVector.a[2] = val ;
		mMBFITSimulationRecord->mIsModified = true ;
	}	

	mVoltage->GetEditText(text) ;
	val = atof(text) ;
	if (fabs(val-mMBFITSimulationRecord->mMicroscope.mVoltage) > tolerance) {
		mMBFITSimulationRecord->mMicroscope.mVoltage = val ;
		mMBFITSimulationRecord->mIsModified = true ;
	}
	
	// Tilt
	mTiltH->GetEditText(text) ;
	val = atof(text) ;
	if (fabs(val-mMBFITSimulationRecord->mTiltH) > tolerance) {
		mMBFITSimulationRecord->mTiltH = val ;
		mMBFITSimulationRecord->mIsModified = true ;
	}	

	mTiltK->GetEditText(text) ;
	val = atof(text) ;
	if (fabs(val-mMBFITSimulationRecord->mTiltG) > tolerance) {
		mMBFITSimulationRecord->mTiltG = val ;
		mMBFITSimulationRecord->mIsModified = true ;
	}	

	// Calculation type range
	if (m2D_Disk->GetState()) mMBFITSimulationRecord->mCalcType = kMBFIT_CALC_KXY_RANGE_Disk ;
	else if (m2D_Square->GetState()) mMBFITSimulationRecord->mCalcType = kMBFIT_CALC_KXY_RANGE_Square ;
	else if (m1D->GetState()) mMBFITSimulationRecord->mCalcType = kMBFIT_CALC_KXY_RANGE_1D ;
	else if (m0D->GetState()) mMBFITSimulationRecord->mCalcType = kMBFIT_CALC_KXY_RANGE_0D ;

	mPixelsAcrossDisk->GetEditText(text) ;
	ival = atoi(text) ;
	if (ival != mMBFITSimulationRecord->mNumCalculationPoints) {
		mMBFITSimulationRecord->mNumCalculationPoints = ival ;
		mMBFITSimulationRecord->mIsModified = true ;
	}	

	mH_Min->GetEditText(text) ;
	val = atof(text) ;
	if (fabs(val-mMBFITSimulationRecord->mMinH) > tolerance) {
		mMBFITSimulationRecord->mMinH = val ;
		mMBFITSimulationRecord->mIsModified = true ;
	}	

	mH_Max->GetEditText(text) ;
	val = atof(text) ;
	if (fabs(val-mMBFITSimulationRecord->mMaxH) > tolerance) {
		mMBFITSimulationRecord->mMaxH = val ;
		mMBFITSimulationRecord->mIsModified = true ;
	}	

	mG_Min->GetEditText(text) ;
	val = atof(text) ;
	if (fabs(val-mMBFITSimulationRecord->mMinG) > tolerance) {
		mMBFITSimulationRecord->mMinG = val ;
		mMBFITSimulationRecord->mIsModified = true ;
	}	

	mG_Max->GetEditText(text) ;
	val = atof(text) ;
	if (fabs(val-mMBFITSimulationRecord->mMaxG) > tolerance) {
		mMBFITSimulationRecord->mMaxG = val ;
		mMBFITSimulationRecord->mIsModified = true ;
	}	

	mXYRange_XPoints->GetEditText(text) ;
	ival = atoi(text) ;
	if (ival != mMBFITSimulationRecord->mNumCalculationPointsX) {
		mMBFITSimulationRecord->mNumCalculationPointsX = ival ;
		mMBFITSimulationRecord->mIsModified = true ;
	}	

	mXYRange_XBegin->GetEditText(text) ;
	val = atof(text) ;
	if (fabs(val-mMBFITSimulationRecord->mMinX) > tolerance) {
		mMBFITSimulationRecord->mMinX = val ;
		mMBFITSimulationRecord->mIsModified = true ;
	}	

	mXYRange_XEnd->GetEditText(text) ;
	val = atof(text) ;
	if (fabs(val-mMBFITSimulationRecord->mMaxX) > tolerance) {
		mMBFITSimulationRecord->mMaxX = val ;
		mMBFITSimulationRecord->mIsModified = true ;
	}	

	mXYRange_YPoints->GetEditText(text) ;
	ival = atoi(text) ;
	if (ival != mMBFITSimulationRecord->mNumCalculationPointsY) {
		mMBFITSimulationRecord->mNumCalculationPointsY = ival ;
		mMBFITSimulationRecord->mIsModified = true ;
	}	

	mXYRange_YBegin->GetEditText(text) ;
	val = atof(text) ;
	if (fabs(val-mMBFITSimulationRecord->mMinY) > tolerance) {
		mMBFITSimulationRecord->mMinY = val ;
		mMBFITSimulationRecord->mIsModified = true ;
	}	

	mXYRange_YEnd->GetEditText(text) ;
	val = atof(text) ;
	if (fabs(val-mMBFITSimulationRecord->mMaxY) > tolerance) {
		mMBFITSimulationRecord->mMaxY = val ;
		mMBFITSimulationRecord->mIsModified = true ;
	}	

	// Dispersion surface branches
	mBranchStart->GetEditText(text) ;
	ival = atoi(text) ;
	if (ival != mMBFITSimulationRecord->branchBegin) {
		mMBFITSimulationRecord->branchBegin = ival ;
		mMBFITSimulationRecord->mIsModified = true ;
	}	

	mBranchEnd->GetEditText(text) ;
	ival = atoi(text) ;
	if (ival != mMBFITSimulationRecord->branchEnd) {
		mMBFITSimulationRecord->branchEnd = ival ;
		mMBFITSimulationRecord->mIsModified = true ;
	}	
	
	// Beam search method
	// Calculation type range
	if (mAutomaticBeamSearch->GetState()) {
		if (mMBFITSimulationRecord->mBeamSearchType != kMBFIT_AutoBeamSearch_SG) mMBFITSimulationRecord->mIsModified = true ;
		mMBFITSimulationRecord->mBeamSearchType = kMBFIT_AutoBeamSearch_SG ;
		mAutoMaxReflections->GetEditText(text) ;
		ival = atoi(text) ;
		if (ival != mMBFITSimulationRecord->mBeamSearchAuto.num_g_limit) {
			mMBFITSimulationRecord->mBeamSearchAuto.num_g_limit = ival ;
			mMBFITSimulationRecord->mIsModified = true ;
		}	
		mAutoGMax->GetEditText(text) ;
		val = atof(text) ;
		if (fabs(val-mMBFITSimulationRecord->mBeamSearchAuto.g_max) > tolerance) {
			mMBFITSimulationRecord->mBeamSearchAuto.g_max = val ;
			mMBFITSimulationRecord->mIsModified = true ;
		}	
		mAutoSG_Exact_Max->GetEditText(text) ;
		val = atof(text) ;
		if (fabs(val-mMBFITSimulationRecord->mBeamSearchAuto.sg_exact_max) > tolerance) {
			mMBFITSimulationRecord->mBeamSearchAuto.sg_exact_max = val ;
			mMBFITSimulationRecord->mIsModified = true ;
		}	
		mAutoX1_G_Max->GetEditText(text) ;
		val = atof(text) ;
		if (fabs(val-mMBFITSimulationRecord->mBeamSearchAuto.xi_g_max) > tolerance) {
			mMBFITSimulationRecord->mBeamSearchAuto.xi_g_max = val ;
			mMBFITSimulationRecord->mIsModified = true ;
		}	

		mAutoZ_Min->GetEditText(text) ;
		ival = atoi(text) ;
		if (ival != mMBFITSimulationRecord->mBeamSearchAuto.lz_min) {
			mMBFITSimulationRecord->mBeamSearchAuto.lz_min = ival ;
			mMBFITSimulationRecord->mIsModified = true ;
		}	
		mAutoZ_Max->GetEditText(text) ;
		ival = atoi(text) ;
		if (ival != mMBFITSimulationRecord->mBeamSearchAuto.lz_max) {
			mMBFITSimulationRecord->mBeamSearchAuto.lz_max = ival ;
			mMBFITSimulationRecord->mIsModified = true ;
		}	
		mAutoSG_Bethe_Max->GetEditText(text) ;
		val = atof(text) ;
		if (fabs(val-mMBFITSimulationRecord->mBeamSearchAuto.sg_bethe_max) > tolerance) {
			mMBFITSimulationRecord->mBeamSearchAuto.sg_bethe_max = val ;
			mMBFITSimulationRecord->mIsModified = true ;
		}	
	}
	else if (mManualBeamSearch->GetState()) {
		if (mMBFITSimulationRecord->mBeamSearchType != kMBFIT_ManualBeamSearch) mMBFITSimulationRecord->mIsModified = true ;
		mMBFITSimulationRecord->mBeamSearchType = kMBFIT_ManualBeamSearch ;
		mMBFITSimulationRecord->mNumExact = mExactList->GetNumberOfListItems() ;
		if (mMBFITSimulationRecord->mExact != nil) {
			delete[] mMBFITSimulationRecord->mExact ;
			mMBFITSimulationRecord->mExact = nil ;
		}
		if (mMBFITSimulationRecord->mNumExact > 0) mMBFITSimulationRecord->mExact = new LVector[mMBFITSimulationRecord->mNumExact] ;
		for(i=0; i < mMBFITSimulationRecord->mNumExact; i++) {
			theCell.row = i+1 ;
			mExactList->FetchCellData(theCell,text) ;
			sscanf(text,"%d %d %d",&mMBFITSimulationRecord->mExact[i].a[0],&mMBFITSimulationRecord->mExact[i].a[1],&mMBFITSimulationRecord->mExact[i].a[2]) ;
		}
		mMBFITSimulationRecord->mNumBethe = mBetheList->GetNumberOfListItems() ;
		if (mMBFITSimulationRecord->mBethe != nil) {
			delete[] mMBFITSimulationRecord->mBethe ;
			mMBFITSimulationRecord->mBethe = nil ;
		}
		if (mMBFITSimulationRecord->mNumBethe > 0) mMBFITSimulationRecord->mBethe = new LVector[mMBFITSimulationRecord->mNumBethe] ;
		for(i=0; i < mMBFITSimulationRecord->mNumBethe; i++) {
			theCell.row = i+1 ;
			mBetheList->FetchCellData(theCell,text) ;
			sscanf(text,"%d %d %d",&mMBFITSimulationRecord->mBethe[i].a[0],&mMBFITSimulationRecord->mBethe[i].a[1],
				&mMBFITSimulationRecord->mBethe[i].a[2]) ;
		}
	}
	// Thickness
	mMBFITSimulationRecord->mNumThicknesses = mThicknessList->GetNumberOfListItems() ;
	if (mMBFITSimulationRecord->mThicknesses != nil) {
		delete[] mMBFITSimulationRecord->mThicknesses ;
		mMBFITSimulationRecord->mThicknesses = nil ;
	}
	if (mMBFITSimulationRecord->mNumThicknesses > 0) mMBFITSimulationRecord->mThicknesses = new float[mMBFITSimulationRecord->mNumThicknesses] ;
	for(i=0; i < mMBFITSimulationRecord->mNumThicknesses; i++) {
		theCell.row = i+1 ;
		mThicknessList->FetchCellData(theCell,text) ;
		sscanf(text,"%f",&mMBFITSimulationRecord->mThicknesses[i]) ;
	}
	// Output reflections
	if (mOutputAllBeams->GetState()) {
		mMBFITSimulationRecord->mNumReflectionsToOutput = 0 ;
	}
	else {
		mMBFITSimulationRecord->mNumReflectionsToOutput = mBeamList->GetNumberOfListItems() ;
		if (mMBFITSimulationRecord->mOutputReflection != nil) {
			delete[] mMBFITSimulationRecord->mOutputReflection ;
			mMBFITSimulationRecord->mOutputReflection = nil ;
		}
		if (mMBFITSimulationRecord->mNumReflectionsToOutput > 0) mMBFITSimulationRecord->mOutputReflection = new LVector[mMBFITSimulationRecord->mNumReflectionsToOutput] ;
		for(i=0; i < mMBFITSimulationRecord->mNumReflectionsToOutput; i++) {
			theCell.row = i+1 ;
			mBeamList->FetchCellData(theCell,text) ;
			sscanf(text,"%d %d %d",&mMBFITSimulationRecord->mOutputReflection[i].a[0],&mMBFITSimulationRecord->mOutputReflection[i].a[1],
				&mMBFITSimulationRecord->mOutputReflection[i].a[2]) ;
		}
	}
}

char	*FindMBFITFileTag(char *stream,char *tag) ;
// Finds the tag and advances to the next line
char	*FindMBFITFileTag(char *stream,char *tag) {
	char	inputLine[256] ;
	char	*pos = stream ;
	Boolean foundTag = false ;
	while( (pos = ReadLine(pos,inputLine,256)) != NULL) {
		if (strstr(inputLine,tag) != NULL) {
			foundTag = true ;
			break ;
		}
	}
	if (foundTag) return pos ;
	return NULL ;
}

TCrystal::TCrystal(TCrystalWindow *crystalWindow,FSInfo &fsInfo) {
	
	mDocWindow  = crystalWindow ;
	mSelection  = nil ;
	mHasAtomsSelected = false ;
	mSecondSetting 		= false ;
	mHexagonalIndices 	= false ; 
	mAlternateSetting 	= false ; 
	mCentroSymmetric 	= false ; 
	
	mSymOps = nil ;

	mZone.a[0]		= 0. ;
	mZone.a[1]		= 0. ;
	mZone.a[2]		= 1. ;
	mZone.a[3]		= 0. ;
	
	mAlpha  = 90. ;
	mBeta   = 90. ;
	
	if (!InitializeFromNCEMSSPrintFile(fsInfo))
		throw "Failed to Initialize structure from the NCEMSS Print file" ;
	
	DegreeToRad(&mAlpha) ;
	DegreeToRad(&mBeta) ;
	DegreeToRad(&mGamma) ;
	
	mLightX 		= 1. ;
	mLightY 		= 1. ;
	mLightZ			= 3. ;
	mXViewCells		= 1 ;
	mXViewCells		= 1 ;
	mXViewCells		= 1 ;
	mShiftX = 0 ;
	mShiftY = 0 ;
	
	mAtomScaleFactor = 0.5 ;	
	
	mScreenAtom		= nil ;
	mDrawOutline = true ;
	mDrawOrigin = true ;
	mPlanePt		= kNoPlanePt ;
	
	mWAtom = kFirstAtom ;
	mAnglePt = kNoAnglePt ;
	
	mDisplayRect	= mDocWindow->mCrystalRect ;
	
	mWidth = mDisplayRect.right-mDisplayRect.left ;
	mHeight = mDisplayRect.bottom-mDisplayRect.top ;
	
 	// Allocate the atoms required for the view
	CreateView(1,1,1) ;	
    
	SetTMatrix() ;
	SetSMatrix() ;
	
	mMagnification = 1. ;
	
	TVector tZone = mZone ;
	
	SetTransformationMatrix(tZone) ;
	
	CreateUnitCellViewAtoms() ;
	mRotMatrix = 1.0 ; 
	
}


// ------------------------------------------------------
// Create an instance of the class TCrystal from a file
//

TCrystal::TCrystal(FSInfo &fsInfo) {

	mSelection 		= nil ;
	mHasAtomsSelected = false ;
	mSecondSetting 		= false ;
	mHexagonalIndices 	= false ; 
	mAlternateSetting 	= false ; 
	mCentroSymmetric 	= false ; 
	
	mZone.a[0]		= 0. ;
	mZone.a[1]		= 0. ;
	mZone.a[2]		= 1. ;
	mZone.a[3]		= 0. ;
	mLightX 		= 1. ;
	mLightY 		= 1. ;
	mLightZ			= 3. ;
	mXViewCells		= 1 ;
	mXViewCells		= 1 ;
	mXViewCells		= 1 ;
	mShiftX = 0 ;
	mShiftY = 0 ;
	
	mAtomScaleFactor = 0.5 ;	
	
	mScreenAtom		= nil ;
	mDrawOutline = true ;
	mDrawOrigin = true ;
	mPlanePt		= kNoPlanePt ;
	
	mWAtom = kFirstAtom ;
	mAnglePt = kNoAnglePt ;
	
	if (strstr(fsInfo.mFileName,".at") != NULL || strstr(fsInfo.mFileName,".AT") != NULL ) {
		if (!InitializeFromNCEMSSFile(fsInfo))
			throw "Failed to Initialize structure from the NCEMSS file" ;
		// Try to set the dimension for the simulation as a test
		mTEMRecord.SetCalculationSize(this) ;
		mDocWindow 		= new TCrystalWindow(this,kNCEMSS_FILE) ; 
	}
	else if (strstr(fsInfo.mFileName,".txt") != NULL || strstr(fsInfo.mFileName,".TXT") != NULL) {
		// Try to read it as a Tsuda CBED calculation file
		if (!InitializeFromMBFITFile(fsInfo))
			throw "Failed to Initialize structure from the MBFIT file" ;
		mDocWindow 		= new TCrystalWindow(this,kMBFIT_FILE) ; 
	} 

	mDisplayRect	= mDocWindow->mCrystalRect ;
	
	char	windowName[256] ;
	strcpy(windowName,fsInfo.mFileName) ;
	char *cp = strrchr(windowName,'.') ;
	*cp = '\0' ;
	if ((cp = strrchr(windowName,'/')) == NULL)
		mDocWindow->SetWindowTitle(windowName) ;
	else
		mDocWindow->SetWindowTitle(cp+1) ;
     
 	// Allocate the atoms required for the view
	CreateView(1,1,1) ;	
    
	SetTMatrix() ;
	SetSMatrix() ;
    
	mWidth = mDisplayRect.right-mDisplayRect.left ;
	mHeight = mDisplayRect.bottom-mDisplayRect.top ;
	    
	mMagnification = 1. ;
	
	TVector tZone = mZone ;
			
	SetTransformationMatrix(tZone) ;
	
	CreateUnitCellViewAtoms() ;
	mRotMatrix = 1.0 ; 


}

void	TCrystal::Update() {
     
 	// Allocate the atoms required for the view
	CreateView(1,1,1) ;	
    
	SetTMatrix() ;
	SetSMatrix() ;
    
	mWidth = mDisplayRect.right-mDisplayRect.left ;
	mHeight = mDisplayRect.bottom-mDisplayRect.top ;
	    
	mMagnification = 1. ;
	
	TVector tZone = mZone ;
			
	SetTransformationMatrix(tZone) ;
	
	CreateUnitCellViewAtoms() ;
	mRotMatrix = 1.0 ; 


}

// This is here so that we can edit the structure
// We do not care about anything but the structure itself
TCrystal::TCrystal(TCrystal *s) {
	long	i,j ;
	mSelection 		= nil ;
	mHasAtomsSelected = false ;
	mSecondSetting 		= s->mSecondSetting ;
	mHexagonalIndices 	= s->mHexagonalIndices ; 
	mAlternateSetting 	= s->mAlternateSetting ; 
	mCentroSymmetric 	= s->mCentroSymmetric ; 
	
	mAtomScaleFactor = s->mAtomScaleFactor ;	
	
	mScreenAtom		= nil ;	
	
	mA		= s->mA ;
	mB		= s->mB ;
	mC		= s->mC ;
	mAlpha		= s->mAlpha ;
	mBeta		= s->mBeta ;
	mGamma		= s->mGamma ;
	mSpacegroup		= s->mSpacegroup ;

	mNumSymOps		= s->mNumSymOps ;
	mNumBasisAtoms	= s->mNumBasisAtoms ;
	mNumUnitCellAtoms	= s->mNumUnitCellAtoms ;
	mNumTypes		= s->mNumTypes ;

	mBasisPtr = new TAtom[mNumBasisAtoms] ;
	for (i=0;i<mNumBasisAtoms;i++)    {
		mBasisPtr[i] = s->mBasisPtr[i] ;
	}


	mSymOps = new TMatrix[mNumSymOps] ;
	for (i=0;i<mNumSymOps;i++)    {
		mSymOps[i] = s->mSymOps[i] ;
		strcpy(mSymopTxt[i],s->mSymopTxt[i]) ;
	}
	
	// Apply the symmetry operators to the basis
	mAtomPtr = new TAtom[mNumBasisAtoms*200] ;	// This is the maximum we can have, so we don't need
	Assert(mAtomPtr) ;							// to reallocate this storage
	mAllocatedAtoms = mNumBasisAtoms*200 ;
	for (i=0;i<mNumUnitCellAtoms;i++)    {
		mAtomPtr[i] = s->mAtomPtr[i] ;
	}
	// That should be all we need
}

// We are only really interested in copying the actual structure
TCrystal& TCrystal::operator=(const TCrystal &s) {
	long	i ;
	mSecondSetting 		= s.mSecondSetting ;
	mHexagonalIndices 	= s.mHexagonalIndices ; 
	mAlternateSetting 	= s.mAlternateSetting ; 
	mCentroSymmetric 	= s.mCentroSymmetric ; 
	
	mA		= s.mA ;
	mB		= s.mB ;
	mC		= s.mC ;
	mAlpha		= s.mAlpha ;
	mBeta		= s.mBeta ;
	mGamma		= s.mGamma ;
	mSpacegroup		= s.mSpacegroup ;
	
	if (mNumBasisAtoms != s.mNumBasisAtoms) {
		delete[] mBasisPtr ;
		mNumBasisAtoms = s.mNumBasisAtoms ;		
		mBasisPtr = new TAtom[mNumBasisAtoms] ;
		delete[] mAtomPtr ;
		mAtomPtr = new TAtom[mNumBasisAtoms*200] ;	// This is the maximum we can have, so we don't need
		mAllocatedAtoms = mNumBasisAtoms*200 ;
	}
	mNumTypes		= s.mNumTypes ;

	for (i=0;i<mNumBasisAtoms;i++)    {
		mBasisPtr[i] = s.mBasisPtr[i] ;
	}

	if (mNumSymOps != s.mNumSymOps) {
		delete[] mSymOps ;
		mNumSymOps = s.mNumSymOps ;
		mSymOps = new TMatrix[mNumSymOps] ;
	}
	for (i=0;i<mNumSymOps;i++)    {
		mSymOps[i] = s.mSymOps[i] ;
		strcpy(mSymopTxt[i],s.mSymopTxt[i]) ;
	}
	
	mNumUnitCellAtoms = s.mNumUnitCellAtoms ;
	for (i=0;i<mNumUnitCellAtoms;i++)    {
		mAtomPtr[i] = s.mAtomPtr[i] ;
	}
	return *this ;
}

void	TCrystal::WriteNCEMSSControlFile() {
	
	FILE	*fp ;
	long	i,j ;
	TAtom   *thisAtom ;
	
	float   alpha,beta,gamma ;
	alpha = mAlpha ;
	beta = mBeta ;
	gamma = mGamma ;
	RadToDegree(&alpha) ;
	RadToDegree(&beta) ;
	RadToDegree(&gamma) ;
	
	if ((fp = fopen(mFSInfo.mFileName,"w")) == NULL) {
		PutMessage("Could not update the NCEMSS file") ;
		return ;
	}
	fprintf(fp,"%s\n",mTEMRecord.mTitle) ;
	fprintf(fp,"%d\n",mSpacegroup) ;
	fprintf(fp,"%f %f %f %f %f %f\n",mA,mB,mC,alpha,beta,gamma) ;
	fprintf(fp,"%f\n",mTEMRecord.mGMax) ; 
	fprintf(fp,"%d %d %d\n",near(mTEMRecord.mZoneAxis.a[0]),near(mTEMRecord.mZoneAxis.a[1]),near(mTEMRecord.mZoneAxis.a[2])) ;

	if (mTEMRecord.mNumSubSlices != 1) {
		fprintf(fp,"%d %d %d\n",mNumSymOps,mTEMRecord.mNumSubSlices,mTEMRecord.m3DSlicing) ;
	}
	else {
		fprintf(fp,"%d %d\n",mNumSymOps,mTEMRecord.mNumSubSlices) ;
	}
	fprintf(fp,"%d %d\n",mNumBasisAtoms,mNumTypes) ;

	for (j=0;j<mNumBasisAtoms;j++) {
		thisAtom = mBasisPtr+j ;
		fprintf(fp,"%d %s %f %f %f %f %f\n",
			thisAtom->type,thisAtom->name,
			thisAtom->xc,thisAtom->yc,
			thisAtom->zc,thisAtom->dw,
			thisAtom->occ) ;	
	}
	fprintf(fp,"%s %f %f %f\n",mTEMRecord.mMicroscope.mName,mTEMRecord.mMicroscope.mCs,
			mTEMRecord.mMicroscope.mSpreadOfFocus,mTEMRecord.mMicroscope.mConvergence) ;
	
	fprintf(fp,"%f\n",mTEMRecord.mMicroscope.mVoltage) ;	
	fprintf(fp,"%f %f\n",mTEMRecord.mMicroscope.mBeamTilt.a[0],mTEMRecord.mMicroscope.mBeamTilt.a[1]) ;

	if (mTEMRecord.mNumThicknesses > 1) {
		fprintf(fp,"%f,%f,%f\n",mTEMRecord.mStartThickness,mTEMRecord.mEndThickness,mTEMRecord.mDeltaThickness) ;
	}
	else {
		fprintf(fp,"%f\n",mTEMRecord.mStartThickness) ;
	}
	
	if (mTEMRecord.mStoreReflections) {
		fprintf(fp,"YES\n") ;
		fprintf(fp,"%d\n",mTEMRecord.mNumReflectionsToStore) ;
		for(j = 0; j < mTEMRecord.mNumReflectionsToStore; j++) {
			fprintf(fp,"%d %d %d\n",mTEMRecord.mReflectionsToStore[j].a[0],mTEMRecord.mReflectionsToStore[j].a[1],mTEMRecord.mReflectionsToStore[j].a[2]) ;
		}
	}
	else {
		fprintf(fp,"NO\n") ;
	}

	if (mTEMRecord.mNumFocii > 1) {
		fprintf(fp,"%f,%f,%f\n",mTEMRecord.mMicroscope.mStartFocus,mTEMRecord.mMicroscope.mEndFocus,mTEMRecord.mMicroscope.mDeltaFocus) ;
	}
	else {
		fprintf(fp,"%f\n",mTEMRecord.mMicroscope.mStartFocus) ;
	}
	

	fprintf(fp,"%f\n %f %f %f %f\n",mTEMRecord.mMicroscope.mAperture,
			mTEMRecord.mMicroscope.mApertureH,mTEMRecord.mMicroscope.mApertureK,mTEMRecord.mMicroscope.mOpticAxisH,mTEMRecord.mMicroscope.mOpticAxisK) ;
	for (i=0;i<mNumSymOps;fprintf(fp,"%s\n",txtptr_symop[i++])) ;
	
	if (mTEMRecord.mCalculationStatus < 1) mTEMRecord.mCalculationStatus = 1 ;
	fprintf(fp,"%d\n",mTEMRecord.mCalculationStatus) ;
	fclose(fp) ;
	
	mTEMRecord.mIsModified = false ;
}

Boolean	TCrystal::InitializeFromNCEMSSFile(FSInfo &fsInfo) {
	FILE 	*fp ;
	char	name[32],plots[32] ;
	
	mFSInfo = fsInfo ;

	fp = fopen(fsInfo.mFileName,"r") ;
	if (fp == nil) return false ;
	
	long i,iz,j,n,k,numVariablesAssigned ;
	char c,buffer[256],text[256] ;


	// File is ncemss structure file
	// read in the title
	if (ReadLine(mTEMRecord.mTitle,256,fp) == NULL) {
		PutMessage("Error Reading the title") ;
		fprintf(stderr,"\nRead Line: %s",buffer) ;
		fclose(fp) ;
		return false ;
	}
	if (gDebugMode) fprintf(stderr,"\nRead: %s",mTEMRecord.mTitle) ;
	// read in the space group number
	if (ReadLine(buffer,256,fp) == NULL) {
		PutMessage("Error Reading the space group number") ;
		fprintf(stderr,"\nRead Line: %s",buffer) ;
		fclose(fp) ;
		return false ;
	}
	if (gDebugMode) fprintf(stderr,"\nRead: %s",buffer) ;
	sscanf(buffer,"%hd",&mSpacegroup) ;

	// read in a,b,c,alpha,beta,gamma
	if (ReadLine(buffer,256,fp) == NULL) {
		PutMessage("Error Reading the lattice parameters") ;
		fprintf(stderr,"\nRead Line: %s",buffer) ;
		return false ;
	}
	if (gDebugMode) fprintf(stderr,"\nRead: %s",buffer) ;
	sscanf(buffer,"%f %f %f %f %f %f",&mA,&mB,&mC,&mAlpha,&mBeta,&mGamma) ;
	DegreeToRad(&mAlpha) ;
	DegreeToRad(&mBeta) ;
	DegreeToRad(&mGamma) ;

	// read in gmax
	if (ReadLine(buffer,256,fp) == NULL) {
		PutMessage("Error Reading gmax") ;
		fprintf(stderr,"\nRead Line: %s",buffer) ;
		fclose(fp) ;
		return false ;
	}
	if (gDebugMode) fprintf(stderr,"\nRead: %s",buffer) ;
	sscanf(buffer,"%f",&mTEMRecord.mGMax) ;
		
	// read in the incoming beam direction
	if (ReadLine(buffer,256,fp) == NULL) {
		fclose(fp) ;
		return false ;
	}
	if (gDebugMode) fprintf(stderr,"\nRead: %s",buffer) ;
	sscanf(buffer,"%f %f %f",&mTEMRecord.mZoneAxis.a[0],&mTEMRecord.mZoneAxis.a[1],&mTEMRecord.mZoneAxis.a[2]) ;
	
	// Number of symmetry operators , number of sub-slices and 3d slicing */
	if (ReadLine(buffer,256,fp) == NULL) {
		PutMessage("Error Reading number of symmetry operators") ;
		fprintf(stderr,"\nRead Line: %s",buffer) ;
		fclose(fp) ;
		return false ;
	}
	if (gDebugMode) fprintf(stderr,"\nRead: %s",buffer) ;
	sscanf(buffer,"%hd %hd %d",&mNumSymOps,&mTEMRecord.mNumSubSlices,&mTEMRecord.m3DSlicing) ;

	// number of basis atoms and types
	if (ReadLine(buffer,256,fp) == NULL) {
		PutMessage("Error Reading number of atoms and types of atoms") ;
		fprintf(stderr,"\nRead Line: %s",buffer) ;
		fclose(fp) ;
		return false ;
	}
	if (gDebugMode) fprintf(stderr,"\nRead: %s",buffer) ;
	sscanf(buffer,"%d %hd",&mNumBasisAtoms,&mNumTypes) ;
	
	mBasisPtr = new TAtom[mNumBasisAtoms] ;
	for (n=0;n<mNumBasisAtoms;n++)    {
		if (ReadLine(buffer,256,fp) == NULL) {
			PutMessage("Error Reading the atom information") ;
			fprintf(stderr,"\nRead Line: %s",buffer) ;
			fclose(fp) ;
			return false ;
		}
		mBasisPtr[n].dw		= 0. ;
		mBasisPtr[n].occ	= 1.0 ;
		if (gDebugMode) fprintf(stderr,"\nRead: %s",buffer) ;
		if ((numVariablesAssigned=sscanf(buffer,"%hd %s %f %f %f %f %f",
					&mBasisPtr[n].type,name,
					&mBasisPtr[n].xc,&mBasisPtr[n].yc,&mBasisPtr[n].zc,
					&mBasisPtr[n].dw,&mBasisPtr[n].occ)) < 5) {	
			fprintf(stderr,"\nERROR Read Line: %s for atom[%d]\n",buffer,n+1) ;
			PutMessage("Error Reading the atom information") ;
			fclose(fp) ;
			return false ;
		}
		mBasisPtr[n].no 	= (iz=atomno(name)) ;		// atomic number
		mBasisPtr[n].name 	= chemsymb[iz-1] ;          // chemical symbol
		mBasisPtr[n].rad 	= covrad[mBasisPtr[n].no-1]  ;
		mBasisPtr[n].color 	= RGBgreen ;
		mBasisPtr[n].refine = kRefineX + kRefineY ;
	}

	if (ReadLine(buffer,256,fp) == NULL) {
		PutMessage("Error Reading the microscope information") ;
		fprintf(stderr,"\nRead Line: %s",buffer) ;
		fclose(fp) ;
		return false ;
	}
	if (gDebugMode) fprintf(stderr,"\nRead: %s",buffer) ;
	sscanf(buffer,"%s %f %f %f %f %f %f %f",
		mTEMRecord.mMicroscope.mName,&mTEMRecord.mMicroscope.mCs,&mTEMRecord.mMicroscope.mSpreadOfFocus,&mTEMRecord.mMicroscope.mConvergence,
		&mTEMRecord.mMicroscope.m2Astigmatism,&mTEMRecord.mMicroscope.m2AstigmatismAngle,
		&mTEMRecord.mMicroscope.m3Astigmatism,&mTEMRecord.mMicroscope.m3AstigmatismAngle) ;
		
	if (ReadLine(buffer,256,fp) == NULL) {
		PutMessage("Error Reading the voltage information") ;
		fprintf(stderr,"\nRead Line: %s",buffer) ;
		fclose(fp) ;
		return false ;
	}
	if (gDebugMode) fprintf(stderr,"\nRead: %s",buffer) ;
	sscanf(buffer,"%f",&mTEMRecord.mMicroscope.mVoltage) ;

	if (ReadLine(buffer,256,fp) == NULL) {
		PutMessage("Error Reading the Lh,Lk information") ;
		fclose(fp) ;
		return false ;
	}
	if (gDebugMode) fprintf(stderr,"\nRead: %s",buffer) ;
	sscanf(buffer,"%f %f",&mTEMRecord.mMicroscope.mBeamTilt.a[0],&mTEMRecord.mMicroscope.mBeamTilt.a[1]) ;
	
	// Get the thickness data
	if (ReadLine(buffer,256,fp) == NULL) {
		PutMessage("Error Reading the thickness information") ;
		fprintf(stderr,"\nRead Line: %s",buffer) ;
		fclose(fp) ;
		return false ;
	}
	if (gDebugMode) fprintf(stderr,"\nRead: %s",buffer) ;
    if (strchr(buffer,',') != NULL) {
        sscanf(buffer,"%f,%f,%f",&mTEMRecord.mStartThickness,&mTEMRecord.mEndThickness,&mTEMRecord.mDeltaThickness) ;
		mTEMRecord.mNumThicknesses = (mTEMRecord.mEndThickness-mTEMRecord.mStartThickness)/mTEMRecord.mDeltaThickness + 1 ;
	}
    else {
        sscanf(buffer,"%f",&mTEMRecord.mStartThickness) ;
        mTEMRecord.mEndThickness = mTEMRecord.mStartThickness ;
        mTEMRecord.mDeltaThickness = 0. ;
		mTEMRecord.mNumThicknesses = 1 ;
	}
	
	if (ReadLine(plots,256,fp) == NULL) {
		PutMessage("Error Reading the pendellossung information") ;
		fprintf(stderr,"\nRead Line: %s",buffer) ;
		fclose(fp) ;
		return false ;
	}
	if (gDebugMode) fprintf(stderr,"\nRead: %s",plots) ;
	mTEMRecord.mStoreReflections = false ;
	if (plots[0] == 'y' || plots[0] == 'Y') {
		if (ReadLine(buffer,256,fp) == NULL) {
			PutMessage("Error reading the Pendellosung Option Indices") ;
			fprintf(stderr,"\nRead Line: %s",buffer) ;
			fclose(fp) ;
			throw ;
		}
		mTEMRecord.mStoreReflections = true ;
	if (gDebugMode) fprintf(stderr,"\nRead: %s",buffer) ;
		if ((numVariablesAssigned = sscanf(buffer,"%d",&mTEMRecord.mNumReflectionsToStore)) != 1) {
			PutMessage("Error reading the number of reflections to store") ;
			fclose(fp) ;
			return false ;
		}
		mTEMRecord.mReflectionsToStore = new SVector[mTEMRecord.mNumReflectionsToStore] ;
		for (k=0;k<mTEMRecord.mNumReflectionsToStore;k++) {
			if (ReadLine(buffer,256,fp) == NULL) {
				PutMessage("Error reading the Pendellosung Option Indices") ;
				fclose(fp) ;
				fprintf(stderr,"\nRead Line: %s",buffer) ;
				return false ;
			}
			if (gDebugMode) fprintf(stderr,"\nRead: %s",buffer) ;
			if ((numVariablesAssigned = sscanf(buffer,"%hd %hd %hd",&mTEMRecord.mReflectionsToStore[k].a[0],
					&mTEMRecord.mReflectionsToStore[k].a[1],&mTEMRecord.mReflectionsToStore[k].a[2])) != 3) {
				PutMessage("Error reading the Pendellosung Option Indices") ;
				fclose(fp) ;
				return false ;
			}
		}
	}
	if (ReadLine(buffer,256,fp) == NULL) {
		PutMessage("Error reading the defocus information") ;
		fprintf(stderr,"\nRead Line: %s",buffer) ;
		fclose(fp) ;
		return false ;
	}
	if (gDebugMode) fprintf(stderr,"\nRead: %s",buffer) ;
    if (strchr(buffer,',') != NULL) {
        sscanf(buffer,"%f,%f,%f",&mTEMRecord.mMicroscope.mStartFocus,&mTEMRecord.mMicroscope.mEndFocus,&mTEMRecord.mMicroscope.mDeltaFocus) ;
		mTEMRecord.mNumFocii = (mTEMRecord.mMicroscope.mEndFocus-mTEMRecord.mMicroscope.mStartFocus)/mTEMRecord.mMicroscope.mDeltaFocus ;
		if (mTEMRecord.mNumFocii < 0) mTEMRecord.mNumFocii = -mTEMRecord.mNumFocii ;
		mTEMRecord.mNumFocii++ ;
	}
    else {
        sscanf(buffer,"%f",&mTEMRecord.mMicroscope.mStartFocus) ;
   		mTEMRecord.mMicroscope.mEndFocus = mTEMRecord.mMicroscope.mStartFocus ;
        mTEMRecord.mMicroscope.mDeltaFocus = 0. ;
		mTEMRecord.mNumFocii = 1 ;
	}


	if (ReadLine(buffer,256,fp) == NULL) {
		PutMessage("Error reading microscope objective lens aperture") ;
		fprintf(stderr,"\nRead Line: %s",buffer) ;
		fclose(fp) ;
		throw ;
	}
	if (gDebugMode) fprintf(stderr,"\nRead: %s",buffer) ;
	sscanf(buffer,"%f",&mTEMRecord.mMicroscope.mAperture) ;
	
	if (ReadLine(buffer,256,fp) == NULL) {
		PutMessage("Error reading microscope information") ;
		fprintf(stderr,"\nRead Line: %s",buffer) ;
		fclose(fp) ;
		return false ;
	}
	if (gDebugMode) fprintf(stderr,"\nRead: %s",buffer) ;
    if ((numVariablesAssigned = sscanf(buffer,"%f %f %f %f",&mTEMRecord.mMicroscope.mApertureH,&mTEMRecord.mMicroscope.mApertureK,     
                            &mTEMRecord.mMicroscope.mOpticAxisH,&mTEMRecord.mMicroscope.mOpticAxisK)) != 4) {
		PutMessage("Error reading the center of the Laue circle and optic axis parameters") ;
		fprintf(stderr,"\nRead Line: %s",buffer) ;
		fclose(fp) ;
		return false ;
	}
	
	mSymOps = new TMatrix[mNumSymOps] ;

    for (i=0;i<mNumSymOps;i++) {
		if (ReadLine(buffer,256,fp) == NULL) {
    		sprintf(text,"Error reading symmetry operator #%d",i+1) ;
			PutMessage(text) ;
			fclose(fp) ;
			return false ;
		}
		if (gDebugMode) fprintf(stderr,"\nRead: %s",buffer) ;
        do {
            for (j=0,n=0;(c=buffer[n])!=EOS;n++)
               if (c!=BLANK) symop_txt[i][j++] = c ;
		} while (j==0) ;
        symop_txt[i][j] = EOS ;
		txtptr_symop[i] = &symop_txt[i][0] ;
		decode_symop(i) ;
		strcpy(mSymopTxt[i],symop_txt[i]) ;
		for(j=0; j<4 ; j++)
			for(n=0; n < 4; n++)
				mSymOps[i].m[j][n] = symop[i].matr[j][n] ;
	}
	// Apply the symmetry operators to the basis
	mAtomPtr = new TAtom[mNumBasisAtoms*200] ;	// This is the maximum we can have, so we don't need
	Assert(mAtomPtr) ;							// to reallocate this storage
	mAllocatedAtoms = mNumBasisAtoms*200 ;
	ApplySymmops() ; 

	mTEMRecord.mCalculationStatus = 1 ; // Nothing has been done
	if (ReadLine(buffer,256,fp) != NULL) {	
		sscanf(buffer,"%d",&mTEMRecord.mCalculationStatus) ;
		if (gDebugMode) fprintf(stderr,"\nRead: %s",buffer) ;
	}
	
	mTEMRecord.mMicroscope.Update() ;	// So we can set the wavelength and the interaction parameter
	
	fclose(fp) ;
	return true ;
}

Boolean	TCrystal::InitializeFromNCEMSSPrintFile(FSInfo &fsInfo) {
	// Find the dimensions of the width of the calculation from the crystal and the zone-axis
	// We will do this by reading in the files that ncemss created for this structure
	TAtom   *theAtom ;
	char	baseName[256],fileName[256] ;
	Boolean hasA=false,hasAtoms=false,hasGamma=false ;
	
	strcpy(baseName,fsInfo.mFileName) ;
		
	char *cp = strstr(baseName,".at") ;
	if (cp == nil) return false ;
	
	*cp = '\0' ;
	
	// Open the phase-grating print file
	sprintf(fileName,"%s.p_prnt",baseName) ;
	FILE *fp ;
	fp = fopen(fileName,"r") ;
	if (fp == nil) return false ;
	
	mNumBasisAtoms = mDocWindow->mCrystal->mNumUnitCellAtoms ;
	mNumUnitCellAtoms = mNumBasisAtoms ;
	mBasisPtr   = new TAtom[mNumBasisAtoms] ;
	mAtomPtr	= new TAtom[mNumUnitCellAtoms] ;
	mNumSymOps = 1 ;
	
	// Read until we get to the line containing the calculation dimensions
	char	buffer[256],name[32] ;
	long	ival,iz ;
	Boolean done = false ;
	while (!done) {
		if (ReadLine(buffer,256,fp) == nil)
			break ;
		if (!hasA) {
			cp = strstr(buffer,"A =") ;
			if (cp != nil) {
				sscanf(cp,"A = %f ,   B = %f ,   C = %f",&mA,&mB,&mC) ;
				hasA = true ;
			}
		}
		if (!hasGamma) {
			cp = strstr(buffer,"GAMMA =") ;
			if (cp != nil) {
				sscanf(cp,"GAMMA = %f Degrees",&mGamma) ;
				hasGamma = true ;
			}
		}
		if (!hasAtoms) {
			cp = strstr(buffer,"Atom positions are as listed") ;
			if (cp != nil) {
				cp = buffer ;
				ReadLine(buffer,256,fp) ; // Blank line
				// Read in all the atoms in the structure
				for(long i = 0 ; i < mNumUnitCellAtoms; i++) {
					ReadLine(buffer,256,fp) ;
					theAtom = mBasisPtr+i ;
					sscanf(buffer," %d %s %f %f %f %f %f",&ival,name,&mBasisPtr[i].xc,&mBasisPtr[i].yc,&mBasisPtr[i].zc,
						   &mBasisPtr[i].dw,&mBasisPtr[i].occ) ;
					mBasisPtr[i].type   = 1 ; // Does not matter
					mBasisPtr[i].no 	= (iz=atomno(name)) ;		// atomic number
					mBasisPtr[i].name 	= chemsymb[iz-1] ;          // chemical symbol
					mBasisPtr[i].rad 	= covrad[mBasisPtr[i].no-1]  ;
					mBasisPtr[i].color 	= RGBgreen ;
					mBasisPtr[i].refine = kRefineX + kRefineY ;
					mAtomPtr[i] = mBasisPtr[i] ;
				}
				hasAtoms = true ;
			}
		}
		if (hasA && hasGamma && hasAtoms) done = true ;
	}
	fclose(fp) ;

	strcpy(mSymopTxt[0],"x,y,z") ;
	
	return true ;
	
}

Boolean	TCrystal::InitializeFromMBFITFile(FSInfo &fsInfo) {
	FILE 	*fp ;
	char	*cpStartCrystalBlock, *cpEndCrystalBlock ;
	char	*cpStartLatticeBlock, *cpEndLatticeBlock ;
	char	*cpStartAtomBlock, *cpEndAtomBlock ;
	char	*cpStartAtomSiteBlock, *cpEndAtomSiteBlock ;
	char	*bufferStart ;
	char	name[256] ;
	char	inputLine[256],mode[256] ;
	
	fp = fopen(fsInfo.mFileName,"r") ;
	if (fp == nil) return false ;

	mFSInfo = fsInfo ;

	long i,iz,j,n ;
	char buffer[256] ;
	
	TFile theFile(fsInfo,kRead) ;
	unsigned char *fileData = (unsigned char *)theFile.Read() ;
	if (fileData == nil) {
		return false ;
	}
	
	// Process all the data that is in the buffer "fileData"
	// ReadLine will skip comment lines and comment blocks
	bufferStart = ReadLine((char *)fileData,inputLine,256) ;	
	sscanf(inputLine,"%s",mode) ;
	if (strcmp(mode,"SIM_MODE") == 0) mMBFITRecord.mMode = kMBFIT_SIM_MODE ;
	else if (strcmp(mode,"BLOCH_MODE") == 0) mMBFITRecord.mMode = kMBFIT_BLOCH_MODE ;
	else if (strcmp(mode,"DISPSURF_MODE") == 0) mMBFITRecord.mMode = kMBFIT_DISPERSION_MODE ;
	
	// 	-------- START PROCESSING CRYSTAL_STRUCTURE_DATA --------------------------------
	cpStartCrystalBlock = FindMBFITFileTag((char *)fileData,"CRYSTAL_STRUCTURE_DATA{") ;
	if (cpStartCrystalBlock == nil) return false ;

	cpEndCrystalBlock = FindMBFITFileTag(cpStartCrystalBlock,"}CRYSTAL_STRUCTURE_DATA") ;
	if (cpEndCrystalBlock == nil) return false ;		


	// 	-------- START PROCESSING LATTICE --------------------------------
	cpStartLatticeBlock = FindMBFITFileTag(cpStartCrystalBlock,"LATTICE{") ;
	if (cpStartLatticeBlock == nil) return false ;
		
	cpEndLatticeBlock = FindMBFITFileTag(cpStartLatticeBlock,"}LATTICE") ;
	if (cpEndLatticeBlock == nil) return false ;		

	bufferStart = ReadLine(cpStartLatticeBlock,buffer,256) ;
	if (bufferStart == nil) return false ;
	sscanf(buffer,"%hd",&mSpacegroup) ;

	bufferStart = ReadLine(bufferStart,buffer,256) ;
	if (bufferStart == nil) return false ;
	sscanf(buffer,"%f",&mA) ;

	bufferStart = ReadLine(bufferStart,buffer,256) ;
	if (bufferStart == nil) return false ;
	sscanf(buffer,"%f",&mB) ;
	
	bufferStart = ReadLine(bufferStart,buffer,256) ;
	if (bufferStart == nil) return false ;
	sscanf(buffer,"%f",&mC) ;
	
	bufferStart = ReadLine(bufferStart,buffer,256) ;
	if (bufferStart == nil) return false ;
	sscanf(buffer,"%f",&mAlpha) ;

	bufferStart = ReadLine(bufferStart,buffer,256) ;
	if (bufferStart == nil) return false ;
	sscanf(buffer,"%f",&mBeta) ;

	bufferStart = ReadLine(bufferStart,buffer,256) ;
	if (bufferStart == nil) return false ;
	sscanf(buffer,"%f",&mGamma) ;

	// 	-------- START PROCESSING ATOM --------------------------------
	cpStartAtomBlock = FindMBFITFileTag(cpStartLatticeBlock,"ATOM{") ;
	if (cpStartAtomBlock == nil) return false ;		
		
	cpEndAtomBlock = FindMBFITFileTag(cpStartAtomBlock,"}ATOM") ;
	if (cpEndAtomBlock == nil) return false ;		
		
	bufferStart = cpStartAtomBlock  ;	
	bufferStart = ReadLine(bufferStart,buffer,256) ;	// # Number of independent sites.
	if (bufferStart == nil) return false ;
	
	sscanf(buffer,"%d",&mNumBasisAtoms) ;

	mSymOps = nil ;
	mBasisPtr = nil ;
	mAllocatedAtoms = 0 ;
	mAtomPtr = nil ;
	
	// Allocate space for the initial basis
    mBasisPtr = new TAtom[mNumBasisAtoms] ;
	Assert(mBasisPtr) ;

	for(i = 0; i < mNumBasisAtoms; i++) {
		// 	-------- START PROCESSING SITE --------------------------------
		cpStartAtomSiteBlock = FindMBFITFileTag(cpStartAtomBlock,"SITE{") ;
		if (cpStartAtomSiteBlock == nil) return false ;
			
		cpEndAtomSiteBlock = FindMBFITFileTag(cpStartAtomSiteBlock,"}SITE") ;
		if (cpEndAtomSiteBlock == nil) return false ;		
			
		bufferStart = cpStartAtomSiteBlock  ;

		bufferStart = ReadLine(bufferStart,buffer,256) ;
		if (bufferStart == nil) return false ;
		sscanf(buffer,"%s",name) ;
		iz = atomno(name) ;
		mBasisPtr[i].no = iz ;
		mBasisPtr[i].name = chemsymb[iz-1] ;
		mBasisPtr[i].rad = covrad[iz-1] ;
		
		bufferStart = ReadLine(bufferStart,buffer,256) ;	//  # x
		if (bufferStart == nil) return false ;
		sscanf(buffer,"%f",&mBasisPtr[i].xc) ;

		bufferStart = ReadLine(bufferStart,buffer,256) ;	//  # y
		if (bufferStart == nil) return false ;
		sscanf(buffer,"%f",&mBasisPtr[i].yc) ;

		bufferStart = ReadLine(bufferStart,buffer,256) ;	//  # z
		if (bufferStart == nil) return false ;
		sscanf(buffer,"%f",&mBasisPtr[i].zc) ;

		bufferStart = ReadLine(bufferStart,buffer,256) ;	// # ISO: isotropic-B / ANISO: anisotropic-B
		if (bufferStart == nil) return false ;
		// sscanf(buffer,"%s",&mNumBasisAtoms) ;			
		
		bufferStart = ReadLine(bufferStart,buffer,256) ;	// # B [A^{2}]
		if (bufferStart == nil) return false ;
		sscanf(buffer,"%f",&mBasisPtr[i].dw) ;

		bufferStart = ReadLine(bufferStart,buffer,256) ;	// # occupancy [0.0-1.0]
		if (bufferStart == nil) return false ;
		sscanf(buffer,"%f",&mBasisPtr[i].occ) ;

		bufferStart = ReadLine(bufferStart,buffer,256) ;	// # f table (0:Mott+BK, 1:BK+BK)
		if (bufferStart == nil) return false ;
		// sscanf(buffer,"%d",&f) ;		

		cpStartAtomBlock = cpEndAtomSiteBlock ;
	}		

	// Find the number low order reflections;
	bufferStart = FindMBFITFileTag((char *)fileData,"LOW_ORDER_FG{") ;
	bufferStart = ReadLine(bufferStart,inputLine,256) ;	// # low order reflections
	sscanf(inputLine,"%hd",&mMBFITRecord.mNumLowOrderG) ;
	if (mMBFITRecord.mNumLowOrderG > 0) {
		mMBFITRecord.mLowOrderG = new LVector[mMBFITRecord.mNumLowOrderG ] ;
		for(i = 0 ; i < mMBFITRecord.mNumLowOrderG; i++) {
			bufferStart = ReadLine(bufferStart,inputLine,256) ;	// hkls
			sscanf(inputLine,"%d %d %d",&mMBFITRecord.mLowOrderG[i].a[0],&mMBFITRecord.mLowOrderG[i].a[1],&mMBFITRecord.mLowOrderG[i].a[2]) ;
		}
	}
	
	// Set the incident beam direction
	bufferStart = FindMBFITFileTag((char *)fileData,"INCIDENT_BEAM{") ;
	if (bufferStart == nil) return false ;
	bufferStart = ReadLine(bufferStart,inputLine,256) ;	// Incident beam
	sscanf(inputLine,"%d %d %d",&mMBFITRecord.mZoneAxis.a[0],&mMBFITRecord.mZoneAxis.a[1],&mMBFITRecord.mZoneAxis.a[2]) ;
	bufferStart = ReadLine(bufferStart,inputLine,256) ;	// surface normal
	sscanf(inputLine,"%d %d %d",&mMBFITRecord.mSurfaceNormal.a[0],&mMBFITRecord.mSurfaceNormal.a[1],&mMBFITRecord.mSurfaceNormal.a[2]) ;
	bufferStart = ReadLine(bufferStart,inputLine,256) ;	// ZOLZ base vector
	sscanf(inputLine,"%d %d %d",&mMBFITRecord.mZoneBaseVector.a[0],&mMBFITRecord.mZoneBaseVector.a[1],&mMBFITRecord.mZoneBaseVector.a[2]) ;
	bufferStart = ReadLine(bufferStart,inputLine,256) ;	// H tilt
	sscanf(inputLine,"%f",&mMBFITRecord.mTiltH) ;
	bufferStart = ReadLine(bufferStart,inputLine,256) ;	// G tilt
	sscanf(inputLine,"%f",&mMBFITRecord.mTiltG) ;
	bufferStart = ReadLine(bufferStart,inputLine,256) ;	// Accel voltage
	sscanf(inputLine,"%f",&mMBFITRecord.mMicroscope.mVoltage) ;

	// Set the calculation range
	bufferStart = FindMBFITFileTag((char *)fileData,"CALC_KXY_RANGE{") ;
	if (bufferStart == nil) return false ;
	bufferStart = ReadLine(bufferStart,inputLine,256) ;	// dummy parameter
	bufferStart = ReadLine(bufferStart,inputLine,256) ;	// Calc. type
	sscanf(inputLine,"%d",&mMBFITRecord.mCalcType) ;
	bufferStart = ReadLine(bufferStart,inputLine,256) ;	// numh: number of calc. points
	sscanf(inputLine,"%hd",&mMBFITRecord.mNumCalculationPoints) ;
	bufferStart = ReadLine(bufferStart,inputLine,256) ;	// hbgn, hend
	sscanf(inputLine,"%f %f",&mMBFITRecord.mMinH,&mMBFITRecord.mMaxH) ;
	bufferStart = ReadLine(bufferStart,inputLine,256) ;	// gbgn, gend
	sscanf(inputLine,"%f %f",&mMBFITRecord.mMinG,&mMBFITRecord.mMaxG) ;
	
	// Set the Beam search method
	bufferStart = FindMBFITFileTag((char *)fileData,"BEAM_SEARCH{") ;
	if (bufferStart == nil) return false ;
	bufferStart = ReadLine(bufferStart,inputLine,256) ;	// beam search method
	sscanf(inputLine,"%d",&mMBFITRecord.mBeamSearchType) ;
	if (mMBFITRecord.mBeamSearchType == kMBFIT_ManualBeamSearch) {
		bufferStart = ReadLine(bufferStart,inputLine,256) ;	// num. of exact, num. of bethe
		sscanf(inputLine,"%hd  %hd",&mMBFITRecord.mBeamSearchManual.num_exact,&mMBFITRecord.mBeamSearchManual.num_bethe) ;
		for(i = 0 ; i < mMBFITRecord.mBeamSearchManual.num_exact; i++) {
			bufferStart = ReadLine(bufferStart,inputLine,256) ;	// num. of exact, num. of bethe
			sscanf(inputLine,"%d %d	%d",
				&mMBFITRecord.mBeamSearchManual.hklExact[i].a[0],
				&mMBFITRecord.mBeamSearchManual.hklExact[i].a[1],
				&mMBFITRecord.mBeamSearchManual.hklExact[i].a[2]) ;
		}
		for(i = 0 ; i < mMBFITRecord.mBeamSearchManual.num_bethe; i++) {
			bufferStart = ReadLine(bufferStart,inputLine,256) ;	// num. of exact, num. of bethe
			sscanf(inputLine,"%d %d	%d",
				&mMBFITRecord.mBeamSearchManual.hklBethe[i].a[0],
				&mMBFITRecord.mBeamSearchManual.hklBethe[i].a[1],
				&mMBFITRecord.mBeamSearchManual.hklBethe[i].a[2]) ;
		}
	}
	else if (mMBFITRecord.mBeamSearchType == kMBFIT_AutoBeamSearch_WG) {
		// Don't know what to do here yet
	}
	else if (mMBFITRecord.mBeamSearchType == kMBFIT_AutoBeamSearch_SG) {
		bufferStart = ReadLine(bufferStart,inputLine,256) ;	
		sscanf(inputLine,"%hd %f %hd %hd",
			&mMBFITRecord.mBeamSearchAuto.num_g_limit,&mMBFITRecord.mBeamSearchAuto.g_max,
			&mMBFITRecord.mBeamSearchAuto.lz_min,&mMBFITRecord.mBeamSearchAuto.lz_max) ;
		bufferStart = ReadLine(bufferStart,inputLine,256) ;	
		sscanf(inputLine,"%f %f %f",
			&mMBFITRecord.mBeamSearchAuto.sg_exact_max,&mMBFITRecord.mBeamSearchAuto.sg_bethe_max,
			&mMBFITRecord.mBeamSearchAuto.xi_g_max) ;
	}

	switch(mMBFITRecord.mMode ) {
		case kMBFIT_SIM_MODE:
			// Set the Output reflections
			bufferStart = FindMBFITFileTag((char *)fileData,"OUTPUT_REFLEX{") ;
			if (bufferStart == nil) return false ;
			bufferStart = ReadLine(bufferStart,inputLine,256) ;	// Number of reflections to be calculated (0:all)
			sscanf(inputLine,"%hd",&mMBFITRecord.mNumReflectionsToOutput) ;
			if (mMBFITRecord.mNumReflectionsToOutput > 0) {
				mMBFITRecord.mOutputReflection = new LVector[mMBFITRecord.mNumReflectionsToOutput] ;
				for(i = 0; i < mMBFITRecord.mNumReflectionsToOutput; i++) {
					bufferStart = ReadLine(bufferStart,inputLine,256) ;	// hkls
					sscanf(inputLine,"%d %d	%d",
						&mMBFITRecord.mOutputReflection[i].a[0],
						&mMBFITRecord.mOutputReflection[i].a[1],
						&mMBFITRecord.mOutputReflection[i].a[2]) ;
				}
			}

			// Set the Output thicknesses
			bufferStart = FindMBFITFileTag((char *)fileData,"OUTPUT_THICKNESS{") ;
			if (bufferStart == nil) return false ;
			bufferStart = ReadLine(bufferStart,inputLine,256) ;	// Number of thicknesses
			sscanf(inputLine,"%hd",&mMBFITRecord.mNumThicknesses) ;
			if (mMBFITRecord.mNumThicknesses > 0) {
				mMBFITRecord.mThicknesses = new float[mMBFITRecord.mNumThicknesses] ;
				for(i = 0; i < mMBFITRecord.mNumThicknesses; i++) {
					bufferStart = ReadLine(bufferStart,inputLine,256) ;	// thickness
					sscanf(inputLine,"%f",&mMBFITRecord.mThicknesses[i]) ;
				}
			}
			break ;
		case kMBFIT_DISPERSION_MODE:
			// Set the Output Branch
			bufferStart = FindMBFITFileTag((char *)fileData,"OUTPUT_BRANCH{") ;
			if (bufferStart == nil) return false ;
			bufferStart = ReadLine(bufferStart,inputLine,256) ;	// # Number of branches to be calculated (begin, end)
			sscanf(inputLine,"%hd %hd",&mMBFITRecord.branchBegin,&mMBFITRecord.branchEnd) ;
			
			// Set the Output reflections
			bufferStart = FindMBFITFileTag((char *)fileData,"OUTPUT_REFLEX{") ;
			if (bufferStart == nil) return false ;
			bufferStart = ReadLine(bufferStart,inputLine,256) ;	// Number of reflections to be calculated (0:all)
			sscanf(inputLine,"%hd",&mMBFITRecord.mNumReflectionsToOutput) ;
			if (mMBFITRecord.mNumReflectionsToOutput > 0) {
				mMBFITRecord.mOutputReflection = new LVector[mMBFITRecord.mNumReflectionsToOutput] ;
				for(i = 0; i < mMBFITRecord.mNumReflectionsToOutput; i++) {
					bufferStart = ReadLine(bufferStart,inputLine,256) ;	// hkls
					sscanf(inputLine,"%d %d	%d",
						&mMBFITRecord.mOutputReflection[i].a[0],
						&mMBFITRecord.mOutputReflection[i].a[1],
						&mMBFITRecord.mOutputReflection[i].a[2]) ;
				}
			}

			// Set the Output thicknesses
			bufferStart = FindMBFITFileTag((char *)fileData,"OUTPUT_THICKNESS{") ;
			if (bufferStart == nil) return false ;
			bufferStart = ReadLine(bufferStart,inputLine,256) ;	// Number of thicknesses
			sscanf(inputLine,"%hd",&mMBFITRecord.mNumThicknesses) ;
			if (mMBFITRecord.mNumThicknesses > 0) {
				mMBFITRecord.mThicknesses = new float[mMBFITRecord.mNumThicknesses] ;
				for(i = 0; i < mMBFITRecord.mNumThicknesses; i++) {
					bufferStart = ReadLine(bufferStart,inputLine,256) ;	// thickness
					sscanf(inputLine,"%f",&mMBFITRecord.mThicknesses[i]) ;
				}
			}
			break ;
		case kMBFIT_BLOCH_MODE:
			// Set the XY calculation range
			bufferStart = FindMBFITFileTag((char *)fileData,"CALC_RXY_RANGE{") ;
			if (bufferStart == nil) return false ;
			bufferStart = ReadLine(bufferStart,inputLine,256) ;	// numh: number of calc. points along x and y
			sscanf(inputLine,"%hd %hd",&mMBFITRecord.mNumCalculationPointsX,&mMBFITRecord.mNumCalculationPointsY) ;
			bufferStart = ReadLine(bufferStart,inputLine,256) ;	// hbgn, hend
			sscanf(inputLine,"%f %f",&mMBFITRecord.mMinX,&mMBFITRecord.mMaxX) ;
			bufferStart = ReadLine(bufferStart,inputLine,256) ;	// gbgn, gend
			sscanf(inputLine,"%f %f",&mMBFITRecord.mMinY,&mMBFITRecord.mMaxY) ;
			// Set the Output Branch
			bufferStart = FindMBFITFileTag((char *)fileData,"OUTPUT_BRANCH{") ;
			if (bufferStart == nil) return false ;
			bufferStart = ReadLine(bufferStart,inputLine,256) ;	// # Number of branches to be calculated (begin, end)
			sscanf(inputLine,"%hd %hd",&mMBFITRecord.branchBegin,&mMBFITRecord.branchEnd) ;
			break ;
	}
	
	mXViewCells		= 1 ;
	mXViewCells		= 1 ;
	mXViewCells		= 1 ;
	mShiftX = 0. ;
	mShiftY = 0. ;
	

	mSymOps = new TMatrix[MAXSYMOPS] ;	// We don't need to keep on allocating symmetry operators
    assm_spcgrp(mSpacegroup) ;
    mNumSymOps = gNumSymops ;

	for(i=0; i < mNumSymOps; i++) {
		decode_symop(i) ;
		strcpy(mSymopTxt[i],symop_txt[i]) ;
		for(j=0; j<4 ; j++)
			for(n=0; n < 4; n++)
				mSymOps[i].m[j][n] = symop[i].matr[j][n] ;
	}
			
	DegreeToRad(&mAlpha) ;
    DegreeToRad(&mBeta) ;
    DegreeToRad(&mGamma) ;
    
	// Apply the symmetry operators to the basis
	mAtomPtr = new TAtom[mNumBasisAtoms*MAXSYMOPS] ;	// This is the maximum we can have, so we don't need
	Assert(mAtomPtr) ;							// to reallocate this storage
	mAllocatedAtoms = mNumBasisAtoms*MAXSYMOPS ;
	ApplySymmops() ; 
    
	return true ;

}

void	TCrystal::WriteMBFITControlFile() {
	
	FILE	*fp ;
	long	i,j ;
	TAtom   *thisAtom ;
	char	element[4] ;
	
	float   alpha,beta,gamma ;
	alpha = mAlpha ;
	beta = mBeta ;
	gamma = mGamma ;
	RadToDegree(&alpha) ;
	RadToDegree(&beta) ;
	RadToDegree(&gamma) ;
	
	if ((fp = fopen(mFSInfo.mFileName,"w")) == NULL) {
		PutMessage("Could not update the MBFIT file") ;
		return ;
	}
	fprintf(fp,"#%s\n",mFSInfo.mFileName) ;
	switch(mMBFITRecord.mMode) {
		case kMBFIT_SIM_MODE:
			fprintf(fp,"# mbfit: a sample data for MBFIT simulation\n") ;
			fprintf(fp,"# Comment lines:  # or \% or COMMENT{  }COMMENT\n") ;
			fprintf(fp,"# ----------------------------------------------------------------------\n") ;
			fprintf(fp,"# job type\n") ;
			fprintf(fp,"# SIM_MODE: MBFIT simulations\n") ;
			fprintf(fp,"# ----------\n") ;
			fprintf(fp,"SIM_MODE\n") ;
			break ;
		case kMBFIT_DISPERSION_MODE:
			fprintf(fp,"# mbfit: a sample data for DISPERSION SURFACE calculations\n") ;
			fprintf(fp,"# Comment lines:  # or \% or COMMENT{  }COMMENT\n") ;
			fprintf(fp,"# ----------------------------------------------------------------------\n") ;
			fprintf(fp,"# job type\n") ;
			fprintf(fp,"#  DISPSURF_MODE: dispersion surface\n") ;
			fprintf(fp,"# ----------\n") ;
			fprintf(fp,"DISPSURF_MODE\n") ;
			break ;
		case kMBFIT_BLOCH_MODE:
			fprintf(fp,"# mbfit: a sample data for BLOCH WAVE simulation\n") ;
			fprintf(fp,"# Comment lines:  # or \% or COMMENT{  }COMMENT\n") ;
			fprintf(fp,"# ----------------------------------------------------------------------\n") ;
			fprintf(fp,"# job type\n") ;
			fprintf(fp,"# BLOCH_MODE: BLOCH WAVE simulations\n") ;
			fprintf(fp,"# ----------\n") ;
			fprintf(fp,"BLOCH_MODE\n") ;
			break ;
	}
	fprintf(fp,"# ----------------------------------------------------------------------\n") ;
	fprintf(fp,"# For changing internal control parameters\n") ;
	fprintf(fp,"EPS_CONTROL_PARAMETERS{\n") ;
	
	// Put in internal control flags right here
	
	fprintf(fp,"}EPS_CONTROL_PARAMETERS\n") ;
	fprintf(fp,"# ----------------------------------------------------------------------\n") ;

	// Start Crystal Structure Block
	fprintf(fp,"# Crystal structural data block\n") ;
	fprintf(fp,"CRYSTAL_STRUCTURE_DATA{\n") ;
	fprintf(fp,"#\n") ;
	// Start Crystal lattice Block
	fprintf(fp,"LATTICE{\n") ;
	fprintf(fp,"%d     # Number of space group\n",mSpacegroup) ;
	fprintf(fp,"# lattice parameter [in Angstrom units]\n") ;
	fprintf(fp,"%.4f    # a\n",mA) ;
	fprintf(fp,"%.4f    # b\n",mB) ;
	fprintf(fp,"%.4f    # c\n",mC) ;
	fprintf(fp,"%.2f    # alpha\n",alpha) ;
	fprintf(fp,"%.2f    # beta\n",beta) ;
	fprintf(fp,"%.2f    # gamma\n",gamma) ;
	fprintf(fp,"}LATTICE\n") ;
	// End Crystal lattice Block

	// Start ATOM List block
	fprintf(fp,"#\n") ;
	fprintf(fp,"ATOM{\n") ;
	fprintf(fp,"%d  # Number of independent sites\n",mNumBasisAtoms) ;
	fprintf(fp,"##\n") ;
	for (j=0;j<mNumBasisAtoms;j++) {
		thisAtom = mBasisPtr+j ;
		// Start SITE List block
		fprintf(fp,"SITE{\n") ;
		for(i=0; i < strlen(thisAtom->name); i++) element[i] = toupper(thisAtom->name[i]) ; element[strlen(thisAtom->name)] = '\0' ;
		fprintf(fp,"%s\n",element) ;
		fprintf(fp,"%f    # x (fractional coordinate)\n",thisAtom->xc) ;
		fprintf(fp,"%f    # y\n",thisAtom->yc) ;
		fprintf(fp,"%f    # z\n",thisAtom->zc) ;
		fprintf(fp,"ISO    # ISO: isotropic-B / ANISO: anisotropic-B\n") ;
		fprintf(fp,"%f    # B [A^{2}]\n",thisAtom->dw) ;
		fprintf(fp,"%f    # occupancy [0.0-1.0]\n",thisAtom->occ) ;
		fprintf(fp,"0      # f table (0:Mott+BK, 1:BK+BK)\n") ;
		fprintf(fp,"}SITE\n") ;
		// End SITE List block
	}
	fprintf(fp,"}ATOM\n") ;
	// END ATOM List block
	fprintf(fp,"##\n") ;
	// START LOW_ORDER_FG block
	fprintf(fp,"LOW_ORDER_FG{\n") ;
		fprintf(fp,"0    # number of low-order g\n") ;
		// Below follows a list of reflections 
	fprintf(fp,"}LOW_ORDER_FG\n") ;
	// End LOW_ORDER_FG block
	fprintf(fp,"}CRYSTAL_STRUCTURE_DATA\n") ;
	// End CRYSTAL_STRUCTURE_DATA block

	fprintf(fp,"# ----------------------------------------------------------------------\n") ;
	fprintf(fp,"# ----------------------------------------------------------------------\n") ;
	fprintf(fp,"# Simulation: calculation condition block\n") ;
	switch(mMBFITRecord.mMode) {
		case kMBFIT_SIM_MODE:
			// START SIM_CALCULATION_CONDITION block
			fprintf(fp,"SIM_CALCULATION_CONDITION{\n") ;
			fprintf(fp,"#\n") ;
			// START INCIDENT_BEAM block
			fprintf(fp,"INCIDENT_BEAM{\n") ;
			fprintf(fp,"%d %d %d    # Incidence	(upward)\n",
				mMBFITRecord.mZoneAxis.a[0],mMBFITRecord.mZoneAxis.a[1],mMBFITRecord.mZoneAxis.a[2]) ;
			fprintf(fp,"%d %d %d    # Surface normal (upward)\n",
				mMBFITRecord.mSurfaceNormal.a[0],mMBFITRecord.mSurfaceNormal.a[1],mMBFITRecord.mSurfaceNormal.a[2]) ;
			fprintf(fp,"%d %d %d    # ZOLZ base vector\n",
				mMBFITRecord.mZoneBaseVector.a[0],mMBFITRecord.mZoneBaseVector.a[1],mMBFITRecord.mZoneBaseVector.a[2]) ;
			fprintf(fp,"%f          # H tilt\n",mMBFITRecord.mTiltH) ;
			fprintf(fp,"%f          # G tilt\n",mMBFITRecord.mTiltG) ;
			fprintf(fp,"%.2f        # Accelerating Voltage [kV]\n",mMBFITRecord.mMicroscope.mVoltage) ;
			fprintf(fp,"}INCIDENT_BEAM\n") ;
			// End INCIDENT_BEAM block
			fprintf(fp,"#\n") ;
			// START CALC_KXY_RANGE block
			fprintf(fp,"CALC_KXY_RANGE{\n") ;
			fprintf(fp,"0     # cal_range_flag  [dummy: always 0]\n") ;
			fprintf(fp,"%d     # Calc. type [0:0D, 1:1D, 2:2D disk, 3:2D square]\n",mMBFITRecord.mCalcType) ;
			fprintf(fp,"%d   # numh: number of calc. points  > 5  (total~sq(numh))\n",mMBFITRecord.mNumCalculationPoints) ;
			fprintf(fp,"%.2f %.2f    # hbgn, hend\n",mMBFITRecord.mMinH,mMBFITRecord.mMaxH) ;
			fprintf(fp,"%.2f %.2f    # gbgn, gend\n",mMBFITRecord.mMinG,mMBFITRecord.mMaxG) ;
			fprintf(fp,"}CALC_KXY_RANGE\n") ;
			// End CALC_KXY_RANGE block
			fprintf(fp,"#\n") ;
			fprintf(fp,"#\n") ;
			// START BEAM_SEARCH block
			if (mMBFITRecord.mBeamSearchType == kMBFIT_ManualBeamSearch) {
				fprintf(fp,"BEAM_SEARCH{\n") ;
				fprintf(fp,"# Manual beam input\n") ;
				fprintf(fp,"0     # Beam search method [0:manual, 1:auto[wg], 2:[sg]]\n") ;
				fprintf(fp,"%d %d  # num. of exact, num. of bethe\n",
					mMBFITRecord.mBeamSearchManual.num_exact,mMBFITRecord.mBeamSearchManual.num_bethe) ;
				for(i = 0 ; i < mMBFITRecord.mBeamSearchManual.num_exact; i++) {
					fprintf(fp,"%d %d %d\n",
						mMBFITRecord.mBeamSearchManual.hklExact[i].a[0],
						mMBFITRecord.mBeamSearchManual.hklExact[i].a[1],
						mMBFITRecord.mBeamSearchManual.hklExact[i].a[2]) ;
				}
				for(i = 0 ; i < mMBFITRecord.mBeamSearchManual.num_bethe; i++) {
					fprintf(fp,"%d %d %d\n",
						mMBFITRecord.mBeamSearchManual.hklBethe[i].a[0],
						mMBFITRecord.mBeamSearchManual.hklBethe[i].a[1],
						mMBFITRecord.mBeamSearchManual.hklBethe[i].a[2]) ;
				}
				fprintf(fp,"}BEAM_SEARCH\n") ;
				fprintf(fp,"#\n") ;
				fprintf(fp,"COMMENT{  --------------------------------------------------\n") ;
				fprintf(fp,"BEAM_SEARCH{\n") ;
				fprintf(fp,"# Automatic SG beam search\n") ;
				fprintf(fp,"2                  # Beam search [0:manual, 1:automatic-wg, 2:automatic-sg]\n") ;
				fprintf(fp,"200 4.0 0 0     # num_g_limit, g_max[A^{-1}], lz_min, lz_max\n") ;
				fprintf(fp,"0.02 0.03 1e+11  # sg_exact_max,  sg_bethe_max,  xi_g_max\n") ;
				fprintf(fp,"}BEAM_SEARCH\n") ;
				fprintf(fp,"}COMMENT  --------------------------------------------------\n") ;
				fprintf(fp,"#\n") ;
			}
			else if (mMBFITRecord.mBeamSearchType == kMBFIT_AutoBeamSearch_WG) {
				fprintf(fp,"BEAM_SEARCH{\n") ;
				fprintf(fp,"# Automatic WG beam search\n") ;
				fprintf(fp,"1                  # Beam search [0:manual, 1:automatic-wg, 2:automatic-sg]\n") ;
				fprintf(fp,"%d %.4f %d %d     # num_g_limit, g_max[A^{-1}], lz_min, lz_max\n",
					mMBFITRecord.mBeamSearchAuto.num_g_limit,mMBFITRecord.mBeamSearchAuto.g_max,
					mMBFITRecord.mBeamSearchAuto.lz_min,mMBFITRecord.mBeamSearchAuto.lz_max) ;
				fprintf(fp,"%.4f %.4f %e  # sg_exact_max,  sg_bethe_max,  xi_g_max\n",
					mMBFITRecord.mBeamSearchAuto.sg_exact_max,mMBFITRecord.mBeamSearchAuto.sg_bethe_max,
					mMBFITRecord.mBeamSearchAuto.xi_g_max) ;
				fprintf(fp,"}BEAM_SEARCH\n") ;
				fprintf(fp,"#\n") ;
			}
			else if (mMBFITRecord.mBeamSearchType == kMBFIT_AutoBeamSearch_SG) {
				fprintf(fp,"BEAM_SEARCH{\n") ;
				fprintf(fp,"# Automatic SG beam search\n") ;
				fprintf(fp,"2                  # Beam search [0:manual, 1:automatic-wg, 2:automatic-sg]\n") ;
				fprintf(fp,"%d %.4f %d %d     # num_g_limit, g_max[A^{-1}], lz_min, lz_max\n",
					mMBFITRecord.mBeamSearchAuto.num_g_limit,mMBFITRecord.mBeamSearchAuto.g_max,
					mMBFITRecord.mBeamSearchAuto.lz_min,mMBFITRecord.mBeamSearchAuto.lz_max) ;
				fprintf(fp,"%.4f %.4f %e  # sg_exact_max,  sg_bethe_max,  xi_g_max\n",
					mMBFITRecord.mBeamSearchAuto.sg_exact_max,mMBFITRecord.mBeamSearchAuto.sg_bethe_max,
					mMBFITRecord.mBeamSearchAuto.xi_g_max) ;
				fprintf(fp,"}BEAM_SEARCH\n") ;
				fprintf(fp,"#\n") ;
			}
			// End BEAM_SEARCH block
			// START OUTPUT_REFLEX block  
			fprintf(fp,"OUTPUT_REFLEX{\n") ;
			fprintf(fp,"%d  # Number of reflections to be calculated (0:all)\n",mMBFITRecord.mNumReflectionsToOutput) ;
			fprintf(fp,"# indices hkl\n") ;
			for(i=0 ; i < mMBFITRecord.mNumReflectionsToOutput; i++) {
				fprintf(fp,"%d %d %d\n",
					mMBFITRecord.mOutputReflection[i].a[0],
					mMBFITRecord.mOutputReflection[i].a[1],
					mMBFITRecord.mOutputReflection[i].a[2]) ;
			}
			fprintf(fp,"#\n") ;
			fprintf(fp,"}OUTPUT_REFLEX\n") ;
			// End OUTPUT_REFLEX block
			// START OUTPUT_THICKNESS block  
			fprintf(fp,"OUTPUT_THICKNESS{\n") ;
			fprintf(fp,"# Number of thicknesses for output\n") ;
			fprintf(fp,"%d\n",mMBFITRecord.mNumThicknesses) ;
			fprintf(fp,"# thicknesses (in Angstrom units)\n") ;
			for(i=0 ; i < mMBFITRecord.mNumThicknesses; i++) {
				fprintf(fp,"%.0f\n",mMBFITRecord.mThicknesses[i]) ;
			}
			fprintf(fp,"}OUTPUT_THICKNESS\n") ;
			// End OUTPUT_THICKNESS block
			fprintf(fp,"}SIM_CALCULATION_CONDITION\n") ;
			// End SIM_CALCULATION_CONDITION block
			break ;
		case kMBFIT_DISPERSION_MODE:
			// START DISPSURF_CALCULATION_CONDITION block
			fprintf(fp,"DISPSURF_CALCULATION_CONDITION{\n") ;
			fprintf(fp,"#\n") ;
			// START INCIDENT_BEAM block
			fprintf(fp,"INCIDENT_BEAM{\n") ;
			fprintf(fp,"%d %d %d    # Incidence	(upward)\n",
				mMBFITRecord.mZoneAxis.a[0],mMBFITRecord.mZoneAxis.a[1],mMBFITRecord.mZoneAxis.a[2]) ;
			fprintf(fp,"%d %d %d    # Surface normal (upward)\n",
				mMBFITRecord.mSurfaceNormal.a[0],mMBFITRecord.mSurfaceNormal.a[1],mMBFITRecord.mSurfaceNormal.a[2]) ;
			fprintf(fp,"%d %d %d    # ZOLZ base vector\n",
				mMBFITRecord.mZoneBaseVector.a[0],mMBFITRecord.mZoneBaseVector.a[1],mMBFITRecord.mZoneBaseVector.a[2]) ;
			fprintf(fp,"%f          # H tilt\n",mMBFITRecord.mTiltH) ;
			fprintf(fp,"%f          # G tilt\n",mMBFITRecord.mTiltG) ;
			fprintf(fp,"%.2f        # Accelerating Voltage [kV]\n",mMBFITRecord.mMicroscope.mVoltage) ;
			fprintf(fp,"}INCIDENT_BEAM\n") ;
			// End INCIDENT_BEAM block
			fprintf(fp,"#\n") ;
			// START CALC_KXY_RANGE block
			fprintf(fp,"CALC_KXY_RANGE{\n") ;
				fprintf(fp,"0     # cal_range_flag  [dummy: always 0]\n") ;
				fprintf(fp,"1     # Calc. type [should be 1 for DISPSURF_MODE]\n") ;
				fprintf(fp,"%d   # numh: number of calc. points  > 5  (total~sq(numh))\n",mMBFITRecord.mNumCalculationPoints) ;
				fprintf(fp,"%.2f %.2f    # hbgn, hend\n",mMBFITRecord.mMinH,mMBFITRecord.mMaxH) ;
				fprintf(fp,"%.2f %.2f    # gbgn, gend\n",mMBFITRecord.mMinG,mMBFITRecord.mMaxG) ;
			fprintf(fp,"}CALC_KXY_RANGE\n") ;
			// End CALC_KXY_RANGE block
			fprintf(fp,"#\n") ;
			fprintf(fp,"#\n") ;
			// START BEAM_SEARCH block
			if (mMBFITRecord.mBeamSearchType == kMBFIT_ManualBeamSearch) {
				fprintf(fp,"BEAM_SEARCH{\n") ;
				fprintf(fp,"# Manual beam input\n") ;
				fprintf(fp,"0     # Beam search method [0:manual, 1:auto[wg], 2:[sg]]\n") ;
				fprintf(fp,"%d %d  # num. of exact, num. of bethe\n",
					mMBFITRecord.mBeamSearchManual.num_exact,mMBFITRecord.mBeamSearchManual.num_bethe) ;
				for(i = 0 ; i < mMBFITRecord.mBeamSearchManual.num_exact; i++) {
					fprintf(fp,"%d %d %d\n",
						mMBFITRecord.mBeamSearchManual.hklExact[i].a[0],
						mMBFITRecord.mBeamSearchManual.hklExact[i].a[1],
						mMBFITRecord.mBeamSearchManual.hklExact[i].a[2]) ;
				}
				for(i = 0 ; i < mMBFITRecord.mBeamSearchManual.num_bethe; i++) {
					fprintf(fp,"%d %d %d\n",
						mMBFITRecord.mBeamSearchManual.hklBethe[i].a[0],
						mMBFITRecord.mBeamSearchManual.hklBethe[i].a[1],
						mMBFITRecord.mBeamSearchManual.hklBethe[i].a[2]) ;
				}
				fprintf(fp,"}BEAM_SEARCH\n") ;
				fprintf(fp,"#\n") ;
				fprintf(fp,"COMMENT{  --------------------------------------------------\n") ;
				fprintf(fp,"BEAM_SEARCH{\n") ;
				fprintf(fp,"# Automatic SG beam search\n") ;
				fprintf(fp,"2                  # Beam search [0:manual, 1:automatic-wg, 2:automatic-sg]\n") ;
				fprintf(fp,"200 4.0 0 0     # num_g_limit, g_max[A^{-1}], lz_min, lz_max\n") ;
				fprintf(fp,"0.02 0.03 1e+11  # sg_exact_max,  sg_bethe_max,  xi_g_max\n") ;
				fprintf(fp,"}BEAM_SEARCH\n") ;
				fprintf(fp,"}COMMENT  --------------------------------------------------\n") ;
				fprintf(fp,"#\n") ;
			}
			else if (mMBFITRecord.mBeamSearchType == kMBFIT_AutoBeamSearch_WG) {
				fprintf(fp,"BEAM_SEARCH{\n") ;
				fprintf(fp,"# Automatic WG beam search\n") ;
				fprintf(fp,"1                  # Beam search [0:manual, 1:automatic-wg, 2:automatic-sg]\n") ;
				fprintf(fp,"%d %.4f %d %d     # num_g_limit, g_max[A^{-1}], lz_min, lz_max\n",
					mMBFITRecord.mBeamSearchAuto.num_g_limit,mMBFITRecord.mBeamSearchAuto.g_max,
					mMBFITRecord.mBeamSearchAuto.lz_min,mMBFITRecord.mBeamSearchAuto.lz_max) ;
				fprintf(fp,"%.4f %.4f %e  # sg_exact_max,  sg_bethe_max,  xi_g_max\n",
					mMBFITRecord.mBeamSearchAuto.sg_exact_max,mMBFITRecord.mBeamSearchAuto.sg_bethe_max,
					mMBFITRecord.mBeamSearchAuto.xi_g_max) ;
				fprintf(fp,"}BEAM_SEARCH\n") ;
				fprintf(fp,"#\n") ;
			}
			else if (mMBFITRecord.mBeamSearchType == kMBFIT_AutoBeamSearch_SG) {
				fprintf(fp,"BEAM_SEARCH{\n") ;
				fprintf(fp,"# Automatic SG beam search\n") ;
				fprintf(fp,"2                  # Beam search [0:manual, 1:automatic-wg, 2:automatic-sg]\n") ;
				fprintf(fp,"%d %.4f %d %d     # num_g_limit, g_max[A^{-1}], lz_min, lz_max\n",
					mMBFITRecord.mBeamSearchAuto.num_g_limit,mMBFITRecord.mBeamSearchAuto.g_max,
					mMBFITRecord.mBeamSearchAuto.lz_min,mMBFITRecord.mBeamSearchAuto.lz_max) ;
				fprintf(fp,"%.4f %.4f %e  # sg_exact_max,  sg_bethe_max,  xi_g_max\n",
					mMBFITRecord.mBeamSearchAuto.sg_exact_max,mMBFITRecord.mBeamSearchAuto.sg_bethe_max,
					mMBFITRecord.mBeamSearchAuto.xi_g_max) ;
				fprintf(fp,"}BEAM_SEARCH\n") ;
				fprintf(fp,"#\n") ;
			}
			
			// START OUTPUT_BRANCH block  
			fprintf(fp,"OUTPUT_BRANCH{\n") ;
			fprintf(fp,"%d %d    # Number of branches to be calculated (begin, end)\n",mMBFITRecord.branchBegin , mMBFITRecord.branchEnd) ;
			fprintf(fp,"}OUTPUT_BRANCH\n") ;
			fprintf(fp,"#\n") ;
			// End OUTPUT_BRANCH block

			// START OUTPUT_REFLEX block  
			fprintf(fp,"OUTPUT_REFLEX{\n") ;
				fprintf(fp,"%d  # Number of reflections (0 means no output of epsCg and intensities)\n",mMBFITRecord.mNumReflectionsToOutput) ;
				fprintf(fp,"# indices hkl\n") ;
				for(i=0 ; i < mMBFITRecord.mNumReflectionsToOutput; i++) {
					fprintf(fp,"%d %d %d\n",
						mMBFITRecord.mOutputReflection[i].a[0],
						mMBFITRecord.mOutputReflection[i].a[1],
						mMBFITRecord.mOutputReflection[i].a[2]) ;
				}
			fprintf(fp,"}OUTPUT_REFLEX\n") ;
			fprintf(fp,"#\n") ;
			// End OUTPUT_REFLEX block
			// START OUTPUT_THICKNESS block  
			fprintf(fp,"OUTPUT_THICKNESS{\n") ;
				fprintf(fp,"# Number of thicknesses for output \n") ;
				fprintf(fp,"%d #(0: no intensity output)\n",mMBFITRecord.mNumThicknesses) ;
				fprintf(fp,"# thicknesses (in Angstrom units)\n") ;
				for(i=0 ; i < mMBFITRecord.mNumThicknesses; i++) {
					fprintf(fp,"%.0f\n",mMBFITRecord.mThicknesses[i]) ;
				}
			fprintf(fp,"}OUTPUT_THICKNESS\n") ;
			// End OUTPUT_THICKNESS block
			fprintf(fp,"}DISPSURF_CALCULATION_CONDITION\n") ;
			// End DISPSURF_CALCULATION_CONDITION block
			break ;
		case kMBFIT_BLOCH_MODE:
			// START BLOCH_CALCULATION_CONDITION block
			fprintf(fp,"BLOCH_CALCULATION_CONDITION{\n") ;
			fprintf(fp,"#\n") ;
			// START INCIDENT_BEAM block
			fprintf(fp,"INCIDENT_BEAM{\n") ;
			fprintf(fp,"%d %d %d    # Incidence	(upward)\n",
				mMBFITRecord.mZoneAxis.a[0],mMBFITRecord.mZoneAxis.a[1],mMBFITRecord.mZoneAxis.a[2]) ;
			fprintf(fp,"%d %d %d    # Surface normal (upward)\n",
				mMBFITRecord.mSurfaceNormal.a[0],mMBFITRecord.mSurfaceNormal.a[1],mMBFITRecord.mSurfaceNormal.a[2]) ;
			fprintf(fp,"%d %d %d    # ZOLZ base vector\n",
				mMBFITRecord.mZoneBaseVector.a[0],mMBFITRecord.mZoneBaseVector.a[1],mMBFITRecord.mZoneBaseVector.a[2]) ;
			fprintf(fp,"%f          # H tilt\n",mMBFITRecord.mTiltH) ;
			fprintf(fp,"%f          # G tilt\n",mMBFITRecord.mTiltG) ;
			fprintf(fp,"%.2f        # Accelerating Voltage [kV]\n",mMBFITRecord.mMicroscope.mVoltage) ;
			fprintf(fp,"}INCIDENT_BEAM\n") ;
			// End INCIDENT_BEAM block
			fprintf(fp,"#\n") ;
			// START CALC_KXY_RANGE block
			fprintf(fp,"CALC_KXY_RANGE{\n") ;
			fprintf(fp,"0     # cal_range_flag  [dummy: always 0]\n") ;
			fprintf(fp,"0     # Calc. type [0:0D]\n") ;
			fprintf(fp,"1   # numh: number of calc. points\n") ;
			fprintf(fp,"0.0 0.0    # hbgn, hend\n") ;
			fprintf(fp,"0.0 0.0    # gbgn, gend\n") ;
			/*
			fprintf(fp,"%d   # numh: number of calc. points\n",mMBFITRecord.mNumCalculationPoints) ;
			fprintf(fp,"%.2f  %.2f    # hbgn, hend\n",mMBFITRecord.mMinH,mMBFITRecord.mMaxH) ;
			fprintf(fp,"%.2f  %.2f    # gbgn, gend\n",mMBFITRecord.mMinG,mMBFITRecord.mMaxG) ;
			*/
			fprintf(fp,"}CALC_KXY_RANGE\n") ;
			// End CALC_KXY_RANGE block
			fprintf(fp,"#\n") ;
			fprintf(fp,"#\n") ;
			// START BEAM_SEARCH block
			if (mMBFITRecord.mBeamSearchType == kMBFIT_ManualBeamSearch) {
				fprintf(fp,"BEAM_SEARCH{\n") ;
				fprintf(fp,"# Manual beam input\n") ;
				fprintf(fp,"0     # Beam search method [0:manual, 1:auto[wg], 2:[sg]]\n") ;
				fprintf(fp,"%d %d  # num. of exact, num. of bethe\n",
					mMBFITRecord.mBeamSearchManual.num_exact,mMBFITRecord.mBeamSearchManual.num_bethe) ;
				for(i = 0 ; i < mMBFITRecord.mBeamSearchManual.num_exact; i++) {
					fprintf(fp,"%d %d %d\n",
						mMBFITRecord.mBeamSearchManual.hklExact[i].a[0],
						mMBFITRecord.mBeamSearchManual.hklExact[i].a[1],
						mMBFITRecord.mBeamSearchManual.hklExact[i].a[2]) ;
				}
				for(i = 0 ; i < mMBFITRecord.mBeamSearchManual.num_bethe; i++) {
					fprintf(fp,"%d %d %d\n",
						mMBFITRecord.mBeamSearchManual.hklBethe[i].a[0],
						mMBFITRecord.mBeamSearchManual.hklBethe[i].a[1],
						mMBFITRecord.mBeamSearchManual.hklBethe[i].a[2]) ;
				}
				fprintf(fp,"}BEAM_SEARCH\n") ;
				fprintf(fp,"#\n") ;
				fprintf(fp,"COMMENT{  --------------------------------------------------\n") ;
				fprintf(fp,"BEAM_SEARCH{\n") ;
				fprintf(fp,"# Automatic SG beam search\n") ;
				fprintf(fp,"2                  # Beam search [0:manual, 1:automatic-wg, 2:automatic-sg]\n") ;
				fprintf(fp,"200 4.0 0 0     # num_g_limit, g_max[A^{-1}], lz_min, lz_max\n") ;
				fprintf(fp,"0.02 0.03 1e+11  # sg_exact_max,  sg_bethe_max,  xi_g_max\n") ;
				fprintf(fp,"}BEAM_SEARCH\n") ;
				fprintf(fp,"}COMMENT  --------------------------------------------------\n") ;
				fprintf(fp,"#\n") ;
			}
			else if (mMBFITRecord.mBeamSearchType == kMBFIT_AutoBeamSearch_WG) {
				fprintf(fp,"BEAM_SEARCH{\n") ;
				fprintf(fp,"# Automatic WG beam search\n") ;
				fprintf(fp,"1                  # Beam search [0:manual, 1:automatic-wg, 2:automatic-sg]\n") ;
				fprintf(fp,"%d %.4f %d %d     # num_g_limit, g_max[A^{-1}], lz_min, lz_max\n",
					mMBFITRecord.mBeamSearchAuto.num_g_limit,mMBFITRecord.mBeamSearchAuto.g_max,
					mMBFITRecord.mBeamSearchAuto.lz_min,mMBFITRecord.mBeamSearchAuto.lz_max) ;
				fprintf(fp,"%.4f %.4f %e  # sg_exact_max,  sg_bethe_max,  xi_g_max\n",
					mMBFITRecord.mBeamSearchAuto.sg_exact_max,mMBFITRecord.mBeamSearchAuto.sg_bethe_max,
					mMBFITRecord.mBeamSearchAuto.xi_g_max) ;
				fprintf(fp,"}BEAM_SEARCH\n") ;
				fprintf(fp,"#\n") ;
			}
			else if (mMBFITRecord.mBeamSearchType == kMBFIT_AutoBeamSearch_SG) {
				fprintf(fp,"BEAM_SEARCH{\n") ;
				fprintf(fp,"# Automatic SG beam search\n") ;
				fprintf(fp,"2                  # Beam search [0:manual, 1:automatic-wg, 2:automatic-sg]\n") ;
				fprintf(fp,"%d %.4f %d %d     # num_g_limit, g_max[A^{-1}], lz_min, lz_max\n",
					mMBFITRecord.mBeamSearchAuto.num_g_limit,mMBFITRecord.mBeamSearchAuto.g_max,
					mMBFITRecord.mBeamSearchAuto.lz_min,mMBFITRecord.mBeamSearchAuto.lz_max) ;
				fprintf(fp,"%.4f %.4f %e  # sg_exact_max,  sg_bethe_max,  xi_g_max\n",
					mMBFITRecord.mBeamSearchAuto.sg_exact_max,mMBFITRecord.mBeamSearchAuto.sg_bethe_max,
					mMBFITRecord.mBeamSearchAuto.xi_g_max) ;
				fprintf(fp,"}BEAM_SEARCH\n") ;
				fprintf(fp,"#\n") ;
			}
			// End BEAM_SEARCH block
			// START CALC_RXY_RANGE block  

			fprintf(fp,"CALC_RXY_RANGE{\n") ;
			fprintf(fp,"%d %d    # Number of calculation points for x, y (odd number)\n",mMBFITRecord.mNumCalculationPointsX,mMBFITRecord.mNumCalculationPointsY) ;
			fprintf(fp,"%f %f   # x begin, end (along the ZOLZ base vector)\n",mMBFITRecord.mMinX,mMBFITRecord.mMaxX) ;
			fprintf(fp,"%f %f   # y begin, end (perpendicular to the ZOLZ base vector)\n",mMBFITRecord.mMinY,mMBFITRecord.mMaxY) ;
			fprintf(fp,"}CALC_RXY_RANGE\n") ;
			fprintf(fp,"#\n") ;
			// End CALC_RXY_RANGE block

			// START OUTPUT_BRANCH block  
			fprintf(fp,"OUTPUT_BRANCH{\n") ;
			fprintf(fp,"%d %d    # Number of branches to be calculated (begin, end)\n",mMBFITRecord.branchBegin,mMBFITRecord.branchEnd) ;
			fprintf(fp,"}OUTPUT_BRANCH\n") ;
			fprintf(fp,"#\n") ;
			// End OUTPUT_BRANCH block
			fprintf(fp,"}BLOCH_CALCULATION_CONDITION\n") ;
			// End BLOCH_CALCULATION_CONDITION block
			break ;
	}
	fclose(fp) ;
	mTEMRecord.mIsModified = false ;
	
}

void	TCrystal::Edit() {
}

void	TCrystal::UpdateWindowSize() {
	mDisplayRect	= mDocWindow->mCrystalRect ;
	mWidth = mDisplayRect.right-mDisplayRect.left ;
	mHeight = mDisplayRect.bottom-mDisplayRect.top ;
	
	SetCenter() ;


	mShiftX = 0 ;
	mShiftY = 0 ;

	UpdateViewAtoms() ;
}

Boolean
TCrystal::NewSymop(GENERATOR *p_symop,long num_symop) {
	register long i,j ;
	long new_gen ;
	float *p_sym,*p_old_sym,test ;

	new_gen = true ;
	for (i=0;i< num_symop;i++) {
		p_sym = (float *)(&p_symop->matr[0][0]) ;
		p_old_sym = (float *)&mSymOps[i] ;
		for(j=0; j < 16 ;j++) {
			test = *p_sym - *p_old_sym ;
			test = (test >= 0. ? test : -test) ;
			if (test > 0.01) 
				break ;
			p_sym++ ;
			p_old_sym++ ;
		}
		if (j==16) {
			new_gen = false ;
			break ;
		}
	}
	return(new_gen) ;
}

void 
TCrystal::LoadSpacegroup(char *spacegroup,short spcgrp_number) {	
	strcpy(spacegroup,gSpacegroupStrings[spcgrp_number-1]) ;
}

Boolean TCrystal::Equivalent(TAtom *atomList,float x,float y,float z,long istart,long ndx) {
  /* ----------------------------------------------------------	*/
  /* Routine to check if the atom whose coordinates were just	*/
  /* passed is a new one. The search is only through the list 	*/
  /* of atoms of the same type. Thus if too many atoms were put	*/
  /* into the basis, the application of the symmetry operators	*/
  /* will produce too many atoms in the total structure. Beware	*/
  /* Returns (false) if this is a 'new' atom, (true) otherwise  */
  /* ----------------------------------------------------------	*/

	long 	k ;
	Boolean xtest,ytest,ztest ;
	float 	xdiff,ydiff,zdiff ;

	for (k=istart;k<ndx;k++) {
      		xdiff = (x-(atomList+k)->xc)*mA ;
      		ydiff = (y-(atomList+k)->yc)*mB ;
      		zdiff = (z-(atomList+k)->zc)*mC ;      
      		xtest = false ; 
      		ytest = false ; 
      		ztest = false ;
      		if (fabs(xdiff) < 0.1) xtest = true ;
      		if (fabs(ydiff) < 0.1) ytest = true ;
      		if (fabs(zdiff) < 0.1) ztest = true ;
      		if (xtest && ytest && ztest) 
         		return(true) ;              /* The atom already exists */
	}
	return(false) ;                   	/* New atom                */
} 
	
void TCrystal::Assert(void *ptr) {
	if (ptr == nil) {
		::PutMessage("Assertion Failure in TCrystal. We ran out of memory") ;
		exit(0) ;
	}
}

TCrystal::~TCrystal() {

	
	if (mAtomPtr != nil) 
		delete[] mAtomPtr ;
	if (mBasisPtr != nil && mBasisPtr != mAtomPtr)
		delete[] mBasisPtr ;
	if (mScreenAtom != nil) 
		delete[] mScreenAtom ;
	if (mSymOps != nil) 
		delete[] mSymOps ;	
}

void 
TCrystal::ApplySymmops() {

  	/* ------------------------------------------------------------	*/
  	/* Applies the symmetry operators on the atoms in the basis 	*/
  	/* ------------------------------------------------------------	*/
    
	register long i,j,ndx,icount ;
	float xc,yc,zc ;
	float xnew,ynew,znew ;
	
	ndx = 0 ;
	for (j=0;j<mNumBasisAtoms;j++) {
		icount = ndx ;
		xc = (mBasisPtr+j)->xc ;
		yc = (mBasisPtr+j)->yc ; 
		zc = (mBasisPtr+j)->zc ;
		for (i=0;i<mNumSymOps;i++) {
			xnew = (mSymOps+i)->m[0][0]*xc + (mSymOps+i)->m[0][1]*yc +
				   (mSymOps+i)->m[0][2]*zc + (mSymOps+i)->m[0][3] ;
			ynew = (mSymOps+i)->m[1][0]*xc + (mSymOps+i)->m[1][1]*yc +
				   (mSymOps+i)->m[1][2]*zc + (mSymOps+i)->m[1][3] ;
			znew = (mSymOps+i)->m[2][0]*xc + (mSymOps+i)->m[2][1]*yc +
				   (mSymOps+i)->m[2][2]*zc + (mSymOps+i)->m[2][3] ;
			xnew = one(xnew) ;
			ynew = one(ynew) ;
			znew = one(znew) ;
			if (!Equivalent(mAtomPtr,xnew,ynew,znew,icount,ndx)) {
				*(mAtomPtr+ndx) = *(mBasisPtr+j) ;
				mAtomPtr[ndx].xc = xnew ;
				mAtomPtr[ndx].yc = ynew ;
				mAtomPtr[ndx].zc = znew ;
				ndx++ ;
			}
		}
	}
	mNumUnitCellAtoms = ndx ;
}

void	
TCrystal::WriteInfo(char *text) {
	mDocWindow->WriteMessage(text) ;	
}

void	
TCrystal::SizeAtoms(Boolean reduceSize) {
	if (reduceSize) {
		mAtomScaleFactor /= 1.1 ;
	}
	else {
		mAtomScaleFactor *= 1.1 ;
	}
	UpdateViewAtoms() ;
	DrawCrystal() ;
}

// The vector izone is to be interpreted as a plane normal, thus it is a vector
// in reciprocal space (hkl)
void 
TCrystal::CalculateRealZoneFromReciprocal(TVector vec) {
	TVector 	recZone ;
	long		 i ;
	float 	lengthRecipVector,dSpacing,lengthRealVector,
			minVal,factor,absVal ;		
	
			
	recZone = vec ;
	
	RecipToRlVector(recZone,vec) ;
	
	lengthRecipVector = LengthOfRecipVector(recZone) ;
	dSpacing = 1./lengthRecipVector ;
	lengthRealVector = LengthOfRealVector(vec) ;
	
	factor = dSpacing/lengthRealVector ;
	
	vec *= factor ;
	
	// We probably still want to reduce so that we have no more than 1 
	// non-integer value ;
	minVal = 10000. ;
	for(i=0;i<3;i++) {
		absVal = fabs(vec.a[i]) ;
		minVal = (absVal < minVal && absVal != 0. ? absVal : minVal) ;
	}
	vec /= minVal ;
	
	// Now one of the numbers are an integer.  
	// Multiply with an integer number until all are or until one more is
	// integer and we have exceeded a maximum on the integer value
	long	maxInt = 10 ;
	Boolean allIntegers ;
	for(long j=1;j<=maxInt;j++) {
		vec.a[i] *= j ;
		allIntegers = false ;
		for (i=0;i<3;i++) {
			allIntegers = (NumberIsInteger(vec.a[i]) && allIntegers) ;
		}
		if (allIntegers) 
			break ;
	}
}

void 
TCrystal::RecipToRlVector(TVector &rcpVec,TVector &rlVec) {
	TMatrix sinv,sinvt ;

	/* We have (a,b,c) = S (ast,bst,cst), such that (ast,bst,cst) = S(-1) (a,b,c)	*/
	/* For coordinates we then have (xa,xb,xc) = [S(-1)] (T) (xast,xbst,xcst)		*/
	InvertMatrix(mSmatrix,sinv) ;
	TransposeMatrix(sinv,sinvt) ;
	rlVec = sinvt*rcpVec ;
	
}

void 
TCrystal::RlToRecipVector(TVector &rlVec,TVector &rcpVec) {
    TMatrix st ;

    /* We have (a,b,c) = S (ast,bst,cst)								*/
    /* For coordinates we then have (xast,xbst,xcst) = S (T) (xa,xb,xc)	*/
	TransposeMatrix(mSmatrix,st) ;
	rcpVec = st*rlVec ;
}

float 
TCrystal::LengthOfRecipVector(TVector &vec) {

  	float 	gsq,ast,bst,cst,csgast,csalst,csbest,h,k,l ;
  	float 	astsq,bstsq,cstsq,Vol ;
	float	sial,sibe,siga,csal,csbe,csga ;
	
  	sial 	= sin(mAlpha) ;
	sibe 	= sin(mBeta) ;
	siga 	= sin(mGamma) ;
	csal 	= cos(mAlpha) ;
	csbe 	= cos(mBeta) ;
	csga 	= cos(mGamma) ;
	Vol		= mA*mB*mC*sqrt(1.-csal*csal-
						csbe*csbe-csga*csga+2.*csal*csbe*csga) ;
	ast 	= mB*mC*sial/Vol ;
	bst 	= mC*mA*sibe/Vol ;
	cst 	= mA*mB*siga/Vol ;
	csalst	= (csbe*csga-csal)/(sibe*siga) ;
	csbest	= (csga*csal-csbe)/(siga*sial) ;
	csgast	= (csal*csbe-csga)/(sial*sibe) ;

	astsq = ast*ast ;
	bstsq = bst*bst ;
	cstsq = cst*cst ;
    
	h = vec.a[0] ;
	k = vec.a[1] ;
	l = vec.a[2] ;
    
	gsq =  h*h*astsq + k*k*bstsq + l*l*cstsq +
				2*h*k*ast*bst*csgast+
				2*h*l*ast*cst*csbest+
				2*k*l*bst*cst*csalst  ;
	return(sqrt(gsq)) ;
}

float 
TCrystal::LengthOfRealVector(TVector &vec) {

  	float 	lsq,u,v,w ;
  	float 	a,b,c,asq,bsq,csq ;
	float	csal,csbe,csga ;
	
  	csal 	= cos(mAlpha) ;
	csbe 	= cos(mBeta) ;
	csga 	= cos(mGamma) ;
	
	a		= mA ;
	b		= mB ;
	c		= mC ;
	asq 	= a*a ;
	bsq		= b*b ;
	csq		= c*c ;
	
	u 		= vec.a[0] ;
	v		= vec.a[1] ;
	w		= vec.a[2] ;
	lsq =  u*u*asq + v*v*bsq + w*w*csq +
				2*u*v*a*b*csga+
				2*u*w*a*c*csbe+
				2*v*w*b*c*csal  ;
	return(sqrt(lsq)) ;
}


void
TCrystal::SetScale() {
	float 	scale,ax,bx,cx,ay,by,cy,
			xlength,ylength,scalex,scaley ;
	// Draw the outline.  The projection of the three axis on the screen
	ax = mXyzToScreenMatrix.m[0][0]; ay = mXyzToScreenMatrix.m[1][0]; 
	bx = mXyzToScreenMatrix.m[0][1]; by = mXyzToScreenMatrix.m[1][1]; 
	cx = mXyzToScreenMatrix.m[0][2]; cy = mXyzToScreenMatrix.m[1][2] ;

	xlength = fabs(mXViewCells*ax) + fabs(mYViewCells*bx) + fabs(mZViewCells*cx) ;	
	/* Span of view in ngstrm horizontally	*/
	
	ylength = fabs(mXViewCells*ay) + fabs(mYViewCells*by) + fabs(mZViewCells*cy) ;	
	/* Span of unit cell in ngstrm vertically		*/

	scale = max(xlength,ylength) ;
	scalex = 0.6*mWidth/scale ;				/* Num pixels per Angstrom 	*/
	scaley = 0.6*mHeight/scale ;			/* Num pixels per Angstrom 	*/
	scale  = min(scalex,scaley) ;			/* Final scale				*/
	
	scale *= mMagnification ;
	mXSection.AngstrPrPixel	= 1./scale ;
	 	
	SetCenter() ;
	
}
void 
TCrystal::ResetView() {
	mShiftX = 0 ;
	mShiftY = 0 ;
}

void	
TCrystal::PanView(Point &loc) {
	Point lastPoint,currentPoint ;
	long dx,dy ;
	
	lastPoint = loc ;
	currentPoint = loc ;
	while(Button()) {
		GetMouse(&currentPoint) ;
		dx = currentPoint.x - lastPoint.x ;
		dy = -(currentPoint.y - lastPoint.y) ;
		if (dx != 0 || dy != 0) {
			mShiftX += dx ;
			mShiftY += dy ;
			
			mXyzToScreenMatrix.m[0][3] += dx ;
			mXyzToScreenMatrix.m[1][3] += dy ;
			
		    mXSection.Origin.x		+= dx ;
		    mXSection.Origin.y		+= dy ;
		    UpdateViewAtoms() ;
			DrawCrystal() ;
			if (mSelection != nil) {
				mSelection->Pan(dx,-dy) ;
				mSelection->Update(&mDisplayRect) ;
			}			
			lastPoint = currentPoint ;
		}
	}
}

void 
TCrystal::SetCenter(void)
{
	short 	iorigox,iorigoy,iorigoz ;
	float	scale,xpr,ypr,zpr ;
	
	scale = 1./mXSection.AngstrPrPixel ;
	
	/* We want the coordinates 1/2*nA, 1/2*nB, 1/2*nC to fall in the center of the screen */
	xpr = 0.5*(mXyzToScreenMatrix.m[0][0]*mXViewCells + mXyzToScreenMatrix.m[0][1]*mYViewCells + 
			mXyzToScreenMatrix.m[0][2]*mZViewCells) ;
	ypr = 0.5*(mXyzToScreenMatrix.m[1][0]*mXViewCells + mXyzToScreenMatrix.m[1][1]*mYViewCells +
			 mXyzToScreenMatrix.m[1][2]*mZViewCells) ;
	zpr = 0.5*(mXyzToScreenMatrix.m[2][0]*mXViewCells + mXyzToScreenMatrix.m[2][1]*mYViewCells + 
			mXyzToScreenMatrix.m[2][2]*mZViewCells) ;
	iorigox = (short)(mWidth/2 - xpr*scale) + mShiftX ;
	iorigoy = (short)(mHeight/2 - ypr*scale) + mShiftY ;	
	iorigoz = (short)(-zpr*scale) ;
	mXyzToScreenMatrix.m[0][3] = iorigox ;
	mXyzToScreenMatrix.m[1][3] = iorigoy ;
	mXyzToScreenMatrix.m[2][3] = iorigoz ;
	
	mXSection.Origin.x		= iorigox ;
	mXSection.Origin.y		= iorigoy ;

}

void
TCrystal::SetTransformationMatrix(TVector &view) {

	TVector 	screenUVW ;
	TMatrix 	temp ;
	short 		i,j ;
	float		ax,ay,pi,phipr ;
	float 		uvwmag,theta,phi,xymag ;
	
	
	for(i=0;i<4;i++) 				// Reset rotation matrix
		for(j=0;j<4;j++)
			mRotMatrix.m[i][j] = 0. ;
	for(i=0;i<4;i++) mRotMatrix.m[i][i] = 1. ;
	
	temp = mTmatrix*mRotMatrix ;
	TransposeMatrix(temp,mXyzToScreenMatrix) ;
	
	
	// Find the appropriate angles to rotate the crystal
	screenUVW = mXyzToScreenMatrix*view ;	// The UVW given in screen ccordinates
		
	uvwmag = screenUVW.length() ;
	
	xymag = screenUVW.a[0]*screenUVW.a[0]+screenUVW.a[1]*screenUVW.a[1] ;
	xymag = xymag > 0. ? sqrt(xymag) : 0. ;
	
	pi = 2*asin(1.0) ;
	if (screenUVW.a[0] == 0. && screenUVW.a[1] == 0.) {
		phi = 0. ;
		theta = 0. ;
	}
	else if (screenUVW.a[0] == 0.) {
		phi = (screenUVW.a[1] > 0. ?  pi/2. : -pi/2. ) ;
		theta = acos(screenUVW.a[2]/uvwmag) ;
	}
	else if (screenUVW.a[1] == 0.) {
		phi = (screenUVW.a[0] > 0. ?  0. : pi ) ;
		theta = acos(screenUVW.a[2]/uvwmag) ;
	}
	else {
		theta = acos(screenUVW.a[2]/uvwmag) ;
		phi = acos(screenUVW.a[0]/xymag) ;
		phi = (screenUVW.a[1] > 0. ?  phi : -phi ) ;
	}
	
	SetDisRotMatrix(zRotAxis,-phi) ;
	SetDisRotMatrix(yRotAxis,-theta) ;
		
	
	// Find the angle that a makes with the x-axis
	ax = mXyzToScreenMatrix.m[0][0]; ay = mXyzToScreenMatrix.m[1][0]; 
	xymag = ax*ax+ay*ay ;
	xymag = xymag > 0. ? sqrt(xymag) : 0. ;
	if (ax == 0. && ay == 0.) {
		phipr = 0. ;
	}
	else if (ay == 0.) {
		phipr = (ax > 0. ?  0. : pi ) ;
	}
	else if (ax == 0.) {
		phipr = (ay > 0. ?  pi/2. : -pi/2. ) ;
	}
	else {
		phipr = acos(ax/xymag) ;
	}
	
	SetDisRotMatrix(zRotAxis,-phipr) ;
	temp = mTmatrix*mRotMatrix ;	// MultiplyMatrix(gTMatrix,gDisplayRotMatrix,temp) ;
	
	TransposeMatrix(temp,mXyzToScreenMatrix) ;

	SetScale() ;
	
    
}



void 
TCrystal::CreateUnitCellViewAtoms()
{
	short 		i,n,m,l ;
	float		TZ0,fn,fm,fl ;
	float 		x,y,z,scale,rad ;
	RGBColor	color ;
	
	
	scale	=	1/mXSection.AngstrPrPixel ;	
	TZ0		= 	1.0 ;
	
	SetUCellFaces() ;
	
	mSnumAtoms = 0 ;
	fn = (float)mXViewCells ;
	fm = (float)mYViewCells ;
	fl = (float)mZViewCells ;
	for(l=0;l<mZViewCells;l++) {
		for(m=0;m<mYViewCells;m++) {
			for(n=0;n<mXViewCells;n++) {
				/* Transform all the coordinates of the atoms in the cell to screen coordinates */
				for(i=0;i< mNumUnitCellAtoms;i++) {
					x = (mAtomPtr+i)->xc + n ;
					y = (mAtomPtr+i)->yc + m ;
					z = (mAtomPtr+i)->zc + l ;
					rad = (mAtomPtr+i)->rad;	// bjf 2/96  for custom radii
					color	= (mAtomPtr+i)->color ;
					
					if (fabs(x) < 0.00001) x = 0. ;
					if (fabs(y) < 0.00001) y = 0. ;
					if (fabs(z) < 0.00001) z = 0. ;
					
					AddViewAtom(x,y,z,TZ0,scale,i,rad,&color) ;
					
					if (x == 0. || y == 0. || z == 0.) {
						if (x==0. && y != 0. && z != 0.) {
							AddViewAtom(fn,y,z,TZ0,scale,i,rad,&color) ;
						}
						else if (x==0. && y==0. && z!=0.) {
							AddViewAtom(fn,y,z,TZ0,scale,i,rad,&color) ;
							AddViewAtom(x,fm,z,TZ0,scale,i,rad,&color) ;
							AddViewAtom(fn,fm,z,TZ0,scale,i,rad,&color) ;
						}
						else if (x==0. && y==0. && z==0.) {
							AddViewAtom(fn,y,z,TZ0,scale,i,rad,&color) ;
							AddViewAtom(fn,fm,z,TZ0,scale,i,rad,&color) ;
							AddViewAtom(x,fm,z,TZ0,scale,i,rad,&color) ;
							AddViewAtom(x,y,fl,TZ0,scale,i,rad,&color) ;
							AddViewAtom(fn,y,fl,TZ0,scale,i,rad,&color) ;
							AddViewAtom(x,fm,fl,TZ0,scale,i,rad,&color) ;
							AddViewAtom(fn,fm,fl,TZ0,scale,i,rad,&color) ;
						}
						else if (y==0. && x!=0. &&z!=0.) {
							AddViewAtom(x,fm,z,TZ0,scale,i,rad,&color) ;
						}
						else if (y==0. && x!=0. && z==0.) {
							AddViewAtom(x,fm,z,TZ0,scale,i,rad,&color) ;
							AddViewAtom(x,fm,fl,TZ0,scale,i,rad,&color) ;
							AddViewAtom(x,y,fl,TZ0,scale,i,rad,&color) ;
						}
						else if (z==0. && x!=0. && y!=0.) {
							AddViewAtom(x,y,fl,TZ0,scale,i,rad,&color) ;
						}
						else if (z==0. && x==0. && y!=0.) {
							AddViewAtom(fn,y,z,TZ0,scale,i,rad,&color) ;
							AddViewAtom(x,y,fl,TZ0,scale,i,rad,&color) ;
							AddViewAtom(fn,y,fl,TZ0,scale,i,rad,&color) ;
						}
					}
				}
			}
		}
	}
		
	// Before we can draw the atoms, we must sort the atoms according to the z-position
	SortScreenAtoms() ;
}

void 
TCrystal::DrawUnitCell() 
{
	if (mDrawOutline) {
		DrawInvisibleUCellFaces() ;
	}
	// Draw the Crystal 
	Draw2DScreenBalls() ;
	
	if (mDrawOutline) {
		DrawVisibleUCellFaces() ;
	}
	
	if(mDrawOrigin)
		DrawOrigin(); // bjf 9/96
	
}

void TCrystal::DrawCrystal() {
	SetPort(mDocWindow) ;			
	Draw() ;
	
}

void 
TCrystal::Draw() 
{
	
	InsetRect(&mDisplayRect,1,1) ;
	TStOffscreen offScreen(mDocWindow,&mDisplayRect) ;
	if (offScreen.GetPixmap() != nil) {
		InsetRect(&mDisplayRect,-1,-1) ;
		DrawUnitCell() ;
	}
	
}

void
TCrystal::DrawOrigin() // test function to see if it is easy to id the origin
{

	Point screenPt0, screenPtA, screenPtB, screenPtC, A, B, C;
	TVector V;
	
	V.a[0] = V.a[1] = V.a[2] = 0.0;
	screenPt0 = VectorToPoint(V);
	
	V.a[0] = 0.5;
	V.a[1] = V.a[2] = 0.0;
	screenPtA = VectorToPoint(V);
	V.a[0] = 0.6;
	A = VectorToPoint(V);
	
	V.a[1] = 0.5;
	V.a[0] = V.a[2] = 0.0;
	screenPtB = VectorToPoint(V);
	V.a[1] = 0.6;
	B = VectorToPoint(V);

	V.a[2] = 0.5;
	V.a[0] = V.a[1] = 0.0;
	screenPtC = VectorToPoint(V);
	V.a[2] = 0.6;
	C = VectorToPoint(V);
		
	::PenSize(3,3);
	SetFgColor(gBluePixel) ;
	DrawLine(screenPt0.x,screenPt0.y,screenPtA.x,screenPtA.y); 
	DrawLine(screenPt0.x,screenPt0.y,screenPtB.x,screenPtB.y);
	DrawLine(screenPt0.x,screenPt0.y,screenPtC.x,screenPtC.y); 

	::MoveTo(A.x, A.y);
	::DrawString("A");
	//::DrawText("A", 0, 1);
	::MoveTo(B.x, B.y);
	::DrawString("B");
	//::DrawText("B", 0, 1);
	::MoveTo(C.x, C.y);
	//::DrawText("C", 0, 1);
	::DrawString("C");
	::PenNormal();	
	SetFgColor(gBlackPixel) ;

}


// this gets called alot so I made a subroutine 
// just takes a vector in crystal space and maps to screen coords
Point 
TCrystal::VectorToPoint(TVector inV)
{
	Point outPt;
	float x,y,z,xt,yt,zt, scale	= 1/mXSection.AngstrPrPixel;
	
	x = inV.a[0] ;
	y = inV.a[1] ;
	z = inV.a[2] ;
	
	xt = mXyzToScreenMatrix.m[0][0]*x + mXyzToScreenMatrix.m[0][1]*y + mXyzToScreenMatrix.m[0][2]*z ;
	yt = mXyzToScreenMatrix.m[1][0]*x + mXyzToScreenMatrix.m[1][1]*y + mXyzToScreenMatrix.m[1][2]*z ;
	zt = mXyzToScreenMatrix.m[2][0]*x + mXyzToScreenMatrix.m[2][1]*y + mXyzToScreenMatrix.m[2][2]*z ;	
	xt = scale*xt+mXyzToScreenMatrix.m[0][3] ;
	yt = scale*yt+mXyzToScreenMatrix.m[1][3] ;
	zt = scale*zt+mXyzToScreenMatrix.m[2][3] ;	
	outPt.x = near(xt) ;
	outPt.y = mHeight - near(yt) ;
	
	//outPt.x += mDisplayRect.left ;
	//outPt.y += mDisplayRect.top ;
	return outPt;
}


void
TCrystal::SortScreenAtoms(void) {
	long gap,n ;
	register long i,j;
	SAtom tAtom ; 
	
	n = mSnumAtoms ;
	for (gap=n/2;gap>0;gap/=2) {
		for(i=gap;i<n;i++) {
			for(j=i-gap;j>=0;j-=gap) {
				if ((mScreenAtom+j)->location.d <= (mScreenAtom+j+gap)->location.d)
					break ;					
				tAtom = *(mScreenAtom+j) ;				
				*(mScreenAtom+j) = *(mScreenAtom+j+gap) ;				
				*(mScreenAtom+j+gap) = tAtom ;
			}
		}
	}
}

void TCrystal::AddViewAtom(float x,float y, float z,float TZ0,float scale,short i,float atomrad,RGBColor *color)
{
	float 	xt,yt,zt ;
	short	x0,y0,z0 ;
	float 	rad ;
	SAtom *whichAtom ;
	
	xt = mXyzToScreenMatrix.m[0][0]*x + mXyzToScreenMatrix.m[0][1]*y + mXyzToScreenMatrix.m[0][2]*z ;
	yt = mXyzToScreenMatrix.m[1][0]*x + mXyzToScreenMatrix.m[1][1]*y + mXyzToScreenMatrix.m[1][2]*z ;
	zt = mXyzToScreenMatrix.m[2][0]*x + mXyzToScreenMatrix.m[2][1]*y + mXyzToScreenMatrix.m[2][2]*z ;

	rad = mAtomScaleFactor*atomrad*scale;	// Easier to work with
	if(atomrad <= 0.0)
	{
		rad = mAtomScaleFactor*covrad[(mAtomPtr+i)->no-1]*scale ; 
	}
	whichAtom = (mScreenAtom+mSnumAtoms) ;
	(mScreenAtom+mSnumAtoms)->atom = *(mAtomPtr+i) ;
	(mScreenAtom+mSnumAtoms)->atom.xc = x ; 
	(mScreenAtom+mSnumAtoms)->atom.yc = y ; 
	(mScreenAtom+mSnumAtoms)->atom.zc = z ;
	(mScreenAtom+mSnumAtoms)->position.h = xt ;
	(mScreenAtom+mSnumAtoms)->position.v = yt ;
	(mScreenAtom+mSnumAtoms)->position.d = zt ;
	(mScreenAtom+mSnumAtoms)->index = i ; 			// index into (mAtomPtr+i) ;
	
	xt = scale*xt+mXyzToScreenMatrix.m[0][3] ;
	yt = scale*yt+mXyzToScreenMatrix.m[1][3] ;
	zt = scale*zt+mXyzToScreenMatrix.m[2][3] ;
	
	x0 = near(xt) ;
	y0 = near(yt) ;
	z0 = near(zt) ;
	(mScreenAtom+mSnumAtoms)->location.h = x0 ;
	(mScreenAtom+mSnumAtoms)->location.v = y0 ;
	(mScreenAtom+mSnumAtoms)->location.d = z0 ;
	(mScreenAtom+mSnumAtoms)->rad = near(rad) ;
	(mScreenAtom+mSnumAtoms)->state = (mAtomPtr+i)->state ;
	(mScreenAtom+mSnumAtoms)->groupnumber = (mAtomPtr+i)->groupnumber ;
	mSnumAtoms++ ;
}

void TCrystal::SetTMatrix() {
	float csa,csb,csg,sia,sib,sig ;

	csa = CheckZeroOne(cos(mAlpha)) ;
	csb = CheckZeroOne(cos(mBeta)) ;
	csg = CheckZeroOne(cos(mGamma)) ;
	sia = CheckZeroOne(sin(mAlpha)) ;
	sib = CheckZeroOne(sin(mBeta)) ;
	sig = CheckZeroOne(sin(mGamma)) ;
    
    
	mTmatrix.m[0][0] = mA ;
	mTmatrix.m[0][1] = 0. ;
	mTmatrix.m[0][2] = 0. ;
	mTmatrix.m[0][3] = 0. ;
	mTmatrix.m[1][0] = mB*csg ;
	mTmatrix.m[1][1] = mB*sig ;
	mTmatrix.m[1][2] = 0. ;
	mTmatrix.m[1][3] = 0. ;
	mTmatrix.m[2][0] = mC*csb ;
	mTmatrix.m[2][1] = mC*(csa-csg*csb)/sig ;
	mTmatrix.m[2][2] = mC*sqrt(1.-csa*csa-csb*csb-csg*csg+2.*csa*csb*csg)/sig ;
	mTmatrix.m[2][3] = 0. ;
	mTmatrix.m[3][0] = 0. ;
	mTmatrix.m[3][1] = 0. ;
	mTmatrix.m[3][2] = 0. ;
	mTmatrix.m[3][3] = 1. ;
}

void TCrystal::SetSMatrix() {
	float csa,csb,csg,aa,ab,ac,bb,bc,cc ;
	csa = CheckZeroOne(cos(mAlpha)) ;
	csb = CheckZeroOne(cos(mBeta)) ;
	csg = CheckZeroOne(cos(mGamma)) ;
	aa 	= mA*mA ;
	ab	= mA*mB ;
	ac 	= mA*mC ;
	bb	= mB*mB ;
	bc	= mB*mC ;
	cc	= mC*mC ;
	
	mSmatrix.m[0][0] 	= aa ;
	mSmatrix.m[0][1]	= ab*csg ;
	mSmatrix.m[0][2]	= ac*csb ;
	mSmatrix.m[0][3] 	= 0. ;
	mSmatrix.m[1][0]	= ab*csg ;
	mSmatrix.m[1][1]	= bb ;
	mSmatrix.m[1][2]	= bc*csa ;
	mSmatrix.m[1][3] 	= 0. ;
	mSmatrix.m[2][0]	= ac*csb ;
	mSmatrix.m[2][1]	= bc*csa ;
	mSmatrix.m[2][2]	= cc ;
	mSmatrix.m[2][3] 	= 0. ;
	mSmatrix.m[3][0] 	= 0. ;
	mSmatrix.m[3][1] 	= 0. ;
	mSmatrix.m[3][2] 	= 0. ;
	mSmatrix.m[3][3] 	= 1. ;
}

void TCrystal::SetDisRotMatrix(TypeRotAxis whichAxis,float angle) {
	float 	csangle,siangle ;
	TMatrix 	tRotMatrix(1.0),oldMatrix ;
	
	oldMatrix = mRotMatrix ;
			
	csangle = cos(angle) ;
	siangle = sin(angle) ;
	
	switch(whichAxis) {
		case xRotAxis:
			tRotMatrix.m[0][0] = 1. ;
			tRotMatrix.m[0][1] = 0. ;
			tRotMatrix.m[0][2] = 0. ;
			tRotMatrix.m[1][0] = 0. ;
			tRotMatrix.m[2][0] = 0. ;
			tRotMatrix.m[1][1] = csangle ;
			tRotMatrix.m[1][2] = siangle ;
			tRotMatrix.m[2][1] = -siangle ;
			tRotMatrix.m[2][2] = csangle ;
			break ;
		case yRotAxis :
			tRotMatrix.m[1][1] = 1. ;
			tRotMatrix.m[0][1] = 0. ;
			tRotMatrix.m[1][0] = 0. ;
			tRotMatrix.m[1][2] = 0. ;
			tRotMatrix.m[2][1] = 0. ;
			tRotMatrix.m[0][0] = csangle ;
			tRotMatrix.m[2][0] = siangle ;
			tRotMatrix.m[0][2] = -siangle ;
			tRotMatrix.m[2][2] = csangle ;
			break ;
		case zRotAxis:
			tRotMatrix.m[2][2] = 1. ;
			tRotMatrix.m[0][2] = 0. ;
			tRotMatrix.m[1][2] = 0. ;
			tRotMatrix.m[2][0] = 0. ;
			tRotMatrix.m[2][1] = 0. ;
			tRotMatrix.m[0][0] = csangle ;
			tRotMatrix.m[0][1] = siangle ;
			tRotMatrix.m[1][0] = -siangle ;
			tRotMatrix.m[1][1] = csangle ;
			break ;
	}
	
	mRotMatrix = oldMatrix*tRotMatrix ;
}



void TCrystal::Draw2DScreenBalls(void) {
	register long i ;
	
	SetFgColor(gBlackPixel) ;
	for(i=0;i< mSnumAtoms;i++) {
		DrawAtom(mScreenAtom+i) ;
	}
	SetFgColor(gBlackPixel) ;

}


void TCrystal::DrawInvisibleUCellFaces() {
	long i ;
	
	
	SetLineStyle(DOT) ;
	for (i=0;i<6;i++) {
		if (!IsFaceVisible(mUCellFace[i])) 
			DrawFace(mUCellFace[i]) ;
	}
	SetLineStyle(SOLID) ;
}


void TCrystal::DrawVisibleUCellFaces() {
	long i ;
	for (i=0;i<6;i++) {
		if (IsFaceVisible(mUCellFace[i])) 
			DrawFace(mUCellFace[i]) ;
	}
}

// -------------------------------------------------------------------
// Sorts the faces according to their height on the screen           
// 
void TCrystal::SortFaces() {
  

	long 	i,j,gap,n ;
	UCellFace	temp ;
 
 	n = 6 ;
	for (gap = n/2; gap > 0 ;gap /= 2)
		for (i = gap; i < n; i++)
			for (j = i-gap;j >= 0;j -= gap) {
          			if ((mUCellFace[j].cz-mUCellFace[j+gap].cz) <= 0.) 
             				break ;  
          			temp = mUCellFace[j] ;
          			mUCellFace[j] = mUCellFace[j+gap] ;
          			mUCellFace[j+gap] = temp ;
        		}
}


Boolean TCrystal::IsFaceVisible(UCellFace face) {
	return(face.normal[2] > 0.) ;
}


void TCrystal::DrawFace(UCellFace face) {
	
	DrawLine(near(face.corner[0].h),near(mHeight-face.corner[0].v),
				near(face.corner[1].h),near(mHeight-face.corner[1].v)) ;
	DrawLine(near(face.corner[1].h),near(mHeight-face.corner[1].v),
				near(face.corner[2].h),near(mHeight-face.corner[2].v)) ;
	DrawLine(near(face.corner[2].h),near(mHeight-face.corner[2].v),
				near(face.corner[3].h),near(mHeight-face.corner[3].v)) ;
	DrawLine(near(face.corner[3].h),near(mHeight-face.corner[3].v),
				near(face.corner[0].h),near(mHeight-face.corner[0].v)) ;
}

void TCrystal::SetUCellFaces() {
	register long i,j,k ;
	register float ax,bx,cx ;
	float x1,y1,x2,y2,z1,z2 ;
	TMatrix M ;
	
	M = mXyzToScreenMatrix ;
	
	mUCellFace[0].corner[0].h = 0. ;
	mUCellFace[0].corner[0].v = 0. ;
	mUCellFace[0].corner[0].d = 0. ;
	mUCellFace[0].corner[1].h = 0. ;
	mUCellFace[0].corner[1].v = mYViewCells ;
	mUCellFace[0].corner[1].d = 0. ;
	mUCellFace[0].corner[2].h = mXViewCells ;
	mUCellFace[0].corner[2].v = mYViewCells ;
	mUCellFace[0].corner[2].d = 0. ;
	mUCellFace[0].corner[3].h = mXViewCells ;
	mUCellFace[0].corner[3].v = 0. ;
	mUCellFace[0].corner[3].d = 0. ;
	
	
	mUCellFace[1].corner[0].h = 0. ;
	mUCellFace[1].corner[0].v = 0. ;
	mUCellFace[1].corner[0].d = 0. ;
	mUCellFace[1].corner[1].h = mXViewCells ;
	mUCellFace[1].corner[1].v = 0. ;
	mUCellFace[1].corner[1].d = 0. ;
	mUCellFace[1].corner[2].h = mXViewCells ;
	mUCellFace[1].corner[2].v = 0. ;
	mUCellFace[1].corner[2].d = mZViewCells ;
	mUCellFace[1].corner[3].h = 0. ;
	mUCellFace[1].corner[3].v = 0. ;
	mUCellFace[1].corner[3].d = mZViewCells ;
	
	mUCellFace[2].corner[0].h = 0. ;
	mUCellFace[2].corner[0].v = 0. ;
	mUCellFace[2].corner[0].d = 0. ;
	mUCellFace[2].corner[1].h = 0. ;
	mUCellFace[2].corner[1].v = 0. ;
	mUCellFace[2].corner[1].d = mZViewCells ;
	mUCellFace[2].corner[2].h = 0. ;
	mUCellFace[2].corner[2].v = mYViewCells ;
	mUCellFace[2].corner[2].d = mZViewCells ;
	mUCellFace[2].corner[3].h = 0. ;
	mUCellFace[2].corner[3].v = mYViewCells ;
	mUCellFace[2].corner[3].d = 0. ;
	
	mUCellFace[3].corner[0].h = 0. ;
	mUCellFace[3].corner[0].v = 0. ;
	mUCellFace[3].corner[0].d = mZViewCells ;
	mUCellFace[3].corner[1].h = mXViewCells ;
	mUCellFace[3].corner[1].v = 0. ;
	mUCellFace[3].corner[1].d = mZViewCells ;
	mUCellFace[3].corner[2].h = mXViewCells ;
	mUCellFace[3].corner[2].v = mYViewCells ;
	mUCellFace[3].corner[2].d = mZViewCells ;
	mUCellFace[3].corner[3].h = 0. ;
	mUCellFace[3].corner[3].v = mYViewCells ;
	mUCellFace[3].corner[3].d = mZViewCells ;
	
	mUCellFace[4].corner[0].h = mXViewCells ;
	mUCellFace[4].corner[0].v = 0. ;
	mUCellFace[4].corner[0].d = 0. ;
	mUCellFace[4].corner[1].h = mXViewCells ;
	mUCellFace[4].corner[1].v = mYViewCells ;
	mUCellFace[4].corner[1].d = 0. ;
	mUCellFace[4].corner[2].h = mXViewCells ;
	mUCellFace[4].corner[2].v = mYViewCells ;
	mUCellFace[4].corner[2].d = mZViewCells ;
	mUCellFace[4].corner[3].h = mXViewCells ;
	mUCellFace[4].corner[3].v = 0. ;
	mUCellFace[4].corner[3].d = mZViewCells ;
	
	mUCellFace[5].corner[0].h = mXViewCells ;
	mUCellFace[5].corner[0].v = mYViewCells ;
	mUCellFace[5].corner[0].d = 0. ;
	mUCellFace[5].corner[1].h = 0. ;
	mUCellFace[5].corner[1].v = mYViewCells ;
	mUCellFace[5].corner[1].d = 0. ;
	mUCellFace[5].corner[2].h = 0. ;
	mUCellFace[5].corner[2].v = mYViewCells ;
	mUCellFace[5].corner[2].d = mZViewCells ;
	mUCellFace[5].corner[3].h = mXViewCells ;
	mUCellFace[5].corner[3].v = mYViewCells ;
	mUCellFace[5].corner[3].d = mZViewCells ;
	
	for(j=0;j<6;j++) {
		mUCellFace[j].cx = 0. ;
		mUCellFace[j].cy = 0. ;
		mUCellFace[j].cz = 0. ;
		for(i=0;i<4;i++) {
			mUCellFace[j].cx += mUCellFace[j].corner[i].h/4 ;
			mUCellFace[j].cy += mUCellFace[j].corner[i].v/4 ;
			mUCellFace[j].cz += mUCellFace[j].corner[i].d/4 ;
		}
	}
	
	// Transform the vertices
	for(i=0;i<6;i++) {
		for (k=0;k<4;k++) {
			ax = M.m[0][0]*mUCellFace[i].corner[k].h + M.m[0][1]*mUCellFace[i].corner[k].v + M.m[0][2]*mUCellFace[i].corner[k].d  ;
			bx = M.m[1][0]*mUCellFace[i].corner[k].h + M.m[1][1]*mUCellFace[i].corner[k].v + M.m[1][2]*mUCellFace[i].corner[k].d  ;
			cx = M.m[2][0]*mUCellFace[i].corner[k].h + M.m[2][1]*mUCellFace[i].corner[k].v + M.m[2][2]*mUCellFace[i].corner[k].d  ;
			mUCellFace[i].corner[k].h = ax/mXSection.AngstrPrPixel + M.m[0][3] ;
			mUCellFace[i].corner[k].v = bx/mXSection.AngstrPrPixel + M.m[1][3];
			mUCellFace[i].corner[k].d = cx/mXSection.AngstrPrPixel + M.m[2][3];
		}
		x1 = mUCellFace[i].corner[1].h - mUCellFace[i].corner[0].h ;
		y1 = mUCellFace[i].corner[1].v - mUCellFace[i].corner[0].v ;
		z1 = mUCellFace[i].corner[1].d - mUCellFace[i].corner[0].d ;
		x2 = mUCellFace[i].corner[2].h - mUCellFace[i].corner[1].h ;
		y2 = mUCellFace[i].corner[2].v - mUCellFace[i].corner[1].v ;
		z2 = mUCellFace[i].corner[2].d - mUCellFace[i].corner[1].d ;
		mUCellFace[i].normal[0] = y1*z2-z1*y2 ;
		mUCellFace[i].normal[1] = x2*z1-x1*z2 ;
		mUCellFace[i].normal[2] = x1*y2-y1*x2 ;
	}
	
	// Translate the faces to the appropriate location
	for(i=0;i<6;i++) {
		for (k=0;k<4;k++) {
			// mUCellFace[i].corner[k].h += mDisplayRect.left ;
			// mUCellFace[i].corner[k].v -= mDisplayRect.top ;	// Because it is flipped
		}
	}
	SortFaces() ; 
}

void TCrystal::DrawAtom(SAtom *sAtom) {
	Rect theRect ;
	long 	x,y,colorIndex,rad ;

	if ((sAtom->atom).occ == 0.) 
		return ;					// No drawing of atoms with occupancy factor 0
		
	x = sAtom->location.h ;
	y = mHeight - sAtom->location.v ;
	rad = sAtom->rad ;
	colorIndex = (sAtom->atom).no - 1 ;
	
	theRect.top = y - rad ;
	theRect.left = x - rad ;
	theRect.bottom = theRect.top+2*rad ;
	theRect.right = theRect.left+2*rad ;

	
	switch(sAtom->state) {
		case kNormalState:
			colorIndex = (sAtom->atom).no % NUM_COLORS ;
			SetFgColor(gColors[gNumGrayLevels+colorIndex].pixel) ;
			FillCircle(x,y,rad) ;
			SetFgColor(gWhitePixel) ;
			FillCircle(x+rad/3,y-rad/3,rad/4) ;
			SetFgColor(gBlackPixel) ;
			DrawCircle(x,y,rad) ;
			break ;
		case kDeSelectedState:	
			colorIndex = (sAtom->atom).no % NUM_COLORS ;
			SetFgColor(gColors[gNumGrayLevels+colorIndex].pixel) ;
			FillCircle(x,y,rad) ;
			SetFgColor(gWhitePixel) ;
			FillCircle(x+rad/3,y-rad/3,rad/4) ;
			DrawCircle(x,y,rad-1) ;
			SetFgColor(gGreenPixel) ;
			DrawCircle(x,y,rad) ;
			SetFgColor(gBlackPixel) ;
			DrawCircle(x,y,rad+1) ;
			break ;
		case kSelectedState:
			colorIndex = (sAtom->atom).no % NUM_COLORS ;
			SetFgColor(gColors[gNumGrayLevels+colorIndex].pixel) ;
			PenPat() ;
			FillCircle(x,y,rad) ;
			PenNormal() ;
			SetFgColor(gWhitePixel) ;
			FillCircle(x+rad/3,y-rad/3,rad/4) ;
			DrawCircle(x,y,rad-1) ;
			SetFgColor(gRedPixel) ;
			DrawCircle(x,y,rad) ;
			SetFgColor(gBlackPixel) ;
			DrawCircle(x,y,rad+1) ;
			break ;
		default:
			break ;
	}
}

void TCrystal::Print() {
	
	GrafPtr currPort ;
	GetPort(&currPort) ;
		
	if (mDrawOutline) {
		DrawInvisibleUCellFaces() ;
	}
	
	/* Draw the Crystal  */
	Draw2DScreenBalls() ;
	
	if (mDrawOutline) {
		DrawVisibleUCellFaces() ;
	}
	
}

void TCrystal::CreateView(short na,short nb,short nc)
{
	short numUCells ;
	long numAtomsInCells,atomsToAllocate,atomsInUCell ;
	
	numUCells = na*nb*nc ;	
	atomsInUCell = mNumUnitCellAtoms ;	// We set off enough to hold space for 
									// either the basis or all the atoms
	
	
	numAtomsInCells = numUCells*atomsInUCell ;	// Total number of atoms in the
												// expanded unit cells, excluding
												// atoms at the faces added for 
												// viewing
	
	mNumFaceAtoms = CalculateNumFaceAtoms() ;
	atomsToAllocate = numAtomsInCells + numUCells*mNumFaceAtoms ;
	
	if (mScreenAtom != nil) 
		delete[] mScreenAtom ;
	mScreenAtom = new SAtom[atomsToAllocate] ;
	Assert(mScreenAtom) ;
		
	mXViewCells = na ;
	mYViewCells = nb ;
	mZViewCells = nc ;
	
}

void	
TCrystal::SetMagnification(float magnification) { 

	mMagnification = magnification ;
	SetScale() ;
	UpdateViewAtoms() ; // CreateUnitCellViewAtoms() ; 
	DrawCrystal() ; 
}
		
void	
TCrystal::ChangeMagnification(float factor) { 
	float	zoomfactor  = 1./factor ;

	mXSection.AngstrPrPixel *= factor ;
	mMagnification /= factor ;
	SetCenter() ; 
	
	UpdateViewAtoms() ; // CreateUnitCellViewAtoms() ; 
	DrawCrystal() ; 
	if (mSelection != nil) {
		mSelection->Zoom(zoomfactor,0,0) ;
		mSelection->Update(&mDisplayRect) ;
	}			

}

long TCrystal::CalculateNumFaceAtoms(void)
{
	register 	long i,j ;
	register	float x,y,z ;
	 
	j = mNumUnitCellAtoms ;
	
	for(i=0;i<mNumUnitCellAtoms;i++) {
		x = (mAtomPtr+i)->xc ;
		y = (mAtomPtr+i)->yc ;
		z = (mAtomPtr+i)->zc ;
		if (fabs(x) < 0.001) x = 0. ;
		if (fabs(y) < 0.001) y = 0. ;
		if (fabs(z) < 0.001) z = 0. ; 
		
		if (x == 0. || y == 0. || z == 0.) {
			if (x==0. && y != 0. && z != 0.) {
				j++ ;
			}
			else if (x==0. && y==0. && z!=0.) {
				j += 3 ;
			}
			else if (x==0. && y==0. && z==0.) {
				j += 7 ;
			}
			else if (y==0. && x!=0. &&z!=0.) {
				j++ ;
			}
			else if (y==0. && x!=0. && z==0.) {
				 j += 3 ;
			}
			else if (z==0. && x!=0. && y!=0.) {
				j++ ;
			}
			else if (z==0. && x==0. && y!=0.) {
				j += 3 ;
			}
		}
	}
	return (j-mNumUnitCellAtoms) ;
}



void TCrystal::UpdateViewAtoms() {
	float 	x,y,z,xt,yt,zt ;
	short	x0,y0,z0 ;
	float 	rad,iorigox,iorigoy,iorigoz,scale ;
	float	txx,txy,txz,tyx,tyy,tyz,tzx,tzy,tzz ;
	long 	i ;

	scale	=	1/mXSection.AngstrPrPixel ;	
	
	txx = mXyzToScreenMatrix.m[0][0] ;
	txy = mXyzToScreenMatrix.m[0][1] ;
	txz = mXyzToScreenMatrix.m[0][2] ;
	
	tyx = mXyzToScreenMatrix.m[1][0] ;
	tyy = mXyzToScreenMatrix.m[1][1] ;
	tyz = mXyzToScreenMatrix.m[1][2] ;
	
	tzx = mXyzToScreenMatrix.m[2][0] ;
	tzy = mXyzToScreenMatrix.m[2][1] ;
	tzz = mXyzToScreenMatrix.m[2][2] ;
	
	SetUCellFaces() ;
	iorigox	=	mXyzToScreenMatrix.m[0][3] ;
	iorigoy =	mXyzToScreenMatrix.m[1][3] ;
	iorigoz =	mXyzToScreenMatrix.m[2][3] ;
	
	
	for(i=0;i < mSnumAtoms;i++) {
		x = (mScreenAtom+i)->atom.xc ;
		y = (mScreenAtom+i)->atom.yc ; ;
		z = (mScreenAtom+i)->atom.zc ; ;
		xt = txx*x + txy*y + txz*z ;
		yt = tyx*x + tyy*y + tyz*z ;
		zt = tzx*x + tzy*y + tzz*z ;
		(mScreenAtom+i)->position.h = xt ;
		(mScreenAtom+i)->position.v = yt ;
		(mScreenAtom+i)->position.d = zt ;
		
		xt = scale*xt + iorigox ;
		yt = scale*yt + iorigoy ;
		zt = scale*zt + iorigoz ;
		rad = mAtomScaleFactor*scale*(mScreenAtom+i)->atom.rad ;
		
		x0 = near(xt) ;
		y0 = near(yt) ;
		z0 = near(zt) ;
		(mScreenAtom+i)->location.h = x0 ;
		(mScreenAtom+i)->location.v = y0 ;
		(mScreenAtom+i)->location.d = z0 ;
		(mScreenAtom+i)->rad = near(rad) ;
	}
	SortScreenAtoms() ;

}

void TCrystal::SetAtomTypes() {

	long	i,j,numTypes,equal ;
	float	debye[100] ;
	char	*symb[100] ;
	long	type[100] ;

	if (mNumBasisAtoms == 0) return ;
	
	numTypes 	= 1 ;
	symb[0] 	= (mBasisPtr)->name ;
	debye[0] 	= (mBasisPtr)->dw ;
	type[0] 	= 1 ;
	
	(mBasisPtr)->type 	= 1 ;
	for (j=1;j < mNumBasisAtoms;j++) {
		equal = false ;
		for (i=0;i<numTypes;i++) {
			if ((mBasisPtr+j)->name == symb[i] &&
			    	(mBasisPtr+j)->dw == debye[i]) {
				equal = true ;
				(mBasisPtr+j)->type = type[i] ;
				break ;
			}
		}
		if (!equal) {				 
			symb[numTypes] 	= (mBasisPtr+j)->name ;
			debye[numTypes] = (mBasisPtr+j)->dw ;
			type[numTypes] = numTypes+1 ;
			(mBasisPtr+j)->type = type[numTypes] ;
			numTypes++ ;
		}
	}

	SortAtomsByType() ;
	
	mNumTypes = numTypes ;
}

void TCrystal::SortAtomsByType() {
	long gap ;
	register long i,j ;
	TAtom	temp ;
	
	for (gap=mNumBasisAtoms/2;gap>0;gap/=2)
		for(i=gap;i<mNumBasisAtoms;i++)
			for(j=i-gap;j>=0;j-=gap) {
					if ((mBasisPtr+j)->type <= (mBasisPtr+j+gap)->type)
						break ;
					temp = *(mBasisPtr+j) ;
					*(mBasisPtr+j) = *(mBasisPtr+j+gap) ;
					*(mBasisPtr+j+gap) = temp ;
			}

}

void		
TCrystal::ShowSelectedAtom(TAtom *atom) {
	// Find the atom(s) in the view that has this coordinate in the unit cell
	// If there are any, highlight them
	float 	x,y,z ;
	long		l,m,n ;
	float 	scale,TZ0,fn,fm,fl ;
	TAtom	theAtom ;
	
	Boolean removeOldHighlight = true ;
	theAtom.color = atom->color ;
	theAtom.rad = atom->rad ;
	scale = 1./mXSection.AngstrPrPixel ;	
	TZ0		= 1.0 ;
	
	// We need to set the proper port for our drawing
	GrafPtr		originalPort ;
	GetPort(&originalPort) ;
	SetPort((GrafPtr)mDocWindow) ;

	fn = (float)mXViewCells ;
	fm = (float)mYViewCells ;
	fl = (float)mZViewCells ;

	for(l=0;l<mZViewCells;l++) {
		for(m=0;m<mYViewCells;m++) {
			for(n=0;n<mXViewCells;n++) {
				/* Transform the coordinates of the atom(s) in the cell to screen coordinates */
				x = atom->xc + n ;
				y = atom->yc + m ;
				z = atom->zc + l ;
				if (fabs(x) < 0.0001) x = 0. ;
				if (fabs(y) < 0.0001) y = 0. ;
				if (fabs(z) < 0.0001) z = 0. ;	
				
				if (removeOldHighlight) {
					DrawCrystal() ;
					removeOldHighlight = false ;
				}
				SelectThisAtom(x,y,z,TZ0,scale,&theAtom) ;
				if (x == 0. || y == 0. || z == 0.) {
					if (x==0. && y != 0. && z != 0.) {
						SelectThisAtom(fn,y,z,TZ0,scale,&theAtom) ;
					}
					else if (x==0. && y==0. && z!=0.) {
						SelectThisAtom(fn,y,z,TZ0,scale,&theAtom) ;
						SelectThisAtom(x,fm,z,TZ0,scale,&theAtom) ;
						SelectThisAtom(fn,fm,z,TZ0,scale,&theAtom) ;
					}
					else if (x==0. && y==0. && z==0.) {
						SelectThisAtom(fn,y,z,TZ0,scale,&theAtom) ;
						SelectThisAtom(fn,fm,z,TZ0,scale,&theAtom) ;
						SelectThisAtom(x,fm,z,TZ0,scale,&theAtom) ;
						SelectThisAtom(x,y,fl,TZ0,scale,&theAtom) ;
						SelectThisAtom(fn,y,fl,TZ0,scale,&theAtom) ;
						SelectThisAtom(x,fm,fl,TZ0,scale,&theAtom) ;
						SelectThisAtom(fn,fm,fl,TZ0,scale,&theAtom) ;
					}
					else if (y==0. && x!=0. &&z!=0.) {
						SelectThisAtom(x,fm,z,TZ0,scale,&theAtom) ;
					}
					else if (y==0. && x!=0. && z==0.) {
						SelectThisAtom(x,fm,z,TZ0,scale,&theAtom) ;
						SelectThisAtom(x,fm,fl,TZ0,scale,&theAtom) ;
						SelectThisAtom(x,y,fl,TZ0,scale,&theAtom) ;
					}
					else if (z==0. && x!=0. && y!=0.) {
						SelectThisAtom(x,y,fl,TZ0,scale,&theAtom) ;
					}
					else if (z==0. && x==0. && y!=0.) {
						SelectThisAtom(fn,y,z,TZ0,scale,&theAtom) ;
						SelectThisAtom(x,y,fl,TZ0,scale,&theAtom) ;
						SelectThisAtom(fn,y,fl,TZ0,scale,&theAtom) ;
					}
				}

			}
		}
	}
	SetPort(originalPort) ;
}

void
TCrystal::SelectThisAtom(float x,float y,float z,float TZ0,float scale,TAtom *theAtom) {
	float 	xt,yt,zt ;
	short	x0,y0 ;
	float 	rad ;
	Point	where ;
		
	xt = mXyzToScreenMatrix.m[0][0]*x + mXyzToScreenMatrix.m[0][1]*y + mXyzToScreenMatrix.m[0][2]*z ;
	yt = mXyzToScreenMatrix.m[1][0]*x + mXyzToScreenMatrix.m[1][1]*y + mXyzToScreenMatrix.m[1][2]*z ;
	zt = mXyzToScreenMatrix.m[2][0]*x + mXyzToScreenMatrix.m[2][1]*y + mXyzToScreenMatrix.m[2][2]*z ;
	xt = scale*xt+mXyzToScreenMatrix.m[0][3] ;
	yt = scale*yt+mXyzToScreenMatrix.m[1][3] ;
	zt = scale*zt+mXyzToScreenMatrix.m[2][3] ;
	rad = theAtom->rad ;
	
	theAtom->xc = xt ;
	theAtom->yc = yt ;
	theAtom->zc = zt ;
	theAtom->rad = rad ;
	x0 = near(xt) ;
	y0 = near(yt) ;
	where.x = x0 ;
	where.y = mHeight - y0 ;
	HiliteAtom(theAtom, where);
}

SAtom	*TCrystal::GetAtom(Point& mouseLoc)  {
	long i,dsq,dx,dy,radsq,rad,ndx,num ;
	Boolean found ;
	Point	pt ;
	SAtom *whichAtom ;

	ndx = 0 ;
	whichAtom = nil ;
	
	// Position as it is referred to for the atom positions
	pt.x = mouseLoc.x -  mDisplayRect.left ;
	pt.y = mHeight - (mouseLoc.y - mDisplayRect.top) ;

	num = 1 ;
	for (i=mSnumAtoms-1;i >= 0;i--) {
		dx = pt.x-(mScreenAtom+i)->location.h ;
		dy = pt.y-(mScreenAtom+i)->location.v ;
		dsq = dx*dx+dy*dy ;
		rad = (mScreenAtom+i)->rad ;
		radsq = rad*rad ;
		if (dsq < radsq) {
			found = true ;
			ndx = i ;
			mouseLoc.x = (mScreenAtom+i)->location.h + mDisplayRect.left ;
			mouseLoc.y = mHeight - (mScreenAtom+i)->location.v + mDisplayRect.top ;
			whichAtom = (mScreenAtom+i) ;
			break ;
		}
	}
	return (whichAtom) ; 

}

Boolean TCrystal::FindAtom(Point& mouseLoc,SAtom *whichAtom) {
	long i,dsq,dx,dy,radsq,rad,ndx,num ;
	Boolean found ;
	Point	pt ;

	found = false ;
	ndx = 0 ;
	
	// Position as it is referred to for the atom positions
	pt.x = mouseLoc.x -  mDisplayRect.left ;
	pt.y = mHeight - (mouseLoc.y - mDisplayRect.top) ;

	num = 1 ;
	for (i=mSnumAtoms-1;i >= 0;i--) {
		dx = pt.x-(mScreenAtom+i)->location.h ;
		dy = pt.y-(mScreenAtom+i)->location.v ;
		dsq = dx*dx+dy*dy ;
		rad = (mScreenAtom+i)->rad ;
		radsq = rad*rad ;
		if (dsq < radsq) {
			found = true ;
			ndx = i ;
			mouseLoc.x = (mScreenAtom+i)->location.h + mDisplayRect.left ;
			mouseLoc.y = mHeight - (mScreenAtom+i)->location.v + mDisplayRect.top ;
			*whichAtom = *(mScreenAtom+i) ;
			break ;
		}
	}
	return found ;
}

// -----------------------------------------------------------
// void TCrystal::Inquire(Point *mouseLoc)
// Report back information on the clicked atom	
//
void TCrystal::Inquire(const Point& loc) {
	float 	x,y,z,xt,yt,zt ;
	char 	text[80] ;
	Point	where ;
	where = loc ;
	TMatrix iXyzToScreenMatrix ;
	SAtom	whichAtom ;
	
	if (!FindAtom(where,&whichAtom)) {
		return ;
	}
	
	DrawCrystal(); // to remove previous hilite
	HiliteAtom(&whichAtom.atom, where);
	
	//EncodeAtomName(name,whichAtom.atom.name,whichAtom.atom.tag) ;

	x = whichAtom.position.h ;
	y = whichAtom.position.v ;
	z = whichAtom.position.d ;
	
	InvertMatrix(mXyzToScreenMatrix,iXyzToScreenMatrix) ;
	xt = iXyzToScreenMatrix.m[0][0]*x + iXyzToScreenMatrix.m[0][1]*y + iXyzToScreenMatrix.m[0][2]*z ;
	yt = iXyzToScreenMatrix.m[1][0]*x + iXyzToScreenMatrix.m[1][1]*y + iXyzToScreenMatrix.m[1][2]*z ;
	zt = iXyzToScreenMatrix.m[2][0]*x + iXyzToScreenMatrix.m[2][1]*y + iXyzToScreenMatrix.m[2][2]*z ;

	//sprintf(text,"%s: at (%.3f,%.3f,%.3f)",name,xt,yt,zt) ;
	sprintf(text,"%s: at (%.3f,%.3f,%.3f)",whichAtom.atom.name,xt,yt,zt) ;
	WriteInfo(text) ;
}

void TCrystal::DrawCrystal(TVector &zone) {
	mMagnification = 1. ;
	
	mZone = zone ;			
	SetTransformationMatrix(zone) ;
	
	CreateUnitCellViewAtoms() ;
	DrawCrystal() ;

}
void TCrystal::HiliteAtom(TAtom *theAtom, Point AtomPt) 
{
	
	float scale	=	1/mXSection.AngstrPrPixel ;	
	float atomrad = theAtom->rad * scale;
	if(atomrad <= 0.0)
	{
		short ndx = (mAtomPtr+theAtom->no)->no-1 ;
		atomrad = covrad[ndx]*scale ; 
	}
	long Rad = near(mAtomScaleFactor*atomrad) + 2;
	
	RgnHandle oldClip = NewRgn() ;
	RectRgn(oldClip,&mDocWindow->GetPortRect()) ;
	//GetClip(oldClip) ;
	
	InsetRect(&mDisplayRect,1,1) ;
	RgnHandle newClip = NewRgn() ;
	RectRgn(newClip,&mDisplayRect) ;
	SetClip(newClip) ;	
	InsetRect(&mDisplayRect,-1,-1) ;
	
	RGBColor Hcolor = {56000, 6000, 1500}; // nice red color
	::RGBForeColor(&Hcolor);
	DrawCircle(AtomPt.x+1, AtomPt.y+1, Rad);
	RGBColor black = {0, 0, 0}; 
	::RGBForeColor(&black);
	
	SetClip(oldClip) ;
	DisposeRgn(oldClip) ;
	DisposeRgn(newClip) ;
}


// -----------------------------------------------------------
// void TCrystal::measureDistance(Point *mouseLoc)
// Measure distance between two atoms that are marked
// consecutively	
//
void TCrystal::MeasureDistance(const Point& loc) {
	float 	distance,x1,y1,z1,x2,y2,z2,projDist,dx,dy,dz ;
	SAtom	whichAtom ;
	char 	text[80] ;
	Point	where ;
	where = loc ;

	switch(mWAtom) {
		case kFirstAtom:
			if (!FindAtom(where,&whichAtom)) {
				//SysBeep(1) ;
				return ;
			}
			mWAtom = kSecondAtom ;
			mFirstAtom = whichAtom ;
			SetClip(mDisplayRect) ;
			DrawCross(where.x,where.y) ;
			mFirstPt = where ;
			SetClip(gCurrentPort->GetPortRect()) ;
			return ;
			break ;
		case kSecondAtom:
			if (!FindAtom(where,&whichAtom)) {
				//SysBeep(1) ;
				return ;
			}
			mWAtom = kFirstAtom ;
			mSecondAtom = whichAtom ;
			SetClip(mDisplayRect) ;
			DrawCross(where.x,where.y) ;
			mSecondPt = where ;
			break ;
		default:
			break ;
	}
    
	DrawLine(mFirstPt.x,mFirstPt.y,mSecondPt.x,mSecondPt.y) ;
	SetClip(gCurrentPort->GetPortRect()) ;

	//EncodeAtomName(name1,mFirstAtom.atom.name,mFirstAtom.atom.tag) ;
	//EncodeAtomName(name2,mSecondAtom.atom.name,mSecondAtom.atom.tag) ;
	x1 = mFirstAtom.position.h ;
	y1 = mFirstAtom.position.v ;
	z1 = mFirstAtom.position.d ;
	x2 = mSecondAtom.position.h ;
	y2 = mSecondAtom.position.v ;
	z2 = mSecondAtom.position.d ;
	dx = x2-x1 ;
	dy = y2-y1 ;
	dz = z2-z1 ;
	distance = sqrt(dx*dx+dy*dy+dz*dz) ;
	projDist = sqrt(dx*dx+dy*dy) ;
	sprintf(text,"Dist. between %s' and '%s' :Real: %.2fA,  Proj : %.2fA",mFirstAtom.atom.name,
    		mSecondAtom.atom.name,distance,projDist) ;
	WriteInfo(text) ;
}

void TCrystal::MeasureAngle(const Point &loc) {
	float 	x1,y1,z1,x2,y2,z2,x3,y3,z3,dx1,dy1,dx2,dy2,dz1,dz2,
			length1,length2,csth,theta,projTheta ;
	SAtom 	whichAtom ;
	char 	text[80] ;
	Point	endPoint ;
	Point	where ;
	where = loc ;

	switch(mAnglePt) {
		case kNoAnglePt:
			if (!FindAtom(where,&whichAtom)) {
				//SysBeep(1)  ;
				return ;
			}
			mAnglePt = kfirstEndPoint ;
			SetClip(mDisplayRect) ;
			mFirstAtom = whichAtom ;
			DrawCross(where.x,where.y) ;
			SetClip(gCurrentPort->GetPortRect()) ;
			mFirstPt = where ;
			return ;
			break ;
		case kfirstEndPoint:
			if (!FindAtom(where,&whichAtom)) {
				//SysBeep(1)  ;
				return ;
			}
			mAnglePt = kVertex ;
			mSecondAtom = whichAtom ;
			SetClip(mDisplayRect) ;
			DrawCross(where.x,where.y) ;
			mSecondPt = where ;
			DrawLine(mFirstPt.x,mFirstPt.y,mSecondPt.x,mSecondPt.y) ;
			SetClip(gCurrentPort->GetPortRect()) ;
			return ;
			break ;
		case kVertex:
			if (!FindAtom(where,&whichAtom)) {
				//SysBeep(1)  ;
				return ;
			}
			endPoint = where ;
			SetClip(mDisplayRect) ;
			DrawCross(where.x,where.y) ;
			DrawLine(mSecondPt.x,mSecondPt.y,endPoint.x,endPoint.y) ;
			SetClip(gCurrentPort->GetPortRect()) ;
			mAnglePt = kNoAnglePt ;
			break ;
	}
	
	x1 = mFirstAtom.position.h ;
	y1 = mFirstAtom.position.v ;
	z1 = mFirstAtom.position.d ;
	x2 = mSecondAtom.position.h ;
	y2 = mSecondAtom.position.v ;
	z2 = mSecondAtom.position.d ;
	x3 = whichAtom.position.h ;
	y3 = whichAtom.position.v ;
	z3 = whichAtom.position.d ;
	
	dx1 = x1-x2 ;
	dy1 = y1-y2 ;
	dx2 = x3-x2 ;
	dy2 = y3-y2 ;
	dz1 = z1-z2 ;
	dz2 = z3-z2 ;

	
	length1 = sqrt(dx1*dx1+dy1*dy1+dz1*dz1) ;
	length2 = sqrt(dx2*dx2+dy2*dy2+dz2*dz2) ;
	csth = (dx1*dx2+dy1*dy2+dz1*dz2)/(length1*length2) ;
	theta = acos(csth)*180./3.1415926 ;
	length1 = sqrt(dx1*dx1+dy1*dy1) ;
	length2 = sqrt(dx2*dx2+dy2*dy2) ;
	csth = (dx1*dx2+dy1*dy2)/(length1*length2) ;
	projTheta = acos(csth)*180./3.1415926 ;
	sprintf(text,"Measured Angle:  %.2f Deg.",theta) ;
	WriteInfo(text) ;
}



void TCrystal::Rotate() {
	float 	dAngle = 0.75,angle ;
	Point 	newLoc,oldLoc ;
	TMatrix	t,tr ;
	long	dx,dy ;
	TMatrix	temp,tInv ; ;
	TVector	zone ;
	TVector xyzZone(0.,0.,1.,0.) ; 
	
	GetMouse(&oldLoc) ;
	while (Button()) {
		GetMouse(&newLoc) ;
		if (newLoc.x != oldLoc.x || newLoc.y != oldLoc.y) {
			dx = newLoc.x - oldLoc.x ;
			dy = newLoc.y - oldLoc.y ;
			if (ShiftKeyDown()) {	/* Rotation about the z-axis	*/
				angle = (dx-dy)*dAngle ;
				if (angle != 0.) {
					DegreeToRad(&angle) ;
					SetDisRotMatrix(zRotAxis,angle) ; 
				}
			}
			else { 	// Rotation about the other two axis
					// Always do mA 'horizontal' rotation first
				angle = (newLoc.x - oldLoc.x)*dAngle ;
				if (angle != 0.) {
					DegreeToRad(&angle) ;
					SetDisRotMatrix(yRotAxis,angle) ; 
				}
				angle = (newLoc.y - oldLoc.y)*dAngle ;
				if (angle != 0.) {
					DegreeToRad(&angle) ;
					SetDisRotMatrix(xRotAxis,angle) ; 
				}
			}
			// Draw the axis
			oldLoc = newLoc ;
			temp = mTmatrix*mRotMatrix ;				// Current transformation matrix of	
			TransposeMatrix(temp,mXyzToScreenMatrix) ;	// between mA,mB,mC and screen system
			InvertMatrix(mXyzToScreenMatrix,tInv) ;
			zone = tInv*xyzZone ;
			zone *= LengthOfRealVector(zone) ;
			SetZone(zone) ;
			mZone = zone ;
			mDocWindow->UpdateZoneAxis(zone) ;
			SetCenter() ;
			UpdateViewAtoms() ;
			DrawCrystal() ;
			
		}
	}
	DrawCrystal() ; 
}


void 
TCrystal::RecipToCartVector(TVector &reciprocalVector,TVector &cartesianVector) {
	TMatrix sInverse,tt,t ;

	InvertMatrix(mSmatrix,sInverse) ;
	tt = sInverse*mTmatrix ;
	TransposeMatrix(tt,t) ;
	cartesianVector = t*reciprocalVector ;
}

MBFITManualBeamSearchRec::MBFITManualBeamSearchRec(short numExact, short numBethe) {
		hklExact=nil ; hklBethe = nil ; 
		if (numExact > 0) hklExact = new LVector[numExact] ;
		if (num_bethe > 0) hklBethe = new LVector[num_bethe] ;
}

MBFITManualBeamSearchRec::~MBFITManualBeamSearchRec() {
		if (hklExact != nil) delete[] hklExact ;
		if (hklBethe != nil) delete[] hklBethe ;
}


TBasisAtomDisplayList::TBasisAtomDisplayList(Rect	*theListRect) 
	: TList(theListRect,sizeof(TAtom)) {
		mHasColLabels = true ;
		mSelectedAtom = 1 ;
}

void 	TBasisAtomDisplayList::DrawCellValue(ListCellT &inCell,Rect &cellFrame) {
	char str[256] ;
	TAtom   theAtom ;
	short	width = (mCanvasRect.right-mCanvasRect.left)/6 ;
	
	TextSize(0);
	FetchCellData(inCell,&theAtom) ;

	MoveTo(cellFrame.left + 8, cellFrame.bottom - 4);
	sprintf(str,"%3d    %s",inCell.row,theAtom.name) ;
	DrawString(str) ;

	MoveTo(cellFrame.left + 8 + width , cellFrame.bottom - 4);
	sprintf(str,"%.4f",theAtom.xc) ;
	DrawString(str) ;

	MoveTo(cellFrame.left + 8 + 2*width , cellFrame.bottom - 4);
	sprintf(str,"%.4f",theAtom.yc) ;
	DrawString(str) ;

	MoveTo(cellFrame.left + 8 + 3*width , cellFrame.bottom - 4);
	sprintf(str,"%.4f",theAtom.zc) ;
	DrawString(str) ;

	MoveTo(cellFrame.left + 8 + 4*width , cellFrame.bottom - 4);
	sprintf(str,"%.2f",theAtom.dw) ;
	DrawString(str) ;

	MoveTo(cellFrame.left + 8 + 5*width , cellFrame.bottom - 4);
	sprintf(str,"%.2f",theAtom.occ) ;
	DrawString(str) ;
}


void TBasisAtomDisplayList::ClickCell(
	ListCellT &inCell) 
{
	TAtom   theAtom ;
	TEditCrystalDialog *theEditCrystalDialog = (TEditCrystalDialog *)GetParentWindow() ;
	if (inCell.row > 0 && inCell.row <= mRows) {
		mSelectedCell = inCell ;
		// Get the atom associated with this position and display parameters
		FetchCellData(inCell,&theAtom) ;
		theEditCrystalDialog->ShowAtom(theAtom) ;
	}
	else {
		mSelectedCell.row = mSelectedCell.col = 0 ;
	}
	Refresh() ;
}

void 	TBasisAtomDisplayList::DrawColLabel(short colNumber) {
	short	width = (mCanvasRect.right-mCanvasRect.left)/6 ;
	
	Rect	cellFrame;
	
	if (mHasRowLabels)
		SetRect(&cellFrame,colNumber*mColWidth,0,(colNumber+1)*mColWidth,mRowHeight) ;
	else
		SetRect(&cellFrame,(colNumber-1)*mColWidth,0,colNumber*mColWidth,mRowHeight) ;
		
	if (colNumber != 0) 
		::OffsetRect(&cellFrame,near(mListOrigin.x),0) ;
	InsetRect(&cellFrame,1,1) ;
	SetFgColor(gGrayPixel) ;
	PaintRect(&cellFrame) ;
	SetFgColor(gDarkerGray) ;
	MoveTo(cellFrame.left,cellFrame.bottom) ;
	LineTo(cellFrame.left,cellFrame.top) ;
	LineTo(cellFrame.right,cellFrame.top) ;
	SetFgColor(gBlackPixel) ;
	MoveTo(cellFrame.right,cellFrame.top) ;
	LineTo(cellFrame.right,cellFrame.bottom) ;
	LineTo(cellFrame.left,cellFrame.bottom) ;
	InsetRect(&cellFrame,-1,-1) ;
	TextSize(0);
	
	MoveTo(cellFrame.left + 12, cellFrame.bottom - 4);
	DrawString("Symb") ;

	MoveTo(cellFrame.left + width + 14, cellFrame.bottom - 4);
	DrawString("X") ;

	MoveTo(cellFrame.left + 2*width + 14, cellFrame.bottom - 4);
	DrawString("Y") ;
	
	MoveTo(cellFrame.left + 3*width + 14, cellFrame.bottom - 4);
	DrawString("Z") ;

	MoveTo(cellFrame.left + 4*width + 8, cellFrame.bottom - 4);
	DrawString("DW") ;

	MoveTo(cellFrame.left + 5*width + 4, cellFrame.bottom - 4);
	DrawString("OCC") ;
}


TUnitCellAtomDisplayList::TUnitCellAtomDisplayList(Rect	*theListRect) 
	: TList(theListRect,sizeof(TAtom)) {
		mHasColLabels = true ;
}

void 	TUnitCellAtomDisplayList::DrawCellValue(ListCellT &inCell,Rect &cellFrame) {
	char str[256] ;
	TAtom   theAtom ;
	short	width = (mCanvasRect.right-mCanvasRect.left)/6 ;
	
	TextSize(0);
	FetchCellData(inCell,&theAtom) ;

	MoveTo(cellFrame.left + 8, cellFrame.bottom - 4);
	sprintf(str,"%3d    %s",inCell.row,theAtom.name) ;
	DrawString(str) ;

	MoveTo(cellFrame.left + 8 + width , cellFrame.bottom - 4);
	sprintf(str,"%.4f",theAtom.xc) ;
	DrawString(str) ;

	MoveTo(cellFrame.left + 8 + 2*width , cellFrame.bottom - 4);
	sprintf(str,"%.4f",theAtom.yc) ;
	DrawString(str) ;

	MoveTo(cellFrame.left + 8 + 3*width , cellFrame.bottom - 4);
	sprintf(str,"%.4f",theAtom.zc) ;
	DrawString(str) ;

	MoveTo(cellFrame.left + 8 + 4*width , cellFrame.bottom - 4);
	sprintf(str,"%.2f",theAtom.dw) ;
	DrawString(str) ;

	MoveTo(cellFrame.left + 8 + 5*width , cellFrame.bottom - 4);
	sprintf(str,"%.2f",theAtom.occ) ;
	DrawString(str) ;
}


void 	TUnitCellAtomDisplayList::DrawColLabel(short colNumber) {
	short	width = (mCanvasRect.right-mCanvasRect.left)/6 ;
	
	Rect	cellFrame;
	
	if (mHasRowLabels)
		SetRect(&cellFrame,colNumber*mColWidth,0,(colNumber+1)*mColWidth,mRowHeight) ;
	else
		SetRect(&cellFrame,(colNumber-1)*mColWidth,0,colNumber*mColWidth,mRowHeight) ;
		
	if (colNumber != 0) 
		::OffsetRect(&cellFrame,near(mListOrigin.x),0) ;
	InsetRect(&cellFrame,1,1) ;
	SetFgColor(gGrayPixel) ;
	PaintRect(&cellFrame) ;
	SetFgColor(gDarkerGray) ;
	MoveTo(cellFrame.left,cellFrame.bottom) ;
	LineTo(cellFrame.left,cellFrame.top) ;
	LineTo(cellFrame.right,cellFrame.top) ;
	SetFgColor(gBlackPixel) ;
	MoveTo(cellFrame.right,cellFrame.top) ;
	LineTo(cellFrame.right,cellFrame.bottom) ;
	LineTo(cellFrame.left,cellFrame.bottom) ;
	InsetRect(&cellFrame,-1,-1) ;
	TextSize(0);
	
	MoveTo(cellFrame.left + 12, cellFrame.bottom - 4);
	DrawString("Symb") ;

	MoveTo(cellFrame.left + width + 14, cellFrame.bottom - 4);
	DrawString("X") ;

	MoveTo(cellFrame.left + 2*width + 14, cellFrame.bottom - 4);
	DrawString("Y") ;
	
	MoveTo(cellFrame.left + 3*width + 14, cellFrame.bottom - 4);
	DrawString("Z") ;

	MoveTo(cellFrame.left + 4*width + 8, cellFrame.bottom - 4);
	DrawString("DW") ;

	MoveTo(cellFrame.left + 5*width + 4, cellFrame.bottom - 4);
	DrawString("OCC") ;
}


TSymmopDisplayList::TSymmopDisplayList(Rect	*theListRect) 
	: TList(theListRect,64) {
		mHasColLabels = false ;
}

void 	TSymmopDisplayList::DrawCellValue(ListCellT &inCell,Rect &cellFrame) {
	char	str[256] ;
	char	symb[64] ;
	
	TextSize(0);
	FetchCellData(inCell,symb) ;
	
	sprintf(str,"%3d    %s",inCell.row,symb) ;

	MoveTo(cellFrame.left + 8, cellFrame.bottom - 4);
	DrawString(str) ;
}

TEditCrystalDialog::TEditCrystalDialog(TCrystal *theCrystal) 
	: TDialog(200,200,620,540,"Crystal Parameters") {
	
	mCrystal = new TCrystal(theCrystal) ;
	mSourceCrystal = theCrystal ;
	
	mCrystalIsModified = false ;
	
	Rect	theRect,theRect1,theRect2,theRect3 ;
	char	text[256] ;
	
	long	width = mPortRect.right - mPortRect.left ;
	long	height = mPortRect.bottom - mPortRect.top ;
	
	TGroupBox   *theGroupBox = new TGroupBox(16,16,300,16+100,"Unit Cell Parameters") ;
	AddObject(theGroupBox) ;
	theRect1 = theRect = theGroupBox->GetRect() ;

	theRect.right = (theRect1.left + theRect1.right)/2 ;
	theRect.left = theRect.right - 60 ;
	theRect.top = theRect1.top + 10 ;
	theRect.bottom = theRect.top+20 ;
	
	float	alpha = mCrystal->mAlpha ;
	float	beta = mCrystal->mBeta ;
	float	gamma = mCrystal->mGamma ;
	
	RadToDegree(&alpha) ;
	RadToDegree(&beta) ;
	RadToDegree(&gamma) ;
	
	// Define and set the controls 
	sprintf(text,"%.4f",mCrystal->mA) ;
	mTextA = new TEditText(&theRect,"A - Axis [Ang]",text);	
	OffsetRect(&theRect,0,30) ;

	sprintf(text,"%.4f",mCrystal->mB) ;
	mTextB = new TEditText(&theRect,"B - Axis [Ang]",text);	
	OffsetRect(&theRect,0,30) ;

	sprintf(text,"%.4f",mCrystal->mC) ;
	mTextC = new TEditText(&theRect,"C - Axis [Ang]",text);	
	
	theRect.right = theRect1.right - 16 ;
	theRect.left = theRect.right - 60 ;
	theRect.top = theRect1.top + 10 ;
	theRect.bottom = theRect.top+20 ;

	sprintf(text,"%.4f",alpha) ;
	mTextAlpha = new TEditText(&theRect,"Alpha [Deg]",text);	
	OffsetRect(&theRect,0,30) ;

	sprintf(text,"%.4f",beta) ;
	mTextBeta = new TEditText(&theRect,"Beta [Deg]",text);	
	OffsetRect(&theRect,0,30) ;

	sprintf(text,"%.4f",gamma) ;
	mTextGamma = new TEditText(&theRect,"Gamma [Deg]",text);	
	OffsetRect(&theRect,0,30) ;

	theRect.right = theRect1.right + 160 ;
	theRect.left = theRect.right - 60 ;
	theRect.top = theRect1.top + 10 ;
	theRect.bottom = theRect.top+20 ;

	sprintf(text,"%d",mCrystal->mSpacegroup) ;
	mTextSpacegroup = new TEditText(&theRect,"Space Group",text);	

	mApplySpacegroup = new TButton(width-16-120,theRect.top,width-16,theRect.bottom,"Apply Spacegroup") ;
	AddObject(mApplySpacegroup) ;
	
	TStaticText	*theText ;

	theRect1 = theRect = theGroupBox->GetRect() ;
	OffsetRect(&theRect,0,(theRect1.bottom-theRect1.top)+20) ;
	theRect1.bottom += 60 ;
	theRect.bottom += 60 ;
	mBasisDisplayList = new TBasisAtomDisplayList(&theRect) ;
	AddObject(mBasisDisplayList) ;
	
	theRect = mBasisDisplayList->GetRect() ;
	theText = new TStaticText(theRect.left,theRect.top-20,theRect.left+120,theRect.top,"Atoms in Motif") ;
	AddObject(theText) ;
	
	OffsetRect(&theRect,0,(theRect1.bottom-theRect1.top)+20) ;
	mAtomDisplayList = new TUnitCellAtomDisplayList(&theRect) ;
	AddObject(mAtomDisplayList) ;
	theRect = mAtomDisplayList->GetRect() ;
	theText = new TStaticText(theRect.left,theRect.top-20,theRect.left+120,theRect.top,"Atoms in Unit Cell") ;
	AddObject(theText) ;
	theRect = mBasisDisplayList->GetRect() ;
	
	theRect = mAtomDisplayList->GetRect() ;
	OffsetRect(&theRect,(theRect1.right-theRect1.left)+16,0) ;
	mSymmopDisplayList = new TSymmopDisplayList(&theRect) ;
	AddObject(mSymmopDisplayList) ;
	theRect = mSymmopDisplayList->GetRect() ;
	theText = new TStaticText(theRect.left,theRect.top-20,theRect.left+120,theRect.top,"Symmetry Operators") ;
	AddObject(theText) ;

	theRect1 = mBasisDisplayList->GetRect() ;
	theRect.left = theRect1.right + 16 ;
	theRect.right = theRect.left + 50 ;
	theRect.top = theRect1.top ;
	theRect.bottom = theRect.top + 20 ;
	mTextAtomSymbol = new TEditText(&theRect,"","Cu") ;	
	
	OffsetRect(&theRect,0,30) ;
	mTextAtomX = new TEditText(&theRect,"","0.0000") ;	
	
	OffsetRect(&theRect,58,0) ;
	mTextAtomY = new TEditText(&theRect,"","0.0000") ;	

	OffsetRect(&theRect,58,0) ;
	mTextAtomZ = new TEditText(&theRect,"","0.0000") ;	

	OffsetRect(&theRect,58,0) ;
	mTextAtomDW = new TEditText(&theRect,"","0.5") ;	

	OffsetRect(&theRect,58,0) ;
	mTextAtomOCC = new TEditText(&theRect,"","1.0") ;	

	theRect = mTextAtomX->GetRect() ;
	OffsetRect(&theRect,0,25) ;
	mAddAtom =new TButton(&theRect,"Add") ;
	AddObject(mAddAtom) ;
	OffsetRect(&theRect,0,25) ;
	mDeleteAtom = new TButton(&theRect,"Delete") ;
	AddObject(mDeleteAtom) ;
	theRect = mTextAtomY->GetRect() ;
	OffsetRect(&theRect,0,50) ;
	mModifyAtom = new TButton(&theRect,"Modify") ;
	AddObject(mModifyAtom) ;
	
	// Set up the lists
	long i ;
	TAtom	*theAtom ;
	for(i=0; i < mCrystal->mNumBasisAtoms; i++) {
		theAtom = mCrystal->mBasisPtr+i ;
		mBasisDisplayList->AddRow(theAtom) ;
	}
	for(i=0; i < mCrystal->mNumUnitCellAtoms; i++) {
		theAtom = mCrystal->mAtomPtr+i ;
		mAtomDisplayList->AddRow(theAtom) ;
	}
	
	for(i=0; i < mCrystal->mNumSymOps; i++) {
		mSymmopDisplayList->AddRow(mCrystal->mSymopTxt[i]) ;
	}
	
	if (mCrystal->mNumBasisAtoms > 0) {
		mBasisDisplayList->SelectRow(1) ;
		theAtom = mCrystal->mBasisPtr ;
		mTextAtomSymbol->SetEditText(theAtom->name) ;
		sprintf(text,"%.4f",theAtom->xc) ;
		mTextAtomX->SetEditText(text) ;
		sprintf(text,"%.4f",theAtom->yc) ;
		mTextAtomY->SetEditText(text) ;
		sprintf(text,"%.4f",theAtom->zc) ;
		mTextAtomZ->SetEditText(text) ;
		sprintf(text,"%.2f",theAtom->dw) ;
		mTextAtomDW->SetEditText(text) ;
		sprintf(text,"%.2f",theAtom->occ) ;
		mTextAtomOCC->SetEditText(text) ;
	}
	
	AddTextEditObject(mTextAtomOCC) ;
	AddTextEditObject(mTextAtomDW) ;
	AddTextEditObject(mTextAtomZ) ;
	
	AddTextEditObject(mTextAtomY) ;
	AddTextEditObject(mTextAtomX) ;
	AddTextEditObject(mTextAtomSymbol) ;
	AddTextEditObject(mTextSpacegroup) ;	
	
	AddTextEditObject(mTextGamma) ;
	AddTextEditObject(mTextBeta) ;
	AddTextEditObject(mTextAlpha) ;
	AddTextEditObject(mTextC) ;
	AddTextEditObject(mTextB) ;
	AddTextEditObject(mTextA) ;
	
	mOKButton = new TButton(width-16-70,height-16-20,width-16,height-16,"OK") ;
	AddObject(mOKButton) ;
	theRect = mOKButton->GetRect() ;
	OffsetRect(&theRect,-(theRect.right-theRect.left+10),0) ;
	mCancelButton = new TButton(&theRect,"Cancel") ;
	AddObject(mCancelButton) ;
	SetDefaultButton(mOKButton) ;
}

TEditCrystalDialog::~TEditCrystalDialog() {
}

void	TEditCrystalDialog::RetrieveDialogItems() {
	// Assemble the crystal from what we have
	char	text[256] ;
	float	a,b,c,alpha,beta,gamma ;
		
	mTextA->GetEditText(text) ;
	a = atof(text) ;
	if (fabs(mCrystal->mA - a) > 0.00001) {
		mCrystal->mA = a ;
		mCrystalIsModified = true ;
	}
	mTextB->GetEditText(text) ;
	b = atof(text) ;
	if (fabs(mCrystal->mB - b) > 0.00001) {
		mCrystal->mB = b ;
		mCrystalIsModified = true ;
	}
	mTextC->GetEditText(text) ;
	c = atof(text) ;
	if (fabs(mCrystal->mC - c) > 0.00001) {
		mCrystal->mC = c ;
		mCrystalIsModified = true ;
	}
	mTextAlpha->GetEditText(text) ;
	alpha = atof(text) ;
	DegreeToRad(&alpha) ;
	if (fabs(mCrystal->mAlpha - alpha) > 0.00001) {
		mCrystal->mAlpha = alpha ;
		mCrystalIsModified = true ;
	}
	mTextBeta->GetEditText(text) ;
	beta = atof(text) ;
	DegreeToRad(&beta) ;
	if (fabs(mCrystal->mBeta - beta) > 0.00001) {
		mCrystal->mBeta = beta ;
		mCrystalIsModified = true ;
	}
	mTextGamma->GetEditText(text) ;
	gamma = atof(text) ;
	DegreeToRad(&gamma) ;
	if (fabs(mCrystal->mGamma - gamma) > 0.00001) {
		mCrystal->mGamma = gamma ;
		mCrystalIsModified = true ;
	}
	if (mCrystalIsModified) {
		*mSourceCrystal = *mCrystal ;
	}
}

void	TEditCrystalDialog::UpdateMotif() {
	int i ;
	ListCellT inCell ;
	inCell.col = 1 ;
	TAtom	*theAtom ;
	
	for(i=0; i < mBasisDisplayList->GetNumberOfListItems();i++) {
		inCell.row = i+1 ;
		theAtom = mCrystal->mBasisPtr+i ;
		mBasisDisplayList->FetchCellData(inCell,theAtom) ;		
	}
	mCrystal->SetAtomTypes() ;
	// Apply the symmetry operators
	mCrystal->ApplySymmops() ;
	// Update the atoms in the unit cell
	mAtomDisplayList->RemoveAllRows() ;
	for(i=0; i < mCrystal->mNumUnitCellAtoms; i++) {
		inCell.row = i+1 ;
		theAtom = mCrystal->mAtomPtr+i ;
		mAtomDisplayList->AddRow(theAtom) ;		
	}
	mAtomDisplayList->Refresh() ;
	
}

void	TEditCrystalDialog::UpdateSpaceGroup() {
	char	text[256] ;
	float	alpha,beta,gamma ;
	long	number,i,j,k ;
	TAtom	*theAtom ;
	
	mTextSpacegroup->GetEditText(text) ;
	number = atoi(text) ;

	mCrystal->mSpacegroup = number ;
	
	// Set the angle alpha based upon the spacegroup
	if (!(number <= 2)  && !(number >= 147 && 
											number <= 167 
											&& mCrystal->mSecondSetting)) {
		
		alpha = 90. ;
		sprintf(text,"%.2f",alpha) ;
		mTextAlpha->SetEditText(text) ;
	}
	// Set the angle beta based upon the spacegroup
	if (number >= 16  ) { // Not Monoclinic 
		beta = alpha ;
		sprintf(text,"%.2f",beta) ;
		mTextBeta->SetEditText(text) ;
	}
	// Set the angle gamma based upon the spacegroup
	// Cubic,Orthorombic or Tetragonal spacegroup 
	if (number >= 195 || (number > 15 && number < 143)) {
		gamma = 90. ;
		sprintf(text,"%.2f",gamma) ;
		mTextGamma->SetEditText(text) ;
	}
	else if (number > 167 && number < 195) {
		gamma = 120. ;
		mCrystal->mHexagonalIndices = true ;
		sprintf(text,"%.2f",gamma) ;
		mTextGamma->SetEditText(text) ;
	}
	else if (number <= 167 && number >= 143) {
		if (!mCrystal->mSecondSetting) {
			gamma = 120. ;
			mCrystal->mHexagonalIndices = true ;
			sprintf(text,"%.2f",gamma) ;
			mTextGamma->SetEditText(text) ;
		}
		else {
			gamma = beta ;
			sprintf(text,"%.2f",gamma) ;
			mTextGamma->SetEditText(text) ;
		}
	}
	if (number >= 75) {
		mTextA->GetEditText(text) ;
		mTextB->SetEditText(text) ;
	}
	if (number >= 195 ) {
		mTextA->GetEditText(text) ;
		mTextC->SetEditText(text) ;
	}
	
	assm_spcgrp(number) ;
	mCrystal->mNumSymOps = gNumSymops ;
	
	
	for(i=0; i < gNumSymops; i++) {
		for(j=0; j < 4; j++) {
			for(k=0; k < 4; k++) {
				mCrystal->mSymOps[i].m[j][k] = symop[i].matr[j][k] ;
			}
		}
		strcpy(mCrystal->mSymopTxt[i],txtptr_symop[i]) ;
	}
	mCrystal->ApplySymmops() ;

	mSymmopDisplayList->RemoveAllRows() ;
	for(i=0; i < mCrystal->mNumSymOps; i++) {
		mSymmopDisplayList->AddRow(mCrystal->mSymopTxt[i]) ;
	}

	mAtomDisplayList->RemoveAllRows() ;
	for(i=0; i < mCrystal->mNumUnitCellAtoms; i++) {
		theAtom = mCrystal->mAtomPtr+i ;
		mAtomDisplayList->AddRow(theAtom) ;
	}
	
}

void	TEditCrystalDialog::DoKeyDownInObject(TObject *obj) {
	char	text[256] ;
	short	planeGroup ;
	if (obj == mTextA) {
		mTextA->GetEditText(text) ;
		if (mCrystal->mSpacegroup >= 75) {
			mTextB->SetEditText(text) ;
		}
		if (mCrystal->mSpacegroup >= 195 ) {
			mTextC->SetEditText(text) ;
		}
		else if (mCrystal->mSpacegroup <= 167 && 
			mCrystal->mSpacegroup >= 143) {	// Trigonal with Rhombohedral axes
				if (mCrystal->mSecondSetting ) { // Rhombohedral 
					mTextC->SetEditText(text) ;
				}
		}
	}
	else if (obj == mTextB) {
		mTextB->GetEditText(text) ;
		if (mCrystal->mSpacegroup >= 75) {
			mTextA->SetEditText(text) ;
		}
		if (mCrystal->mSpacegroup >= 195) {
			mTextC->SetEditText(text) ;
		}
		else if (mCrystal->mSpacegroup <= 167 && mCrystal->mSpacegroup >= 143) {	
				// Trigonal with Rhombohedral axes 
				if (mCrystal->mSecondSetting) {  // Rhombohedral 
					mTextC->SetEditText(text) ;
				}
		}
	}
	else if (obj == mTextC) {
		mTextC->GetEditText(text) ;
		if (mCrystal->mSpacegroup >= 195) {
			mTextA->SetEditText(text) ;
			mTextB->SetEditText(text) ;
		}
		else if (mCrystal->mSpacegroup <= 167 && mCrystal->mSpacegroup >= 143) {
			if (mCrystal->mSecondSetting ) { // Rhombohedral
				mTextA->SetEditText(text) ;
				mTextB->SetEditText(text) ;
			}
		}
	}
}

void	TEditCrystalDialog::DoMouseDownInObject(TObject *obj,Point *pt) {
	char	text[256] ;
	TAtom	theAtom ;
	short	theSelectedRow ;
	ListCellT	theCell ;
	
	if (pt != nil)
		if (!obj->DoMouseDown(pt)) return ;
	if (obj == mOKButton) {
		mDone = true ;
		mDialogReturnCode = kDialogOK ;
	}
	else if (obj == mCancelButton) {
		mDone = true ;
		mDialogReturnCode = kDialogCancel ;
	}
	else if (obj == mApplySpacegroup) {
		// Apply the symmetry operators to get the atoms in the unit cell
		mCrystalIsModified = true ;
		UpdateSpaceGroup() ;
	}
	else if (obj == mAddAtom) {
		// Add a row based on the information in the atom fields
		mTextAtomSymbol->GetEditText(text) ;
		theAtom.name = chemsymb[atomno(text)-1] ;
		theAtom.no = atomno(text) ;
		mTextAtomX->GetEditText(text) ;
		theAtom.xc = atof(text) ;
		mTextAtomY->GetEditText(text) ;
		theAtom.yc = atof(text) ;
		mTextAtomZ->GetEditText(text) ;
		theAtom.zc = atof(text) ;
		mTextAtomDW->GetEditText(text) ;
		theAtom.dw = atof(text) ;
		mTextAtomOCC->GetEditText(text) ;
		theAtom.occ = atof(text) ;
		mBasisDisplayList->AddRow(&theAtom) ;
		mCrystal->mNumBasisAtoms++ ;
		delete[] mCrystal->mBasisPtr ;
		mCrystal->mBasisPtr = new TAtom[mCrystal->mNumBasisAtoms] ;
		delete[] mCrystal->mAtomPtr ;
		mCrystal->mAtomPtr = new TAtom[mCrystal->mNumBasisAtoms*200] ;
		// Collect all the atoms and apply the symmetry operators to get the atoms in the unit cell
		mCrystalIsModified = true ;
		UpdateMotif() ;
	}
	else if (obj == mDeleteAtom) {
		// Delete selected row
		theSelectedRow = mBasisDisplayList->GetSelectedRow() ;
		mBasisDisplayList->RemoveRow(theSelectedRow) ;
		if (theSelectedRow <= mBasisDisplayList->GetNumberOfListItems()) {
			theCell.col = 1 ;
			theCell.row = theSelectedRow ;
			mBasisDisplayList->ClickCell(theCell) ;
		}
		// Collect all the atoms and apply the symmetry operators to get the atoms in the unit cell
		mCrystal->mNumBasisAtoms-- ;
		delete[] mCrystal->mBasisPtr ;
		mCrystal->mBasisPtr = new TAtom[mCrystal->mNumBasisAtoms] ;
		delete[] mCrystal->mAtomPtr ;
		mCrystal->mAtomPtr = new TAtom[mCrystal->mNumBasisAtoms*200] ;
		mCrystalIsModified = true ;
		UpdateMotif() ;
	}
	else if (obj == mModifyAtom) {
		// Modify a row based on the information in the atom fields
		// Add a row based on the information in the atom fields
		theSelectedRow = mBasisDisplayList->GetSelectedRow() ;
		if (theSelectedRow <= mBasisDisplayList->GetNumberOfListItems()) {
			theCell.col = 1 ;
			theCell.row = theSelectedRow ;
			mTextAtomSymbol->GetEditText(text) ;
			theAtom.name = chemsymb[atomno(text)-1] ;
			theAtom.no = atomno(text) ;
			mTextAtomX->GetEditText(text) ;
			theAtom.xc = atof(text) ;
			mTextAtomY->GetEditText(text) ;
			theAtom.yc = atof(text) ;
			mTextAtomZ->GetEditText(text) ;
			theAtom.zc = atof(text) ;
			mTextAtomDW->GetEditText(text) ;
			theAtom.dw = atof(text) ;
			mTextAtomOCC->GetEditText(text) ;
			theAtom.occ = atof(text) ;
			mBasisDisplayList->SetCellData(theCell,&theAtom) ;
			mBasisDisplayList->Refresh() ;
			// Collect all the atoms and apply the symmetry operators to get the atoms in the unit cell
			mCrystalIsModified = true ;
			UpdateMotif() ;
		}
	}

}

void	TEditCrystalDialog::ShowAtom(TAtom &theAtom) {
	char	text[256] ;
	mTextAtomSymbol->SetEditText(theAtom.name) ;
	
	sprintf(text,"%.4f",theAtom.xc) ;
	mTextAtomX->SetEditText(text) ;
	
	sprintf(text,"%.4f",theAtom.yc) ;
	mTextAtomY->SetEditText(text) ;
	
	sprintf(text,"%.4f",theAtom.zc) ;
	mTextAtomZ->SetEditText(text) ;
	
	sprintf(text,"%.4f",theAtom.dw) ;
	mTextAtomDW->SetEditText(text) ;
	
	sprintf(text,"%.4f",theAtom.occ) ;
	mTextAtomOCC->SetEditText(text) ;
	
}

TMicroscope::TMicroscope()  {
	
	strcpy(mName,"") ;

	mVoltage = 200. ;
	mCs = 1.0e+7*2.0 ;
	mConvergence = 1.0e-3*0.5 ;
	mSpreadOfFocus = 80. ;
	
	m2Astigmatism 		= 0. ;
	m2AstigmatismAngle 	= 0. ;
	m3Astigmatism 		= 0. ;
	m3AstigmatismAngle 	= 0. ;
	m5Astigmatism 		= 0. ;
	m5AstigmatismAngle 	= 0. ;
	mComa 				= 0. ;
	mComaAngle 			= 0. ;
	mAperture 			= 0.7 ;
	mVibrationX 		= 0. ;
	mVibrationY 		= 0. ;
	mVibrationAngle 	= 0. ;
	mApertureH 			= 0. ;
	mApertureK 			= 0. ;
	mOpticAxisH 		= 0. ;
	mOpticAxisK			= 0. ;
	
	mBeamTilt.a[0] = mOpticAxisH ;
	mBeamTilt.a[1] = mOpticAxisK ;
	mBeamTilt.a[2] = 0. ;
	mBeamTilt.a[3] = 0. ;

	mAperturePosition.a[0] = mApertureH ;
	mAperturePosition.a[1] = mApertureK ;
	mAperturePosition.a[2] = 0. ;
	mAperturePosition.a[3] = 0. ;
	
    mWavelength = 12.263652/sqrt((mVoltage+9.785025e-4*mVoltage*mVoltage)*1000.) ;
    mSigma = 2.08886e-2*mWavelength*sqrt(1.0+5.8886579e-4/(mWavelength*mWavelength)) ;
    
	mStartFocus = -600. ;
	mEndFocus =  -600. ;
	mDeltaFocus = 0. ;

	mMinIntensityInLens = 8 ;

}

TMicroscope::~TMicroscope() {
}

TMicroscope::TMicroscope(TMicroscope *microscope) {
    
	strcpy(mName,microscope->mName) ;
	mVoltage = microscope->mVoltage ;
	mCs = microscope->mCs ;
	mConvergence = microscope->mConvergence ;
	mSpreadOfFocus = microscope->mSpreadOfFocus ;
	
	m2Astigmatism 		=  microscope->m2Astigmatism ;
	m2AstigmatismAngle 	=  microscope->m2AstigmatismAngle ;
	m3Astigmatism 		=  microscope->m3Astigmatism ;
	m3AstigmatismAngle 	=  microscope->m3AstigmatismAngle ;
	m5Astigmatism 		=  microscope->m5Astigmatism ;
	m5AstigmatismAngle 	=  microscope->m5AstigmatismAngle ;
	mComa 				=  microscope->mComa ;
	mComaAngle 			=  microscope->mComaAngle ;
	mAperture 			=  microscope->mAperture ;
	mVibrationX 		=  microscope->mVibrationX ;
	mVibrationY 		=  microscope->mVibrationY ;
	mVibrationAngle 	=  microscope->mVibrationAngle ;
	mApertureH 			=  microscope->mApertureH ;
	mApertureK 			=  microscope->mApertureK ;
	mOpticAxisH 		=  microscope->mOpticAxisH ;
	mOpticAxisK			=  microscope->mOpticAxisK ;
	

	mBeamTilt =  microscope->mBeamTilt ;

	mAperturePosition = microscope->mAperturePosition ;
	
    mWavelength = microscope->mWavelength ;
    mSigma =  microscope->mSigma ;
    
	mStartFocus = microscope->mStartFocus ;
	mEndFocus =  microscope->mEndFocus ;
	mDeltaFocus = microscope->mDeltaFocus ;

	mMinIntensityInLens = microscope->mMinIntensityInLens ;

}

void TMicroscope:: Update() {
    float   electronWavelength ;
    
    electronWavelength = 1.2263652e-7/sqrt((mVoltage+9.785025e-4*mVoltage*mVoltage)*1000.) ;    
	mWavelength = 1.0e+8*electronWavelength ;	
	mSigma = 2.08886e+6*electronWavelength*sqrt(1.0+5.886579e-20/(electronWavelength*electronWavelength)) ;
}

void	TMicroscope::SetVoltage(float voltage) {
	mVoltage = voltage ;
    mWavelength = 12.263652/sqrt((mVoltage+9.785025e-4*mVoltage*mVoltage)*1000.) ;
    mSigma = 2.08886e-2*mWavelength*sqrt(1.0+5.8886579e-4/(mWavelength*mWavelength)) ;
}
